/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2002 Kelly Grizzle (TJDO)
2003 Erik Bengtson - refactored the persistent id generator System property
2003 Erik Bengtson - added PMFConfiguration, PMFContext
2003 Andy Jefferson - commented, and added ApplicationId to supported options
2003 Andy Jefferson - introduction of localiser
2004 Andy Jefferson - update to getProperties() method
2004 Andy Jefferson - added LifecycleListener
2004 Andy Jefferson - added Level 2 Cache
2005 Marco Schulze - implemented copying the lifecycle listeners in j2ee environment
    ...
**********************************************************************/
package org.jpox;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Set;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.datastore.DataStoreCache;
import javax.jdo.datastore.Sequence;
import javax.jdo.listener.InstanceLifecycleListener;
import javax.jdo.spi.JDOPermission;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.jpox.cache.Level2Cache;
import org.jpox.cache.NullLevel2Cache;
import org.jpox.datasource.DriverManagerDataSource;
import org.jpox.exceptions.ConnectionFactoryNotFoundException;
import org.jpox.exceptions.UnsupportedConnectionFactoryException;
import org.jpox.store.rdbms.JDBCUtils;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * A factory used to obtain {@link javax.jdo.PersistenceManager} instances.
 * The factory is configurable up to a point when it is frozen. Thereafter
 * nothing can be changed.
 *
 * @version $Revision: 1.94 $
 */
public abstract class AbstractPersistenceManagerFactory extends PMFConfiguration
{
    private static final String VERSION_NUMBER_PROPERTY = "VersionNumber";
    private static final String VENDOR_NAME_PROPERTY = "VendorName";
    private static final String VENDOR_NAME="JPOX";

    /** Version of JPOX being used. Read in at startup from properties. */
    private static String jpoxVersion=null;

    /** Vendor of this version of JPOX. */
    private static String jpoxVendor=null;

    /** Transactional data source to use. Possible to have more than one datasource for failover */
    private transient DataSource tds[] = null;

    /** Non-transactional data source (or just the transactional variant). Possible to have more than one datasource for failover */
    private transient DataSource ntds[] = null;

    /** The cache of PM's in use */
    private Set pmCache = new HashSet();
   
    /** The context that this PMF uses. */
    protected PMFContext pmfContext;

    /** Level 2 Cache */
    protected Level2Cache cache;

    /** Lifecycle Listeners */
    protected List lifecycleListeners;

    /** Whether the PMF is closed */
    private boolean closed;

    /** Map of user-defined sequences keyed by the factory class name. */
    private Map sequenceByFactoryClass;

    /**
     * Constructor for the PMF.
     */
    public AbstractPersistenceManagerFactory()
    {
        super();
        pmfContext = new PMFContext(this);

        if (!ClassUtils.hasJDO2JarInClasspath())
        {
            throw new JDOFatalUserException(LOCALISER.msg("PMF.JDO2Inaccessible"));
        }
    }

    /**
     * Accessor to the ImplementationCreator
     * @return The ImplementationCreator
     */
    public ImplementationCreator getImplementationCreator()
    {
        return getPMFContext().getImplementationCreator();
    }

    /**
     * Asserts that's the PMF is open.
     * @throws JDOUserException if it is already closed
     */
    protected void assertIsOpen()
    {
        if (isClosed())
        {
            // Comply with Section 11.4 of the JDO2 spec (throw JDOUserException if already closed)
            throw new JDOUserException(LOCALISER.msg("PMF.PMFIsClosed"));
        }
    }

    /**
     * Looks up a DataSource object in JNDI.
     *
     * @param name The JNDI name of the DataSource.
     * @return  The DataSource object.
     * @exception ConnectionFactoryNotFoundException If a JNDI lookup failure occurs.
     * @exception UnsupportedConnectionFactoryException If the object is not a javax.sql.DataSource.
     */
    private DataSource lookupDataSource(String name)
    {
        Object obj;

        try
        {
            obj = new InitialContext().lookup(name);
        }
        catch (NamingException e)
        {
            throw new ConnectionFactoryNotFoundException(name, e);
        }

        if (!(obj instanceof DataSource))
        {
            throw new UnsupportedConnectionFactoryException(obj);
        }

        return (DataSource)obj;
    }

    /**
     * Freezes the current configuration.
     *
     * @exception UnsupportedConnectionFactoryException
     *      If the configured connection factory is not a javax.sql.DataSource.
     */
    protected void freezeConfiguration()
    {
        if (configurable)
        {
            // Set the transactional DataSource using connection factory (1)
            if (getConnectionFactory() != null)
            {
                if (!(getConnectionFactory() instanceof DataSource))
                {
                    throw new UnsupportedConnectionFactoryException(getConnectionFactory());
                }
                tds = new DataSource[1];
                tds[0] = (DataSource)getConnectionFactory();
            }
            else if (getConnectionFactoryName() != null)
            {
                String[] connectionFactoryNames = StringUtils.split(getConnectionFactoryName(),",");
                tds = new DataSource[connectionFactoryNames.length];
                for (int i=0; i<connectionFactoryNames.length; i++)
                {
                    tds[i] = lookupDataSource(connectionFactoryNames[i]);
                }
            }
            else
            {
                tds = new DataSource[1];
                String poolingType = getConnectionPoolingType();
                if (poolingType != null && !poolingType.equalsIgnoreCase("None"))
                {
                    // User has requested internal database connection pooling so check the registered plugins
                    try
                    {
                        // Create the DataSource to be used
                        Object dataSourceFactory = pmfContext.getPluginManager().createExecutableExtension("org.jpox.datasource", "name", poolingType, "class-name", null, null);
                        if (dataSourceFactory == null)
                        {
                            // User has specified a pool plugin that has not registered
                            throw new JDOFatalUserException(LOCALISER.msg("DataSource.ConnectionPoolPluginNotFound", poolingType));
                        }
                        // Create the JPOXDataSourceFactory
                        Class[] argTypes = new Class[] 
                                                     {String.class, String.class, String.class, String.class, String.class, ClassLoaderResolver.class};
                        Object[] args = new Object[]
                                                   {
                                getConnectionDriverName(), getConnectionURL(), getConnectionUserName(), 
                                getConnectionPassword(), getConnectionPoolingConfigurationFile(), pmfContext.getClassLoaderResolver(null) };


                        Method m = dataSourceFactory.getClass().getMethod("makePooledDataSource", argTypes);
                        tds[0] = (DataSource)m.invoke(dataSourceFactory, args);
                        setConnectionUserName(null); // Not needed when we go via the pool
                        setConnectionPassword(null); // Not needed when we go via the pool
                    }
                    catch (ClassNotFoundException cnfe)
                    {
                        throw new JDOFatalUserException(LOCALISER.msg("DataSource.ConnectionPoolPluginNotFound", poolingType),cnfe);
                    }
                    catch (Exception e)
                    {
                        if (e instanceof InvocationTargetException)
                        {
                            InvocationTargetException ite = (InvocationTargetException)e;
                            JPOXLogger.JDO.error(ite.getTargetException());
                            throw new JDOFatalInternalException(LOCALISER.msg("DataSource.ConnectionPoolPluginError", poolingType, 
                                ite.getTargetException().getMessage()), ite.getTargetException());
                        }
                        else
                        {
                            JPOXLogger.JDO.error(e);
                            throw new JDOFatalInternalException(LOCALISER.msg("DataSource.ConnectionPoolPluginError", poolingType, 
                                e.getMessage()),e);
                        }
                    }
                }
                else
                {
                    // Use our internal connection source
                    tds[0] = new DriverManagerDataSource(getConnectionDriverName(), getConnectionURL(), getPMFContext().getClassLoaderResolver(null));
                }
            }

            // Set non-transactional DataSource based on connection factory 2
            if (getConnectionFactory2() != null)
            {
                if (!(getConnectionFactory2() instanceof DataSource))
                {
                    throw new UnsupportedConnectionFactoryException(getConnectionFactory2());
                }

                ntds    = new DataSource[1];
                ntds[0] = (DataSource)getConnectionFactory2();
            }
            else if (getConnectionFactory2Name() != null)
            {
                String[] connectionFactory2Names = StringUtils.split(getConnectionFactory2Name(),",");
                ntds    = new DataSource[connectionFactory2Names.length];
                for( int i=0; i<connectionFactory2Names.length; i++)
                {
                    ntds[i] = lookupDataSource(connectionFactory2Names[i]);
                }
            }
            else
            {
                ntds = tds;
            }

            // Set up the Level 2 Cache
            initialiseLevel2Cache();

            configurable = false;

            // Log the configuration
            JPOXLogger.JDO.info(LOCALISER.msg("PMF.Version",getVendorName(),getVersionNumber()));
            JPOXLogger.JDO.info(LOCALISER.msg("PMF.Initialised", getConnectionURL(),getConnectionDriverName(), getConnectionUserName()));
            if (JPOXLogger.JDO.isDebugEnabled())
            {
                JPOXLogger.JDO.debug("================ PMF Configuration ================");
                JPOXLogger.JDO.debug("JDK : " + System.getProperty("java.version") + " on " + System.getProperty("os.name"));
                if (getPersistenceUnitName() != null)
                {
                    JPOXLogger.JDO.debug("Persistence-Unit : " + getPersistenceUnitName());
                }
                JPOXLogger.JDO.debug("JDO Standard Options : " + (getMultithreaded() ? "multithreaded" : "singlethreaded") +
                    (getRetainValues() ? ", retainValues" : "") +
                    (getRestoreValues() ? ", restoreValues" : "") +
                    (getNontransactionalRead() ? ", nontransactionalRead" : "") +
                    (getNontransactionalWrite() ? ", nontransactionalWrite" : "") +
                    (getIgnoreCache() ? ", ignoreCache" : ""));
                JPOXLogger.JDO.debug("Persistence Options :" +
                    (getPersistenceByReachabilityAtCommit() ? " reachability-at-commit" : "") +
                    (getDetachAllOnCommit() ? " detach-all-on-commit" : "") +
                    (getDetachOnClose() ? " detach-on-close" : "") +
                    " deletion-policy=" + getDeletionPolicy());

                StringBuffer autoCreateOptions = null;
                if (getAutoCreateTables() || getAutoCreateColumns() || getAutoCreateConstraints())
                {
                    autoCreateOptions = new StringBuffer();
                    boolean first = true;
                    if (getAutoCreateTables())
                    {
                        if (!first)
                        {
                            autoCreateOptions.append(",");
                        }
                        autoCreateOptions.append("Tables");
                        first = false;
                    }
                    if (getAutoCreateColumns())
                    {
                        if (!first)
                        {
                            autoCreateOptions.append(",");
                        }
                        autoCreateOptions.append("Columns");
                        first = false;
                    }
                    if (getAutoCreateConstraints())
                    {
                        if (!first)
                        {
                            autoCreateOptions.append(",");
                        }
                        autoCreateOptions.append("Constraints");
                        first = false;
                    }
                }
                
                StringBuffer validateOptions = null;
                if (getValidateTables() || getValidateColumns() || getValidateConstraints())
                {
                    validateOptions = new StringBuffer();
                    boolean first = true;
                    if (getValidateTables())
                    {
                        validateOptions.append("Tables");
                        first = false;
                    }
                    if (getValidateColumns())
                    {
                        if (!first)
                        {
                            validateOptions.append(",");
                        }
                        validateOptions.append("Columns");
                        first = false;
                    }
                    if (getValidateConstraints())
                    {
                        if (!first)
                        {
                            validateOptions.append(",");
                        }
                        validateOptions.append("Constraints");
                        first = false;
                    }
                }
                JPOXLogger.JDO.debug("Datastore : " +
                    "AutoCreate(" + (autoCreateOptions != null ? autoCreateOptions.toString() : "None") + ")" +
                    ", Validate(" + (validateOptions != null ? validateOptions.toString() : "None") + ")" +
                    (getReadOnlyDatastore() ? ", READONLY" : "") +
                    (getFixedDatastore() ? ", FIXED" : "") +
                    (getUseUpdateLock() ? ", useUpdateLock" : "") +
                    (getCheckExistTablesOrViews() ? ", CheckTableViewExistence" : "") +
                    ", foreignKeyCreateMode=" + getForeignKeyCreateMode());
                JPOXLogger.JDO.debug("Datastore Identifiers :" +
                    " Factory=\"" + getIdentifierFactory() + "\"" +
                    (getCatalog() != null ? (" catalog=" + getCatalog()) : "") + // TODO These are RDBMS-specific so need moving somewhere
                    (getSchema() != null ? (" schema=" + getSchema()) : "")); // TODO These are RDBMS-specific so need moving somewhere
                JPOXLogger.JDO.debug("MetaData : XML Validation=" + getMetaDataValidate() +
                    ", File Suffixes (JDO = " + getJdoMetaDataFileExtension() +
                    ", ORM = " + getOrmMetaDataFileExtension() +
                    ", JDOQUERY = " + getJdoqueryMetaDataFileExtension() + ")" +
                    ", ORM Mapping name=" + (getMapping() != null ? getMapping() : "(none)"));
                if (getDatastoreAdapterClassName() != null)
                {
                    JPOXLogger.JDO.debug("Datastore Adapter : class " + getDatastoreAdapterClassName());
                }
                JPOXLogger.JDO.debug("Datastore Connection Pooling : " + getConnectionPoolingType());
                JPOXLogger.JDO.debug("Types Control : stringDefaultLength=" + getStringDefaultLength() + ", dateTimezone=" + getDateTimezone());
                JPOXLogger.JDO.debug("Initialisation : mechanism=" + getAutoStartMechanism() + ", mode=" + getAutoStartMechanismMode() +
                    ((getAutoStartClassNames() != null) ? (", classes=" + getAutoStartClassNames()) : ""));
                JPOXLogger.JDO.debug("Transactions : type=" + getTransactionType() +
                    ", mode=" + (getOptimistic() ? "optimistic" : "pessimistic") +
                    ", isolation=" + JDBCUtils.getNameForJDBCIsolationLevel(getTransactionIsolation()) +
                    "  POID - isolation=" + JDBCUtils.getNameForJDBCIsolationLevel(getPoidTransactionIsolationLevel()) +
                    ", connection=" + (getPoidTransactionAttribute().equalsIgnoreCase("New") ? "New" : "PM"));
                JPOXLogger.JDO.debug("ClassLoading : " + getClassLoaderResolverName());
                JPOXLogger.JDO.debug("Cache : Level1 (" + getJdoCacheLevel1Type() + ")" +
                    (getJdoCacheLevel2() ? (", Level2 (" + getJdoCacheLevel2Type() + ")") : "") +
                    (getJdoCacheCollections() ? ", Collections/Maps " : ""));
                JPOXLogger.JDO.debug("Queries : fetchSize=" + getQueryFetchSize() + ", fetchDirection=" + getQueryFetchDirection() + 
                    ", resultSetType=" + getQueryResultSetType() + ", resultSetConcurrency=" + getQueryResultSetConcurrency());
                JPOXLogger.JDO.debug("===================================================");
            }
        }
    }

    /**
     * Method to initialise the L2 cache.
     */
    protected void initialiseLevel2Cache()
    {
        if (getJdoCacheLevel2())
        {
            // Find the L2 cache class name from its plugin name
            String level2Type = getJdoCacheLevel2Type();
            String level2ClassName = pmfContext.getPluginManager().getAttributeValueForExtension("org.jpox.cache_level2", 
                "name", level2Type, "class-name");
            if (level2ClassName == null)
            {
                // Plugin of this name not found
                throw new JDOFatalUserException(LOCALISER.msg("Cache.Level2.PluginNotFound", level2Type));
            }

            try
            {
                // Create an instance of the L2 Cache
                Properties cacheProps = new Properties();
                if (getJdoCacheLevel2CacheName() != null)
                {
                    cacheProps.put("CacheName", getJdoCacheLevel2CacheName());
                }
                if (getJdoCacheLevel2ConfigurationFile() != null)
                {
                    cacheProps.put("ConfigurationFile", getJdoCacheLevel2ConfigurationFile());
                }

                Class level2CacheClass = Class.forName(level2ClassName);
                Class[] ctrArgsClasses = new Class[] {Properties.class};
                Object[] ctrArgs = new Object[] {cacheProps};

                Constructor ctr = level2CacheClass.getConstructor(ctrArgsClasses);
                cache = (Level2Cache)ctr.newInstance(ctrArgs);
                JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.Initialised", level2ClassName));
            }
            catch (Exception e)
            {
                // Class name for this L2 cache plugin is not found!
                throw new JDOFatalUserException(LOCALISER.msg("Cache.Level2.PluginClassNotFound", level2Type, level2ClassName),e);
            }
        }
        else
        {
            cache = new NullLevel2Cache();
        }
    }

    /**
     * Returns the data source to be used for transactional access.
     *
     * @return  The data source to be used for transactional access. Possible to have more than one datasource for failover
     */
    public synchronized DataSource[] getTransactionalDataSource()
    {
        freezeConfiguration();

        return tds;
    }

    /**
     * Returns the data source to be used for non-transactional access.
     *
     * @return  The data source to be used for non-transactional access. Possible to have more than one datasource for failover
     */
    public synchronized DataSource[] getNontransactionalDataSource()
    {
        freezeConfiguration();

        return ntds;
    }

    /**
     * Return non-configurable properties of this PersistenceManagerFactory.
     * Properties with keys VendorName and VersionNumber are required.  Other
     * keys are optional.
     * @return the non-configurable properties of this
     * PersistenceManagerFactory.
     */
    public Properties getProperties()
    {
        Properties props = new Properties();

        props.setProperty(VENDOR_NAME_PROPERTY,getVendorName());
        props.setProperty(VERSION_NUMBER_PROPERTY,getVersionNumber());

        return props;
    }

    /**
     * Utility to get the version of JPOX.
     * @return Version number for JPOX.
     **/
    public static String getVersionNumber()
    {
        if (jpoxVersion != null)
        {
            return jpoxVersion;
        }
        
        String version="Unknown";
        try
        {
            ResourceBundle bundle=ResourceBundle.getBundle("org.jpox.JPOXVersion");
            try
            {
                version = bundle.getString("jpox.version");
            }
            catch (Exception e1)
            {
            }
        }
        catch (Exception e)
        {
        }
        
        return jpoxVersion = version;
    }

    /**
     * Utility to get the vendor of JPOX.
     * @return Vendor name for JPOX.
     **/
    public static String getVendorName()
    {
        if (jpoxVendor != null)
        {
            return jpoxVendor;
        }
        
        String vendor=VENDOR_NAME;
        try
        {
            ResourceBundle bundle=ResourceBundle.getBundle("org.jpox.JPOXVersion");
            try
            {
                vendor = bundle.getString("jpox.vendor");
            }
            catch (Exception e1)
            {
            }
        }
        catch (Exception e)
        {
        }

        return jpoxVendor = vendor;
    }

    /**
     * The application can determine from the results of this method which
     * optional features, and which query languages are supported by the JDO
     * implementation. Se esection 11.6 of the JDO 2 specification.
     * @return A Collection of String representing the supported options.
     */
    public Collection supportedOptions()
    {
        return Collections.unmodifiableList(Arrays.asList(OPTION_ARRAY));
    }

    /**
     * The JDO spec optional features that JPOX supports.
     * See JDO 2.0 spec section 11.6 for the full list of possibilities.
     **/
    private static final String[] OPTION_ARRAY =
    {
        "javax.jdo.option.TransientTransactional",
        // "javax.jdo.option.NontransactionalWrite",
        "javax.jdo.option.NontransactionalRead",
        "javax.jdo.option.RetainValues",
        "javax.jdo.option.Optimistic",
        "javax.jdo.option.ApplicationIdentity",
        "javax.jdo.option.DatastoreIdentity",
        // "javax.jdo.option.NonDurableIdentity",
        "javax.jdo.option.ArrayList",
        "javax.jdo.option.LinkedList",
        "javax.jdo.option.TreeSet",
        "javax.jdo.option.TreeMap",
        "javax.jdo.option.Vector",
        "javax.jdo.option.List",
        "javax.jdo.option.Stack", // Not a listed JDO2 feature
        "javax.jdo.option.Map", // Not a listed JDO2 feature
        "javax.jdo.option.HashMap", // Not a listed JDO2 feature
        "javax.jdo.option.Hashtable", // Not a listed JDO2 feature
        "javax.jdo.option.SortedSet", // Not a listed JDO2 feature
        "javax.jdo.option.SortedMap", // Not a listed JDO2 feature
        "javax.jdo.option.Array",
        "javax.jdo.option.NullCollection",
        // "javax.jdo.option.ChangeApplicationIdentity",
        "javax.jdo.option.BinaryCompatibility",
        "javax.jdo.option.GetDataStoreConnection",
        "javax.jdo.option.GetJDBCConnection",
        "javax.jdo.query.SQL",
        "javax.jdo.option.UnconstrainedQueryVariables",
        "javax.jdo.option.version.DateTime",
        "javax.jdo.option.PreDirtyEvent",
        "javax.jdo.option.mapping.HeterogeneousObjectType",
        "javax.jdo.option.mapping.HeterogeneousInterfaceType",
        "javax.jdo.option.mapping.JoinedTablePerClass",
        "javax.jdo.option.mapping.JoinedTablePerConcreteClass",
        "javax.jdo.option.mapping.NonJoinedTablePerConcreteClass",
        // "javax.jdo.option.mapping.RelationSubclassTable", // Not yet supported for multiple subclasses
        "javax.jdo.query.JDOQL", // Not a listed optional feature
        "javax.jdo.query.JPOXSQL"
    };

    /**
     * Return the Set of PersistenceManagers actually in cache
     * @return Set this contains instances of PersistenceManagers in the cache
     */
    protected Set getPmCache()
    {
        return pmCache;
    }

    /**
     * Remove a PersistenceManager from the cache
     * Only the PersistenceManager is allowed to call this method
     * @param pm  the PersistenceManager to be removed from cache
     */
    public void releasePersistenceManager(PersistenceManager pm)
    {
        getPmCache().remove(pm);
    }

    /**
     * Close this PersistenceManagerFactory. Check for
     * JDOPermission("closePersistenceManagerFactory") and if not authorized,
     * throw SecurityException.
     * <P>If the authorization check succeeds, check to see that all
     * PersistenceManager instances obtained from this PersistenceManagerFactory
     * have no active transactions. If any PersistenceManager instances have
     * an active transaction, throw a JDOUserException, with one nested
     * JDOUserException for each PersistenceManager with an active Transaction.
     * <P>If there are no active transactions, then close all PersistenceManager
     * instances obtained from this PersistenceManagerFactory, mark this
     * PersistenceManagerFactory as closed, disallow getPersistenceManager
     * methods, and allow all other get methods. If a set method or
     * getPersistenceManager method is called after close, then
     * JDOUserException is thrown.
     * 
     * @see javax.jdo.PersistenceManagerFactory#close()
     */
    public synchronized void close()
    {
        assertIsOpen();
        
        SecurityManager secmgr = System.getSecurityManager();
        if (secmgr != null)
        {
            // checkPermission will throw SecurityException if not authorized
            secmgr.checkPermission(JDOPermission.CLOSE_PERSISTENCE_MANAGER_FACTORY);
        }
        // iterate though the list of pms to release resources
        Iterator pms = new HashSet(getPmCache()).iterator();
        Set exceptions = new HashSet();
        while (pms.hasNext())
        {
            try
            {
                ((AbstractPersistenceManager)pms.next()).close();
            }
            catch(JDOUserException ex)
            {
                exceptions.add(ex);
            }
        }
        if( !exceptions.isEmpty() )
        {
            throw new JDOUserException(LOCALISER.msg("PMF.CloseError"),(Throwable[])exceptions.toArray(new Throwable[exceptions.size()]));
        }

        // Clear the Cache
        if (getJdoCacheLevel2() && cache != null)
        {
            cache.clear();
            JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.Cleared"));
        }

        // Clear the Sequence factory cache
        if (sequenceByFactoryClass != null)
        {
            sequenceByFactoryClass.clear();
        }
		pmfContext.close();
		pmfContext = null;
		ntds = null;
		tds = null;
        closed = true;
    }

    /**
     * Utility to return whether the PMF is closed or not.
     * @return Whether it is closed.
     */
    public boolean isClosed()
    {
        return closed;
    }

    /**
     * Gets the context for this PersistenceManagerFactory
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return pmfContext;
    }

    /**
     * Accessor for the PMFConfiguration.
     * @return Returns the PMF config.
     */
    public PMFConfiguration getPMFConfiguration()
    {
        return this;
    }

    /**
     * Accessor for the DataStore (level 2) Cache
     * @return The datastore cache
     * @since 1.1
     */
    public DataStoreCache getDataStoreCache()
    {
        freezeConfiguration();

        return cache;
    }

    // -------------------------------- Lifecycle Listeners -------------------------------

    /**
     * @return Returns either <tt>null</tt> or a <tt>List</tt> with instances of
     *		<tt>LifecycleListenerSpecification</tt>.
     */
    public List getLifecycleListenerSpecifications()
    {
    	return lifecycleListeners;
    }

    /**
     * Method to add lifecycle listeners for particular classes.
     * Adds the listener to all PMs already created.
     * @param listener The listener
     * @param classes The classes that the listener relates to
     * @since 1.1
     */
    public void addInstanceLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        if (listener == null)
        {
            return;
        }

        synchronized (pmCache)
        {
            // Add to any PMs - make sure that the PM Cache isnt changed while doing so
            Iterator pms = getPmCache().iterator();
            while (pms.hasNext())
            {
                AbstractPersistenceManager pm = (AbstractPersistenceManager)pms.next();
                pm.addInstanceLifecycleListener(listener, classes);
            }
        }

        // Register the lifecycle listener
        if (lifecycleListeners == null)
        {
            lifecycleListeners = new ArrayList(5);
        }
        lifecycleListeners.add(new LifecycleListenerForClass(listener, classes));
    }

    /**
     * Method to remove a lifecycle listener. Removes the listener from all PM's as well.
     * @param listener The Listener
     * @since 1.1
     */
    public void removeInstanceLifecycleListener(InstanceLifecycleListener listener)
    {
        if (listener == null || lifecycleListeners == null)
        {
            return;
        }

        synchronized (pmCache)
        {
            // Remove from any PMs - make sure that the PM Cache isnt changed while doing so
            Iterator pms = getPmCache().iterator();
            while (pms.hasNext())
            {
                AbstractPersistenceManager pm = (AbstractPersistenceManager)pms.next();
                pm.removeInstanceLifecycleListener(listener);
            }
        }

        // Remove from the PMF
        Iterator iter = lifecycleListeners.iterator();
        while (iter.hasNext())
        {
            LifecycleListenerForClass classListener = (LifecycleListenerForClass) iter.next();
            if (classListener.getListener() == listener)
            {
                iter.remove();
            }
        }
    }

    // --------------------------- Sequences ----------------------------------

    /**
     * Method to register a sequence for a particular factory class.
     * @param factoryClassName Name of the factory class
     * @param sequence The sequence
     */
    void addSequenceForFactoryClass(String factoryClassName, Sequence sequence)
    {
        if (sequenceByFactoryClass == null)
        {
            sequenceByFactoryClass = new HashMap();
        }

        sequenceByFactoryClass.put(factoryClassName, sequence);
    }

    /**
     * Accessor for the sequence for a factory class.
     * @param factoryClassName The name of the factory class
     * @return The sequence
     */
    Sequence getSequenceForFactoryClass(String factoryClassName)
    {
        if (sequenceByFactoryClass == null)
        {
            return null;
        }

        return (Sequence)sequenceByFactoryClass.get(factoryClassName);
    }
}