/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2004 Andy Jefferson - added description of addition process.
2004 Andy Jefferson - added constructor, and try-catch on initialisation
2004 Andy Jefferson - added various new properties
    ...
**********************************************************************/
package org.jpox;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TimeZone;

import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;

import org.jpox.plugin.PluginRegistry;
import org.jpox.store.AutoStartMechanism;
import org.jpox.store.rdbms.JDBCUtils;
import org.jpox.store.rdbms.SchemaAutoStarter;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * User configuration for the PMF.
 * All properties should have setter/getter, and should have a "PropertySetter"
 * to allow loading of the property from the PMF Properties.
 * @version $Revision: 1.99 $
 */
public abstract class PMFConfiguration
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    // ------------------------------ JDO standard properties ---------------------------------

    /** Datastore driver name property. */
    public static final String JDO_DATASTORE_DRIVERNAME_PROPERTY = "javax.jdo.option.ConnectionDriverName";

    /** Datastore URL property. */
    public static final String JDO_DATASTORE_URL_PROPERTY = "javax.jdo.option.ConnectionURL";

    /** Datastore connection username property. */
    public static final String JDO_DATASTORE_USERNAME_PROPERTY = "javax.jdo.option.ConnectionUserName";

    /** Datastore connection password property. */
    public static final String JDO_DATASTORE_PASSWORD_PROPERTY = "javax.jdo.option.ConnectionPassword";

    /** Datastore connection factory name property. */
    public static final String JDO_CONNECTION_FACTORY_NAME_PROPERTY = "javax.jdo.option.ConnectionFactoryName";

    /** Datastore connection factory 2 name property. */
    public static final String JDO_CONNECTION_FACTORY2_NAME_PROPERTY = "javax.jdo.option.ConnectionFactory2Name";

    /** Datastore connection factory property. */
    public static final String JDO_CONNECTION_FACTORY_PROPERTY = "javax.jdo.option.ConnectionFactory";

    /** Datastore connection factory 2 property. */
    public static final String JDO_CONNECTION_FACTORY2_PROPERTY = "javax.jdo.option.ConnectionFactory2";

    /** Ignore Cache property. */
    public static final String JDO_IGNORECACHE_PROPERTY = "javax.jdo.option.IgnoreCache";

    /** Optimistic transaction property. */
    public static final String JDO_OPTIMISTIC_PROPERTY = "javax.jdo.option.Optimistic";

    /** Multithreaded property. */
    public static final String JDO_MULTITHREADED_PROPERTY = "javax.jdo.option.Multithreaded";

    /** Retain values property. */
    public static final String JDO_RETAINVALUES_PROPERTY = "javax.jdo.option.RetainValues";

    /** Restore values property. */
    public static final String JDO_RESTOREVALUES_PROPERTY = "javax.jdo.option.RestoreValues";

    /** Non-transactional Read property. */
    public static final String JDO_NONTRANSACTIONAL_READ_PROPERTY = "javax.jdo.option.NontransactionalRead";

    /** Non-transactional Write property. */
    public static final String JDO_NONTRANSACTIONAL_WRITE_PROPERTY = "javax.jdo.option.NontransactionalWrite";

    /** Mapping suffix property. */
    public static final String JDO_MAPPING_PROPERTY = "javax.jdo.option.Mapping";

    /** Catalog name property. */
    public static final String JDO_MAPPING_CATALOG_PROPERTY = "javax.jdo.mapping.Catalog";

    /** Schema name property. */
    public static final String JDO_MAPPING_SCHEMA_PROPERTY = "javax.jdo.mapping.Schema";

    /** Detach-all-on-commit property */
    public static final String JDO_DETACHALLONCOMMIT_PROPERTY = "javax.jdo.option.DetachAllOnCommit";

    /** Transaction type property. */
    public static final String JDO_TRANSACTION_TYPE_PROPERTY = "javax.jdo.option.TransactionType";

    // ------------------------------ JPOX extension properties ---------------------------------

    /** Name of the "persistence-unit" (likely to become JDO property). */
    public static final String PERSISTENCE_UNIT_NAME = "org.jpox.persistenceUnitName";

    /** ClassLoaderResolver name property. */
    public static final String CLASS_LOADER_RESOLVER_NAME_PROPERTY = "org.jpox.classLoaderResolverName";

    /** Implementation Creator name property. */
    public static final String IMPLEMENTATION_CREATOR_NAME_PROPERTY = "org.jpox.implementationCreatorName";

    /** Detach-on-close property. */
    public static final String DETACH_ON_CLOSE_PROPERTY = "org.jpox.DetachOnClose";

    /** Property defining the classname of the datastore adapter to use. */
    public static final String DATASTORE_ADAPTER_CLASSNAME_PROPERTY = "org.jpox.datastoreAdapterClassName";

    /** Property for whether to check if the table/view exists. */
    public static final String CHECK_EXIST_TABLES_VIEWS_PROPERTY = "org.jpox.rdbms.CheckExistTablesOrViews";

    /** Property for the maximum number of statements that can be batched (0 = no batching) */
    public static final String STATEMENT_BATCH_LIMIT_PROPERTY = "org.jpox.rdbms.statementBatchLimit";

    /** Property for whether to validate the tables upon loading. */
    public static final String VALIDATE_TABLES_PROPERTY = "org.jpox.validateTables";

    /** Property for whether to validate the columns upon loading. */
    public static final String VALIDATE_COLUMNS_PROPERTY = "org.jpox.validateColumns";

    /** Property for whether the validate the constraints upon loading */
    public static final String VALIDATE_CONSTRAINTS_PROPERTY = "org.jpox.validateConstraints";

    /** Property for whether to warn only when receiving an error on auto creation. */
    public static final String AUTO_CREATE_WARN_ON_ERROR_PROPERTY = "org.jpox.autoCreateWarnOnError";

    /** Property for whether to auto create the schema upon loading. */
    public static final String AUTO_CREATE_SCHEMA_PROPERTY = "org.jpox.autoCreateSchema";

    /** Property for whether to auto create the tables upon loading. */
    public static final String AUTO_CREATE_TABLES_PROPERTY = "org.jpox.autoCreateTables";

    /** Property for whether to auto create the columns upon loading. */
    public static final String AUTO_CREATE_COLUMNS_PROPERTY = "org.jpox.autoCreateColumns";

    /** The system property that selects the default value for the AutoCreateConstraints setting. */
    public static final String AUTO_CREATE_CONSTRAINTS_PROPERTY = "org.jpox.autoCreateConstraints";

    /** Property for whether to validate the tables upon loading. */
    public static final String INITILIAZE_PRIMARY_KEY_COLUMN_INFO = "org.jpox.initializePrimaryKeyColumnInfo";

    /** Whether the datastore schema is read-only and so writes should not be performed. @deprecated */
    public static final String READ_ONLY_SCHEMA_PROPERTY = "org.jpox.readOnlySchema";

    /** Whether the datastore schema is read-only and so writes should not be performed. @deprecated */
    public static final String FIXED_SCHEMA_PROPERTY = "org.jpox.fixedSchema";

    /** Whether the datastore schema is read-only and so writes should not be performed. */
    public static final String READ_ONLY_DATASTORE_PROPERTY = "org.jpox.readOnlyDatastore";

    /** Whether the datastore schema is read-only and so writes should not be performed. */
    public static final String FIXED_DATASTORE_PROPERTY = "org.jpox.fixedDatastore";

    /** The system property that defines the mode of creation of foreign keys. */
    public static final String FOREIGN_KEY_CREATE_MODE_PROPERTY = "org.jpox.foreignKeyCreateMode";

    /** Auto-creation mode, including users MetaData specifications. */
    public static final String FOREIGN_KEY_AUTO_MODE="AUTO";

    /** Strict mode, only using users MetaData specifications. */
    public static final String FOREIGN_KEY_STRICT_MODE="STRICT";

    /** The system property that defines the process for deletions. */
    public static final String DELETION_POLICY_PROPERTY = "org.jpox.deletionPolicy";

    /** Property defining the transaction isolation mode to use. */
    public static final String TRANSACTION_ISOLATION_PROPERTY = "org.jpox.transactionIsolation";

    /** Property defining the transaction isolation to use when obtaining POIDs. */
    public static final String POID_TRANSACTION_ISOLATION_PROPERTY = "org.jpox.poid.transactionIsolation";

    /** Property defining how to obtain a datastore connection when generating POIDs */
    public static final String POID_TRANSACTION_ATTRIBUTE_PROPERTY = "org.jpox.poid.transactionAttribute";

    /** Property defining the auto-start mechanism to use (if any).  */
    public static final String AUTO_START_MECHANISM_PROPERTY = "org.jpox.autoStartMechanism";

    /** Property defining the mode of operation of any auto-start mechanism. */
    public static final String AUTO_START_MECHANISM_MODE_PROPERTY = "org.jpox.autoStartMechanismMode";

    /** Property defining the name of the file used when using the XML AutoStartMechanism. */
    public static final String AUTO_START_MECHANISM_XML_FILE_PROPERTY = "org.jpox.autoStartMechanismXmlFile";

    /** Property defining the names of the classes whose MetaData is to be loaded at startup. */
    public static final String AUTO_START_CLASS_NAMES_PROPERTY = "org.jpox.autoStartClassNames";

    /** Property defining how to pool connections. */
    public static final String CONNECTION_POOLING_TYPE_PROPERTY = "org.jpox.connectionPoolingType";

    /** Property defining the configuration file for connection pooling. */
    public static final String CONNECTION_POOLING_CONFIGURATION_FILE_PROPERTY = "org.jpox.connectionPoolingConfigurationFile";

    /** Property defining the type of Level 1 Cache to use. */
    public static final String CACHE_LEVEL_1_TYPE_PROPERTY = "org.jpox.cache.level1.type";

    /** Property defining whether to use a Level 2 Cache. */
    public static final String CACHE_LEVEL_2_PROPERTY = "org.jpox.cache.level2";

    /** Property defining the type of Level 2 Cache to use. */
    public static final String CACHE_LEVEL_2_TYPE_PROPERTY = "org.jpox.cache.level2.type";

    /** Property defining the symbolic name of the cache to use for Level 2. */
    public static final String CACHE_LEVEL_2_CACHE_NAME_PROPERTY = "org.jpox.cache.level2.cacheName";

    /** Property defining the configuration file of the cache to use for Level 2. */
    public static final String CACHE_LEVEL_2_CONFIGURATION_FILE_PROPERTY = "org.jpox.cache.level2.configurationFile";

    /** Property defining whether to cache collections. */
    public static final String CACHE_COLLECTIONS_PROPERTY = "org.jpox.cache.collections";

    /** Property defining whether to lazy load collections (when cached). */
    public static final String CACHE_COLLECTIONS_LAZY_PROPERTY = "org.jpox.cache.collections.lazy";

    /** Property defining the file extension for the JDO metadata files. */
    public static final String METADATA_JDO_FILE_EXTENSION_PROPERTY = "org.jpox.metadata.jdoFileExtension";

    /** Property defining the file extension for the ORM metadata files. */
    public static final String METADATA_ORM_FILE_EXTENSION_PROPERTY = "org.jpox.metadata.ormFileExtension";

    /** Property defining the file extension for the JDO Query metadata files.*/
    public static final String METADATA_JDOQUERY_FILE_EXTENSION_PROPERTY = "org.jpox.metadata.jdoqueryFileExtension";

    /** Property defining whether to validate the metadata files.*/
    public static final String METADATA_VALIDATE_PROPERTY = "org.jpox.metadata.validate";

    /** Property defining whether to use "SELECT ... FOR UPDATE" to fetch objects when using READ COMMITTED or lower. */
    public static final String USE_UPDATE_LOCK_PROPERTY = "org.jpox.useUpdateLock";

    /** Property defining the default strategy for inheritance, allowing change of behaviour from JDO2. */
    public static final String DEFAULT_INHERITANCE_STRATEGY_PROPERTY = "org.jpox.rdbms.defaultInheritanceStrategy";

    /** Property defining if datastore ops are delated til commit. In optimistic tx this is automatic. */
    public static final String DATASTORE_DELAY_OPERATIONS_UNTIL_COMMIT_PROPERTY = "org.jpox.delayDatastoreOperationsUntilCommit";

    /** Property for how many times to retry the addition of a class to the store management. */
    public static final String DATASTORE_CLASS_ADDITION_MAX_RETRIES_PROPERTY = "org.jpox.store.classAdditionMaxRetries";

    /** Property defining whether to add unique constraints to the element table for a map inverse field */
    public static final String ADD_UNIQUE_CONSTRAINT_MAP_INVERSE_PROPERTY = "org.jpox.rdbms.uniqueConstraints.mapInverse";

    /** Property defining whetehr to use the FetchPlan when executing the query. */
    public static final String QUERY_USE_FETCH_PLAN_PROPERTY = "org.jpox.query.useFetchPlan";

    /** Property defining the timeout (seconds) for any queries. */
    public static final String QUERY_TIMEOUT_PROPERTY = "org.jpox.query.timeout";

    /** Property defining the number of rows of a result set that should be fetched when more rows are needed. */
    public static final String QUERY_FETCH_SIZE_PROPERTY = "org.jpox.query.fetchSize";

    /** Property defining the direction in which a result set will be navigated. */
    public static final String QUERY_FETCH_DIRECTION_PROPERTY = "org.jpox.query.fetchDirection";

    /** Property defining the type of result set to create. */
    public static final String QUERY_RESULT_SET_TYPE_PROPERTY = "org.jpox.query.resultSetType";

    /** Property defining the concurrency of the result set */
    public static final String QUERY_RESULT_SET_CONCURRENCY_PROPERTY = "org.jpox.query.resultSetConcurrency";

    /** Property defining the whether to use inner joins only on queries. */
    public static final String QUERY_USE_INNER_JOINS_ONLY_PROPERTY = "org.jpox.jdoql.useInnerJoinsOnly";

    /** Property for allow/disallow all SQL statements. */
    public static final String QUERY_ALLOW_ALL_SQL_STATEMENTS = "org.jpox.query.allowAllSQLStaments";

    /** Property defining the default string length when storing a String type in the datastore and no length given. */
    public static final String STRING_DEFAULT_LENGTH_PROPERTY = "org.jpox.rdbms.stringDefaultLength";

    /** Property defining the timezone name to assume all dates are in in the datastore. */
    public static final String DATE_TIMEZONE_PROPERTY = "org.jpox.rdbms.dateTimezone";
    
    /** Property defining whether PBR is run at commit time (JDO default is true). */
    public static final String PERSISTENCE_BY_REACHABILITY_AT_COMMIT = "org.jpox.persistenceByReachabilityAtCommit";

    /** Property defining the maximum fetch depth to use by default. */
    public static final String MAX_FETCH_DEPTH_PROPERTY = "org.jpox.maxFetchDepth";

    /** Property defining the identifier factory to use. */
    public static final String IDENTIFIER_FACTORY_PROPERTY = "org.jpox.identifierFactory";

    /** Property defining the identifier case to be used. */
    public static final String IDENTIFIER_CASE_PROPERTY = "org.jpox.identifier.case";

    /** Property defining the word separator for use in identifiers (if supported by the factory). */
    public static final String IDENTIFIER_WORD_SEPARATOR_PROPERTY = "org.jpox.identifier.wordSeparator";

    /** Property defining the table prefix for use in identifiers (if supported by the factory). */
    public static final String IDENTIFIER_TABLE_PREFIX_PROPERTY = "org.jpox.identifier.tablePrefix";

    /** Property defining the table suffix for use in identifiers (if supported by the factory). */
    public static final String IDENTIFIER_TABLE_SUFFIX_PROPERTY = "org.jpox.identifier.tableSuffix";

    /** Property defining a file that contains properties that will be loaded by JPOX. */
    public static final String PROPERTIES_FILE = "org.jpox.propertiesFile";

    /** Plug-in registry class name */
    public static final String PLUGIN_REGISTRY_CLASS_NAME = "org.jpox.plugin.pluginRegistryClassName";

    /** Interface allowing setting properties on a PMF with String property values. */
    protected interface StringPropertySetter
    {
        /**
         * Set this property on the given PersistenceManagerFactoryImpl to the String value.
         * @param configuration The PMFConfiguration to set the value on.
         * @param value The String value to set.
         */
        void set(PMFConfiguration configuration, String value);
    }

    /** Interface allowing setting properties on a PMF with Object property values. */
    protected interface ObjectPropertySetter
    {
        /**
         * Set this property on the given PersistenceManagerFactoryImpl to the Object value.
         * @param configuration The PMFConfiguration to set the value on.
         * @param value The Object value to set.
         */
        void set(PMFConfiguration configuration, Object value);
    }

    /**
     * A Map mapping property name to the corresponding PropertySetter.
     */
    private final Map PROPERTY_SETTERS = initPropertySetters();

    /** options given when this PMF is constructed **/
    private transient Map options;
        
    private Calendar dateTimezoneCalendar = null;

    private TimeZone DATE_TIMEZONE;

    // JDO standard properties
    private String driverName = null;
    private String url = null;
    private String userName = null;
    private String password = null;
    private Object connectionFactory = null;
    private String connectionFactoryName = null;
    private Object connectionFactory2 = null;
    private String connectionFactory2Name = null;
    private boolean multithreaded = false;
    private boolean optimistic = false;
    private boolean retainValues = false;
    private boolean restoreValues = false;
    private boolean nontransactionalRead = false;
    private boolean nontransactionalWrite = false;
    private boolean ignoreCache = false;
    private boolean detachAllOnCommit = false;
    private String catalogName = null;
    private String schemaName = null;
    private String mapping = null;
    private String transactionType = null;

    private String persistenceUnitName = null;

    // Schema
    private String datastoreAdapterClassName = null;
    private boolean useUpdateLock = false;
    private boolean datastoreDelayOperationsEnabled = false;
    private int datastoreClassAdditionMaxRetries = 3;
    private boolean uniqueConstraintsMapInverse = true;
    private String defaultInheritanceStrategy = "JDO2";
    private String identifierFactory = "jpox";
    private String identifierCase = "UpperCase";
    private String identifierWordSeparator = null;
    private String identifierTablePrefix = null;
    private String identifierTableSuffix = null;

    // Query
    private boolean queryUseFetchPlan = true;
    private int queryTimeout = 0;
    private int queryFetchSize = 0;
    private String queryFetchDirection = "forward";
    private String queryResultSetType = "forward-only";
    private String queryResultSetConcurrency = "read-only";
    private boolean queryAllowAllSQLStatements = false;
    private boolean queryUseInnerJoinsOnly = false;

    // Types
    private final static int JDO2_STRING_DEFAULT_LENGTH = 256;
    private int stringDefaultLength = JDO2_STRING_DEFAULT_LENGTH;
    private String dateTimezone = "UTC";

    // Persistence control
    private boolean detachOnClose = false;
    private boolean persistenceByReachabilityAtCommit = true;
    private int maxFetchDepth = 1; // JDO2 default is 1
    private String classLoaderResolverName = "jdo";
    private String implementationCreatorName = null;

    // Schema
    private boolean checkExistTablesOrViews = true;
    private boolean validateTables = true;
    private boolean validateColumns = true;
    private boolean validateConstraints = true;
    private boolean autoCreateTables = false;
    private boolean autoCreateColumns = false;
    private boolean autoCreateConstraints = false;
    private boolean autoCreateWarnOnError = false;
    private boolean initializePrimaryKeyColumnInfo = false;
    private int statementBatchLimit = -1; // Unlimited

    private boolean readOnlyDatastore = false;
    private boolean fixedDatastore = false;
    private String foreignKeyCreateMode = FOREIGN_KEY_AUTO_MODE;
    private String deletionPolicy = "JDO2";

    // Autostart
    private String autoStartMechanism = SchemaAutoStarter.SCHEMA_TABLE;
    private String autoStartMechanismXmlFile = "jpoxAutoStart.xml";
    private String autoStartMechanismMode = AutoStartMechanism.MODE_QUIET;
    private String autoStartClassNames = null;

    private int isolationLevel = Connection.TRANSACTION_READ_COMMITTED;
    private int poidIsolationLevel = Connection.TRANSACTION_READ_COMMITTED;
    private String poidTransactionAttribute = "New";
    private String persistentIdGenerator = "org.jpox.poid.DefaultIdentifierGenerator";

    private String connectionPoolingType = "None";
    private String connectionPoolingConfigurationFile = null;

    private boolean jdoCacheCollections = true;
    private Boolean jdoCacheCollectionsLazy;
    private String jdoCacheLevel1Type = "weak";
    private boolean jdoCacheLevel2 = false; // L2 cache defaults to OFF
    private String jdoCacheLevel2Type = "default";
    private String jdoCacheLevel2CacheName = null;
    private String jdoCacheLevel2ConfigurationFile = null;

    // MetaData
    private String jdoMetaDataFileExtension = "jdo";
    private String ormMetaDataFileExtension = "orm";
    private String jdoqueryMetaDataFileExtension = "jdoquery";
    private boolean metadataValidate = true;

    // Properties
    private File propertiesFile = null;

    // Plug-in
    private String pluginRegistryClassName = null;
    protected transient boolean configurable=true;

    /**
     * Constructor.
     * Initialises all properties using the System property values. This behaviour
     * is not specified by the JDO spec(s) but is a nice to have capability allowing
     * users to define command line args to set up their (single) PMF.
     */
    public PMFConfiguration()
    {
        try
        {
            // Schema Management
            datastoreAdapterClassName = System.getProperty(DATASTORE_ADAPTER_CLASSNAME_PROPERTY);

            validateTables = new Boolean(System.getProperty(VALIDATE_TABLES_PROPERTY, "true")).booleanValue();
            validateColumns = new Boolean(System.getProperty(VALIDATE_COLUMNS_PROPERTY, "true")).booleanValue();
            validateConstraints = new Boolean(System.getProperty(VALIDATE_CONSTRAINTS_PROPERTY, "true")).booleanValue();
            initializePrimaryKeyColumnInfo = new Boolean(System.getProperty(INITILIAZE_PRIMARY_KEY_COLUMN_INFO, "false")).booleanValue();
            if (!validateTables)
            {
                // Can't validate columns if not validating the table!
                validateColumns = false;
            }

            autoCreateTables = new Boolean(System.getProperty(AUTO_CREATE_TABLES_PROPERTY, "false")).booleanValue();
            autoCreateColumns = new Boolean(System.getProperty(AUTO_CREATE_COLUMNS_PROPERTY, "false")).booleanValue();
            autoCreateConstraints = new Boolean(System.getProperty(AUTO_CREATE_CONSTRAINTS_PROPERTY, "false")).booleanValue();
            setAutoCreateSchema(new Boolean(System.getProperty(AUTO_CREATE_SCHEMA_PROPERTY, "false")).booleanValue());

            autoCreateWarnOnError = new Boolean(System.getProperty(AUTO_CREATE_WARN_ON_ERROR_PROPERTY, "false")).booleanValue();

            // TODO These first two are deprecated, replaced by the latter two, so remove them soon
            boolean val = new Boolean(System.getProperty(READ_ONLY_SCHEMA_PROPERTY)).booleanValue();
            if (val)
            {
                setReadOnlySchema(val);
            }
            val = new Boolean(System.getProperty(FIXED_SCHEMA_PROPERTY)).booleanValue();
            if (val)
            {
                setFixedSchema(val);
            }
            setReadOnlyDatastore(readOnlyDatastore = new Boolean(System.getProperty(READ_ONLY_DATASTORE_PROPERTY, "false")).booleanValue());
            setFixedDatastore(new Boolean(System.getProperty(FIXED_DATASTORE_PROPERTY, "false")).booleanValue());

            String datastoreClassAdditionMaxRetriesString = System.getProperty(DATASTORE_CLASS_ADDITION_MAX_RETRIES_PROPERTY);
            if (datastoreClassAdditionMaxRetriesString != null)
            {
                try
                {
                    datastoreClassAdditionMaxRetries = (new Integer(datastoreClassAdditionMaxRetriesString)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing and use default.
                }
            }

            mapping = System.getProperty(JDO_MAPPING_PROPERTY);
            catalogName = System.getProperty(JDO_MAPPING_CATALOG_PROPERTY);
            schemaName = System.getProperty(JDO_MAPPING_SCHEMA_PROPERTY);
            transactionType = System.getProperty(JDO_TRANSACTION_TYPE_PROPERTY);
            persistenceUnitName = System.getProperty(PERSISTENCE_UNIT_NAME);

            if (readOnlyDatastore || fixedDatastore)
            {
                if (autoCreateTables)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateTables"));
                    autoCreateTables = false;
                }
                if (autoCreateColumns)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateColumns"));
                    autoCreateColumns = false;
                }
                if (autoCreateConstraints)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateConstraints"));
                    autoCreateConstraints = false;
                }
            }
            String fkCreateMode = System.getProperty(FOREIGN_KEY_CREATE_MODE_PROPERTY);
            if (fkCreateMode != null && fkCreateMode.equals("" + FOREIGN_KEY_AUTO_MODE))
            {
                foreignKeyCreateMode = FOREIGN_KEY_AUTO_MODE;
            }
            else if (fkCreateMode != null && fkCreateMode.equals("" + FOREIGN_KEY_STRICT_MODE))
            {
                foreignKeyCreateMode = FOREIGN_KEY_STRICT_MODE;
            }
            else
            {
                foreignKeyCreateMode = FOREIGN_KEY_AUTO_MODE;
            }
            deletionPolicy = System.getProperty(DELETION_POLICY_PROPERTY, "JDO2");

            identifierFactory = System.getProperty(IDENTIFIER_FACTORY_PROPERTY, "jpox");
            identifierCase = System.getProperty(IDENTIFIER_CASE_PROPERTY, "UpperCase");
            if (System.getProperty(IDENTIFIER_WORD_SEPARATOR_PROPERTY) != null)
            {
                setIdentifierWordSeparator(System.getProperty(IDENTIFIER_WORD_SEPARATOR_PROPERTY));
            }
            if (System.getProperty(IDENTIFIER_TABLE_PREFIX_PROPERTY) != null)
            {
                setIdentifierTablePrefix(System.getProperty(IDENTIFIER_TABLE_PREFIX_PROPERTY));
            }
            if (System.getProperty(IDENTIFIER_TABLE_SUFFIX_PROPERTY) != null)
            {
                setIdentifierTableSuffix(System.getProperty(IDENTIFIER_TABLE_SUFFIX_PROPERTY));
            }

            // JDO operation
            classLoaderResolverName = System.getProperty(CLASS_LOADER_RESOLVER_NAME_PROPERTY, "jdo");
            connectionPoolingType = System.getProperty(CONNECTION_POOLING_TYPE_PROPERTY, "None");
            connectionPoolingConfigurationFile = System.getProperty(CONNECTION_POOLING_CONFIGURATION_FILE_PROPERTY);

            defaultInheritanceStrategy = System.getProperty(DEFAULT_INHERITANCE_STRATEGY_PROPERTY, "JDO2");
            autoStartMechanism = System.getProperty(AUTO_START_MECHANISM_PROPERTY, SchemaAutoStarter.SCHEMA_TABLE);
            autoStartMechanismXmlFile = System.getProperty(AUTO_START_MECHANISM_XML_FILE_PROPERTY, "jpoxAutoStart.xml");
            autoStartMechanismMode = System.getProperty(AUTO_START_MECHANISM_MODE_PROPERTY, AutoStartMechanism.MODE_QUIET);
            autoStartClassNames = System.getProperty(AUTO_START_CLASS_NAMES_PROPERTY);
            poidTransactionAttribute = System.getProperty(POID_TRANSACTION_ATTRIBUTE_PROPERTY, "New");

            // Schema identifiers
            if (System.getProperty(IDENTIFIER_CASE_PROPERTY) != null)
            {
                setIdentifierCase(System.getProperty(IDENTIFIER_CASE_PROPERTY));
            }

            // MetaData files
            String extension = System.getProperty(METADATA_JDO_FILE_EXTENSION_PROPERTY);
            if (!StringUtils.isWhitespace(extension))
            {
                jdoMetaDataFileExtension = extension;
            }
            extension = System.getProperty(METADATA_ORM_FILE_EXTENSION_PROPERTY);
            if (!StringUtils.isWhitespace(extension))
            {
                ormMetaDataFileExtension = extension;
            }
            extension = System.getProperty(METADATA_JDOQUERY_FILE_EXTENSION_PROPERTY);
            if (!StringUtils.isWhitespace(extension))
            {
                jdoqueryMetaDataFileExtension = extension;
            }
            metadataValidate = new Boolean(System.getProperty(METADATA_VALIDATE_PROPERTY,"true")).booleanValue();

            // Caching
            jdoCacheCollections = new Boolean(System.getProperty(CACHE_COLLECTIONS_PROPERTY,"true")).booleanValue();
            if (System.getProperty(CACHE_COLLECTIONS_LAZY_PROPERTY) != null)
            {
                jdoCacheCollectionsLazy = new Boolean(System.getProperty(CACHE_COLLECTIONS_LAZY_PROPERTY));
            }
            jdoCacheLevel1Type = System.getProperty(CACHE_LEVEL_1_TYPE_PROPERTY, "weak");
            jdoCacheLevel2 = new Boolean(System.getProperty(CACHE_LEVEL_2_PROPERTY,"false")).booleanValue();
            jdoCacheLevel2Type = System.getProperty(CACHE_LEVEL_2_TYPE_PROPERTY, "default");
            jdoCacheLevel2CacheName = System.getProperty(CACHE_LEVEL_2_CACHE_NAME_PROPERTY);
            jdoCacheLevel2ConfigurationFile = System.getProperty(CACHE_LEVEL_2_CONFIGURATION_FILE_PROPERTY);
            
            // Queries
            queryUseFetchPlan = new Boolean(System.getProperty(QUERY_USE_FETCH_PLAN_PROPERTY,"true")).booleanValue();
            String queryTimeoutString = System.getProperty(QUERY_TIMEOUT_PROPERTY);
            if (queryTimeoutString != null)
            {
                try
                {
                    queryTimeout = (new Integer(queryTimeoutString)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing and use default.
                }
            }
            String queryFetchSizeString = System.getProperty(QUERY_FETCH_SIZE_PROPERTY);
            if (queryFetchSizeString != null)
            {
                try
                {
                    queryFetchSize = (new Integer(queryFetchSizeString)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing and use default.
                }
            }
            queryFetchDirection = System.getProperty(QUERY_FETCH_DIRECTION_PROPERTY, "forward");
            queryResultSetType = System.getProperty(QUERY_RESULT_SET_TYPE_PROPERTY, "forward-only");
            queryResultSetConcurrency = System.getProperty(QUERY_RESULT_SET_CONCURRENCY_PROPERTY, "read-only");
            queryAllowAllSQLStatements = new Boolean(System.getProperty(QUERY_ALLOW_ALL_SQL_STATEMENTS, "false")).booleanValue();
            queryUseInnerJoinsOnly = new Boolean(System.getProperty(QUERY_USE_INNER_JOINS_ONLY_PROPERTY, "false")).booleanValue();

            // Types
            String stringLengthTemp = System.getProperty(STRING_DEFAULT_LENGTH_PROPERTY);
            if (stringLengthTemp != null)
            {
                try
                {
                    stringDefaultLength = (new Integer(stringLengthTemp)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    stringDefaultLength = JDO2_STRING_DEFAULT_LENGTH;
                }
            }
            dateTimezone = System.getProperty(DATE_TIMEZONE_PROPERTY, "UTC");
            DATE_TIMEZONE = TimeZone.getTimeZone(dateTimezone);
            if (DATE_TIMEZONE == null)
            {
                DATE_TIMEZONE = TimeZone.getTimeZone("UTC");
            }

            setPropertiesFile(System.getProperty(PROPERTIES_FILE));
            
        }
        catch (Exception e)
        {
            // May arrive here if the server throws a SecurityException
            JPOXLogger.GENERAL.error(e);
        }
    }

    /**
     * Equality operator.
     * @param obj Object to compare against.
     * @return Whether the objects are equal.
     **/
    public synchronized boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }
        if (!(obj instanceof PMFConfiguration))
        {
            return false;
        }

        PMFConfiguration pmfConfiguration = (PMFConfiguration)obj;

        boolean isEquals = true;
        isEquals = isEquals & equalsVariable(driverName, pmfConfiguration.driverName);
        isEquals = isEquals & equalsVariable(url, pmfConfiguration.url);
        isEquals = isEquals & equalsVariable(userName, pmfConfiguration.userName);
        isEquals = isEquals & equalsVariable(password, pmfConfiguration.password);
        isEquals = isEquals & equalsVariable(connectionFactory, pmfConfiguration.connectionFactory);
        isEquals = isEquals & equalsVariable(connectionFactoryName, pmfConfiguration.connectionFactoryName);
        isEquals = isEquals & equalsVariable(connectionFactory2, pmfConfiguration.connectionFactory2);
        isEquals = isEquals & equalsVariable(connectionFactory2Name, pmfConfiguration.connectionFactory2Name);
        isEquals = isEquals & equalsVariable(mapping, pmfConfiguration.mapping);
        isEquals = isEquals & equalsVariable(catalogName, pmfConfiguration.catalogName);
        isEquals = isEquals & equalsVariable(schemaName, pmfConfiguration.schemaName);
        isEquals = isEquals & equalsVariable(transactionType, pmfConfiguration.transactionType);

        isEquals = isEquals & equalsVariable(persistenceUnitName, pmfConfiguration.persistenceUnitName);

        isEquals = isEquals & equalsVariable(jdoCacheLevel1Type, pmfConfiguration.jdoCacheLevel1Type);
        isEquals = isEquals & equalsVariable(jdoCacheLevel2Type, pmfConfiguration.jdoCacheLevel2Type);
        isEquals = isEquals & equalsVariable(jdoCacheLevel2CacheName, pmfConfiguration.jdoCacheLevel2CacheName);
        isEquals = isEquals & equalsVariable(jdoCacheLevel2ConfigurationFile, pmfConfiguration.jdoCacheLevel2ConfigurationFile);
        isEquals = isEquals & (jdoCacheLevel2 == pmfConfiguration.jdoCacheLevel2);
        isEquals = isEquals & (jdoCacheCollections == pmfConfiguration.jdoCacheCollections);
        isEquals = isEquals & (jdoCacheCollectionsLazy == pmfConfiguration.jdoCacheCollectionsLazy);

        isEquals = isEquals & equalsVariable(datastoreAdapterClassName, pmfConfiguration.datastoreAdapterClassName);
        isEquals = isEquals & equalsVariable(persistentIdGenerator, pmfConfiguration.persistentIdGenerator);
        isEquals = isEquals & equalsVariable(autoStartMechanism, pmfConfiguration.autoStartMechanism);
        isEquals = isEquals & equalsVariable(autoStartMechanismXmlFile, pmfConfiguration.autoStartMechanismXmlFile);
        isEquals = isEquals & equalsVariable(autoStartMechanismMode, pmfConfiguration.autoStartMechanismMode);
        isEquals = isEquals & equalsVariable(autoStartClassNames, pmfConfiguration.autoStartClassNames);

        isEquals = isEquals & equalsVariable(poidTransactionAttribute,pmfConfiguration.poidTransactionAttribute);
        isEquals = isEquals & (checkExistTablesOrViews == pmfConfiguration.checkExistTablesOrViews);
        isEquals = isEquals & equalsVariable(defaultInheritanceStrategy, pmfConfiguration.defaultInheritanceStrategy);
        isEquals = isEquals & (stringDefaultLength == pmfConfiguration.stringDefaultLength);
        isEquals = isEquals & equalsVariable(dateTimezone, pmfConfiguration.dateTimezone);
        isEquals = isEquals & (statementBatchLimit == pmfConfiguration.statementBatchLimit);

        // MetaData
        isEquals = isEquals & equalsVariable(jdoMetaDataFileExtension, pmfConfiguration.jdoMetaDataFileExtension);
        isEquals = isEquals & equalsVariable(ormMetaDataFileExtension, pmfConfiguration.ormMetaDataFileExtension);
        isEquals = isEquals & equalsVariable(jdoqueryMetaDataFileExtension, pmfConfiguration.jdoqueryMetaDataFileExtension);
        isEquals = isEquals & (metadataValidate == pmfConfiguration.metadataValidate);

        // query
        isEquals = isEquals & (queryUseFetchPlan == pmfConfiguration.queryUseFetchPlan);
        isEquals = isEquals & (queryAllowAllSQLStatements == pmfConfiguration.queryAllowAllSQLStatements);
        isEquals = isEquals & (queryTimeout == pmfConfiguration.queryTimeout);
        isEquals = isEquals & (queryFetchSize == pmfConfiguration.queryFetchSize);
        isEquals = isEquals & equalsVariable(queryFetchDirection, pmfConfiguration.queryFetchDirection);
        isEquals = isEquals & equalsVariable(queryResultSetConcurrency, pmfConfiguration.queryResultSetConcurrency);
        isEquals = isEquals & equalsVariable(queryResultSetType, pmfConfiguration.queryResultSetType);
        isEquals = isEquals & (queryUseInnerJoinsOnly == pmfConfiguration.queryUseInnerJoinsOnly);

        isEquals = isEquals & (persistenceByReachabilityAtCommit == pmfConfiguration.persistenceByReachabilityAtCommit);
        isEquals = isEquals & (maxFetchDepth == pmfConfiguration.maxFetchDepth);
        isEquals = isEquals & equalsVariable(classLoaderResolverName, pmfConfiguration.classLoaderResolverName);
        isEquals = isEquals & equalsVariable(implementationCreatorName, pmfConfiguration.implementationCreatorName);

        isEquals = isEquals & equalsVariable(connectionPoolingType, pmfConfiguration.connectionPoolingType);
        isEquals = isEquals & equalsVariable(connectionPoolingConfigurationFile, pmfConfiguration.connectionPoolingConfigurationFile);

        isEquals = isEquals & (datastoreDelayOperationsEnabled == pmfConfiguration.datastoreDelayOperationsEnabled);
        isEquals = isEquals & (datastoreClassAdditionMaxRetries == pmfConfiguration.datastoreClassAdditionMaxRetries);
        isEquals = isEquals & (isolationLevel == pmfConfiguration.isolationLevel);
        isEquals = isEquals & (poidIsolationLevel == pmfConfiguration.poidIsolationLevel);
        isEquals = isEquals & (multithreaded == pmfConfiguration.multithreaded);
        isEquals = isEquals & (optimistic == pmfConfiguration.optimistic);
        isEquals = isEquals & (retainValues == pmfConfiguration.retainValues);
        isEquals = isEquals & (restoreValues == pmfConfiguration.restoreValues);
        isEquals = isEquals & (nontransactionalRead == pmfConfiguration.nontransactionalRead);
        isEquals = isEquals & (nontransactionalWrite == pmfConfiguration.nontransactionalWrite);
        isEquals = isEquals & (ignoreCache == pmfConfiguration.ignoreCache);
        isEquals = isEquals & (detachOnClose == pmfConfiguration.detachOnClose);
        isEquals = isEquals & (detachAllOnCommit == pmfConfiguration.detachAllOnCommit);
        
        //schema
        isEquals = isEquals & (validateTables == pmfConfiguration.validateTables);
        isEquals = isEquals & (validateColumns == pmfConfiguration.validateColumns);
        isEquals = isEquals & (validateConstraints == pmfConfiguration.validateConstraints);
        isEquals = isEquals & (autoCreateTables == pmfConfiguration.autoCreateTables);
        isEquals = isEquals & (autoCreateColumns == pmfConfiguration.autoCreateColumns);
        isEquals = isEquals & (autoCreateConstraints == pmfConfiguration.autoCreateConstraints);
        isEquals = isEquals & (autoCreateWarnOnError == pmfConfiguration.autoCreateWarnOnError);
        isEquals = isEquals & (initializePrimaryKeyColumnInfo == pmfConfiguration.initializePrimaryKeyColumnInfo);
        isEquals = isEquals & (readOnlyDatastore == pmfConfiguration.readOnlyDatastore);
        isEquals = isEquals & (fixedDatastore == pmfConfiguration.fixedDatastore);
        isEquals = isEquals & (foreignKeyCreateMode == pmfConfiguration.foreignKeyCreateMode);
        isEquals = isEquals & (deletionPolicy == pmfConfiguration.deletionPolicy);
        isEquals = isEquals & (uniqueConstraintsMapInverse == pmfConfiguration.uniqueConstraintsMapInverse);
        isEquals = isEquals & (propertiesFile == pmfConfiguration.propertiesFile);

        return isEquals;
    }

    /**
     * Equals two variables testing for nulls
     * @param var1 the first variable
     * @param var2 the second variable
     * @return true if equals
     */
    private boolean equalsVariable(Object var1, Object var2)
    {
        if (var1 == null)
        {
            if (var2 != null)
            {
                return false;
            }
        }
        else if (!var1.equals(var2))
        {
            return false;
        }
        return true;
    }

    // ------------------------------ JDO Standard Getters/Setters --------------------------------

    /**
     * Set the user name for the data store connection.
     * @param userName the user name for the data store connection.
     */
    public synchronized void setConnectionUserName(String userName)
    {
        assertConfigurable();
        this.userName = userName;
    }

    /**
     * Get the user name for the data store connection.
     * @return the user name for the data store connection.
     */
    public String getConnectionUserName()
    {
        return userName;
    }

    /**
     * Set the password for the data store connection.
     * @param password the password for the data store connection.
     */
    public synchronized void setConnectionPassword(String password)
    {
        assertConfigurable();
        this.password = password;
    }

    /**
     * Set the URL for the data store connection.
     * @param url the URL for the data store connection.
     */
    public synchronized void setConnectionURL(String url)
    {
        assertConfigurable();
        this.url = url;
    }

    /**
     * Get the URL for the data store connection.
     * @return the URL for the data store connection.
     */
    public String getConnectionURL()
    {
        return url;
    }

    /**
     * Set the driver name for the data store connection.
     * @param driverName the driver name for the data store connection.
     */
    public synchronized void setConnectionDriverName(String driverName)
    {
        assertConfigurable();
        this.driverName = driverName;
    }

    /**
     * Get the driver name for the data store connection.
     * @return the driver name for the data store connection.
     */
    public String getConnectionDriverName()
    {
        return driverName;
    }

    /**
     * Get the password for the data store connection.
     * @return the password for the data store connection.
     */
    public String getConnectionPassword()
    {
        return password;
    }

    /**
     * Set the name for the data store connection factory.
     * @param connectionFactoryName name of the data store connection factory.
     */
    public synchronized void setConnectionFactoryName(String connectionFactoryName)
    {
        assertConfigurable();
        this.connectionFactoryName = connectionFactoryName;
    }

    /**
     * Get the name for the data store connection factory.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactoryName()
    {
        return connectionFactoryName;
    }

    /**
     * Set the data store connection factory.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public synchronized void setConnectionFactory(Object connectionFactory)
    {
        assertConfigurable();
        this.connectionFactory = connectionFactory;
    }

    /**
     * Get the data store connection factory.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory()
    {
        return connectionFactory;
    }

    /**
     * Set the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @param connectionFactoryName name of the data store connection factory.
     */
    public synchronized void setConnectionFactory2Name(String connectionFactoryName)
    {
        assertConfigurable();
        this.connectionFactory2Name = connectionFactoryName;
    }

    /**
     * Get the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactory2Name()
    {
        return connectionFactory2Name;
    }

    /**
     * Set the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public synchronized void setConnectionFactory2(Object connectionFactory)
    {
        assertConfigurable();
        this.connectionFactory2 = connectionFactory;
    }

    /**
     * Get the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory2()
    {
        return connectionFactory2;
    }

    /**
     * Set the default Multithreaded setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default Multithreaded setting.
     */
    public synchronized void setMultithreaded(boolean flag)
    {
        assertConfigurable();
        multithreaded = flag;
    }

    /**
     * Get the default Multithreaded setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default Multithreaded setting.
     */
    public boolean getMultithreaded()
    {
        return multithreaded;
    }

    /**
     * Set the default Optimistic setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default Optimistic setting.
     */
    public synchronized void setOptimistic(boolean flag)
    {
        assertConfigurable();
        optimistic = flag;
    }

    /**
     * Get the default Optimistic setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default Optimistic setting.
     */
    public boolean getOptimistic()
    {
        return optimistic;
    }

    /**
     * Set the default RetainValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default RetainValues setting.
     */
    public synchronized void setRetainValues(boolean flag)
    {
        assertConfigurable();
        retainValues = flag;
    }

    /**
     * Get the default RetainValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default RetainValues setting.
     */
    public boolean getRetainValues()
    {
        return retainValues;
    }

    /**
     * Set the default RestoreValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default RestoreValues setting.
     */
    public synchronized void setRestoreValues(boolean flag)
    {
        assertConfigurable();
        restoreValues = flag;
    }

    /**
     * Get the default RestoreValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default RestoreValues setting.
     */
    public boolean getRestoreValues()
    {
        return restoreValues;
    }

    /**
     * Set the default NontransactionalRead setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default NontransactionalRead setting.
     */
    public synchronized void setNontransactionalRead(boolean flag)
    {
        assertConfigurable();
        nontransactionalRead = flag;
    }

    /**
     * Get the default NontransactionalRead setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default NontransactionalRead setting.
     */
    public boolean getNontransactionalRead()
    {
        return nontransactionalRead;
    }

    /**
     * Set the default NontransactionalWrite setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default NontransactionalWrite setting.
     */
    public synchronized void setNontransactionalWrite(boolean flag)
    {
        if (flag)
        {
            throw new JDOUnsupportedOptionException("nontransactionalWrite is not supported by JPOX");
        }
        assertConfigurable();
        nontransactionalWrite = flag;
    }

    /**
     * Get the default NontransactionalWrite setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default NontransactionalWrite setting.
     */
    public boolean getNontransactionalWrite()
    {
        return nontransactionalWrite;
    }

    /**
     * Set the default IgnoreCache setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default IgnoreCache setting.
     */
    public synchronized void setIgnoreCache(boolean flag)
    {
        assertConfigurable();
        ignoreCache = flag;
    }

    /**
     * Get the default IgnoreCache setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the IgnoreCache setting.
     */
    public boolean getIgnoreCache()
    {
        return ignoreCache;
    }

    /**
     * Set the default DetachOnClose setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default DetachOnClose setting.
     * @since 1.1
     */
    public synchronized void setDetachOnClose(boolean flag)
    {
        assertConfigurable();
        detachOnClose = flag;
    }

    /**
     * Get the default DetachOnClose setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the DetachOnClose setting.
     * @since 1.1
     */
    public boolean getDetachOnClose()
    {
        return detachOnClose;
    }

    /**
     * Mutator for the DetachAllOnCommit setting.
     * @param flag the default DetachAllOnCommit setting.
     * @since 1.1
     */
    public synchronized void setDetachAllOnCommit(boolean flag)
    {
        assertConfigurable();
        detachAllOnCommit = flag;
    }

    /**
     * Accessor for the DetachAllOnCommit setting.
     * @return the DetachAllOnCommit setting.
     * @since 1.1
     */
    public boolean getDetachAllOnCommit()
    {
        return detachAllOnCommit;
    }

    /**
     * Set the name for any mapping, used in searching for ORM/Query metadata
     * files.
     * @param mapping the mapping name
     */
    public synchronized void setMapping(String mapping)
    {
        assertConfigurable();
        this.mapping = mapping;
    }

    /**
     * Get the name for any mapping, used in retrieving metadata files for
     * ORM/Query data.
     * @return the name for the mapping.
     */
    public String getMapping()
    {
        return mapping;
    }

    /**
     * Mutator for the catalog to use for this PMF.
     * @param catalog Name of the catalog
     * @since 1.1
     */
    public synchronized void setCatalog(String catalog)
    {
        assertConfigurable();
        this.catalogName = catalog;
    }

    /**
     * Accessor for the catalog to use for this PMF.
     * @return the name of the catalog
     * @since 1.1
     */
    public String getCatalog()
    {
        return catalogName;
    }

    /**
     * Mutator for the schema to use for this PMF.
     * @param schema Name of the schema
     * @since 1.1
     */
    public synchronized void setSchema(String schema)
    {
        assertConfigurable();
        this.schemaName = schema;
    }

    /**
     * Accessor for the schema to use for this PMF.
     * @return the name of the schema
     * @since 1.1
     */
    public String getSchema()
    {
        return schemaName;
    }

    /**
     * Mutator for the transaction type to use for this PMF.
     * @param type Transaction type
     */
    public synchronized void setTransactionType(String type)
    {
        assertConfigurable();
        if (type == null)
        {
            return;
        }
        if (type.equalsIgnoreCase("JTA") || type.equalsIgnoreCase("ResourceLocal"))
        {
            this.transactionType = type;
        }
        else
        {
            // Localise this message
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", JDO_TRANSACTION_TYPE_PROPERTY, type));
        }
    }

    /**
     * Accessor for the transaction type to use with this PMF.
     * @return transaction type
     */
    public String getTransactionType()
    {
        return transactionType;
    }

    // --------------------------- JPOX Extension Getters/Setters ---------------------

    /**
     * Mutator for the name of the persistence unit.
     * @param name Name of the persistence unit
     */
    public synchronized void setPersistenceUnitName(String name)
    {
        assertConfigurable();
        this.persistenceUnitName = name;
    }

    /**
     * Accessor for the name of the persistence unit
     * @return the name of the persistence unit
     */
    public String getPersistenceUnitName()
    {
        return persistenceUnitName;
    }

    /**
     * Set the name of the datastore adapter to use.
     * @param adapterClassName Name of the class of the datastore adapter to use.
     * @since 1.1
     */
    public synchronized void setDatastoreAdapterClassName(String adapterClassName)
    {
        assertConfigurable();
        datastoreAdapterClassName = adapterClassName;
    }

    /**
     * Accessor for the datastore adapter class name (null implies autodetect)
     * @return the datastore adapter class name to use.
     * @since 1.1
     */
    public String getDatastoreAdapterClassName()
    {
        return datastoreAdapterClassName;
    }

    /**
     * Set the connection pooling type.
     * @param type The connection pooling type
     * @since 1.1
     */
    public synchronized void setConnectionPoolingType(String type)
    {
        assertConfigurable();
        if (type == null)
        {
            type = "None";
        }
        this.connectionPoolingType = type;
    }

    /**
     * Get the connection pooling type.
     * @return the connection pooling type to be used.
     * @since 1.1
     */
    public String getConnectionPoolingType()
    {
        return connectionPoolingType;
    }

    /**
     * Set the connection pooling configuration file to use.
     * @param file The connection pooling configuration file
     * @since 1.1
     */
    public synchronized void setConnectionPoolingConfigurationFile(String file)
    {
        assertConfigurable();

        this.connectionPoolingConfigurationFile = file;
    }

    /**
     * Get the connection pooling configuration file.
     * @return the connection pooling configuration file to be used.
     * @since 1.1
     */
    public String getConnectionPoolingConfigurationFile()
    {
        return connectionPoolingConfigurationFile;
    }

    /**
     * Set the limit on number of SQL statements that can be batched in one go.
     * This supports values of 0 for no batching and -1 for unlimited batching.
     * @param limit Max number of statements to batch
     */
    public synchronized void setStatementBatchLimit(int limit)
    {
        assertConfigurable();
        statementBatchLimit = limit;
    }

    /**
     * Accessor for the limit on number of SQL statements that can be batched in one go.
     * @return Max number of statements to batch
     */
    public int getStatementBatchLimit()
    {
        return statementBatchLimit;
    }

    /**
     * Set the default CheckExistTablesOrViews setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default CheckExistTablesOrViews setting.
     */
    public synchronized void setCheckExistTablesOrViews(boolean flag)
    {
        assertConfigurable();
        checkExistTablesOrViews = flag;
    }

    /**
     * Get the default CheckExistTablesOrViews setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default CheckExistTablesOrViews setting.
     */
    public boolean getCheckExistTablesOrViews()
    {
        return checkExistTablesOrViews;
    }

    /**
     * Set the default ValidateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default ValidateTables setting.
     */
    public synchronized void setValidateTables(boolean flag)
    {
        assertConfigurable();
        validateTables = flag;
        if (!validateTables)
        {
            // Can't validate the columns if we aren't validating the table
            setValidateColumns(false);
        }
    }

    /**
     * Get the default ValidateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default ValidateTables setting.
     */
    public boolean getValidateTables()
    {
        return validateTables;
    }

    /**
     * Set the default InitializePrimaryKeyColumnInfo setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default InitializePrimaryKeyColumnInfo setting.
     */
    public synchronized void setInitializePrimaryKeyColumnInfo(boolean flag)
    {
        assertConfigurable();
        initializePrimaryKeyColumnInfo = flag;
    }

    /**
     * Get the default InitializePrimaryKeyColumnInfo setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default InitializePrimaryKeyColumnInfo setting.
     */
    public boolean getInitializePrimaryKeyColumnInfo()
    {
        return initializePrimaryKeyColumnInfo;
    }
    
    /**
     * Set the default ValidateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default ValidateColumns setting.
     * @since 1.1
     */
    public synchronized void setValidateColumns(boolean flag)
    {
        assertConfigurable();
        validateColumns = flag;
    }

    /**
     * Get the default ValidateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default ValidateColumns setting.
     * @since 1.1
     */
    public boolean getValidateColumns()
    {
        return validateColumns;
    }

    /**
     * Set the default ValidateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default ValidateConstraints setting.
     */
    public synchronized void setValidateConstraints(boolean flag)
    {
        assertConfigurable();
        validateConstraints = flag;
    }

    /**
     * Get the default ValidateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default ValidateConstraints setting.
     */
    public boolean getValidateConstraints()
    {
        return validateConstraints;
    }

    /**
     * Set whether this datastore schema is read-only.
     * @param flag the read-only setting.
     */
    public synchronized void setReadOnlySchema(boolean flag)
    {
        throw new JDOUserException("org.jpox.readOnlySchema is deprecated, replaced by org.jpox.readOnlyDatastore");
    }

    /**
     * Set whether this datastore schema is fixed
     * @param flag the fixed setting.
     */
    public synchronized void setFixedSchema(boolean flag)
    {
        throw new JDOUserException("org.jpox.fixedSchema is deprecated, replaced by org.jpox.fixedDatastore");
    }

    /**
     * Set whether this datastore is readOnly
     * @param flag the readOnly setting.
     */
    public synchronized void setReadOnlyDatastore(boolean flag)
    {
        assertConfigurable();
        readOnlyDatastore = flag;
        if (flag)
        {
            setAutoCreateTables(false);
            setAutoCreateColumns(false);
            setAutoCreateConstraints(false);
        }
    }

    /**
     * Accessor for whether the datastore is read-only (no structural or content changes allowed).
     * @return the readOnly setting.
     */
    public boolean getReadOnlyDatastore()
    {
        return readOnlyDatastore;
    }

    /**
     * Set whether this datastore schema is fixed
     * @param flag the fixed setting.
     */
    public synchronized void setFixedDatastore(boolean flag)
    {
        assertConfigurable();
        fixedDatastore = flag;
        if (flag)
        {
            setAutoCreateTables(false);
            setAutoCreateColumns(false);
            setAutoCreateConstraints(false);
        }
    }

    /**
     * Accessor for whether the datastore is read-only (no structural changes allowed).
     * @return the fixed setting.
     */
    public boolean getFixedDatastore()
    {
        return fixedDatastore;
    }

    /**
     * Mutator for the foreign key creation mode
     * @param mode the FK create mode
     * @since 1.1
     */
    public synchronized void setForeignKeyCreateMode(String mode)
    {
        assertConfigurable();
        if (mode.equals(FOREIGN_KEY_AUTO_MODE) || mode.equals(FOREIGN_KEY_STRICT_MODE))
        {
            foreignKeyCreateMode = mode;
        }
    }

    /**
     * Accessor for the FK creation mode.
     * @return the FK creation mode
     * @since 1.1
     */
    public String getForeignKeyCreateMode()
    {
        return foreignKeyCreateMode;
    }

    /**
     * Mutator for the deletion policy
     * @param policy the FK create mode
     * @since 1.1
     */
    public synchronized void setDeletionPolicy(String policy)
    {
        assertConfigurable();
        if (policy.equals("JPOX") || policy.equals("JDO2"))
        {
            deletionPolicy = policy;
        }
    }

    /**
     * Accessor for the deletion policy.
     * @return the deletion policy
     * @since 1.1
     */
    public String getDeletionPolicy()
    {
        return deletionPolicy;
    }

    /**
     * Mutator for the identifier factory
     * @param factory the identifier factory
     */
    public synchronized void setIdentifierFactory(String factory)
    {
        assertConfigurable();
        identifierFactory = factory;
    }

    /**
     * Accessor for the identifier factory
     * @return the identifier factory
     */
    public String getIdentifierFactory()
    {
        return identifierFactory;
    }

    /**
     * Mutator for the identifier case
     * @param idCase the identifier case
     */
    public synchronized void setIdentifierCase(String idCase)
    {
        assertConfigurable();
        if (idCase.equalsIgnoreCase("UpperCase") || idCase.equalsIgnoreCase("LowerCase") || idCase.equalsIgnoreCase("PreserveCase"))
        {
            identifierCase = idCase;
        }
    }

    /**
     * Accessor for the identifier case
     * @return the identifier case
     */
    public String getIdentifierCase()
    {
        return identifierCase;
    }

    /**
     * Mutator for the identifier word separator
     * @param word the identifier word separator
     */
    public synchronized void setIdentifierWordSeparator(String word)
    {
        assertConfigurable();
        identifierWordSeparator = word;
    }

    /**
     * Accessor for the identifier word separator
     * @return the identifier word separator
     */
    public String getIdentifierWordSeparator()
    {
        return identifierWordSeparator;
    }

    /**
     * Mutator for the identifier table prefix
     * @param prefix the identifier table prefix
     */
    public synchronized void setIdentifierTablePrefix(String prefix)
    {
        assertConfigurable();
        identifierTablePrefix = prefix;
    }

    /**
     * Accessor for the identifier table prefix
     * @return the identifier table prefix
     */
    public String getIdentifierTablePrefix()
    {
        return identifierTablePrefix;
    }

    /**
     * Mutator for the identifier table suffix
     * @param suffix the identifier table suffix
     */
    public synchronized void setIdentifierTableSuffix(String suffix)
    {
        assertConfigurable();
        identifierTableSuffix = suffix;
    }

    /**
     * Accessor for the identifier table suffix
     * @return the identifier table suffix
     */
    public String getIdentifierTableSuffix()
    {
        return identifierTableSuffix;
    }

    /**
     * Set the default UniqueConstraintsMapInverse setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default UniqueConstraintsMapInverse setting.
     */
    public synchronized void setUniqueConstraintsMapInverse(boolean flag)
    {
        assertConfigurable();
        uniqueConstraintsMapInverse = flag;
    }

    /**
     * Get the default UniqueConstraintsMapInverse setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default UniqueConstraintsMapInverse setting.
     */
    public boolean getUniqueConstraintsMapInverse()
    {
        return uniqueConstraintsMapInverse;
    }

    /**
     * Get the use update lock flag for fetches.
     * @return the default useUpdateLock setting.
     */
    public boolean getUseUpdateLock()
    {
        return useUpdateLock;
    }

    /**
     * Set whether to use the update lock for fetches.
     * @param flag the useUpdateLock setting.
     */
    public synchronized void setUseUpdateLock(boolean flag)
    {
        assertConfigurable();
        useUpdateLock = flag;
    }

    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     * @return true if datastore operations are delayed until commit
     */
    public boolean getDatastoreDelayOperationsEnabled()
    {
        return datastoreDelayOperationsEnabled;
    }

    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     * @param flag true if datastore operations should be delayed until commit
     */
    public synchronized void setDatastoreDelayOperationsEnabled(boolean flag)
    {
        assertConfigurable();
        datastoreDelayOperationsEnabled = flag;
    }

    /**
     * Accessor for the maximum number of retries for adding a class to the store manager.
     * @return The max number of retries
     * @since 1.1
     */
    public int getDatastoreClassAdditionMaxRetries()
    {
        return datastoreClassAdditionMaxRetries;
    }

    /**
     * Set the max number of times to retry adding a class to the store manager.
     * @param max The max number of retries
     * @since 1.1
     */
    public void setDatastoreClassAdditionMaxRetries(int max)
    {
        this.datastoreClassAdditionMaxRetries = max;
    }

    /**
     * Accessor for whether "persistence-by-reachability" is run at commit time.
     * @return Whether to run PBR at commit time
     */
    public boolean getPersistenceByReachabilityAtCommit()
    {
        return persistenceByReachabilityAtCommit;
    }

    /**
     * Mutator for whether to run "persistence-by-reachability" at commit time.
     * @param flag Whether to run PBR at commit time
     */
    public synchronized void setPersistenceByReachabilityAtCommit(boolean flag)
    {
        assertConfigurable();
        persistenceByReachabilityAtCommit = flag;
    }

    /**
     * Accessor for the maximum fetch depth to use by default.
     * @return Max fetch depth to use
     */
    public int getMaxFetchDepth()
    {
        return maxFetchDepth;
    }

    /**
     * Mutator for the default max fetch depth
     * @param value Max fetch depth to use
     */
    public synchronized void setMaxFetchDepth(int value)
    {
        assertConfigurable();
        maxFetchDepth = value;
    }

    /**
     * Accessor for the name of the class loader resolver.
     * @return Name of the class loader resolver to use.
     */
    public String getClassLoaderResolverName()
    {
        return classLoaderResolverName;
    }

    /**
     * Mutator for the class loader resolver name to use.
     * @param clrName Class Loader resolver name to use
     */
    public synchronized void setClassLoaderResolverName(String clrName)
    {
        assertConfigurable();
        classLoaderResolverName = clrName;
    }

    /**
     * Accessor for the name of the implementation creator class.
     * @return Name of the implementation creator class to use.
     */
    public String getImplementationCreatorName()
    {
        return implementationCreatorName;
    }

    /**
     * Mutator for the implementation creator name to use.
     * @param implCreatorName Name of implementation creator class to use
     */
    public synchronized void setImplementationCreatorName(String implCreatorName)
    {
        assertConfigurable();
        implementationCreatorName = implCreatorName;
    }

    /**
     * Accessor for the default inheritance strategy to use.
     * @return the default inheritance strategy
     * @since 1.1
     */
    public String getDefaultInheritanceStrategy()
    {
        return defaultInheritanceStrategy;
    }

    /**
     * Mutator for the default inheritance strategy.
     * @param strategy the default inheritance strategy
     * @since 1.1
     */
    public synchronized void setDefaultInheritanceStrategy(String strategy)
    {
        assertConfigurable();
        if (strategy == null)
        {
            return;
        }
        else if (!strategy.equalsIgnoreCase("JDO2") && !strategy.equalsIgnoreCase("JPOX"))
        {
            // Illegal value
            return;
        }
        defaultInheritanceStrategy = strategy;
    }

    /**
     * Set the default AutoCreateSchema setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateSchema setting.
     */
    public synchronized void setAutoCreateSchema(boolean flag)
    {
        assertConfigurable();
        if (flag)
        {
            // This flag is a shortcut to set the tables and constraints
            setAutoCreateTables(true);
            setAutoCreateConstraints(true);
        }

        if ((readOnlyDatastore || fixedDatastore) && flag)
        {
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateSchema"));
        }
    }

    /**
     * Get the default AutoCreateSchema setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateSchema setting.
     */
    public boolean getAutoCreateSchema()
    {
        return (autoCreateTables && autoCreateConstraints);
    }

    /**
     * Set the default AutoCreateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateTables setting.
     */
    public synchronized void setAutoCreateTables(boolean flag)
    {
        assertConfigurable();
        autoCreateTables = flag;
        if ((readOnlyDatastore || fixedDatastore) && flag)
        {
            autoCreateTables = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateTables"));
        }
    }

    /**
     * Get the default AutoCreateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateTables setting.
     */
    public boolean getAutoCreateTables()
    {
        return autoCreateTables;
    }

    /**
     * Set the default AutoCreateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateColumns setting.
     * @since 1.1
     */
    public synchronized void setAutoCreateColumns(boolean flag)
    {
        assertConfigurable();
        autoCreateColumns = flag;
        if ((readOnlyDatastore || fixedDatastore) && flag)
        {
            autoCreateTables = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateColumns"));
        }
    }

    /**
     * Get the default AutoCreateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateColumns setting.
     * @since 1.1
     */
    public boolean getAutoCreateColumns()
    {
        return autoCreateColumns;
    }

    /**
     * Set the default AutoCreateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateConstraints setting.
     */
    public synchronized void setAutoCreateConstraints(boolean flag)
    {
        assertConfigurable();
        autoCreateConstraints = flag;
        if ((readOnlyDatastore || fixedDatastore) && flag)
        {
            autoCreateConstraints = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateConstraints"));
        }
    }

    /**
     * Get the default AutoCreateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateConstraints setting.
     */
    public boolean getAutoCreateConstraints()
    {
        return autoCreateConstraints;
    }

    /**
     * Set the default AutoCreateWarnOnError setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateWarnOnError setting.
     */
    public synchronized void setAutoCreateWarnOnError(boolean flag)
    {
        assertConfigurable();
        autoCreateWarnOnError = flag;
    }

    /**
     * Get the default AutoCreateWarnOnError setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateWarnOnError setting.
     */
    public boolean getAutoCreateWarnOnError()
    {
        return autoCreateWarnOnError;
    }

    /**
     * Accessor for the Auto-Start Mechanism for management of which classes
     * are supported in this data store.
     * @return the Auto-Start Mechanism setting.
     */
    public String getAutoStartMechanism()
    {
        if (autoStartMechanism == null)
        {
            return AutoStartMechanism.NONE;
        }

        return autoStartMechanism;
    }

    /**
     * Set the AutoStartMechanism setting to this PMF.
     * @param mechanism The autostart mechanism plugin name
     */
    public synchronized void setAutoStartMechanism(String mechanism)
    {
        assertConfigurable();
        autoStartMechanism = mechanism;
    }

    /**
     * Accessor for the name of the XML file to use when using the XML AutoStarter
     * @return The XML file name
     */
    public String getAutoStartMechanismXmlFile()
    {
        return autoStartMechanismXmlFile;
    }

    /**
     * Mutator for the XML filename when using the XML AutoStarter.
     * @param xmlFile Name of the XML file
     */
    public synchronized void setAutoStartMechanismXmlFile(String xmlFile)
    {
        autoStartMechanismXmlFile = xmlFile;
    }

    /**
     * Accessor for the Auto-Start Mechanism Mode for how the auto start
     * mechanism operates. Currently supports 3 possible values for this
     * parameter "Checked", "Ignored", and "Quiet" (default).
     * @return the Auto-Start Mechanism Mode setting.
     */
    public String getAutoStartMechanismMode()
    {
        if (autoStartMechanismMode != null &&
            autoStartMechanismMode.equalsIgnoreCase(AutoStartMechanism.MODE_CHECKED))
        {
            return AutoStartMechanism.MODE_CHECKED;
        }
        else if (autoStartMechanismMode != null &&
                 autoStartMechanismMode.equalsIgnoreCase(AutoStartMechanism.MODE_IGNORED))
        {
            return AutoStartMechanism.MODE_IGNORED;
        }
        else    // Default
        {
            return AutoStartMechanism.MODE_QUIET;
        }
    }

    /**
     * Set the default AutoStartMechanismMode setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param mode Checked = raise exceptions on errors
     *             Ignored = proceed without corrections
     *             Quiet   = correct errors and proceed (default)
     */
    public synchronized void setAutoStartMechanismMode(String mode)
    {
        assertConfigurable();
        if (!mode.equals(AutoStartMechanism.MODE_CHECKED) &&
            !mode.equals(AutoStartMechanism.MODE_IGNORED) &&
            !mode.equals(AutoStartMechanism.MODE_QUIET))
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", AUTO_START_MECHANISM_MODE_PROPERTY, mode));
        }
        autoStartMechanismMode = mode;
    }

    /**
     * Accessor for the names of the classes to be loaded at startup.
     * @return Names of classes to load
     * @since 1.1
     */
    public String getAutoStartClassNames()
    {
        return autoStartClassNames;
    }

    /**
     * Mutator for the names of classes to load at startup.
     * @param classNames Name of the classes to load.
     * @since 1.1
     */
    public synchronized void setAutoStartClassNames(String classNames)
    {
        autoStartClassNames = classNames;
    }

    /**
     * Get the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     *
     * @return the default transaction isolation level for
     *         <tt>PoidGenerator</tt> instances.
     */
    public int getPoidTransactionIsolationLevel()
    {
        return poidIsolationLevel;
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     * @param isolationLevelName
     *      One of the values "read uncommitted", "read committed", "repeatable
     *      read", or "serializable".  The name is case-insensitive, and spaces
     *      and underscores are equivalent.
     */
    protected synchronized void setPoidTransactionIsolation(String isolationLevelName)
    {
        String iln = isolationLevelName.trim().replace(' ', '_').toUpperCase();
        poidIsolationLevel = JDBCUtils.getJDBCIsolationLevelForName(iln);
        if (poidIsolationLevel < 0)
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", POID_TRANSACTION_ISOLATION_PROPERTY, isolationLevelName));
        }
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     * @param isolationLevel the default transaction isolation level.
     */
    public synchronized void setPoidTransactionIsolation(int isolationLevel)
    {
        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", POID_TRANSACTION_ISOLATION_PROPERTY, "" + isolationLevel));
        }

        this.poidIsolationLevel = isolationLevel;
    }

    /**
     * Whether to use the PM connection or open a new connection
     * @return Returns the poidTransactionAttribute.
     */
    public String getPoidTransactionAttribute()
    {
        return poidTransactionAttribute;
    }

    /**
     * Set whether to use the PM connection or open a new connection
     * @param poidTransactionAttribute The poidTransactionAttribute to set.
     */
    public void setPoidTransactionAttribute(String poidTransactionAttribute)
    {
        if (poidTransactionAttribute == null)
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", 
                "org.jpox.poid.transactionAttribute", poidTransactionAttribute));
        }
        if (!poidTransactionAttribute.equals("New") && 
            !poidTransactionAttribute.equals("UsePM"))
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", 
                "org.jpox.poid.transactionAttribute", poidTransactionAttribute));
        }
        this.poidTransactionAttribute = poidTransactionAttribute;
    }

    /**
     * Accessor for whether to use the FetchPlan in queries.
     * @return Whether to use the FetchPlan in queries
     * @since 1.1
     */
    public boolean getQueryUseFetchPlan()
    {
        return queryUseFetchPlan;
    }

    /**
     * Set whether to use the FetchPlan in queries
     * @param useFetchPlan Whether to use the FetchPlan in queries
     * @since 1.1
     */
    public void setQueryUseFetchPlan(boolean useFetchPlan)
    {
        queryUseFetchPlan = useFetchPlan;
    }

    /**
     * Whether to allow running any SQL statement in SQL queries.
     * @return true if allowed to execute any SQL stament in SQL queries
     * @since 1.1
     */
    public boolean isQueryAllowAllSQLStatements()
    {
        return queryAllowAllSQLStatements;
    }

    /**
     * Whether to allow running any SQL statement in SQL queries.
     * @param allow true if allowed to execute any SQL stament in SQL queries
     * @since 1.1
     */
    public void setQueryAllowAllSQLStatements(boolean allow)
    {
        this.queryAllowAllSQLStatements = allow;
    }
    
    /**
     * Accessor for the timeout for queries.
     * @return The timeout
     * @since 1.1
     */
    public int getQueryTimeout()
    {
        return queryTimeout;
    }

    /**
     * Set the timeout for queries
     * @param timeout The timeout to use
     * @since 1.1
     */
    public void setQueryTimeout(int timeout)
    {
        this.queryTimeout = timeout;
    }

    /**
     * Accessor for the fetch size to use for ResultSet's.
     * @return The fetch size
     * @since 1.1
     */
    public int getQueryFetchSize()
    {
        return queryFetchSize;
    }

    /**
     * Set the fetch size to use for ResultSet's.
     * @param size The fetch size to use
     * @since 1.1
     */
    public void setQueryFetchSize(int size)
    {
        this.queryFetchSize = size;
    }

    /**
     * Accessor for the fetch direction to use for ResultSet's.
     * @return The fetch direction.
     * @since 1.1
     */
    public String getQueryFetchDirection()
    {
        return queryFetchDirection;
    }

    /**
     * Set the fetch direction to use for ResultSet's.
     * @param dir The fetch direction to use
     * @since 1.1
     */
    public void setQueryFetchDirection(String dir)
    {
        if (dir == null ||
            !dir.equals("forward") ||
            !dir.equals("reverse") ||
            !dir.equals("unknown"))
        {
            return;
        }
        queryFetchDirection = dir;
    }

    /**
     * Accessor for the type of ResultSet
     * @return The ResultSet type.
     * @since 1.1
     */
    public String getQueryResultSetType()
    {
        return queryResultSetType;
    }

    /**
     * Set the type of the ResultSet
     * @param type The ResultSet type
     * @since 1.1
     */
    public void setQueryResultSetType(String type)
    {
        if (type == null ||
            (!type.equals("forward-only") &&
            !type.equals("scroll-sensitive") &&
            !type.equals("scroll-insensitive")))
        {
            return;
        }
        queryResultSetType = type;
    }

    /**
     * Accessor for the concurrency of ResultSet
     * @return The ResultSet concurrency.
     * @since 1.1
     */
    public String getQueryResultSetConcurrency()
    {
        return queryResultSetConcurrency;
    }

    /**
     * Set the concurrency of the ResultSet
     * @param concur The ResultSet concurrency
     * @since 1.1
     */
    public void setQueryResultSetConcurrency(String concur)
    {
        if (concur == null ||
            (!concur.equals("read-only") &&
             !concur.equals("updateable")))
        {
            return;
        }
        queryResultSetConcurrency = concur;
    }

    /**
     * Accessor for whether to use inner joins only in JDOQL.
     * @return Whether to use inner joins only
     * @since 1.1
     */
    public boolean getQueryUseInnerJoinsOnly()
    {
        return queryUseInnerJoinsOnly;
    }

    /**
     * Set whether to use inner joins only
     * @param useInner Whether to use inner joins only
     * @since 1.1
     */
    public void setQueryUseInnerJoinsOnly(boolean useInner)
    {
        queryUseInnerJoinsOnly = useInner;
    }

    /**
     * Accessor for the default max length for strings.
     * @return The default length for strings
     */
    public int getStringDefaultLength()
    {
        return stringDefaultLength;
    }

    /**
     * Set the default length for strings.
     * @param len Default length for strings
     */
    public void setStringDefaultLength(int len)
    {
        this.stringDefaultLength = len;
    }

    /**
     * Accessor for the timezone in which all dates are stored in the datastore.
     * @return The timezone for all dates in the datastore
     * @since 1.1
     */
    public String getDateTimezone()
    {
        return dateTimezone;
    }

    /**
     * Set the timezone to use for dates in the datastore.
     * "JDK_DEFAULT_TIMEZONE" allows users to use the local JDK timezone.
     * @param timezone The timezone to use for all dates in the datastore
     * @since 1.1
     */
    public void setDateTimezone(String timezone)
    {
        this.dateTimezone = timezone;
        if (timezone != null && !timezone.equals("JDK_DEFAULT_TIMEZONE"))
        {
            DATE_TIMEZONE = TimeZone.getTimeZone(timezone);
            if (DATE_TIMEZONE == null)
            {
                DATE_TIMEZONE = TimeZone.getTimeZone("UTC");
            }
        }
    }

    /**
     * Accessor for the Calendar to be used in handling all timezone issues with the datastore.
     * The Calendar will be null if the user has selected "JDK_DEFAULT_TIMEZONE" as the DateTimezone.
     * @return The calendar to use for dateTimezone issues.
     */
    public Calendar getCalendarForDateTimezone()
    {
        if (dateTimezoneCalendar == null && !dateTimezone.equals("JDK_DEFAULT_TIMEZONE"))
        {
            dateTimezoneCalendar = new GregorianCalendar(DATE_TIMEZONE);
        }
        return dateTimezoneCalendar;
    }

    /**
     * Accessor for whether to cache collections.
     * @return whether to cache collections.
     */
    public boolean getJdoCacheCollections()
    {
        return jdoCacheCollections;
    }

    /**
     * Set whether to cache collections for this PMF.
     * @param cache Whether to cache collections.
     */
    public synchronized void setJdoCacheCollections(boolean cache)
    {
        assertConfigurable();
        jdoCacheCollections = cache;
    }

    /**
     * Accessor for whether to lazy load any cached collections.
     * @return whether to lazy load any cache collections.
     */
    public Boolean getJdoCacheCollectionsLazy()
    {
        return jdoCacheCollectionsLazy;
    }

    /**
     * Set whether to lazy load any cached collections for this PMF.
     * @param lazy Whether to lazy load any cache collections.
     */
    public synchronized void setJdoCacheCollectionsLazy(Boolean lazy)
    {
        assertConfigurable();
        if (lazy != null)
        {
            jdoCacheCollectionsLazy = lazy;
        }
    }

    /**
     * Accessor for the JDO Level 1 Cache Type
     * @return the JDO Level 1 Cache Type
     * @since 1.1
     */
    public String getJdoCacheLevel1Type()
    {
        return jdoCacheLevel1Type;
    }

    /**
     * Set the default Level 1 Cache Type for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param type the Level 1 Cache Type
     * @since 1.1
     */
    public synchronized void setJdoCacheLevel1Type(String type)
    {
        assertConfigurable();
        jdoCacheLevel1Type = type;
    }

    /**
     * Accessor for whether to use a level 2 Cache.
     * @return whether to use a level 2 Cache.
     * @since 1.1
     */
    public boolean getJdoCacheLevel2()
    {
        return jdoCacheLevel2;
    }

    /**
     * Set whether to use a Level 2 Cache for this PMF.
     * @param cache Whether to use a level 2 Cache
     * @since 1.1
     */
    public synchronized void setJdoCacheLevel2(boolean cache)
    {
        assertConfigurable();
        jdoCacheLevel2 = cache;
    }

    /**
     * Accessor for the JDO Level 2 Cache Type
     * @return the JDO Level 2 Cache Type
     * @since 1.1
     */
    public String getJdoCacheLevel2Type()
    {
        return jdoCacheLevel2Type;
    }

    /**
     * Set the default Level 2 Cache Type for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param type the Level 2 Cache Type
     * @since 1.1
     */
    public synchronized void setJdoCacheLevel2Type(String type)
    {
        assertConfigurable();
        jdoCacheLevel2Type = type;
    }

    /**
     * Accessor for the JDO Level 2 Cache Name
     * @return the JDO Level 2 Cache Name
     * @since 1.1
     */
    public String getJdoCacheLevel2CacheName()
    {
        return jdoCacheLevel2CacheName;
    }

    /**
     * Set the default Level 2 Cache Name for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param name Name of the cache
     * @since 1.1
     */
    public synchronized void setJdoCacheLevel2CacheName(String name)
    {
        assertConfigurable();
        jdoCacheLevel2CacheName = name;
    }

    /**
     * Accessor for the JDO Level 2 Configuration File
     * @return the JDO Level 2 Configuration File
     * @since 1.1
     */
    public String getJdoCacheLevel2ConfigurationFile()
    {
        return jdoCacheLevel2ConfigurationFile;
    }

    /**
     * Set the default Level 2 Configuration File for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param confFile The File Name of the configuration file
     * @since 1.1
     */
    public synchronized void setJdoCacheLevel2ConfigurationFile(String confFile)
    {
        assertConfigurable();
        jdoCacheLevel2ConfigurationFile = confFile;
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param isolationLevelName
     *      One of the values "none", "read uncommitted", "read committed", "repeatable
     *      read", or "serializable".  The name is case-insensitive, and spaces
     *      and underscores are equivalent.
     */
    protected synchronized void setTransactionIsolation(String isolationLevelName)
    {
        assertConfigurable();

        String iln = isolationLevelName.trim().replace(' ', '_').toUpperCase();
        isolationLevel = JDBCUtils.getJDBCIsolationLevelForName(iln);
        if (isolationLevel < 0)
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", TRANSACTION_ISOLATION_PROPERTY, isolationLevelName));
        }
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param isolationLevel the default transaction isolation level.
     */
    public synchronized void setTransactionIsolation(int isolationLevel)
    {
        assertConfigurable();

        switch (isolationLevel)
        {
            case Connection.TRANSACTION_NONE:
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", TRANSACTION_ISOLATION_PROPERTY, "" + isolationLevel));
        }

        this.isolationLevel = isolationLevel;
    }

    /**
     * Get the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default transaction isolation level.
     */
    public int getTransactionIsolation()
    {
        return isolationLevel;
    }

    /**
     * Accessor for the suffix for table identifiers
     * @param value the suffix for table identifiers
     */
    public synchronized void setPropertiesFile(String value)
    {
        assertConfigurable();
        if (value == null)
        {
            return;
        }
        if (this.propertiesFile != null)
        {
            // Already have a properties file read in
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyFileAlreadyLoaded", this.propertiesFile, value));
            return;
        }

        Properties  props=new Properties();

        //try to load the PMF properties file
        File file = new File(value);
        if( file.exists() )
        {
            this.propertiesFile = file;
            try
            {
                props.load(new FileInputStream(file));
            }
            catch (FileNotFoundException e)
            {
                this.propertiesFile = null;
                throw new JDOFatalUserException(LOCALISER.msg("PMF.PropertyFileLoadError", value), e);
            }
            catch (IOException e)
            {
                this.propertiesFile = null;
                throw new JDOFatalUserException(LOCALISER.msg("PMF.PropertyFileLoadError", value), e);
            }
        }
        setOptions(props);
    }

    /**
     * Accessor for the properties file that is loaded by this
     * @return the properties file that is loaded by this
     */
    public File getPropertiesFile()
    {
        return propertiesFile;
    }

    /**
     * Accessor for the JDO MetaData file extension.
     * @return Returns the JDO metadata File Extension.
     */
    public String getJdoMetaDataFileExtension()
    {
        return jdoMetaDataFileExtension;
    }

    /**
     * Mutator for the JDO MetaData file extension.
     * @param metadataFileExtension The JDO metadata File Extension to set.
     */
    public void setJdoMetaDataFileExtension(String metadataFileExtension)
    {
        this.jdoMetaDataFileExtension = metadataFileExtension;
    }

    /**
     * Accessor for the ORM MetaData file extension.
     * @return Returns the ORM metadata File Extension.
     */
    public String getOrmMetaDataFileExtension()
    {
        return ormMetaDataFileExtension;
    }

    /**
     * Mutator for the ORM MetaData file extension.
     * @param metadataFileExtension The ORM metadata File Extension to set.
     */
    public void setOrmMetaDataFileExtension(String metadataFileExtension)
    {
        this.ormMetaDataFileExtension = metadataFileExtension;
    }

    /**
     * Accessor for the JDO Query MetaData file extension.
     * @return Returns the JDO Query metadata File Extension.
     */
    public String getJdoqueryMetaDataFileExtension()
    {
        return jdoqueryMetaDataFileExtension;
    }

    /**
     * Mutator for the JDO Query MetaData file extension.
     * @param metadataFileExtension The JDO Query metadata File Extension to set.
     */
    public void setJdoqueryMetaDataFileExtension(String metadataFileExtension)
    {
        this.jdoqueryMetaDataFileExtension = metadataFileExtension;
    }

    /**
     * Accessor for whether to validate the metadata.
     * @return Returns whether to validate the metadata
     */
    public boolean getMetaDataValidate()
    {
        return metadataValidate;
    }

    /**
     * Mutator for whether to validate the metadata
     * @param validate Whether to validate the metadata
     */
    public void setMetaDataValidate(boolean validate)
    {
        this.metadataValidate = validate;
    }

    /**
     * Mutator for the {@link PluginRegistry} class name. If one is provided, it will be used as registry for plug-ins
     * @return the fully qualified class {@link PluginRegistry}
     */
    public String getPluginRegistryClassName()
    {
        return pluginRegistryClassName;
    }

    /**
     * Mutator for the {@link PluginRegistry} class name. If one is provided, it will be used as registry for plug-ins
     * @param pluginRegistryClassName the fully qualified class {@link PluginRegistry}
     */
    public void setPluginRegistryClassName(String pluginRegistryClassName)
    {
        this.pluginRegistryClassName = pluginRegistryClassName;
    }
    
    /**
     * Asserts that a change to a configuration property is allowed.
     */
    protected void assertConfigurable()
    {
        if (!configurable)
        {
            throw new JDOUserException(LOCALISER.msg("PMF.NotConfigurableAfterGeneration"));
        }
    }
    
    /**
     * Initialize the PROPERTY_SETTERS Map.
     * @return  The PROPERTY_SETTERS Map.
     */
    protected Map initPropertySetters()
    {
        final Map map = new HashMap();

        // JDO Standard properties
        map.put(PMFConfiguration.JDO_OPTIMISTIC_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setOptimistic(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_RETAINVALUES_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setRetainValues(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_RESTOREVALUES_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setRestoreValues(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_IGNORECACHE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIgnoreCache(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_DETACHALLONCOMMIT_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDetachAllOnCommit(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.JDO_NONTRANSACTIONAL_READ_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setNontransactionalRead(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_NONTRANSACTIONAL_WRITE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setNontransactionalWrite(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_MULTITHREADED_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setMultithreaded(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionUserName(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionPassword(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionDriverName(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionURL(s);
            }
        });
        map.put(PMFConfiguration.JDO_CONNECTION_FACTORY_NAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionFactoryName(s);
            }
        });
        map.put(PMFConfiguration.JDO_CONNECTION_FACTORY2_NAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionFactory2Name(s);
            }
        });
        map.put(PMFConfiguration.JDO_CONNECTION_FACTORY_PROPERTY, new ObjectPropertySetter()
        {
            public void set(PMFConfiguration configuration, Object obj)
            {
                configuration.setConnectionFactory(obj);
            }
        });
        map.put(PMFConfiguration.JDO_CONNECTION_FACTORY2_PROPERTY, new ObjectPropertySetter()
        {
            public void set(PMFConfiguration configuration, Object obj)
            {
                configuration.setConnectionFactory2(obj);
            }
        });
        map.put(PMFConfiguration.JDO_MAPPING_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setMapping(s);
            }
        });
        map.put(PMFConfiguration.JDO_MAPPING_CATALOG_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setCatalog(s);
            }
        });
        map.put(PMFConfiguration.JDO_MAPPING_SCHEMA_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setSchema(s);
            }
        });
        map.put(PMFConfiguration.JDO_TRANSACTION_TYPE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setTransactionType(s);
            }
        });

        // JPOX Extension properties
        map.put(PMFConfiguration.PERSISTENCE_UNIT_NAME, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPersistenceUnitName(s);
            }
        });

        map.put(PMFConfiguration.DETACH_ON_CLOSE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDetachOnClose(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.DATASTORE_ADAPTER_CLASSNAME_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDatastoreAdapterClassName(s);
                }
            });
        map.put(PMFConfiguration.STATEMENT_BATCH_LIMIT_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                try
                {
                    int len = (new Integer(s)).intValue();
                    configuration.setStatementBatchLimit(len);
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing
                }
            }
        });
        map.put(PMFConfiguration.CHECK_EXIST_TABLES_VIEWS_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setCheckExistTablesOrViews(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.VALIDATE_TABLES_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setValidateTables(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.VALIDATE_COLUMNS_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setValidateColumns(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.ADD_UNIQUE_CONSTRAINT_MAP_INVERSE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setUniqueConstraintsMapInverse(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setValidateConstraints(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_SCHEMA_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateSchema(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_TABLES_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateTables(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateColumns(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateConstraints(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_WARN_ON_ERROR_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setAutoCreateWarnOnError(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.INITILIAZE_PRIMARY_KEY_COLUMN_INFO, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setInitializePrimaryKeyColumnInfo(Boolean.valueOf(s).booleanValue());
                }
            });

        map.put(PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanism(s);
            }
        });
        map.put(PMFConfiguration.AUTO_START_MECHANISM_XML_FILE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanismXmlFile(s);
            }
        });
        map.put(PMFConfiguration.AUTO_START_MECHANISM_MODE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanismMode(s);
            }
        });
        map.put(PMFConfiguration.AUTO_START_CLASS_NAMES_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setAutoStartClassNames(s);
                }
            });

        map.put(PMFConfiguration.READ_ONLY_SCHEMA_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setReadOnlySchema(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.FIXED_SCHEMA_PROPERTY, new StringPropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setFixedSchema(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.READ_ONLY_DATASTORE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setReadOnlyDatastore(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.FIXED_DATASTORE_PROPERTY, new StringPropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setFixedDatastore(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.FOREIGN_KEY_CREATE_MODE_PROPERTY, new StringPropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setForeignKeyCreateMode(s);
            }
        });
        map.put(PMFConfiguration.DELETION_POLICY_PROPERTY, new StringPropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setDeletionPolicy(s);
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_FACTORY_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIdentifierFactory(s);
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_CASE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIdentifierCase(s);
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_WORD_SEPARATOR_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIdentifierWordSeparator(s);
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_TABLE_PREFIX_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIdentifierTablePrefix(s);
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_TABLE_SUFFIX_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIdentifierTableSuffix(s);
            }
        });

        map.put(PMFConfiguration.CONNECTION_POOLING_TYPE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionPoolingType(s);
            }
        });
        map.put(PMFConfiguration.CONNECTION_POOLING_CONFIGURATION_FILE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionPoolingConfigurationFile(s);
            }
        });

        map.put(PMFConfiguration.DEFAULT_INHERITANCE_STRATEGY_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDefaultInheritanceStrategy(s);
                }
            });
        map.put(PMFConfiguration.USE_UPDATE_LOCK_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setUseUpdateLock(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.DATASTORE_DELAY_OPERATIONS_UNTIL_COMMIT_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setDatastoreDelayOperationsEnabled(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.DATASTORE_CLASS_ADDITION_MAX_RETRIES_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                try
                {
                    int len = (new Integer(s)).intValue();
                    configuration.setDatastoreClassAdditionMaxRetries(len);
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing
                }
            }
        });
        map.put(PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setTransactionIsolation(s);
            }
        });
        map.put(PMFConfiguration.POID_TRANSACTION_ATTRIBUTE_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPoidTransactionAttribute(s);
            }
        });
        map.put(PMFConfiguration.POID_TRANSACTION_ISOLATION_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPoidTransactionIsolation(s);
            }
        });

        map.put(PMFConfiguration.PERSISTENCE_BY_REACHABILITY_AT_COMMIT, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPersistenceByReachabilityAtCommit(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.MAX_FETCH_DEPTH_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setMaxFetchDepth(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.CLASS_LOADER_RESOLVER_NAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setClassLoaderResolverName(s);
            }
        });
        map.put(PMFConfiguration.IMPLEMENTATION_CREATOR_NAME_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setImplementationCreatorName(s);
            }
        });

        // Query
        map.put(PMFConfiguration.QUERY_USE_FETCH_PLAN_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        boolean use = (new Boolean(s)).booleanValue();
                        configuration.setQueryUseFetchPlan(use);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.QUERY_TIMEOUT_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setQueryTimeout(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.QUERY_FETCH_SIZE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setQueryFetchSize(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.QUERY_FETCH_DIRECTION_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryFetchDirection(s);
                }
            });
        map.put(PMFConfiguration.QUERY_RESULT_SET_TYPE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryResultSetType(s);
                }
            });
        map.put(PMFConfiguration.QUERY_RESULT_SET_CONCURRENCY_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryResultSetConcurrency(s);
                }
            });
        map.put(PMFConfiguration.QUERY_USE_INNER_JOINS_ONLY_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryUseInnerJoinsOnly(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.QUERY_ALLOW_ALL_SQL_STATEMENTS, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        boolean allow = (new Boolean(s)).booleanValue();
                        configuration.setQueryAllowAllSQLStatements(allow);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });

        // Types
        map.put(PMFConfiguration.STRING_DEFAULT_LENGTH_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setStringDefaultLength(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.DATE_TIMEZONE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDateTimezone(s);
                }
            });

        // Cache
        map.put(PMFConfiguration.CACHE_COLLECTIONS_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setJdoCacheCollections(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.CACHE_COLLECTIONS_LAZY_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    boolean lazy = Boolean.valueOf(s).booleanValue();
                    if (lazy)
                    {
                        configuration.setJdoCacheCollectionsLazy(Boolean.TRUE);
                    }
                    else
                    {
                        configuration.setJdoCacheCollectionsLazy(Boolean.FALSE);
                    }
                }
            });
        map.put(PMFConfiguration.CACHE_LEVEL_1_TYPE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel1Type(s);
                }
            });
        map.put(PMFConfiguration.CACHE_LEVEL_2_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.CACHE_LEVEL_2_TYPE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2Type(s);
                }
            });
        map.put(PMFConfiguration.CACHE_LEVEL_2_CACHE_NAME_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2CacheName(s);
                }
            });
        map.put(PMFConfiguration.CACHE_LEVEL_2_CONFIGURATION_FILE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2ConfigurationFile(s);
                }
            });

        map.put(PMFConfiguration.METADATA_JDO_FILE_EXTENSION_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setJdoMetaDataFileExtension(s);
            }
        });
        map.put(PMFConfiguration.METADATA_ORM_FILE_EXTENSION_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setOrmMetaDataFileExtension(s);
                }
            });
        map.put(PMFConfiguration.METADATA_JDOQUERY_FILE_EXTENSION_PROPERTY, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setJdoqueryMetaDataFileExtension(s);
            }
        });
        map.put(PMFConfiguration.METADATA_VALIDATE_PROPERTY, new StringPropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setMetaDataValidate(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.PROPERTIES_FILE, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPropertiesFile(s);
            }
        });
        map.put(PMFConfiguration.PLUGIN_REGISTRY_CLASS_NAME, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPluginRegistryClassName(s);
            }
        });

        return map;
    }
 
    /**
     * Set the options for this PersistenceManagerFactory based on the
     * given Properties.
     * @param props  The Properties to set the options from.
     */
    protected void setOptions(Properties props)
    {
        this.options = props;
        for (Enumeration e = props.propertyNames(); e.hasMoreElements();)
        {
            String prop = (String) e.nextElement();

            StringPropertySetter setter = (StringPropertySetter) PROPERTY_SETTERS.get(prop);
            if (prop != null && setter != null)
            {
                String value = props.getProperty(prop);
                setter.set(this, value);
            }
            else if (prop != null && setter == null)
            {
                // Omit the ones that we dont have setters for
                if (!prop.equals("javax.jdo.PersistenceManagerFactoryClass"))
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.UnknownProperty",prop));
                }
            }
        }
    }
 
    /**
     * Set the options for this PersistenceManagerFactory based on the
     * given map of properties.
     * @param props The Properties to set the options from.
     */
    protected void setOptions(Map props)
    {
        this.options = props;
        Set keys = props.keySet();
        Iterator keyIter = keys.iterator();
        while (keyIter.hasNext())
        {
            Object keyObj = keyIter.next();
            if (keyObj instanceof String)
            {
                String key = (String)keyObj;
                Object valueObj = props.get(keyObj);
                if (valueObj instanceof String)
                {
                    StringPropertySetter setter = (StringPropertySetter) PROPERTY_SETTERS.get(key);
                    if (key != null && setter != null)
                    {
                        setter.set(this, (String)valueObj);
                    }
                    else if (key != null && setter == null)
                    {
                        // Omit the ones that we dont have setters for
                        if (!key.equals("javax.jdo.PersistenceManagerFactoryClass"))
                        {
                            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.UnknownProperty", key));
                        }
                    }
                }
                else
                {
                    ObjectPropertySetter setter = (ObjectPropertySetter) PROPERTY_SETTERS.get(key);
                    if (key != null && setter != null)
                    {
                        setter.set(this, valueObj);
                    }
                    else if (key != null && setter == null)
                    {
                        // Omit the ones that we dont have setters for
                        if (!key.equals("javax.jdo.PersistenceManagerFactoryClass"))
                        {
                            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.UnknownProperty", key));
                        }
                    }
                }
            }
            else
            {
                // Ignore this key since we only accept String keys
            }
        }
    }    
    
    /**
     * Get the options for this PersistenceManagerFactory 
     * @return the PersistenceManagerFactory 
     */
    protected Map getOptions()
    {
        return options;
    }    
}