/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 

Contributors:
2004 Andy Jefferson - added MetaDataManager
2006 Andy Jefferson - added ClassLoaderResolver
    ...
**********************************************************************/
package org.jpox;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalUserException;

import org.jpox.api.ApiAdapter;
import org.jpox.api.ApiAdapterFactory;
import org.jpox.management.ManagementManager;
import org.jpox.metadata.JDOMetaDataManager;
import org.jpox.metadata.JPAMetaDataManager;
import org.jpox.metadata.MetaDataManager;
import org.jpox.plugin.Extension;
import org.jpox.plugin.PluginManager;
import org.jpox.store.StoreManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Represents the context of a PMF instance.
 *
 * @version $Revision: 1.32 $ 
 */
public class PMFContext
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** ImplementationCreator for any persistent interfaces. */
    private ImplementationCreator ic;

    /** Hash for StoreManagers, key: user+password */
    private Map storeMgrs = new HashMap();

    /** MetaDataManager for handling the MetaData for this PMF. */
    private MetaDataManager metaDataManager = null;

    /** The PMF Configuration. */
    private final PMFConfiguration pmfConfiguration;

    /** Manager for JMX features. */
    private ManagementManager mgmtManager = null;

    /** Manager for Plug-ins **/
    private final PluginManager pluginManager;
    
    /** Manager for types and JavaTypeMappings **/
    private final TypeManager typeManager;
    
    /** the ApiAdapter to be used by the persistence engine **/
    private ApiAdapter apiAdapter;
    
    /** api name **/
    private String apiName = "JDO";

    /** Map of the ClassLoaderResolver, keyed by the clr class and the primaryLoader name. */
    private Map classLoaderResolverMap = new HashMap();

    /** Name of the class providing the ClassLoaderResolver. */
    private String classLoaderResolverClassName = null;

    /**
     * Constructor for the context.
     * @param pmfConfiguration The PMF configuration
     */
    public PMFContext(PMFConfiguration pmfConfiguration)
    {
        this.pmfConfiguration = pmfConfiguration;

        // We use JDOClassLoaderResolver here since we need the plugin mechanism before being able to create our specified CLR
        ClassLoaderResolver clr = new JDOClassLoaderResolver();
        this.pluginManager = new PluginManager(this.pmfConfiguration,clr);
        this.pluginManager.registerExtensionPoints();
        this.pluginManager.registerExtensions();

        this.typeManager = new TypeManager(this.pluginManager, getClassLoaderResolver(null));
        ApiAdapterFactory.getInstance().initializeApiAdapterExtensions(getClassLoaderResolver(null), this.pluginManager);
        apiAdapter = ApiAdapterFactory.getInstance().getApiAdapter(apiName);
    }

    /**
     * Clear out resources
     */
    public synchronized void close()
    {
        for (Iterator it = storeMgrs.values().iterator(); it.hasNext(); )
        {
            ((StoreManager)it.next()).close();
        }
        storeMgrs.clear();
        storeMgrs = null;

        if (metaDataManager != null)
        {
            metaDataManager.close();
            metaDataManager = null;
        }

        if (mgmtManager != null)
        {
            mgmtManager.close();
            mgmtManager = null;
        }

        classLoaderResolverMap.clear();
        classLoaderResolverMap = null;
        classLoaderResolverClassName = null;
    }

    /**
     * Accessor for the management manager.
     * @return The management manager
     */
    public ManagementManager getManagement()
    {
        if (mgmtManager == null)
        {
            // Create if not already present
            mgmtManager = new ManagementManager(this);
        }
        return mgmtManager;
    }

    /**
     * Accessor for a ClassLoaderResolver to use in resolving classes.
     * @param primaryLoader Loader to use as the primary loader.
     * @return The ClassLoader resolver
     */
    public ClassLoaderResolver getClassLoaderResolver(ClassLoader primaryLoader)
    {
// Commented out since it is possible to find the JDOClassLoaderResolver and then the users specified loader
//      if (classLoaderResolverClassName == null)
//      {
            // See what the PMF has been specified to use
            String clrName = pmfConfiguration.getClassLoaderResolverName();
            classLoaderResolverClassName = pluginManager.getAttributeValueForExtension("org.jpox.classloader_resolver", 
                "name", clrName, "class-name");
            if (classLoaderResolverClassName == null)
            {
                // User has specified a classloader_resolver plugin that has not registered
                throw new JDOFatalUserException(LOCALISER.msg("ClassLoaderResolver.PluginNotFound", clrName));
            }
//      }

        // Set the key we will refer to this loader by
        String key = classLoaderResolverClassName;
        if (primaryLoader != null)
        {
            key += ":[" + StringUtils.toJVMIDString(primaryLoader) + "]"; 
        }

        // See if we have the loader cached
        ClassLoaderResolver clr = (ClassLoaderResolver)classLoaderResolverMap.get(key);
        if (clr == null)
        {
            // Create the ClassLoaderResolver of this type with this primary loader
            try
            {
                Class cls = Class.forName(classLoaderResolverClassName);
                Class[] ctrArgs = null;
                Object[] ctrParams = null;
                if (primaryLoader != null)
                {
                    ctrArgs = new Class[] {ClassLoader.class};
                    ctrParams = new Object[] {primaryLoader};
                }
                Constructor ctor = cls.getConstructor(ctrArgs);
                clr = (ClassLoaderResolver)ctor.newInstance(ctrParams);
            }
            catch (ClassNotFoundException cnfe)
            {
                throw new JDOFatalUserException(LOCALISER.msg("ClassLoaderResolver.PluginClassNotFound", classLoaderResolverClassName), cnfe);
            }
            catch (Exception e)
            {
                throw new JDOFatalUserException(LOCALISER.msg("ClassLoaderResolver.Error", classLoaderResolverClassName), e);
            }
            classLoaderResolverMap.put(key, clr);
        }

        return clr;
    }

    /**
     * Accessor for the implementation creator for this context.
     * @return The implementation creator
     */
    public ImplementationCreator getImplementationCreator()
    {
        if (ic == null)
        {
            // Create if not already present
            try
            {
                Class cls = null;

                //if provided as PMF setting
                if (pmfConfiguration.getImplementationCreatorName() != null)
                {
                    cls = Class.forName(pmfConfiguration.getImplementationCreatorName(), true, 
                        AbstractPersistenceManagerFactory.class.getClassLoader());
                    ic = (ImplementationCreator)cls.newInstance();
                }
                // TODO This is wrong. It should pick the implementation creator of a particular name (this loads one by one and takes the last one!!)
                //if not provided as PMF setting search for extension point extensions
                if (ic == null)
                {
                    Extension[] exts = getPluginManager().getExtensionPoint("org.jpox.implementation_creator").getExtensions();
                    for (int i=0; ic==null && i<exts.length; i++)
                    {
                        for (int j=0; ic==null && j<exts[i].getConfigurationElements().length; j++)
                        {
                            cls = Class.forName(exts[i].getConfigurationElements()[j].getAttribute("class-name"), true, 
                                AbstractPersistenceManagerFactory.class.getClassLoader());
                            ic = (ImplementationCreator)cls.newInstance();
                        }
                    }
                }
                if (JPOXLogger.GENERAL.isDebugEnabled())
                {
                    if (ic == null)
                    {
                        JPOXLogger.GENERAL.debug(LOCALISER.msg("PMF.ImplementationCreatorNotConfigured"));
                    }
                    else
                    {
                        JPOXLogger.GENERAL.debug(LOCALISER.msg("PMF.ImplementationCreatorInUse",cls.getName()));
                    }
                }
            }
            catch (InstantiationException e)
            {
                throw new JDOFatalException(e.toString(),e);
            }
            catch (IllegalAccessException e)
            {
                throw new JDOFatalException(e.toString(),e);
            }
            catch (ClassNotFoundException e)
            {
                throw new JDOFatalException(LOCALISER.msg("PMF.ImplementationCreatorNotFound", pmfConfiguration.getImplementationCreatorName()),e);
            }
        }
        return ic;
    }

    /**
     * Accessor for the Meta-Data Manager.
     * @return Returns the MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        if (metaDataManager == null)
        {
            // Allocate a MetaData manager handling our type of MetaData files
            if (apiName.equalsIgnoreCase("JPA"))
            {
                metaDataManager = new JPAMetaDataManager(this);
            }
            else
            {
                metaDataManager = new JDOMetaDataManager(this);
            }
        }

        return metaDataManager;
    }

    /**
     * Accessor for the Store Managers. 
     * @return Returns the storeMgrs.
     */
    public Map getStoreMgrs()
    {
        return storeMgrs;
    }

    /**
     * Accessor for the PMF Configuration.
     * @return Returns the pmfConfiguration.
     */
    public PMFConfiguration getPmfConfiguration()
    {
        return pmfConfiguration;
    }
    
    /**
     * Accessor for the Plugin Manager
     * @return the PluginManager
     */
    public PluginManager getPluginManager()
    {
        return pluginManager;
    }
    
    /**
     * Accessor for the Type Manager
     * @return the TypeManager
     */
    public TypeManager getTypeManager()
    {
        return typeManager;
    }

    /**
     * Accessor for the ApiAdapter
     * @return the ApiAdapter
     */
    public ApiAdapter getApiAdapter()
    {
        return apiAdapter;
    }

    /**
     * Accessor for the API name
     * @return the api
     */
    public String getApi()
    {
        return apiName;
    }
    
    /**
     * Configure the API to be used
     * @param name the API name
     */
    public void setApi(String name)
    {
        this.apiName = name;
        ApiAdapter adapter = ApiAdapterFactory.getInstance().getApiAdapter(name);
        if (adapter != null)
        {
            this.apiAdapter = adapter;
        }
        else
        {
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.ApiAdapterNotFound", name));
        }
    }
}
