/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 
Contributors:
2004 Joerg Van Frantzius - changes to support a form of DDL output
2004 Erik Bengtson - dbinfo() mode
2004 Andy Jefferson - added "mapping" property to allow ORM files
    ...
**********************************************************************/
package org.jpox;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.jdo.JDOHelper;
import javax.jdo.PersistenceManagerFactory;

import org.jpox.metadata.MetaDataUtils;
import org.jpox.store.StoreManager;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.columninfo.TableInfo;
import org.jpox.util.ClassUtils;
import org.jpox.util.CommandLine;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Provides user-domain utilities for the maintenace of schemas.
 * These utilities cover the creation of a set of tables representing classes
 * specified in input JDO meta-data files. Additionally the tables representing
 * these classes can be deleted. 
 *
 * @version $Revision: 1.52 $
 */
public class SchemaTool extends PMFConfiguration
{
    /** Localiser for messages. */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** PMFContext **/
    private PMFContext pmfContext;

    /** Command Line **/
    private CommandLine cmd;

    /** default command line arguments **/
    private String[] defaultArgs;

    /** verbose **/
    private boolean verbose;

    /** help **/
    private boolean help;

    /** dump ddl file **/
    private String dumpDdl;

    /** mode **/
    private String mode;

    /** create mode **/
    public static final int SCHEMATOOL_CREATE_MODE = 1;

    /** delete mode **/
    public static final int SCHEMATOOL_DELETE_MODE = 2;

    /** validate mode **/
    public static final int SCHEMATOOL_VALIDATE_MODE = 3;

    /** database info mode **/
    public static final int SCHEMATOOL_DATABASE_INFO_MODE = 4;

    /** schema info mode **/
    public static final int SCHEMATOOL_SCHEMA_INFO_MODE = 5;

    /** schema tool mode **/
    public static final String SCHEMATOOL_OPTION_MODE = "org.jpox.schemaTool.mode";

    /** schema tool in verbose **/
    public static final String SCHEMATOOL_OPTION_VERBOSE = "org.jpox.schemaTool.verbose";

    /** properties file for schema tool **/
    public static final String SCHEMATOOL_OPTION_PROPERTIES_FILE = "org.jpox.schemaTool.propertiesFile";

    /** dump DDL to file **/
    public static final String SCHEMATOOL_OPTION_DUMPDDL = "org.jpox.schemaTool.dumpDdl";

    /** output help for schema tool **/
    public static final String SCHEMATOOL_OPTION_HELP = "org.jpox.schemaTool.help";

    /**
     * Constructor
     */
    public SchemaTool()
    {
        pmfContext = new PMFContext(this);
        cmd = new CommandLine();
        cmd.addOption("create", "create", null, LOCALISER.msg("SchemaTool.HelpOption.Create"));
        cmd.addOption("delete", "delete", null, LOCALISER.msg("SchemaTool.HelpOption.Delete"));
        cmd.addOption("validate", "validate", null, LOCALISER.msg("SchemaTool.HelpOption.Validate"));
        cmd.addOption("dbinfo", "dbinfo", null, LOCALISER.msg("SchemaTool.HelpOption.DBInfo"));
        cmd.addOption("schemainfo", "schemainfo", null, LOCALISER.msg("SchemaTool.HelpOption.SchemaInfo"));
        cmd.addOption("dumpDdl", "dumpDdl", "dumpDdl", LOCALISER.msg("SchemaTool.HelpOption.DumpDdl"));
        cmd.addOption("help", "help", null, LOCALISER.msg("SchemaTool.HelpOption.Help"));
        cmd.addOption("v", "verbose", null, "verbose output");
        cmd.addOption("props", "jpoxproperties", "props", "path to a properties file");
    }

    /**
     * Convenience method that takes the passed in jdo/class file URLs and returns a list of the class names.
     * @param jdoFiles Array of JDO metadata file URLs
     * @param classFiles Array of class file URLs
     * @return The List of class names
     * @throws Exception Thrown if some of the files passed in are erroneous
     */
    private static List getClassNamesForFileURLs(URL[] jdoFileURLs, URL[] classFileURLs)
    throws Exception
    {
        // Build up set of classes
        List classNames = null;

        // A). Extract the class names implied by the JDO files
        try
        {
            classNames = MetaDataUtils.getInstance().getOrderedClassNamesFromFileList(jdoFileURLs, null);
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.MetaDataParseError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            // Add the exception to the log
            JPOXLogger.SCHEMATOOL.error(e);

            throw e;
        }

        // B). Add on the class names for the class files
        if (classFileURLs != null)
        {
            for (int i=0;i<classFileURLs.length;i++)
            {
                String className = ClassUtils.getClassNameForFileURL(classFileURLs[i]);
                classNames.add(className);
            }
        }

        return classNames;
    }

    /**
     * Method to handle the creation of a schema's tables.
     * @param jdoFiles JDO metadata files defining the classes to be persisted (and hence the tables to be created).
     * @param classFiles Class files for classes to be persisted (and hence the tables to be created)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @param ddlFilename Name of the DDL file (optional)
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined.
     */
    public static int createSchemaTables(String[] jdoFiles, String[] classFiles, File propsFile, boolean verbose, String ddlFilename)
    throws Exception
    {
        URL[] jdoFileURLs = transformFilenamesToURLs(jdoFiles);
        URL[] classFileURLs = transformFilenamesToURLs(classFiles);
        return createSchemaTables(jdoFileURLs, classFileURLs, propsFile, verbose, ddlFilename);
    }

    /**
     * Method to handle the creation of a schema's tables.
     * @param jdoFileURLs JDO metadata file URLs defining the classes to be persisted (and hence the tables to be created).
     * @param classFileURLs Class file URLs for classes to be persisted (and hence the tables to be created)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @param ddlFilename Name of the DDL file (optional)
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined.
     */
    public static int createSchemaTables(URL[] jdoFileURLs, URL[] classFileURLs, File propsFile, boolean verbose, String ddlFilename)
    throws Exception
    {
        // Build up set of classes
        List classNames = getClassNamesForFileURLs(jdoFileURLs, classFileURLs);

        if (classNames != null && classNames.size() > 0)
        {
            // Set up properties for the data store
            PersistenceManagerFactory pmf = null;
            Properties props = null;
            if (propsFile == null)
            {
                try
                {
                    props = getPropertiesForDatastore(verbose);

                    // Respect any setting of AUTO_CREATE_CONSTRAINTS in case they dont want the constraints
                    String autoCreateConstraints = System.getProperty(PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY);
                    if (autoCreateConstraints != null)
                    {
                        addProperty(props, PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, autoCreateConstraints, verbose);
                    }
                    else
                    {
                        addProperty(props, PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, "true", verbose);
                    }

                    String autoCreateColumns = System.getProperty(PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY);
                    if (autoCreateColumns != null)
                    {
                        addProperty(props, PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, autoCreateColumns, verbose);
                    }
                    else
                    {
                        addProperty(props, PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, "false", verbose);
                    }

                    // Respect any setting of VALIDATE_TABLES
                    String validateTables = System.getProperty(PMFConfiguration.VALIDATE_TABLES_PROPERTY);
                    if (validateTables != null)
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, validateTables, verbose);
                    }
                    else
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "false", verbose);
                    }

                    // Respect any setting of VALIDATE_COLUMNS
                    String validateColumns = System.getProperty(PMFConfiguration.VALIDATE_COLUMNS_PROPERTY);
                    if (validateColumns != null)
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_COLUMNS_PROPERTY, validateColumns, verbose);
                    }
                    else
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_COLUMNS_PROPERTY, "false", verbose);
                    }

                    // Respect any setting of VALIDATE_CONSTRAINTS
                    String validateConstraints = System.getProperty(PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY);
                    if (validateConstraints != null)
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, validateConstraints, verbose);
                    }
                    else
                    {
                        addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, "false", verbose);
                    }
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    System.err.println(e.getMessage());
                    throw e;
                }
            }
            else
            {
                props = getPropertiesFromFile(propsFile);
            }

            if (ddlFilename != null)
            {
                // the tables must not be created in the DB, so do not validate (SQL is being dumped to a file)
                addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, "false", verbose);
                addProperty(props, PMFConfiguration.VALIDATE_COLUMNS_PROPERTY, "false", verbose);
                addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "false", verbose);
            }

            // Tag on the mandatory props that we must have
            addProperty(props, PMFConfiguration.AUTO_CREATE_TABLES_PROPERTY, "true", false);
            addProperty(props, PMFConfiguration.FIXED_DATASTORE_PROPERTY, "false", false);
            addProperty(props, PMFConfiguration.READ_ONLY_DATASTORE_PROPERTY, "false", false);
            addProperty(props, PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, "None", false);

            pmf = JDOHelper.getPersistenceManagerFactory(props);
            if (verbose)
            {
                System.out.println("");
            }

            // Create a PersistenceManager for this store and create the tables
            PersistenceManager pm = (PersistenceManager) pmf.getPersistenceManager();
            StoreManager storeMgr = pm.getStoreManager();

            FileWriter ddlFileWriter = null;
            if (ddlFilename != null)
            {
                // Open the DDL file for writing
                File ddlFile = StringUtils.getFileForFilename(ddlFilename);
                if (ddlFile.exists())
                {
                    ddlFile.delete();
                }
                ddlFile.createNewFile();
                ddlFileWriter = new FileWriter(ddlFile);

                SimpleDateFormat fmt = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss");
                ddlFileWriter.write("------------------------------------------------------------------\n");
                ddlFileWriter.write("-- JPOX SchemaTool " + 
                    "(version " + AbstractPersistenceManagerFactory.getVersionNumber() + ")" +
                    " ran at " + fmt.format(new java.util.Date()) +
                    "\n");
                ddlFileWriter.write("------------------------------------------------------------------\n");
                ddlFileWriter.write("-- Schema required for the following classes:-\n");
                Iterator classNameIter = classNames.iterator();
                while (classNameIter.hasNext())
                {
                    ddlFileWriter.write("--     " + classNameIter.next() + "\n");
                }
                ddlFileWriter.write("--\n");
            }

            try
            {
                String[] classNameArray = (String[])classNames.toArray(new String[classNames.size()]);
                storeMgr.addClasses(classNameArray, pm.getClassLoaderResolver(), ddlFileWriter);
            }
            finally
            {
                if (ddlFileWriter != null)
                {
                    ddlFileWriter.close();
                }
            }

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /**
     * Method to handle the deletion of a schema's tables.
     * @param jdoFiles JDO files defining the classes to be unpersisted (and hence the tables to be deleted).
     * @param classFiles Class files for classes to be persisted (and hence the tables to be created)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined.
     */
    public static int deleteSchemaTables(String[] jdoFiles, String[] classFiles, File propsFile, boolean verbose)
    throws Exception
    {
        URL[] jdoFileURLs = transformFilenamesToURLs(jdoFiles);
        URL[] jdoClassFileURLs = transformFilenamesToURLs(classFiles);
        return deleteSchemaTables(jdoFileURLs, jdoClassFileURLs, propsFile, verbose);
    }

    /**
     * Method to handle the deletion of a schema's tables.
     * @param jdoFileURLs JDO metadata file URLs defining the classes to be unpersisted (and hence the tables to be deleted).
     * @param classFileURLs Class file URLs for classes to be persisted (and hence the tables to be deleted)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined. 
     */
    public static int deleteSchemaTables(URL[] jdoFileURLs, URL[] classFileURLs, File propsFile, boolean verbose) throws Exception
    {
        // Build up set of classes
        List classNames = getClassNamesForFileURLs(jdoFileURLs, classFileURLs);

        if (classNames != null && classNames.size() > 0)
        {
            // Set up properties for the data store
            PersistenceManagerFactory pmf = null;
            Properties props = null;
            if (propsFile == null)
            {
                try
                {
                    props = getPropertiesForDatastore(verbose);
                }
                catch (Exception e)
                {
                    JPOXLogger.SCHEMATOOL.error(e.getMessage());
                    System.err.println(e.getMessage());

                    throw e;
                }
            }
            else
            {
                props = getPropertiesFromFile(propsFile);
            }

            // Add on the mandatory props that we must have
            addProperty(props, PMFConfiguration.FIXED_DATASTORE_PROPERTY, "false", false);
            addProperty(props, PMFConfiguration.READ_ONLY_DATASTORE_PROPERTY, "false", false);
            addProperty(props, PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, "None", false);

            pmf = JDOHelper.getPersistenceManagerFactory(props);
            if (verbose)
            {
                System.out.println("");
            }

            // Create a PersistenceManager for this store and delete the tables
            PersistenceManager pm = (PersistenceManager) pmf.getPersistenceManager();
            StoreManager storeMgr = pm.getStoreManager();

            String[] classNameArray = (String[])classNames.toArray(new String[classNames.size()]);
            storeMgr.addClasses(classNameArray, pm.getClassLoaderResolver(), null); // Add them to mgr first
            storeMgr.removeAllClasses(pm.getClassLoaderResolver());

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.info(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /**
     * Method to handle the validation of a schema's tables.
     * @param jdoFiles JDO files defining the classes to be validated
     * @param classFiles Class files for classes to be persisted (and hence the tables to be created)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return  Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined.
     */
    public static int validateSchemaTables(String[] jdoFiles, String[] classFiles, File propsFile, boolean verbose) throws Exception
    {
        URL[] jdoFileURLs = transformFilenamesToURLs(jdoFiles);
        URL[] jdoClassFileURLs = transformFilenamesToURLs(classFiles);
        return validateSchemaTables(jdoFileURLs, jdoClassFileURLs, propsFile, verbose);
    }

    /**
     * Method to handle the validation of a schema's tables.
     * @param jdoFileURLs JDO metadata file URLs defining the classes to be validated
     * @param classFileURLs Class file URLs for classes to be persisted (and hence the tables to be validated)
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the MetaData, or the DB definition is not defined. 
     */
    public static int validateSchemaTables(URL[] jdoFileURLs, URL[] classFileURLs, File propsFile, boolean verbose) throws Exception
    {
        // Build up set of classes
        List classNames = getClassNamesForFileURLs(jdoFileURLs, classFileURLs);

        if (classNames != null && classNames.size() > 0)
        {
            // Set up properties for the data store
            PersistenceManagerFactory pmf = null;
            Properties props = null;
            if (propsFile == null)
            {
                try
                {
                    props = getPropertiesForDatastore(verbose);
                }
                catch (Exception e)
                {
                    JPOXLogger.SCHEMATOOL.error(e.getMessage());
                    System.err.println(e.getMessage());

                    throw e;
                }
            }
            else
            {
                props = getPropertiesFromFile(propsFile);
            }
            addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "true", verbose);
            addProperty(props, PMFConfiguration.VALIDATE_COLUMNS_PROPERTY, "true", verbose);
            addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, " true", verbose);
            addProperty(props, PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, "None", false);

            pmf = JDOHelper.getPersistenceManagerFactory(props);
            if (verbose)
            {
                System.out.println("");
            }

            // Create a PersistenceManager for this store and validate
            PersistenceManager pm = (PersistenceManager) pmf.getPersistenceManager();
            StoreManager storeMgr = pm.getStoreManager();

            String[] classNameArray = (String[])classNames.toArray(new String[classNames.size()]);
            storeMgr.addClasses(classNameArray, pm.getClassLoaderResolver(), null); // Validates since we have the flags set

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /**
     * Method to output the database information.
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when the DB definition is not defined. 
     */
    public static int outputDBInfo(File propsFile, boolean verbose)
    throws Exception
    {
        // Set up properties for the data store
        PersistenceManagerFactory pmf = null;
        Properties props = null;
        if (propsFile == null)
        {
            try
            {
                props = getPropertiesForDatastore(verbose);
            }
            catch (Exception e)
            {
                System.err.println(e.getMessage());
                throw e;
            }
        }
        else
        {
            props = getPropertiesFromFile(propsFile);
        }
        addProperty(props, PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, "None", false);

        pmf = JDOHelper.getPersistenceManagerFactory(props);
        if (verbose)
        {
            System.out.println("");
        }

        // Create a PersistenceManager for this store
        RDBMSAdapter dba = (RDBMSAdapter) ((PersistenceManager) pmf.getPersistenceManager()).getStoreManager().getDatastoreAdapter();

        // Print out the DBA
        String msg = dba.toString();
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);
        System.out.println();

        // Print out the type info
        msg = LOCALISER.msg("SchemaTool.DatabaseTypeInfo");
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);

        Iterator typeInfosIter = dba.iteratorTypeInfo();
        while (typeInfosIter.hasNext())
        {
            Object typeInfo = typeInfosIter.next();
            JPOXLogger.SCHEMATOOL.info(typeInfo);
            System.out.println(typeInfo);
        }
        System.out.println("");

        // Print out the keywords info
        msg = LOCALISER.msg("SchemaTool.DatabaseKeywords");
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);

        Iterator reservedWordsIter = dba.iteratorReservedWords();
        while (reservedWordsIter.hasNext())
        {
            Object words = reservedWordsIter.next();
            JPOXLogger.SCHEMATOOL.info(words);
            System.out.println(words);
        }
        System.out.println("");

        return 0;
    }

    /**
     * Method to output the schema information.
     * @param propsFile File containing properties to use (optional)
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when the DB definition is not defined. 
     */
    public static int outputSchemaInfo(File propsFile, boolean verbose) 
    throws Exception
    {
        // Set up properties for the data store
        PersistenceManagerFactory pmf = null;
        Properties props = null;
        if (propsFile == null)
        {
            try
            {
                props = getPropertiesForDatastore(verbose);
            }
            catch (Exception e)
            {
                System.err.println(e.getMessage());
                throw e;
            }
        }
        else
        {
            props = getPropertiesFromFile(propsFile);
        }
        addProperty(props, PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, "None", false);
        pmf = JDOHelper.getPersistenceManagerFactory(props);
        if (verbose)
        {
            System.out.println("");
        }

        // Create a PersistenceManager for this store
        RDBMSAdapter dba = (RDBMSAdapter) ((PersistenceManager) pmf.getPersistenceManager()).getStoreManager().getDatastoreAdapter();

        // Print out the DBA
        String msg = dba.toString();
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);
        System.out.println();

        JPOXLogger.SCHEMATOOL.info("TABLES");
        Iterator itTables = ((RDBMSManager) ((PersistenceManager) pmf.getPersistenceManager()).getStoreManager()).getTableInfo(null).iterator();
        while (itTables.hasNext())
        {
            TableInfo ti = (TableInfo) itTables.next();
            JPOXLogger.SCHEMATOOL.info(ti);
            System.out.println(ti);
            Iterator itColumns = ((RDBMSManager) ((PersistenceManager) pmf.getPersistenceManager()).getStoreManager()).getColumnInfo(ti.getTableSchem(),
                ti.getTableName()).iterator();
            while (itColumns.hasNext())
            {
                ColumnInfo ci = (ColumnInfo) itColumns.next();
                JPOXLogger.SCHEMATOOL.info(ci);
                System.out.println(ci);
            }
        }
        System.out.println("");

        return 0;
    }

    /**
     * Utility to set up the properties defining the datastore. Uses the
     * DatabaseProperties object to obtain them from System.properties.
     * @param verbose Verbose output
     * @return The properties definition for the datastore.
     * @throws Exception Thrown when the datastore is not fully specified.
     */
    public static Properties getPropertiesForDatastore(boolean verbose) throws Exception
    {
        Properties props = new Properties();

        if (System.getProperty(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY) == null || System.getProperty(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY) == null)
        {
            //try to load a default PMF properties file
            File file = new File(System.getProperty("user.home") + "/.jdo/PMFProperties.properties");
            if (file.exists())
            {
                props.load(new FileInputStream(file));
            }
            else
            {
                throw new Exception(LOCALISER.msg("SchemaTool.DatabaseDriverNotSpecifiedError"));
            }
        }

        // Add the properties necessary
        if (verbose)
        {
            String msg = LOCALISER.msg("SchemaTool.Properties");
            JPOXLogger.SCHEMATOOL.info(msg);
            System.out.println(msg);
        }

        addProperty(props, "javax.jdo.PersistenceManagerFactoryClass", "org.jpox.PersistenceManagerFactoryImpl", verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY, System.getProperty(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_URL_PROPERTY, System.getProperty(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY, System.getProperty(PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY, System.getProperty(PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY), false);
        addProperty(props, PMFConfiguration.JDO_MAPPING_PROPERTY, System.getProperty(PMFConfiguration.JDO_MAPPING_PROPERTY), verbose);

        return props;
    }

    /**
     * Utility to generate a Properties object from the specified properties file.
     * @param propsFile Properties file
     * @return The Properties to use with the PMF
     */
    public static Properties getPropertiesFromFile(File propsFile)
    {
        Properties props = new Properties();

        try
        {
            FileInputStream fis = new FileInputStream(propsFile);
            props.load(fis);
        }
        catch (FileNotFoundException fnfe)
        {
            // Cant happen since we checked earlier
        }
        catch (IOException ioe)
        {
        }

        return props;
    }

    /**
     * Convenience utility to add a property to the specified properties list.
     * @param props The current properties list to add it to
     * @param name Name of the property
     * @param value Value of the property
     * @param verbose Whether to give verbose logging info
     */
    private static void addProperty(Properties props, String name, String value, boolean verbose)
    {
        if (name == null || value == null)
        {
            return;
        }

        props.setProperty(name, value);

        if (verbose)
        {
            String msg = LOCALISER.msg("SchemaTool.Property", name, value);
            JPOXLogger.SCHEMATOOL.info(msg);
            System.out.println(msg);
        }
    }

    /**
     * Entry method when invoked from the command line.
     * @param args List of options for processing by the available methods in this class.
     * @throws Exception
     */
    public static void main(String[] args) throws Exception
    {
        SchemaTool tool = new SchemaTool();
        tool.setCommandLineArgs(args);

        if (tool.isHelp())
        {
            System.out.println(LOCALISER.msg("SchemaTool.HelpTitle"));
            System.out.println(LOCALISER.msg("SchemaTool.HelpCommand"));
            System.out.println(LOCALISER.msg("SchemaTool.HelpOptions"));
            System.out.println(tool.cmd.toString());
            System.out.println(LOCALISER.msg("SchemaTool.HelpJDOFiles"));
            System.exit(0);
        }

        // Mode of operation
        int mode = SCHEMATOOL_CREATE_MODE;
        if (tool.getMode() != null)
        {
            if (tool.getMode().equals("create"))
            {
                mode = SCHEMATOOL_CREATE_MODE;
                System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaCreation", AbstractPersistenceManagerFactory.getVersionNumber()));
            }
            else if (tool.getMode().equals("delete"))
            {
                mode = SCHEMATOOL_DELETE_MODE;
                System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaDeletion", AbstractPersistenceManagerFactory.getVersionNumber()));
            }
            else if (tool.getMode().equals("validate"))
            {
                mode = SCHEMATOOL_VALIDATE_MODE;
                System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaValidation", AbstractPersistenceManagerFactory.getVersionNumber()));
            }
            else if (tool.getMode().equals("dbinfo"))
            {
                mode = SCHEMATOOL_DATABASE_INFO_MODE;
                System.out.println(LOCALISER.msg("SchemaTool.Title.DBInfo", AbstractPersistenceManagerFactory.getVersionNumber()));
            }
            else if (tool.getMode().equals("schemainfo"))
            {
                mode = SCHEMATOOL_SCHEMA_INFO_MODE;
                System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaInfo", AbstractPersistenceManagerFactory.getVersionNumber()));
            }
        }
        System.out.println();

        if (tool.isVerbose())
        {
            // Verbose output block
            // Classpath
            String msg = LOCALISER.msg("SchemaTool.Classpath");
            JPOXLogger.SCHEMATOOL.info(msg);
            System.out.println(msg);
            StringTokenizer tokeniser = new StringTokenizer(System.getProperty("java.class.path"), File.pathSeparator);
            while (tokeniser.hasMoreTokens())
            {
                String entry = LOCALISER.msg("SchemaTool.Classpath.Entry", tokeniser.nextToken());
                JPOXLogger.SCHEMATOOL.info(entry);
                System.out.println(entry);
            }
            System.out.println();

            // Properties file
            if (tool.getPropertiesFile() != null)
            {
                System.out.println(LOCALISER.msg("SchemaTool.PropertyFile", tool.getPropertiesFile()));
                System.out.println();
            }

            // DDL file
            if (tool.getDumpDdl() != null)
            {
                System.out.println(LOCALISER.msg("SchemaTool.OutputDDL", tool.getDumpDdl()));
                System.out.println();
            }

            // Input files
            System.out.println(LOCALISER.msg("SchemaTool.InputJdoFiles"));
            for (int i = 0; i < tool.getDefaultArgs().length; i++)
            {
                String jdo_file = tool.getDefaultArgs()[i];
                System.out.println(">> " + jdo_file);
            }
            System.out.println();
        }

        // Check for no input files
        if (mode != SCHEMATOOL_SCHEMA_INFO_MODE && mode != SCHEMATOOL_DATABASE_INFO_MODE)
        {
            if (tool.getDefaultArgs().length == 0)
            {
                String msg = LOCALISER.msg("SchemaTool.NoMetaDataFilesError");
                JPOXLogger.SCHEMATOOL.error(msg);
                System.err.println(msg);
                System.exit(3);
            }
        }

        // Set up DDL dumping (optional)
        String ddlFilename = tool.getDumpDdl();

        // Split the input files into jdo/class.
        // TODO Detect non-jdo/class files
        String[] inputFiles = tool.getDefaultArgs();
        String[] jdoFiles = null;
        String[] classFiles = null;
        if (inputFiles != null)
        {
            int numberOfJDOFiles = 0;
            int numberOfClassFiles = 0;
            for (int i=0;i<inputFiles.length;i++)
            {
                if (inputFiles[i].endsWith(".class"))
                {
                    numberOfClassFiles++;
                }
                else
                {
                    // Assume JDO file (but could be anything!)
                    numberOfJDOFiles++;
                }
            }
            jdoFiles = new String[numberOfJDOFiles];
            int num = 0;
            for (int i=0;i<inputFiles.length;i++)
            {
                if (!inputFiles[i].endsWith(".class"))
                {
                    jdoFiles[num++] = inputFiles[i];
                }
            }
            classFiles = new String[numberOfClassFiles];
            num = 0;
            for (int i=0;i<inputFiles.length;i++)
            {
                if (inputFiles[i].endsWith(".class"))
                {
                    classFiles[num++] = inputFiles[i];
                }
            }
        }

        // Run SchemaTool
        int code = 0;
        try
        {
            if (mode == SCHEMATOOL_CREATE_MODE)
            {
                code = createSchemaTables(jdoFiles, classFiles, tool.getPropertiesFile(), tool.isVerbose(), ddlFilename);
            }
            else if (mode == SCHEMATOOL_DELETE_MODE)
            {
                code = deleteSchemaTables(jdoFiles, classFiles, tool.getPropertiesFile(), tool.isVerbose());
            }
            else if (mode == SCHEMATOOL_VALIDATE_MODE)
            {
                code = validateSchemaTables(jdoFiles, classFiles, tool.getPropertiesFile(), tool.isVerbose());
            }
            else if (mode == SCHEMATOOL_DATABASE_INFO_MODE)
            {
                code = outputDBInfo(tool.getPropertiesFile(), tool.isVerbose());
            }
            else if (mode == SCHEMATOOL_SCHEMA_INFO_MODE)
            {
                code = outputSchemaInfo(tool.getPropertiesFile(), tool.isVerbose());
            }
            if (code == 0)
            {
                String msg = LOCALISER.msg("SchemaTool.Success");
                JPOXLogger.SCHEMATOOL.info(msg);
                System.out.println(msg);
            }
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.Error", e.getMessage());
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            // Log the exception
            JPOXLogger.SCHEMATOOL.error(e);

            code = 2;
        }

        System.exit(code);
    }

    /**
     * Convenience method that converts filenames to URLs.
     * @param filenames the filenames
     * @return The URLs
     * @throws MalformedURLException
     */
    private static URL[] transformFilenamesToURLs(String[] filenames) throws MalformedURLException
    {
        URL[] urls = new URL[filenames.length];
        for (int i = 0; i < filenames.length; i++)
        {
            urls[i] = new URL("file:" + filenames[i]);
        }
        return urls;
    }

    /**
     * Initialize the command line arguments
     * @param args
     */
    public void setCommandLineArgs(String[] args)
    {
        cmd.parse(args);

        defaultArgs = cmd.getDefaultArgs();

        // populate options
        Map map = new HashMap();

        if (cmd.hasOption("create"))
        {
            map.put(SCHEMATOOL_OPTION_MODE, "create");
        }
        else if (cmd.hasOption("delete"))
        {
            map.put(SCHEMATOOL_OPTION_MODE, "delete");
        }
        else if (cmd.hasOption("validate"))
        {
            map.put(SCHEMATOOL_OPTION_MODE, "validate");
        }
        else if (cmd.hasOption("dbinfo"))
        {
            map.put(SCHEMATOOL_OPTION_MODE, "dbinfo");
        }
        else if (cmd.hasOption("schemainfo"))
        {
            map.put(SCHEMATOOL_OPTION_MODE, "schemainfo");
        }
        if (cmd.hasOption("help"))
        {
            map.put(SCHEMATOOL_OPTION_HELP, Boolean.TRUE.toString());
        }
        if (cmd.hasOption("dumpDdl"))
        {
            map.put(SCHEMATOOL_OPTION_DUMPDDL, cmd.getOptionArg("dumpDdl"));
        }
        if (cmd.hasOption("props"))
        {
            map.put(PMFConfiguration.PROPERTIES_FILE, cmd.getOptionArg("props"));
        }
        if (cmd.hasOption("v"))
        {
            map.put(SCHEMATOOL_OPTION_VERBOSE, Boolean.TRUE.toString());
        }
        setOptions(map);

    }

    /**
     * @return the help
     */
    public boolean isHelp()
    {
        return help;
    }

    /**
     * @param help the help to set
     */
    public void setHelp(boolean help)
    {
        this.help = help;
    }

    /**
     * @return the mode
     */
    public String getMode()
    {
        return mode;
    }

    /**
     * @param mode the mode to set
     */
    public void setMode(String mode)
    {
        this.mode = mode;
    }

    /**
     * @return the verbose
     */
    public boolean isVerbose()
    {
        return verbose;
    }

    /**
     * @param verbose the verbose to set
     */
    public void setVerbose(boolean verbose)
    {
        this.verbose = verbose;
    }

    /**
     * @return the defaultArgs
     */
    public String[] getDefaultArgs()
    {
        return defaultArgs;
    }

    /**
     * @return the pmfContext
     */
    public PMFContext getPmfContext()
    {
        return pmfContext;
    }

    /**
     * @return the dumpDdl
     */
    public String getDumpDdl()
    {
        return dumpDdl;
    }

    /**
     * @param dumpDdl the dumpDdl to set
     */
    public void setDumpDdl(String dumpDdl)
    {
        this.dumpDdl = dumpDdl;
    }

    /**
     * Initialize the PROPERTY_SETTERS Map.
     * @return  The PROPERTY_SETTERS Map.
     */
    protected Map initPropertySetters()
    {
        final Map map = super.initPropertySetters();

        map.put(SchemaTool.SCHEMATOOL_OPTION_DUMPDDL, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                ((SchemaTool) configuration).setDumpDdl(s);
            }
        });
        map.put(SchemaTool.SCHEMATOOL_OPTION_HELP, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                ((SchemaTool) configuration).setHelp(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(SchemaTool.SCHEMATOOL_OPTION_MODE, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                ((SchemaTool) configuration).setMode(s);
            }
        });
        map.put(SchemaTool.SCHEMATOOL_OPTION_VERBOSE, new StringPropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                ((SchemaTool) configuration).setVerbose(Boolean.valueOf(s).booleanValue());
            }
        });

        return map;
    }
}