/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
     ...
***********************************************************************/
package org.jpox.management;

import javax.jdo.JDOFatalInternalException;

import org.jpox.PMFContext;

/**
 * Management interface for JPOX. Management operations and attributes
 * are exposed through this interface that holds statistics linked to a PMF instance.
 * 
 * The mechanics for starting and stopping JMX servers are not defined here,
 * and must be done by plug-ins, by providing the implementation of
 * {@link ManagementServer}.
 * 
 * This Manager controls the lifecycle of management servers.
 * 
 * A management server is started when an instance of this class is created,
 * and its shutdown when the close operation is invoked
 * 
 * The management server startup is triggered when the Manager gets enabled.
 */
public class ManagementManager
{
    /** hold the statistics count **/
    final private long[] statistics;
    
    /** the pmfContext that owns this management **/
    final private PMFContext pmfContext;
    
    /** whether this is closed **/
    private boolean closed = false;

    /** whether this Manager is enabled to update statistics **/
    private boolean enabled = true;
    
    /** the Management Server **/
    private ManagementServer mgmtServer;
    
    /**
     * Constructor for Management
     * @param pmfContext the pmfContext that owns this management
     */
    public ManagementManager(PMFContext pmfContext)
    {
        this.statistics = new long[ManagementCategory.MANAGEMENT_CATEGORY_TOTAL];
        this.pmfContext = pmfContext;
        startManagementServer();
    }
    
    /**
     * Disable the update of statistics
     * @throws JDOFatalInternalException if the manager is closed
     */
    public void disable()
    {
        assertNotClosed();
        enabled = false;
    }
    
    /**
     * Enable the update of statistics
     * @throws JDOFatalInternalException if the manager is closed
     */
    public void enable()
    {
        assertNotClosed();
        enabled = true;
        startManagementServer();
    }

    /**
     * Whether this Manager is enabled
     * @throws JDOFatalInternalException if the manager is closed
     */
    public boolean isEnabled()
    {
        assertNotClosed();
        //DO not thread safe to increase performance
        //the trade-off here is very low
        return enabled;
    }
    
    /**
     * Increment in one the statistics of the given category.
     * 
     * Before calling this method, it's recommended to check 
     * if the Manager is enabled in order to improve performance
     * on disabled cases. This pattern is very similar to Log frameworks.
     * 
     * @param cat the management category
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized void increment(ManagementCategory cat)
    {
        assertNotClosed();
        if (!enabled)
        {
            return;
        }
        statistics[cat.getType()]++;
    }

    /**
     * Increment in one the statistics of the given categories
     * 
     * Before calling this method, it's recommended to check 
     * if the Manager is enabled in order to improve performance
     * on disabled cases. This pattern is very similar to Log frameworks.
     * 
     * @param cats the management categories
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized void increment(ManagementCategory[] cats)
    {
        assertNotClosed();
        if (!enabled)
        {
            return;
        }
        for( int i=0; i<cats.length; i++)
        {
            statistics[cats[i].getType()]++;
        }
    }
    
    /**
     * Accessor for the current statistic value for the given category
     * @param cat the management category
     * @return the current value for the statistic
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized long getStatisticValue(ManagementCategory cat)
    {
        assertNotClosed();
        return statistics[cat.getType()];
    }    
    
    /**
     * Accessor for the statistic values
     * @return the current value for the statistic
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized long[] getStatisticValues()
    {
        assertNotClosed();
        return statistics;
    }
    
    /**
     * Reset the statistics values
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized void reset()
    {
        assertNotClosed();
        for( int i=0; i<statistics.length; i++)
        {
            statistics[i] = 0;
        }
    }

    /**
     * Accessor for the PMFContext
     * @return the PMFContext
     * @throws JDOFatalInternalException if the manager is closed
     */
    public PMFContext getPmfContext()
    {
        assertNotClosed();
        return pmfContext;
    }
    
    /**
     * Start Management Server
     */
    private void startManagementServer()
    {
        if (mgmtServer==null)
        {
            //TODO get registered extension
            if (mgmtServer!=null)
            {
                mgmtServer.start();
            }
        }
    }    
    
    /**
     * Shutdown Management Server
     */
    private void stopManagementServer()
    {
        if (mgmtServer!=null)
        {
            mgmtServer.stop();
        }
    }    

    /**
     * Close a instance.
     * @throws JDOFatalInternalException if the manager is closed
     */
    public synchronized void close()
    {
        assertNotClosed();
        stopManagementServer();
        this.closed = true;
    }    
    
    /**
     * Assert that this instance is open
     */
    private void assertNotClosed()
    {
        if( this.closed )
        {
            throw new JDOFatalInternalException("Management instance is closed and cannot be used. You must adquire a new PMFContext");
        }
    }     
}

