/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.ClassNameConstants;
import org.jpox.api.ApiAdapter;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.MacroString;
import org.jpox.util.StringUtils;
import org.jpox.util.ViewUtils;

/**
 * Abstract representation of the MetaData of a class/interface. 
 * Has a parent PackageMetaData that can contain the metadata for several 
 * classes/interfaces. Is extended by ClassMetaData and InterfaceMetaData.
 * @since 1.1
 * @version $Revision: 1.49 $
 */
public abstract class AbstractClassMetaData extends ExtendableMetaData
{
    protected static Localiser LOCALISER_API=Localiser.getInstance("org.jpox.Localisation");

    /** Named queries */
    protected Collection queries = null;

    /** JoinMetaData */
    protected JoinMetaData[] joinMetaData;

    /** IndexMetaData */
    protected IndexMetaData[] indexMetaData;

    /** ForeignKeyMetaData */
    protected ForeignKeyMetaData[] foreignKeyMetaData;

    /** UniqueMetaData */
    protected UniqueMetaData[] uniqueMetaData;

    /** Identity-type tag value. */
    protected IdentityType identityType = IdentityType.DATASTORE;

    /** Class name */
    protected final String name;


    /** Entity name. Required by JPA  4.3.1 */
    protected final String entityName;
    
    /** persistence-modifier tag value. */
    protected ClassPersistenceModifier persistenceModifier = null;
    
    /** persistence-capable-superclas tag value. */
    protected String persistenceCapableSuperclass;
    
    /** objectid-class tag value. */
    protected String objectidClass;

    /** requires-extent tag value. */
    protected boolean requiresExtent = true;

    /** detachable tag value. */
    protected boolean detachable = false;

    /** embedded-only tag value. */
    protected boolean embeddedOnly = false;

    /** Catalog name (O/R mapping). */
    protected String catalog;

    /** Schema name (O/R mapping). */
    protected String schema;

    /** Table name (O/R mapping). This may be of the form "[database].[catalog].[schema].table" */
    protected String table;

    /** Full name. (e.g org.jpox.MyClass) */
    protected final String fullName;

    /** Version MetaData */
    protected VersionMetaData versionMetaData;

    /** Identity MetaData */
    protected IdentityMetaData identityMetaData;

    /** Flag whether the identity was specified by the user. */
    protected boolean identitySpecified = false;
    
    /** Inheritance MetaData */
    protected InheritanceMetaData inheritanceMetaData;

    /** PrimaryKey MetaData */
    protected PrimaryKeyMetaData primaryKeyMetaData;

    /** List of properties */
    protected List properties = new ArrayList();
    
    /** The columns that are present in the datastore yet not mapped to fields in this class. */
    protected List unmappedColumns = null;

    /** Callbacks method names keyed by callback callback type */
    protected Map callbacks = null;

    /** Listener classes names. Use a list to preserve ordering. */
    protected List listenerClassNames = null;

    /** Flag to exclude superclass listeners. */
    protected boolean excludeSuperClassListeners = false;

    /** Flag to exclude default listeners. */
    protected boolean excludeDefaultListeners = false;

    // -------------------------------------------------------------------------
    // These fields are only used when the MetaData is read by the parser and
    // elements are dynamically added to the other elements. At initialise()
    // they are cleared and nulled so shouldn't be used internally.

    /** List of joins */
    protected List joins = new ArrayList();

    /** List of fetch-group */
    protected List fetchGroups = new ArrayList();

    /** List of foreign-key */
    protected List foreignKeys = new ArrayList();

    /** List of indexes */
    protected List indexes = new ArrayList();

    /** List of uniqueConstraints */
    protected List uniqueConstraints = new ArrayList();

    /**
     * A contained fetch-group element indicates that the named group is to be
     * included in the group being defined. Nested fetch group elements are
     * limited to only the name attribute.
     */
    protected FetchGroupMetaData[] fetchGroupMetaData;

    protected Map fetchGroupMetaDataByName;

    /**
     * Class MetaData for the Persistence-Capable superclass (if any)
     */
    protected AbstractClassMetaData pcSuperclassMetaData=null;
    
    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.
 
    /** convenience PropertyMetaData **/
    protected AbstractPropertyMetaData[] propertyMetaData;    
    
    /**
     * JDO managed fields. Contains the same AbstractPropertyMetaData objects as are in
     * "fields".
     */
    protected AbstractPropertyMetaData[] managedFields;

    /**
     * Fields for superclasse sthat are overridden in this class.
     */
    protected AbstractPropertyMetaData[] overriddenFields;

    /**
     * Field numbers mapped by the name of the field.
     */
    protected Map fieldNumbersByName;

    /**
     * Field numbers of all fields (inc superclasses).
     */
    protected int[] allFieldNumbers;

    /**
     * Field numbers of the primary key fields (inc superclasses).
     */
    protected int[] primaryKeyFieldNumbers;

    /**
     * Field numbers of the non-primary key fields (inc superclasses).
     */
    protected int[] nonPrimaryKeyFieldNumbers;
    
    /**
     * Field numbers of the non-primary key fields (inc superclasses).
     */
    protected boolean[] nonPrimaryKeyFieldFlags;    

    /**
     * Field numbers of the default fetch group fields (inc superclasses).
     */
    protected int[] defaultFetchGroupFieldNumbers;

    /**
     * Flags of the default fetch group state for all fields.
     */
    protected boolean[] defaultFetchGroupFieldFlags;

    /**
     * Field numbers of the second class mutable fields (inc superclasses).
     */
    protected int[] secondClassMutableFieldNumbers;

    /**
     * Flags of the second class mutable state for all fields.
     */
    protected boolean[] secondClassMutableFieldFlags;

    /**
     * Field numbers of the PersistenceCapable fields (inc superclasses).
     */
    protected int[] persistenceCapableFieldNumbers;    

    /**
     * No of managed fields in superclasses, that are inherited by this class.
     */
    protected int noOfInheritedManagedFields=0;

    /**
     * if this PersistenceCapable class uses SingleFieldIdentity
     */
    protected boolean usesSingleFieldIdentityClass;    

    // ----------------------------- Constructors ------------------------------
 
    /**
     * Constructor.
     * Takes the basic string information found in the MetaData file.
     * @param parent The package to which this class/interface belongs
     * @param name Name of class
     * @param identityType identity-type flag
     * @param objectidClass Primary key class name
     * @param requiresExtent Whether the class requires an extent
     * @param detachable Whether the class can be detached
     * @param embeddedOnly embedded-only tag
     * @param modifier persistence-modifier tag 
     * @param catalog The catalog name default for this class
     * @param schema The schema name default for this class
     * @param table RDBMS table to store the class in
     * @param entityName the entity name required by JPA 4.3.1
     */
    protected AbstractClassMetaData(final PackageMetaData parent,
                         final String name,
                         final String identityType,
                         final String objectidClass,
                         final String requiresExtent,
                         final String detachable,
                         final String embeddedOnly,
                         final String modifier,
                         final String persistenceCapableSuperclass,
                         final String catalog,
                         final String schema,
                         final String table,
                         final String entityName)
    {
        super(parent);

        if (name == null)
        {
            throw new JDOUserException(LOCALISER.msg("MetaData.Parser.ExpectedValueInAttribute","name",getPackageName(),"class"));
        }
        if (StringUtils.isWhitespace(name))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.NameNotSpecified.Error",parent.name);
        }

        this.name = name;
        this.fullName = ClassUtils.createFullClassName(parent.name, name);
        this.objectidClass = (StringUtils.isWhitespace(objectidClass) ? null : ClassUtils.createFullClassName(parent.name, objectidClass));
        this.identityType = IdentityType.getIdentityType(identityType);
        this.requiresExtent = (requiresExtent != null && requiresExtent.equalsIgnoreCase("false")) ? false : true;
        this.detachable = (detachable != null && detachable.equalsIgnoreCase("false")) ? false : true;
        this.embeddedOnly = (embeddedOnly != null && embeddedOnly.equalsIgnoreCase("true")) ? true : false;
        this.catalog = (StringUtils.isWhitespace(catalog) ? null : catalog);
        this.schema = (StringUtils.isWhitespace(schema) ? null : schema);
        this.table = (StringUtils.isWhitespace(table) ? null : table);
        this.persistenceModifier = ClassPersistenceModifier.getClassPersistenceModifier(modifier);
        this.entityName = entityName;
        
        // "persistence-capable-superclass"
        this.persistenceCapableSuperclass = persistenceCapableSuperclass;
        if (!StringUtils.isWhitespace(persistenceCapableSuperclass))
        {
            this.persistenceCapableSuperclass =
                ClassUtils.createFullClassName(parent.name, persistenceCapableSuperclass);
        }        
    }

    /** Implementation of "persistent-interface" needing table setting from superclass. */
    protected transient boolean persistentInterfaceImplNeedingTableFromSuperclass = false;

    /** Implementation of "persistent-interface" needing table setting from subclass. */
    protected transient boolean persistentInterfaceImplNeedingTableFromSubclass = false;

    /**
     * Constructor for creating the ClassMetaData for an implementation of a "persistent-interface".
     * @param imd MetaData for the "persistent-interface"
     * @param implClassName Name of the implementation class
     * @param copyFields Whether to copy the fields of the interface too
     */
    public AbstractClassMetaData(InterfaceMetaData imd, String implClassName, boolean copyFields)
    {
        this((PackageMetaData)imd.parent, implClassName, imd.identityType.toString(), imd.objectidClass, 
            imd.requiresExtent ? "true":"false",
            imd.detachable ? "true":"false",
            imd.embeddedOnly ? "true":"false",
            ClassPersistenceModifier.PERSISTENCE_CAPABLE.toString(),
            null, // Ignore class inheritance since we are just implementing this interface
            imd.catalog,
            imd.schema,
            imd.table,
            imd.entityName);

        if (copyFields)
        {
            copyFieldsFromInterface(imd);
        }

        setVersionMetaData(imd.versionMetaData);
        setIdentityMetaData(imd.identityMetaData);
        setPrimaryKeyMetaData(imd.primaryKeyMetaData);

        if (imd.inheritanceMetaData != null)
        {
            if (imd.inheritanceMetaData.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
            {
                // Flag the table as requiring setting based on the next superclass
                persistentInterfaceImplNeedingTableFromSuperclass = true;
            }
            else if (imd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUBCLASS_TABLE)
            {
                // Flag the table as requiring setting based on the next subclass
                persistentInterfaceImplNeedingTableFromSubclass = true;
            }

            InheritanceMetaData inhmd = new InheritanceMetaData(this,InheritanceStrategy.NEW_TABLE.toString());
            if( imd.inheritanceMetaData.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE )
            {
                AbstractClassMetaData acmd = imd.getSuperAbstractClassMetaData();
                while( acmd != null )
                {
                    if( acmd.getInheritanceMetaData() != null )
                    {
                        if (acmd.getInheritanceMetaData().getStrategyValue()==InheritanceStrategy.NEW_TABLE)
                        {
                            if( acmd.getInheritanceMetaData().getDiscriminatorMetaData() != null )
                            {
                                inhmd.setDiscriminatorMetaData(new DiscriminatorMetaData(inhmd,acmd.getInheritanceMetaData().getDiscriminatorMetaData()));
                            }
                            inhmd.setJoinMetaData(acmd.getInheritanceMetaData().getJoinMetaData());
                            break;
                        }
                    }
                    acmd = acmd.getSuperAbstractClassMetaData();
                }
            }
            else if( imd.inheritanceMetaData.getStrategyValue() == InheritanceStrategy.NEW_TABLE )
            {
            	if( imd.getInheritanceMetaData().getDiscriminatorMetaData() != null )
            	{
            		inhmd.setDiscriminatorMetaData(new DiscriminatorMetaData(inhmd,imd.getInheritanceMetaData().getDiscriminatorMetaData()));
            	}
            	inhmd.setJoinMetaData(imd.getInheritanceMetaData().getJoinMetaData());
            }
            setInheritanceMetaData(inhmd);
        }
        
        if (imd.joinMetaData != null)
        {
            for (int i=0;i<imd.joinMetaData.length;i++)
            {
                addJoin(imd.joinMetaData[i]);
            }
        }
        if (imd.foreignKeyMetaData != null)
        {
            for (int i=0;i<imd.foreignKeyMetaData.length;i++)
            {
                addForeignKey(imd.foreignKeyMetaData[i]);
            }
        }
        if (imd.indexMetaData != null)
        {
            for (int i=0;i<imd.indexMetaData.length;i++)
            {
                addIndex(imd.indexMetaData[i]);
            }
        }
        if (imd.uniqueMetaData != null)
        {
            for (int i=0;i<imd.uniqueMetaData.length;i++)
            {
                addUniqueConstraint(imd.uniqueMetaData[i]);
            }
        }
        if (imd.fetchGroupMetaData != null)
        {
            for (int i=0;i<imd.fetchGroupMetaData.length;i++)
            {
                addFetchGroup(imd.fetchGroupMetaData[i]);
            }
        }
        if (queries != null)
        {
            Iterator iter = queries.iterator();
            {
                addQuery((QueryMetaData)iter.next());
            }
        }

        if (imd.listenerClassNames != null)
        {
            if (listenerClassNames == null)
            {
                listenerClassNames = new ArrayList();
            }
            listenerClassNames.addAll(imd.listenerClassNames);
        }
        if (imd.callbacks != null)
        {
            if (callbacks == null)
            {
                callbacks = new HashMap();
            }
            callbacks.putAll(imd.callbacks);
        }
    }

    /**
     * Convenience method to copy the fields from an existing interface
     * @param imd The class/interface that we copy from
     */
    protected void copyFieldsFromInterface(InterfaceMetaData imd)
    {
        for (int i=0; i<imd.getNoOfInheritedManagedFields()+imd.getNoOfManagedFields(); i++)
        {
            //generate FieldMetaData out of PropertyMetaData
            FieldMetaData fmd = new FieldMetaData(this, imd.getManagedFieldAbsolute(i));
            addProperty(fmd);
        }
    }

    /**
     * Method to check whether the Meta-Data has been initialised.
     * @throws JDOFatalInternalException Thrown if the Meta-Data hasn't been initialised. 
     **/
    protected void checkInitialised()
    {
        if (!isInitialised())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.NotInitialisedError",fullName));
        }
    }

    /**
     * Method to check whether the Meta-Data has been populated.
     * @throws JDOFatalInternalException Thrown if the Meta-Data hasn't been populated. 
     **/
    protected void checkPopulated()
    throws JDOFatalInternalException
    {
        if (!isPopulated() && !isInitialised())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.NotPopulatedError",fullName));
        }
    }

    /**
     * Load the persistent interface/class
     * @param clr the ClassLoader
     * @param primary the primary ClassLoader to use (or null)
     * @return the loaded class
     */
    protected Class loadClass(ClassLoaderResolver clr, ClassLoader primary)
    {
        // No class loader, so use default
        if (clr == null)
        {
            JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Class.UsingDefaultClassLoaderResolver",fullName));
            clr = getMetaDataManager().getPMFContext().getClassLoaderResolver(null);
        }
    
        // Load the class we are modelling
        Class cls;
        try
        {
            cls = clr.classForName(fullName,primary,false);
        }
        catch (ClassNotResolvedException cnre)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.ClassNotFoundError",fullName));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.ClassNotFoundError",
                fullName);
        }
        if( cls == null )
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.ClassNotFoundError",fullName));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.ClassNotFoundError",
                fullName);
        }
        return cls;
    }

    /**
     * Determines the identity based on MetaData defaults or user defined MetaData
     */
    protected void determineIdentity()
    {
        // Provide a default identity type if not supplied
        if (this.identityType == null)
        {
            if (objectidClass != null)
            {
                // PK provided so we use application-identity
                identityType = IdentityType.APPLICATION;
            }
            else
            {
                int noOfPkKeys = 0;
                Iterator fieldsIter = properties.iterator();
                while (fieldsIter.hasNext())
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)fieldsIter.next();
                    if (fmd.isPrimaryKey())
                    {
                        noOfPkKeys++;
                    }
                }
                if (noOfPkKeys == 1)
                {
                    // 1 PK so we use application-identity (SingleFieldIdentity)
                    identityType = IdentityType.APPLICATION; // SingleFieldIdentity
                }
                else
                {
                    // Use datastore-identity
                    identityType = IdentityType.DATASTORE;
                }
            }
        }
    }

    /**
     * Add PropertyMetaData of properties not declared in MetaData
     * @param cls This class
     */
    protected void addPropertiesNotInMetaData(Class cls)
    {
        // Add fields for the class that aren't in the PropertyMetaData.
        // We use Reflection here since JDOImplHelper would only give use info
        // for enhanced files (and the enhancer needs unenhanced as well). 
        // NOTE 1 : We ignore fields in superclasses
        // NOTE 2 : We ignore "enhanced" fields (added by the JDO enhancer)
        // NOTE 3 : We ignore inner class fields (containing "$") 
        // NOTE 4 : We sort the fields into ascending alphabetical order
        Collections.sort(properties);
        try
        {
            // Get all (reflected) fields in the populating class
            Method[] clsMethods=cls.getDeclaredMethods();
            for (int i=0;i<clsMethods.length;i++)
            {
                // Limit to methods in this class, that aren't enhanced fields
                // that aren't inner class fields, and that aren't static
                if (clsMethods[i].getDeclaringClass().getName().equals(fullName) &&
                    ( clsMethods[i].getName().startsWith("get") || clsMethods[i].getName().startsWith("is") ) &&
                    !ClassUtils.isInnerClass(clsMethods[i].getName()))
                {
                    String fieldName = ClassUtils.getFieldNameForJavaBeanGetter(clsMethods[i].getName() );
                    // Find if there is a PropertyMetaData for this field.
                    // This is possible as PropertyMetaData implements Comparable
                    if (Collections.binarySearch(properties,fieldName) < 0)
                    {
                        // Add a default AbstractPropertyMetaData for this field.
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.AddingUnspecifiedField",fieldName,name));
                        AbstractPropertyMetaData fmd = newDefaultedProperty(fieldName);
                        properties.add(fmd);
                        Collections.sort(properties);
                    }
                }
            }
        }
        catch (Exception e)
        {
            JPOXLogger.METADATA.error(e.getMessage(), e);
            throw new RuntimeException(e.getMessage());
        }
    }
    
    /**
     * Determine the super PersistenceCapable class.
     * The persistence-capable-superclass attribute is deprecated for JDO2.0. The
     * attribute will be ignored so metadata files from previous releases can be used.
     * The persistenceCapableSuperclass we use will be retrieved from the class itself.
     * @param clr The ClassLoaderResolver
     * @param cls This class
     * @throws InvalidMetaDataException if the declared <code>persistence-capable-superclass</code> super class 
     *                                  cannot be loaded by the <code>clr</code>. 
     * @throws InvalidMetaDataException if the declared <code>persistence-capable-superclass</code> has
     *                                  no reachable MetaData
     */
    protected void determineSuperClassName(ClassLoaderResolver clr, Class cls)
    {
        String realPcSuperclassName = null;

        Collection superclasses = null;
        if (cls.isInterface())
        {
            superclasses = ClassUtils.getSuperinterfaces(cls);
        }
        else
        {
            superclasses = ClassUtils.getSuperclasses(cls);
        }

        if (!superclasses.isEmpty())
        {
            Iterator superclassesIter = superclasses.iterator();
            while (superclassesIter.hasNext())
            {
                Class superclass = (Class) superclassesIter.next();
                AbstractClassMetaData superCmd = getMetaDataManager().getMetaDataForClassInternal(superclass, clr);
                if (superCmd != null && superCmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                {
                    realPcSuperclassName = superclass.getName();
                    break;
                }
            }
        }

        if (persistenceCapableSuperclass != null && !persistenceCapableSuperclass.equals(realPcSuperclassName))
        {
            // Valifies the defined persistenceCapableSuperclass
            try
            {
                AbstractClassMetaData superCmd = getMetaDataManager().getMetaDataForClassInternal(clr.classForName(persistenceCapableSuperclass), clr);
                if (superCmd == null || superCmd.getPersistenceModifier() != ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                {
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.IsNotPCError", name,
                        persistenceCapableSuperclass);
                }
            }
            catch (ClassNotResolvedException cnre)
            {
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.ClassNotFoundError", name,
                    persistenceCapableSuperclass);
            }
            if (realPcSuperclassName != null)
            {
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.InvalidDefinition", name, realPcSuperclassName,
                    persistenceCapableSuperclass);
            }
            else
            {
                // The defined persistenceCapableSuperclass could NOT have been a superclass
                // of this class, otherwise it would have been equal to the realPcSuperclassName.
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.IsNotSuperclassError", name,
                    persistenceCapableSuperclass);
            }
        }
        if (persistenceCapableSuperclass == null && realPcSuperclassName != null)
        {
            persistenceCapableSuperclass = realPcSuperclassName;
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Class.PersistenceCapableSuperclass.Set", name, persistenceCapableSuperclass));
        }
        validateSuperClass(clr, cls, superclasses);
    }

    /**
     * Check if the previously determined super class implements PersistenceCapable and has 
     * reachable MetaData file. The super class MetaData is populated if not yet populated or initilised.
     * @param clr the ClassLoaderResolver
     * @param cls This class
     * @param superclasses A List of super classes of this class, starting from the least derived super class up to the root class
     * @throws InvalidMetaDataException if the super class cannot be loaded by the <code>clr</code>. 
     * @throws InvalidMetaDataException if the declared <code>persistence-capable-superclass</code> is not 
     *                                  actually assignable from <code>cls</code> 
     * @throws InvalidMetaDataException if any of the super classes implements <code>PersistenceCapable</code> interface, 
     *                                  but the MetaData says that class is not persistent. 
     */
    private void validateSuperClass(ClassLoaderResolver clr, Class cls, Collection superclasses)
    {
        // "persistence-capable-superclass"
        // Check that the class can be loaded, and is a true superclass 
        if (persistenceCapableSuperclass != null)
        {
            Class pcsc=null;
            try
            {
                // Load using same class loader resolver as this class
                pcsc = clr.classForName(persistenceCapableSuperclass);
            }
            catch (ClassNotResolvedException cnre)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.NotFoundError",
                    name,persistenceCapableSuperclass);
            }

            if (persistenceCapableSuperclass.equals(fullName) ||
                !(pcsc.isAssignableFrom(cls)))
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.IsNotSuperclassError",
                    name,persistenceCapableSuperclass);
            }

            // Retrieve the Meta-Data for the superclass
            if (getMetaDataManager() != null)
            {
                // JPOX normal operation will have a MetaDataManager and so we
                // can ask that for the MetaData of the superclass.
                pcSuperclassMetaData = getMetaDataManager().getMetaDataForClassInternal(pcsc, clr);
                if (pcSuperclassMetaData == null)
                {
                    // Superclass isn't persistence capable since no MetaData
                    // could be found.
                    throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.PersistenceCapableSuperclass.IsNotPCError",
                        name,persistenceCapableSuperclass);
                }
            }
            else
            {
                // The enhancer doesn't need MetaDataManager so we just
                // navigate up to the FileMetaData and find it.
                // NOTE : assumes that the class is specified in the same file 
                String superclass_pkg_name=persistenceCapableSuperclass.substring(0,persistenceCapableSuperclass.lastIndexOf('.'));
                PackageMetaData pmd=getPackageMetaData().getFileMetaData().getPackage(superclass_pkg_name);
                if (pmd != null)
                {
                    String superclass_class_name=persistenceCapableSuperclass.substring(persistenceCapableSuperclass.lastIndexOf('.')+1);
                    pcSuperclassMetaData = pmd.getClass(superclass_class_name);
                }
            }
            if (pcSuperclassMetaData == null)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.NoSuperclassMetaDataError",
                                        name, persistenceCapableSuperclass);
            }
            if (!pcSuperclassMetaData.isPopulated() &&
                !pcSuperclassMetaData.isInitialised())
            {
                // Populate the superclass since we will be needing it
                pcSuperclassMetaData.populate(clr,cls.getClassLoader());
            }
        }
        else
        {
            // Check that none of our real superclasses is PersistenceCapable since this
            // class hasn't declared a PC superclass
            if (!superclasses.isEmpty())
            {
                Iterator superclassesIter = superclasses.iterator();
                while (superclassesIter.hasNext())
                {
                    Class superclass = (Class)superclassesIter.next();
                    AbstractClassMetaData superCmd = getMetaDataManager().getMetaDataForClassInternal(superclass, clr);
                    if (superCmd != null && superCmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                    {
                        throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.NoSuperclassYetSuperclassIsPC",
                            fullName, superclass.getName());
                    }
                }
            }
        }
    }    
    
    /**
     * Check for conflicts on super class and this class MetaData identity
     * @throws InvalidMetaDataException if the user tries to overwrite a super class identity / strategy
     */
    protected void validateUserInputForIdentity()
    {
        // Check that the user isn't trying to override the datastore-identity strategy!
        if (pcSuperclassMetaData != null)
        {
            AbstractClassMetaData baseCmd = getBaseAbstractClassMetaData();
            IdentityMetaData imd = baseCmd.getIdentityMetaData();
            if (baseCmd.identitySpecified && identitySpecified &&
                imd != null && imd.getStrategyValue() != null &&
                identityMetaData != null && identityMetaData.getStrategyValue() != null &&
                identityMetaData.getStrategyValue() != imd.getStrategyValue())
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.IdentityStrategyNotEqualsSuperClassIdentityStrategy",
                    fullName, identityMetaData.getStrategyValue(), imd.getStrategyValue());
            }
        }
    }
    
    /**
     * Convenience accessor for the AbstractClassMetaData of the base object in this hierarchy.
     * @return The AbstractClassMetaData for the base object.
     */
    public AbstractClassMetaData getBaseAbstractClassMetaData()
    {
        if (pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getBaseAbstractClassMetaData();
        }
        else
        {
            return this;
        }
    }
    
    /**
     * Inherit detach settings from super classes. This class is forced to be detachable
     * if the super classes is.
     */
    protected void inheritDetachableSettings()
    {
        // "persistence-capable-superclass"
        // Check that the class can be loaded, and is a true superclass 
        if (persistenceCapableSuperclass != null)
        {
            if (!isDetachable() && pcSuperclassMetaData.isDetachable())
            {
                // If the superclass is detachable then we are by definition since we extend
                detachable = true;
            }
        }
    }
    
    /**
     * Inherit the identity definition from super classes. 
     * @throws InvalidMetaDataException if the MetaData of this class conflicts with super classes definition
     */
    protected void inheritIdentity()
    {
        // "persistence-capable-superclass"
        // Check that the class can be loaded, and is a true superclass 
        if (persistenceCapableSuperclass != null)
        {
            // Class has superclass, yet has objectid-class defined! ops, this might result in user errors
            if (objectidClass != null)
            {
                String superObjectIdClass = pcSuperclassMetaData.getObjectidClass();
                if (superObjectIdClass == null || !objectidClass.equals(superObjectIdClass))
                {
                    throw new InvalidMetaDataException(LOCALISER,
                      "MetaData.Class.PersistenceCapableSuperclass.ObjectIdClassDiffersFromSuperclassError",
                                        name, persistenceCapableSuperclass);
                }
                // by default users should only specify the object-id class in the root persistent class
                JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Class.PersistenceCapableSuperclass.ObjectIdClassSpecifiedWarning",
                    name, persistenceCapableSuperclass));
            } 
            else
            {
                // get the objectid class from superclass
                this.objectidClass = pcSuperclassMetaData.getObjectidClass();
            }

            if (this.identityType == null)
            {
                this.identityType = pcSuperclassMetaData.getIdentityType();
            }

            // Superclass identityType must be the same as this classes identityType
            if (!this.identityType.equals(pcSuperclassMetaData.getIdentityType()))
            {
                // We can't change the identity type from what was specified in the base class
                throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.IdentityTypeNotEqualsSuperClassIdentityType",
                        fullName);
            }
        }
    }    
    
    /**
     * Utility to add a defaulted PropertyMetaData to the class. Provided as a
     * method since then any derived classes can override it (e.g
     * JDOConfigClass can create a JDOConfigField)
     * @param name name of field
     * @return the new PropertyMetaData
     */
    protected AbstractPropertyMetaData newDefaultedProperty(String name)
    {
        return new PropertyMetaData(this, name);
    }
    
    /**
     * Method to merge in the associated metadata from the ORM definition
     * @param ormCmd The ORM Class definition
     */
    public void mergeORMData(AbstractClassMetaData ormCmd)
    {
        if (ormCmd == null)
        {
            return;
        }
        if (isInitialised() || isPopulated())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",name));
        }

        // Update the O/R mapping for this class
        if (ormCmd.getCatalog() != null)
        {
            catalog = ormCmd.getCatalog();
        }
        if (ormCmd.getSchema() != null)
        {
            schema = ormCmd.getSchema();
        }
        if (ormCmd.getTable() != null)
        {
            table = ormCmd.getTable();
        }
        if (ormCmd.getPrimaryKeyMetaData() != null)
        {
            primaryKeyMetaData = ormCmd.getPrimaryKeyMetaData();
        }
        if (ormCmd.getInheritanceMetaData() != null)
        {
            inheritanceMetaData = ormCmd.getInheritanceMetaData();
        }
        if (ormCmd.getIdentityMetaData() != null)
        {
            identityMetaData = ormCmd.getIdentityMetaData();
            identitySpecified = true;
        }
        if (ormCmd.getVersionMetaData() != null)
        {
            versionMetaData = ormCmd.getVersionMetaData();
        }

        if (ormCmd.listenerClassNames != null)
        {
            if (listenerClassNames == null)
            {
                listenerClassNames = new ArrayList();
            }
            listenerClassNames.addAll(ormCmd.listenerClassNames);
        }
        if (ormCmd.callbacks != null)
        {
            if (callbacks == null)
            {
                callbacks = new HashMap();
            }
            callbacks.putAll(ormCmd.callbacks);
        }

        // TODO Should we retain anything from the JDO file and add on the ORM variants ?
        // Queries
        if (ormCmd.queries != null)
        {
            if (queries == null)
            {
                queries = new ArrayList();
            }
            else
            {
                queries.clear();
            }
            queries.addAll(ormCmd.queries);
        }

        // Joins
        joins.clear();
        joins.addAll(ormCmd.joins);

        // Indexes
        indexes.clear();
        indexes.addAll(ormCmd.indexes);

        // Foreign-Keys
        foreignKeys.clear();
        foreignKeys.addAll(ormCmd.foreignKeys);

        // Unique Constraints
        uniqueConstraints.clear();
        uniqueConstraints.addAll(ormCmd.uniqueConstraints);

        // Add on any fields that weren't defined previously
        for (int i=0;i<ormCmd.getNoOfFields();i++)
        {
            AbstractPropertyMetaData ormFmd=ormCmd.getField(i);
            if (!hasField(ormFmd.getName()))
            {
                AbstractPropertyMetaData fmd = null;
                if (ormFmd.className != null)
                {
                    // Overridden field for superclass that we have no JDO field for
                    // Copy the fmd for the actual class (if any).
                    // TODO Replace this with a copy of the JDO version of the field if available
                    AbstractPropertyMetaData jdoFmd = getMetaDataManager().readMetaDataForField(ormFmd.className, ormFmd.name);
                    if (jdoFmd == null)
                    {
                        jdoFmd = getMetaDataManager().readMetaDataForField(ormCmd.getPackageName() + "." + ormFmd.className, ormFmd.name);
                    }
                    if (jdoFmd != null)
                    {
                        if (jdoFmd instanceof FieldMetaData)
                        {
                            // Copy the JDO definition of the superclass since no JDO definition in this class
                            fmd = new FieldMetaData(this, jdoFmd);
                        }
                        else
                        {
                            // Copy the JDO definition of the superclass since no JDO definition in this class
                            fmd = new PropertyMetaData(this, (PropertyMetaData)jdoFmd);
                        }
                    }
                    else
                    {
                        if (ormFmd instanceof FieldMetaData)
                        {
                            // Create default field since no available definition
                            fmd = new FieldMetaData(this, ormFmd.getName());
                        }
                        else
                        {
                            // Create default property since no available definition
                            fmd = new PropertyMetaData(this, ormFmd.getName());
                        }
                    }
                }
                else
                {
                    if (ormFmd instanceof FieldMetaData)
                    {
                        // ORM definition of the JDO field
                        // creates a new fmd with this as parent and add to fields
                        fmd = new FieldMetaData(this, ormFmd.getName());
                    }
                    else
                    {
                        // ORM definition of the JDO property
                        // creates a new fmd with this as parent and add to fields
                        fmd = new PropertyMetaData(this, ormFmd.getName());
                    }
                }
                addProperty(fmd);
            }
        }

        // Update the fields, merging in the ORM definition
        Iterator iter=properties.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData fmd=(AbstractPropertyMetaData)iter.next();
            fmd.mergeORMData(ormCmd.getField(fmd.getName()));
        }
    }

    /**
     * Check if the inheritance MetaData is credible
     * @throws InvalidMetaDataException if the strategy is superclass-table, yet there are no super class
     * @throws InvalidMetaDataException if the strategy is superclass-table, yet the super class has not
     *                                  specified a discriminator
     * @throws InvalidMetaDataException if the strategy is superclass-table and discriminator is "value-map",
     *                                  yet no value for the discriminator has been specified
     */
    protected void validateUserInputForInheritanceMetaData()
    {
        // Check that the inheritance strategy is credible
        if (inheritanceMetaData != null)
        {
            // Check validity of inheritance strategy and discriminator
            if (inheritanceMetaData.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
            {
                AbstractClassMetaData superCmd = getClassManagingTableForClass(this);
                if (superCmd == null)
                {
                    // We need a superclass table yet there is no superclass with its own table!
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoSuperclassError", fullName);
                }

                DiscriminatorMetaData superDismd = superCmd.getInheritanceMetaData().getDiscriminatorMetaData();
                if (superDismd == null)
                {
                    // If we are using "superclass-table" then the superclass should have specified the discriminator.
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoSuperclassDiscriminator", 
                        fullName, superCmd.fullName);
                }

                DiscriminatorMetaData dismd = inheritanceMetaData.getDiscriminatorMetaData();
                if (superDismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP &&
                    (dismd == null || dismd.getValue() == null))
                {
                    // If we are using "superclass-table" and the discriminator uses "value-map" then we must specify a value
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoDiscriminatorValue",
                        fullName, superCmd.fullName, superDismd.getColumnName());
                }
            }
        }
    }
    
    /**
     * Utility to navigate up to superclasses to find the next class with its own table.
     * @param cmd AbstractClassMetaData of the class
     * @return The AbstractClassMetaData of the class managing its own table
     */
    private static AbstractClassMetaData getClassManagingTableForClass(AbstractClassMetaData cmd)
    {
        if (cmd == null)
        {
            return null;
        }

        InheritanceMetaData imd = cmd.getInheritanceMetaData();
        if (imd == null)
        {
            return cmd;
        }
        else if (imd.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            return cmd;
        }
        else if (imd.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
        {
            return getClassManagingTableForClass(cmd.getSuperAbstractClassMetaData());
        }
        return null;
    }    

    /**
     * Accessor for the Meta-Data for the superclass of this class.
     * @return MetaData of the superclass
     */
    public AbstractClassMetaData getSuperAbstractClassMetaData()
    {
        // checkInitialised();

        return pcSuperclassMetaData;
    }    
    
    /**
     * Convenience accessor for the discriminator strategy applying to this class.
     * This is specified on the class managing the table if at all.
     * @return The discriminator strategy
     */
    public final DiscriminatorStrategy getDiscriminatorStrategy()
    {
        if (inheritanceMetaData == null)
        {
            return null;
        }
        else if (inheritanceMetaData.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            if (inheritanceMetaData.getDiscriminatorMetaData() != null)
            {
                return inheritanceMetaData.getDiscriminatorMetaData().getStrategy();
            }
            else
            {
                // Not specified.
                return null;
            }
        }
        else if (getSuperAbstractClassMetaData() != null)
        {
            return getSuperAbstractClassMetaData().getDiscriminatorStrategy();
        }
        else
        {
            // Unknown!
            return null;
        }
    }    

    /**
     * Determine the object id class
     * @param clr the ClassLoader
     * @throws InvalidMetaDataException if the class 0 or more that one primary key field and no <code>objectid-class</code>
     *                                  has been declared in the MetaData
     * @throws InvalidMetaDataException if the <code>objectid-class</code> has not been set and the primary key field does
     *                                  not match a supported SingleFieldIdentity
     * @throws InvalidMetaDataException if the identity type is APPLICATION but not primary key fields have been set
     * @throws InvalidMetaDataException if the <code>objectid-class</code> cannot be loaded by the <code>clr</code>                                                                     
     */
    protected void determineObjectIdClass(ClassLoaderResolver clr)
    {
        // Update "objectid-class" if required yet not specified
        if (objectidClass == null &&
            persistenceCapableSuperclass == null &&
            identityType == IdentityType.APPLICATION)
        {
            int no_of_pk_fields = 0;
            AbstractPropertyMetaData fmd_pk = null;
            Iterator fields_pk_iter=properties.iterator();
            while (fields_pk_iter.hasNext())
            {
                AbstractPropertyMetaData fmd=(AbstractPropertyMetaData)fields_pk_iter.next();
                if (fmd.isPrimaryKey())
                {
                    fmd_pk = fmd;
                    no_of_pk_fields++;
                }
            }

            // Check for more than 1 field
            if (no_of_pk_fields > 1 || no_of_pk_fields == 0)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.SingleFieldIdentityInvalid", fullName, "" + no_of_pk_fields));
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.SingleFieldIdentityInvalid", fullName, "" + no_of_pk_fields);
            }

            // Assign associated SingleField identity class
            Class pk_type = fmd_pk.getType();
            if (Byte.class.isAssignableFrom(pk_type) ||
                byte.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_BYTE_IDENTITY;
            }
            else if (Character.class.isAssignableFrom(pk_type) ||
                     char.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_CHAR_IDENTITY;
            }
            else if (Integer.class.isAssignableFrom(pk_type) ||
                     int.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_INT_IDENTITY;
            }
            else if (Long.class.isAssignableFrom(pk_type) ||
                     long.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_LONG_IDENTITY;
            }
            else if (Short.class.isAssignableFrom(pk_type) ||
                     short.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_SHORT_IDENTITY;
            }
            else if (String.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_STRING_IDENTITY;
            }
            else if (Object.class.isAssignableFrom(pk_type))
            {
                objectidClass = ClassNameConstants.JAVAX_JDO_IDENTITY_OBJECT_IDENTITY;
            }
            else
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.SingleFieldIdentityInvalidType", fullName, pk_type.getName()));
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.SingleFieldIdentityInvalidType", fullName, pk_type.getName());
            }
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Class.SingleFieldIdentityUsed", fullName, objectidClass));
        }

        // Check no of primary key fields (inc superclass)
        int no_of_pk_fields = getNoOfPopulatedPKFields();
        if (no_of_pk_fields == 0 && identityType == IdentityType.APPLICATION)
        {
            // No primary key fields found (even in superclasses)
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyNotSpecifiedError",
                fullName,objectidClass);
        }

        // Update "objectid-class" since has been specified.
        // Only check at the least derived pc class
        if (objectidClass != null && getPersistenceCapableSuperclass() == null)
        {
            Class obj_cls = null;
            try
            {
                // Load the class, using the same class loader resolver as this class
                obj_cls=clr.classForName(objectidClass);
            }
            catch (ClassNotResolvedException cnre)
            {
                // ObjectIdClass not found
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.ObjectIdClass.ClassNotFoundError",
                    fullName,objectidClass);
            }

            boolean validated = false;
            Set errors = new HashSet();
            ApiAdapter apiAdapter = getMetaDataManager().getPMFContext().getApiAdapter();
            try
            {
                // Check against the API Adapter in use for this MetaData
                if (apiAdapter.isValidPrimaryKeyClass(obj_cls, this, clr, no_of_pk_fields))
                {
                    validated = true;
                }
            }
            catch (JDOException ex)
            {
                errors.add(ex);
            }
            if (!validated)
            {
                // Why is this wrapping all exceptions into 1 single exception? 
                // This needs coordinating with the test expectations in the enhancer unit tests.
                throw new JDOUserException(LOCALISER_API.msg("PrimaryKey.InvalidPrimaryKey", obj_cls.getName()), 
                    (Throwable[]) errors.toArray(new Throwable[errors.size()]));
            }
        }
    }

    /**
     * Method to provide the details of the class being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * class being represented.
     * <P>
     * This method must be invoked by subclasses during populate operations  
     * </P>
     * @param clr ClassLoaderResolver to use in loading any classes
     * @param primary the primary ClassLoader to use (or null)
     */
    abstract public void populate(ClassLoaderResolver clr,ClassLoader primary);

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects. If populate() is going to be used it should
     * be used BEFORE calling this method.
     * 
     * Subclasses must overwrite this method and invoke this
     */
    public void initialise()
    {
        propertyMetaData = (AbstractPropertyMetaData[]) properties.toArray(new AbstractPropertyMetaData[properties.size()]);
    }

    /**
     * Impose a default inheritance strategy to match the JDO 2 spec
     */
    protected void determineInheritanceMetaData()
    {
        // Impose a default inheritance strategy to match the JDO 2 spec
        if (inheritanceMetaData == null)
        {
            if (pcSuperclassMetaData != null)
            {
                if (getMetaDataManager().getPMFContext().getPmfConfiguration().getDefaultInheritanceStrategy().equalsIgnoreCase("JPOX"))
                {
                    // JPOX 1.0 behaviour
                    inheritanceMetaData = new InheritanceMetaData(this, InheritanceStrategy.NEW_TABLE.toString());
                }
                else
                {
                    // JDO2 behaviour
                    inheritanceMetaData = new InheritanceMetaData(this, InheritanceStrategy.SUPERCLASS_TABLE.toString());
                }
            }
            else
            {
                inheritanceMetaData = new InheritanceMetaData(this,InheritanceStrategy.NEW_TABLE.toString());
            }
        }
    }

    /**
     * Check usage of deprecated MetaData
     * @throws InvalidMetaDataException if deprecated MetaData is used
     */
    protected void validateDeprecatedMetaData()
    {
        // "use-poid-generator" was used in JPOX 1.0, and 1.1 until beta-2. Now replaced by "value-strategy".
        if (hasExtension("use-poid-generator"))
        {
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.UsePoidGeneratorDeprecated", fullName);
        }
    }
    
    /**
     * Method to return the ClassMetaData records for classes referenced
     * by this object. This adds the entries to orderedCMDs ordered by
     * dependency, and to referencedCMDs for fast lookups.
     * <p>
     * Uses recursion to add all referenced ClassMetaData for any fields,
     * objectid classes, superclasses, and JPOX extension RDBMS "views".
     * </p>
     * <p>
     * This is the entry point for this process, and provides the core of the
     * "persistence-by-reachability" concept.
     * </p> 
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver
     */
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        Map viewReferences = new HashMap();
        getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,viewReferences,clr);
    }

    /**
     * Method to return the ClassMetaData for classes referenced by this
     * object. This method does the actual work of addition.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param viewReferences Map, mapping class name to set of referenced class
     *                       for all views.
     * @param clr the ClassLoaderResolver
     */
    private void getReferencedClassMetaData(final List orderedCMDs,
                                            final Set referencedCMDs,
                                            final String dba_vendor_id,
                                            final Map viewReferences,
                                            final ClassLoaderResolver clr)
    {
        // Recursively call getReferencedClassMetaData(...) before adding them
        // to the orderedCmds and referenced. This will ensure that any
        // classes with dependencies on them are put in the orderedCmds List
        // in the correct order.
        if (!referencedCMDs.contains(this))
        {
            // Go ahead and add this class to the referenced Set, it will
            // get added to the orderedCMDs List after all classes that this
            // depends on have been added.
            referencedCMDs.add(this);

            for (int i=0;i<managedFields.length;i++)
            {
                AbstractPropertyMetaData fmd=managedFields[i];
                fmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }

            // Add on any superclass
            if (persistenceCapableSuperclass != null)
            {
                AbstractClassMetaData super_cmd=getSuperAbstractClassMetaData();
                super_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }

            // Add on any objectid class
            if (objectidClass != null)
            {
                AbstractClassMetaData id_cmd=getMetaDataManager().getMetaDataForClass(objectidClass,clr);
                if (id_cmd != null)
                {
                    id_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
                }
            }

            // Add on any views defined for this class
            String viewDefStr = getValueForExtension("view-definition-"+dba_vendor_id);
            if( viewDefStr == null )
            {
                viewDefStr = getValueForExtension("view-definition");
            }
            if( viewDefStr!= null )
            {
                MacroString viewDef = new MacroString(fullName,
                           getValueForExtension("view-imports"),
                           viewDefStr);
                viewDef.substituteMacros(new MacroString.MacroHandler()
                    {
                        public void onIdentifierMacro(MacroString.IdentifierMacro im)
                        {
                            if (!getFullClassName().equals(im.className)) //ignore itself
                            {
                                addViewReference(viewReferences,im.className);
                                AbstractClassMetaData view_cmd=getMetaDataManager().getMetaDataForClass(im.className,clr);
                                view_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,viewReferences,clr);
                            }
                        }

                        public void onParameterMacro(MacroString.ParameterMacro pm)
                        {
                            throw new JDOUserException("Parameter macros not allowed in view definitions: " + pm);
                        }
                    },clr);
            }           
            
            orderedCMDs.add(this);
        }
    }

    /**
     * Method to add a reference for views. Check the view references for
     * circular dependencies.  If there are any circular dependencies, throw
     * a JDOUserException.
     * @param viewReferences The Map of classname to Set of referenced Classes
     *                       to add the reference to.
     * @param referenced_name Class name of the referenced class
     * @throws JDOFatalUserException Thrown if a circular reference is found
     */
    private void addViewReference(Map viewReferences,String referenced_name)
    throws JDOFatalUserException
    {
        if (fullName.equals(referenced_name))
        {
            // Add this reference to the Map.
            Set referencedSet = (Set)viewReferences.get(referenced_name);
            if (referencedSet == null)
            {
                referencedSet = new HashSet();
                viewReferences.put(fullName,referencedSet);
            }
            referencedSet.add(referenced_name);

            // Check to see if there is a circular dependency.  This will
            // be true if the referenced class references this class.
            ViewUtils.checkForCircularViewReferences(viewReferences,fullName,referenced_name,null);
        }
    }
    
    // ------------------------------ Accessors --------------------------------

    /**
     * Accessor for the number of named queries.
     * @return no of named queries
     */
    public int getNoOfQueries()
    {
        return queries.size();
    }

    /**
     * Accessor for the metadata of the named queries.
     * @return Meta-Data for the named queries.
     */
    public QueryMetaData[] getQueries()
    {
        return (queries == null ? null : ((QueryMetaData[])queries.toArray(new QueryMetaData[queries.size()])));
    }

    /**
     * Accessor for versionMetaData
     * @return Returns the versionMetaData.
     */
    public final VersionMetaData getVersionMetaData()
    {
        return versionMetaData;
    }

    /**
     * Accessor for joinMetaData
     * @return Returns the joinMetaData.
     */
    public final JoinMetaData[] getJoinMetaData()
    {
        return joinMetaData;
    }

    /**
     * Accessor for all MetaData defined for fetch groups for this class.
     * This doesn't include superclasses.
     * @return Returns the fetchGroupMetaData.
     */
    public final FetchGroupMetaData[] getFetchGroupMetaData()
    {
        return fetchGroupMetaData;
    }

    /**
     * Accessor for the full class name.
     * @return full class name.
     */
    public String getFullClassName()
    {
        return fullName;
    }
    
    /**
     * Accessor for fetch group metadata for the specified groups (if present).
     * The returned metadata is what is defined for this class that matches
     * any of the names in the input set.
     * @param groupNames Names of the fetch groups
     * @return MetaData for the groups
     */
    public FetchGroupMetaData[] getFetchGroupMetaData(Collection groupNames)
    {
        Collection results = new HashSet();
        for (Iterator iter = groupNames.iterator(); iter.hasNext();)
        {
            String groupname = (String) iter.next();
            FetchGroupMetaData fgmd = getFetchGroupMetaData(groupname);
            if (fgmd != null)
            {
                results.add(fgmd);
            }
        }
        FetchGroupMetaData[] result = new FetchGroupMetaData[results.size()];
        results.toArray(result);
        return result;
    }

    /**
     * Accessor for the identity-type. 
     * @return identity-type tag value
     */
    public IdentityType getIdentityType()
    {
        return identityType;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData[] getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Accessor for foreignKeyMetaData
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData[] getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    /**
     * Accessor for UniqueMetaData
     * @return Returns the UniqueMetaData.
     */
    public final UniqueMetaData[] getUniqueMetaData()
    {
        return uniqueMetaData;
    }

    /**
     * Accessor for the unmapped columns require for the datastore table.
     * @return The list of unmapped columns
     */
    public final List getUnmappedColumns()
    {
        return unmappedColumns;
    }
    
    /**
     * Accessor for the number of properties
     * @return no of properties
     */
    public int getNoOfProperties()
    {
        return properties.size();
    } 
    
    /**
     * Accessor for the class name.
     * @return class name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the class name.
     * @return class name
     */
    public String getEntityName()
    {
        return entityName;
    }
    
    /**
     * Accessor for the objectid-class tag value.
     * @return the fully qualified objectid class name 
     */
    public String getObjectidClass()
    {
        return objectidClass;
    }

    /**
     * Accessor for the catalog name.
     * @return catalog name
     */
    public String getCatalog()
    {
        if (this.catalog == null)
        {
            // Nothing defined here for catalog, so get from package
            return ((PackageMetaData)parent).getCatalog();
        }
       
        return catalog;
    }

    /**
     * Accessor for the schema name.
     * @return schema name
     */
    public String getSchema()
    {
        if (this.schema == null)
        {
            // Nothing defined here for schema, so get from package
            return ((PackageMetaData)parent).getSchema();
        }

        return schema;
    }

    /**
     * Accessor for the table name.
     * @return table name
     */
    public String getTable()
    {
        return table;
    }

    /**
     * Accessor for the requires-extent tag value.
     * @return requires-extent tag value
     */
    public boolean isRequiresExtent()
    {
        return requiresExtent;
    }

    /**
     * Accessor for the detachable tag value.
     * @return detachable tag value
     */
    public boolean isDetachable()
    {
        return detachable;
    }


    /**
     * Accessor for the embedded-only tag value.
     * @return embedded-only tag value
     */
    public boolean isEmbeddedOnly()
    {
        return embeddedOnly;
    }
    
    /**
     * Accessor for identityMetaData
     * @return Returns the identityMetaData.
     */
    public final IdentityMetaData getIdentityMetaData()
    {
        return identityMetaData;
    }

    /**
     * Accessor for inheritanceMetaData
     * @return Returns the inheritanceMetaData.
     */
    public final InheritanceMetaData getInheritanceMetaData()
    {
        return inheritanceMetaData;
    }

    /**
     * Accessor for primaryKeyMetaData
     * @return Returns the primaryKey MetaData.
     */
    public final PrimaryKeyMetaData getPrimaryKeyMetaData()
    {
        return primaryKeyMetaData;
    }

    /**
     * Convenience accessor for the MetaData Manager in use.
     * @return MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        if (parent != null && parent.getParent() != null)
        {
            return getPackageMetaData().getFileMetaData().getMetaDataManager();
        }
        return null;
    }

    /**
     * Convenience accessor for the parent Package MetaData.
     * @return MetaData for parent package.
     */
    public PackageMetaData getPackageMetaData()
    {
        if (parent != null)
        {
            return (PackageMetaData)parent;
        }
        return null;
    }

    /**
     * Convenience accessor for the package name.
     * @return package name.
     */
    public String getPackageName()
    {
        return getPackageMetaData().getName();
    }

    /**
     * Accessor for the number of fields.
     * This is the total number of fields (inc static, final etc) in this class 
     * @return no of fields
     */
    public int getNoOfFields()
    {
        return properties.size();
    }

    /**
     * Accessor for the metadata of a field. Does not include superclasses
     * @param index field index relative to this class only starting from 0
     * @return Meta-Data for the field.
     */
    public AbstractPropertyMetaData getField(int index)
    {
        if (index < 0 || index >= properties.size())
        {
            return null;
        }
        return (AbstractPropertyMetaData)properties.get(index);
    }

    /**
     * Accessor for the persistence-modifier
     * @return persistence-modifier tag value
     */
    public ClassPersistenceModifier getPersistenceModifier()
    {
        return persistenceModifier;
    }

    /**
     * Accessor for the persistence-capable-superclass tag value.
     * @return persistence-capable-superclass tag value
     */
    public String getPersistenceCapableSuperclass()
    {
        return persistenceCapableSuperclass;
    }

    /**
     * Check if the argument cmd is the same as this or a descedent.
     * @param cmd the AbstractClassMetaData to be verify if this is an ancestor
     * @return true if the argument is a child or same as this
     */
    public boolean isSameOrAncestorOf(AbstractClassMetaData cmd)
    {
        checkInitialised();
        
        if (cmd == null)
        {
            return false;
        }
        
        if (fullName.equals(cmd.fullName))
        {
            return true;
        }
        
        AbstractClassMetaData parent = cmd.getSuperAbstractClassMetaData();
        while( parent != null )
        {
            if (fullName.equals(parent.fullName))
            {
                return true;
            }
            parent = parent.getSuperAbstractClassMetaData();
        }
        return false;
    }

    /**
     * Accessor for the fetch group metadata for the group specified.
     * Overrides the version in AbstractClassMetaData so that we can allow for
     * superclasses.
     * @param groupname Name of the fetch group
     * @return MetaData for this group
     */
    public FetchGroupMetaData getFetchGroupMetaData(String groupname)
    {
        FetchGroupMetaData fgmd = (FetchGroupMetaData) fetchGroupMetaDataByName.get(groupname);
        if (fgmd == null && pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getFetchGroupMetaData(groupname);
        }
        return fgmd;
    }

    /**
     * Method to check if a field exists in this classes definition.
     * @param fieldName Nme of field
     * @return return true if exists.
     */
    public boolean hasField(String fieldName)
    {
        Iterator iter=properties.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData field=(AbstractPropertyMetaData)iter.next();
            if (field.getName().equals(fieldName))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the Meta-Data for a field. Include superclasses
     * @param name the name of the field
     * @return Meta-Data for the field.
     */
    public AbstractPropertyMetaData getField(String name)
    {
        if (name == null)
        {
            return null;
        }

        Iterator iter=properties.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)iter.next();
            if (fmd.getName().equals(name))
            {
                return fmd;
            }
        }

        // Check superclass for the field with this name
        if (pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getField(name);
        }
        return null;
    }

    /**
     * Accessor for the number of JDO managed fields. 
     * The number of fields is for this class only.
     * @return no of JDO managed fields in this class
     */
    public int getNoOfManagedFields()
    {
        // checkInitialised();

        if (managedFields == null)
        {
            return 0;
        }
        return managedFields.length;
    }

    /**
     * Accessor for the managed fields in this class (not including superclass).
     * @return The JDO managed fields in this class
     */
    public AbstractPropertyMetaData[] getManagedFields()
    {
        checkInitialised();

        return managedFields;
    }

    /**
     * Accessor for a managed field in this class. The position is relative to
     * the first field in this class (ignores superclasses).
     * @param position The position of the managed field. 0 = first in the class
     * @return The JDO managed field at that position
     */
    public AbstractPropertyMetaData getManagedField(int position)
    {
        checkInitialised();

        if (managedFields == null)
        {
            return null;
        }
        if (position < 0 || position >= managedFields.length)
        {
            return null;
        }

        return managedFields[position];
    }

    /**
     * Accessor for the number of JDO overridden fields. 
     * The number of fields is for this class only.
     * @return no of JDO overridden fields in this class
     */
    public int getNoOfOverriddenFields()
    {
        if (overriddenFields == null)
        {
            return 0;
        }
        return overriddenFields.length;
    }

    /**
     * Accessor for the overridden fields in this class.
     * @return The JDO overridden fields in this class
     */
    public AbstractPropertyMetaData[] getOverriddenFields()
    {
        checkInitialised();

        return overriddenFields;
    }

    /**
     * Accessor for an overridden field in this class.
     * @param position The position of the overridden field.
     * @return The JDO managed field at that position
     */
    public AbstractPropertyMetaData getOverriddenField(int position)
    {
        checkInitialised();

        if (overriddenFields == null)
        {
            return null;
        }
        if (position < 0 || position >= overriddenFields.length)
        {
            return null;
        }

        return overriddenFields[position];
    }

    /**
     * Accessor for an overridden field with the specified name.
     * @param name Name of the field
     * @return The MetaData for the field
     */
    public AbstractPropertyMetaData getOverriddenField(String name)
    {
        checkInitialised();

        if (overriddenFields == null)
        {
            return null;
        }
        for (int i=0;i<overriddenFields.length;i++)
        {
            if (overriddenFields[i].getName().equals(name))
            {
                return overriddenFields[i];
            }
        }
        return null;
    }

    /**
     * Accessor for the number of inherited managed fields in superclasses.
     * @return No of inherited managed fields in superclasses.
     */
    public int getNoOfInheritedManagedFields()
    {
        checkInitialised();

        return noOfInheritedManagedFields;
    }

    /**
     * Accessor for a managed field including superclass fields.
     * @param abs_position The position of the managed field including the
     * superclass. Fields are numbered from 0 in the root superclass.
     * @return The JDO managed field at this "absolute" position.
     */
    public AbstractPropertyMetaData getManagedFieldAbsolute(int abs_position)
    {
        checkInitialised();

        return getManagedFieldAbsoluteInternal(abs_position);
    }

    /**
     * Internal method to get the field for an absolute field number.
     * If the field for that absolute field position is overridden by a field in this class
     * then this field will be returned.
     * @param abs_position The position of the managed field including the
     * superclass. Fields are numbered from 0 in the root superclass.
     * @return The JDO managed field at this "absolute" position.
     */
    protected AbstractPropertyMetaData getManagedFieldAbsoluteInternal(int abs_position)
    {
        // If the field is in a superclass, go there
        if (abs_position < noOfInheritedManagedFields)
        {
            if (pcSuperclassMetaData == null)
            {
                return null;
            }
            else
            {
                AbstractPropertyMetaData fmd = pcSuperclassMetaData.getManagedFieldAbsoluteInternal(abs_position);
                if (fmd != null)
                {
                    for (int i=0;i<overriddenFields.length;i++)
                    {
                        if (overriddenFields[i].getName().equals(fmd.getName()) &&
                            overriddenFields[i].getClassName().equals(fmd.getClassName()))
                        {
                            // Return the overriding field if we have one (class and field name is the safest comparison)
                            return overriddenFields[i];
                        }
                    }
                    
                    return fmd;
                }
                return null;
            }
        }
        // If the field is in this class, return it
        else if (abs_position - noOfInheritedManagedFields >= managedFields.length)
        {
            return null;
        }
        else
        {
            return managedFields[abs_position - noOfInheritedManagedFields];
        }
    }
    
    /**
     * Accessor for the Absolute field number for a relative field number
     * @param relativeFieldNo The relative field number
     * @return The absolute field number
     */
    public int getAbsoluteFieldNumberForRelativeFieldNumber(int relativeFieldNo)
    {
        return noOfInheritedManagedFields + relativeFieldNo;
    }

    /**
     * Accessor for the position of the field with the specified name.
     * The returned position is relative to this class only.
     * @param field_name Name of the field
     * @return Position of the field in this class.
     */
    public int getFieldNumber(String field_name)
    {
        checkInitialised();

        if (field_name == null)
        {
            return -1;
        }

        Integer i=(Integer)fieldNumbersByName.get(field_name);

        return i == null ? -1 : i.intValue();
    }

    /**
     * Accessor for the absolute position of the field with the specified name.
     * The absolute position has origin of the root superclass, starting at 0.
     * @param field_name Name of the field
     * @return Absolute position of the field.
     */
    public int getFieldNumberAbsolute(String field_name)
    {
        checkInitialised();

        if (field_name == null)
        {
            return -1;
        }

        int i=getFieldNumber(field_name);
        if (i < 0)
        {
            if (pcSuperclassMetaData != null)
            {
                i = pcSuperclassMetaData.getFieldNumberAbsolute(field_name);
            }
        }
        else
        {
            i += noOfInheritedManagedFields;
        }

        return i;
    }

    /**
     * Accessor for the absolute position of the field with the specified name.
     * The absolute position has origin of the root superclass, starting at 0.
     * Will only retrieve the absolute field number if the <code>field_name</code>
     * is of type given by <code>class_name</code>.
     * @param class_name Name of the class
     * @param field_name Name of the field
     * @return Absolute position of the field.
     */
    public int getFieldNumberAbsolute(String class_name, String field_name)
    {
        checkInitialised();

        if (field_name == null)
        {
            return -1;
        }
        
        int i=-1;
        if( class_name.equals(getFullClassName()))
        {
            i=getFieldNumber(field_name);
        }
        if (i < 0)
        {
            if (pcSuperclassMetaData != null)
            {
                i = pcSuperclassMetaData.getFieldNumberAbsolute(class_name,field_name);
            }
        }
        else
        {
            i += noOfInheritedManagedFields;
        }

        return i;
    }
    
    /**
     * Convenience method to check the number of fields in this class that have been populated and
     * that are primary key fields. This is only ever called during populate() since the accessor
     * for PK fields cant be used yet due to lack of initialisation.
     * Recurses to its superclass if it has a superclass.
     * @return The number of PK fields
     */
    private int getNoOfPopulatedPKFields()
    {
        if (pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getNoOfPopulatedPKFields();
        }

        Iterator fields_iter = properties.iterator();
        int noOfPks = 0;
        while (fields_iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)fields_iter.next();
            if (fmd.isPrimaryKey())
            {
                noOfPks++;
            }
        }
        return noOfPks;
    }

    /**
     * Accessor for the number of primary key fields.
     * @return no of primary key fields
     */
    public int getNoOfPrimaryKeyFields()
    {
        if (primaryKeyFieldNumbers == null)
        {
            return 0;
        }
        else
        {
            return primaryKeyFieldNumbers.length;
        }
    }

    /**
     * Accessor for all field numbers. These are absolute numbers and include
     * superclasses and are really just 0, 1, 2, ... n.
     * @return The field numbers of all (managed) fields.
     **/
    public int[] getAllFieldNumbers()
    {
        checkInitialised();
        return allFieldNumbers;
    }
 
    /**
     * Accessor for the field numbers of the primary key fields. These are
     * absolute numbers (including superclasses).
     * @return The field numbers of the primary key fields
     **/
    public int[] getPrimaryKeyFieldNumbers()
    {
        checkInitialised();
        return primaryKeyFieldNumbers;
    }
 
    /**
     * Accessor for the field numbers of the non primary key fields (inc
     * superclass fields).
     * @return The field numbers
     **/
    public int[] getNonPrimaryKeyFieldNumbers()
    {
        checkInitialised();
        return nonPrimaryKeyFieldNumbers;
    }    

    /**
     * Accessor for the flags of the non primary key fields (inc
     * superclass fields).
     * @return The flags whether the field is non primary key
     **/
    public boolean[] getNonPrimaryKeyFieldFlags()
    {
        checkInitialised();
        return nonPrimaryKeyFieldFlags;
    } 
    
    /**
     * Accessor for the absolute field numbers of the default fetch group
     * fields (inc superclasses).
     * @return The field numbers of the DFG fields (inc superclasses).
     **/
    public int[] getDefaultFetchGroupFieldNumbers()
    {
        checkInitialised();
        return defaultFetchGroupFieldNumbers;
    }

    /**
     * Accessor for the flags of the default fetch group fields (inc superclass
     * fields).
     * @return The flags whether the field is in the DFG
     **/
    public boolean[] getDefaultFetchGroupFieldFlags()
    {
        checkInitialised();
        return defaultFetchGroupFieldFlags;
    }

    /**
     * Accessor for the absolute field numbers of the second class mutable
     * fields
     * @return The field numbers of the second class mutable fields (inc
     *         superclasses).
     **/
    public int[] getSecondClassMutableFieldNumbers()
    {
        checkInitialised();
        return secondClassMutableFieldNumbers;
    }

    /** Field numbers of all SCO fields that arent containers. */
    protected int[] secondClassNonContainerFieldNumbers = null;

    /**
     * Accessor for the absolute field numbers of all SCO fields that are
     * NOT containers (e.g Dates, Points, etc)
     * @return Field numbers of all SCO non-container fields
     */
    public int[] getSecondClassNonContainerFieldNumbers()
    {
        checkInitialised();
        if (secondClassNonContainerFieldNumbers == null)
        {
            int[] noncontainerFieldNumbers;
            int numberNonContainerSCOFields = 0;
            for (int i=0;i<secondClassMutableFieldNumbers.length;i++)
            {
                AbstractPropertyMetaData fmd = getManagedFieldAbsolute(secondClassMutableFieldNumbers[i]);
                if (!(java.util.Collection.class.isAssignableFrom(fmd.getType())) &&
                    !(java.util.Map.class.isAssignableFrom(fmd.getType())))
                {
                    numberNonContainerSCOFields++;
                }
            }

            noncontainerFieldNumbers = new int[numberNonContainerSCOFields];
            int nonContNum = 0;
            for (int i=0;i<secondClassMutableFieldNumbers.length;i++)
            {
                AbstractPropertyMetaData fmd = getManagedFieldAbsolute(secondClassMutableFieldNumbers[i]);
                if (!(java.util.Collection.class.isAssignableFrom(fmd.getType())) &&
                    !(java.util.Map.class.isAssignableFrom(fmd.getType())))
                {
                    noncontainerFieldNumbers[nonContNum++] = secondClassMutableFieldNumbers[i];
                }
            }

            secondClassNonContainerFieldNumbers = noncontainerFieldNumbers;
        }
        return secondClassNonContainerFieldNumbers;
    }

    /** Field numbers of all SCO fields that are containers. */
    protected int[] secondClassContainerFieldNumbers = null;

    /**
     * Accessor for the absolute field numbers of all SCO fields that are containers.
     * @return Field numbers of all SCO container fields
     */
    public int[] getSecondClassContainerFieldNumbers()
    {
        checkInitialised();
        if (secondClassContainerFieldNumbers == null)
        {
            int[] containerFieldNumbers;
            int numberContainerSCOFields = 0;
            for (int i=0;i<secondClassMutableFieldNumbers.length;i++)
            {
                AbstractPropertyMetaData fmd = getManagedFieldAbsolute(secondClassMutableFieldNumbers[i]);
                if (java.util.Collection.class.isAssignableFrom(fmd.getType()) ||
                    java.util.Map.class.isAssignableFrom(fmd.getType()))
                {
                    numberContainerSCOFields++;
                }
            }

            containerFieldNumbers = new int[numberContainerSCOFields];
            int contNum = 0;
            for (int i=0;i<secondClassMutableFieldNumbers.length;i++)
            {
                AbstractPropertyMetaData fmd = getManagedFieldAbsolute(secondClassMutableFieldNumbers[i]);
                if (java.util.Collection.class.isAssignableFrom(fmd.getType()) ||
                    java.util.Map.class.isAssignableFrom(fmd.getType()))
                {
                    containerFieldNumbers[contNum++] = secondClassMutableFieldNumbers[i];
                }
            }

            secondClassContainerFieldNumbers = containerFieldNumbers;
        }
        return secondClassContainerFieldNumbers;
    }

    /**
     * Accessor for the absolute field numbers of the PersistenceCapable
     * fields
     * @return The field numbers of the PersistenceCapable fields (inc
     *         superclasses).
     **/
    public int[] getPersistenceCapableFieldNumbers()
    {
        checkInitialised();
        return persistenceCapableFieldNumbers;
    }

    /**
     * Accessor for the flags of the second class mutable fields (inc
     * superclass fields).
     * @return The flags whether the field is second class mutable
     **/
    public boolean[] getSecondClassMutableFieldFlags()
    {
        checkInitialised();
        return secondClassMutableFieldFlags;
    }    

    /**
     * Whether this PersistenceCapable class uses SingleFieldIdentity
     * @return true if using SingleFieldIdentity as objectid class
     */
    public boolean usesSingleFieldIdentityClass()
    {
        return usesSingleFieldIdentityClass;
    } 
    
    // ------------------------------- Mutators --------------------------------

    /**
     * Method to add a named query to this class. Rejects the addition of
     * duplicate named queries.
     * @param qmd Meta-Data for the query.
     */
    public void addQuery(QueryMetaData qmd)
    {
        if (qmd == null)
        {
            return;
        }
        if (queries == null)
        {
            queries = new HashSet();
        }

        queries.add(qmd);
    }

    /**
     * Method to add an index to this class.
     * @param idxmd Meta-Data for the index.
     */
    public void addIndex(IndexMetaData idxmd)
    {
        if (idxmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        indexes.add(idxmd);
    }

    /**
     * Method to add an foreign-key to this class.
     * @param fkmd Meta-Data for the foreign-key.
     */
    public void addForeignKey(ForeignKeyMetaData fkmd)
    {
        if (fkmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        foreignKeys.add(fkmd);
    }

    /**
     * Method to add a unique constraint to this class.
     * @param unimd Meta-Data for the unique constraint.
     */
    public void addUniqueConstraint(UniqueMetaData unimd)
    {
        if (unimd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        uniqueConstraints.add(unimd);
    }

    /**
     * Method to add an unmapped column.
     * @param colmd The metadata for the unmapped column
     */
    public final void addUnmappedColumn(ColumnMetaData colmd)
    {
        if (unmappedColumns == null)
        {
            unmappedColumns = new ArrayList();
        }
        unmappedColumns.add(colmd);
    }

    /**
     * Method to add a property (field) to this interface.
     * Rejects the addition of duplicate named properties (fields).
     * @param pmd Property
     */
    public void addProperty(AbstractPropertyMetaData pmd)
    {
        if (pmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            // TODO Localise this message
            throw new RuntimeException("adding property (field) " + pmd.getName() + " when already initialised!");
        }

        Iterator iter = properties.iterator();
        while (iter.hasNext())
        {
            AbstractPropertyMetaData md = (AbstractPropertyMetaData)iter.next();
            if (pmd.getName().equals(md.getName()))
            {
                throw new RuntimeException(LOCALISER.msg("MetaData.Class.DuplicateFieldError", fullName, pmd.getName()));
            }
        }
        properties.add(pmd);
    }

    /**
     * Method to add a fetch-group to this class.
     * @param fgmd Meta-Data for the fetch-group.
     */
    public void addFetchGroup(FetchGroupMetaData fgmd)
    {
        if (fgmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        fetchGroups.add(fgmd);
    }
    
    /**
     * Method to add a join to this class.
     * Rejects the addition of duplicate named fields.
     * @param jnmd Meta-Data for the join.
     */
    public void addJoin(JoinMetaData jnmd)
    {
        if (jnmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        joins.add(jnmd);
    }

    /**
     * Add a method callback 
     * @param callbackClassName the callback class name
     * @param methodName the method name
     */
    public void addCallback(String callbackClassName, String methodName)
    {
        if (callbacks == null)
        {
            callbacks = new HashMap();
        }
        callbacks.put(callbackClassName, methodName);    
    }
    
    /**
     * Get the callback method name for a given callback class name 
     * @param callbackClassName the callback class name
     * @return null if there is no method for the callback
     */
    public String getCallback(String callbackClassName)
    {
        return callbacks == null ? null : (String) callbacks.get(callbackClassName);    
    }

    /**
     * Add a listener class name
     * @param listenerClassName the fully qualified listener class name. Duplicated classes are ignored
     */
    public void addListener(String listenerClassName)
    {
        if (listenerClassNames == null)
        {
            listenerClassNames = new ArrayList();
        }
        if (!listenerClassNames.contains(listenerClassName))
        {
            listenerClassNames.add(listenerClassName);
        }
    }

    /**
     * Get the listener class names
     * @return the listener class names
     */
    public List getListenerClassNames()
    {
        return listenerClassNames;    
    }

    /**
     * Toogle exclude super class listeners
     */
    public void excludeSuperClassListeners()
    {
        this.excludeSuperClassListeners = true;
    }
    
    /**
     * Whether super classes listeners are not going to be invoked
     * @return true if super class listeners are not invoked
     */
    public boolean isExcludeSuperClassListeners()
    {
        return this.excludeSuperClassListeners;
    }

    /**
     * Method to exclude default listeners.
     */
    public void excludeDefaultListeners()
    {
        this.excludeDefaultListeners = true;
    }

    /**
     * Whether default listeners are not going to be invoked
     * @return true if default listeners are not invoked
     */
    public boolean isExcludeDefaultListeners()
    {
        return this.excludeDefaultListeners;
    }

    /**
     * Mutator for the Version MetaData.
     * @param versionMetaData The versionMetaData to set.
     */
    public final void setVersionMetaData(VersionMetaData versionMetaData)
    {
        this.versionMetaData = versionMetaData;
        if (this.versionMetaData != null)
        {
            this.versionMetaData.parent = this;
        }
    }

    /**
     * Mutator for the identity MetaData.
     * @param identityMetaData The identityMetaData to set.
     */
    public final void setIdentityMetaData(IdentityMetaData identityMetaData)
    {
        this.identityMetaData = identityMetaData;
        if (this.identityMetaData != null)
        {
            this.identityMetaData.parent = this;
        }
        identitySpecified = true;
    }

    /**
     * Mutator for the inheritance MetaData.
     * @param inheritanceMetaData The inheritanceMetaData to set.
     */
    public final void setInheritanceMetaData(InheritanceMetaData inheritanceMetaData)
    {
        this.inheritanceMetaData = inheritanceMetaData;
        if (this.inheritanceMetaData != null)
        {
            this.inheritanceMetaData.parent = this;
        }
    }

    /**
     * Mutator for the PrimaryKey MetaData.
     * @param primaryKeyMetaData The PrimaryKey MetaData to set.
     */
    public final void setPrimaryKeyMetaData(PrimaryKeyMetaData primaryKeyMetaData)
    {
        this.primaryKeyMetaData = primaryKeyMetaData;
        if (this.primaryKeyMetaData != null)
        {
            this.primaryKeyMetaData.parent = this;
        }
    }

    /**
     * Mutator for the table name.
     * @param table_name Name of the table
     */
    public final void setTable(String table_name)
    {
        this.table = table_name;
    }
}