/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.sco.SCOUtils;
import org.jpox.store.DatastoreField;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * The property element declares mapping between a virtual field of an implemented 
 * interface and the corresponding persistent field of a persistence-capable class. 
 * The name attribute is required, and declares the name for the property. The naming 
 * conventions for JavaBeans property names is used: the property name is the same as 
 * the corresponding get method for the property with the get removed and the resulting
 * name lowercased. 
 * The field-name attribute is required; it associates a persistent field with the 
 * named property.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * <!ELEMENT property ((collection|map|array|column)? , extension*)>
 * <!ATTLIST property name CDATA #REQUIRED>
 * <!ATTLIST property column CDATA #IMPLIED>
 * </PRE>
 * 
 * @version $Revision$
 */
public abstract class AbstractPropertyMetaData extends ExtendableMetaData  implements Comparable, ColumnMetaDataContainer
{
    /** Contains the metadata for column(s). */
    protected ColumnMetaData[] columnMetaData;

    /** Meta-Data of any container. */
    protected ContainerMetaData container;

    /** EmbeddedMetaData. */
    protected EmbeddedMetaData embeddedMetaData;

    /** JoinMetaData. */
    protected JoinMetaData joinMetaData;

    /** ElementMetaData. */
    protected ElementMetaData elementMetaData;

    /** KeyMetaData. */
    protected KeyMetaData keyMetaData;

    /** ValueMetaData. */
    protected ValueMetaData valueMetaData;

    /** IndexMetaData. */
    protected IndexMetaData indexMetaData;

    /** The indexing value */
    protected IndexedValue indexed=null;

    /** UniqueMetaData. */
    protected UniqueMetaData uniqueMetaData;

    /** Whether to add a unique constraint. */
    protected final boolean uniqueConstraint;

    /** OrderMetaData. */
    protected OrderMetaData orderMetaData;

    /** ForeignKeyMetaData. */
    protected ForeignKeyMetaData foreignKeyMetaData;

    /** default-fetch-group tag value. */
    protected Boolean defaultFetchGroup;

    /** column tag value.  */
    protected String column;

    /** mapped-by tag value. */
    protected String mappedBy;

    /** embedded tag value. */
    protected Boolean embedded;

    /**
     * dependent tag value.
     * 
     * The dependent attribute indicates that the field contains a reference
     * that is to be deleted from the datastore if the referring instance in
     * which the field is declared is deleted, or if the referring field is
     * nullified.
     */
    protected Boolean dependent;
    
    /** serialized tag value. */
    protected Boolean serialized;

    /** load-fetch-group value. */
    protected final String loadFetchGroup;

    /** Default recursion-depth according to proposed final draft spec, [12.7.2]. */
    public static final int DEFAULT_RECURSION_DEPTH = 1;

    /** Indicates the recursion-depth is not defined.  Use default value. */
    public static final int UNDEFINED_RECURSION_DEPTH = 0;

    /** recursion-depth value. */
    protected int recursionDepth = UNDEFINED_RECURSION_DEPTH;

    /** Field name. */
    protected final String name;

    /** null-value tag value (default is NONE). */
    protected NullValue nullValue=NullValue.NONE;

    /** persistence-modifier tag value. */
    protected FieldPersistenceModifier persistenceModifier=FieldPersistenceModifier.DEFAULT;

    /** primary key tag value. */
    protected Boolean primaryKey;

    /** Table name for this field. */
    protected String table;

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     */
    protected IdentityStrategy valueStrategy;
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     */
    protected String sequence;

    /**
     * Name of the class to which this field really belongs. Will be null if the field belongs
     * to the parent ClassMetaData, and will have a value if it is an overriding field.
     */
    protected String className = null;

    /**
     * The field-type attribute is used to specify a more restrictive type than the field definition
     * in the class. This might be required in order to map the field to the datastore. To be
     * portable, specify the name of a single type that is itself able to be mapped to the datastore
     * (e.g. a field of type Object can specify field-type=Integer). To specify multiple
     * types that the field might contain, use a comma-separated list of types, although this cannot
     * be portably mapped. Rules for type names are as specified in collection elementtype.
     */
    protected String fieldType;

    /** Field type being represented by this MetaData. */
    protected Class type;

    /** Field modifiers */
    protected int modifiers;

    /**
     * Id of the field in its class (only for fields managed by JDO).
     * If the value is -1, the field is NOT managed by JDO or the object hasn't
     * been populated.
     */
    protected int fieldId=-1;

    /** The relation type of this field (1-1, 1-N, M-N, N-1). */
    protected int relationType = -1;

    /** FieldMetaData for the other end of a relation when this field is a bidirectional relation. */
    protected AbstractPropertyMetaData relatedFieldMetaData = null;

    // -------------------------------------------------------------------------
    // These fields are only used when the MetaData is read by the parser and
    // elements are dynamically added to the other elements. At runtime, "columnMetaData"
    // should be used.

    /**
     * Columns ColumnMetaData
     */
    protected List columns = new ArrayList();

    /**
     * Convenience constructor taking defaults
     * @param parent Parent component
     * @param name Name of the field
     */
    public AbstractPropertyMetaData(MetaData parent,
                         final String name)
    {
        this(parent, name, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Convenience constructor to copy the specification from the passed field.
     * This is used when we have an overriding field and we make a copy of the baseline
     * field as a starting point.
     * @param parent The parent
     * @param fmd The field to copy
     */
    public AbstractPropertyMetaData(MetaData parent, AbstractPropertyMetaData fmd)
    {
        super(parent);

        // Copy the simple field values
        this.name = fmd.name;
        this.primaryKey = fmd.primaryKey;
        this.defaultFetchGroup = fmd.defaultFetchGroup;
        this.column = fmd.column;
        this.mappedBy = fmd.mappedBy;
        this.dependent = fmd.dependent;
        this.embedded = fmd.embedded;
        this.serialized = fmd.serialized;
        this.nullValue = fmd.nullValue;
        this.persistenceModifier = fmd.persistenceModifier;
        this.table = fmd.table;
        this.indexed = fmd.indexed;
        this.valueStrategy = fmd.valueStrategy;
        this.sequence = fmd.sequence;
        this.uniqueConstraint = fmd.uniqueConstraint;
        this.loadFetchGroup = fmd.loadFetchGroup;
        this.fieldType = fmd.fieldType;

        // Create copies of the object fields
        if (fmd.joinMetaData != null)
        {
            this.joinMetaData = new JoinMetaData(this, fmd.joinMetaData);
        }
        if (fmd.elementMetaData != null)
        {
            this.elementMetaData = new ElementMetaData(this, fmd.elementMetaData);
        }
        if (fmd.keyMetaData != null)
        {
            this.keyMetaData = new KeyMetaData(this, fmd.keyMetaData);
        }
        if (fmd.valueMetaData != null)
        {
            this.valueMetaData = new ValueMetaData(this, fmd.valueMetaData);
        }
        if (fmd.orderMetaData != null)
        {
            this.orderMetaData = new OrderMetaData(this, fmd.orderMetaData);
        }
        if (fmd.indexMetaData != null)
        {
            this.indexMetaData = new IndexMetaData(this, fmd.indexMetaData);
        }
        if (fmd.uniqueMetaData != null)
        {
            this.uniqueMetaData = new UniqueMetaData(this, fmd.uniqueMetaData);
        }
        if (fmd.foreignKeyMetaData != null)
        {
            this.foreignKeyMetaData = new ForeignKeyMetaData(this, fmd.foreignKeyMetaData);
        }
        if (fmd.embeddedMetaData != null)
        {
            this.embeddedMetaData = new EmbeddedMetaData(this, fmd.embeddedMetaData);
        }
        if (fmd.container != null)
        {
            if (fmd.container instanceof CollectionMetaData)
            {
                container = new CollectionMetaData(this, (CollectionMetaData)fmd.container);
            }
            else if (fmd.container instanceof MapMetaData)
            {
                container = new MapMetaData(this, (MapMetaData)fmd.container);
            }
            else if (fmd.container instanceof ArrayMetaData)
            {
                container = new ArrayMetaData(this, (ArrayMetaData)fmd.container);
            }
        }
        for (int i=0;i<fmd.columns.size();i++)
        {
            addColumn(new ColumnMetaData(this,(ColumnMetaData)fmd.columns.get(i)));
        }
    }

    /**
     * Constructor. Saves the MetaData with the specified values. The object is
     * then in an "unpopulated" state. It can become "populated" by calling the
     * <B>populate()</B> method which compares it against the field it is to
     * represent and updates any unset attributes and flags up any errors.
     * @param parent parent MetaData instance
     * @param name field name 
     * @param pk attribute primary-key value
     * @param modifier attribute persistence-modifier value
     * @param defaultFetchGroup attribute default-fetch-group value
     * @param nullValue attribute null-value value 
     * @param embedded attribute embedded value
     * @param serialized attribute serialized value
     * @param dependent attribute dependent value
     * @param mappedBy attribute mapped-by value
     * @param column attribute column value
     * @param table attribute table value
     * @param deleteAction attribute delete-action value
     * @param indexed Whether this is indexed
     * @param unique Apply a unique constraint
     * @param recursionDepth The depth of fetch to use when recursing
     * @param loadFetchGroup Name of the additional fetch group to use when loading
     * @param valueStrategy attribute value-strategy value
     * @param sequence attribute sequence value
     * @param fieldType Implementation type(s) for field.
     */
    public AbstractPropertyMetaData(MetaData parent,
                         final String name,
                         final String pk,
                         final String modifier,
                         final String defaultFetchGroup,
                         final String nullValue,
                         final String embedded,
                         final String serialized,
                         final String dependent,
                         final String mappedBy,
                         final String column,
                         final String table,
                         final String deleteAction,
                         final String indexed,
                         final String unique,
                         final String recursionDepth,
                         final String loadFetchGroup,
                         final String valueStrategy,
                         final String sequence,
                         final String fieldType)
    {
        super(parent);

        if (name == null)
        {
            throw new JDOUserException(LOCALISER.msg("MetaData.Parser.ExpectedValueInAttribute","name", getClassName(true), "field"));
        }
        if (name.indexOf('.') >= 0)
        {
            // TODO Check if this is a valid className
            this.className = name.substring(0,name.lastIndexOf('.'));
            this.name = name.substring(name.lastIndexOf('.')+1);
        }
        else
        {
            this.name = name;
        }

        // "primary-key"
        if (!StringUtils.isWhitespace(pk) && pk.equalsIgnoreCase("true"))
        {
            this.primaryKey = Boolean.TRUE;
            this.defaultFetchGroup = Boolean.TRUE;
        }
        else
        {
            this.primaryKey = Boolean.FALSE;
        }

        this.column = (StringUtils.isWhitespace(column) ? null : column);
        this.mappedBy = (StringUtils.isWhitespace(mappedBy) ? null : mappedBy);

        // "dependent"
        if (dependent != null)
        {
            if (dependent.equalsIgnoreCase("true"))
            {
                this.dependent = Boolean.TRUE;
            }
            else if (dependent.equalsIgnoreCase("false"))
            {
                this.dependent = Boolean.FALSE;
            }
        }

        // "embedded"
        if (embedded != null)
        {
            if (embedded.equalsIgnoreCase("true"))
            {
                this.embedded = Boolean.TRUE;
            }
            else if (embedded.equalsIgnoreCase("false"))
            {
                this.embedded = Boolean.FALSE;
            }
        }

        // "serialized"
        if (serialized != null)
        {
            if (serialized.equalsIgnoreCase("true"))
            {
                this.serialized = Boolean.TRUE;
            }
            else if (serialized.equalsIgnoreCase("false"))
            {
                this.serialized = Boolean.FALSE;
            }
        }

        // "null-value"
        this.nullValue = NullValue.getNullValue(nullValue);

        // "persistence-modifier"
        if (!StringUtils.isWhitespace(modifier))
        {
            persistenceModifier = FieldPersistenceModifier.getFieldPersistenceModifier(modifier);
        }
        if( persistenceModifier == null )
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.PopulateWithWrongValue",name,getClassName(false),modifier,"persistence-modifier"));
            throw new InvalidMetaDataException(LOCALISER,
                                               "MetaData.Field.PopulateWithWrongValue",
                                               name,getClassName(false),
                                               modifier,
                                               "persistence-modifier");            
        }
        
        // "default-fetch-group"
        if (defaultFetchGroup != null)
        {
            if (defaultFetchGroup.equalsIgnoreCase("true"))
            {
                this.defaultFetchGroup = Boolean.TRUE;
            }
            else if (defaultFetchGroup.equalsIgnoreCase("false"))
            {
                this.defaultFetchGroup = Boolean.FALSE;
            }
        }

        // "table"
        this.table = (StringUtils.isWhitespace(table) ? null : table);

        if (deleteAction != null)
        {
            foreignKeyMetaData = new ForeignKeyMetaData(null, deleteAction, null, null, null, null);
        }

        this.indexed = IndexedValue.getIndexedValue(indexed);

        if (unique != null && unique.equalsIgnoreCase("true"))
        {
            uniqueConstraint = true;
        }
        else
        {
            uniqueConstraint = false;
        }

        // Fetch Group
        this.loadFetchGroup = (StringUtils.isWhitespace(loadFetchGroup) ? null : loadFetchGroup);
        if (recursionDepth != null)
        {
            try
            {
                this.recursionDepth = (new Integer(recursionDepth)).intValue();
            }
            catch (Exception e)
            {
                // Not parseable so set as "undefined"
                this.recursionDepth = UNDEFINED_RECURSION_DEPTH;
            }
        }

        this.valueStrategy = valueStrategy == null ? null : IdentityStrategy.getIdentityStrategy(valueStrategy);
        this.sequence = (StringUtils.isWhitespace(sequence) ? null : sequence);
        this.fieldType = (StringUtils.isWhitespace(fieldType) ? null : fieldType);
    }

    /**
     * Method to merge in the associated metadata from the ORM definition
     * @param ormFmd The ORM Field definition
     */
    public void mergeORMData(AbstractPropertyMetaData ormFmd)
    {
        if (ormFmd == null)
        {
            return;
        }

        if (isInitialised() || isPopulated())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Field.AlreadyPopulatedError",name,ormFmd.getAbstractClassMetaData().getFullClassName()));
        }

        if (ormFmd.className != null)
        {
            // If the ORM is an overriding field, make sure we have the (real) class name
            this.className = ormFmd.className;
        }
        if (ormFmd.container != null)
        {
            container = ormFmd.container;
            container.parent = this;
        }

        // Update our O/R mapping details
        if (primaryKey == Boolean.FALSE && ormFmd.primaryKey == Boolean.TRUE)
        {
            primaryKey = ClassUtils.booleanValueOf(ormFmd.isPrimaryKey());
        }
        if (ormFmd.getTable() != null)
        {
            table = ormFmd.getTable();
        }
        if (ormFmd.getColumn() != null)
        {
            column = ormFmd.getColumn();
        }
        if (ormFmd.dependent != null)
        {
            dependent = ormFmd.dependent;
        }
        if (ormFmd.getMappedBy() != null)
        {
            mappedBy = ormFmd.getMappedBy();
        }
        if (ormFmd.getValueStrategy() != null)
        {
            valueStrategy = ormFmd.getValueStrategy();
        }
        if (ormFmd.getSequence() != null)
        {
            sequence = ormFmd.getSequence();
        }

        if (ormFmd.getJoinMetaData() != null)
        {
            joinMetaData = ormFmd.getJoinMetaData();
            if (joinMetaData != null)
            {
                joinMetaData.parent = this;
            }
        }
        if (ormFmd.getEmbeddedMetaData() != null)
        {
            embeddedMetaData = ormFmd.getEmbeddedMetaData();
            if (embeddedMetaData != null)
            {
                embeddedMetaData.parent = this;
            }
        }
        if (ormFmd.getElementMetaData() != null)
        {
            elementMetaData = ormFmd.getElementMetaData();
            if (elementMetaData != null)
            {
                elementMetaData.parent = this;
            }
        }
        if (ormFmd.getKeyMetaData() != null)
        {
            keyMetaData = ormFmd.getKeyMetaData();
            if (keyMetaData != null)
            {
                keyMetaData.parent = this;
            }
        }
        if (ormFmd.getValueMetaData() != null)
        {
            valueMetaData = ormFmd.getValueMetaData();
            if (valueMetaData != null)
            {
                valueMetaData.parent = this;
            }
        }
        if (ormFmd.getOrderMetaData() != null)
        {
            orderMetaData = ormFmd.getOrderMetaData();
            if (orderMetaData != null)
            {
                orderMetaData.parent = this;
            }
        }
        if (ormFmd.getForeignKeyMetaData() != null)
        {
            foreignKeyMetaData = ormFmd.getForeignKeyMetaData();
            if (foreignKeyMetaData != null)
            {
                foreignKeyMetaData.parent = this;
            }
        }
        if (ormFmd.indexed != null)
        {
            indexed = ormFmd.indexed;
        }
        if (ormFmd.getIndexMetaData() != null)
        {
            indexMetaData = ormFmd.getIndexMetaData();
            if (indexMetaData != null)
            {
                indexMetaData.parent = this;
            }
        }
        if (ormFmd.getUniqueMetaData() != null)
        {
            uniqueMetaData = ormFmd.getUniqueMetaData();
            if (uniqueMetaData != null)
            {
                uniqueMetaData.parent = this;
            }
        }

        // Add the columns (assumes none existing)
        // TODO Allow for others being specified in the basic file
        columns.clear();
        ColumnMetaData[] ormColumns=ormFmd.getColumnMetaData();
        if (ormColumns != null)
        {
            for (int i = 0; i < ormColumns.length; i++)
            {
                columns.add(ormColumns[i]);
            }
        }

        // Add any extensions supplied in the ORM file
        ExtensionMetaData[] ormExtensions = ormFmd.getExtensions();
        if (ormExtensions != null)
        {
            for (int i=0;i<ormExtensions.length;i++)
            {
                addExtension(ormExtensions[i].vendorName, ormExtensions[i].key, ormExtensions[i].value);
            }
        }
    }

    /**
     * Method to provide the details of the field being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * field being represented.
     * @param clr ClassLoaderResolver to use for any class loading 
     * @param fieldType The type of the field being represented
     * @param modifiers The modifiers of the field being represented (Static, Abstract, Final, Volatile, etc)
     */
    public synchronized void populate(ClassLoaderResolver clr, Class fieldType, int modifiers)
    {
        if (isPopulated() || isInitialised())
        {
            return;
            /*JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.AlreadyPopulatedError",name,getClassName(false)));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Field.AlreadyPopulatedError",
                name,getClassName(false));*/
        }

        if (fieldType == null)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.PopulateWithNullFieldError",name,getClassName(false)));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Field.PopulateWithNullError",
                name,getClassName(false));
        }

        // No class loader, so use System
        if (clr == null)
        {
            JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Class.UsingDefaultClassLoaderResolver",name,getClassName(true)));
            clr = getAbstractClassMetaData().getMetaDataManager().getPMFContext().getClassLoaderResolver(null);
        }

        this.type = fieldType;
        this.modifiers = modifiers;
        
        if (className != null)
        {
            // If the field is overriding a superclass field, check that it is valid
            Class thisClass = null;
            try
            {
                thisClass = clr.classForName(getAbstractClassMetaData().getPackageName() + "." + getAbstractClassMetaData().getName());
            }
            catch (ClassNotResolvedException cnre)
            {
                // Do nothing
            }
            
            Class fieldClass = null;
            try
            {
                fieldClass = clr.classForName(className);
            }
            catch (ClassNotResolvedException cnre)
            {
                try
                {
                    fieldClass = clr.classForName(getAbstractClassMetaData().getPackageName() + "." + className);
                    className = getAbstractClassMetaData().getPackageName() + "." + className;
                }
                catch (ClassNotResolvedException cnre2)
                {
                    JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.ClassNameInvalid", getAbstractClassMetaData().getName(), name, className));
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Field.ClassNameInvalid", getAbstractClassMetaData().getName(), name, className);
                }
            }
            if (fieldClass != null && !fieldClass.isAssignableFrom(thisClass))
            {
                // TODO We could also check if PersistenceCapable, but won't work when enhancing
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.ClassNotSuperclass", getAbstractClassMetaData().getName(), name, className));
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Field.ClassNotSuperclass", getAbstractClassMetaData().getName(), name, className);
            }
        }

        // Update "embedded" based on type
        if (primaryKey == Boolean.FALSE && embedded == null)
        {
            Class element_type=getType();
            if (element_type.isArray())
            {
                element_type = element_type.getComponentType();
                if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(element_type))
                {
                    embedded = Boolean.TRUE;
                }
            }
            else if (getMetaDataManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(element_type))
            {
                embedded = Boolean.TRUE;
            }
        }
        if (embedded == null)
        {
            embedded = Boolean.FALSE;
        }

        // Update "persistence-modifier" according to type etc
        if (FieldPersistenceModifier.DEFAULT.equals(persistenceModifier))
        {
            boolean isPcClass = getType().isArray() ? isFieldArrayTypePersistenceCapable() : isFieldTypePersistenceCapable(); 
            if (!isPcClass)
            {
                if (getType().isArray() && getType().getComponentType().isInterface())
                {
                    isPcClass = 
                        (getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(getType().getComponentType(), clr) != null);
                }
                else if (getType().isInterface())
                {
                    isPcClass = (getAbstractClassMetaData().getMetaDataManager().getMetaDataForClassInternal(getType(), clr) != null);
                }
            }

            persistenceModifier =
                getMetaDataManager().getPMFContext().getTypeManager().getDefaultFieldPersistenceModifier(
                    getType(),modifiers,isPcClass);
        }

        // TODO If this field is NONE in superclass, make it NONE here too
        // Check whether the type is supported by JPOX
        if (!persistenceModifier.equals(FieldPersistenceModifier.NONE) && getType().isArray() &&
            !getAbstractClassMetaData().getMetaDataManager().isEnhancing())
        {
            // JPOX supports arrays specified in the jpoxtypes.xml, arrays of PCs, and arrays of reference types
            if (!getMetaDataManager().getPMFContext().getTypeManager().isSupportedType(getType().getName()) &&
                !PersistenceCapable.class.isAssignableFrom(getType().getComponentType()) &&
                !getType().getComponentType().isInterface() &&
                !getType().getComponentType().getName().equals("java.lang.Object"))
            {
                JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Field.InvalidArrayType",name,getClassName(false)));
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Field.InvalidArrayType", name,getClassName(false));
            }
        }

        // Update "default-fetch-group" according to type
        if (defaultFetchGroup == null &&
            persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            defaultFetchGroup = Boolean.FALSE;
        }
        else if (defaultFetchGroup == null &&
                persistenceModifier.equals(FieldPersistenceModifier.TRANSACTIONAL))
        {
            defaultFetchGroup = Boolean.FALSE;
        }
        else if (defaultFetchGroup == null)
        {
            defaultFetchGroup = Boolean.FALSE;
            if (!primaryKey.equals(Boolean.TRUE) &&
                    getMetaDataManager().getPMFContext().getTypeManager().isDefaultFetchGroup(getType()))
            {
                defaultFetchGroup = Boolean.TRUE;
            }
        }

        // Field is not specified as "persistent" yet has DFG or primary-key !
        if (persistenceModifier.equals(FieldPersistenceModifier.TRANSACTIONAL) ||
            persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            if (defaultFetchGroup == Boolean.TRUE || primaryKey == Boolean.TRUE)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Field.NotPersistentWithIllegalAttributesError",
                    name, getClassName(false), this.getType().getName(),
                    persistenceModifier.toString());
            }
        }

        // Check for invalid dependent field specifications
        if (this.container != null && this.dependent != null)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.DependentVsContainerDependentError",
                name, getClassName(false), ((ClassMetaData)this.parent).getName()));            
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Field.DependentVsContainerDependentError",
                name, getClassName(false), ((ClassMetaData)this.parent).getName());
        }

        // Populate any element object
        if (elementMetaData != null)
        {
            elementMetaData.populate(clr);
        }

        // Populate any key/value object
        if (keyMetaData != null)
        {
            keyMetaData.populate(clr);
        }
        if (valueMetaData != null)
        {
            valueMetaData.populate(clr);
        }

        // Populate any embedded object
        if (embeddedMetaData != null)
        {
            embeddedMetaData.populate(clr);
            embedded = Boolean.TRUE;
        }

        if (elementMetaData != null)
        {
            if (elementMetaData.mappedBy != null && mappedBy == null)
            {
                // User has specified "mapped-by" on element instead of field so pull it up to this level
                // <element mapped-by="..."> is the same as <field mapped-by="..."> for a collection field
                mappedBy = elementMetaData.mappedBy;
            }
        }

        // Populate any attached container
        /*if (container == null)
        {
            if (type.isArray())
            {
                // User hasnt specifed <array> but the field is an array so add it for them
                Class arrayCls = type.getComponentType();
                ArrayMetaData arrmd = new ArrayMetaData(this, arrayCls.getName(), null, null, null);
                container = arrmd;
            }
            else if (java.util.Collection.class.isAssignableFrom(type))
            {
                // User hasnt specifed <collection> but the field is a collection so add it for them
                // TODO Do this when using JDK 1.5 and generics since we can get the element type
            }
            else if (java.util.Map.class.isAssignableFrom(type))
            {
                // User hasnt specifed <map> but the field is a map so add it for them
                // TODO Do this when using JDK 1.5 and generics since we can get the key,value types
            }
        }*/
        if (container != null)
        {
            if (container instanceof CollectionMetaData)
            {
                ((CollectionMetaData)container).populate(clr);
            }
            else if (container instanceof MapMetaData)
            {
                ((MapMetaData)container).populate(clr);
            }
            else if (container instanceof ArrayMetaData)
            {
                ((ArrayMetaData)container).populate(clr);
            }
        }

        if (hasExtension("implementation-classes"))
        {
            // Check the validity of the implementation-classes and qualify them where required.
            StringBuffer str = new StringBuffer();
            String[] implTypes = getValuesForExtension("implementation-classes");
            for (int i=0;i<implTypes.length;i++)
            {
                String implTypeName = ClassUtils.createFullClassName(getAbstractClassMetaData().getPackageName(), implTypes[i]);
                if (i > 0)
                {
                    str.append(",");
                }

                try
                {
                    clr.classForName(implTypeName);
                    str.append(implTypeName);
                }
                catch (ClassNotResolvedException cnre)
                {
                    try
                    {
                        // Maybe the user specified a java.lang class without fully-qualifying it
                        // This is beyond the scope of the JDO spec which expects java.lang cases to be fully-qualified
                        String langClassName = ClassUtils.getJavaLangClassForType(implTypeName);
                        clr.classForName(langClassName);
                        str.append(langClassName);
                    }
                    catch (ClassNotResolvedException cnre2)
                    {
                        // Implementation type not found
                        throw new InvalidMetaDataException(LOCALISER, "MetaData.Field.ImplementationClassInvalid",
                            name, getClassName(false), implTypes[i]);
                    }
                }
            }
            addExtension(JPOX_VENDOR_NAME, "implementation-classes", str.toString()); // Replace with this new value
        }

        setPopulated();
    }

    /**
     * Initialisation method. This should be called AFTER using the populate
     * method if you are going to use populate. It creates the internal
     * convenience arrays etc needed for normal operation.
     */
    public synchronized void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
            columnMetaData[0].initialise();
        }
        else if (columns.size() == 1 && column != null)
        {
            // Cater for user specifying <column> and <field column="...">
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = (ColumnMetaData) columns.get(0);
            if (columnMetaData[0].getName() == null)
            {
                columnMetaData[0].setName(column);
            }
            columnMetaData[0].initialise();
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        // Initialise all sub-objects
        if (container != null)
        {
            container.initialise();
        }
        if (embeddedMetaData != null)
        {
            embeddedMetaData.initialise();
        }
        if (joinMetaData != null)
        {
            joinMetaData.initialise();
        }
        if (elementMetaData != null)
        {
            elementMetaData.initialise();
        }
        if (keyMetaData != null)
        {
            keyMetaData.initialise();
        }
        if (valueMetaData != null)
        {
            valueMetaData.initialise();
        }

        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }
        if (uniqueMetaData == null && uniqueConstraint)
        {
            uniqueMetaData = new UniqueMetaData(this, null, column, null);
            for (int i=0;i<columnMetaData.length;i++)
            {
                uniqueMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (uniqueMetaData != null)
        {
            uniqueMetaData.initialise();
        }

        if (foreignKeyMetaData != null)
        {
            foreignKeyMetaData.initialise();
        }
        if (orderMetaData != null)
        {
            orderMetaData.initialise();
        }

        if ((hasCollection() && SCOUtils.collectionHasSerialisedElements(this)) ||
            (hasMap() && SCOUtils.mapHasSerialisedKeysAndValues(this)) ||
            (hasArray() && SCOUtils.arrayIsStoredInSingleColumn(this)))
        {
            // Collection/Array with serialised elements or Map with serialised keys and values so set as serialised for later
            serialized = Boolean.TRUE;
        }
        
        setInitialised();
    }

    /**
     * Utility to return if this field is PersistenceCapable.
     * Not valid for use by the enhancer. Must be overridden for that mode.
     * @return Whether the field type is persistence capable.
     */
    public boolean isFieldTypePersistenceCapable()
    {
        return ClassUtils.isPersistenceCapableClass(type);
    }

    /**
     * Utility to return if this array field has elements that are PersistenceCapable.
     * Not valid for use by the enhancer. Must be overridden for that mode.
     * If the field is not an array will return false.
     * @return Whether the field type is persistence capable.
     */
    public boolean isFieldArrayTypePersistenceCapable()
    {
        if (!type.isArray())
        {
            return false;
        }
        return ClassUtils.isPersistenceCapableClass(type.getComponentType());
    }

    /**
     * Convenience method to return if this field represents a static field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is static
     */
    public boolean isStatic()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isStatic(modifiers);
    }

    /**
     * Convenience method to return if this field represents a final field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is final
     */
    public boolean isFinal()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isFinal(modifiers);
    }

    /**
     * Convenience method to return if this field represents a transient field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is transient
     */
    public boolean isTransient()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isTransient(modifiers);
    }

    // ------------------------------- Accessors -------------------------------

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     * @return the value strategy
     */
    public IdentityStrategy getValueStrategy()
    {
        return valueStrategy;
    }
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     * @return the sequence name 
     */
    public String getSequence()
    {
        return sequence;
    }
    
    /**
     * Accessor for the (additional) fetch group for loading this field
     * @return Returns the load fetch group
     */
    public String getLoadFetchGroup()
    {
        return loadFetchGroup;
    }

    /**
     * Accessor for the depth of the fetch when recursing
     * @return Returns the depth of the fetch when recursing
     */
    public int getRecursionDepth()
    {
        return recursionDepth;
    }

    /**
     * Convenience method to navigate back through the parents to find the overall
     * ClassMetaData handling this object. This is to cater specifically for nested
     * embedded fields where you can nest object several levels deep.
     * @param metadata The metadata to check
     * @return The overall class metadata for this element
     */
    protected MetaData getOverallParentClassMetaData(MetaData metadata)
    {
        if (metadata == null)
        {
            return null;
        }
        else if (metadata instanceof AbstractClassMetaData)
        {
            return metadata;
        }
        else
        {
            return getOverallParentClassMetaData(metadata.getParent());
        }
    }

    /**
     * Convenience accessor for the MetaData of the parent class.
     * @return Returns the MetaData of the parent class.
     */
    public AbstractClassMetaData getAbstractClassMetaData()
    {
        // TODO Consider replacing this method with the getOverallParentClassMetaData above since its generalised
        if (parent == null)
        {
            return null;
        }
        else if (parent instanceof AbstractClassMetaData)
        {
            return (AbstractClassMetaData)parent;
        }
        else if (parent instanceof EmbeddedMetaData)
        {
            // <embedded> is contained in a <field>, <element>, <key>, <value>
            // but could be multiple levels deep so adopt a generic strategy for finding the parent class
            MetaData parentMd = ((EmbeddedMetaData)parent).getParent();
            return (AbstractClassMetaData)getOverallParentClassMetaData(parentMd.getParent());
        }
        return null;
    }

    /**
     * Accessor for orderMetaData
     * @return Returns the orderMetaData.
     */
    public final OrderMetaData getOrderMetaData()
    {
        return orderMetaData;
    }
    
    /**
     * Accessor for the field name 
     * @return field name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the full field name. This prepends the class name. 
     * @return full field name.
     */
    public String getFullFieldName()
    {
        if (className != null)
        {
            return className + "." + name;
        }
        return getClassName(true) + "." + name;
    }

    /**
     * Accessor for whether the field is for a superclass, and not for this class.
     * @return Whether the field belongs to a superclass
     */
    public boolean fieldBelongsToClass()
    {
        return (className == null);
    }

    /**
     * Accessor for the fully-qualified class name owning this field.
     * @return The class name
     */
    public String getClassName()
    {
        return getClassName(true);
    }

    /**
     * Convenience method so that ClassMetaData can update the name of the superclass
     * to which this field belongs.
     * @param className Name of the class
     */
    void setClassName(String className)
    {
        this.className = className;
    }

    /**
     * Convenience to return the class name that this a field of.
     * @param fully_qualified Whether the name should be fully qualified.
     * @return Class name
     */
    public String getClassName(boolean fully_qualified)
    {
        if (className != null)
        {
            return className;
        }

        if (parent == null)
        {
            return null;
        }
        else if (parent instanceof ClassMetaData)
        {
            ClassMetaData cmd=(ClassMetaData)parent;
            if (fully_qualified)
            {
                return cmd.getFullClassName();
            }
            else
            {
                return cmd.getName();
            }
        }
        else if (parent instanceof EmbeddedMetaData)
        {
            // <embedded> is contained in a <field>, <element>, <key>, <value>
            MetaData parentMd = ((EmbeddedMetaData)parent).getParent();
            if (parentMd instanceof AbstractPropertyMetaData)
            {
                return ((AbstractPropertyMetaData)parentMd).getTypeName();
            }
            else if (parentMd instanceof ElementMetaData)
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)((ElementMetaData)parentMd).getParent();
                return fmd.getCollection().getElementType();
            }
            else if (parentMd instanceof KeyMetaData)
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)((KeyMetaData)parentMd).getParent();
                return fmd.getMap().getKeyType();
            }
            else if (parentMd instanceof ValueMetaData)
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)((ValueMetaData)parentMd).getParent();
                return fmd.getMap().getValueType();
            }
            else
            {
                // Should be impossible
                return null;
            }
        }
        else if (parent instanceof UniqueMetaData)
        {
            MetaData grandparent = ((UniqueMetaData)parent).getParent();
            if (grandparent instanceof AbstractClassMetaData)
            {
                return ((AbstractClassMetaData)grandparent).getFullClassName();
            }
            // TODO Cater for other parent options
        }
        return null;
    }

    /**
     * Accessor for the null-value tag
     * @return null-value tag
     */
    public NullValue getNullValue()
    {
        return nullValue;
    }

    /**
     * Accessor for the persistence-modifier tag value
     * @return persistence-modifier tag value
     */
    public FieldPersistenceModifier getPersistenceModifier()
    {
        return persistenceModifier;
    }

    /**
     * Accessor for the default-fetch-group tag value
     * @return default-fetch-group tag value
     */
    public boolean isDefaultFetchGroup()
    {
        if (defaultFetchGroup == null)
        {
            return false;
        }
        else
        {
            return defaultFetchGroup.booleanValue();
        }
    }
    
    /**
     * Accessor for the dependent attribute indicates that the field contains a
     * reference that is to be deleted from the datastore if the referring
     * instance in which the field is declared is deleted, or if the referring
     * field is nullified.
     * 
     * @return dependent tag value
     */
    public boolean isDependent()
    {
        if (dependent == null)
        {
            return false;
        }
        else
        {
            return dependent.booleanValue();
        }
    }

    /**
     * Accessor for the embedded tag value.
     * This value is a hint only to the implementation so will not mean
     * that the type is certainly embedded.
     * @return embedded tag value
     */
    public boolean isEmbedded()
    {
        if (embedded == null)
        {
            return false;
        }
        else
        {
            return embedded.booleanValue();
        }
    }
    
    /**
     * Accessor for the serialized tag value
     * @return serialized tag value
     */
    public boolean isSerialized()
    {
        if (serialized == null)
        {
            return false;
        }
        else
        {
            return serialized.booleanValue();
        }
    }

    /**
     * Accessor for the primary-key tag value.
     * @return primary-key tag value.
     */
    public boolean isPrimaryKey()
    {
        if (primaryKey == null)
        {
            return false;
        }
        else
        {
            return primaryKey.booleanValue();
        }
    }

    /**
     * Whether this uses getter/setter accessors (Property) or
     * used field based access (Field)
     * @return true if this is a property
     */
    public abstract boolean isProperty();
    
    /**
     * Accessor for the table name 
     * @return table name
     */
    public String getTable()
    {
        return table;
    }

    /**
     * Accessor for the field id.
     * Not set when the field is an overriding field.
     * @return field id
     */
    public int getFieldId()
    {
        return fieldId;
    }

    /**
     * Accessor for the implementation type(s) that can be stored in this field when it is a reference type.
     * When more than one type is valid will be comma-separated.
     * @return Returns the implementation type(s) for the field.
     */
    public final String getFieldType()
    {
        return fieldType;
    }
    
    /**
     * Accessor for the field id
     * @return field id
     */
    public int getAbsoluteFieldNumber()
    {
        if (className == null)
        {
            // Normal field, parented by its true class
            return fieldId + getAbstractClassMetaData().getNoOfInheritedManagedFields();
        }
        else
        {
            // Overriding field, parented by a foster class
            return getAbstractClassMetaData().getFieldNumberAbsolute(name);
        }
    }
    
    /**
     * Accessor for the field type
     * @return Reflection field type
     */
    public Class getType()
    {
        return type;
    }

    /**
     * Accessor for the field type name
     * @return Reflection field type name
     */
    public String getTypeName()
    {
        if (type == null)
        {
            return null;
        }
        return type.getName();
    }

    /**
     * Accessor for the container for this field.
     * @return The MetaData of the container for this field.
     **/
    public ContainerMetaData getContainer()
    {
        return container;
    }

    /**
     * Accessor for an array container for this field. Returns null if no array
     * attached.
     * @return The MetaData of the container for this field if an array
     **/
    public ArrayMetaData getArray()
    {
        if (container != null && container instanceof ArrayMetaData)
        {
            return (ArrayMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for a collection container for this field. Returns null if no
     * collection attached.
     * @return The MetaData of the container for this field if a Collection.
     **/
    public CollectionMetaData getCollection()
    {
        if (container != null && container instanceof CollectionMetaData)
        {
            return (CollectionMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for a map container for this field. Returns null if no map
     * attached.
     * @return The MetaData of the container for this field if a Map.
     **/
    public MapMetaData getMap()
    {
        if (container != null && container instanceof MapMetaData)
        {
            return (MapMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for the column 
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for mappedBy
     * @return Returns the mappedBy.
     */
    public final String getMappedBy()
    {
        return mappedBy;
    }

    /**
     * Acessor for the columns
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for elementMetaData
     * @return Returns the elementMetaData.
     */
    public final ElementMetaData getElementMetaData()
    {
        return elementMetaData;
    }

    /**
     * Accessor for keyMetaData
     * @return Returns the keyMetaData.
     */
    public final KeyMetaData getKeyMetaData()
    {
        return keyMetaData;
    }

    /**
     * Accessor for valueMetaData
     * @return Returns the valueMetaData.
     */
    public final ValueMetaData getValueMetaData()
    {
        return valueMetaData;
    }

    /**
     * Accessor for embeddedMetaData
     * @return Returns the embeddedMetaData.
     */
    public final EmbeddedMetaData getEmbeddedMetaData()
    {
        return embeddedMetaData;
    }

    /**
     * Accessor for foreignKeyMetaData
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Accessor for uniqueMetaData
     * @return Returns the uniqueMetaData.
     */
    public final UniqueMetaData getUniqueMetaData()
    {
        return uniqueMetaData;
    }

    /**
     * Accessor for joinMetaData
     * @return Returns the joinMetaData.
     */
    public final JoinMetaData getJoinMetaData()
    {
        return joinMetaData;
    }

    /**
     * Add a new ColumnMetaData element
     * @param colmd the ColumnMetaData to add
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
        colmd.parent = this;
        columnMetaData = new ColumnMetaData[columns.size()];
        for (int i=0; i<columnMetaData.length; i++)
        {
            columnMetaData[i] = (ColumnMetaData) columns.get(i);
        }
    }

    /**
     * Accessor for whether the field has a container.
     * @return Whether it represents a container.
     */
    public boolean hasContainer()
    {
        return (container != null);
    }

    /**
     * Accessor for whether the field has an array
     * @return return true if has array
     */
    public boolean hasArray()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof ArrayMetaData);
    }

    /**
     * Accessor for whether the field has a collection
     * @return return true if has collection
     */
    public boolean hasCollection()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof CollectionMetaData);
    }

    /**
     * Accessor for whether the field has a map.
     * @return return true if has map
     */
    public boolean hasMap()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof MapMetaData);
    }

    /**
     * Accessor for whether the field is to be managed by JDO.
     * @return Whether it is JDO managed
     */
    public boolean isJdoField()
    {
        if (isPopulated())
        {
            if (isStatic() || isFinal())
            {
                return false;
            }
        }

        if (persistenceModifier == null)
        {
            return false;
        }
        else if (persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            return false;
        }
        return true;
    }

    // ------------------------------ Mutators ---------------------------------

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     * @param valueStrategy the value strategy
     */
    public void setValueStrategy(IdentityStrategy valueStrategy)
    {
        this.valueStrategy = valueStrategy;
    }
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     * @param sequence the sequence name 
     */
    public void setSequence(String sequence)
    {
        this.sequence = sequence;
    }

    /**
     * Mutator for dependent attribute.
     * @param dependent Whether it is dependent.
     */
    public void setDependent(boolean dependent)
    {
        this.dependent = new Boolean(dependent);
    }

    /**
     * Mutator for mappedBy
     * @param mappedby The mapped-by field.
     */
    public void setMappedBy(String mappedby)
    {
        mappedBy = mappedby;
    }

    /**
     * Method to set the container for this field (if this field represents a
     * container (collection, map, array).
     * @param conmd The MetaData of the container for this field.
     **/
    public void setContainer(ContainerMetaData conmd)
    {
        container = conmd;
        container.parent = this;
    }

    /**
     * Mutator for the element MetaData 
     * @param elementMetaData The elementMetaData to set.
     */
    public final void setElementMetaData(ElementMetaData elementMetaData)
    {
        this.elementMetaData = elementMetaData;
        this.elementMetaData.parent = this;
    }

    /**
     * Mutator for the key MetaData 
     * @param keyMetaData The keyMetaData to set.
     */
    public final void setKeyMetaData(KeyMetaData keyMetaData)
    {
        this.keyMetaData = keyMetaData;
        this.keyMetaData.parent = this;
    }

    /**
     * Mutator for the order MetaData 
     * @param orderMetaData The orderMetaData to set.
     */
    public final void setOrderMetaData(OrderMetaData orderMetaData)
    {
        this.orderMetaData = orderMetaData;
        this.orderMetaData.parent = this;
    }

    /**
     * Mutator for the value MetaData 
     * @param valueMetaData The valueMetaData to set.
     */
    public final void setValueMetaData(ValueMetaData valueMetaData)
    {
        this.valueMetaData = valueMetaData;
        this.valueMetaData.parent = this;
    }

    /**
     * Mutator for the embedded MetaData 
     * @param embeddedMetaData The embeddedMetaData to set.
     */
    public final void setEmbeddedMetaData(EmbeddedMetaData embeddedMetaData)
    {
        this.embeddedMetaData = embeddedMetaData;
        this.embeddedMetaData.parent = this;
    }

    /**
     * Mutator for the foreignKey MetaData 
     * @param foreignKeyMetaData The foreignKeyMetaData to set.
     */
    public final void setForeignKeyMetaData(ForeignKeyMetaData foreignKeyMetaData)
    {
        this.foreignKeyMetaData = foreignKeyMetaData;
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    /**
     * Mutator for the unique MetaData 
     * @param uniqueMetaData The uniqueMetaData to set.
     */
    public final void setUniqueMetaData(UniqueMetaData uniqueMetaData)
    {
        this.uniqueMetaData = uniqueMetaData;
    }

    /**
     * Mutator for the join MetaData 
     * @param joinMetaData The joinMetaData to set.
     */
    public final void setJoinMetaData(JoinMetaData joinMetaData)
    {
        this.joinMetaData = joinMetaData;
        this.joinMetaData.parent = this;
    }

    /**
     * Mutator for the field id.
     * Given package access since updated by ClassMetaData typically.
     * Only used when the field is not an overriding field.
     * @param field_id Id of the field
     */
    void setFieldId(int field_id)
    {
        fieldId = field_id;
    }

    /**
     * Mutator for the table name
     * @param table The table name
     */
    public void setTable(String table)
    {
        this.table = table;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Convenience method that sets up the relation type of this field, and the reference to 
     * any related field when it is bidirectional. If the relation is bidirectional then will also
     * set the other side of the relation (to relate to this side).
     * @param clr ClassLoader resolver
     */
    protected void setRelation(ClassLoaderResolver clr)
    {
        if (getPersistenceModifier() != FieldPersistenceModifier.PERSISTENT)
        {
            // Only of use for persistent fields
            relationType = Relation.NONE;
            relatedFieldMetaData = null;
            return;
        }

        // Find the metadata for the field object
        AbstractClassMetaData otherCmd = null;
        if (hasCollection())
        {
            otherCmd = getCollection().getElementClassMetaData();
            if (otherCmd == null)
            {
                // Maybe a reference field
                Class elementCls = clr.classForName(getCollection().getElementType());
                if (getMetaDataManager().getPMFContext().getTypeManager().isReferenceType(elementCls))
                {
                    String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(this, 
                        DatastoreField.ROLE_COLLECTION_ELEMENT, clr);
                    if (implNames != null && implNames.length > 0)
                    {
                        // Take the first implementation
                        otherCmd = getMetaDataManager().getMetaDataForClass(implNames[0], clr);
                    }
                }
            }
        }
        else if (hasMap())
        {
            otherCmd = ((MapMetaData) container).getValueClassMetaData();
            //TODO [CORE-2585] valueCMD may be null because its type is an interface (non persistent interface), 
            //so we should handle the implementation classes
            if (otherCmd == null)
            {
                // Value not PC so use the Key if it is specified
                otherCmd = ((MapMetaData)container).getKeyClassMetaData();
            }
            if (otherCmd == null)
            {
                // Maybe a reference key/value
            }
        }
        else if (hasArray())
        {
            otherCmd = ((ArrayMetaData)container).getElementClassMetaData();
        }
        else
        {
            if (getType().isInterface())
            {
                // Reference field - take the metadata of the first implementation
                String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(this, DatastoreField.ROLE_FIELD, clr);
                if (implNames != null && implNames.length > 0)
                {
                    otherCmd = getMetaDataManager().getMetaDataForClass(implNames[0], clr);
                }
                /*otherCmd = getMetaDataManager().getMetaDataForImplementationOfReference(getType(), null, clr);*/
                /*otherCmd = getAbstractClassMetaData().getMetaDataManager().getMetaDataForInterface(getType(), clr);*/
            }
            else
            {
                otherCmd = getMetaDataManager().getMetaDataForClass(getType(), clr);
            }
        }

        //TODO [CORE-2585] when the element or value type is an interface (non persistent interface), 
        //we should look the implementation classes for the "otherCmd"
        //for now the relationType will be Relation.NONE in these cases because the otherCmd is null
        if (otherCmd == null)
        {
            // Field cannot have a relation
            relatedFieldMetaData = null;
            relationType = Relation.NONE;
        }
        else
        {
            // Field is relational
            if (mappedBy != null)
            {
                // This class has the "mapped-by" specified
                relatedFieldMetaData = otherCmd.getField(mappedBy);
                if (relatedFieldMetaData == null)
                {
                    // "mapped-by" field doesnt exist at the target
                    throw new JDOFatalUserException(LOCALISER.msg("MetaData.Field.MappedByInvalid", 
                        getAbstractClassMetaData().getFullClassName(), name, mappedBy, otherCmd.getFullClassName()));
                }
                relatedFieldMetaData.relatedFieldMetaData = this;
                if (hasContainer() && relatedFieldMetaData.hasContainer())
                {
                    relationType = Relation.MANY_TO_MANY_BI;
                    relatedFieldMetaData.relationType = Relation.MANY_TO_MANY_BI;
                }
                else if (hasContainer() && !relatedFieldMetaData.hasContainer())
                {
                    relationType = Relation.ONE_TO_MANY_BI;
                    relatedFieldMetaData.relationType = Relation.MANY_TO_ONE_BI;
                }
                else if (!hasContainer() && relatedFieldMetaData.hasContainer())
                {
                    relationType = Relation.MANY_TO_ONE_BI;
                    relatedFieldMetaData.relationType = Relation.ONE_TO_MANY_BI;
                }
                else
                {
                    relationType = Relation.ONE_TO_ONE_BI;
                    relatedFieldMetaData.relationType = Relation.ONE_TO_ONE_BI;
                }
            }
            else
            {
                // The "other" class maybe has "mapped-by" across to this field so navigate through the fields to find this one
                int[] otherFieldNumbers = otherCmd.getAllFieldNumbers();
                for (int i=0;i<otherFieldNumbers.length;i++)
                {
                    AbstractPropertyMetaData otherFmd = otherCmd.getManagedFieldAbsolute(otherFieldNumbers[i]);
                    if (otherFmd.getMappedBy() != null && otherFmd.getMappedBy().equals(name))
                    {
                        if (otherFmd.hasContainer())
                        {
                            if ((otherFmd.hasCollection() && otherFmd.getCollection().getElementType().equals(getClassName(true))) ||
                                (otherFmd.hasArray() && otherFmd.getArray().getElementType().equals(getClassName(true))) ||
                                (otherFmd.hasMap() && otherFmd.getMap().getKeyType().equals(getClassName(true))) ||
                                (otherFmd.hasMap() && otherFmd.getMap().getValueType().equals(getClassName(true))))
                            {
                                relatedFieldMetaData = otherFmd;
                                relatedFieldMetaData.relatedFieldMetaData = this;
                                if (hasContainer())
                                {
                                    // Should we mark Arrays, Lists, Maps as M-N ?
                                    relationType = Relation.MANY_TO_MANY_BI;
                                    relatedFieldMetaData.relationType = Relation.MANY_TO_MANY_BI;
                                    break;
                                }
                                else
                                {
                                    relationType = Relation.MANY_TO_ONE_BI;
                                    relatedFieldMetaData.relationType = Relation.ONE_TO_MANY_BI;
                                    break;
                                }
                            }
                        }
                        else if (otherFmd.getTypeName().equals(getClassName(true)))
                        {
                            relatedFieldMetaData = otherFmd;
                            relatedFieldMetaData.relatedFieldMetaData = this;
                            if (hasContainer())
                            {
                                relationType = Relation.ONE_TO_MANY_BI;
                                relatedFieldMetaData.relationType = Relation.MANY_TO_ONE_BI;
                            }
                            else
                            {
                                relationType = Relation.ONE_TO_ONE_BI;
                                relatedFieldMetaData.relationType = Relation.ONE_TO_ONE_BI;
                            }
                            break;
                        }
                    }
                }

                if (relatedFieldMetaData == null)
                {
                    // No "mapped-by" found at either end so is unidirectional
                    if (hasContainer())
                    {
                        relationType = Relation.ONE_TO_MANY_UNI;
                    }
                    else
                    {
                        relationType = Relation.ONE_TO_ONE_UNI;
                    }
                }
            }
        }
    }

    /**
     * Accessor for the relation type for this field.
     * @param clr ClassLoader resolver
     * @return The relation type.
     */
    public int getRelationType(ClassLoaderResolver clr)
    {
        if (relationType == -1)
        {
            // Is possible that this could be done in the populate() step but depends on availability of the related class
            setRelation(clr);
        }
        return relationType;
    }

    /**
     * Accessor for the FieldMetaData of any related field (where this field is
     * part of a bidirectional relation). Allows for 1-1, 1-N, and M-N. If this field
     * is not part of a bidirectional relation (no "mapped-by" at either end) then it
     * returns null.
     * @param clr the ClassLoaderResolver
     * @return The FieldMetaData for the "other end".
     */
    public AbstractPropertyMetaData getRelatedFieldMetaData(ClassLoaderResolver clr)
    {
        if (relationType == -1)
        {
            // Is possible that this could be done in the populate() step but depends on availability of the related class
            setRelation(clr);
        }
        // TODO This doesnt allow for 1-1, 1-N reference fields where there are likely multiple "other fields".
        return relatedFieldMetaData;
    }

    /**
     * Accessor for all ClassMetaData referenced by this Field.
     * <p>
     * Part of the "persistence-by-reachability" concept. 
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of referenced ClassMetaData objects (added to)
     * @param dba_vendor_id Vendor id of the DBA
     * @param clr the ClassLoaderResolver
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        AbstractClassMetaData type_cmd=getAbstractClassMetaData().getMetaDataManager().getMetaDataForClass(getType(), clr);
        if (type_cmd != null)
        {
            type_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }

        if (container != null)
        {
            if (container instanceof CollectionMetaData)
            {
                ((CollectionMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
            else if (container instanceof MapMetaData)
            {
                ((MapMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
            else if (container instanceof ArrayMetaData)
            {
                ((ArrayMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
        }
    }

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }
    
    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        return super.toString(prefix, indent);
    }

    /**
     * Comparator method. This allows the ClassMetaData to search for a
     * FieldMetaData with a particular name.
     * @param o The object to compare against
     * @return The comparison result
     */ 
    public int compareTo(Object o)
    {
        if (o instanceof AbstractPropertyMetaData)
        {
            AbstractPropertyMetaData c = (AbstractPropertyMetaData)o;
            return this.name.compareTo(c.name);
        }
        else if (o instanceof String)
        {
            return this.name.compareTo((String)o);
        }
        else if (o == null)
        {
            throw new ClassCastException("object is null");
        }
        throw new ClassCastException(this.getClass().getName() + " != " + o.getClass().getName());
    }
    
    /**
     * Convenience accessor for the MetaData Manager in use.
     * @return MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        if (parent != null && parent.getParent() != null && getAbstractClassMetaData() != null)
        {
            return getAbstractClassMetaData().getPackageMetaData().getFileMetaData().getMetaDataManager();
        }
        return null;
    }    
}
