/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigClass
2004 Marco Schulze (NightLabs) - changed the behaviour to warn only if
        an inherited class declares an own objectid and it equals the
        one of the superclass.
2004 Andy Jefferson - Added discriminator/inheritance checks
2004 Erik Bengtson - changes for application identity
2004 Andy Jefferson - moved PK class checks out into JDOUtils
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.AIDUtils;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;

/**
 * Representation of the MetaData of a class. Extends the abstract definition to include
 * implementations, fields, embedded-only tags. Has a parent PackageMetaData that
 * can contain the metadata for several classes.
 *
 * <H3>Lifecycle state</H3>
 * This object supports 3 lifecycle states. The first is the raw
 * constructed object which represents pure MetaData (maybe from a MetaData
 * file). The second is a "populated" object which represents MetaData for a
 * Class with the metadata aligned to be appropriate for that Class.
 * The third is "initialised" once the internal arrays are created.
 * This object, once populated, will represent ALL fields in the class
 * (including static, final and transient fields).
 *
 * <H3>Fields</H3>
 * This object keeps a list of FieldMetaData objects for the fields of this
 * class (stored as "fields"). In addition it has an array of FieldMetaData
 * objects representing those that are actually managed by JDO
 * ("managedFields"). This second set does not contain things like static, final
 * or transient fields since JDO doesn't support those yet.
 * <P>Fields are of 2 types. The first are normal fields of this class.
 * These have their own "relative" field number, relative to this class.
 * The second type are "overriding" fields which override the baseline field
 * in a superclass. These fields have no "relative" field number since they are
 * relative to this class (and such a relative field number would make no sense).
 * Fields are all added through addField() during the parse process, and
 * are updated during the populate/initialise process to define their relative field
 * numbers. Please refer to FieldMetaData for more details of fields.
 *
 * <H3>Numbering of fields</H3>
 * Fields of the class are numbered in 2 ways. The first way is the numbering
 * within a class. In a class, the field 'id's will start at 0. If a class is
 * inherited, it will also have a second numbering for its fields - the
 * "absolute" numbering. With "absolute" numbering, the fields start at the
 * first field in the root superclass which has absolute number 0, and they are
 * numbered from there, navigating down the hierarchy. In terms of what is
 * stored in the records, the FieldMetaData stores fieldId as the first
 * method (relative to the class it is in). The "absolute" numbering is 
 * always derived from this and the inheritance hierarchy.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT class (datastore-identity?, implements*, inheritance?, version?, join*,
 *      foreign-key*, index*, unique*, field*, column*, query*, fetch-group*, extension*)&gt;
 * &lt;!ATTLIST class name CDATA #REQUIRED&gt;
 * &lt;!ATTLIST class identity-type (application|datastore|nondurable) #IMPLIED&gt;
 * &lt;!ATTLIST class catalog CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class schema CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class persistence-capable-superclass CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class objectid-class CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class requires-extent (true|false) 'true'&gt;
 * &lt;!ATTLIST class detachable (true|false) 'true'&gt;
 * &lt;!ATTLIST class embedded-only (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST class persistence-modifier 
 *      (persistence-capable|persistence-aware|non-persistent) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.102 $
 */
public class ClassMetaData extends AbstractClassMetaData
{
    /**
     * List of implements
     */
    protected List implementations = new ArrayList();

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * ImplementsMetaData
     */
    protected ImplementsMetaData[] implementsMetaData;

    /**
     * is abstract the PersistenceCapable class
     */
    protected boolean isAbstractPersistenceCapable;
    
    /** whether the populate method is running **/
    private boolean populating = false;

    // ----------------------------- Constructors ------------------------------
 
    /**
     * Constructor.
     * Takes the basic string information found in the MetaData file.
     * @param parent The package to which this class belongs
     * @param name Name of class
     * @param identityType identity-type flag
     * @param objectidClass Primary key class name
     * @param requiresExtent Whether the class requires an extent
     * @param detachable Whether the class can be detached
     * @param modifier persistence-modifier tag
     * @param embeddedOnly embedded-only tag
     * @param persistenceCapableSuperclass Name of PC superclass
     * @param catalog Name for catalog
     * @param schema Name for schema
     * @param table RDBMS table to store the class in
     * @param entityName the entity name required by JPA 4.3.1
     */
    public ClassMetaData(final PackageMetaData parent,
                         final String name,
                         final String identityType,
                         final String objectidClass,
                         final String requiresExtent,
                         final String detachable,
                         final String embeddedOnly,
                         final String modifier,
                         final String persistenceCapableSuperclass,
                         final String catalog,
                         final String schema,
                         final String table,
                         final String entityName)
    {
        super(parent, name, identityType, objectidClass, requiresExtent, detachable, embeddedOnly, modifier, persistenceCapableSuperclass,
            catalog, schema, table,entityName);
    }

    /**
     * Constructor for creating the ClassMetaData for an implementation of a "persistent-interface".
     * @param imd MetaData for the "persistent-interface"
     * @param implClassName Name of the implementation class
     * @param copyFields Whether to copy the fields of the interface too
     */
    public ClassMetaData(final InterfaceMetaData imd, String implClassName, boolean copyFields)
    {
        super(imd, implClassName, copyFields);
    }

    /**
     * Method to provide the details of the class being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * class being represented.
     * <P>
     * One possible use of this method would be to take a basic ClassMetaData
     * for a class and call this, passing in the users class. This would then
     * add FieldMetaData for all fields in this class providing defaults for
     * all of these.
     *
     * @param clr ClassLoaderResolver to use in loading any classes
     * @param primary the primary ClassLoader to use (or null)
     */
    public synchronized void populate(ClassLoaderResolver clr, ClassLoader primary)
    {
        if (isInitialised() || isPopulated())
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",name));
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",fullName));
        }
        if (populating)
        {
            return;
        }

        try
        {
            if (JPOXLogger.METADATA.isDebugEnabled())
            {
                JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.PopulatingMetaData",fullName));
            }
            populating = true;

            Class cls = loadClass(clr,primary);

            isAbstractPersistenceCapable = Modifier.isAbstract(cls.getModifiers());
            
            // Load any ORM definition for this class
            getMetaDataManager().addORMDataToClass(cls, clr);

            // If a class is an inner class and is non-static it is invalid
            if (ClassUtils.isInnerClass(fullName) && !Modifier.isStatic(cls.getModifiers()) &&
                persistenceModifier == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
            {
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.NonStaticInnerClassNotPersistable", fullName);
            }

            determineSuperClassName(clr, cls);

            inheritDetachableSettings();
            
            inheritIdentity();
            
            determineIdentity();
            
            validateUserInputForIdentity();
            
            addPropertiesNotInMetaData(cls);

            if (objectidClass == null)
            {
                // No user-defined objectid-class but potentially have SingleFieldIdentity so make sure PK fields are set
                populatePropertyMetaData(clr, cls, true); // Make sure all PK fields (and superclasses) are set before objectid-class
                determineObjectIdClass(clr);
                populatePropertyMetaData(clr, cls, false); // Populate non-PK fields
            }
            else
            {
                populatePropertyMetaData(clr, cls, true);
                populatePropertyMetaData(clr, cls, false);
                determineObjectIdClass(clr);
            }

            validateUserInputForInheritanceMetaData();
            
            determineInheritanceMetaData();

            validateDeprecatedMetaData();

            // populate the implements
            for(int i=0; i<implementations.size(); i++)
            {
                ((ImplementsMetaData)implementations.get(i)).populate(clr);
            }

            if (persistentInterfaceImplNeedingTableFromSuperclass)
            {
                // Need to go up to next superinterface and make sure its metadata is populated
                // until we find the next interface with metadata with inheritance strategy of "new-table".
                AbstractClassMetaData acmd = getMetaDataForSuperinterfaceManagingTable(cls, clr);
                if (acmd != null)
                {
                    table = acmd.table;
                    schema = acmd.schema;
                    catalog = acmd.catalog;
                }
                persistentInterfaceImplNeedingTableFromSuperclass = false;
            }
            else if (persistentInterfaceImplNeedingTableFromSubclass)
            {
                // TODO Cater for finding the subclass-table that manages our table
                persistentInterfaceImplNeedingTableFromSubclass = false;
            }

            setPopulated();
        }
        catch (RuntimeException e)
        {
            JPOXLogger.METADATA.debug(e);
            throw e;
        }
        finally
        {
            populating = false;
        }
    }

    /**
     * Method to find a superinterface with MetaData that specifies NEW_TABLE inheritance strategy
     * @param cls The class
     * @param clr ClassLoader resolver
     * @return The AbstractClassMetaData for the class managing the table
     */
    private AbstractClassMetaData getMetaDataForSuperinterfaceManagingTable(Class cls, ClassLoaderResolver clr)
    {
        Collection superinterfaces = ClassUtils.getSuperinterfaces(cls);
        Iterator iter = superinterfaces.iterator();
        while (iter.hasNext())
        {
            Class superintf = (Class)iter.next();
            AbstractClassMetaData acmd = getMetaDataManager().getMetaDataForInterface(superintf, clr);
            if (acmd != null && acmd.getInheritanceMetaData() != null)
            {
                if (acmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
                {
                    // Found it
                    return acmd;
                }
                else if (acmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
                {
                    // Try further up the hierarchy
                    return getMetaDataForSuperinterfaceManagingTable(superintf, clr);
                }
            }
        }
        return null;
    }

    /**
     * Add FieldMetaData of fields not declared in MetaData
     * @param cls This class
     */
    protected void addPropertiesNotInMetaData(Class cls)
    {
        // Add fields for the class that aren't in the FieldMetaData.
        // We use Reflection here since JDOImplHelper would only give use info
        // for enhanced files (and the enhancer needs unenhanced as well). 
        // NOTE 1 : We ignore fields in superclasses
        // NOTE 2 : We ignore "enhanced" fields (added by the JDO enhancer)
        // NOTE 3 : We ignore inner class fields (containing "$") 
        // NOTE 4 : We sort the fields into ascending alphabetical order
        Collections.sort(properties);
        try
        {
            // Get all (reflected) fields in the populating class
            Field[] cls_fields=cls.getDeclaredFields();
            for (int i=0;i<cls_fields.length;i++)
            {
                // Limit to fields in this class, that aren't enhanced fields
                // that aren't inner class fields, and that aren't static
                if (cls_fields[i].getDeclaringClass().getName().equals(fullName) &&
                    !cls_fields[i].getName().startsWith("jdo") &&
                    !ClassUtils.isInnerClass(cls_fields[i].getName()) &&
                    !Modifier.isStatic(cls_fields[i].getModifiers()))
                {
                    // Find if there is a FieldMetaData for this field.
                    // This is possible as FieldMetaData implements Comparable
                    if (Collections.binarySearch(properties,cls_fields[i].getName()) < 0)
                    {
                        // Add a default FieldMetaData for this field.
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.AddingUnspecifiedField",cls_fields[i].getName(),name));
                        AbstractPropertyMetaData fmd = newDefaultedProperty(cls_fields[i].getName());
                        properties.add(fmd);
                        Collections.sort(properties);
                    }
                }
            }
        }
        catch (Exception e)
        {
            JPOXLogger.METADATA.error(e.getMessage(), e);
            throw new RuntimeException(e.getMessage());
        }
    }

    /**
     * Populate FieldMetaData
     * @param clr The ClassLoaderResolver
     * @param cls This class
     * @param pkFields Process pk fields (or non-PK fields if false)
     * @throws InvalidMetaDataException if the Class for a declared type in a field cannot be loaded by the <code>clr</code>
     * @throws InvalidMetaDataException if a field declared in the MetaData does not exist in the Class
     */
    protected void populatePropertyMetaData(ClassLoaderResolver clr, Class cls, boolean pkFields)
    {
        Collections.sort(properties);
        
        // Populate the FieldMetaData with their real field values
        // This will populate any containers in these fields also
        Iterator fields_iter=properties.iterator();
        while (fields_iter.hasNext())
        {
            AbstractPropertyMetaData fmd=(AbstractPropertyMetaData)fields_iter.next();
            if (pkFields == fmd.isPrimaryKey())
            {
                Class fieldCls = cls;
                if (!fmd.fieldBelongsToClass())
                {
                    // Field overrides a field in a superclass, so find the class
                    try
                    {
                        fieldCls = clr.classForName(fmd.getClassName());
                    }
                    catch (ClassNotResolvedException cnre)
                    {
                        // Not found at specified location, so try the same package as this class
                        String fieldClassName = getPackageName() + "." + fmd.getClassName();
                        try
                        {
                            fieldCls = clr.classForName(fieldClassName);
                            fmd.setClassName(fieldClassName);
                        }
                        catch (ClassNotResolvedException cnre2)
                        {
                            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.ClassNotFoundError", fieldClassName));
                            throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.ClassNotFoundError", fieldClassName);
                        }
                    }
                }

                Field cls_field=null;
                try
                {
                    cls_field = fieldCls.getDeclaredField(fmd.getName());
                }
                catch (Exception e)
                {
                    // MetaData field doesn't exist in the class!
                    throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.DeclaredFieldNotFoundError",
                        fullName,fmd.getFullFieldName());
                }
                fmd.populate(clr, cls_field.getType(), cls_field.getModifiers());
            }
        }
    }
    
    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects. populate() should be called BEFORE calling this.
     */
    public synchronized void initialise()
    {
        if (populating)
        {
            return;
        }
        checkPopulated();
        if (isInitialised())
        {
            return;
        }

        if (pcSuperclassMetaData != null)
        {
            // We need our superclass to be initialised before us because we rely on information there
            if (!pcSuperclassMetaData.isInitialised())
            {
                pcSuperclassMetaData.initialise();
            }
        }

        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.InitialisingMetaData",fullName));
        }

        // Count the fields of the relevant category
        Iterator fields_iter = properties.iterator();
        int no_of_managed_fields = 0;
        int no_of_overridden_fields = 0;
        while (fields_iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)fields_iter.next();

            // Initialise the FieldMetaData (and its sub-objects)
            fmd.initialise();
            if (fmd.isJdoField())
            {
                if (fmd.fieldBelongsToClass())
                {
                    no_of_managed_fields++;
                }
                else
                {
                    no_of_overridden_fields++;
                }
            }
        }

        // Generate the "managed fields" list
        managedFields = new AbstractPropertyMetaData[no_of_managed_fields];
        overriddenFields = new AbstractPropertyMetaData[no_of_overridden_fields];

        fields_iter = properties.iterator();
        int field_id = 0;
        int overridden_field_id = 0;
        fieldNumbersByName = new HashMap();
        while (fields_iter.hasNext())
        {
            AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)fields_iter.next();

            if (fmd.isJdoField())
            {
                if (fmd.fieldBelongsToClass())
                {
                    fmd.setFieldId(field_id);
                    managedFields[field_id] = fmd;
                    fieldNumbersByName.put(fmd.getName(),new Integer(field_id));
                    field_id++;
                }
                else
                {
                    overriddenFields[overridden_field_id++] = fmd;
                }
            }
        }

        if (pcSuperclassMetaData != null)
        {
            if (!pcSuperclassMetaData.isInitialised())
            {
                pcSuperclassMetaData.initialise();
            }
            noOfInheritedManagedFields = pcSuperclassMetaData.getNoOfInheritedManagedFields() + pcSuperclassMetaData.getNoOfManagedFields();
        }

        // Set up the various convenience arrays of field numbers
        int total_field_count=noOfInheritedManagedFields + managedFields.length;
        allFieldNumbers = new int[total_field_count];
        defaultFetchGroupFieldFlags = new boolean[total_field_count];
        secondClassMutableFieldFlags = new boolean[total_field_count];
        nonPrimaryKeyFieldFlags = new boolean[total_field_count];
        int pk_field_count=0;
        int dfg_field_count=0;
        int scm_field_count=0;
        int pc_field_count=0;
        for (int i=0;i<total_field_count;i++)
        {
            allFieldNumbers[i] = i;

            AbstractPropertyMetaData fmd=getManagedFieldAbsoluteInternal(i);
            if (fmd.isPrimaryKey())
            {
                pk_field_count++;
            }
            else
            {
                nonPrimaryKeyFieldFlags[i] = true;
            }
            if (fmd.isDefaultFetchGroup())
            {
                defaultFetchGroupFieldFlags[i] = true;
                dfg_field_count++;
            }
            if (getMetaDataManager().getPMFContext().getTypeManager().isSecondClassMutableType(fmd.getTypeName()))
            {
                secondClassMutableFieldFlags[i] = true;
                scm_field_count++;
            }
            if (fmd.isFieldTypePersistenceCapable())
            {
                pc_field_count++;
            }
        }
 
        if (pk_field_count > 0 &&
            identityType != IdentityType.APPLICATION)
        {
            // primary key fields found, but not using application identity
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyFieldsFoundButNotUsingApplicationIdentity",
                fullName,new Integer(pk_field_count),identityType);
        }
        else if (pk_field_count > 0)
        {
            primaryKeyFieldNumbers = new int[pk_field_count];
            for (int i=0,pk_num=0;i<total_field_count;i++)
            {
                AbstractPropertyMetaData fmd=getManagedFieldAbsoluteInternal(i);
                if (fmd.isPrimaryKey())
                {
                    primaryKeyFieldNumbers[pk_num++] = i;
                }
            }
        }
        else if (pk_field_count == 0 &&
            identityType == IdentityType.APPLICATION)
        {
            // No primary key fields found
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyNotSpecifiedError",
                fullName,objectidClass);
        }

        nonPrimaryKeyFieldNumbers = new int[total_field_count-pk_field_count];
        persistenceCapableFieldNumbers = new int[pc_field_count];
        for (int i=0,npkf=0,npc=0;i<total_field_count;i++)
        {
            AbstractPropertyMetaData fmd = getManagedFieldAbsoluteInternal(i);
            if (!fmd.isPrimaryKey())
            {
                nonPrimaryKeyFieldNumbers[npkf++] = i;
            }
            if( fmd.isFieldTypePersistenceCapable() )
            {
                persistenceCapableFieldNumbers[npc++] = i;
            }
        }

        defaultFetchGroupFieldNumbers = new int[dfg_field_count];
        secondClassMutableFieldNumbers = new int[scm_field_count];
        for (int i=0,dfg_num=0,scm_num=0;i<total_field_count;i++)
        {
            if (defaultFetchGroupFieldFlags[i])
            {
                defaultFetchGroupFieldNumbers[dfg_num++] = i;
            }
            if (secondClassMutableFieldFlags[i])
            {
                secondClassMutableFieldNumbers[scm_num++] = i;
            }
        }

        // Initialise any sub-objects
        implementsMetaData = new ImplementsMetaData[implementations.size()];
        for (int i=0; i<implementations.size(); i++)
        {
            implementsMetaData[i] = (ImplementsMetaData) implementations.get(i);
            implementsMetaData[i].initialise();
        }
        joinMetaData = new JoinMetaData[joins.size()];
        for (int i=0; i<joinMetaData.length; i++)
        {
            joinMetaData[i] = (JoinMetaData) joins.get(i);
            joinMetaData[i].initialise();
        }
        indexMetaData = new IndexMetaData[indexes.size()];
        for (int i=0; i<indexMetaData.length; i++)
        {
            indexMetaData[i] = (IndexMetaData) indexes.get(i);
            indexMetaData[i].initialise();
        }
        foreignKeyMetaData = new ForeignKeyMetaData[foreignKeys.size()];
        for (int i=0; i<foreignKeyMetaData.length; i++)
        {
            foreignKeyMetaData[i] = (ForeignKeyMetaData) foreignKeys.get(i);
            foreignKeyMetaData[i].initialise();
        }
        uniqueMetaData = new UniqueMetaData[uniqueConstraints.size()];
        for (int i=0; i<uniqueMetaData.length; i++)
        {
            uniqueMetaData[i] = (UniqueMetaData) uniqueConstraints.get(i);
            uniqueMetaData[i].initialise();
        }

        fetchGroupMetaData = new FetchGroupMetaData[fetchGroups.size()];
        fetchGroupMetaDataByName = new HashMap();
        for (int i=0; i<fetchGroupMetaData.length; i++)
        {
            fetchGroupMetaData[i] = (FetchGroupMetaData) fetchGroups.get(i);
            fetchGroupMetaData[i].initialise();
            fetchGroupMetaDataByName.put(fetchGroupMetaData[i].getName(), fetchGroupMetaData[i]);
        }         
        
        // If using datastore id and user hasn't provided the identity element,
        // add a defaulted one (using the superclass if available)
        if (identityType == IdentityType.DATASTORE && identityMetaData == null)
        {
            if (pcSuperclassMetaData != null)
            {
                IdentityMetaData superImd = pcSuperclassMetaData.getIdentityMetaData();
                identityMetaData = new IdentityMetaData(this, superImd.getColumn(), superImd.getStrategyValue().toString(), superImd.getSequence());
            }
            else
            {
                identityMetaData = new IdentityMetaData(this,null,null,null);
            }
        }

        if (versionMetaData != null)
        {
            versionMetaData.initialise();
        }
        if (identityMetaData != null)
        {
            identityMetaData.initialise();
        }
        if (inheritanceMetaData != null)
        {
            inheritanceMetaData.initialise();
        }

        if (identityType == IdentityType.APPLICATION)
        {
            usesSingleFieldIdentityClass = AIDUtils.isSingleFieldIdentityClass(getObjectidClass());
        }

        // Clear out the collections that we used when loading the MetaData
        joins.clear();
        joins = null;
        fetchGroups.clear();
        fetchGroups = null;
        foreignKeys.clear();
        foreignKeys = null;
        indexes.clear();
        indexes = null;
        uniqueConstraints.clear();
        uniqueConstraints = null;
        implementations.clear();
        implementations = null;
        super.initialise();
        setInitialised();
    }

    /**
     * Utility to add a defaulted FieldMetaData to the class. Provided as a
     * method since then any derived classes can override it (e.g
     * JDOConfigClass can create a JDOConfigField)
     * @param name name of field
     * @return the new FieldMetaData
     */
    protected AbstractPropertyMetaData newDefaultedProperty(String name)
    {
        return new FieldMetaData(this, name);
    }

    // ------------------------------ Accessors --------------------------------

    /**
     * Accessor for the implements MetaData
     * @return Returns the implements MetaData.
     */
    public final ImplementsMetaData[] getImplementsMetaData()
    {
        return implementsMetaData;
    }

    /**
     * Convenience accessor for whether this class implements a specified interface
     * @param interfaceName The name of the interface
     * @return Whether it implements the interface
     */
    public boolean implementsInterface(String interfaceName)
    {
        if (implementsMetaData != null)
        {
            for (int i=0;i<implementsMetaData.length;i++)
            {
                if (implementsMetaData[i].getName().equals(interfaceName))
                {
                    return true;
                }
            }
        }
        return false;
    }


    /**
     * Whether the PersistenceCapable class is abstract
     * @return true if the PersistenceCapable class is abstract
     */
    public boolean isAbstractPersistenceCapable()
    {
        return isAbstractPersistenceCapable;
    }   
    
    // ------------------------------- Mutators --------------------------------

    /**
     * Method to add an implements to this class.
     * @param implmd Meta-Data for the implements
     */
    public void addImplements(ImplementsMetaData implmd)
    {
        if (implmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        implementations.add(implmd);
    }

    // ------------------------------ Utilities --------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<class name=\"" + name + "\"\n");
        if (identityType != null)
        {
            sb.append(prefix).append("       identity-type=\"" + identityType + "\"\n");
        }
        if (objectidClass != null)
        {
            sb.append(prefix).append("       objectid-class=\"" + objectidClass + "\"\n");
        }
        if (!requiresExtent)
        {
            sb.append(prefix).append("       requires-extent=\"" + requiresExtent + "\"");
        }
        if (embeddedOnly)
        {
            sb.append(prefix).append("       embedded-only=\"" + embeddedOnly + "\"\n");
        }
        if (persistenceModifier != null)
        {
            sb.append(prefix).append("       persistence-modifier=\"" + persistenceModifier + "\"\n");
        }
        if (catalog != null)
        {
            sb.append(prefix).append("       catalog=\"" + catalog + "\"\n");
        }
        if (schema != null)
        {
            sb.append(prefix).append("       schema=\"" + schema + "\"\n");
        }
        if (table != null)
        {
            sb.append(prefix).append("       table=\"" + table + "\"\n");
        }
        if (detachable)
        {
            sb.append(prefix).append("       detachable=\"" + detachable + "\">\n");
        }

        // Implements
        if (implementsMetaData != null)
        {
            for (int i=0; i<implementsMetaData.length; i++)
            {
                sb.append(implementsMetaData[i].toString(prefix + indent, indent));
            }
        }

        // Identity
        if (identityMetaData != null)
        {
            sb.append(identityMetaData.toString(prefix + indent,indent));
        }

        // PrimaryKey
        if (primaryKeyMetaData != null)
        {
            sb.append(primaryKeyMetaData.toString(prefix + indent,indent));
        }

        // Inheritance
        if (inheritanceMetaData != null)
        {
            sb.append(inheritanceMetaData.toString(prefix + indent,indent));
        }

        // Add Version
        if (versionMetaData != null)
        {
            sb.append(versionMetaData.toString(prefix + indent,indent));
        }

        // Add joins
        if (joinMetaData != null)
        {
            for (int i=0; i<joinMetaData.length; i++)
            {
                sb.append(joinMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add foreign-keys
        if (foreignKeyMetaData != null)
        {
            for (int i=0; i<foreignKeyMetaData.length; i++)
            {
                sb.append(foreignKeyMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add indexes
        if (indexMetaData != null)
        {
            for (int i=0; i<indexMetaData.length; i++)
            {
                sb.append(indexMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add unique constraints
        if (uniqueMetaData != null)
        {
            for (int i=0; i<uniqueMetaData.length; i++)
            {
                sb.append(uniqueMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add fields
        if (managedFields != null)
        {
            for (int i=0; i<managedFields.length; i++)
            {
                sb.append(managedFields[i].toString(prefix + indent,indent));
            }
        }

        // Add unmapped columns
        if (unmappedColumns != null)
        {
            for (int i=0;i<unmappedColumns.size();i++)
            {
                ColumnMetaData col = (ColumnMetaData)unmappedColumns.get(i);
                sb.append(col.toString(prefix + indent, indent));
            }
        }

        // Add queries
        if (queries != null)
        {
            Iterator iter = queries.iterator();
            while (iter.hasNext())
            {
                QueryMetaData q = (QueryMetaData)iter.next();
                sb.append(q.toString(prefix + indent,indent));
            }
        }

        // Add fetch-groups
        if (fetchGroupMetaData != null)
        {
            for (int i=0; i<fetchGroupMetaData.length; i++)
            {
                sb.append(fetchGroupMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent)); 

        sb.append(prefix + "</class>\n");
        return sb.toString();
    }
}