/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.HasExtension
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Base class of any extendable meta-data. In JDO typically all types of
 * meta-data are extendable so extend this. JPOX has a vendor-name of 
 * "jpox" but also supports "JPOX", "JPox", etc.
 *
 * @since 1.1
 * @version $Revision: 1.8 $
 */
public class ExtendableMetaData extends MetaData
{
    /** Vendor name used by JPOX. */
    public static final String JPOX_VENDOR_NAME="jpox";

    /**
     * List of extensions.
     */
    protected List extensions = new ArrayList();

    /**
     * default constructor
     * @param parent The parent of this extendable element
     */
    public ExtendableMetaData(MetaData parent)
    {
        super(parent);
    }

    // -------------------------------- Accessors ------------------------------
 
    /**
     * Accessor for the number of extensions.
     * @return extension length
     */
    public int getNoOfExtensions()
    {
        return extensions.size();
    }

    /**
     * Accessor for the extensions.
     * @return The extensions.
     **/
    public ExtensionMetaData[] getExtensions()
    {
        if (extensions.size() == 0)
        {
            return null;
        }
        ExtensionMetaData result[] = new ExtensionMetaData[extensions.size()];

        Iterator iter=extensions.iterator();
        int i=0;
        while (iter.hasNext())
        {
            result[i++] = (ExtensionMetaData)iter.next();
        }
        return result;
    }

    /**
     * Accessor for an extension
     * @param i index number
     * @return ExtensionMetaData instance
     */
    public ExtensionMetaData getExtension(int i)
    {
        return (ExtensionMetaData)extensions.get(i);
    }

    /**
     * Accessor for whether an extension exists (for JPOX).
     * @param key The key of the extension
     * @return Whether the extension exists
     */
    public boolean hasExtension(String key)
    {
        if (key == null)
        {
            return false;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equalsIgnoreCase(JPOX_VENDOR_NAME))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the value of a particular extension (for JPOX).
     * @param key The key of the extension
     * @return The value of the extension (null if not existing)
     */
    public String getValueForExtension(String key)
    {
        if (key == null)
        {
            return null;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equalsIgnoreCase(JPOX_VENDOR_NAME))
            {
                return ex.getValue();
            }
        }
        return null;
    }

    /**
     * Accessor for the value of extensions having the key starting with the argument (for JPOX).
     * @param key The key of the extension
     * @return The value of the extension (empty list if not existing)
     */
    public List getValuesForExtensionStartingWith(String key)
    {
        List values = new ArrayList();
        if (key == null)
        {
            return values;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().startsWith(key) &&
                ex.getVendorName().equalsIgnoreCase(JPOX_VENDOR_NAME))
            {
                values.add(ex.getValue());
            }
        }
        return values;
    }

    /**
     * Accessor for the value of a particular extension (for JPOX), but
     * splitting it into separate parts. This is for extension tags that have a
     * value as comma separated.
     * @param key The key of the extension
     * @return The value(s) of the extension (null if not existing)
     */
    public String[] getValuesForExtension(String key)
    {
        if (key == null)
        {
            return null;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equalsIgnoreCase(JPOX_VENDOR_NAME))
            {
                return MetaDataUtils.getInstance().getValuesForCommaSeparatedAttribute(ex.getValue());
            }
        }
        return null;
    }

    // -------------------------------- Mutators -------------------------------
 
    /**
     * Append new Extension (for the specified vendor).
     * Will throw an InvalidMetaDataException if the input is invalid.
     * @param vendor vendor-name tag value
     * @param key key tag value
     * @param value value tag value
     */
    public void addExtension(String vendor, String key, String value)
    {
        if (vendor == null ||
            (vendor.equalsIgnoreCase(JPOX_VENDOR_NAME) && (key == null || value == null)))
        {
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Extendable.Invalid", vendor, key, value);
        }

        if (vendor.equals(JPOX_VENDOR_NAME) && hasExtension(key))
        {
            // Remove any existing value
            removeExtension(key);
        }
        extensions.add(new ExtensionMetaData(vendor, key, value));
    }

    /**
     * Append new Extension (for JPOX).
     * Will throw an InvalidMetaDataException if the input is invalid.
     * @param key key tag value
     * @param value value tag value
     */
    public void addExtension(String key, String value)
    {
        addExtension(JPOX_VENDOR_NAME, key, value);
    }

    /**
     * Method to remove a (JPOX) MetaData extension.
     * @param key Key of the tag
     */
    public void removeExtension(String key)
    {
        Iterator iter = extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getVendorName().equalsIgnoreCase(JPOX_VENDOR_NAME) &&
                ex.getKey().equals(key))
            {
                iter.remove();
                break;
            }
        }
    }

    // -------------------------------- Utilities ------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        if (extensions.size() == 0)
        {
            return "";
        }

        StringBuffer sb = new StringBuffer();
        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            sb.append(ex.toString()).append("\n");
        }
        return sb.toString();
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        if (extensions.size() == 0)
        {
            return "";
        }

        StringBuffer sb = new StringBuffer();
        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            sb.append(prefix).append(ex.toString()).append("\n");
        }
        return sb.toString();
    }
}