/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - toString(), "column", javadocs, initialise()
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * For schema generation, it might be useful to specify that a column or columns
 * be indexed, and to provide the name of the index. For this purpose, an index
 * element can be contained within a field, element, key, value, or join
 * element, and this indicates that the column(s) associated with the referenced
 * element should be indexed. Indexes can also be specified at the class level,
 * by including index elements containing column elements. In this case, the
 * column elements are mapped elsewhere, and the column elements contain only
 * the column name.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT index (extension*, (column|field|property)*, extension*)&gt;
 * &lt;!ATTLIST index name CDATA #IMPLIED&gt;
 * &lt;!ATTLIST index table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST index unique (true|false) `false'&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.14 $
 */
public class IndexMetaData extends ExtendableMetaData implements ColumnMetaDataContainer
{
    /**
     * the index name
     */
    final String name;

    /**
     * the table name. If not specified, then applies to the table of the containing class.
     */
    final String table;

    /**
     * You can use UNIQUE constraints to ensure that no duplicate values are
     * entered in specific columns that do not participate in a primary key.
     * Although both a UNIQUE constraint and a PRIMARY KEY constraint enforce
     * uniqueness, use a UNIQUE constraint instead of a PRIMARY KEY constraint
     * when you want to enforce the uniqueness of:
     * <ul>
     * <li>
     * A column, or combination of columns, that is not the primary key.
     * Multiple UNIQUE constraints can be defined on a table, whereas only one
     * PRIMARY KEY constraint can be defined on a table.
     * </li>
     * <li>
     * A column that allows null values. UNIQUE constraints can be defined on
     * columns that allow null values, whereas PRIMARY KEY constraints can be
     * defined only on columns that do not allow null values.
     * </li>
     * </ul>
     * A UNIQUE constraint can also be referenced by a FOREIGN KEY constraint.
     */
    Boolean unique;

    /**
     * Contains the metadata for fields
     */
    protected AbstractPropertyMetaData[] fieldMetaData;

    /**
     * Contains the metadata for columns
     */
    protected ColumnMetaData[] columnMetaData;

    // -------------------------------------------------------------------------
    // Fields below here are used in the metadata parse process where the parser
    // dynamically adds fields/columns as it encounters them in the MetaData files.
    // They are typically cleared at the point of initialise() and not used thereafter.

    /**
     * the field elements to be included in the index. Can be empty.
     */
    protected List fields = new ArrayList();

    /**
     * the columns elements to be included in the index. Suitable to be empty
     * when this metadata is contained within a field, element, key, value, or
     * join elements
     */
    protected List columns = new ArrayList();

    /**
     * Constructor to create a copy of the passed metadata using the provided parent.
     * @param parent The parent
     * @param imd The metadata to copy
     */
    public IndexMetaData(MetaData parent, IndexMetaData imd)
    {
        super(parent);
        this.name = imd.name;
        this.table = imd.table;
        this.unique = imd.unique;
        AbstractPropertyMetaData[] apmd = null;
        if (imd.fields != null)
        {
            apmd = (AbstractPropertyMetaData[]) imd.fields.toArray(new AbstractPropertyMetaData[imd.fields.size()]);
        }
        else
        {
            apmd = imd.fieldMetaData;
        }
        if (apmd != null)
        {
            for (int i = 0; i < apmd.length; i++)
            {
                if (apmd[i] instanceof FieldMetaData)
                {
                    addField(new FieldMetaData(this, apmd[i]));
                }
                else
                {
                    addField(new PropertyMetaData(this, (PropertyMetaData) apmd[i]));
                }
            }
        }
        ColumnMetaData[] colmd = null;
        if (imd.columns != null)
        {
            colmd = (ColumnMetaData[]) imd.columns.toArray(new ColumnMetaData[imd.columns.size()]);
        }
        else
        {
            colmd = imd.columnMetaData;
        }
        if (colmd != null)
        {
            for (int i = 0; i < colmd.length; i++)
            {
                addColumn(new ColumnMetaData(this, colmd[i]));
            }
        }
    }

    /**
     * Constructor. 
     * @param name Name of index
     * @param table Name of the table
     * @param unique Whether it is unique.
     */
    public IndexMetaData(final String name, String table, String unique)
    {
        super(null);    // Ignore parent

        this.name = name;
        this.table = table;
        if (unique != null)
        {
            if (unique.equalsIgnoreCase("true"))
            {
                this.unique = Boolean.TRUE;
            }
            else if (unique.equalsIgnoreCase("false"))
            {
                this.unique = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialise all sub-objects.
     */
    public void initialise()
    {
        if (isInitialised())
        {
            return;
        }

        // Set up the fieldMetaData
        if (fields.size() == 0)
        {
            fieldMetaData = null;
        }
        else
        {
            fieldMetaData = new AbstractPropertyMetaData[fields.size()];
            for (int i=0; i<fieldMetaData.length; i++)
            {
                fieldMetaData[i] = (AbstractPropertyMetaData) fields.get(i);
                fieldMetaData[i].initialise();
            }
        }

        // Set up the columnMetaData
        if (columns.size() == 0)
        {
            columnMetaData = null;
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        // Clean out parsing data
        fields.clear();
        fields = null;
        columns.clear();
        columns = null;

        setInitialised();
    }

    // -------------------------------- Mutators -------------------------------

    /**
     * Add a new AbstractPropertyMetaData element
     * @param fmd The AbstractPropertyMetaData to add
     */
    public void addField(AbstractPropertyMetaData fmd)
    {
        fields.add(fmd);
    }

    /**
     * Add a new ColumnMetaData element
     * @param colmd The ColumMetaData to add
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final AbstractPropertyMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the name of the index.
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for the table
     * @return Returns the table..
     */
    public final String getTable()
    {
        return table;
    }

    /**
     * Accessor for whether the index is unique
     * @return Returns whether the index is unique.
     */
    public final Boolean isUnique()
    {
        return unique;
    }

    // -------------------------------- Utilities ------------------------------

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<index unique=\"" + unique + "\"");
        if (table != null)
        {
            sb.append(" table=\"" + table + "\"");
        }
        sb.append(name != null ? (" name=\"" + name + "\">\n") : ">\n");

        // Add fields
        if (fieldMetaData != null)
        {
            for (int i=0;i<fieldMetaData.length;i++)
            {
                sb.append(fieldMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add columns
        if (columnMetaData != null)
        {
            for (int i=0;i<columnMetaData.length;i++)
            {
                sb.append(columnMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</index>\n");
        return sb.toString();
    }
}