/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.SAXException;

import org.jpox.util.JPOXLogger;

/**
 * Parser handler for JDO MetaData.
 * Implements DefaultHandler and handles the extracting of MetaData for JDO
 * from the XML elements/attributes. This class simply constructs the MetaData
 * representation mirroring what is in the MetaData file. It has no knowledge
 * of the class(es) that it represents, simply the information in the MetaData
 * file. The knowledge of the classes is imposed on the representation at a
 * later stage where necessary.
 * <P>Operates the parse process using a Stack. MetaData components are added
 * to the stack as they are encountered and created. They are then popped off
 * the stack when the end element is encountered.</P>
 *
 * @since 1.1
 * @version $Revision: 1.6 $
 */
public class JDOMetaDataHandler extends AbstractMetaDataHandler
{
    /**
     * Constructor. Protected to prevent instantiation.
     * @param mgr the metadata manager
     * @param filename The name of the file to parse
     * @param resolver Entity Resolver to use (null if not available)
     */
    public JDOMetaDataHandler(MetaDataManager mgr, String filename, EntityResolver resolver)
    {
        super(mgr, filename, resolver);
        metadata = new FileMetaData(filename, mgr, null, null);
        pushStack(metadata);
    }

    /**
     * Utility to create a new class component.
     * @param pmd The parent PackageMetaData
     * @param attrs The attributes
     * @return The ClassMetaData
     */
    protected ClassMetaData newClassObject(PackageMetaData pmd, Attributes attrs)
    {
        return mgr.newClassObject(pmd,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"identity-type"),
                            getAttr(attrs,"objectid-class"),
                            getAttr(attrs,"requires-extent"),
                            getAttr(attrs,"detachable"),
                            getAttr(attrs,"embedded-only"),
                            getAttr(attrs,"persistence-modifier"),
                            getAttr(attrs,"persistence-capable-superclass"),
                            getAttr(attrs,"catalog"),
                            getAttr(attrs,"schema"),
                            getAttr(attrs,"table"),
                            null);
    }

    /**
     * Utility to create a new field component.
     * @param md The parent MetaData
     * @param attrs The attributes
     * @return The FieldMetaData
     */
    protected FieldMetaData newFieldObject(MetaData md, Attributes attrs)
    {
        return mgr.newFieldObject(md,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"primary-key"),
                            getAttr(attrs,"persistence-modifier"),
                            getAttr(attrs,"default-fetch-group"),
                            getAttr(attrs,"null-value"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"dependent"),
                            getAttr(attrs,"mapped-by"),
                            getAttr(attrs,"column"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"),
                            getAttr(attrs,"recursion-depth"),
                            getAttr(attrs,"load-fetch-group"),
                            getAttr(attrs,"value-strategy"),
                            getAttr(attrs,"sequence"),
                            getAttr(attrs,"field-type"));
    }

    /**
     * Utility to create a new property component.
     * @param md The parent MetaData
     * @param attrs The attributes
     * @return The PropertyMetaData
     */
    protected PropertyMetaData newPropertyObject(MetaData md, Attributes attrs)
    {
        return mgr.newPropertyObject(md,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"primary-key"),
                            getAttr(attrs,"persistence-modifier"),
                            getAttr(attrs,"default-fetch-group"),
                            getAttr(attrs,"null-value"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"dependent"),
                            getAttr(attrs,"mapped-by"),
                            getAttr(attrs,"column"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"),
                            getAttr(attrs,"recursion-depth"),
                            getAttr(attrs,"load-fetch-group"),
                            getAttr(attrs,"value-strategy"),
                            getAttr(attrs,"sequence"),
                            getAttr(attrs,"field-type"),
                            getAttr(attrs,"field-name"));
    }    
    
    /**
     * Handler method called at the start of an element.
     * @param uri URI of the tag
     * @param localName Local name
     * @param qName Element name
     * @param attrs Attributes for this element 
     * @throws SAXException in parsing errors
     */
    public void startElement(String uri, 
                             String localName,
                             String qName,
                             Attributes attrs)
    throws SAXException 
    {
        if (charactersBuffer.length() > 0)
        {
            // Cater for subelements that appear before the end of the body text (save the body text with the parent)
            String currentString = getString().trim();
            if (getStack() instanceof QueryMetaData)
            {
                ((QueryMetaData)getStack()).setQuery(currentString.trim());
            }
        }

        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            StringBuffer sb = new StringBuffer();
            sb.append("<" + qName);
            for (int i=0; i<attrs.getLength(); i++)
            {
                sb.append(" ");
                sb.append(attrs.getQName(i)).append("=\"").append(attrs.getValue(i)).append("\"");
            }
            sb.append(">");
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.StartElement",sb.toString(), "" + stack.size()));
        }
        try
        {
            // New JDO file
            if (qName.equals("jdo"))
            {
                FileMetaData filemd = (FileMetaData)getStack();
                filemd.setType(FileMetaData.JDO_FILE);
                filemd.setCatalog(getAttr(attrs, "catalog"));
                filemd.setSchema(getAttr(attrs, "schema"));
            }
            // New JDO O/R mapping file
            else if (qName.equals("orm"))
            {
                FileMetaData filemd = (FileMetaData)getStack();
                filemd.setType(FileMetaData.ORM_FILE);
                filemd.setCatalog(getAttr(attrs, "catalog"));
                filemd.setSchema(getAttr(attrs, "schema"));
            }
            // New JDO Query file
            else if (qName.equals("jdoquery"))
            {
                FileMetaData filemd = (FileMetaData)getStack();
                filemd.setType(FileMetaData.JDOQUERY_FILE);
            }
            // New package for this file
            else if (qName.equals("package"))
            {
                FileMetaData filemd = (FileMetaData)getStack();
                PackageMetaData pmd = new PackageMetaData(filemd, 
                                                          getAttr(attrs, "name"),
                                                          getAttr(attrs, "catalog"),
                                                          getAttr(attrs, "schema"));
                ((FileMetaData)metadata).addPackage(pmd);
                pushStack(pmd);
            }
            // New class for this package
            else if (qName.equals("class"))
            {
                PackageMetaData pmd=(PackageMetaData)getStack();
                ClassMetaData cmd=newClassObject(pmd,attrs);
                pmd.addClass(cmd);
    
                pushStack(cmd);
            }
            // New interface for this package
            else if (qName.equals("interface"))
            {
                PackageMetaData pmd = (PackageMetaData)getStack();
                InterfaceMetaData imd = new InterfaceMetaData(pmd,
                                                  getAttr(attrs, "name"),
                                                  getAttr(attrs, "identity-type"),
                                                  getAttr(attrs, "objectid-class"),
                                                  getAttr(attrs, "requires-extent"),
                                                  getAttr(attrs, "detachable"),
                                                  getAttr(attrs, "embedded-only"),
                                                  getAttr(attrs, "catalog"),
                                                  getAttr(attrs, "schema"),
                                                  getAttr(attrs, "table"),
                                                  null);
                pmd.addInterface(imd);
                pushStack(imd);
            }
            // New primary-key for class/interface
            else if (qName.equals("primary-key"))
            {
                ExtendableMetaData md = (ExtendableMetaData)getStack();
                PrimaryKeyMetaData pkmd = 
                    new PrimaryKeyMetaData(md, getAttr(attrs,"name"), getAttr(attrs, "column"));
                if (md instanceof AbstractClassMetaData)
                {
                    ((AbstractClassMetaData)md).setPrimaryKeyMetaData(pkmd);
                }
                else if (md instanceof JoinMetaData)
                {
                    ((JoinMetaData)md).setPrimaryKeyMetaData(pkmd);
                }
                pushStack(pkmd);
            }
            // New implements for this class
            else if (qName.equals("implements"))
            {
                ClassMetaData cmd = (ClassMetaData)getStack();
                ImplementsMetaData imd = 
                    new ImplementsMetaData(cmd, getAttr(attrs,"name"));
                cmd.addImplements(imd);
                pushStack(imd);
            }
            // New property for this interface
            else if (qName.equals("property"))
            {
                MetaData parent = getStack();
                if (parent instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData acmd = (AbstractClassMetaData)parent;
                    PropertyMetaData propmd = newPropertyObject(acmd, attrs);
                    acmd.addProperty(propmd);
                    pushStack(propmd);
                }
                else if (parent instanceof EmbeddedMetaData)
                {
                    EmbeddedMetaData emd = (EmbeddedMetaData)parent;
                    PropertyMetaData propmd = newPropertyObject(emd, attrs);
                    emd.addField(propmd);
                    pushStack(propmd);
                }
                else if (parent instanceof ImplementsMetaData)
                {
                    ImplementsMetaData implmd = (ImplementsMetaData)parent;
                    PropertyMetaData propmd = newPropertyObject(implmd, attrs);
                    implmd.addProperty(propmd);
                    pushStack(propmd);
                }
            }
            // New datastore-identity for this class
            else if (qName.equals("datastore-identity"))
            {
                AbstractClassMetaData acmd = (AbstractClassMetaData)getStack();
                IdentityMetaData imd = new IdentityMetaData(acmd,
                                getAttr(attrs,"column"),
                                getAttr(attrs,"strategy"),
                                getAttr(attrs,"sequence"));
                acmd.setIdentityMetaData(imd);
                pushStack(imd);
            }
            // New inheritance for this class
            else if (qName.equals("inheritance"))
            {
                MetaData parent = getStack();
                AbstractClassMetaData acmd = (AbstractClassMetaData)parent;
                InheritanceMetaData imd = new InheritanceMetaData(acmd,
                                getAttr(attrs,"strategy"));
                acmd.setInheritanceMetaData(imd);
                pushStack(imd);
            }
            // Discriminator for this inheritance
            else if (qName.equals("discriminator"))
            {
                InheritanceMetaData imd = (InheritanceMetaData)getStack();
                DiscriminatorMetaData dismd = new DiscriminatorMetaData(imd,
                                getAttr(attrs,"column"),
                                getAttr(attrs,"value"),
                                getAttr(attrs,"strategy"),
                                getAttr(attrs,"indexed"));
                imd.setDiscriminatorMetaData(dismd);
                pushStack(dismd);
            }
            // New query for this class
            else if (qName.equals("query"))
            {
                ExtendableMetaData emd = (ExtendableMetaData)getStack();
                QueryMetaData qmd = new QueryMetaData(emd,
                                getAttr(attrs,"name"),
                                getAttr(attrs,"language"),
                                getAttr(attrs,"unmodifiable"),
                                getAttr(attrs,"result-class"),
                                getAttr(attrs,"unique"));
                if (emd instanceof ClassMetaData)
                {
                    ClassMetaData cmd = (ClassMetaData)emd;
                    cmd.addQuery(qmd);
                }
                else if (emd instanceof FileMetaData)
                {
                    FileMetaData fmd = (FileMetaData)emd;
                    fmd.addQuery(qmd);
                }
                pushStack(qmd);
            }
            // New sequence for this class
            else if (qName.equals("sequence"))
            {
                ExtendableMetaData emd = (ExtendableMetaData)getStack();
                SequenceMetaData seqmd = new SequenceMetaData(emd,
                                getAttr(attrs,"name"),
                                getAttr(attrs,"datastore-sequence"),
                                getAttr(attrs,"factory-class"),
                                getAttr(attrs,"strategy"),
                                null, null);
                if (emd instanceof PackageMetaData)
                {
                    PackageMetaData pmd = (PackageMetaData)emd;
                    pmd.addSequence(seqmd);
                }
                pushStack(seqmd);
            }
            // New field for this class
            else if (qName.equals("field"))
            {
                MetaData md = getStack();
                FieldMetaData fmd = newFieldObject(md,attrs);
                if (md instanceof ClassMetaData)
                {
                    ClassMetaData cmd = (ClassMetaData)md;
                    cmd.addProperty(fmd);
                }
                else if (md instanceof FetchGroupMetaData)
                {
                    FetchGroupMetaData fgmd = (FetchGroupMetaData)md;
                    fgmd.addField(fmd);
                }
                else if (md instanceof EmbeddedMetaData)
                {
                    EmbeddedMetaData emd = (EmbeddedMetaData)md;
                    emd.addField(fmd);
                }
                else if (md instanceof ForeignKeyMetaData)
                {
                    ForeignKeyMetaData fkmd = (ForeignKeyMetaData)md;
                    fkmd.addField(fmd);
                }
                else if (md instanceof IndexMetaData)
                {
                    IndexMetaData imd = (IndexMetaData)md;
                    imd.addField(fmd);
                }
                else if (md instanceof UniqueMetaData)
                {
                    UniqueMetaData umd = (UniqueMetaData)md;
                    umd.addField(fmd);
                }
                pushStack(fmd);
            }
            // New join
            else if (qName.equals("join"))
            {
                MetaData parent = getStack();
                JoinMetaData jnmd = new JoinMetaData(parent,
                                                   getAttr(attrs,"table"),
                                                   getAttr(attrs,"column"),
                                                   getAttr(attrs,"outer"),
                                                   getAttr(attrs,"delete-action"));
                if (parent instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)parent;
                    fmd.setJoinMetaData(jnmd);
                }
                else if (parent instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)parent;
                    cmd.addJoin(jnmd);
                }
                else if (parent instanceof InheritanceMetaData)
                {
                    InheritanceMetaData imd = (InheritanceMetaData)parent;
                    imd.setJoinMetaData(jnmd);
                }
                pushStack(jnmd);
            }
            // New map container for this field
            else if (qName.equals("map"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                MapMetaData map = new MapMetaData(fmd,
                                getAttr(attrs,"key-type"),
                                getAttr(attrs,"embedded-key"),
                                getAttr(attrs,"dependent-key"),
                                getAttr(attrs,"serialized-key"),
                                getAttr(attrs,"value-type"),
                                getAttr(attrs,"embedded-value"),
                                getAttr(attrs,"dependent-value"),
                                getAttr(attrs,"serialized-value"));
                fmd.setContainer(map);
                pushStack(map);
            }
            // New array container for this field
            else if (qName.equals("array"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                ArrayMetaData amd = new ArrayMetaData(fmd,
                                getAttr(attrs, "element-type"),
                                getAttr(attrs, "embedded-element"),
                                getAttr(attrs, "dependent-element"),
                                getAttr(attrs, "serialized-element"));
                fmd.setContainer(amd);
                pushStack(amd);
            }
            // New collection container for this field
            else if (qName.equals("collection"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                CollectionMetaData colmd = new CollectionMetaData(fmd,
                                getAttr(attrs,"element-type"),
                                getAttr(attrs,"embedded-element"),
                                getAttr(attrs,"dependent-element"),
                                getAttr(attrs,"serialized-element"));
                fmd.setContainer(colmd);
                pushStack(colmd);
            }
            // New column
            else if (qName.equals("column"))
            {
                ExtendableMetaData extmd = (ExtendableMetaData)getStack();
                ColumnMetaData clnmd = new ColumnMetaData(extmd,
                                getAttr(attrs,"name"),
                                getAttr(attrs,"target"),
                                getAttr(attrs,"target-field"),
                                getAttr(attrs,"jdbc-type"),
                                getAttr(attrs,"sql-type"),
                                getAttr(attrs,"length"),
                                getAttr(attrs,"scale"),
                                getAttr(attrs,"allows-null"),
                                getAttr(attrs,"default-value"),
                                getAttr(attrs,"insert-value"),
                                null, null, null);
                if (extmd instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)extmd;
                    fmd.addColumn(clnmd);
                }
                else if (extmd instanceof AbstractElementMetaData)
                {
                    AbstractElementMetaData elemd = (AbstractElementMetaData)extmd;
                    elemd.addColumn(clnmd);
                }
                else if (extmd instanceof JoinMetaData)
                {
                    JoinMetaData jnmd = (JoinMetaData)extmd;
                    jnmd.addColumn(clnmd);
                }
                else if (extmd instanceof IdentityMetaData)
                {
                    IdentityMetaData idmd = (IdentityMetaData)extmd;
                    idmd.addColumn(clnmd);
                }
                else if (extmd instanceof ForeignKeyMetaData)
                {
                    ForeignKeyMetaData fkmd = (ForeignKeyMetaData)extmd;
                    fkmd.addColumn(clnmd);
                }
                else if (extmd instanceof IndexMetaData)
                {
                    IndexMetaData idxmd = (IndexMetaData)extmd;
                    idxmd.addColumn(clnmd);
                }
                else if (extmd instanceof UniqueMetaData)
                {
                    UniqueMetaData unimd = (UniqueMetaData)extmd;
                    unimd.addColumn(clnmd);
                }
                else if (extmd instanceof OrderMetaData)
                {
                    OrderMetaData ormd = (OrderMetaData)extmd;
                    ormd.addColumn(clnmd);
                }
                else if (extmd instanceof DiscriminatorMetaData)
                {
                    DiscriminatorMetaData dismd = (DiscriminatorMetaData)extmd;
                    dismd.setColumnMetaData(clnmd);
                }
                else if (extmd instanceof VersionMetaData)
                {
                    VersionMetaData vermd = (VersionMetaData)extmd;
                    vermd.addColumn(clnmd);
                }
                else if (extmd instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)extmd;
                    cmd.addUnmappedColumn(clnmd);
                }
                else if (extmd instanceof PrimaryKeyMetaData)
                {
                    PrimaryKeyMetaData pkmd = (PrimaryKeyMetaData)extmd;
                    pkmd.addColumn(clnmd);
                }
                pushStack(clnmd);
            }
            // New element
            else if (qName.equals("element"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                ElementMetaData elemd = new ElementMetaData(fmd,
                                getAttr(attrs, "column"),
                                getAttr(attrs, "delete-action"),
                                getAttr(attrs, "update-action"),
                                getAttr(attrs, "indexed"),
                                getAttr(attrs, "unique"),
                                getAttr(attrs, "mapped-by"));
                fmd.setElementMetaData(elemd);
                pushStack(elemd);
            }
            // New key
            else if (qName.equals("key"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                KeyMetaData keymd = new KeyMetaData(fmd,
                                getAttr(attrs, "column"),
                                getAttr(attrs, "delete-action"),
                                getAttr(attrs, "update-action"),
                                getAttr(attrs, "indexed"),
                                getAttr(attrs, "unique"),
                                getAttr(attrs, "mapped-by"));
                fmd.setKeyMetaData(keymd);
                pushStack(keymd);
            }
            // New value
            else if (qName.equals("value"))
            {
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                ValueMetaData valuemd = new ValueMetaData(fmd,
                                getAttr(attrs, "column"),
                                getAttr(attrs, "delete-action"),
                                getAttr(attrs, "update-action"),
                                getAttr(attrs, "indexed"),
                                getAttr(attrs, "unique"),
                                getAttr(attrs, "mapped-by"));
                fmd.setValueMetaData(valuemd);
                pushStack(valuemd);
            }
            // New fetch-group
            else if (qName.equals("fetch-group"))
            {
                MetaData md = getStack();
                FetchGroupMetaData fgmd = new FetchGroupMetaData(md, getAttr(attrs,"post-load"), getAttr(attrs,"name"));
                if (md instanceof FetchGroupMetaData)
                {
                    FetchGroupMetaData fgmdParent = (FetchGroupMetaData)md;
                    fgmdParent.addFetchGroup(fgmd);
                }
                else if (md instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)md;
                    cmd.addFetchGroup(fgmd);
                }
                pushStack(fgmd);
            }
            // New extension for this tag
            else if (qName.equals("extension"))
            {
                ExtendableMetaData emd = (ExtendableMetaData)getStack();
                emd.addExtension(getAttr(attrs,"vendor-name"),
                                getAttr(attrs,"key"),
                                getAttr(attrs,"value"));
            }
            // New version
            else if (qName.equals("version"))
            {
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                VersionMetaData vermd = new VersionMetaData(
                                getAttr(attrs,"strategy"),
                                getAttr(attrs,"column"),
                                getAttr(attrs,"indexed"));
                cmd.setVersionMetaData(vermd);
                pushStack(vermd);
            }
            // New index
            else if (qName.equals("index"))
            {
                ExtendableMetaData extmd = (ExtendableMetaData)getStack();
                IndexMetaData idxmd = new IndexMetaData(
                                getAttr(attrs,"name"),
                                getAttr(attrs,"table"),
                                getAttr(attrs,"unique"));
                if (extmd instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)extmd;
                    cmd.addIndex(idxmd);
                }
                else if (extmd instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)extmd;
                    fmd.setIndexMetaData(idxmd);
                }
                else if (extmd instanceof JoinMetaData)
                {
                    JoinMetaData jmd = (JoinMetaData)extmd;
                    jmd.setIndexMetaData(idxmd);
                }
                else if (extmd instanceof AbstractElementMetaData)
                {
                    AbstractElementMetaData elmd = (AbstractElementMetaData)extmd;
                    elmd.setIndexMetaData(idxmd);
                }
                else if (extmd instanceof OrderMetaData)
                {
                    OrderMetaData omd = (OrderMetaData)extmd;
                    omd.setIndexMetaData(idxmd);
                }
                else if (extmd instanceof VersionMetaData)
                {
                    VersionMetaData vermd = (VersionMetaData)extmd;
                    vermd.setIndexMetaData(idxmd);
                }
                else if (extmd instanceof DiscriminatorMetaData)
                {
                    DiscriminatorMetaData dismd = (DiscriminatorMetaData)extmd;
                    dismd.setIndexMetaData(idxmd);
                }
                pushStack(idxmd);
            }
            // New unique constraint
            else if (qName.equals("unique"))
            {
                ExtendableMetaData extmd = (ExtendableMetaData)getStack();
                UniqueMetaData unimd = new UniqueMetaData(extmd,
                                                          getAttr(attrs,"name"),
                                                          getAttr(attrs,"table"),
                                                          getAttr(attrs,"deferred"));
                if (extmd instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)extmd;
                    cmd.addUniqueConstraint(unimd);
                }
                else if (extmd instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)extmd;
                    fmd.setUniqueMetaData(unimd);
                }
                else if (extmd instanceof JoinMetaData)
                {
                    JoinMetaData jmd = (JoinMetaData)extmd;
                    jmd.setUniqueMetaData(unimd);
                }
                else if (extmd instanceof AbstractElementMetaData)
                {
                    AbstractElementMetaData elmd = (AbstractElementMetaData)extmd;
                    elmd.setUniqueMetaData(unimd);
                }
                pushStack(unimd);
            }
            // New foreign key
            else if (qName.equals("foreign-key"))
            {
                ExtendableMetaData extmd=(ExtendableMetaData)getStack();
                ForeignKeyMetaData fkmd=new ForeignKeyMetaData(
                                getAttr(attrs,"deferred"),
                                getAttr(attrs,"delete-action"),
                                getAttr(attrs,"update-action"),
                                getAttr(attrs,"table"),
                                getAttr(attrs,"unique"),
                                getAttr(attrs,"name"));
                if (extmd instanceof AbstractClassMetaData)
                {
                    AbstractClassMetaData cmd = (AbstractClassMetaData)extmd;
                    cmd.addForeignKey(fkmd);
                }
                else if (extmd instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)extmd;
                    fmd.setForeignKeyMetaData(fkmd);
                }
                else if (extmd instanceof JoinMetaData)
                {
                    JoinMetaData jmd = (JoinMetaData)extmd;
                    jmd.setForeignKeyMetaData(fkmd);
                }
                else if (extmd instanceof AbstractElementMetaData)
                {
                    AbstractElementMetaData elmd = (AbstractElementMetaData)extmd;
                    elmd.setForeignKeyMetaData(fkmd);
                }
                pushStack(fkmd);
            }
            // New order
            else if (qName.equals("order"))
            {
                OrderMetaData ormd = new OrderMetaData(getAttr(attrs, "column"),
                                                       getAttr(attrs, "indexed"),
                                                       getAttr(attrs, "mapped-by"));
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                fmd.setOrderMetaData(ormd);
                pushStack(ormd);
            }
            // Embedded
            else if (qName.equals("embedded"))
            {
                ExtendableMetaData extmd=(ExtendableMetaData)getStack();
                EmbeddedMetaData emd = new EmbeddedMetaData(extmd,
                                                            getAttr(attrs,"owner-field"),
                                                            getAttr(attrs,"null-indicator-column"),
                                                            getAttr(attrs,"null-indicator-value"));
                if (extmd instanceof AbstractPropertyMetaData)
                {
                    AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)extmd;
                    fmd.setEmbeddedMetaData(emd);
                }
                else if (extmd instanceof KeyMetaData)
                {
                    KeyMetaData kmd = (KeyMetaData)extmd;
                    kmd.setEmbeddedMetaData(emd);
                }
                else if (extmd instanceof ValueMetaData)
                {
                    ValueMetaData vmd = (ValueMetaData)extmd;
                    vmd.setEmbeddedMetaData(emd);
                }
                else if (extmd instanceof ElementMetaData)
                {
                    ElementMetaData elmd = (ElementMetaData)extmd;
                    elmd.setEmbeddedMetaData(emd);
                }
                pushStack(emd);
            }
            else
            {
                String message = LOCALISER.msg("MetaData.Parser.UnknownTag.Error",qName);
                JPOXLogger.METADATA.error(message);
                throw new RuntimeException(message);
            }
        }
        catch(RuntimeException ex)
        {
            //TODO localise
            JPOXLogger.METADATA.error("The following error occurred while parsing <" +qName+"> nested within "+getStack());
            throw ex;
        }
    }

    /**
     * Handler method called at the end of an element.
     * @param uri URI of the tag
     * @param localName local name
     * @param qName Name of element just ending
     * @throws SAXException in parsing errors
     */
    public void endElement(String uri, String localName, String qName)
    throws SAXException
    {
        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.EndElement","<" + qName + ">", "" + stack.size()));
        }

        // Save the current string for elements that have a body value
        String currentString = getString().trim();
        if (currentString.length() > 0)
        {
            ExtendableMetaData extmd = (ExtendableMetaData)getStack();
            if (qName.equals("query"))
            {
                ((QueryMetaData)extmd).setQuery(currentString);
            }
        }

        // Pop the tag
        // If startElement pushes an element onto the stack need a remove here for that type
        if (qName.equals("package") ||
            qName.equals("class") ||
            qName.equals("interface") ||
            qName.equals("implements") ||
            qName.equals("property") ||
            qName.equals("datastore-identity") ||
            qName.equals("inheritance") ||
            qName.equals("primary-key") ||
            qName.equals("version") ||
            qName.equals("unmapped") ||
            qName.equals("query") ||
            qName.equals("sequence") ||
            qName.equals("field") ||
            qName.equals("map") ||
            qName.equals("element") ||
            qName.equals("embedded") ||
            qName.equals("key") ||
            qName.equals("value") ||
            qName.equals("array") ||
            qName.equals("collection") ||
            qName.equals("join") ||
            qName.equals("index") ||
            qName.equals("unique") ||
            qName.equals("foreign-key") ||
            qName.equals("order") ||
            qName.equals("fetch-group") ||
            qName.equals("column") ||
            qName.equals("discriminator"))
        {
            popStack();
        }
    }
}