/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import org.xml.sax.Attributes;
import org.xml.sax.EntityResolver;
import org.xml.sax.SAXException;

import org.jpox.util.JPOXLogger;

/**
 * Parser handler for JPA MetaData.
 * Implements DefaultHandler and handles the extracting of MetaData for JPA
 * from the XML elements/attributes. This class simply constructs the MetaData
 * representation mirroring what is in the MetaData file. It has no knowledge
 * of the class(es) that it represents, simply the information in the MetaData
 * file. The knowledge of the classes is imposed on the representation at a
 * later stage where necessary.
 * <P>Operates the parse process using a Stack. MetaData components are added
 * to the stack as they are encountered and created. They are then popped off
 * the stack when the end element is encountered.</P>
 *
 * @since 1.1
 * @version $Revision: 1.17 $
 */
public class JPAMetaDataHandler extends AbstractMetaDataHandler
{
    /** Default package name for the entity-mappings (if any). */
    String defaultPackageName = null;

    /**
     * Constructor. Protected to prevent instantiation.
     * @param mgr the metadata manager
     * @param filename The name of the file to parse
     * @param resolver Entity Resolver to use (null if not available)
     */
    public JPAMetaDataHandler(MetaDataManager mgr, String filename, EntityResolver resolver)
    {
        super(mgr, filename, resolver);
        metadata = new FileMetaData(filename, mgr, null, null);
        pushStack(metadata); // Start with FileMetaData on the stack
    }

    /**
     * Utility to create a new class component.
     * @param pmd The parent PackageMetaData
     * @param attrs The attributes
     * @return The ClassMetaData
     */
    protected ClassMetaData newClassObject(PackageMetaData pmd, Attributes attrs)
    {
        String className = getAttr(attrs, "class");
        if (className.indexOf('.') > 0)
        {
            // Strip off any package info
            className = className.substring(className.lastIndexOf('.')+1);
        }

        return mgr.newClassObject(pmd,
            className,
            IdentityType.APPLICATION.toString(), // JPA only has "application"
            getAttr(attrs, "objectid-class"), // TODO This is specified elsewhere
            "true", // All classes require an extent
            "true", // All classes are detachable in JPA
            getAttr(attrs, "embedded-only"),
            ClassPersistenceModifier.PERSISTENCE_CAPABLE.toString(), // JPA only has "persistence-capable"
            null, // pc-superclass is not needed
            null, // Catalog specified on <table>
            null, // Schema specified on <table>
            null, // Table specified on <table>
            getAttr(attrs, "name")); // Entity name for use by JPQL
    }

    /**
     * Utility to create a new field component.
     * @param md The parent MetaData
     * @param attrs The attributes
     * @return The FieldMetaData
     */
    protected FieldMetaData newFieldObject(MetaData md, Attributes attrs)
    {
        String fetch = getAttr(attrs, "fetch");
        String dfg = "true";
        if (fetch != null && fetch.equalsIgnoreCase("LAZY"))
        {
            dfg = "false";
        }
        return mgr.newFieldObject(md,
                            getAttr(attrs,"name"),
                            null, // primary-key is specified using "id"
                            FieldPersistenceModifier.PERSISTENT.toString(), // All transient go through a different method
                            dfg,
                            null, // No such concept in JPA
                            null, // Specified via "embedded"
                            null, // Specified via "lob"
                            getAttr(attrs,"dependent"),
                            getAttr(attrs,"mapped-by"),
                            null, // Specified via "column"
                            null, // Specified via "secondary-table" or "join-table"
                            null, // JPA has no delete-action
                            null, // JPA has no index option
                            null, // unique is specified elsewhere
                            null, // JPA has no recursion depth
                            getAttr(attrs,"load-fetch-group"),
                            null, // value strategy is specified elsewhere
                            null, // sequence is specified elsewhere
                            null); // JPA has no field-type
    }

    /**
     * Utility to create a new primary key field component.
     * @param md The parent MetaData
     * @param attrs Attributes of the "id" element
     * @return The FieldMetaData
     */
    protected FieldMetaData newPKFieldObject(MetaData md, Attributes attrs)
    {
        return mgr.newFieldObject(md, getAttr(attrs, "name"), "true", FieldPersistenceModifier.PERSISTENT.toString(), 
            null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Utility to create a new transient field component.
     * @param md The parent MetaData
     * @param name Name of the transient field
     * @return The FieldMetaData
     */
    protected FieldMetaData newTransientFieldObject(MetaData md, String name)
    {
        return mgr.newFieldObject(md, name, null, FieldPersistenceModifier.NONE.toString(), 
            null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Utility to create a new property component.
     * @param md The parent MetaData
     * @param attrs The attributes
     * @return The PropertyMetaData
     */
    protected PropertyMetaData newPropertyObject(MetaData md, Attributes attrs)
    {
        // JPA doesnt support "persistent-interfaces"
        return null;
    }    
    
    /**
     * Handler method called at the start of an element.
     * @param uri URI of the tag
     * @param localName Local name
     * @param qName Element name
     * @param attrs Attributes for this element 
     * @throws SAXException in parsing errors
     */
    public void startElement(String uri, String localName, String qName, Attributes attrs)
    throws SAXException
    {
        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            StringBuffer sb = new StringBuffer();
            sb.append("<" + qName);
            for (int i=0; i<attrs.getLength(); i++)
            {
                sb.append(" ");
                sb.append(attrs.getQName(i)).append("=\"").append(attrs.getValue(i)).append("\"");
            }
            sb.append(">");
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.StartElement",sb.toString(), "" + stack.size()));
        }
        try
        {
            if (qName.equals("entity-mappings"))
            {
                FileMetaData filemd = (FileMetaData)getStack();
                filemd.setType(FileMetaData.JPA_FILE);
            }
            else if (qName.equals("description"))
            {
                // Of no practical use so ignored
            }
            else if (qName.equals("persistence-unit-metadata"))
            {
                // TODO Not yet supported whatever it does
            }
            else if (qName.equals("persistence-unit-defaults"))
            {
                // TODO Not yet supported whatever it does
            }
            else if (qName.equals("package"))
            {
                // Processed elsewhere
            }
            else if (qName.equals("schema"))
            {
                // Processed elsewhere
            }
            else if (qName.equals("catalog"))
            {
                // Processed elsewhere
            }
            else if (qName.equals("access"))
            {
                // TODO Not yet supported
            }
            else if (qName.equals("sequence-generator"))
            {
                // Sequence generator (add to the default package whether they are specified there are not)
                MetaData md = getStack();
                SequenceMetaData seqmd = new SequenceMetaData((ExtendableMetaData)md, getAttr(attrs, "name"),
                    getAttr(attrs, "sequence-name"), null, null,getAttr(attrs, "initial-value"), getAttr(attrs, "allocation-size"));
                PackageMetaData pmd = null;
                if (defaultPackageName != null)
                {
                    pmd = ((FileMetaData)metadata).getPackage(defaultPackageName);
                }
                else
                {
                    if (((FileMetaData)metadata).getNoOfPackages() > 0)
                    {
                        pmd = ((FileMetaData)metadata).getPackage(0);
                    }
                    else
                    {
                        // Add a dummy (root) package to hold our sequences since no default package name set
                        pmd = new PackageMetaData((FileMetaData)metadata, "", null, null);
                        ((FileMetaData)metadata).addPackage(pmd);
                    }
                }
                pmd.addSequence(seqmd);
            }
            else if (qName.equals("table-generator"))
            {
                // Table generator (add to the default package whether they are specified there or not)
                MetaData md = getStack();
                TableGeneratorMetaData tgmd = new TableGeneratorMetaData((ExtendableMetaData)md, getAttr(attrs, "name"), 
                    getAttr(attrs, "table"), getAttr(attrs, "catalog"), getAttr(attrs, "schema"), getAttr(attrs, "pk-column-name"),
                    getAttr(attrs, "value-column-name"), getAttr(attrs, "pk-column-value"), getAttr(attrs, "initial-value"),
                    getAttr(attrs, "allocation-size"));

                PackageMetaData pmd = null;
                if (defaultPackageName != null)
                {
                    pmd = ((FileMetaData)metadata).getPackage(defaultPackageName);
                }
                else
                {
                    if (((FileMetaData)metadata).getNoOfPackages() > 0)
                    {
                        pmd = ((FileMetaData)metadata).getPackage(0);
                    }
                    else
                    {
                        // Add a dummy (root) package to hold our sequences since no default package name set
                        pmd = new PackageMetaData((FileMetaData)metadata, "", null, null);
                        ((FileMetaData)metadata).addPackage(pmd);
                    }
                }
                pmd.addTableGenerator(tgmd);
            }
            else if (qName.equals("named-query"))
            {
                // Named JPQL query
                ClassMetaData cmd = (ClassMetaData)getStack();
                QueryMetaData qmd = new QueryMetaData(cmd, getAttr(attrs, "name"), "javax.jdo.query.JPQL", null, null, null);
                cmd.addQuery(qmd);
                pushStack(qmd);
            }
            else if (qName.equals("named-native-query"))
            {
                // Named SQL query
                ClassMetaData cmd = (ClassMetaData)getStack();
                QueryMetaData qmd = new QueryMetaData(cmd, getAttr(attrs, "name"), "javax.jdo.query.SQL", null, 
                    getAttr(attrs, "result-class"), null);
                // TODO Use the "result-set-mapping" attribute (for what ?)
                cmd.addQuery(qmd);
                pushStack(qmd);
            }
            else if (qName.equals("query"))
            {
                // handled elsewhere
            }
            else if (qName.equals("entity"))
            {
                // New entity for this package
                FileMetaData filemd = (FileMetaData)getStack();
                String className = getAttr(attrs, "class");
                String packageName = null;
                if (className.indexOf('.') > 0)
                {
                    // Fully-qualified so use package name from class
                    packageName = className.substring(0, className.lastIndexOf('.'));
                }
                PackageMetaData pmd = null;
                if (packageName != null)
                {
                    pmd = filemd.getPackage(packageName);
                }
                if (pmd == null)
                {
                    if (packageName != null)
                    {
                        // Class fully qualified so add its package
                        pmd = new PackageMetaData(filemd, packageName, null, null);
                    }
                    else if (defaultPackageName != null)
                    {
                        // Use default package for entity-mappings
                        pmd = filemd.getPackage(defaultPackageName);
                    }
                    else
                    {
                        // Add root package
                        pmd = new PackageMetaData(filemd, "", null, null);
                    }
                }

                ClassMetaData cmd = newClassObject(pmd, attrs);
                pmd.addClass(cmd);

                pushStack(cmd);
            }
            else if (qName.equals("attributes"))
            {
                // Processed elsewhere
            }
            else if (qName.equals("id-class"))
            {
                // Identity class
                ClassMetaData cmd = (ClassMetaData)getStack();
                cmd.objectidClass = getAttr(attrs, "class");
            }
            else if (qName.equals("inheritance"))
            {
                // Inheritance
                ClassMetaData cmd = (ClassMetaData)getStack();
                String strategy = getAttr(attrs, "strategy");
                String strategyType = null;
                if (strategy.equalsIgnoreCase("JOINED"))
                {
                    strategyType = InheritanceStrategy.SUBCLASS_TABLE.toString();
                }
                else if (strategy.equalsIgnoreCase("TABLE_PER_CLASS"))
                {
                    strategyType = InheritanceStrategy.NEW_TABLE.toString();
                }
                else
                {
                    strategyType = InheritanceStrategy.NEW_TABLE.toString();
                    // TODO This should also allow superclass-table option if we have a persistable superclass
                    // but we dont have the class at this point. Maybe leave it unset somehow and set at populate()
                }
                InheritanceMetaData inhmd = new InheritanceMetaData(cmd, getAttr(attrs, strategyType));
                cmd.setInheritanceMetaData(inhmd);
            }
            else if (qName.equals("table"))
            {
                // Table for this entity
                ClassMetaData cmd = (ClassMetaData)getStack();
                cmd.catalog = getAttr(attrs, "catalog");
                cmd.schema = getAttr(attrs, "schema");
                cmd.table = getAttr(attrs, "name");
            }
            else if (qName.equals("id"))
            {
                // Identity field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newPKFieldObject(cmd, attrs);
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("embedded-id"))
            {
                // Embedded identity field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newPKFieldObject(cmd, attrs);
                cmd.addProperty(fmd);
                // TODO Is this correct ? What is "embedded-id" ?

                pushStack(fmd);
            }
            else if (qName.equals("basic"))
            {
                // Basic field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newFieldObject(cmd, attrs);
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("transient"))
            {
                // Transient field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newTransientFieldObject(cmd, getAttr(attrs, "name"));
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("one-to-many"))
            {
                // 1-N field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newFieldObject(cmd, attrs);
                cmd.addProperty(fmd);
                // TODO Add collection, map where appropriate

                pushStack(fmd);
            }
            else if (qName.equals("one-to-one"))
            {
                // 1-1 field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newFieldObject(cmd, attrs);
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("many-to-one"))
            {
                // N-1 field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newFieldObject(cmd, attrs);
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("many-to-many"))
            {
                // M-N field
                ClassMetaData cmd = (ClassMetaData)getStack();
                FieldMetaData fmd = newFieldObject(cmd, attrs);
                cmd.addProperty(fmd);

                pushStack(fmd);
            }
            else if (qName.equals("version"))
            {
                if (getStack() instanceof ClassMetaData)
                {
                    // Version field
                    ClassMetaData cmd = (ClassMetaData)getStack();
                    FieldMetaData fmd = newFieldObject(cmd, attrs);
                    cmd.addProperty(fmd);

                    pushStack(fmd);
                }
            }
            else if (qName.equals("generated-value"))
            {
                // generated value for this field
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                // TODO Support "generator" attribute (use specified table or sequence generator)
//                String generator = getAttr(attrs, "generator");
                IdentityStrategy idStrategy = IdentityStrategy.getIdentityStrategy(getAttr(attrs, "strategy"));
                fmd.setValueStrategy(idStrategy);
            }
            else if (qName.equals("join-table"))
            {
                // Join table for this field
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                JoinMetaData joinmd = new JoinMetaData(fmd, getAttr(attrs, "name"), null, null, null);
                fmd.setJoinMetaData(joinmd);
            }
            else if (qName.equals("column"))
            {
                // Column for the current field
                AbstractPropertyMetaData fmd = (AbstractPropertyMetaData)getStack();
                ColumnMetaData colmd =
                    new ColumnMetaData(fmd,
                        getAttr(attrs, "name"),
                        null, // No "target" in JPA
                        null, // No "target-field" in JPA
                        null, // No "jdbc-type" in JPA (use Lob, Basic, Temporal to decide this somehow)
                        null, // No "sql-type" in JPA
                        getAttr(attrs, "length"), // Should use precision if floating point type
                        getAttr(attrs, "scale"),
                        getAttr(attrs, "nullable"),
                        null, // No default-value in JPA
                        null, // No insert-value in JPA
                        getAttr(attrs, "insertable"),
                        getAttr(attrs, "updatable"),
                        getAttr(attrs, "unique"));
                fmd.addColumn(colmd);
            }
            else if (qName.equals("entity-listeners"))
            {
                // Nothing to add at this point
            }
            else if (qName.equals("entity-listener"))
            {
                ClassMetaData cmd = (ClassMetaData)getStack();
                cmd.addListener(getAttr(attrs, "name"));
            }
            else if (qName.equals("pre-persist"))
            {
                // Pre-create callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PrePersist", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("post-persist"))
            {
                // Post-create callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PostPersist", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("pre-remove"))
            {
                // Pre-delete callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PreRemove", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("post-remove"))
            {
                // Post-delete callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PostRemove", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("pre-update"))
            {
                // Pre-store callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PreUpdate", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("post-update"))
            {
                // Post-store callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PostUpdate", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("post-load"))
            {
                // Post-load callback
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.addCallback("javax.persistence.PostLoad", getAttr(attrs, "method-name"));
            }
            else if (qName.equals("exclude-default-listeners"))
            {
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.excludeDefaultListeners();
            }
            else if (qName.equals("exclude-superclass-listeners"))
            {
                AbstractClassMetaData cmd = (AbstractClassMetaData)getStack();
                cmd.excludeSuperClassListeners();
            }
            else
            {
                String message = LOCALISER.msg("MetaData.Parser.UnknownTag.Error", qName);
                JPOXLogger.METADATA.error(message);
                throw new RuntimeException(message);
            }
        }
        catch(RuntimeException ex)
        {
            //TODO localise
            JPOXLogger.METADATA.error("The following error occurred while parsing <" + qName + "> nested within " + getStack());
            throw ex;
        }
    }

    /**
     * Handler method called at the end of an element.
     * @param uri URI of the tag
     * @param localName local name
     * @param qName Name of element just ending
     * @throws SAXException in parsing errors
     */
    public void endElement(String uri, String localName, String qName)
    throws SAXException
    {
        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.EndElement", "<" + qName + ">", "" + stack.size()));
        }

        // Save the current string for elements that have a body value
        String currentString = getString().trim();
        if (currentString.length() > 0)
        {
            ExtendableMetaData extmd = (ExtendableMetaData)getStack();
            if (qName.equals("schema"))
            {
                if (extmd instanceof FileMetaData)
                {
                    ((FileMetaData)extmd).schema = currentString;
                }
            }
            else if (qName.equals("catalog"))
            {
                if (extmd instanceof FileMetaData)
                {
                    ((FileMetaData)extmd).catalog = currentString;
                }
            }
            else if (qName.equals("package"))
            {
                if (extmd instanceof FileMetaData)
                {
                    // Add the default package
                    FileMetaData filemd = (FileMetaData)extmd;
                    filemd.addPackage(new PackageMetaData(filemd, currentString, null, null));
                    defaultPackageName = currentString;
                }
            }
            else if (qName.equals("discriminator-value"))
            {
                if (extmd instanceof ClassMetaData)
                {
                    // Add the discriminator value
                    ClassMetaData cmd = (ClassMetaData)extmd;
                    InheritanceMetaData inhmd = cmd.getInheritanceMetaData();
                    if (inhmd == null)
                    {
                        // Add an empty inheritance specification
                        inhmd = new InheritanceMetaData(cmd, null);
                        cmd.setInheritanceMetaData(inhmd);
                    }
                    String discrimValue = currentString;
                    DiscriminatorMetaData discmd = new DiscriminatorMetaData(inhmd, null, discrimValue, 
                        DiscriminatorStrategy.VALUE_MAP.toString(), null);
                    inhmd.setDiscriminatorMetaData(discmd);
                }
            }
            else if (qName.equals("query"))
            {
                if (extmd instanceof QueryMetaData)
                {
                    // Named query, so set the query string
                    ((QueryMetaData)extmd).setQuery(currentString);
                }
            }
        }

        // Pop the tag
        // If startElement pushes an element onto the stack need a remove here for that type
        if (qName.equals("entity") ||
            qName.equals("id") ||
            qName.equals("embedded-id") ||
            qName.equals("basic") ||
            qName.equals("transient") ||
            qName.equals("one-to-one") ||
            qName.equals("one-to-many") ||
            qName.equals("many-to-one") ||
            qName.equals("many-to-many") ||
            qName.equals("version") ||
            qName.equals("named-query") ||
            qName.equals("named-native-query"))
        {
            popStack();
        }
    }
}