/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.RegisterClassEvent;
import javax.jdo.spi.RegisterClassListener;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;

/**
 * Manager of MetaData information in JPOX.
 * <P>
 * Acts as a registry of metadata so that metadata files don't need to be 
 * parsed multiple times. MetaData is stored as a FileMetaData, which contains
 * PackageMetaData, which contains ClassMetaData, and so on. This maps exactly
 * to the users model of their metadata. The users access point is 
 * <B>getMetaDataForClass()</B> which will check the known classes without metadata,
 * then check the existing registered metdata, then check the valid locations for 
 * metdata files. This way, the metadata is managed from this single point.
 * </P>
 * <P>
 * Maintains a list of all classes that have been checked for MetaData and
 * don't have any available. This avoids the needs to look up MetaData multiple
 * times finding the same result. Currently this list is for all ClassMetaData 
 * objects keyed by the class name.
 * </P>
 *
 * @version $Revision: 1.102 $ 
 */
public abstract class MetaDataManager
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.metadata.Localisation");

    /** The PMF Context that this metadata manager is operating in. */
    protected PMFContext pmfContext = null;

    /** Manager for annotations. */
    protected AnnotationManager annotationManager = null;

    /** Parser for MetaData. */
    protected MetaDataParser metaDataParser = null;

    /** Flag whether we should validate the metadata files when parsing. */
    protected boolean validateMetaData = true;

    /** Map of FileMetaData for the parsed files, keyed by the URL string. */
    protected Map fileMetaDataByURLString = new HashMap();

    /** Map of ClassMetaData, keyed by the class name. */
    protected Map classMetaDataByClass = new HashMap();

    /** Map of QueryMetaData, keyed by the (class name + query name). */
    protected Map queryMetaDataByName = new HashMap();

    /** Map of SequenceMetaData, keyed by the package name and sequence name. */
    protected Map sequenceMetaDataByPackageSequence = new HashMap();

    /** Map of TableGeneratorMetaData, keyed by the package name and generator name. */
    protected Map tableGeneratorMetaDataByPackageSequence = new HashMap();

    /** Indicator for whether this manager is managing the enhancement process, else it's runtime. */
    protected boolean enhancing = false;

    /**
     * Constructor, specifying the PMFContext used.
     * @param pmfContext PMF Context that this metadata manager operates in
     */
    public MetaDataManager(PMFContext pmfContext)
    {
        this.pmfContext = pmfContext;
        JDOImplHelper.getInstance().removeRegisterClassListener(new MetaDataRegisterClassListener());
        this.setValidate(pmfContext.getPmfConfiguration().getMetaDataValidate());

        if (JavaUtils.isJRE1_5OrAbove())
        {
            try
            {
                ClassLoaderResolver clr = pmfContext.getClassLoaderResolver(null);
                // TODO This is hardcoded to the Java5 manager. Make it configurable ?
                Class annotationReaderClass = clr.classForName("org.jpox.metadata.annotations.AnnotationManagerImpl");

                Class[] ctrArgs = new Class[] {MetaDataManager.class};
                Object[] ctrParams = new Object[] {this};
                Constructor ctor = annotationReaderClass.getConstructor(ctrArgs);
                annotationManager = (AnnotationManager)ctor.newInstance(ctrParams);
            }
            catch (Exception e)
            {
                JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.AnnotationManagerNotFound"));
            }
        }
    }

    /**
     * Utility to create a new class component.
     * This method exists so that the Enhancer can create its own ClassMetaData objects using its class.
     * @param pmd MetaData for package
     * @param name Name of class
     * @param identityType identity type
     * @param objectidClass name of objectid class
     * @param requiresExtent Whether an extent is required
     * @param detachable Whether it is detachable
     * @param embeddedOnly Whether it is embedded only
     * @param modifier persistence modifier
     * @param persistenceCapableSuperclass PC superclass (deprecated)
     * @param catalog catalog to use
     * @param schema schema to use
     * @param table table to use
     * @param entityName the entity name required by JPA 4.3.1
     * @return The MetaData for the class
     */
    public ClassMetaData newClassObject(PackageMetaData pmd, String name, String identityType, String objectidClass,
            String requiresExtent, String detachable, String embeddedOnly, String modifier, String persistenceCapableSuperclass,
            String catalog, String schema, String table, String entityName)
    {
        ClassMetaData cmd = new ClassMetaData(pmd, name, identityType, objectidClass, requiresExtent, detachable,
            embeddedOnly, modifier, persistenceCapableSuperclass, catalog, schema, table, entityName);
        return cmd;
    }

    /**
     * Utility to create a new field component.
     * This method exists so that the Enhancer can create its own FieldMetaData objects using its class.
     * @param md Parent metadata
     * @param name Name of field
     * @param pk Whether it is PK
     * @param modifier persistence modifier
     * @param defaultFetchGroup Whether it is in DFG
     * @param nullValue Behaviour on null value
     * @param embedded Whether it is embedded
     * @param serialized Whether it is serialised
     * @param dependent Whether it is dependent
     * @param mappedBy Field that it is mapped into
     * @param column column to use
     * @param table table to use
     * @param deleteAction Action on FK delete
     * @param indexed whether the column is indexed
     * @param unique Whether the column is unique
     * @param recursionDepth Recursion depth
     * @param loadFetchGroup Whether to load the fetch group
     * @param valueStrategy Value strategy for generating field values
     * @param sequence Sequence name if required
     * @param fieldType Type of the field
     * @return FieldMetaData
     */
    public FieldMetaData newFieldObject(MetaData md, String name, String pk, String modifier, String defaultFetchGroup,
            String nullValue, String embedded, String serialized, String dependent, String mappedBy, String column,
            String table, String deleteAction, String indexed, String unique, String recursionDepth, String loadFetchGroup,
            String valueStrategy, String sequence, String fieldType)
    {
        FieldMetaData fmd = new FieldMetaData(md, name, pk, modifier, defaultFetchGroup, nullValue, embedded, serialized,
            dependent, mappedBy, column, table, deleteAction, indexed, unique, recursionDepth, loadFetchGroup, valueStrategy,
            sequence, fieldType);
        return fmd;
    }

    /**
     * Utility to create a new property component.
     * This method exists so that the Enhancer can create its own PropertyMetaData objects using its class.
     * @param md Parent metadata
     * @param name Name of field
     * @param pk Whether it is PK
     * @param modifier persistence modifier
     * @param defaultFetchGroup Whether it is in DFG
     * @param nullValue Behaviour on null value
     * @param embedded Whether it is embedded
     * @param serialized Whether it is serialised
     * @param dependent Whether it is dependent
     * @param mappedBy Field that it is mapped into
     * @param column column to use
     * @param table table to use
     * @param deleteAction Action on FK delete
     * @param indexed whether the column is indexed
     * @param unique Whether the column is unique
     * @param recursionDepth Recursion depth
     * @param loadFetchGroup Whether to load the fetch group
     * @param valueStrategy Value strategy for generating field values
     * @param sequence Sequence name if required
     * @param fieldType Type of the field
     * @param fieldName Name of the field
     * @return PropertyMetaData
     */
    public PropertyMetaData newPropertyObject(MetaData md, String name, String pk, String modifier, String defaultFetchGroup,
            String nullValue, String embedded, String serialized, String dependent, String mappedBy, String column,
            String table, String deleteAction, String indexed, String unique, String recursionDepth, String loadFetchGroup,
            String valueStrategy, String sequence, String fieldType, String fieldName)
    {
        PropertyMetaData pmd = new PropertyMetaData(md, name, pk, modifier, defaultFetchGroup, nullValue, embedded, serialized,
            dependent, mappedBy, column, table, deleteAction, indexed, unique, recursionDepth, loadFetchGroup, valueStrategy,
            sequence, fieldType, fieldName);
        return pmd;
    }

    /**
     * Mutator for whether to validate the MetaData files for XML compliance.
     * @param validate Whether to validate
     */
    public void setValidate(boolean validate)
    {
        validateMetaData = validate;
    }

    /**
     * Accessor for the PMF Context that this manager is running in.
     * @return The PMF Context
     */
    public PMFContext getPMFContext()
    {
        return pmfContext;
    }

    /**
     * Clear resources
     */
    public void close()
    {
        classMetaDataByClass.clear();
        queryMetaDataByName.clear();
        sequenceMetaDataByPackageSequence.clear();
        tableGeneratorMetaDataByPackageSequence.clear();
        fileMetaDataByURLString.clear();
        classMetaDataByClass = null;
        queryMetaDataByName = null;
        sequenceMetaDataByPackageSequence = null;
        tableGeneratorMetaDataByPackageSequence = null;
        fileMetaDataByURLString = null;
    }

    /**
     * Accessor for whether we are managing the enhancement process.
     * @return Whether we are enhancing
     */
    public boolean isEnhancing()
    {
        return enhancing;
    }

    /**
     * Accessor for the names of the classes with MetaData currently registered with this manager.
     * @return Names of classes with MetaData
     */
    public Collection getClassesWithMetaData()
    {
        return Collections.unmodifiableCollection(classMetaDataByClass.keySet());
    }

    /**
     * Convenience method to check if we have metadata for the specified class
     * @param className The name of the class to check
     * @return Whether the metadata is already registered for this class
     */
    public abstract boolean hasMetaDataForClass(String className);

    /**
     * Accessor for whether a class doesn't have MetaData.
     * @param class_name Name of the class
     * @return Whether it has no metadata.
     */
    protected abstract boolean isClassWithoutMetaData(String class_name);

    /**
     * Accessor for the MetaData for a class given the name and a loader.
     * All MetaData returned from this method will be initialised and ready for full use.
     * If the class can't be loaded, null will be returned. 
     * @param className Name of the class to find MetaData for
     * @param clr ClassLoaderResolver resolver for use in loading the class.
     * @return The ClassMetaData for this class (or null if not found)
     **/
    public synchronized AbstractClassMetaData getMetaDataForClass(String className, ClassLoaderResolver clr)
    {
        if (className == null)
        {
            return null;
        }

        // Check if this class has no MetaData before instantiating its class
        if (isClassWithoutMetaData(className))
        {
            return null;
        }

        // Resolve the class
        Class c = null;
        try
        {
            if (clr == null)
            {
                c = Class.forName(className);
            }
            else
            {
                c = clr.classForName(className, null, false);
            }
        }
        catch (ClassNotFoundException cnfe)
        {
        }
        catch (ClassNotResolvedException cnre)
        {
        }
        if (c == null)
        {
            return null;
        }

        return getMetaDataForClass(c, clr);
    }

    /** Temporary list of the FileMetaData objects utilised in this call for metadata. */
    ArrayList utilisedFileMetaData = new ArrayList();

    /**
     * Main accessor for the MetaData for a class.
     * All MetaData returned from this method will be initialised and ready for full use.
     * @param c The class to find MetaData for
     * @param clr the ClassLoaderResolver
     * @return The ClassMetaData for this class (or null if not found)
     */
    public synchronized AbstractClassMetaData getMetaDataForClass(Class c, ClassLoaderResolver clr)
    {
        if (c == null)
        {
            return null;
        }

        AbstractClassMetaData cmd = null;
        if (c.isInterface())
        {
            // "persistent-interface" - check if it has class built at runtime and return the MetaData for it 
            cmd = getClassMetaDataForImplementationOfPersistentInterface(c.getName());
        }
        else
        {
            // "persistent-class"
            cmd = getMetaDataForClassInternal(c, clr);
        }

        if (cmd != null)
        {
            // Make sure that anything returned is initialised
            if (!cmd.isInitialised())
            {
                // Initialise the class in question
                cmd.initialise();
            }

            if (utilisedFileMetaData.size() > 0)
            {
                // Initialise all FileMetaData that were processed in this call
                Iterator iter = utilisedFileMetaData.iterator();
                while (iter.hasNext())
                {
                    FileMetaData filemd = (FileMetaData)iter.next();
                    initialiseFileMetaData(filemd, clr,c.getClassLoader());
                }
            }
        }

        utilisedFileMetaData.clear();
        return cmd;
    }

    /**
     * Accessor for the MetaData for a class.
     * @param entityName The entity name to find MetaData for
     * @return The ClassMetaData for this entity name (or null if not found)
     */
    public synchronized AbstractClassMetaData getMetaDataForEntityName(String entityName)
    {
        // By default dont support "entity-name"s
        return null;
    }

    /**
     * Method to access the (already known) metadata for the specified class.
     * If the class is not yet known about it returns null.
     * @param className Name of the class
     * @return MetaData for the class
     */
    public ClassMetaData readMetaDataForClass(String className)
    {
        return (ClassMetaData)classMetaDataByClass.get(className);
    }

    /**
     * Method to access the (already known) metadata for the field of the specified class.
     * If the class (or this field) is not yet known about it returns null.
     * @param className Name of the class
     * @param fieldName Name of the field
     * @return MetaData for the field
     */
    public AbstractPropertyMetaData readMetaDataForField(String className, String fieldName)
    {
        AbstractClassMetaData cmd = readMetaDataForClass(className);
        return (cmd != null ? cmd.getField(fieldName) : null);
    }

    /**
     * Internal convenience method for accessing the MetaData for a class.
     * MetaData returned by this method may be uninitialised so should only really
     * be used in initialisation processes.
     * To be implemented by the implementing class.
     * @param c The class to find MetaData for
     * @return The ClassMetaData for this class (or null if not found)
     **/
    public abstract AbstractClassMetaData getMetaDataForClassInternal(Class c, ClassLoaderResolver clr);

    /**
     * Accessor for the subclasses of a particular class
     * @param className Name of the class that we want the known subclasses for.
     * @param includeDescendents Whether to include subclasses of subclasses etc
     * @return Names of the subclasses. return null if there are no subclasses
     */
    public String[] getSubclassesForClass(String className, boolean includeDescendents)
    {
        Collection subclassNames = new HashSet();
        
        provideSubclassesForClass(className, includeDescendents, subclassNames);

        if (subclassNames.size() > 0)
        {
            return (String[])subclassNames.toArray(new String[subclassNames.size()]);
        }

        return null;
    }

    /**
     * Provide the subclasses of a particular class to a given <code>consumer</code>
     * @param className Name of the class that we want the known subclasses for.
     * @param includeDescendents Whether to include subclasses of subclasses etc
     * @param consumer the Collection (Set) where discovered subclasses are added
     */
    private void provideSubclassesForClass(String className, boolean includeDescendents, Collection consumer)
    {
        Collection cmds = classMetaDataByClass.values();
        Iterator cmdIter = cmds.iterator();
        while (cmdIter.hasNext())
        {
            AbstractClassMetaData acmd = (AbstractClassMetaData)cmdIter.next();
            if (acmd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)acmd;
                if (cmd.getPersistenceCapableSuperclass() != null &&
                    cmd.getPersistenceCapableSuperclass().equals(className))
                {
                    if( consumer.add(cmd.getFullClassName()) && includeDescendents)
                    {
                        //go deeper in subclasses
                        provideSubclassesForClass(cmd.getFullClassName(), includeDescendents, consumer);
                    }
                }
            }
        }
    }

    /**
     * Accessor for the list of names of classes that are declared to implement the
     * specified interface (using &lt;implements&gt; in the MetaData). Ignore abstract classes
     * @param interfaceName Name of the interface
     * @param clr The ClassLoaderResolver
     * @return The names of the classes declared as implementing that interface. return null if no classes
     */
    public String[] getClassesImplementingInterface(String interfaceName, ClassLoaderResolver clr)
    {
        Collection cmds = classMetaDataByClass.values();
        Iterator cmdIter = cmds.iterator();
        Collection classNames = new HashSet();
        boolean isPersistentInterface = false;
        while (cmdIter.hasNext())
        {
            AbstractClassMetaData acmd = (AbstractClassMetaData)cmdIter.next();
            if (acmd instanceof ClassMetaData)
            {
                if (!acmd.isInitialised())
                {
                    // Make sure that we are initialised since implementsMetaData wont be set
                    acmd.initialise();
                }
                if (clr.classForName(interfaceName).isAssignableFrom(clr.classForName(((ClassMetaData)acmd).getFullClassName())))
                {
                	if (!((ClassMetaData)acmd).isAbstractPersistenceCapable())
                	{
                        classNames.add(((ClassMetaData)acmd).getFullClassName());
                	}
                }
            }
            else if (acmd instanceof InterfaceMetaData)
            {
                if (clr.classForName(interfaceName).isAssignableFrom(clr.classForName(((InterfaceMetaData)acmd).getFullClassName())))
                {
                    isPersistentInterface = true;
                }
            }
        }
        if (isPersistentInterface)
        {
            //TODO this is very time consuming. got to do some cache
            classNames.add(pmfContext.getImplementationCreator().newInstance(clr.classForName(interfaceName), this, clr).getClass().getName());
        }

        if (classNames.size() > 0)
        {
            return (String[])classNames.toArray(new String[classNames.size()]);
        }
        else
        {
            return null;
        }
    }

    /**
     * Load up and add any O/R mapping info for the specified class to the stored ClassMetaData (if supported).
     * This implementation does nothing so if ORM files are supported then this should be overridden by subclasses
     * @param c The class
     * @param clr the ClassLoaderResolver
     */
    public void addORMDataToClass(Class c, ClassLoaderResolver clr)
    {
        // Default to doing nothing. Specified in subclasses if they support it
        return;
    }

    /**
     * Accessor for the MetaData for an implementation of a reference type.
     * Finds the metadata for the implementation of this reference.
     * @param referenceClass The reference class to find MetaData for
     * @param implValue Object of an implementation class, to return if possible (null=ignore)
     * @param clr ClassLoader resolver
     * @return The ClassMetaData for an implementation of a reference type
     */
    public ClassMetaData getMetaDataForImplementationOfReference(Class referenceClass, Object implValue, ClassLoaderResolver clr)
    {
        if (referenceClass == null || (!referenceClass.isInterface() && referenceClass != java.lang.Object.class))
        {
            return null;
        }

        // Check if this is a "persistent interface"
        Object intfMetaData = getClassMetaDataForImplementationOfPersistentInterface(referenceClass.getName());
        if (intfMetaData != null)
        {
            return (ClassMetaData)intfMetaData;
        }

        ClassMetaData cmd = null;

        // Search for the class required
        Set classMetaDataClasses = classMetaDataByClass.keySet();
        Iterator classMetaDataClassesIter = classMetaDataClasses.iterator();
        while (classMetaDataClassesIter.hasNext())
        {
            String class_name = (String)classMetaDataClassesIter.next();
            AbstractClassMetaData acmd_cls = (AbstractClassMetaData)classMetaDataByClass.get(class_name);

            if (acmd_cls instanceof ClassMetaData)
            {
                try
                {
                    // Check if class is implementation of "implValue" (in the case of java.lang.Object, all will be!)
                    Class cls = referenceClass.getClassLoader().loadClass(class_name);
                    if (referenceClass.isAssignableFrom(cls))
                    {
                        // Find the base class that is an implementation
                        cmd = (ClassMetaData)acmd_cls;
                        if (implValue != null && cmd.getFullClassName().equals(implValue.getClass().getName()))
                        {
                            return cmd;
                        }

                        AbstractClassMetaData cmd_superclass = cmd.getSuperAbstractClassMetaData();
                        while (cmd_superclass != null)
                        {
                            if( !referenceClass.isAssignableFrom(clr.classForName(((ClassMetaData) cmd_superclass).getFullClassName())) )
                            {
                                break;
                            }
                            // TODO Check if superclass is an implementation
                            cmd = (ClassMetaData) cmd_superclass;
                            if (implValue != null && cmd.getFullClassName().equals(implValue.getClass().getName()))
                            {
                                break;
                            }

                            // Go to next superclass
                            cmd_superclass = cmd_superclass.getSuperAbstractClassMetaData();
                            if (cmd_superclass == null)
                            {
                                break;
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                }
            }
        }

        return cmd;
    }

    /**
     * Accessor for the MetaData for a field of a class. Utilises
     * getMetaDataForClass, and then finds the relevant field.
     * @param className The name of the class owning the field
     * @param fieldName The name of the field to find MetaData for
     * @param clr ClassLoaderResolver resolver for any loading of classes
     * @return The AbstractPropertyMetaData for this field (or null if not found)
     */
    public AbstractPropertyMetaData getMetaDataForField(String className, String fieldName, ClassLoaderResolver clr)
    {
        if (className == null || fieldName == null)
        {
            return null;
        }

        AbstractClassMetaData cmd = getMetaDataForClass(className, clr);
        return (cmd != null ? cmd.getField(fieldName) : null);
    }

    /**
     * Accessor for the MetaData for a field of a class.
     * Utilises getMetaDataForClass, and then finds the relevant field.
     * @param c The class owning the field
     * @param clr the ClassLoaderResolver
     * @param field_name The name of the field to find MetaData for
     * @return The AbstractPropertyMetaData for this field (or null if not found)
     **/
    public AbstractPropertyMetaData getMetaDataForField(Class c, ClassLoaderResolver clr, String field_name)
    {
        if (c == null || field_name == null)
        {
            return null;
        }

        AbstractClassMetaData cmd = getMetaDataForClass(c, clr);
        return (cmd != null ? cmd.getField(field_name) : null);
    }

    /**
     * Accessor for the MetaData for a named query for a class.
     * If the class is not specified, searches for the query with this name for any class.
     * Will only return metadata for queries already registered in this implementation.
     * @param cls The class which has the query defined for it
     * @param clr the ClassLoaderResolver
     * @param queryName Name of the query
     * @return The QueryMetaData for the query for this class
     **/
    public QueryMetaData getMetaDataForQuery(Class cls, ClassLoaderResolver clr, String queryName)
    {
        if (queryName == null)
        {
            return null;
        }

        String query_key = queryName;
        if (cls != null)
        {
            query_key = cls.getName() + "_" + queryName;
        }

        return (QueryMetaData)queryMetaDataByName.get(query_key);
    }

    /**
     * Accessor for the MetaData for a Sequence in a package.
     * This implementation simply checks what is already loaded and returns if found
     * @param clr the ClassLoaderResolver
     * @param seqName Name of the package (fully qualified if necessary)
     * @return The SequenceMetaData for this named sequence
     **/
    public SequenceMetaData getMetaDataForSequence(ClassLoaderResolver clr, String seqName)
    {
        if (seqName == null)
        {
            return null;
        }

        return (SequenceMetaData)sequenceMetaDataByPackageSequence.get(seqName);
    }

    /**
     * Accessor for the MetaData for a TableGenerator in a package.
     * This implementation simply checks what is already loaded and returns if found
     * @param clr the ClassLoaderResolver
     * @param genName Name of the package (fully qualified if necessary)
     * @return The TableGenerator for this named generator
     **/
    public TableGeneratorMetaData getMetaDataForTableGenerator(ClassLoaderResolver clr, String genName)
    {
        if (genName == null)
        {
            return null;
        }

        return (TableGeneratorMetaData)tableGeneratorMetaDataByPackageSequence.get(genName);
    }

    // ------------------------------- Persistent Interfaces ---------------------------------------

    /**
     * Accessor for the MetaData for an interface.
     * Part of the support for "persistent-interface".
     * This defaults to returning null since interfaces are only supported by JDO.
     * @param c The interface to find MetaData for
     * @param clr the ClassLoaderResolver
     * @return The InterfaceMetaData for this interface (or null if not found)
     */
    public InterfaceMetaData getMetaDataForInterface(Class c, ClassLoaderResolver clr)
    {
        return null;
    }

    /**
     * Convenience method to return if the passed class name is a "persistent-interface".
     * @param name Name if the interface
     * @return Whether it is a "persistent-interface"
     */
    public boolean isPersistentInterface(String name)
    {
        // Default to not supporting "persistent-interface"s
        return false;
    }

    /**
     * Convenience method to return if the passed class name is an implementation of the passed "persistent-interface".
     * @param interfaceName Name of the persistent interface
     * @param implName The implementation name
     * @return Whether it is a (JPOX-generated) impl of the persistent interface
     */
    public boolean isPersistentInterfaceImplementation(String interfaceName, String implName)
    {
        // Default to not supporting "persistent-interface"s
        return false;
    }

    /**
     * Accessor for the implementation name for the specified "persistent-interface".
     * @param interfaceName The name of the persistent interface
     * @return The name of the implementation class
     */
    public String getImplementationNameForPersistentInterface(String interfaceName)
    {
        // Default to not supporting "persistent-interface"s
        return null;
    }

    /**
     * Accessor for the metadata for the implementation of the specified "persistent-interface".
     * @param interfaceName The name of the persistent interface
     * @return The ClassMetaData of the implementation class
     */
    public ClassMetaData getClassMetaDataForImplementationOfPersistentInterface(String interfaceName)
    {
        // Default to not supporting "persistent-interface"s
        return null;
    }

    /**
     * Method to register a persistent interface and its implementation with the MetaData system.
     * @param imd MetaData for the interface
     * @param implClass The implementation class
     * @param clr ClassLoader Resolver to use
     */
    public void registerPersistentInterface(InterfaceMetaData imd, Class implClass, ClassLoaderResolver clr)
    {
        // Default to not supporting "persistent-interface"s
        return;
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Method to parse the available "persistence.xml" files returning the metadata for all found.
     * Searches for all files "META-INF/persistence.xml" in the CLASSPATH of the current thread.
     * @return The metadata for all "persistence.xml" files
     */
    public PersistenceFileMetaData[] parsePersistenceFiles()
    {
        HashSet metadata = new HashSet();
        try
        {
            // Find all "META-INF/persistence.xml" files in the CLASSPATH of the current thread
            // TODO Use ClassLoaderResolver
            Enumeration files = Thread.currentThread().getContextClassLoader().getResources("META-INF/persistence.xml");
            if (!files.hasMoreElements())
            {
                return null;
            }

            metaDataParser = new MetaDataParser(this, validateMetaData);
            for ( ; files.hasMoreElements() ;)
            {
                // Parse the "persistence.xml"
                URL fileURL = (URL)files.nextElement();
                MetaData permd = metaDataParser.parseMetaDataURL(fileURL, "persistence");
                metadata.add(permd);
            }
        }
        catch (IOException ioe)
        {
            // Do nothing
        }

        return (PersistenceFileMetaData[])metadata.toArray(new PersistenceFileMetaData[metadata.size()]);
    }

    /**
     * Utility to parse a MetaData file.
     * @param file_url URL of the file
     * @return The FileMetaData for this file
     */
    protected abstract FileMetaData parseFile(URL file_url);

    /**
     * Method to take the FileMetaData and register the relevant parts of it with the assorted caches provided.
     * @param fileURLString URL of the metadata file
     * @param filemd The File MetaData
     */
    protected abstract void registerFile(String fileURLString, FileMetaData filemd);

    /**
     * Convenience method to register all sequences found in the passed file.
     * @param filemd MetaData for the file
     */
    protected void registerSequencesForFile(FileMetaData filemd)
    {
        // Register all sequences for the packages in this file
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            SequenceMetaData[] seqmds = pmd.getSequences();
            if (seqmds != null)
            {
                for (int j=0;j<seqmds.length;j++)
                {
                    // Register using its fully qualified name (JDO)
                    sequenceMetaDataByPackageSequence.put(seqmds[j].getFullyQualifiedName(), seqmds[j]);
                }
            }
        }
    }

    /**
     * Convenience method to register all table generators found in the passed file.
     * @param filemd MetaData for the file
     */
    protected void registerTableGeneratorsForFile(FileMetaData filemd)
    {
        // Register all sequences for the packages in this file
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            TableGeneratorMetaData[] tgmds = pmd.getTableGenerators();
            if (tgmds != null)
            {
                for (int j=0;j<tgmds.length;j++)
                {
                    // Register using its fully qualified name (JDO)
                    tableGeneratorMetaDataByPackageSequence.put(tgmds[j].getFullyQualifiedName(), tgmds[j]);
                }
            }
        }
    }

    /**
     * Convenience method to register all queries found in the passed file.
     * @param filemd MetaData for the file
     */
    protected void registerQueriesForFile(FileMetaData filemd)
    {
        // Register all queries for this file
        // Store queries against "queryname"
        QueryMetaData[] queries = filemd.getQueries();
        if (queries != null)
        {
            for (int i=0;i<queries.length;i++)
            {
                queryMetaDataByName.put(queries[i].getName(), queries[i]);
            }
        }

        for (int i = 0; i < filemd.getNoOfPackages(); i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);

            // Register all classes (and their queries) into the respective lookup maps
            for (int j = 0; j < pmd.getNoOfClasses(); j++)
            {
                // Store queries against "classname_queryname"
                ClassMetaData cmd = pmd.getClass(j);
                QueryMetaData[] classQueries = cmd.getQueries();
                if (classQueries != null)
                {
                    for (int k = 0; k < classQueries.length; k++)
                    {
                        queryMetaDataByName.put(cmd.getFullClassName() + "_" + classQueries[k].getName(), classQueries[k]);
                    }
                }
            }

            // Register all interfaces (and their queries) into the respective lookup maps
            for (int j = 0; j < pmd.getNoOfInterfaces(); j++)
            {
                // Store queries against "classname_queryname"
                InterfaceMetaData intfmd = pmd.getInterface(j);
                QueryMetaData[] interfaceQueries = intfmd.getQueries();
                if (interfaceQueries != null)
                {
                    for (int k = 0; k < interfaceQueries.length; k++)
                    {
                        queryMetaDataByName.put(intfmd.getFullClassName() + "_" + interfaceQueries[k].getName(), interfaceQueries[k]);
                    }
                }
            }
        }
    }

    /**
     * Convenience method to populate all classes/interfaces in a Meta-Data file.
     * @param filemd The MetaData file
     * @param clr Class Loader to use in population
     * @param primary the primary ClassLoader to use (or null)
     */
    protected void populateFileMetaData(FileMetaData filemd, ClassLoaderResolver clr, ClassLoader primary)
    {
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            for (int j=0;j<pmd.getNoOfClasses();j++)
            {
                AbstractClassMetaData cmd = pmd.getClass(j);
                if (!cmd.isPopulated() && !cmd.isInitialised())
                {
                    cmd.populate(clr, primary);
                }
            }
            for (int j=0;j<pmd.getNoOfInterfaces();j++)
            {
                AbstractClassMetaData cmd = pmd.getInterface(j);
                if (!cmd.isPopulated() && !cmd.isInitialised())
                {
                    cmd.populate(clr, primary);
                }
            }
        }
    }

    /**
     * Initialise all classes/interfaces in a Meta-Data file.
     * @param filemd the FileMetaData
     * @param clr ClassLoader resolver to use
     * @param primary the primary ClassLoader to use (or null)
     */
    protected void initialiseFileMetaData(FileMetaData filemd, ClassLoaderResolver clr, ClassLoader primary)
    {
        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd = filemd.getPackage(i);
            for (int j=0;j<pmd.getNoOfClasses();j++)
            {
                ClassMetaData cmd = pmd.getClass(j);
                try
                {
                    initialiseClassMetaData(cmd, clr.classForName(cmd.getFullClassName(),primary), clr);
                }
                catch(JDOException jdoex)
                {
                    throw jdoex;
                }
                catch (RuntimeException re)
                {
                    // Do nothing
                }
            }

            for (int j=0;j<pmd.getNoOfInterfaces();j++)
            {
                InterfaceMetaData imd = pmd.getInterface(j);
                try
                {
                    initialiseInterfaceMetaData(imd, clr, primary);
                }
                catch(JDOException jdoex)
                {
                    throw jdoex;
                }
                catch (RuntimeException re)
                {
                    // Do nothing
                }
            }
        }
    }

    /**
     * Utility to initialise the MetaData for a class, using the specified
     * class. This assigns defaults to tags that haven't been assigned.
     * If the class that is being used to populate the MetaData is not
     * enhanced, this will throw a JDOUserException informing them of this. 
     * @param cmd The classes metadata
     * @param cls The class to use as a basis for initialisation
     * @param clr ClassLoader resolver to use
     */
    protected void initialiseClassMetaData(ClassMetaData cmd, Class cls, ClassLoaderResolver clr)
    {
        synchronized(cmd)
        {
            if (cmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE &&
                !ClassUtils.isPersistenceCapableClass(cls))
            {
                throw new JDOUserException(LOCALISER.msg("MetaData.Class.NotEnhancedError", cls.getName()));
            }
            if (!cmd.isPopulated() && !cmd.isInitialised())
            {
                cmd.populate(clr, cls.getClassLoader());
            }
            if (!cmd.isInitialised())
            {
                cmd.initialise();
            }
        }
    }

    /**
     * Utility to initialise the MetaData for a interface, using the specified
     * class. This assigns defaults to tags that haven't been assigned.
     * If the class that is being used to populate the MetaData is not
     * enhanced, this will throw a JDOUserException informing them of this. 
     * @param imd The interface metadata
     * @param clr The loader of the interface
     * @param primary the primary ClassLoader to use (or null)
     */
    protected void initialiseInterfaceMetaData(InterfaceMetaData imd, ClassLoaderResolver clr, ClassLoader primary)
    {
        synchronized(imd)
        {
            if (!imd.isPopulated() && !imd.isInitialised())
            {
                imd.populate(clr, primary);
            }
            if (!imd.isInitialised())
            {
                imd.initialise();
            }
        }
    }

    /**
     * Method to load the annotations for the specified class and return the FileMetaData containing
     * the class. The FileMetaData, PackageMetaData will be dummy records.
     * @param cls The class
     * @param clr ClassLoader resolver
     * @param populate Whether to populate the data
     * @return The FileMetaData
     */
    protected FileMetaData loadAnnotationsForClass(Class cls, ClassLoaderResolver clr, boolean populate)
    {
        if (annotationManager != null)
        {
            // Check for annotations (use dummy file/package so we have a place for it)
            FileMetaData filemd = new FileMetaData(null, this, null, null);
            filemd.setType(FileMetaData.ANNOTATIONS);
            PackageMetaData pmd = new PackageMetaData(filemd, ClassUtils.getPackageNameForClass(cls), null, null);
            filemd.addPackage(pmd);
            ClassMetaData cmd = annotationManager.getMetaDataForClass(cls, pmd, clr);
            if (cmd != null)
            {
                // register before populating to avoid recursive loops when loading referenced classes
                registerFile("annotations:" + cls.getName(), filemd);

                if (populate)
                {
                    // Populate all classes in this file we've just parsed (i.e only 1!)
                    populateFileMetaData(filemd, clr, cls.getClassLoader());
                }
                return filemd;
            }
        }
        return null;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Utility to return all ClassMetaData that is referenced from the supplier
     * class.
     * @param cmd The origin class's MetaData.
     * @param dba_vendor_id The Vendor id of the database adapter in use.
     *                      (Used in handling "views" support)
     * @param clr ClassLoaderResolver resolver for loading any classes.
     * @return List of ClassMetaData referenced by the origin
     */
    public List getReferencedClassMetaData(AbstractClassMetaData cmd,
                                           String dba_vendor_id,
                                           ClassLoaderResolver clr)
    {
        if (cmd == null)
        {
            return null;
        }

        List orderedCMDs=new ArrayList();
        Set referencedCMDs=new HashSet();

        // Use the ClassMetaData to tell us about its classes
        cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);

        return orderedCMDs;
    }
    
    /**
     * Register to persistent class load
     */
    private class MetaDataRegisterClassListener implements RegisterClassListener
    {
        public void registerClass(RegisterClassEvent arg0)
        {
            //register the class / interface in metadata
            getMetaDataForClassInternal(arg0.getRegisteredClass(), pmfContext.getClassLoaderResolver(arg0.getRegisteredClass().getClassLoader()));
        }
        
    }
}