/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2004 Andy Jefferson - toString(), MetaData, javadocs
2004 Andy Jefferson - added index, and indexed
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Representation of Order MetaData - the ordering of the elements
 * of a List.
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT order (extension*, column*, index?, extension*)&gt;
 * &lt;!ATTLIST order column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST order indexed (true|false|unique) #IMPLIED&gt;
 * &lt;!ATTLIST order mapped-by CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @version $Revision: 1.9 $
 */
public class OrderMetaData extends ExtendableMetaData implements ColumnMetaDataContainer
{
    /**
     * the column name
     */
    final String column;

    /**
     * the columns elements 
     */
    final List columns = new ArrayList();

    /**
     * IndexMetaData.
     */
    protected IndexMetaData indexMetaData;

    /**
     * The indexing value
     */
    protected IndexedValue indexed=null;

    /** Name of the field in the element that is the ordering field. */
    protected final String mappedBy;

    /** whether to order ASC or DESC **/
    protected final String ordering;
    
    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor to create a copy of the passed metadata using the provided parent.
     * @param parent The parent
     * @param omd The metadata to copy
     */
    public OrderMetaData(MetaData parent, OrderMetaData omd)
    {
        super(parent);
        this.indexed = omd.indexed;
        this.column = omd.column;
        // TODO Change these to copy rather than reference
        if (omd.indexMetaData != null)
        {
            this.indexMetaData = omd.indexMetaData;
        }
        for (int i=0;i<omd.columns.size();i++)
        {
            addColumn((ColumnMetaData)omd.columns.get(i));
        }
        this.mappedBy = omd.mappedBy;
        this.ordering = omd.ordering;
    }

    /**
     * Constructor. 
     * @param column Name of column
     * @param indexed The indexed value
     * @param mappedBy The field in the element that provides the ordering
     */
    public OrderMetaData(final String column,
                         final String indexed,
                         final String mappedBy)
    {
        super(null);    // Ignore parent
        this.column = column;
        this.indexed = IndexedValue.getIndexedValue(indexed);
        this.mappedBy = mappedBy;
        this.ordering = null;
    }

    /**
     * Constructor. 
     * @param column Name of column
     * @param indexed The indexed value
     * @param mappedBy The field in the element that provides the ordering
     * @param ordering whether it is ASC or DESC (ascending or descending)
     */
    public OrderMetaData(final String column,
                         final String indexed,
                         final String mappedBy,
                         final String ordering)
    {
        super(null);    // Ignore parent
        this.column = column;
        this.indexed = IndexedValue.getIndexedValue(indexed);
        this.mappedBy = mappedBy;
        this.ordering = ordering;
        //TODO validate ordering values
    }    
    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        columnMetaData = new ColumnMetaData[columns.size()];
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
            columnMetaData[0].initialise();
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        setInitialised();
    }

    /**
     * Add a new ColumnMetaData element
     * @param colmd The Column MetaData 
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
        colmd.parent = this;
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    /**
     * Accessor for the field in the element that provides the ordering.
     * @return Field in the value that provides the ordering.
     */
    public String getMappedBy()
    {
        return mappedBy;
    }

    /**
     * Accessor for the ordering.
     * @return DESC, ASC, ascending, descending or null.
     */
    public String getOrdering()
    {
        return mappedBy;
    }

    /**
     * Accessor for the Column MetaData for the columns
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the column name 
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<order");
        if (column != null)
        {
            sb.append(" column=\"" + column + "\"");
        }
        if (indexed != null)
        {
            sb.append(" indexed=\"" + indexed.toString() + "\"");
        }
        if (mappedBy != null)
        {
            sb.append(" mapped-by=\"" + mappedBy + "\"");
        }
        if (ordering != null)
        {
            sb.append(" ordering=\"" + ordering + "\"");
        }
        sb.append(">\n");

        // Add columns
        for (int i=0; i<columns.size(); i++)
        {
            ColumnMetaData c = (ColumnMetaData)columns.get(i);
            sb.append(c.toString(prefix + indent,indent));
        }

        // Add index
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</order>\n");
        return sb.toString();
    }
}