/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import org.jpox.util.StringUtils;

/**
 * Representation of the MetaData of a named Query.
 * <br>
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * <!ELEMENT query (extension*)>
 * <!ATTLIST query name CDATA #IMPLIED>
 * <!ATTLIST query language CDATA #IMPLIED>
 * <!ATTLIST query result-class CDATA #IMPLIED>
 * <!ATTLIST query unique CDATA #IMPLIED>
 * <!ATTLIST query unmodifiable (true|false) 'false'>
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.8 $
 */
public class QueryMetaData extends ExtendableMetaData
{
    /**
     * Query name
     */
    protected final String name;

    /**
     * Language tag value.
     **/
    protected QueryLanguage language;

    /**
     * unmodifiable tag value.
     */
    protected boolean unmodifiable=false;

    /**
     * The single string query
     */
    protected String query;

    /** The result class to use. Only applies to SQL. */
    protected String resultClass = null;

    /** Whether the query returns unique. Onyl applies to SQL. */
    protected boolean unique = false;

    /**
     * Constructor.
     * @param parent the parent of the Query
     * @param name The Query name
     * @param language The language name
     * @param unmodifiable The unmodifiable tag
     * @param resultClass The result class
     * @param unique The unique tag
     */
    public QueryMetaData(final ExtendableMetaData parent,
                         final String name,
                         final String language,
                         final String unmodifiable,
                         final String resultClass,
                         final String unique)
    {
        super(parent);

        this.name = name;
        if (StringUtils.isWhitespace(name))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Query.QueryNameNotSpecifiedError");
        }

        // include-subclasses
        if (unmodifiable != null && unmodifiable.equalsIgnoreCase("true"))
        {
            this.unmodifiable = true;
        }
        else
        {
            this.unmodifiable = false;
        }

        // language
        this.language = QueryLanguage.getQueryLanguage(language);

        this.resultClass = resultClass;

        if (unique != null && unique.equalsIgnoreCase("true"))
        {
            this.unique = true;
        }
    }

    /**
     * Accessor for the class name.
     * @return class name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the language
     * @return language tag value
     */
    public QueryLanguage getLanguage()
    {
        return language;
    }

    /**
     * Accessor for the unmodifiable tag value.
     * @return unmodifiable tag value
     */
    public boolean isUnmodifiable()
    {
        return unmodifiable;
    }

    /**
     * Accessor for the query
     * @return The query
     */
    public String getQuery()
    {
        return query;
    }

    /**
     * Accessor for the result class
     * @return result class
     */
    public String getResultClass()
    {
        return resultClass;
    }

    /**
     * Accessor for the unique tag value.
     * @return unique tag value
     */
    public boolean isUnique()
    {
        return unique;
    }

    /**
     * Mutator for the query
     * @param query The query
     */
    public void setQuery(String query)
    {
        this.query = query;
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<query name=\"" + name + "\"\n");
        sb.append(prefix).append("       language=\"" + language + "\"\n");
        if (unique)
        {
            sb.append(prefix).append("       unique=\"true\"\n");
        }
        if (resultClass != null)
        {
            sb.append(prefix).append("       result-class=\"" + resultClass + "\"\n");
        }
        sb.append(prefix).append("       unmodifiable=\"" + unmodifiable + "\">\n");
        sb.append(prefix).append(query).append("\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix + "</query>\n");
        return sb.toString();
    }
}