/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * MetaData representing a unique constraint.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT unique (extension*, (column|field|property)*, extension*)&gt;
 * &lt;!ATTLIST unique name CDATA #IMPLIED&gt;
 * &lt;!ATTLIST unique table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST unique deferred (true|false) 'false'&gt;
 * </PRE>
 * @since 1.1
 * @version $Revision: 1.11 $
 */
public class UniqueMetaData extends ExtendableMetaData implements ColumnMetaDataContainer
{
    /**
     * the index name
     */
    final String name;

    /**
     * the table name. If not specified, then applies to the table of the containing class.
     */
    final String table;

    /**
     * Whether the unique is initially deferred.
     */
    final boolean deferred;

    /**
     * Contains the metadata for field
     */
    protected AbstractPropertyMetaData[] fieldMetaData;

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    // -------------------------------------------------------------------------
    // Fields below here are used in the metadata parse process where the parser
    // dynamically adds fields/columns as it encounters them in the MetaData files.
    // They are typically cleared at the point of initialise() and not used thereafter.

    /**
     * the fields elements to be included in the index. Can be empty.
     */
    protected List fields = new ArrayList();

    /**
     * the columns elements to be included in the index. Suitable to be empty
     * when this metadata is contained within a field, element, key, value, or
     * join elements
     */
    protected List columns = new ArrayList();

    /**
     * Constructor to create a copy of the passed metadata using the provided parent.
     * @param parent The parent
     * @param umd The metadata to copy
     */
    public UniqueMetaData(MetaData parent, UniqueMetaData umd)
    {
        super(parent);
        this.name = umd.name;
        this.table = umd.table;
        this.deferred = umd.deferred;
        // TODO Change these to copy rather than reference
        for (int i=0;i<umd.fields.size();i++)
        {
            addField((AbstractPropertyMetaData)umd.fields.get(i));
        }
        for (int i=0;i<umd.columns.size();i++)
        {
            addColumn((ColumnMetaData)umd.columns.get(i));
        }
    }

    /**
     * Constructor.
     * @param parent MetaData for parent
     * @param name Name of unique constraint
     * @param table Name of the table
     * @param deferredValue Whether the unique is deferred initially
     */
    public UniqueMetaData(MetaData parent, final String name, final String table, final String deferredValue)
    {
        super(parent);

        this.name = name;
        this.table = table;
        if (deferredValue != null && deferredValue.equalsIgnoreCase("true"))
        {
            this.deferred = true;
        }
        else
        {
            this.deferred = false;
        }
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialise all sub-objects.
     */
    public void initialise()
    {
        if (isInitialised())
        {
            return;
        }

        // Set up the fieldMetaData
        if (fields.size() == 0)
        {
            fieldMetaData = null;
        }
        else
        {
            fieldMetaData = new AbstractPropertyMetaData[fields.size()];
            for (int i=0; i<fieldMetaData.length; i++)
            {
                fieldMetaData[i] = (AbstractPropertyMetaData) fields.get(i);
                fieldMetaData[i].initialise();
            }
        }

        // Set up the columnMetaData
        if (columns.size() == 0)
        {
            columnMetaData = null;
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        // Clear out parsing data
        fields.clear();
        fields = null;
        columns.clear();
        columns = null;

        setInitialised();
    }

    // -------------------------------- Mutators -------------------------------

    /**
     * Add a new Field.
     * @param fmd the FieldMetadata to add
     */
    public void addField(AbstractPropertyMetaData fmd)
    {
        fields.add(fmd);
    }

    /**
     * Add a new ColumnMetaData element
     * @param colmd The ColumnMetaData to add
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final AbstractPropertyMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the name of the unique constraint
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for the table.
     * @return Returns the table..
     */
    public final String getTable()
    {
        return table;
    }

    /**
     * Accessor for whether the unique constraint is deferred.
     * @return Returns whether the unique constraint is deferred
     */
    public final boolean isDeferred()
    {
        return deferred;
    }

    // -------------------------------- Utilities ------------------------------

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<unique");
        if (table != null)
        {
            sb.append(" table=\"" + table + "\"");
        }
        if (deferred)
        {
            sb.append(" deferred=\"true\"");
        }
        sb.append(name != null ? (" name=\"" + name + "\">\n") : ">\n");

        // Add fields
        if (fieldMetaData != null)
        {
            for (int i=0;i<fieldMetaData.length;i++)
            {
                sb.append(fieldMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add columns
        if (columnMetaData != null)
        {
            for (int i=0;i<columnMetaData.length;i++)
            {
                sb.append(columnMetaData[i].toString(prefix + indent,indent));
            }
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</unique>\n");
        return sb.toString();
    }
}