/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2006 Thomas Marti - Added support for configurable plugin file names
    ...
**********************************************************************/
package org.jpox.plugin;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.util.StringUtils;

/**
 * Manages the registry of Extensions and Extension Points outside any
 * OSGI container
 * @version $Revision: 1.7 $
 */
public class NonManagedPluginRegistry implements PluginRegistry
{
    /** ClassLoaderResolver corresponding to the PMF **/
    private final ClassLoaderResolver clr;

    /** directories that are searched for plugin files **/
    // use trailing slash!
    private static final String[] PLUGIN_DIRS = {"/", "/META-INF/plugins/"};

    /** names of additional plugin files have to match this regular expression **/
    private static final String PLUGIN_FILE_REGEX = "plugin(.+)\\.xml$";

    /** filters all accepted plugin file names **/
    private static final FilenameFilter PLUGIN_FILE_FILTER = new FilenameFilter()
    {
        public boolean accept(File dir, String name)
        {
            return name.matches("^" + PLUGIN_FILE_REGEX);
        }
    };

    /** 
     * Character that is used in URLs of jars to separate the 
     * file name from the path of a resource inside the jar.<br/>
     * example: jar:file:foo.jar!/META-INF/manifest.mf
     **/
    private static final char JAR_SEPARATOR = '!';

    /** extension points keyed by Unique Id (plugin.id +"."+ id) **/
    Map extensionPointsByUniqueId = new HashMap();

    /** extension points **/
    ExtensionPoint[] extensionPoints;

    /**
     * Constructor
     * @param clr the ClassLoaderResolver
     */
    public NonManagedPluginRegistry(ClassLoaderResolver clr)
    {
        this.clr = clr;
        extensionPoints = new ExtensionPoint[0];
    }

    /**
     * Acessor for the ExtensionPoint
     * @param id the unique id of the extension point
     * @return null if the ExtensionPoint is not registered
     */
    public ExtensionPoint getExtensionPoint(String id)
    {
        return (ExtensionPoint) extensionPointsByUniqueId.get(id);
    }

    /**
     * Acessor for the currently registed ExtensionPoints
     * @return array of ExtensionPoints
     */
    public ExtensionPoint[] getExtensionPoints()
    {
        return extensionPoints;
    }

    /**
     * Register Extension Points declared in /org/jpox/plugin/plugin.xml
     */
    public void registerExtensionPoints()
    {
        registerExtensions();
    }

    /**
     * Register ExtensionPoints and Extensions declared in plugin files
     */
    public void registerExtensions()
    {
        //use a set to remove any duplicates
        Set set = new HashSet();

        try
        {
            for (int i = 0; i < PLUGIN_DIRS.length; i++)
            {
                // First add all plugin.xml...
                Enumeration paths = clr.getResources(PLUGIN_DIRS[i] + "plugin.xml",null);
                while (paths.hasMoreElements())
                {
                    set.add(paths.nextElement());
                }
                // ...then search for plugins with more complicated 
                // file names in the same directory
                findAdditionalPluginFiles(set, clr.getResources(PLUGIN_DIRS[i],null));
            }
        }
        catch (IOException e)
        { //TODO: Localisation?
            throw new JDOFatalException("Error loading resource", e);
        }

        //parse the files (Extensions are automatically added to ExtensionPoint)
        Iterator it = set.iterator();
        while (it.hasNext())
        {
            URL plugin = (URL) it.next();
            registerExtensionPoints(plugin);
        }

        it = set.iterator();
        while (it.hasNext())
        {
            URL plugin = (URL) it.next();
            registerExtensions(plugin);
        }
    }

    /**
     * Finds additional plugin files in the given paths and adds them to the set.<br/>
     * <strong>NOTE:</strong> Currently only URLs starting with &quot;file:&quot; or 
     * &quot;jar:file:&quot; are supported 
     * @param set all found plugin files will be added to this set
     * @param paths dirs to search for plugin files
     */
    private void findAdditionalPluginFiles(Set set, Enumeration paths)
    {
        while (paths.hasMoreElements())
        {
            URL pathUrl = (URL) paths.nextElement();

            if (pathUrl.getProtocol().equals("jar"))
            {
                String path = pathUrl.getFile();
                String jarPath = path.substring(0, path.indexOf(JAR_SEPARATOR));

                URL jarUrl = null;
                try
                {
                    jarUrl = new URL(jarPath);
                }
                catch (MalformedURLException e)
                {
                    throw new JDOFatalInternalException("Malformed URL for file: " + jarPath, e);
                }

                if (jarUrl.getProtocol().equals("file"))
                {
                    String resourcePath = StringUtils.getDecodedStringFromURLString(path.substring(path.indexOf(JAR_SEPARATOR) + 2));
                    JarFile jarFile;
                    try
                    {
                        jarFile = StringUtils.getJarFileForFilename(jarUrl.getFile());
                    }
                    catch (IOException e)
                    {
                        throw new JDOFatalInternalException("Error creating JAR-File: " + jarUrl.getFile(), e);
                    }

                    for (Enumeration en = jarFile.entries(); en.hasMoreElements();)
                    {
                        JarEntry entry = (JarEntry) en.nextElement();
                        if (entry.getName().matches(resourcePath + PLUGIN_FILE_REGEX))
                        {
                            String pluginFilePath = jarUrl.toString() + JAR_SEPARATOR + entry.getName();
                            try
                            {
                                set.add(new URL(pluginFilePath));
                            }
                            catch (MalformedURLException e)
                            {
                                throw new JDOFatalInternalException("Malformed URL for file: " + pluginFilePath, e);
                            }
                        }
                    }
                }
            }
            else if (pathUrl.getProtocol().equals("file"))
            {
                File[] pluginFiles = StringUtils.getFileForFilename(pathUrl.getPath()).listFiles(PLUGIN_FILE_FILTER);
                for (int i = 0; pluginFiles != null && i < pluginFiles.length; i++)
                {
                    try
                    {
                        set.add(pluginFiles[i].toURL());
                    }
                    catch (MalformedURLException e)
                    {
                        throw new JDOFatalInternalException("Malformed URL for file: " + pluginFiles[i], e);
                    }
                }
            }
        }
    }

    /**
     * Register Extensions and Extension Points declared in plugin
     * @param plugin the plugin file to load extensions
     */
    protected void registerExtensions(URL plugin)
    {
        if (plugin == null)
        {
            return;
        }

        List elements = PluginParser.parseExtensions(this, plugin, clr);
        for (int i = 0; i < elements.size(); i++)
        {
            ExtensionPoint exPoint = (ExtensionPoint) elements.get(i);
            extensionPointsByUniqueId.put(exPoint.getUniqueId(), exPoint);
        }
        extensionPoints = (ExtensionPoint[]) extensionPointsByUniqueId.values().toArray(new ExtensionPoint[extensionPointsByUniqueId.values().size()]);
    }

    /**
     * Register Extensions and Extension Points declared in plugin
     * @param plugin the plugin file to load extensions
     */
    protected void registerExtensionPoints(URL plugin)
    {
        if (plugin == null)
        {
            return;
        }

        List elements = PluginParser.parseExtensionPoints(this, plugin, clr);
        for (int i = 0; i < elements.size(); i++)
        {
            ExtensionPoint exPoint = (ExtensionPoint) elements.get(i);
            extensionPointsByUniqueId.put(exPoint.getUniqueId(), exPoint);
        }
        extensionPoints = (ExtensionPoint[]) extensionPointsByUniqueId.values().toArray(new ExtensionPoint[extensionPointsByUniqueId.values().size()]);
    }
    
    /**
     * Loads a class (do not initialize) from an attribute of {@link ConfigurationElement}
     * @param confElm the configuration element
     * @param name the attribute name
     * @return the Class
     */
    public Object createExecutableExtension(ConfigurationElement confElm, String name, Class[] argsClass, Object[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException
    {
        Class cls = clr.classForName(confElm.getAttribute(name));
        Constructor constructor = cls.getConstructor(argsClass);
        return constructor.newInstance(args);
    }
    
    
    /**
     * Loads a class (do not initialize)
     * @param pluginId the plugin id
     * @param className the class name
     * @return the Class
     * @throws ClassNotFoundException 
     */
    public Class loadClass(String pluginId, String className) throws ClassNotFoundException
    {
        return clr.classForName(className);
    }
    
    /**
     * Converts a URL that uses a user-defined protocol into a URL that uses the file protocol.
     * @param url the url to be converted
     * @return the converted URL
     * @throws IOException
     */
    public URL resolveURLAsFileURL(URL url) throws IOException
    {
        return url;
    }    
}