/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2006 Thomas Marti - Added support for configurable plugin file names
    ...
**********************************************************************/
package org.jpox.plugin;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.exceptions.ClassNotResolvedException;

/**
 * Manages the registry of Extensions and Extension Points. 
 * @version $Revision: 1.26 $
 */
public class PluginManager
{
    /** plugin registry **/
    private PluginRegistry registry;

    /**
     * Constructor
     * @param conf the PMF Configuration
     * @param clr the ClassLoaderResolver
     */
    public PluginManager(PMFConfiguration conf, ClassLoaderResolver clr)
    {
        registry = PluginRegistryFactory.newPluginRegistry(conf, clr);
    }

    /**
     * Acessor for the ExtensionPoint
     * @param id the unique id of the extension point
     * @return null if the ExtensionPoint is not registered
     */
    public ExtensionPoint getExtensionPoint(String id)
    {
        return registry.getExtensionPoint(id);
    }

    /**
     * Acessor for the currently registed ExtensionPoints
     * @return array of ExtensionPoints
     */
    public ExtensionPoint[] getExtensionPoints()
    {
        return registry.getExtensionPoints();
    }

    /**
     * Register Extension Points declared in /org/jpox/plugin/plugin.xml
     */
    public void registerExtensionPoints()
    {
        registry.registerExtensionPoints();
    }

    /**
     * Register ExtensionPoints and Extensions declared in plugin files
     */
    public void registerExtensions()
    {
        registry.registerExtensions();
    }

    /**
     * Convenience accessor for getting the ConfigurationElement for an extension (of an extension point).
     * @param extensionPointName The extension point
     * @param discrimAttrName Attribute on the extension to use as discriminator
     * @param discrimAttrValue Value for discriminator attribute
     * @return The value of the attribute
     */
    public ConfigurationElement getConfigurationElementForExtension(String extensionPointName,
            String discrimAttrName, String discrimAttrValue)
    {
        ExtensionPoint extensionPoint = getExtensionPoint(extensionPointName);
        if (extensionPoint!=null) 
        {
            Extension[] ex = extensionPoint.getExtensions();
            for (int i=0; i<ex.length; i++)
            {
                ConfigurationElement[] confElm = ex[i].getConfigurationElements();
                for (int j=0; j<confElm.length; j++)
                {
                    // Find an extension with this discriminator value
                    if (confElm[j].getAttribute(discrimAttrName).equalsIgnoreCase(discrimAttrValue))
                    {
                        return confElm[j];
                    }
                }
            }
        }
        return null;
    }

    /**
     * Convenience accessor for getting the ConfigurationElement for an extension (of an extension point).
     * @param extensionPointName The extension point
     * @param discrimAttrName1 Attribute on the extension to use as discriminator1
     * @param discrimAttrValue1 Value for discriminator1 attribute
     * @param discrimAttrName2 Attribute on the extension to use as discriminator2
     * @param discrimAttrValue2 Value for discriminator2 attribute
     * @return The value of the attribute
     */
    public ConfigurationElement getConfigurationElementForExtension(String extensionPointName,
            String discrimAttrName1, String discrimAttrValue1,
            String discrimAttrName2, String discrimAttrValue2)
    {
        ExtensionPoint extensionPoint = getExtensionPoint(extensionPointName);
        if (extensionPoint!=null) 
        {
            Extension[] ex = extensionPoint.getExtensions();
            for (int i=0; i<ex.length; i++)
            {
                ConfigurationElement[] confElm = ex[i].getConfigurationElements();
                for (int j=0; j<confElm.length; j++)
                {
                    // Find an extension with this discriminator value
                    if (confElm[j].getAttribute(discrimAttrName1).equalsIgnoreCase(discrimAttrValue1) &&
                        confElm[j].getAttribute(discrimAttrName2).equalsIgnoreCase(discrimAttrValue2))
                    {
                        return confElm[j];
                    }
                }
            }
        }
        return null;
    }

    /**
     * Convenience accessor for getting the value of an attribute for an extension (of an extension point).
     * @param extensionPoint The extension point
     * @param discrimAttrName Attribute on the extension to use as discriminator
     * @param discrimAttrValue Value for discriminator attribute
     * @param attributeName Name of the attribute whose value we want
     * @return The value of the attribute
     */
    public String getAttributeValueForExtension(String extensionPoint,
            String discrimAttrName, String discrimAttrValue, String attributeName)
    {
        ConfigurationElement elem = getConfigurationElementForExtension(extensionPoint, discrimAttrName, discrimAttrValue);
        if (elem != null)
        {
            return elem.getAttribute(attributeName);
        }
        return null;
    }

    /**
     * Convenience accessor for getting the value of an attribute for an extension (of an extension point).
     * @param extensionPoint The extension point
     * @param discrimAttrName1 Attribute on the extension to use as discriminator1
     * @param discrimAttrValue1 Value for discriminator1 attribute
     * @param discrimAttrName2 Attribute on the extension to use as discriminator2
     * @param discrimAttrValue2 Value for discriminator2 attribute
     * @param attributeName Name of the attribute whose value we want
     * @return The value of the attribute
     */
    public String getAttributeValueForExtension(String extensionPoint,
            String discrimAttrName1, String discrimAttrValue1, String discrimAttrName2, String discrimAttrValue2, String attributeName)
    {
        ConfigurationElement elem = getConfigurationElementForExtension(extensionPoint, 
            discrimAttrName1, discrimAttrValue1, discrimAttrName2, discrimAttrValue2);
        if (elem != null)
        {
            return elem.getAttribute(attributeName);
        }
        return null;
    }
    
    /**
     * Convenience accessor for getting the Class of an attribute for an extension (of an extension point).
     * @param extensionPoint The extension point
     * @param discrimAttrName Attribute on the extension to use as discriminator
     * @param discrimAttrValue Value for discriminator attribute
     * @param attributeName Name of the attribute whose value we want
     * @return The value of the attribute
     */
    public Object createExecutableExtension(String extensionPoint,
                                            String discrimAttrName, 
                                            String discrimAttrValue, 
                                            String attributeName, 
                                            Class[] argsClass, 
                                            Object[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException
    {
        ConfigurationElement elem = getConfigurationElementForExtension(extensionPoint, discrimAttrName, discrimAttrValue);
        if (elem != null)
        {
            return registry.createExecutableExtension(elem, attributeName, argsClass, args);
        }
        return null;
    }
    
    /**
     * Loads a class (do not initialize)
     * @param pluginId the plugin id
     * @param className the class name
     * @return the Class
     * @throws ClassNotResolvedException 
     */
    public Class loadClass(String pluginId, String className) throws ClassNotResolvedException
    {
        try
        {
            return registry.loadClass(pluginId, className);
        }
        catch(ClassNotFoundException ex)
        {
            throw new ClassNotResolvedException(ex.getMessage(),ex);
        }
    }
    
    /**
     * Converts a URL that uses a user-defined protocol into a URL that uses the file protocol.
     * @param url the url to be converted
     * @return the converted URL
     * @throws IOException
     */
    public URL resolveURLAsFileURL(URL url) throws IOException
    {
        return registry.resolveURLAsFileURL(url);
    }    
}