/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.plugin;

import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.jpox.ClassLoaderResolver;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;

/**
 * Parses a plug-in file and load elements into the PluginManager 
 */
class PluginParser
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /**
     * Method to parse ExtensionPoints from plug-in file
     * @param mgr the PluginManager
     * @param fileUrl URL of the plugin.xml file
     * @param clr the ClassLoaderResolver
     * @return a List of extensionPoints, if any
     */
    public static List parseExtensionPoints(PluginRegistry mgr, URL fileUrl, ClassLoaderResolver clr)
    {
        if (JPOXLogger.PLUGIN.isDebugEnabled())
        {
            JPOXLogger.PLUGIN.debug(LOCALISER.msg("Plugin.LoadingExtensionPoints", fileUrl.toString()));
        }
        List extensionPoints = new ArrayList();
        try
        {
            final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = factory.newDocumentBuilder();
            // TODO Enable an EntityResolver here
//          db.setEntityResolver(new PluginEntityResolver());
            try
            {
                Element rootElement = db.parse(new InputSource(new InputStreamReader(fileUrl.openStream()))).getDocumentElement();

                Plugin plugin = new Plugin(rootElement.getAttribute("id"), rootElement.getAttribute("name"), 
                    rootElement.getAttribute("provider-name"), rootElement.getAttribute("version"), fileUrl);
                NodeList elements = rootElement.getElementsByTagName("extension-point");
                for (int i = 0; i < elements.getLength(); i++)
                {
                    Element element = (Element) elements.item(i);
                    String id = element.getAttribute("id").trim();
                    String name = element.getAttribute("name");
                    String schema = element.getAttribute("schema");
                    extensionPoints.add(new ExtensionPoint(id, name, clr.getResource(schema,null), plugin));
                }
            }
            catch (JDOException ex)
            {
                throw ex;
            }
            catch (Exception e)
            {   
                JPOXLogger.PLUGIN.error(LOCALISER.msg("Plugin.FileNonExistent", fileUrl.getFile()));
            }
        }
        catch (ParserConfigurationException e1)
        {
            JPOXLogger.PLUGIN.error(LOCALISER.msg("Plugin.ErrorReadingFile", fileUrl.getFile(), e1.getMessage()));
        }
        return extensionPoints;
    }    

    /**
     * Method to parse Extensions in plug-in file
     * @param mgr the PluginManager
     * @param fileUrl URL of the plugin.xml file
     * @param clr the ClassLoaderResolver
     * @return a List of extensionPoints, if any
     */
    public static List parseExtensions(PluginRegistry mgr, URL fileUrl, ClassLoaderResolver clr)
    {
        if (JPOXLogger.PLUGIN.isDebugEnabled())
        {
            JPOXLogger.PLUGIN.debug(LOCALISER.msg("Plugin.LoadingExtensions", fileUrl.toString()));
        }
        List extensionPoints = new ArrayList();
        try
        {
            final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            final DocumentBuilder db = factory.newDocumentBuilder();
            try
            {
                Element rootElement = db.parse(new InputSource(new InputStreamReader(fileUrl.openStream()))).getDocumentElement();

                Plugin plugin = new Plugin(rootElement.getAttribute("id"), rootElement.getAttribute("name"), 
                    rootElement.getAttribute("provider-name"), rootElement.getAttribute("version"), fileUrl);
                NodeList elements = rootElement.getElementsByTagName("extension");
                for (int i = 0; i < elements.getLength(); i++)
                {
                    Element element = (Element) elements.item(i);
                    String point = element.getAttribute("point").trim();
                    ExtensionPoint exPoint = getExtensionPoint(mgr, extensionPoints, point);
                    if (exPoint == null)
                    {
                        JPOXLogger.PLUGIN.warn(LOCALISER.msg("Plugin.ExtensionPointNotRegistered",point,plugin.getId(),fileUrl.toString()));
                        //we just continue processing
                    }
                    else
                    {
                        Extension ex = new Extension(exPoint, plugin);
                        exPoint.addExtension(ex);
                        NodeList elms = element.getChildNodes();
                        for (int e = 0; e < elms.getLength(); e++)
                        {
                            if (elms.item(e) instanceof Element)
                            {
                                ex.addConfigurationElement(parseConfigurationElement(ex,(Element) elms.item(e), null));
                            }
                        }
                    }
                }
            }
            catch (JDOException ex)
            {
                throw ex;
            }
            catch (Exception e)
            {   
                JPOXLogger.PLUGIN.error(LOCALISER.msg("Plugin.FileNonExistent", fileUrl.getFile()));
            }
        }
        catch (ParserConfigurationException e1)
        {
            JPOXLogger.PLUGIN.error(LOCALISER.msg("Plugin.ErrorReadingFile", fileUrl.getFile(), e1.getMessage()));
        }
        return extensionPoints;
    }

    /**
     * Parses the current element and children, creating a ConfigurationElement object
     * @param ex the {@link Extension}
     * @param element the current element
     * @param parent the parent. null if the parent is Extension
     * @return the ConfigurationElement for the element
     */
    private static ConfigurationElement parseConfigurationElement(Extension ex, Element element, ConfigurationElement parent)
    {
        ConfigurationElement confElm = new ConfigurationElement(ex, element.getNodeName(), parent);
        NamedNodeMap attributes = element.getAttributes();
        for (int i = 0; i < attributes.getLength(); i++)
        {
            Node attribute = attributes.item(i);
            confElm.putAttribute(attribute.getNodeName(), attribute.getNodeValue());
        }
        NodeList elements = element.getChildNodes();
        for (int i = 0; i < elements.getLength(); i++)
        {
            if (elements.item(i) instanceof Element)
            {
                Element elm = (Element) elements.item(i);
                ConfigurationElement child = parseConfigurationElement(ex, elm, confElm);
                confElm.addConfigurationElement(child);
            }
            else if(elements.item(i) instanceof Text)
            {
                confElm.setText(elements.item(i).getNodeValue());
            }
        }
        return confElm;
    }

    /**
     * Obtain an extension point from the PluginManager or from a List of
     * ExtensionPoints
     * @param mgr the PluginManager
     * @param fileUrl URL of the plugin.xml file
     * @param id the searched extension point id. unique id
     */
    private static ExtensionPoint getExtensionPoint(PluginRegistry mgr, List elements, String id)
    {
        ExtensionPoint point = mgr.getExtensionPoint(id);
        if (point == null)
        {
            for (int i = 0; i < elements.size(); i++)
            {
                ExtensionPoint elm = (ExtensionPoint) elements.get(i);
                if (elm.getUniqueId().equals(id))
                {
                    return elm;
                }
            }
        }
        return point;
    }
}