/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;

import javax.jdo.JDOHelper;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.state.FetchPlanState;

/**
 * A mutable second-class BitSet object.
 * 
 * @version $Revision: 1.14 $
 */
public class BitSet extends java.util.BitSet implements SCO
{
    private transient PersistenceCapable owner;
    private transient String fieldName;
    
    /**
     * Creates a <tt>BitSet</tt> object. Assigns owning object and field name.
     *
     * @param ownerSM the owning object
     * @param fieldName the owning field name
     */
    public BitSet(StateManager ownerSM, String fieldName)
    {
        super();

        this.owner = ownerSM.getObject();
        this.fieldName = fieldName;
    }

    // ------------------------- SCO Methods -----------------------------

    /**
     * Mutator for the value, using an object.
     * @param o The Object
     * @param forUpdate Whether to update the object in the datastore with this value
     **/
    public void setValueFrom(Object o, boolean forUpdate)
    {
        for( int i=0; i<length(); i++)
        {
            super.clear(i);
        }
        super.or((java.util.BitSet)o);
    }

    /**
     * Utility to unset the owner.
     **/
    public void unsetOwner()
    {
        owner = null;
        fieldName = null;
    }

    /**
     * Accessor for the owner.
     * @return The owner 
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Accessor for the field name
     * @return The field name
     **/ 
    public String getFieldName()
    {
        return this.fieldName;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        if (owner != null)
        {
            JDOHelper.makeDirty(owner, fieldName);
        }
    }

    /**
     * Method to detach a copy.
     * @param state State for detachment process
     * @return A copy of the object
     */
    public Object detachCopy(FetchPlanState state)
    {
        java.util.BitSet detached = new java.util.BitSet();
        detached.or(this);
        return detached;
    }

    /**
     * Method to attached the passed value.
     * @param value The new value
     */
    public void attachCopy(Object value)
    {
        setValueFrom(value, true);
        makeDirty();
    }

    /**
     * Creates and returns a copy of this object.
     *
     * <p>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return A clone of the object
     */
    public Object clone()
    {
        Object obj = super.clone();

        ((BitSet)obj).unsetOwner();

        return obj;
    }
    
    /**
     * The writeReplace method is called when ObjectOutputStream is preparing to write the object to the stream. The
     * ObjectOutputStream checks whether the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its replacement in the stream. The object returned
     * should be either of the same type as the object passed in or an object that when read and resolved will result in
     * an object of a type that is compatible with all references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        java.util.BitSet copy = new java.util.BitSet();
        copy.and(this);
        return copy;
    }         
    
    //  ------------------------- BitSet Methods -----------------------------

    /* (non-Javadoc)
     * @see java.util.BitSet#and(java.util.BitSet)
     */
    public void and(java.util.BitSet set)
    {
        super.and(set);
        makeDirty();
    }

    /* (non-Javadoc)
     * @see java.util.BitSet#andNot(java.util.BitSet)
     */
    public void andNot(java.util.BitSet set)
    {
        super.andNot(set);
        makeDirty();
    }

    /* (non-Javadoc)
     * @see java.util.BitSet#clear(int)
     */
    public void clear(int bitIndex)
    {
        super.clear(bitIndex);
        makeDirty();
    }



    /* (non-Javadoc)
     * @see java.util.BitSet#or(java.util.BitSet)
     */
    public void or(java.util.BitSet set)
    {
        super.or(set);
        makeDirty();
    }

    /* (non-Javadoc)
     * @see java.util.BitSet#set(int)
     */
    public void set(int bitIndex)
    {
        super.set(bitIndex);
        makeDirty();
    }

    /* (non-Javadoc)
     * @see java.util.BitSet#xor(java.util.BitSet)
     */
    public void xor(java.util.BitSet set)
    {
        super.xor(set);
        makeDirty();
    }


    
}
