/**********************************************************************
Copyright (c) 2002 Mike Martin and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Erik Bengtson  - removed ununsed imports
2003 Andy Jefferson - added missing methods
2003 Andy Jefferson - added Queryable implementation
2004 Andy Jefferson - rewritten to always have delegate populated
2004 Andy Jefferson - changed to allow caching
2005 Andy Jefferson - allowed for serialised map
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.AbstractMap;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.spi.Detachable;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.IncompatibleKeyTypeException;
import org.jpox.sco.exceptions.IncompatibleValueTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.state.FetchPlanState;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MapMapping;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.scostore.MapStore;

/**
 * A mutable second-class Map object. 
 * Uses a "delegate" as a local store for the Collection. Uses a "backing store"
 * (SetStore) to represent the datastore. The "delegate" is updated with the
 * "backing store" information at necessary intervals. 
 *
 * <H3>Modes of Operation</H3>
 * The user can operate the collection in 2 modes.
 * The <B>cached</B> mode will use an internal cache of the elements (in the
 * "delegate") reading them at the first opportunity and then using the cache
 * thereafter.
 * The <B>non-cached</B> mode will just go direct to the "backing store" each
 * call.
 *
 * <H3>Mutators</H3>
 * When the backing store is present any updates are passed direct to the
 * datastore as well as to the "delegate". If the "backing store" isn't present
 * the changes are made to the "delegate" only.
 *
 * <H3>Accessors</H3>
 * When any accessor method is invoked, it typically checks whether the
 * collection has been loaded from its backing store and does this as
 * necessary. Some methods (<B>size()</B>, <B>containsKey()</B>) just check if
 * everything is loaded and use the delegate if possible, otherwise going
 * direct to the datastore.
 *
 * @version $Revision: 1.52 $
 */
public class Map extends AbstractMap
    implements java.util.Map, SCOMap, Cloneable, Queryable, java.io.Serializable
{
    private Object owner;
    private StateManager ownerSM;
    private String fieldName;
    private int fieldNumber;
    private Class keyType;
    private Class valueType;
    private boolean allowNulls;

    /** The "backing store" */
    protected MapStore backingStore;

    /** The internal "delegate". */
    protected java.util.Map delegate;

    /** Whether to use "delegate" caching. */
    protected boolean useCache=true;

    /** Status flag whether the collection is loaded into the cache. */
    protected boolean isCacheLoaded=false;

    /**
     * Constructor, using the StateManager of the "owner" and the field name.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    public Map(StateManager ownerSM, String fieldName)
    {
        init(ownerSM, fieldName);
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    private void init(StateManager ownerSM, 
                      String fieldName)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.allowNulls = true;
        
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(c, ownerSM.getPersistenceManager().getClassLoaderResolver());
        boolean serialised = SCOUtils.mapHasSerialisedKeysAndValues(cmd.getField(fieldName));
        this.fieldNumber = cmd.getFieldNumberAbsolute(fieldName);

        if (!serialised && cmd.getManagedFieldAbsolute(fieldNumber).getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
        {
            DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName(), ownerSM.getPersistenceManager().getClassLoaderResolver());
            JavaTypeMapping m = ownerTable.getFieldMapping(ownerSM.getClassMetaData().getField(fieldName));
            if (!(m instanceof MapMapping))
            {
                throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.Map.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c, ownerSM.getPersistenceManager().getClassLoaderResolver(),fieldName).getTypeName());
            }

            this.backingStore = ((MapMapping)m).getBackingStore(ownerSM.getPersistenceManager().getClassLoaderResolver());
            this.keyType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(backingStore.getKeyType());
            this.valueType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(backingStore.getValueType());
        }

        if (keyType != null && !this.keyType.isAssignableFrom(keyType))
        {
            throw new IncompatibleKeyTypeException(ownerSM, fieldName, this.keyType.getName(), keyType.getName());
        }
        if (valueType != null && !this.valueType.isAssignableFrom(valueType))
        {
            throw new IncompatibleValueTypeException(ownerSM, fieldName, this.valueType.getName(), valueType.getName());
        }

        // Set up our "delegate"
        this.delegate = new java.util.HashMap();

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName, cmd); 
        if (useCache && !SCOUtils.useCachedLazyLoading(ownerSM, fieldName, cmd))
        {
            // Load up the collection now if not using lazy loading
            loadFromStore();
        }
    }

    // ------------------------ Implementation of SCO methods ------------------

    /**
     * Method to effect the load of the data in the SCO.
     * Used when the SCO supports lazy-loading to tell it to load all now.
     */
    public void load()
    {
        if (useCache)
        {
            loadFromStore();
        }
    }

    /**
     * Method to update an embedded key in this map.
     * @param key The key
     * @param fieldNumber Number of field in the key
     * @param newValue New value for this field
     */
    public void updateEmbeddedKey(Object key, int fieldNumber, Object newValue)
    {
        if (backingStore != null)
        {
            backingStore.updateEmbeddedKey(ownerSM, key, fieldNumber, newValue);
        }
    }

    /**
     * Method to update an embedded value in this map.
     * @param value The value
     * @param fieldNumber Number of field in the value
     * @param newValue New value for this field
     */
    public void updateEmbeddedValue(Object value, int fieldNumber, Object newValue)
    {
        if (backingStore != null)
        {
            backingStore.updateEmbeddedValue(ownerSM, value, fieldNumber, newValue);
        }
    }

    /**
     * Mutator for setting using an object.
     * @param o  The object to set from
     * @param forUpdate Whether to update the datastore with this value
     */
    public synchronized void setValueFrom(Object o, boolean forUpdate)
    {
        java.util.Map m = (java.util.Map)o;
        if (m != null)
        {
            // Check for the case of serialised maps, and assign StateManagers to any PC keys/values without
            AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
            if (SCOUtils.mapHasSerialisedKeysAndValues(fmd) &&
                (fmd.getMap().getKeyClassMetaData() != null || fmd.getMap().getValueClassMetaData() != null))
            {
                PersistenceManager pm = ownerSM.getPersistenceManager();
                Iterator iter = m.entrySet().iterator();
                while (iter.hasNext())
                {
                    Map.Entry entry = (Map.Entry)iter.next();
                    Object key = entry.getKey();
                    Object value = entry.getValue();
                    if (fmd.getMap().getKeyClassMetaData() != null)
                    {
                        PersistenceCapable pcKey = (PersistenceCapable)key;
                        StateManager objSM = pm.findStateManager(pcKey);
                        if (objSM == null)
                        {
                            objSM = StateManagerFactory.newStateManager(pcKey, pm, false);
                            objSM.addEmbeddedOwner(ownerSM, fieldNumber);
                        }
                    }
                    if (fmd.getMap().getValueClassMetaData() != null)
                    {
                        PersistenceCapable pcValue = (PersistenceCapable)value;
                        StateManager objSM = pm.findStateManager(pcValue);
                        if (objSM == null)
                        {
                            objSM = StateManagerFactory.newStateManager(pcValue, pm, false);
                            objSM.addEmbeddedOwner(ownerSM, fieldNumber);
                        }
                    }
                }
            }
        }

        if (forUpdate)
        {
            clear();
            putAll(m);
        }
        else
        {
            delegate.clear();
            delegate.putAll(m);
        }
    }

    /**
     * Accessor for the field name.
     * @return The field name.
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner.
     * @return The owner.
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            backingStore = null;
        }
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    /**
     * Method to run reachability on this SCO.
     * @param reachables List of StateManagers reachable so far
     */
    public void runReachability(java.util.Set reachables)
    {
        for (Iterator it = entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (key instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)key).runReachability(reachables);
            }
            if (val instanceof PersistenceCapable)
            {
                ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)val).runReachability(reachables);
            }
        }
    }

    /**
     * Method to make transient this map and its contents.
     * @param state State for the fetchplan process
     */
    public void makeTransient(FetchPlanState state)
    {
        for (Iterator it = entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (key instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().internalMakeTransient(key, state);
            }
            if (val instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().internalMakeTransient(val, state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to detach this map and its contents.
     * @param state State for the detachment process
     */
    public void detach(FetchPlanState state)
    {
        for (Iterator it = entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (key instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().detachInternal(key, state);
            }
            if (val instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().detachInternal(val, state);
            }
        }
        ownerSM = null;
        backingStore = null;
    }

    /**
     * Method to load all fields in the fetch plan (for subobjects).
     * @param state State for the fetchplan process
     */
    public void loadFieldsInFetchPlan(FetchPlanState state)
    {
        for (Iterator it = entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (key instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)key).loadFieldsInFetchPlan(state);
            }
            if (val instanceof PersistenceCapable)
            {
               ownerSM.getPersistenceManager().findStateManager((PersistenceCapable)val).loadFieldsInFetchPlan(state);
            }
        }
    }

    /**
     * Method to return a detached copy of the container.
     * Recurse sthrough the keys/values so that they are likewise detached.
     * @param state State for detachment process
     * @return The detached container
     */
    public Object detachCopy(FetchPlanState state)
    {
        java.util.Map detached = new java.util.HashMap();
        for (Iterator it = entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable)
            {
                val = ownerSM.getPersistenceManager().detachCopyInternal(val, state);
            }
            if (key instanceof PersistenceCapable)
            {
                key = ownerSM.getPersistenceManager().detachCopyInternal(key, state);
            }
            detached.put(key, val);
        }
        return detached;
    }

    /**
     * Method to return an attached copy of the passed (detached) value. The returned attached copy
     * is a SCO wrapper. Goes through the existing keys/values in the store for this owner field and
     * removes ones no longer present, and adds new keys/values. All keys/values in the (detached)
     * value are attached.
     * @param value The new (collection) value
     */
    public void attachCopy(Object value)
    {
        java.util.Map m = (java.util.Map) value;

        // Attach all of the keys/values in the new map
        AbstractPropertyMetaData fmd = ownerSM.getClassMetaData().getField(fieldName);
        boolean keysWithoutIdentity = SCOUtils.mapHasKeysWithoutIdentity(fmd);
        boolean valuesWithoutIdentity = SCOUtils.mapHasValuesWithoutIdentity(fmd);

        java.util.Map attachedKeysValues = new java.util.HashMap(m.size());
        Iterator iter = m.entrySet().iterator();
        while (iter.hasNext())
        {
            Map.Entry entry = (Map.Entry) iter.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable && val instanceof Detachable)
            {
                val = ownerSM.getPersistenceManager().attachCopy(val, valuesWithoutIdentity);
            }
            if (key instanceof PersistenceCapable && key instanceof Detachable)
            {
                key = ownerSM.getPersistenceManager().attachCopy(key, keysWithoutIdentity);
            }
            attachedKeysValues.put(key, val);
        }

        // Update the attached collection with the detached elements
        SCOUtils.updateMapWithMapKeysValues(this, attachedKeysValues);
    }

    // ------------------------- Queryable Methods -----------------------------

    /**
     * Method to return a new Query statement for the Map.
     * @return The Query Statement.
     **/
    public synchronized QueryExpression newQueryStatement()
    {
        return newQueryStatement(valueType);
    }

    /**
     * Method to return a new Query statement for the Map given the candidate
     * class.
     * @param candidate_class the value class.
     * @return The Query Statement
     **/
    public synchronized QueryExpression newQueryStatement(Class candidate_class)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a new result object factory for processing of Query
     * statements.
     * @param stmt The Query Statement.
     * @param ignoreCache Whether to ignore the cache.
     * @param resultClass Whether to create objects of a particular class
     * @param useFetchPlan whether to use the fetch plan to retrieve fields in the same query
     * @return The result object factory.
     **/
    public synchronized ResultObjectFactory newResultObjectFactory(
                                        QueryExpression stmt,
                                        boolean ignoreCache,
                                        Class resultClass,
                                        boolean useFetchPlan)
    {
        if (backingStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return backingStore.newResultObjectFactory(ownerSM, stmt,ignoreCache,useFetchPlan);
    }

    // -------------------- Implementation of Map Methods ----------------------
 
    /**
     * Creates and returns a copy of this object.
     *
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return Clone of the object
     */
    public Object clone()
    {
        if (useCache)
        {
            loadFromStore();
        }

        return ((java.util.HashMap)delegate).clone();
    }

    /**
     * Utility to check if a key is contained in the Map.
     * @param key The key to check
     * @return Whether it is contained
     **/
    public synchronized boolean containsKey(Object key)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsKey(key);
        }
        else if (backingStore != null)
        {
            return backingStore.containsKey(ownerSM, key);
        }

        return delegate.containsKey(key);
    }

    /**
     * Utility to check if a value is contained in the Map.
     * @param value The value to check
     * @return Whether it is contained
     **/
    public synchronized boolean containsValue(Object value)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsValue(value);
        }
        else if (backingStore != null)
        {
            return backingStore.containsValue(ownerSM, value);
        }

        return delegate.containsValue(value);
    }

    /**
     * Accessor for the set of entries in the Map.
     * @return Set of entries
     **/
    public synchronized java.util.Set entrySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return new Set(ownerSM, fieldName, false, backingStore.entrySetStore());
        }

        return delegate.entrySet();
    }

    /**
     * Method to check the equality of this map, and another.
     * @param o The map to compare against.
     * @return Whether they are equal.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.Map))
        {
            return false;
        }
        java.util.Map m = (java.util.Map)o;

        return entrySet().equals(m.entrySet());
    }

    /**
     * Accessor for the value stored against a key.
     * @param key The key
     * @return The value.
     **/
    public synchronized Object get(Object key)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            return backingStore.get(ownerSM, key);
        }

        return delegate.get(key);
    }

    /**
     * Method to generate a hashcode for this Map.
     * @return The hashcode.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (backingStore != null)
        {
            int h = 0;
            Iterator i = entrySet().iterator();
            while (i.hasNext())
            {
                h += i.next().hashCode();
            }

            return h;
        }
        return delegate.hashCode();
    }

    /**
     * Method to return if the Map is empty.
     * @return Whether it is empty.
     **/
    public synchronized boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Accessor for the set of keys in the Map.
     * @return Set of keys.
     **/
    public synchronized java.util.Set keySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        if (backingStore != null)
        {
            return new Set(ownerSM, fieldName, false, backingStore.keySetStore(ownerSM.getPersistenceManager().getClassLoaderResolver()));
        }

        return delegate.keySet();
    }

    /**
     * Method to return the size of the Map.
     * @return The size
     **/
    public synchronized int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (backingStore != null)
        {
            return backingStore.entrySetStore().size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Accessor for the set of values in the Map.
     * @return Set of values.
     **/
    public synchronized Collection values()
    {
        if (useCache)
        {
            loadFromStore();
        }
        if (backingStore != null)
        {
            return new Set(ownerSM, fieldName, true, backingStore.valueSetStore(ownerSM.getPersistenceManager().getClassLoaderResolver()));
        }

        return delegate.values();
    }

    /**
     * Method to return a string form of this Map.
     * @return String form of this Map.
     **/
    public String toString()
    {
        StringBuffer s = new StringBuffer("{");

        Iterator i = entrySet().iterator();
        boolean hasNext = i.hasNext();

        while (hasNext)
        {
            Entry e = (Entry) i.next();
            Object key = e.getKey();
            Object val = e.getValue();
            s.append(key).append('=').append(val);
            hasNext = i.hasNext();

            if (hasNext)
            {
                s.append(',');
            }
        }

        s.append("}");

        return s.toString();
    }

    // -------------------------- Mutator methods ------------------------------
 
    /**
     * Method to clear the Map.
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (backingStore != null)
        {
            backingStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to add a value to the Map.
     * @param key The key for the value.
     * @param value The value
     * @return The previous value against this key (if any).
     **/
    public synchronized Object put(Object key, Object value)
    {
        // Reject inappropriate elements
        if (value == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        // Dont load the entries since we dont need them here

        Object oldValue = null;
        if (backingStore != null)
        {
            oldValue = backingStore.put(ownerSM, key, value);
        }
        Object delegateOldValue = delegate.put(key, value);
        if (backingStore == null)
        {
            oldValue = delegateOldValue;
        }
        return oldValue;
    }

    /**
     * Method to add a Map of values to this map.
     * @param m The Map to add
     **/
    public synchronized void putAll(java.util.Map m)
    {
        makeDirty();

        // Dont load the entries since we dont need them here

        if (backingStore != null)
        {
            backingStore.putAll(ownerSM, m);
        }
        delegate.putAll(m);
    }

    /**
     * Method to remove a value from the Map.
     * @param key The key for the value.
     * @return The value removed.
     **/
    public synchronized Object remove(Object key)
    {
        makeDirty();

        // Dont load the entries since we dont need them here

        Object removed = null;
        Object delegateRemoved = delegate.remove(key);
        if (backingStore != null)
        {
            removed = backingStore.remove(ownerSM, key);
        }
        else
        {
            removed = delegateRemoved;
        }
        return removed;
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        if (useCache)
        {
            loadFromStore();
            return new java.util.HashMap(delegate);
        }
        else
        {
            // TODO Cater for non-cached map, load elements in a DB call.
            return new java.util.HashMap(delegate);
        }
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (backingStore != null && !isCacheLoaded)
        {
            delegate.clear();

            // Populate the delegate with the keys/values from the store
            SCOUtils.populateMapDelegateWithStoreData(delegate, backingStore, ownerSM);

            isCacheLoaded = true;
        }
    }
}