/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDOException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ContainerMetaData;
import org.jpox.store.scostore.CollectionStore;
import org.jpox.store.scostore.MapStore;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Collection of utilities for second class objects.
 * 
 * @version $Revision: 1.34 $
 */
public class SCOUtils
{
    /** Localiser for messages. */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /**
     * Method to create a new SCO wrapper for a SCO type.
     * The SCO wrapper will be appropriate for the passed value (which represents the instantiated type of the field)
     * unless it is null when the wrapper will be appropriate for the declared type of the field.
     * @param declaredType The class of the object
     * @param ownerSM State Manager for the owning object
     * @param fmd The Field MetaData for the related field.
     * @param value The value we are wrapping
     * @param forUpdate Whether the SCO needs updating in the datastore with this value
     * @return The Second-Class Object
     */
    public static SCO newSCOInstance(Class declaredType, Object ownerSM, AbstractPropertyMetaData fmd, Object value, boolean forUpdate)
    {
        if (!fmd.getType().isAssignableFrom(declaredType))
        {
            throw new JDOUserException(LOCALISER.msg("SCO.NewInstance.FieldTypeInvalid", declaredType.getName(), fmd.getName(), fmd.getType()));
        }

        String typeName = declaredType.getName();
        if (value != null)
        {
            // If we have a current value, use the instantiated type to define the wrapper type
            typeName = value.getClass().getName();
        }

        Class newType = fmd.getMetaDataManager().getPMFContext().getTypeManager().getSecondClassWrapper(typeName);
        if (newType == null && fmd.getMetaDataManager().getPMFContext().getTypeManager().isSecondClassWrapper(typeName) && value != null)
        {
            // The passed in value is a wrapper type already, so just return it!
            return (SCO)value;
        }

        if (newType == null)
        {
            throw new JDOUserException(LOCALISER.msg("SCO.NewInstance.ClassInvalid", declaredType.getName()));
        }

        SCO obj = (SCO) ClassUtils.newInstance(newType, 
            new Class[]{org.jpox.StateManager.class, String.class}, 
            new Object[]{ownerSM, fmd.getName()});

        if (value != null)
        {
            // Apply the value to this wrapper if we have it
            obj.setValueFrom(value, forUpdate);
        }

        return obj;
    }

    /**
     * Utility to return whether or not to use the container cache for the
     * collection/map for the passed StateManager SCO.
     * @param ownerSM The StateManager for the SCO field
     * @param fieldName Name of the field.
     * @param cmd MetaData for class
     * @return Whether to use the cache.
     */
    public static boolean useContainerCache(StateManager ownerSM, String fieldName, AbstractClassMetaData cmd)
    {
        if (ownerSM == null)
        {
            return false;
        }

        // Get global value for PMF
        boolean useCache = ownerSM.getPersistenceManager().getPMFContext().getPmfConfiguration().getJdoCacheCollections();

        // Check if this container has a MetaData value defined
        ContainerMetaData conmd=ownerSM.getMetaDataManager().getMetaDataForField(ownerSM.getObject().getClass(), ownerSM.getPersistenceManager().getClassLoaderResolver(),fieldName).getContainer();
        if (conmd != null && conmd.hasExtension("cache"))
        {
            useCache = (new Boolean(conmd.getValueForExtension("cache"))).booleanValue();
        }

        boolean useLazy = true;
        if (useCache)
        {
            useLazy = useCachedLazyLoading(ownerSM, fieldName, cmd);
        }
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            if (useCache)
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("SCO.Collection.Cached", fieldName, ownerSM.getObject().getClass().getName(), "" + useLazy));
            }
            else
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("SCO.Collection.PassThrough", fieldName, ownerSM.getObject().getClass().getName()));
            }
        }

        return useCache;
    }

    /**
     * Accessor for whether the use lazy loading when caching the collection.
     * @param ownerSM StateManager of the owning object
     * @param fieldName Name of the collection/map field
     * @param cmd MetaData for the class
     * @return Whether to use lazy loading when caching the collection
     */
    public static boolean useCachedLazyLoading(StateManager ownerSM, String fieldName, AbstractClassMetaData cmd)
    {
        if (ownerSM == null)
        {
            return false;
        }

        boolean lazy = false;

        AbstractPropertyMetaData fmd = cmd.getField(fieldName);
        if (ownerSM.getPersistenceManager().getPMFContext().getPmfConfiguration().getJdoCacheCollectionsLazy() != null)
        {
            // Global setting for PMF
            lazy = ownerSM.getPersistenceManager().getPMFContext().getPmfConfiguration().getJdoCacheCollectionsLazy().booleanValue();
        }
        else if (fmd.getContainer() != null && fmd.getContainer().hasExtension("cache-lazy-loading"))
        {
            // Check if this container has a MetaData value defined
            lazy = (new Boolean(fmd.getContainer().getValueForExtension("cache-lazy-loading"))).booleanValue();
        }
        else
        {
            // Check if this SCO is in the current FetchPlan
            boolean inFP = false;
            int[] fpFields = ((FetchPlanImpl)ownerSM.getPersistenceManager().getFetchPlan()).getFetchPlanForClass(cmd).getFieldsInActualFetchPlan();
            int fieldNo = fmd.getAbsoluteFieldNumber();
            if (fpFields != null && fpFields.length > 0)
            {
                for (int i=0;i<fpFields.length;i++)
                {
                    if (fpFields[i] == fieldNo)
                    {
                        inFP = true;
                        break;
                    }
                }
            }
            // Default to lazy loading when not in FetchPlan, and non-lazy when in FetchPlan
            lazy = !inFP;
        }

        return lazy;
    }

    /**
     * Convenience method to return if a collection field has elements without their own identity.
     * Checks if the elements are embedded in a join table, or in the main table, or serialised.
     * @param fmd MetaData for the field
     * @return Whether the elements have their own identity or not
     */
    public static boolean collectionHasElementsWithoutIdentity(AbstractPropertyMetaData fmd)
    {
        boolean elementsWithoutIdentity = false;
        if (fmd.isSerialized())
        {
            // Elements serialised into main table
            elementsWithoutIdentity = true;
        }
        else if (fmd.getElementMetaData() != null && fmd.getElementMetaData().getEmbeddedMetaData() != null && fmd.getJoinMetaData() != null)
        {
            // Elements embedded in join table using embedded mapping
            elementsWithoutIdentity = true;
        }
        else if (fmd.getCollection() != null && fmd.getCollection().isEmbeddedElement())
        {
            // Elements are embedded (either serialised, or embedded in join table)
            elementsWithoutIdentity = true;
        }

        return elementsWithoutIdentity;
    }

    /**
     * Convenience method to return if a map field has keys without their own identity.
     * Checks if the keys are embedded in a join table, or in the main table, or serialised.
     * @param fmd MetaData for the field
     * @return Whether the keys have their own identity or not
     */
    public static boolean mapHasKeysWithoutIdentity(AbstractPropertyMetaData fmd)
    {
        boolean keysWithoutIdentity = false;
        if (fmd.isSerialized())
        {
            // Keys (and values) serialised into main table
            keysWithoutIdentity = true;
        }
        else if (fmd.getKeyMetaData() != null && fmd.getKeyMetaData().getEmbeddedMetaData() != null && fmd.getJoinMetaData() != null)
        {
            // Keys embedded in join table using embedded mapping
            keysWithoutIdentity = true;
        }
        else if (fmd.getMap() != null && fmd.getMap().isEmbeddedKey())
        {
            // Keys are embedded (either serialised, or embedded in join table)
            keysWithoutIdentity = true;
        }

        return keysWithoutIdentity;
    }

    /**
     * Convenience method to return if a map field has values without their own identity.
     * Checks if the values are embedded in a join table, or in the main table, or serialised.
     * @param fmd MetaData for the field
     * @return Whether the values have their own identity or not
     */
    public static boolean mapHasValuesWithoutIdentity(AbstractPropertyMetaData fmd)
    {
        boolean valuesWithoutIdentity = false;
        if (fmd.isSerialized())
        {
            // Values (and keys) serialised into main table
            valuesWithoutIdentity = true;
        }
        else if (fmd.getValueMetaData() != null && fmd.getValueMetaData().getEmbeddedMetaData() != null && fmd.getJoinMetaData() != null)
        {
            // Values embedded in join table using embedded mapping
            valuesWithoutIdentity = true;
        }
        else if (fmd.getMap() != null && fmd.getMap().isEmbeddedValue())
        {
            // Values are embedded (either serialised, or embedded in join table)
            valuesWithoutIdentity = true;
        }

        return valuesWithoutIdentity;
    }

    /**
     * Convenience method to return if a collection field has the elements serialised into the
     * table of the field as a single BLOB.
     * @param fmd MetaData for the field
     * @return Whether the elements are serialised (either explicitly or implicitly)
     */
    public static boolean collectionHasSerialisedElements(AbstractPropertyMetaData fmd)
    {
        boolean serialised = fmd.isSerialized();
        if (fmd.getCollection() != null && fmd.getCollection().isEmbeddedElement() && fmd.getJoinMetaData() == null)
        {
            // Elements are embedded but no join table so we serialise
            serialised = true;
        }

        return serialised;
    }

    /**
     * Convenience method to return if an array field has the elements stored into the
     * table of the field as a single (BLOB) column.
     * @param fmd MetaData for the field
     * @return Whether the elements are stored in a single column
     */
    public static boolean arrayIsStoredInSingleColumn(AbstractPropertyMetaData fmd)
    {
        boolean singleColumn = fmd.isSerialized();
        if (!singleColumn && fmd.getArray() != null && fmd.getJoinMetaData() == null)
        {
            if (fmd.getArray().isEmbeddedElement())
            {
                // Elements are embedded but no join table so we store in a single column
                singleColumn = true;
            }

            Class elementClass = fmd.getType().getComponentType();
            if (!elementClass.isInterface() && !PersistenceCapable.class.isAssignableFrom(elementClass))
            {
                // Array of non-PC with no join table  so store in single column of main table
                singleColumn = true;
            }
        }

        return singleColumn;
    }

    /**
     * Convenience method to return if a map field has the keys/values serialised into the
     * table of the field as a single BLOB.
     * @param fmd MetaData for the field
     * @return Whether the keys and values are serialised (either explicitly or implicitly)
     */
    public static boolean mapHasSerialisedKeysAndValues(AbstractPropertyMetaData fmd)
    {
        boolean inverseKeyField = false;
        if (fmd.getKeyMetaData() != null && fmd.getKeyMetaData().getMappedBy() != null)
        {
            inverseKeyField = true;
        }
        boolean inverseValueField = false;
        if (fmd.getValueMetaData() != null && fmd.getValueMetaData().getMappedBy() != null)
        {
            inverseValueField = true;
        }
        boolean serialised = fmd.isSerialized();
        if (fmd.getMap() != null && fmd.getJoinMetaData() == null &&
            (fmd.getMap().isEmbeddedKey() || fmd.getMap().isEmbeddedValue()) &&
            !inverseKeyField && !inverseValueField)
        {
            // Keys/values are embedded but no join table so we serialise the whole map
            // Note that we explicitly excluded the JPOX extension 1-N Map with the key stored in the value
            serialised = true;
        }

        return serialised;
    }

    /**
     * Convenience method to update a collection to contain the elements in another collection.
     * Performs the updates by calling the necessary add(), remove() methods just for the
     * elements that have changed. Allows for some elements in one collection being attached
     * and some being detached (so having same id, but different state)
     * @param coll The collection to update
     * @param newColl The new collection whose elements we need in "coll"
     */
    public static void updateCollectionWithCollection(Collection coll, Collection newColl)
    {
        if (coll == null)
        {
            return;
        }

        if (newColl == null)
        {
            coll.clear();
            return;
        }

        // Remove all elements no longer in the Collection
        Iterator iter = coll.iterator();
        while (iter.hasNext())
        {
            Object element = iter.next();
            if (element instanceof PersistenceCapable)
            {
                Object id = JDOHelper.getObjectId(element);

                if (id != null)
                {
                    // Element has an id so compare the id
                    boolean present = false;
                    Iterator newIter = newColl.iterator();
                    while (newIter.hasNext())
                    {
                        Object newElement = newIter.next();
                        Object newId = JDOHelper.getObjectId(newElement);

                        if (id.equals(newId))
                        {
                            present = true;
                            break;
                        }
                    }

                    if (!present)
                    {
                        iter.remove();
                    }
                }
                else
                {
                    if (!newColl.contains(element))
                    {
                        iter.remove();
                    }
                }
            }
            else
            {
                if (!newColl.contains(element))
                {
                    iter.remove();
                }
            }
        }

        // Add all new elements
        Iterator newIter = newColl.iterator();
        while (newIter.hasNext())
        {
            Object newElement = newIter.next();

            if (newElement instanceof PersistenceCapable)
            {
                Object newId = JDOHelper.getObjectId(newElement);
                if (newId != null)
                {
                    boolean present = false;
                    iter = coll.iterator();
                    while (iter.hasNext())
                    {
                        Object element = iter.next();
                        Object id = JDOHelper.getObjectId(element);
                        if (newId.equals(id))
                        {
                            present = true;
                            break;
                        }
                    }

                    if (!present)
                    {
                        coll.add(newElement);
                    }
                }
                else
                {
                    if (!coll.contains(newElement))
                    {
                        coll.add(newElement);
                    }
                }
            }
            else
            {
                if (!coll.contains(newElement))
                {
                    coll.add(newElement);
                }
            }
        }
    }

    /**
     * Convenience method to update a Store collection to contain the elements in another collection.
     * Performs the updates by calling the necessary add(), remove() methods just for the
     * elements that have changed. Allows for some elements in one collection being attached
     * and some being detached (so having same id, but different state)
     * @param store The store to apply changes to
     * @param ownerSM StateManager of the owner
     * @param newColl The new collection whose elements we need in "coll"
     */
    public static void updateStoreWithCollection(CollectionStore store, StateManager ownerSM, Collection newColl)
    {
        if (store == null || ownerSM == null)
        {
            return;
        }

        if (newColl == null)
        {
            store.clear(ownerSM);
            return;
        }

        // Retrieve the current elements from the store
        Collection coll = new java.util.HashSet();
        Iterator iter = store.iterator(ownerSM);
        while (iter.hasNext())
        {
            coll.add(iter.next());
        }

        // Remove all elements no longer in the store
        iter = coll.iterator();
        while (iter.hasNext())
        {
            Object element = iter.next();
            if (element instanceof PersistenceCapable)
            {
                Object id = JDOHelper.getObjectId(element);

                if (id != null)
                {
                    // Element has an id so compare the id
                    boolean present = false;
                    Iterator newIter = newColl.iterator();
                    while (newIter.hasNext())
                    {
                        Object newElement = newIter.next();
                        Object newId = JDOHelper.getObjectId(newElement);

                        if (id.equals(newId))
                        {
                            present = true;
                            break;
                        }
                    }

                    if (!present)
                    {
                        store.remove(ownerSM, element);
                    }
                }
                else
                {
                    if (!newColl.contains(element))
                    {
                        store.remove(ownerSM, element);
                    }
                }
            }
            else
            {
                if (!newColl.contains(element))
                {
                    store.remove(ownerSM, element);
                }
            }
        }

        // Add all new elements to the store
        Iterator newIter = newColl.iterator();
        while (newIter.hasNext())
        {
            Object newElement = newIter.next();

            if (newElement instanceof PersistenceCapable)
            {
                Object newId = JDOHelper.getObjectId(newElement);
                if (newId != null)
                {
                    boolean present = false;
                    iter = coll.iterator();
                    while (iter.hasNext())
                    {
                        Object element = iter.next();
                        Object id = JDOHelper.getObjectId(element);
                        if (newId.equals(id))
                        {
                            present = true;
                            break;
                        }
                    }

                    if (!present)
                    {
                        store.add(ownerSM, newElement);
                    }
                }
                else
                {
                    if (!coll.contains(newElement))
                    {
                        store.add(ownerSM, newElement);
                    }
                }
            }
            else
            {
                if (!coll.contains(newElement))
                {
                    store.add(ownerSM, newElement);
                }
            }
        }
    }

    /**
     * Convenience method for use by Collection/Set/HashSet attachCopy methods to
     * update the passed (attached) collection using the (attached) elements passed.
     * @param coll The current (attached) collection
     * @param elements The collection of (attached) elements needed.
     * @return If the Collection was updated
     */
    static boolean updateCollectionWithCollectionElements(Collection coll, Collection elements)
    {
        boolean updated = false;

        // Delete any elements that are no longer in the collection
        Iterator attachedIter = coll.iterator();
        while (attachedIter.hasNext())
        {
            Object attachedElement = attachedIter.next();
            if (!elements.contains(attachedElement))
            {
                // No longer present so remove it
                attachedIter.remove();
                updated = true;
            }
        }

        // Add any new elements
        Iterator elementsIter = elements.iterator();
        while (elementsIter.hasNext())
        {
            Object element = elementsIter.next();
            if (!coll.contains(element))
            {
                // Not present so add it
                coll.add(element);
                updated = true;
            }
        }
        return updated;
    }

    /**
     * Convenience method for use by List attachCopy methods to update the
     * passed (attached) list using the (attached) list elements passed.
     * @param list The current (attached) list
     * @param elements The list of (attached) elements needed.
     * @return If the List was updated
     */
    static boolean updateListWithListElements(List list, List elements)
    {
        boolean updated = false;
        // This method needs to take the existing list and generate a list
        // of add/remove/set/clear operations that change the list to the passed
        // elements in as efficient a way as possible. The simplest is
        // clear() then addAll()!, but if there are many objects and very little
        // has changed this would be very inefficient.
        // What we do currently is remove all elements no longer present, and then
        // add any missing elements, correcting the ordering. This can be non-optimal
        // in some situations.
        // TODO Optimise the process

        // Delete any elements that are no longer in the list
        java.util.ArrayList newCopy = new java.util.ArrayList(elements);
        Iterator attachedIter = list.iterator();
        while (attachedIter.hasNext())
        {
            Object attachedElement = attachedIter.next();
            if (!newCopy.remove(attachedElement))
            {
                // No longer present, so remove it
                attachedIter.remove();
                updated = true;
            }
        }

        // Add any new elements that have been added
        java.util.ArrayList oldCopy = new java.util.ArrayList(list);
        Iterator elementsIter = elements.iterator();
        while (elementsIter.hasNext())
        {
            Object element = elementsIter.next();
            if (!oldCopy.remove(element))
            {
                // Now present, so add it
                list.add(element);
                updated = true;
            }
        }

        // Update position of elements in the list to match the new order
        elementsIter = elements.iterator();
        int position = 0;
        while (elementsIter.hasNext())
        {
            Object element = elementsIter.next();
            if (!(list.get(position).equals(element)))
            {
                // Update the position, taking care not to have dependent-field deletes taking place
                ((SCOList)list).set(position, element, false);
                updated = true;
            }

            position++;
        }

        return updated;
    }

    /**
     * Convenience method for use by Map attachCopy methods to update the
     * passed (attached) map using the (attached) map keys/values passed.
     * @param map The current (attached) map
     * @param keysValues The keys/values required
     * @return If the map was updated
     */
    static boolean updateMapWithMapKeysValues(Map map, Map keysValues)
    {
        boolean updated = false;

        // Take a copy of the map so we can call remove() on the map itself
        // TODO Change this to use EntrySet in the future.
        // EntrySet.iterator().remove() doesn't seem to feed through to the DB at the moment
        Map copy = new HashMap(map);

        // Delete any keys that are no longer in the Map
        Iterator attachedIter = copy.entrySet().iterator();
        while (attachedIter.hasNext())
        {
            Map.Entry entry = (Map.Entry) attachedIter.next();
            Object key = entry.getKey();
            if (!keysValues.containsKey(key))
            {
                map.remove(key);
                updated = true;
            }
        }

        // Add any new keys/values and update any changed values
        Iterator keysIter = keysValues.entrySet().iterator();
        while (keysIter.hasNext())
        {
            Map.Entry entry = (Map.Entry) keysIter.next();
            Object key = entry.getKey();
            Object value = entry.getValue();
            if (!map.containsKey(key))
            {
                // Not present so add it
                map.put(key, keysValues.get(key));
                updated = true;
            }
            else
            {
                // Update any values
                if (value instanceof PersistenceCapable)
                {
                    // In case they have changed the PC value for this key
                    // TODO Check if the PC has changed ?
                    map.put(key, value);
                }
                else
                {
                    Object oldValue = map.get(key);
                    if ((oldValue == null && value != null) || (oldValue != null && !oldValue.equals(value)))
                    {
                        map.put(key, value);
                    }
                }
            }
        }

        return updated;
    }

    /**
     * Convenience method to populate the passed delegate Map with the keys/values from
     * the associated Store.
     * <P>
     * The issue here is that we need to load the keys and values in as few calls as possible.
     * The method employed here reads in the keys (if PersistenceCapable), then the values
     * (if PersistenceCapable), and then the "entries" (ids of keys and values) so we can 
     * associate the keys to the values.
     * @param delegate The delegate
     * @param store The Store
     * @param ownerSM State Manager of the owner of the map.
     */
    public static void populateMapDelegateWithStoreData(Map delegate, MapStore store, StateManager ownerSM)
    {
        java.util.HashSet keys = new java.util.HashSet();
        if (!store.keysAreEmbedded() && !store.keysAreSerialised())
        {
            // Retrieve the PersistenceCapable keys
            SetStore keystore = store.keySetStore(ownerSM.getPersistenceManager().getClassLoaderResolver());
            Iterator keyIter = keystore.iterator(ownerSM);
            while (keyIter.hasNext())
            {
                keys.add(keyIter.next());
            }
        }

        java.util.HashSet values = new java.util.HashSet();
        if (!store.valuesAreEmbedded() && !store.valuesAreSerialised())
        {
            // Retrieve the PersistenceCapable values
            SetStore valuestore = store.valueSetStore(ownerSM.getPersistenceManager().getClassLoaderResolver());
            Iterator valueIter = valuestore.iterator(ownerSM);
            while (valueIter.hasNext())
            {
                values.add(valueIter.next());
            }
        }

        // Retrieve the entries (key-value pairs so we can associate them)
        SetStore entries = store.entrySetStore();
        Iterator entryIter = entries.iterator(ownerSM);
        while (entryIter.hasNext())
        {
            Map.Entry entry = (Map.Entry)entryIter.next();
            Object key = entry.getKey();
            Object value = entry.getValue();

            Object trueKey = null;
            if (store.keysAreEmbedded() || store.keysAreSerialised())
            {
                // The value has been retrieved as part of the entry
                trueKey = key;
            }
            else
            {
                // Find the true key (with its fields populated)
                // The key in the entry contains just the identity
                Object pcKeyId = ((PersistenceCapable)key).jdoGetObjectId();
                Iterator keyIter = keys.iterator();
                while (keyIter.hasNext())
                {
                    Object obj = keyIter.next();
                    if (((PersistenceCapable)obj).jdoGetObjectId().equals(pcKeyId))
                    {
                        trueKey = obj;
                        break;
                    }
                }
            }

            Object trueValue = null;
            if (store.valuesAreEmbedded() || store.valuesAreSerialised())
            {
                // The value has been retrieved as part of the entry
                trueValue = value;
            }
            else
            {
                // Find the true value (with its fields populated)
                // The value in the entry contains just the identity
                Object pcValueId = ((PersistenceCapable)value).jdoGetObjectId();
                Iterator valueIter = values.iterator();
                while (valueIter.hasNext())
                {
                    Object obj = valueIter.next();
                    if (((PersistenceCapable)obj).jdoGetObjectId().equals(pcValueId))
                    {
                        trueValue = obj;
                    }
                }
            }
            delegate.put(trueKey, trueValue);
        }
        keys.clear();
        values.clear();
    }
    
    /**
     * Returns <tt>true</tt> if this collection contains the specified
     * element.  More formally, returns <tt>true</tt> if and only if this
     * collection contains at least one element <tt>it</tt> such that
     * <tt>(o==null ? it==null : o.equals(it))</tt>.<p>
     * <p>
     * This implementation iterates over the elements in the collection,
     * checking each element in turn for equality with the specified element.
     * @param backingStore the Store
     * @param sm the StateManager
     * @return <tt>true</tt> if this collection contains the specified element.
     */
     public static Object[] toArray(CollectionStore backingStore, StateManager sm)
     {
         Object[] result = new Object[backingStore.size(sm)];
         Iterator it = backingStore.iterator(sm);
         for (int i=0; it.hasNext(); i++)
         {
             result[i] = it.next();
         }        
         return result;
    }

    /**
     * Returns an array containing all of the elements in this collection; 
     * 
     * @param  backingStore the Store
     * @param  sm the StateManager
     * @param  a the array into which the elements of the collection are to
     *         be stored, if it is big enough; otherwise, a new array of the
     *         same runtime type is allocated for this purpose.
     * @return an array containing the elements of the collection.
     * 
     * @throws NullPointerException if the specified array is <tt>null</tt>.
     * 
     * @throws ArrayStoreException if the runtime type of the specified array
     *         is not a supertype of the runtime type of every element in this
     *         collection.
     */
    public static Object[] toArray(CollectionStore backingStore, StateManager sm, Object a[])
    {
        int size = backingStore.size(sm);
        if (a.length < size)
        {
            a = (Object[])java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
        }
        Iterator it=backingStore.iterator(sm);
        for (int i=0; i<size; i++)
        {
            a[i] = it.next();
        }

        if (a.length > size)
        {
            a[size] = null;
        }

        return a;        
    }

    /**
     * Convenience method for creating a Comparator using extension metadata tags for the specified field.
     * Uses the extension key "comparator-name" (and allows for "comparatorName" (deprecated).
     * @param fmd The field that needs the comparator
     * @param clr ClassLoader resolver
     * @return The Comparator
     */
    static Comparator getComparator(AbstractPropertyMetaData fmd, ClassLoaderResolver clr)
    {
        Comparator comparator = null;
        String comparatorName = null;
        if (fmd.hasMap() && fmd.getMap().hasExtension("comparator-name"))
        {
            comparatorName = fmd.getMap().getValueForExtension("comparator-name");
        }
        else if (fmd.hasCollection() && fmd.getCollection().hasExtension("comparator-name"))
        {
            comparatorName = fmd.getCollection().getValueForExtension("comparator-name");
        }
        else if (fmd.hasMap() && fmd.getMap().hasExtension("comparatorName"))
        {
            comparatorName = fmd.getMap().getValueForExtension("comparatorName");
        }
        else if (fmd.hasCollection() && fmd.getCollection().hasExtension("comparatorName"))
        {
            comparatorName = fmd.getCollection().getValueForExtension("comparatorName");
        }

        if (comparatorName != null)
        {
            Class comparatorCls = null;
            try
            {
                comparatorCls = clr.classForName(comparatorName);
                comparator = (Comparator)ClassUtils.newInstance(comparatorCls, null, null);
            }
            catch (JDOException jdoe)
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("SCO.Comparator.ClassNotFound", fmd.getFullFieldName(), comparatorName));
            }
        }
        return comparator;
    }
}