/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
2005 Jorg von Frantzius - updates for fetch-depth
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.sco.SCO;
import org.jpox.store.fieldmanager.SingleValueFieldManager;

/**
 * Utility class to handle the detach of fields.
 *
 * @version $Revision: 1.21 $
 **/
public class DetachFieldManager extends AbstractFetchFieldManager
{
    /**
     * Constructor for a field manager for detachment.
     * @param sm the StateManager of the instance being detached. An instance in Persistent or Transactional state
     * @param secondClassMutableFields The second class mutable fields for the class of this object
     * @param fpClass Fetch Plan for the class of this instance
     * @param state State object to hold any pertinent controls for the detachment process
     */
    public DetachFieldManager(StateManager sm, boolean[] secondClassMutableFields, FetchPlanForClass fpClass, FetchPlanState state)
    {
        super(sm, secondClassMutableFields, fpClass, state);
    }

    /**
     * Method to fetch an object field whether it is SCO collection, PC, or whatever for the detachment process.
     * @param fieldNumber Number of the field
     * @return The object
     */
    protected Object internalFetchObjectField(int fieldNumber)
    {
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);

        if (value == null)
        {
            return null;
        }
        else if (secondClassMutableFields[fieldNumber])
        {
            SCO sco;
            if (value instanceof SCO)
            {
                sco = (SCO) value;
            }
            else
            {
                // Replace with a SCO wrapper so that we can detach it
                sco = (SCO) sm.replaceSCOField(fieldNumber, value);
            }

            return sco.detachCopy(state);
        }
        else if (value.getClass().isArray() &&
                 PersistenceCapable.class.isAssignableFrom(value.getClass().getComponentType()))
        {
            // Array of PersistenceCapable objects
            // Return a new array with all elements detached (use the existing element if already detached, otherwise
            // detach a copy and put it in the new array).
            Object[] arrValue = (Object[])value;
            Object[] arrDetached = new Object[arrValue.length];
            for (int j=0;j<arrValue.length;j++)
            {
                PersistenceCapable pc = (PersistenceCapable)arrValue[j];
                if (!pc.jdoIsDetached())
                {
                    PersistenceCapable pcDetached =
                        (PersistenceCapable)((PersistenceManager)pc.jdoGetPersistenceManager()).detachCopyInternal(pc, state);
                    arrDetached[j] = pcDetached;
                }
                else
                {
                    arrDetached[j] = pc;
                }
            }
            return arrDetached;
        }
        else if (value instanceof PersistenceCapable)
        {
            PersistenceCapable pc = (PersistenceCapable) value;
            if (pc.jdoIsDetached())
            {
                return pc;
            }
            return ((PersistenceManager)pc.jdoGetPersistenceManager()).detachCopyInternal(pc, state);
        }
        else
        {
            return value;
        }
    }

    /**
     * Method to throw and EndOfFetchPlanGraphException since we're at the end of a branch in the tree.
     * @param fieldNumber Number of the field
     * @return Object to return
     */
    protected Object endOfGraphOperation(int fieldNumber)
    {
        // check if the object here is PC and is in the detached cache anyway
        SingleValueFieldManager sfv = new SingleValueFieldManager();
        sm.provideFields(new int[]{fieldNumber}, sfv);
        Object value = sfv.fetchObjectField(fieldNumber);
        if (value instanceof PersistenceCapable)
        {
            PersistenceCapable detached = ((DetachState)state).getDetachedCopyObject(value);
            if (detached != null)
            {
                // While we are at the end of a branch and this would go beyond the depth limits, 
                // the object here *is* already detached so just return it
                return detached;
            }
        }

        // we reached a leaf of the object graph to detach
        throw new EndOfFetchPlanGraphException();
    }
}