/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
 **********************************************************************/
package org.jpox.state;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.api.ApiAdapter;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.store.FieldValues;

/**
 * Factory of StateManagers.
 * @version $Revision: 1.2 $
 */
public class StateManagerFactory
{
    /**
     * Constructs a state manager to manage a hollow instance having the
     * given object ID.  This constructor is used for creating new instances of
     * existing persistent objects (i.e. via {@link PersistenceManager#getObjectById(Object, boolean, boolean)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     */
    public static StateManager newStateManager(PersistenceManager pm, Class pcClass, Object id)
    {
        return new StateManagerImpl(pm,pcClass,id);
    }
    
    /**
     * Constructs a state manager to manage a PersistenceCapable instance that will
     * be EMBEDDED/SERIALISED into another PersistenceCapable object. The instance will not be
     * assigned an identity in the process since it is a SCO.
     * @param pc The PersistenceCapable to manage (see copyPc also)
     * @param pm The persistence manager controlling this state manager.
     * @param copyPc Whether the SM should manage a copy of the passed PC or that one
     */
    public static StateManager newStateManager(PersistenceCapable pc, PersistenceManager pm, boolean copyPc)
    {
        return new StateManagerImpl(pc,pm,copyPc);
    }

    /**
     * Constructs a state manager to manage a hollow or pclean instance having the given
     * FieldValues. This constructor is used for creating new instances of
     * existing persistent objects using application identity.
     * {@link PersistenceManager#getObjectByAID}
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param fv the initial field values of the object.
     */
    public static StateManager newStateManager(PersistenceManager pm, Class pcClass, FieldValues fv)
    {
        return new StateManagerImpl(pm,pcClass,fv);
    }

    /**
     * Constructs a state manager to manage a recently populated hollow instance having the
     * given object ID and the given field values.  This constructor is used for
     * creating new instances of persistent objects obtained e.g. via a Query or
     * backed by a view (i.e. via {@link PersistenceManager#getObjectById(Object,FieldValues)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     * @param fv the initial field values of the object.
     */
    public static StateManager newStateManager(PersistenceManager pm, Class pcClass, Object id, FieldValues fv)
    {
        return new StateManagerImpl(pm,pcClass,id,fv);
    }

    /**
     * Constructs a state manager to manage a transient instance that is 
     * becoming newly persistent.  A new object ID for the
     * instance is obtained from the store manager and the object is inserted
     * in the data store.
     * <p>This constructor is used for assigning state managers to existing
     * instances that are transitioning to a persistent state (i.e. via
     * {@link PersistenceManager#makePersistent(Object)}).
     *
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     * @param preInsertChanges Any changes to make before inserting
     */
    public static StateManager newStateManager(PersistenceManager pm, PersistenceCapable pc, FieldValues preInsertChanges)
    {
        return new StateManagerImpl(pm,pc,preInsertChanges);
    }

    /**
     * Constructs a state manager to manage a Transactional Transient instance.
     * A new object ID for the instance is obtained from the store manager and
     * the object is inserted in the data store.
     * <p>
     * This constructor is used for assigning state managers to Transient
     * instances that are transitioning to a transient clean state (i.e. via
     * {@link PersistenceManager#makeTransactional(Object)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     */
    public static StateManager newStateManager(PersistenceManager pm, PersistenceCapable pc)
    {
        return new StateManagerImpl(pm,pc);
    }

    /**
     * Constructor for creating SM instances to manage PersistenceCapable
     * objects in detached state
     * @param apiAdapter the ApiAdapter
     * @param pc the detach object.
     * @param cmd the AbstractClassMetaData
     * @param id the JDO identity of the object.
     * @param version the detached version
     * @since 1.1
     */
    public static StateManager newStateManager(ApiAdapter apiAdapter, PersistenceCapable pc, AbstractClassMetaData cmd, Object id, Object version)
    {
        return new StateManagerImpl(apiAdapter,pc,cmd,id,version);
    }

    /**
     * Constructor to create a StateManager for a PersistenceCapable object, assigning the
     * specified id to the object. This is used when getting objects out of the L2 Cache,
     * where they have no StateManager assigned, and returning them as associated with a
     * particular PM.
     * @param pm PersistenceManager managing this object
     * @param id Id to assign to the PersistenceCapable object
     * @param pc The PersistenceCapable object
     * @param loaded The list of loaded fields (when put in the cache)
     */
    public static StateManager newStateManager(PersistenceManager pm, Object id, PersistenceCapable pc, boolean loaded[])
    {
        return new StateManagerImpl(pm,id,pc,loaded);
    }
}