/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
2004 Erik Bengtson - AdapterTime methods
   ...
**********************************************************************/
package org.jpox.store;

import java.sql.Timestamp;
import java.util.HashSet;

import org.jpox.ClassLoaderResolver;
import org.jpox.plugin.PluginManager;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingManager;
import org.jpox.util.Localiser;

/**
 * Generalised datastore representation. Each datastore is assumed to have a
 * "name", and a "version" (of the form major.minor.revision) and will store
 * "identifiers". In addition, each field will have a Mapping to the datastore
 * <P>
 * This will typically be extended to provide e.g RDBMS connection using JDBC,
 * or XML file connection using a DOM parser.
 * </P>
 * @version $Revision: 1.25 $
 */
public abstract class AbstractDatastoreAdapter implements DatastoreAdapter
{
    protected static final Localiser LOCALISER_BASE = Localiser.getInstance("org.jpox.store.Localisation");

    /** The set of reserved keywords for this datastore. */
    protected final HashSet reservedKeywords = new HashSet();

    /** The product name of the underlying datastore. */
    protected String datastoreProductName;

    /** The version number of the underlying datastore as a string. */
    protected String datastoreProductVersion;

    /** The major version number of the underlying datastore. */
    protected int datastoreMajorVersion;

    /** The minor version number of the underlying datastore. */
    protected int datastoreMinorVersion;

    /** The revision version number of the underlying datastore. */
    protected int datastoreRevisionVersion = 0;

    /** The String used to quote identifiers. */
    protected String identifierQuoteString;

    /** Manager for the mapping between Java and datastore types. */
    protected MappingManager mappingManager;

    /**
     * Constructor.
     */
    public AbstractDatastoreAdapter()
    {
    }

    /**
     * Accessor for a new mapping manager.
     * Must be implemented by subclasses.
     * @return The new mapping manager
     */
    protected abstract MappingManager getNewMappingManager();

    /**
     * Load the datastore mapping declared as Plug-in
     * @param mgr the PluginManager
     * @param clr the ClassLoaderResolver
     */
    public void loadDatastoreMapping(PluginManager mgr, ClassLoaderResolver clr)
    {
        getMappingManager().loadDatastoreMapping(mgr, clr, getVendorID());
    }

    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = getNewMappingManager();
        }
        return mappingManager;
    }

    /**
     * Convenience accessor for the mapping for the specified class.
     * Provides a wrapper to the method on the MappingManager.
     * @param c Java type
     * @param storeMgr the StoreManager
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c, StoreManager storeMgr)
    {
        return getMappingManager().getMapping(c, false, false, null, this, storeMgr.getPMFContext().getTypeManager());
    }

    /**
     * Convenience accessor for the mapping for the specified class.
     * Provides a wrapper to the method on the MappingManager.
     * @param c Java type
     * @param serialised Whether the type is serialised
     * @param embedded Whether the type is embedded
     * @param fieldName Name of field (for logging only).
     * @param storeMgr the StoreManager
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c, boolean serialised, boolean embedded, String fieldName, StoreManager storeMgr)
    {
        return getMappingManager().getMapping(c, serialised, embedded, fieldName, this, storeMgr.getPMFContext().getTypeManager());
    }

    /**
     * Convenience accessor for the mapping for the specified class.
     * Provides a wrapper to the method on the MappingManager.
     * @param c Java type
     * @param storeMgr The Store manager
     * @param clr ClassLoader resolver
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c, StoreManager storeMgr, ClassLoaderResolver clr)
    {
        return getMappingManager().getMapping(c, false, false, storeMgr, clr, this);
    }

    /**
     * Convenience accessor for the mapping for the specified class.
     * Provides a wrapper to the method on the MappingManager.
     * @param c Java type
     * @param expr the ScalarExpression
     * @return The mapping for the class.
     **/
    protected final JavaTypeMapping getMapping(Class c, ScalarExpression expr)
    {
        return getMappingManager().getMapping(c, false, false, expr.getQueryExpression().getStoreManager(), expr.getQueryExpression().getClassLoaderResolver(), this);
    }
    
    /**
     * Convenience accessor for the mapping for the specified class.
     * Provides a wrapper to the method on the MappingManager.
     * @param c Java type
     * @param serialised Whether the type is serialised
     * @param embedded Whether the type is embedded
     * @param storeMgr The Store manager
     * @param clr ClassLoader resolver
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c, boolean serialised, boolean embedded, StoreManager storeMgr, ClassLoaderResolver clr)
    {
        return getMappingManager().getMapping(c, serialised, embedded, storeMgr, clr, this);
    }

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#getVendorID()
     */
    public String getVendorID()
    {
        return null;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#isReservedKeyword(java.lang.String)
     */
    public boolean isReservedKeyword(String word)
    {
        return reservedKeywords.contains(word);
    }

    /**
     * Accessor for an identifier quote string.
     * @return Identifier quote string.
     **/
    public String getIdentifierQuoteString()
    {
        return identifierQuoteString;
    }


    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#newQueryStatement(org.jpox.store.DatastoreContainerObject, org.jpox.ClassLoaderResolver)
     */
    public abstract QueryExpression newQueryStatement(DatastoreContainerObject container, ClassLoaderResolver clr);

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#newQueryStatement(org.jpox.store.DatastoreContainerObject, org.jpox.store.DatastoreIdentifier, org.jpox.ClassLoaderResolver)
     */
    public abstract QueryExpression newQueryStatement(DatastoreContainerObject container, DatastoreIdentifier rangeVar, ClassLoaderResolver clr);

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#getAdapterTime(java.sql.Timestamp)
     */
    public long getAdapterTime(Timestamp time)
    {
        long timestamp = getTime(time.getTime(), time.getNanos());
        int ms = getMiliseconds(time.getNanos());

        return timestamp + ms;
    }

    protected long getTime(long time, long nanos)
    {
        if (nanos < 0)
        {
            return (((time / 1000) - 1) * 1000);
        }
        return (time / 1000) * 1000;
    }

    protected int getMiliseconds(long nanos)
    {
        return (int) (nanos / 1000000);
    }

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#getDatastoreMajorVersion()
     */
    public int getDatastoreMajorVersion()
    {
        return datastoreMajorVersion;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#getDatastoreMinorVersion()
     */
    public int getDatastoreMinorVersion()
    {
        return datastoreMinorVersion;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.IDatastoreAdapter#modOperator(org.jpox.store.expression.ScalarExpression, org.jpox.store.expression.ScalarExpression)
     */
    public NumericExpression modOperator(ScalarExpression operand1, ScalarExpression operand2)
    {
        return new NumericExpression(operand1, ScalarExpression.OP_MOD, operand2);
    }

    /**
     * Accessor for whether autoincrementing fields are supported.
     * @return Whether we support autoincrementing fields
     */
    public boolean supportsAutoIncrementFields()
    {
        return false;
    }

    /**
     * Accessor for whether sequences are supported.
     * @return whether we support sequences.
     **/
    public boolean supportsSequences()
    {
        return false;
    }

    /**
     * Whether the datastore supports setting a fetch size lower than one.
     * Usually, the datastore should ignore settting fetch size lower than one.
     * @return whether we support fetch size lower than one
     **/
    public boolean supportsQueryFetchSizeLowerThanOne()
    {
        return true;
    }
}