/**********************************************************************
 Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 Contributors:
 ...
 **********************************************************************/
package org.jpox.store;

import javax.jdo.JDOFatalUserException;

import org.jpox.store.IdentifierFactory;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Abstract representation of an identifier factory.
 * To be extended to generate the identifiers.
 * 
 * @version $Revision: 1.7 $
 */
public abstract class AbstractIdentifierFactory implements IdentifierFactory
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");

    public static final int CASE_PRESERVE = 1;
    public static final int CASE_UPPER = 2;
    public static final int CASE_LOWER = 3;

    /** Adapter for the datastore. */
    protected DatastoreAdapter dba;

    /** Case to use for identifiers. */
    protected int identifierCase;

    /** Range to use for creating hased ending when truncating identifiers. */
    private static final int HASH_RANGE = Character.MAX_RADIX * Character.MAX_RADIX / 2;

    /**
     * Constructor.
     * @param dba Datastore adapter
     * @param requiredCase The case the user requires
     */
    public AbstractIdentifierFactory(DatastoreAdapter dba, String requiredCase)
    {
        this.dba = dba;

        // Set the identifier case to be used based on what the user has requested and what the datastore supports
        int userIdentifierCase = CASE_UPPER;
        if (requiredCase.equalsIgnoreCase("UpperCase"))
        {
            userIdentifierCase = CASE_UPPER;
        }
        else if (requiredCase.equalsIgnoreCase("LowerCase"))
        {
            userIdentifierCase = CASE_LOWER;
        }
        else if (requiredCase.equalsIgnoreCase("PreserveCase"))
        {
            userIdentifierCase = CASE_PRESERVE;
        }

        if (userIdentifierCase == CASE_UPPER)
        {
            if (dba.isStoresUpperCaseIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE;
            }
            else if (dba.isStoresUpperCaseQuotedIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE_QUOTED;
            }
            else if (dba.isStoresMixedCaseIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE;
            }
            else if (dba.isStoresMixedCaseQuotedIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE_QUOTED;
            }
            else
            {
                if (dba.isStoresLowerCaseIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative","UPPERCASE", "LOWERCASE"));
                }
                else if (dba.isStoresLowerCaseQuotedIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE_QUOTED;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative","UPPERCASE", "LOWERCASEQUOTED"));
                }
                else
                {
                    // Should never happen since we've tried all options
                    throw new JDOFatalUserException(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastore", "UPPERCASE"));
                }
            }
        }
        else if (userIdentifierCase == CASE_LOWER)
        {
            if (dba.isStoresLowerCaseIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE;
            }
            else if (dba.isStoresLowerCaseQuotedIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE_QUOTED;
            }
            else if (dba.isStoresMixedCaseIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE;
            }
            else if (dba.isStoresMixedCaseQuotedIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE_QUOTED;
            }
            else
            {
                if (dba.isStoresUpperCaseIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative","LOWERCASE", "UPPERCASE"));
                }
                else if (dba.isStoresUpperCaseQuotedIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE_QUOTED;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative","LOWERCASE", "UPPERCASEQUOTED"));
                }
                else
                {
                    // Should never happen since we've tried all options
                    throw new JDOFatalUserException(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastore", "LOWERCASE"));
                }
            }
        }
        else if (userIdentifierCase == CASE_PRESERVE)
        {
            if (dba.isStoresMixedCaseIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_MIXED_CASE;
            }
            else if (dba.isStoresMixedCaseQuotedIdentifiers())
            {
                identifierCase = IdentifierFactory.IDENTIFIER_MIXED_CASE_QUOTED;
            }
            else
            {
                if (dba.isStoresLowerCaseIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative","MIXEDCASE", "LOWERCASE"));
                }
                else if (dba.isStoresLowerCaseQuotedIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_LOWER_CASE_QUOTED;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative", "MIXEDCASE", "LOWERCASEQUOTED"));
                }
                else if (dba.isStoresUpperCaseIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative", "MIXEDCASE", "UPPERCASE"));
                }
                else if (dba.isStoresUpperCaseQuotedIdentifiers())
                {
                    identifierCase = IdentifierFactory.IDENTIFIER_UPPER_CASE_QUOTED;
                    JPOXLogger.JDO.warn(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastoreUsingAlternative", "MIXEDCASE", "UPPERCASEQUOTED"));
                }
                else
                {
                    // Should never happen since we've tried all options
                    throw new JDOFatalUserException(LOCALISER.msg("IdentifierFactory.CaseNotAvailableForDatastore", "MIXEDCASE"));
                }
            }
        }
        else
        {
            // Case not supported
            throw new JDOFatalUserException(LOCALISER.msg("IdentifierFactory.CaseNotSupported", userIdentifierCase));
        }
    }

    /**
     * Accessor for the datastore adapter that we are creating identifiers for.
     * @return The datastore adapter
     */
    public DatastoreAdapter getDatastoreAdapter()
    {
        return dba;
    }

    /**
     * Accessor for the identifier case being used.
     * @return The identifier case
     */
    public int getIdentifierCase()
    {
        return identifierCase;
    }

    /**
     * Convenience method to return the name for the identifier case.
     * @return Identifier case name
     */
    public String getNameOfIdentifierCase()
    {
        if (identifierCase == IDENTIFIER_LOWER_CASE)
        {
            return "lowercase";
        }
        else if (identifierCase == IDENTIFIER_LOWER_CASE_QUOTED)
        {
            return "\"lowercase\"";
        }
        else if (identifierCase == IDENTIFIER_MIXED_CASE)
        {
            return "\"MixedCase\"";
        }
        else if (identifierCase == IDENTIFIER_MIXED_CASE_QUOTED)
        {
            return "\"MixedCase\"";
        }
        else if (identifierCase == IDENTIFIER_UPPER_CASE)
        {
            return "UPPERCASE";
        }
        else if (identifierCase == IDENTIFIER_UPPER_CASE_QUOTED)
        {
            return "\"UPPERCASE\"";
        }
        else
        {
            return "UNKNOWN";
        }
    }

    /**
     * Method to truncate an identifier to fit within the specified identifier length.
     * If truncation is necessary will use a 2 char hashcode (at the end) to attempt to create uniqueness.
     * @param identifier The identifier
     * @param length The (max) length to use
     * @return The truncated identifier.
     */
    protected static String truncate(String identifier, int length)
    {
        if (identifier.length() > length)
        {
            // Truncation is necessary so cut down to "maxlength-2" and add 2 char hashcode
            int tailIndex = length - 2;
            int tailHash = identifier.substring(tailIndex).hashCode();

            // Scale the hash code down to the range 0 - 1295
            if (tailHash < 0)
            {
                tailHash = tailHash % HASH_RANGE + (HASH_RANGE - 1);
            }
            else
            {
                tailHash = tailHash % HASH_RANGE + HASH_RANGE;
            }

            String suffix = "0" + Integer.toString(tailHash, Character.MAX_RADIX);

            return identifier.substring(0, tailIndex) + suffix.substring(suffix.length() - 2);
        }
        else
        {
            return identifier;
        }
    }

    /**
     * Method to return the maximum permitted length of an identifier of the 
     * specified type. Returns -1 if the identifier type is not supported.
     * @param identifierType Type of identifier
     * @return The maximum length
     */
    protected abstract int getMaxLengthForIdentifierType(int identifierType);
}