/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import java.sql.Timestamp;

import org.jpox.ClassLoaderResolver;
import org.jpox.plugin.PluginManager;
import org.jpox.store.expression.ExpressionConversionAdapter;
import org.jpox.store.expression.ExpressionLogicSetAdapter;
import org.jpox.store.expression.ExpressionMethodAdapter;
import org.jpox.store.expression.ExpressionOperatorAdapter;
import org.jpox.store.expression.ExpressionPatternAdapter;
import org.jpox.store.expression.ExpressionSupportedFeaturesAdapter;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingManager;

/**
 * Definition of a datastore adapter.
 * 
 * @version $Revision: 1.30 $
 */
public interface DatastoreAdapter extends ExpressionConversionAdapter, ExpressionOperatorAdapter, ExpressionMethodAdapter, ExpressionSupportedFeaturesAdapter, ExpressionPatternAdapter, ExpressionLogicSetAdapter
{
    /**
     * Accessor for the Mapping Manager for field mapping management.
     * @return The Mapping Manager.
     */
    public abstract MappingManager getMappingManager();

    /**
     * Accessor for the Vendor ID for this datastore.
     * @return Vendor id for this datastore
     */
    public abstract String getVendorID();
    
    /**
     * Load the datastore mappings declared as Plug-in
     * @param mgr the PluginManager
     * @param clr the ClassLoaderResolver
     */
    public void loadDatastoreMapping(PluginManager mgr, ClassLoaderResolver clr);

    /**
     * Method to check if a word is reserved for this datastore.
     * @param word The word
     * @return Whether it is reserved
     */
    public abstract boolean isReservedKeyword(String word);

    /**
     * Creates the auxiliary functions/procedures in the datastore 
     * @param conn the connection to the datastore
     */
    public abstract void initialiseDatastore(Object conn);

    /**
     * Accessor for the quote string to use when quoting identifiers.
     * @return The quote string for the identifier
     */
    public abstract String getIdentifierQuoteString();

    /**
     * Accessor for a new query statement.
     * @param container The table to query 
     * @param clr the ClassLoaderResolver
     * @return The Query Statement
     **/
    public abstract QueryExpression newQueryStatement(DatastoreContainerObject container, ClassLoaderResolver clr);

    /**
     * Accessor for a new query statement.
     * @param container The table to query 
     * @param rangeVar A range variable for the query
     * @param clr the ClassLoaderResolver
     * @return The Query Statement
     **/
    public abstract QueryExpression newQueryStatement(DatastoreContainerObject container, DatastoreIdentifier rangeVar, ClassLoaderResolver clr);

    /**
     * Accessor for the mapping for the specified class.
     * @param c Java type
     * @param storeMgr the StoreManager
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c, StoreManager storeMgr);

    /**
     * Accessor for the mapping for the specified class.
     * @param c Java type
     * @param serialised Whether the type is serialised
     * @param embedded Whether the type is embedded
     * @param fieldName Name of field (for logging only)
     * @param storeMgr the StoreManager
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c, boolean serialised, boolean embedded, String fieldName, StoreManager storeMgr);

    /**
     * Convenience accessor for the mapping for the specified class.
     * Simply directs the request to the same method taking serialised and embedded as false.
     * @param c Class to query
     * @param storeMgr The Store Manager
     * @param clr The ClassLoaderResolver
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c, StoreManager storeMgr, ClassLoaderResolver clr);

    /**
     * Accessor for the mapping for the specified class.
     * @param c Class to query
     * @param serialised Whether the type is serialised
     * @param embedded Whether the type is embedded
     * @param storeMgr The Store Manager
     * @param clr The ClassLoaderResolver
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c, boolean serialised, boolean embedded, StoreManager storeMgr, ClassLoaderResolver clr);

    /**
     * Utility to return the adapter time in case there are rounding issues with
     * millisecs etc.
     * @param time The timestamp
     * @return The time in millisecs
     */
    public abstract long getAdapterTime(Timestamp time);

    /**
     * Accessor for the datastore major version number.
     * @return Major version number
     */
    public abstract int getDatastoreMajorVersion();

    /**
     * Accessor for the datastore minor version number.
     * @return Minor version number
     */
    public abstract int getDatastoreMinorVersion();
    
    /**
     * Method to generate a modulus expression. The binary % operator is said to
     * yield the remainder of its operands from an implied division; the
     * left-hand operand is the dividend and the right-hand operand is the
     * divisor.
     * @param operand1 the left expression
     * @param operand2 the right expression
     * @return The Expression for modulus
     */
    public abstract NumericExpression modOperator(ScalarExpression operand1, ScalarExpression operand2);

    /**
     * Verifies if the given <code>columnDef</code> is auto incremented by the datastore.
     * @param columnDef the datastore type name
     * @return true when the <code>columnDef</code> has values auto incremented by the datastore
     **/
    boolean isAutoIncrementingDataType(String columnDef);

    /**
     * Whether this datastore adapter support auto-incrementing fields.
     * The column could be things like "AUTOINCREMENT", "IDENTITY", "SERIAL" in SQL.
     * @return whether autoincrementing fields are supported.
     **/
    public abstract boolean supportsAutoIncrementFields();

    /**
     * Whether this datastore adapter supports sequences.
     * @return whether sequences are supported.
     **/
    public abstract boolean supportsSequences();

    /**
     * Whether the datastore supports setting a fetch size lower than one.
     * Usually, the datastore should ignore settting fetch size lower than one.
     * @return whether we support fetch size lower than one
     **/
    boolean supportsQueryFetchSizeLowerThanOne();

    /**
     * Accessor for whether this adapter stores lower case identifiers.
     * @return Whether lowercase identifiers are stored
     */
    public boolean isStoresLowerCaseIdentifiers();

    /**
     * Accessor for whether this adapter stores "lowercase" identifiers.
     * @return Whether "lowercase" identifiers are stored
     */
    public boolean isStoresLowerCaseQuotedIdentifiers();

    /**
     * Accessor for whether this adapter stores MixedCase identifiers.
     * @return Whether MixedCase identifiers are stored
     */
    public boolean isStoresMixedCaseIdentifiers();

    /**
     * Accessor for whether this adapter "MixedCase" identifiers.
     * @return Whether "MixedCase" identifiers are stored
     */
    public boolean isStoresMixedCaseQuotedIdentifiers();

    /**
     * Accessor for whether this adapter stores UPPERCASE identifiers.
     * @return Whether UPPERCASE identifiers are stored
     */
    public boolean isStoresUpperCaseIdentifiers();

    /**
     * Accessor for whether this adapter stores "UPPERCASE" identifiers.
     * @return Whether "UPPERCASE" identifiers are stored
     */
    public boolean isStoresUpperCaseQuotedIdentifiers();

    /**
     * Method to return this object as a string.
     * @return String version of this object.
     */
    public abstract String toString();
}