/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;

/**
 * Representation of a Java class in a datastore.
 * In the case of RDBMS this will be a table (primary or secondary).
 * In the case of a file-based structure this may be a directory.
 * In the case of an XML-based structure this may be an element.
 *
 * @version $Revision: 1.28 $
 **/
public interface DatastoreClass extends DatastoreContainerObject
{
    /**
     * Accessor for the primary class represented.
     * @return Name of the class
     **/
    String getType();

    /**
     * Accessor for the identity-type used by this table.
     * @return identity-type tag value
     */    
    public IdentityType getIdentityType();

    /**
     * Accessor for whether the object id will be attributed by the datastore
     * directly, or whether values have to be supplied.
     * @return Whether it is attributed in the datastore
     */
    boolean isObjectIDDatastoreAttributed();

    /**
     * Accessor for whether this datastore class is the base datastore class
     * for this inheritance hierarchy.
     * @return Whether it is the base.
     */
    boolean isBaseDatastoreClass();

    /**
     * Method to return the base DatastoreClass that persists the
     * specified field. This navigates up through the superclass
     * tables to find a table that manages the field.
     * @param fmd MetaData for the field required
     * @return The DatastoreClass managing that field
     */
    public DatastoreClass getBaseDatastoreClassWithField(AbstractPropertyMetaData fmd);

    /**
     * Accessor for the supertable for this table.
     * This is only relevant if the DatastoreClass in use supports supertables.
     * If supertables arent supported by the datastore then null is returned.
     * @return The supertable (if any)
     **/
    public DatastoreClass getSuperDatastoreClass();

    /**
     * Accessor for whether this table manages the specified class
     * @param className Name of the class
     * @return Whether it is managed by this table
     */
    public boolean managesClass(String className);

    /**
     * Method to insert an object for this class.
     * Will insert any superclass records as well (recursive).
     * @param sm StateManager for the object to insert.
     **/
    void insert(StateManager sm);
    
    /**
     * Method to fetch an object for this class.
     * Will fetch any superclass records as well (recursive).
     * @param sm StateManager for the object to fetch.
     * @param fieldMetaData MetaData for the fields to be fetched
     */
    void fetch(StateManager sm, AbstractPropertyMetaData[] fieldMetaData);

    /**
     * Method to update an object for this class.
     * Will update any superclass records as well (recursive).
     * @param sm StateManager for the object to update.
     * @param fieldMetaData MetaData for the fields to be updated
     */
    void update(StateManager sm, AbstractPropertyMetaData[] fieldMetaData);

    /**
     * Method to delete an object for this class.
     * Will delete any superclass records as well (recursive).
     * @param sm StateManager for the object to delete.
     **/
    void delete(StateManager sm);

    /**
     * Locates this object in the datastore.
     * @param sm The StateManager for the object to be found 
     * @throws javax.jdo.JDOObjectNotFoundException If the instance does not exist in
     * the datastore
     */
    void locate(StateManager sm);

    /**
     * Delete dependent objects of a persistent object.
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldMetaData MetaData for the fields to delete.
     */
    void deleteDependent(StateManager sm, AbstractPropertyMetaData[] fieldMetaData);

    /**
     * Accessor for the name of the datastore class (table).
     * @return The name
     */
    String toString();

    // --------------------------- Mapping Access ------------------------------

    /**
     * Accessor for a mapping for the datastore ID (OID) for this object.
     * @return The (OID) mapping.
     **/
    JavaTypeMapping getDataStoreObjectIdMapping();

    /**
     * Accessor for the mapping for the specified field name.
     * Doesn't cope with fields of the same name in different subclasses - you
     * should call the equivalent method passing FieldMetaData for those.
     * @param fieldName Name of field
     * @return The Mapping for the field.
     */
    JavaTypeMapping getFieldMapping(String fieldName);

    /**
     * Accessor for the mapping for the specified field.
     * @param fmd AbstractPropertyMetaData of the field
     * @return The Mapping for the field.
     */
    JavaTypeMapping getFieldMapping(AbstractPropertyMetaData fmd);

    /**
     * Accessor for a mapping for the datastore ID (OID) for this table.
     * @param consumer Consumer for the mappings
     **/
    void provideDatastoreIdMappings(MappingConsumer consumer);

    /**
     * Provide the mappings to the consumer for all primary-key fields mapped to
     * this table (for application identity).
     * @param consumer Consumer for the mappings
     */
    void providePrimaryKeyMappings(MappingConsumer consumer);

    /**
     * Provide the mappings to the consumer for all non primary-key fields
     * mapped to this table.
     * @param consumer Consumer for the mappings
     */
    void provideNonPrimaryKeyMappings(MappingConsumer consumer);

    /**
     * Provide the mappings to the consumer for all absolute field Numbers in this table
     * that are container in the fieldNumbers parameter.
     * @param consumer Consumer for the mappings
     * @param fieldMetaData MetaData of the fields to provide mappings for
     * @param includeSecondaryTables Whether to supply fields in secondary tables
     */
    void provideMappingsForFields(MappingConsumer consumer, AbstractPropertyMetaData[] fieldMetaData, boolean includeSecondaryTables);

    /**
     * Provide the mappings to version mappings
     * @param consumer Consumer for the version mappings
     */
    void provideVersionMappings(MappingConsumer consumer);

    /**
     * Provide the mappings to discriminator mappings
     * @param consumer Consumer for the mappings
     */
    void provideDiscriminatorMappings(MappingConsumer consumer);

    /**
     * Instruction to provide all external mappings to the passed consumer.
     * @param consumer The consumer for the mappings
     * @param mappingType Type of external mapping to provide
     */
    void provideExternalMappings(MappingConsumer consumer, int mappingType);

    /**
     * Create a new OID. 
     * OID can be datastore identity or application identity
     * @param pm The PersistenceManager
     * @param pcClass The Persistence-Capable class
     * @return New OID
     */
    Object newOID(PersistenceManager pm, PersistenceCapable pcClass);

    /**
     * Accessor for the external mapping for the specified field of the specified type.
     * An external mapping is a mapping for which there is no field in the actual class
     * to represent it (part of a relation).
     * The type can be FK, FK discriminator, order, etc
     * @param fmd MetaData for the (external) field
     * @param mappingType The type of mapping
     * @return The external mapping
     */
    JavaTypeMapping getExternalMapping(AbstractPropertyMetaData fmd, int mappingType);
}