/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import org.jpox.metadata.ExtendableMetaData;
import org.jpox.store.mapping.JavaTypeMapping;

/**
 * Representation of a Java field in a datastore.
 * In the case of RDBMS this will be a column.
 * In the case of a file-based structure this may be a file.
 * In the case of an XML-based structure this may be an node. 
 *
 * @version $Revision: 1.7 $
 **/ 
public interface DatastoreField extends DatastoreObject
{
    public static final int ROLE_CUSTOM = -1;
    public static final int ROLE_NONE = 0;
    public static final int ROLE_OWNER = 1;
    public static final int ROLE_FIELD = 2;
    public static final int ROLE_COLLECTION_ELEMENT = 3;
    public static final int ROLE_ARRAY_ELEMENT = 4;
    public static final int ROLE_MAP_KEY = 5;
    public static final int ROLE_MAP_VALUE = 6;
    public static final int ROLE_INDEX = 7;

    /**
     * Accessor for the type of data stored in this field.
     * @return The type of data in the field.
     **/
    String getStoredJavaType();

    /**
     * Mutator to make the field the primary key.
     **/
    void setAsPrimaryKey();

    /**
     * Accessor for whether the field is the primary key
     * @return whether the field is (part of) the primary key
     **/
    boolean isPrimaryKey();

    /**
     * Accessor for whether the field is nullable
     * @return whether the field is nullable
     **/
    boolean isNullable();

    /**
     * Accessor for the Mapping for this field. 
     * @return The Mapping
     */
    JavaTypeMapping getMapping();
    
    /**
     * Accessor for the DatastoreContainerObject container of this field
     * @return The DatastoreContainerObject
     */
    DatastoreContainerObject getDatastoreContainerObject();
    
    /**
     * Wraps the column name with a FUNCTION.
     * <PRE>example: SQRT(?) generates: SQRT(columnName)</PRE>
     * @param replacementValue the replacement to ?. Probably it's a column name, that may be fully qualified name or not
     * @return a String with function taking as parameter the replacementValue
     */
    String applySelectFunction(String replacementValue);

    /**
     * Copy the configuration of this field to another field
     * @param col the datastore field
     */
	void copyConfigurationTo(DatastoreField col);
	
    /**
     * Mutator for the nullability of the datastore field.
     *
     * @return The datastore field with the updated info
     **/
    DatastoreField setNullable();
    
    /**
     * Mutator for the defaultability of the datastore field.
     *
     * @return The datastore field with the updated info
     **/
    DatastoreField setDefaultable(); 
    
    /**
     * Access the metadata definition for this DatastoreField. For RDBMS
     * mapping it will return an instance of ColumnMetaData
     * @return the ExtendableMetaData
     */
    ExtendableMetaData getExtendableMetaData();

}
