/******************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Erik Bengtson - changed to use org.w3c.dom
2004 Andy Jefferson - added table-owner. Changed table to be optional
   ...
*****************************************************************/
package org.jpox.store;

import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.store.exceptions.DatastoreInitialisationException;
import org.jpox.util.JPOXLogger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * An auto-starter mechanism storing its definition in an XML file.
 * Is independent of the datastore since it is stored as a file and not
 * in the actual datastore.
 *
 * TODO Add a JPOXAutoStart DTD to validate the file automatically.
 * TODO If we have one per PMF, need to guarantee unique naming of file.
 * 
 * @version $Revision: 1.9 $
 */
public class XMLAutoStarter extends AbstractAutoStartMechanism
{
    protected final URL fileUrl;
    protected Document doc;
    protected Element rootElement;

    /**
     * Constructor, taking the XML file URL.
     * @param storeMgr The StoreManager managing the store that we are auto-starting.
     * @param clr The ClassLoaderResolver
     * @throws MalformedURLException 
     */
    public XMLAutoStarter(StoreManager storeMgr, ClassLoaderResolver clr) throws MalformedURLException
    {
        super();

        this.fileUrl = new URL("file:" + storeMgr.getPMFContext().getPmfConfiguration().getAutoStartMechanismXmlFile());

        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final DocumentBuilder db;

        try
        {
            db = factory.newDocumentBuilder();
            try
            {
                db.setEntityResolver(new XMLAutoStarterEntityResolver());
                rootElement = db.parse(new InputSource(new InputStreamReader(fileUrl.openStream()))).getDocumentElement();
                doc = rootElement.getOwnerDocument();
            }
            catch (Exception e)
            {
                JPOXLogger.JDO.info(LOCALISER.msg("XMLAutoStarter.StartupFileNonExistent", fileUrl.getFile()));

                // File doesn't exist, so create it
                doc = db.newDocument();
                rootElement = doc.createElement("jpox");
                doc.appendChild(rootElement);

                writeToFile();
            }
        }
        catch (ParserConfigurationException e1)
        {
            JPOXLogger.JDO.error(LOCALISER.msg("XMLAutoStarter.ErrorReadingStartupFile", fileUrl.getFile(), e1.getMessage()));
        }
    }

    /**
     * Accessor for all auto start data for this starter.
     * @return The class auto start data. Collection of StoreData elements
     * @throws DatastoreInitialisationException
     */
    public Collection getAllClassData()
    throws DatastoreInitialisationException
    {
        Collection classes=new HashSet();

        NodeList classElements = rootElement.getElementsByTagName("class");
        for( int i=0; i<classElements.getLength(); i++)
        {
            Element element = (Element) classElements.item(i);

            // Default to owner unless specified
            boolean owner=true;
            if (element.getAttribute("table-owner") != null && element.getAttribute("table-owner").equals("false"))
            {
                owner = false;
            }
            StoreData classData = new StoreData(element.getAttribute("name"),
                								element.getAttribute("table"),
                								owner,
                								element.getAttribute("type").equals("FCO") ? StoreData.FCO_TYPE : StoreData.SCO_TYPE,
                								element.getAttribute("interface-name"));
            classes.add(classData);
        }
        
        return classes;
    }

    /**
     * Whether it's open for writing (add/delete) classes to the auto start
     * mechanism
     * @return whether this is open for writing 
     */
    public boolean isOpen()
    {
        // Always open
        return true;
    }

    /**
     * Closes a transaction for writing (add/delete) classes to the auto start
     * mechanism 
     */
    public void close()
    {
        writeToFile();
        super.close();
    }

    /**
     * Method to add a class to the starter.
     * @param data The store data to add
     */
    public void addClass(StoreData data)
    {
        // TODO Check for existence of this class
        Element classElement = doc.createElement("class");
        classElement.setAttribute("name", data.getName());
        if (data.hasTable())
        {
            classElement.setAttribute("table", data.getTableName());
            classElement.setAttribute("table-owner", data.isTableOwner() ? "true" : "false");
        }

        classElement.setAttribute("type", data.isFCO() ? "FCO" : "SCO");
        classElement.setAttribute("version", AbstractPersistenceManagerFactory.getVersionNumber());
        if (data.getInterfaceName() != null)
        {
            classElement.setAttribute("interface-name", data.getInterfaceName());
        }

        rootElement.appendChild(classElement);
    }

    /**
     * Method to remove a class from the starter
     * @param className The name of the class to remove.
     */
    public void deleteClass(String className)
    {
        NodeList classElements = rootElement.getElementsByTagName("class");
        for( int i=0; i<classElements.getLength(); i++)
        {
            Element element = (Element) classElements.item(i);
            String attr = element.getAttribute("name");
            if (attr != null && attr.equals(className))
            {
                rootElement.removeChild(element);
            }
        }
    }

    /**
     * Method to remove all classes from the starter.
     */
    public void deleteAllClasses()
    {
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final DocumentBuilder db;

        try
        {
        	// Dont do any deletion, just reconstruct the DOM with an empty root element
            db = factory.newDocumentBuilder();
            doc = db.newDocument();
            rootElement = doc.createElement("jpox");
            doc.appendChild(rootElement);
        }
        catch (ParserConfigurationException e)
        {
            JPOXLogger.JDO.error(LOCALISER.msg("XMLAutoStarter.ErrorWritingStartupFile", fileUrl.getFile(), e.getMessage()));
        }
    }

    /**
     * Method to give a descriptive name for the starter process.
     * @return Description of the starter process.
     */
    public String getStorageDescription()
    {
        return LOCALISER.msg("XMLAutoStarter.StorageDescription");
    }

    /**
     * Method to write the DOM to its file.
     */
    private synchronized void writeToFile()
    {
        // Write the DOM back to file
        try
        {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer m = tf.newTransformer();
            DOMSource source = new DOMSource(doc);
            FileOutputStream os = new FileOutputStream(fileUrl.getFile());
            StreamResult result = new StreamResult(os);
            m.setOutputProperty(OutputKeys.INDENT, "yes");
            m.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC, XMLAutoStarterEntityResolver.PUBLIC_ID_KEY);
            m.transform(source, result);
            os.close();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            JPOXLogger.JDO.error(LOCALISER.msg("XMLAutoStarter.ErrorWritingStartupFile", fileUrl.getFile(), e.getMessage()));
        }
    }
}