/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

import javax.jdo.JDOUserException;

import org.jpox.store.mapping.JavaTypeMapping;

/**
 * An expression representing a class.
 * This is used in JDOQL when wanting to
 * <UL>
 * <LI>access public static final fields</LI>
 * <LI>use the "instanceof" operator</LI>
 * </UL>
 * In the first case we invoke accessField on the ClassExpression.
 * In the second case we invoke instanceOf on an ObjectExpression, passing in a ClassExpression.
 *
 * @version $Revision: 1.5 $
 */
public class ClassExpression extends ScalarExpression
{
    /** The class being represented. */
    private Class cls;

    /**
     * Constructor.
     * @param qs The Query Statement
     * @param cls The class
     **/
    public ClassExpression(QueryExpression qs, Class cls)
    {
        super(qs);

        this.cls = cls;
    }

    /**
     * Accessor for the class being represented.
     * @return The class
     */
    public Class getCls()
    {
        return cls;
    }

    /**
     * Method called when wanting to call public static final methods on the class.
     * @param fieldName Name of the public static final field
     * @param innerJoin Not used
     * @return Expression for the field access
     */
    public ScalarExpression accessField(String fieldName, boolean innerJoin)
    {
        try
        {
            Field fld = cls.getField(fieldName);
            if (!Modifier.isStatic(fld.getModifiers()) || !Modifier.isFinal(fld.getModifiers()) || !Modifier.isPublic(fld.getModifiers()))
            {
                throw new JDOUserException(LOCALISER.msg("ClassExpression.MethodNotPublicStaticFinal", fieldName, cls.getName()));
            }
            Object value = fld.get(null);
            if (value == null)
            {
                return new NullLiteral(qs);
            }
            JavaTypeMapping m = qs.getStoreManager().getDatastoreAdapter().getMapping(value.getClass(), 
                qs.getStoreManager(), qs.getClassLoaderResolver());
            return m.newLiteral(qs, value);
        }
        catch (IllegalAccessException iae)
        {
            // Will not happen since we already checked for it
        }
        catch (NoSuchFieldException nsfe)
        {
            throw new JDOUserException(LOCALISER.msg("ClassExpression.MethodNotFoundForClass", fieldName, cls.getName()));
        }
        return null;
    }
}