/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.expression;

import org.jpox.ClassLoaderResolver;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.query.StatementText;

/**
 * QueryExpression instances contain query statements in language-dependent form. 
 * A statement used as criteria for perfoming a search
 * @version $Revision: 1.16 $
 */
public interface QueryExpression
{
    /**
     * Whether this query will return a meta data expression in the SELECT clause
     * @return hasMetaDataExpression
     */
    boolean hasMetaDataExpression();

    /**
     * Accessor to the table expression for the given range identifier by <code>rangeVar</code>
     * @param rangeVar the range variable identifier
     * @return the TableExpression
     */
    LogicSetExpression getTableExpression(DatastoreIdentifier rangeVar);

    /**
     * Accessor to the default table expression
     * @return the TableExpression
     */
    LogicSetExpression getDefaultTableExpression();

    /**
     * Accessor for the store manager associated with this query.
     * @return The store manager
     */
    StoreManager getStoreManager();

    /**
     * Accessor for the ClassLoaderResolver to use with this query statement.
     * @return ClassLoader resolver
     */
    ClassLoaderResolver getClassLoaderResolver();

    /**
     * Set whether this statement returns distinct results.
     * @param distinctResults Whether we return distinct results
     */
    void setDistinctResults(boolean distinctResults);

    /**
     * Mutator for whether to only use inner joins.
     * @param innerJoin Whether to only use inner joins
     */
    void setUseInnerJoinsOnly(boolean innerJoin);

    /**
     * Creates a table expression
     * @param mainTable the main table
     * @param rangeVar the range var
     * @return TableExpression
     */
    LogicSetExpression newTableExpression(DatastoreContainerObject mainTable, DatastoreIdentifier rangeVar);
    
    /**
     * Creates a table expression
     * @param mainTable the main table
     * @param rangeVar the range var
     * @param unionQueries Whether to add to any union
     * @return TableExpression[]
     */
    LogicSetExpression[] newTableExpression(DatastoreContainerObject mainTable, DatastoreIdentifier rangeVar, boolean unionQueries);

    /**
     * Select the datastore identity column.
     * @param alias Alias to use for this column
     * @param unionQueries Whether to select the datastore id column of all unioned tables
     * @return The position of the column in the result set
     */
    int[] selectDatastoreIdentity(String alias, boolean unionQueries);

    /**
     * Select the version column.
     * @param alias Alias to use for this column
     * @param unionQueries Whether to select the version column of all unioned tables
     * @return The position of the column in the result set
     */
    int[] selectVersion(String alias, boolean unionQueries);

    /**
     * Select the column(s) for the specified field in the primary table of the query.
     * @param fieldName Name of the field
     * @param alias Alias to use for these column(s)
     * @param unionQueries Whether to select the field column(s) of all unioned queries.
     * @return The position of the columns in the result set.
     */
    int[] selectField(String fieldName, String alias, boolean unionQueries);

    /**
     * Select the columns for a mapping
     * @param mapping The mapping
     * @return The index of the columns in the select
     **/
    int[] select(JavaTypeMapping mapping);
    
    /**
     * select a new column, add to union queries, if unionQueries is true
     * @param mapping The mapping
     * @param unionQueries Whether to add to any union
     * @return The index of the columns in the select
     */
    int[] select(JavaTypeMapping mapping, boolean unionQueries);
    
    /**
     * select an expression; eg: "'Text' as alias"
     * @param expr The expression to add to the select statement 
     * @return The index of the expression in the select
     */
    int selectScalarExpression(ScalarExpression expr);
    
    /**
     * select an expression; eg: "'Text' as alias"
     * @param expr The expression to add to the select statement
     * @param unionQueries whether to apply the select in all queries unified by the union clause 
     * @return The index of the expression in the select
     */
    int selectScalarExpression(ScalarExpression expr, boolean unionQueries);
    
    /**
     * Select columns, add to union queries, if unionQueries is true
     * @param rangeVar The range variable
     * @param mapping The mapping
     * @return The index of the columns in the select
     */
    int[] select(DatastoreIdentifier rangeVar, JavaTypeMapping mapping);
    
    /**
     * select columns, add to union queries, if unionQueries is true
     * @param rangeVar The range variable
     * @param mapping The mapping
     * @param unionQueries Whether to add to any union
     * @return The index of the column in the select
     */
    int[] select(DatastoreIdentifier rangeVar, JavaTypeMapping mapping, boolean unionQueries);

    /**
     * add an condition to the query.
     * @param condition the Boolean expression
     */
    void andCondition(BooleanExpression condition);

    /**
     * add an condition to the query and queries involved in the union if unionQuery is true
     * @param condition the Boolean expression 
     * @param unionQueries whether to apply the condition in all queries unified by the union clause 
     */
    void andCondition(BooleanExpression condition, boolean unionQueries);

    /**
     * Method to add tables/crossJoin but no joins to they, will be output as FROM TABLE1,TABLE2,TABLE3
     * @param tableExpr table expression
     * @param unionQueries Whether to apply the alias to unions of this query.
     **/
    void crossJoin(LogicSetExpression tableExpr, boolean unionQueries);

    /**
     * Method to add tables/alias but no joins to they, will be output as FROM TABLE1,TABLE2,TABLE3
     * Checks parent expressions until reach the root expression
     * @param tableExpr table expression
     **/
    boolean hasCrossJoin(LogicSetExpression tableExpr);

    /**
     * Sets the parent QueryExpression of this query. In SQL it can be exemplified as
     * <code> 
     * SELECT 1 FROM PARENT WHERE EXISTS (SELECT 1 FROM THIS)
     * </code>
     * The parent QueryExpression is the nesting SQL.
     * 
     * @param parentQueryExpr the parent of this query
     **/
    void setParent(QueryExpression parentQueryExpr);
    
    /**
     * Method to do an inner join to another table, and optionally apply it to
     * any unions for this query.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr the 
     * @param equals if the join is applied as filter, if use equals or not equals
     * @param unionQueries whether to apply the inner join in all queries unified by the union clause 
     **/
    void innerJoin(ScalarExpression expr, ScalarExpression expr2, 
                   LogicSetExpression tblExpr, boolean equals, boolean unionQueries);

    /**
     * Method to do an inner join to another table.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression for the table to apply the join
     * @param equals if the join is applied as filter, if use equals or not equals
     **/
    void innerJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr, boolean equals);

    /**
     * Method to do a left outer join to another table, and optionally apply it
     * to any unions for this query.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression for the table to apply the join
     * @param equals if the join is applied as filter, if use equals or not equals
     * @param unionQueries Whether to apply to unions of this query.
     **/
    void leftOuterJoin(ScalarExpression expr, ScalarExpression expr2, 
                       LogicSetExpression tblExpr, boolean equals, boolean unionQueries);

    /**
     * Method to do a left outer join to another table.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression
     * @param equals if the join is applied as filter, if use equals or not equals
     **/
    void leftOuterJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr, boolean equals);

    /**
     * Method to do a right outer join to another table, and optionally apply it
     * to any unions for this query.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression for the table to apply the join
     * @param equals if the join is applied as filter, if use equals or not equals
     * @param unionQueries Whether to apply to unions of this query.
     **/
    void rightOuterJoin(ScalarExpression expr, ScalarExpression expr2, 
                        LogicSetExpression tblExpr, boolean equals, boolean unionQueries);

    /**
     * Method to do a right outer join to another table.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression for the table to apply the join
     * @param equals if the join is applied as filter, if use equals or not equals
     **/
    void rightOuterJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr, boolean equals);

    /**
     * Method to add a grouping clause to the statement.
     * Grouping clauses that are implied by the selected columns will be added automatically
     * so this provides a means to supplement them.
     * @param expr The group by expression
     */
    void addGroupingExpression(ScalarExpression expr);

    /**
     * Accessor for the grouping expressions of this statement.
     * @return The grouping expressions
     */
    ScalarExpression[] getGroupingExpressions();

    /**
     * Method to set the having clause of the statement.
     * @param expr The having expression
     */
    void setHaving(BooleanExpression expr);

    /**
     * Accessor for the having expression of this statement.
     * @return The having expression
     */
    BooleanExpression getHavingExpression();

    /**
     * Mutator for the ordering criteria.
     * @param exprs The expressions to order by
     * @param descending Whether each expression is ascending/descending
     **/
    void setOrdering(ScalarExpression[] exprs, boolean[] descending);

    /**
     * Accessor for the ordering expressions of this statement.
     * @return The ordering expressions
     */
    ScalarExpression[] getOrderingExpressions();

    /**
     * Method to convert the criteria into the statement text.
     * @param lock whether to lock the instances using this statement
     * @return The StatementText
     **/
    StatementText toStatementText(boolean lock); 
    
    /**
     * Union two QueryExpressions <code>this</code> and <code>qe</code>.
     * Both QueryExpressions must have the same ScalarExpressions selected,
     * and they must be in the same select order. 
     * valid:
     * e.g. a) fieldA, fieldB, fieldE, fieldC
     *      b) fieldA, fieldB, fieldE, fieldC
     * invalid:
     * e.g. a) fieldA, fieldE, fieldB, fieldC      
     *      b) fieldA, fieldB, fieldE, fieldC
     *      
     * @param qe the QueryExpression
     */
    void union(QueryExpression qe);

    /**
     * add an condition to the query.
     * @param condition the Boolean expression 
     */
    void iorCondition(BooleanExpression condition);

    /**
     * add an condition to the query and queries involved in the union if unionQuery is true
     * @param condition the Boolean expression 
     * @param unionQueries whether to apply the condition in all queries unified by the union clause 
     */
    void iorCondition(BooleanExpression condition, boolean unionQueries);

    /**
     * Method to add a range constraint on any SELECT.
     * This typically will use LIMIT/OFFSET where they are supported by
     * the underlying RDBMS.
     * @param offset The offset to start from
     * @param count The number of records to return
     */
    void setRangeConstraint(long offset, long count);

    /**
     * Set this query is to be used as a as set for the Exists function.
     * example WHERE EXISTS( QUERY )
     * @param isExistsSubQuery The isExistsSubQuery to set.
     */
    void setExistsSubQuery(boolean isExistsSubQuery);
    
    /**
     * Method to convert the criteria into a delete statement text.
     * @return The StatementText
     **/
    StatementText toDeleteStatementText();
    
    /**
     * Accessor for the number of ScalarExpression projected.
     * @return The number of columns in the SELECT
     **/
    int getNumberOfScalarExpressions();
    
    /**
     * Allows reseting the compiled expression
     */
    void reset();
}