/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.util.HashMap;

import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.exceptions.TableMismatchException;

/**
 * A SQL table expression that joins superclass tables by joining them directly
 * to the surrounding QueryStatement.
 *
 * @see QueryExpression
 * @version $Revision: 1.20 $ 
 */
public class TableExprAsJoins extends LogicSetExpression
{
    protected final HashMap rangeVarsByTable = new HashMap();

    /**
     * 
     * @param qs the QueryExpression
     * @param mainTable the main table in the query
     * @param mainRangeVar The alias/identifier of the table in the query
     */
    public TableExprAsJoins(QueryExpression qs, DatastoreContainerObject mainTable, DatastoreIdentifier mainRangeVar)
    {
        super(qs, mainTable, mainRangeVar);

        rangeVarsByTable.put(mainTable, mainRangeVar);
    }

    public String referenceColumn(DatastoreField col)
    {
        assertNotFrozen();

        DatastoreContainerObject table = col.getDatastoreContainerObject();
        DatastoreIdentifier rangeVar = (DatastoreIdentifier)rangeVarsByTable.get(table);

        if (rangeVar == null)
        {
            if (!(mainTable instanceof DatastoreClass) || !(table instanceof DatastoreClass))
            {
                throw new TableMismatchException(col, mainTable);
            }

            /*
             * Since both tables are ClassTables we assume that the column
             * is a superclass field, meaning 'table' is a supertable of
             * 'mainTable'.  We join the supertable to the query statement using
             * range variable derived from the main range variable (an
             * underscore and a counter are appended).
             *
             * So even though the caller may have been looking for column FOO
             * relative to rangevar THIS he may actually get back something like
             * THIS_1.FOO.
             */
            IdentifierFactory idFactory = qs.getStoreManager().getIdentifierFactory();
            rangeVar = idFactory.newIdentifier(mainRangeVar, String.valueOf(rangeVarsByTable.size()));
            rangeVarsByTable.put(table, rangeVar);

            /* mt... = "main table" */
            /* st... = "supertable" */
            DatastoreClass mt = (DatastoreClass)mainTable;
            DatastoreClass st = (DatastoreClass)table;
            LogicSetExpression stTblExpr = qs.newTableExpression(st, rangeVar);

            ScalarExpression mtExpr = mt.getIDMapping().newScalarExpression(qs,this);
            ScalarExpression stExpr = st.getIDMapping().newScalarExpression(qs,stTblExpr);

            qs.innerJoin(mtExpr,stExpr, stTblExpr, true, true);
        }
        String rangeVarString = rangeVar.toString();
        if (rangeVarString.length() > 0)
        {
            return col.applySelectFunction(rangeVarString + "." + col.getIdentifier());
        }
        else
        {
            return col.applySelectFunction(col.getIdentifier().toString());
        }
    }

    public String toString()
    {
        if (sqlText == null)
        {
            DatastoreIdentifier mainTableName = mainTable.getIdentifier();
            StringBuffer sb = new StringBuffer(mainTable.toString());
            if (!mainRangeVar.equals(mainTableName))
            {
                sb.append(' ').append(mainRangeVar);
            }

            sqlText = sb.toString();
        }

        return sqlText;
    }
}