/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.extent;

import java.util.HashMap;
import java.util.Iterator;

import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;

import org.jpox.PersistenceManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.store.StoreManager;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.util.Localiser;

/**
 * Abstract representation of an Extent for stores. Individual stores
 * should extend this to implement Extent's for their purpose.
 * 
 * @version $Revision: 1.9 $
 */
public abstract class AbstractExtent implements Extent
{
    /** Localised messages source */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");

    /** Persistence Manager for the Extent. */
    protected final PersistenceManager pm;

    /** The candidate class. We store the class since we need to retain it for class loading. */
    protected final Class candidateClass;

    /** Whether to include subclasses. */
    protected final boolean subclasses;

    /** ClassMetaData for the candidate class. */
    protected final AbstractClassMetaData cmd;

    /** Manager for the Store. */
    protected final StoreManager storeMgr;

    /** Query executed for returning the Extent. */
    protected final Query query;

    /** Map of the iterators of the Extents accessed. */
    protected HashMap queryResultsByIterator = new HashMap();

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param cls candidate class
     * @param subclasses Whether to include subclasses
     */
    public AbstractExtent(PersistenceManager pm, Class cls, boolean subclasses)
    {
        if (cls == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("Extent.ClassNameNotSpecified"));
        }

        // Find the MetaData for this class
        storeMgr = pm.getStoreManager();
        cmd = storeMgr.getMetaDataManager().getMetaDataForClass(cls, pm.getClassLoaderResolver());
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("Extent.ClassHasNoMetaData", cls.getName()));
        }

        this.pm = pm;
        this.candidateClass = cls;
        this.subclasses = subclasses;

        query = (Query) pm.newQuery();
        query.setClass(candidateClass);
        query.setCandidates(this);
    }

    /**
     * Returns whether this Extent was defined to contain subclasses.
     * @return true if this Extent was defined to include subclasses.
     */
    public boolean hasSubclasses()
    {
        return subclasses;
    }

    /**
     * Accessor for the class of instances in this Extent.
     * @return the Class of instances of this Extent
     */
    public Class getCandidateClass()
    {
        return candidateClass;
    }

    /**
     * Accessor for the owning PersistenceManager.
     * @return the owning PersistenceManager
     */
    public javax.jdo.PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * Returns an iterator over all the instances in the Extent.
     * @return an iterator over all the instances in the Extent.
     */
    public Iterator iterator()
    {
        QueryResult qr = (QueryResult)query.execute();
        Iterator iter = qr.iterator();

        queryResultsByIterator.put(iter, qr);

        return iter;
    }

    /**
     * Close an Iterator associated with this Extent instance. Iterators closed
     * by this method will return false to hasNext() and will throw
     * NoSuchElementException on next().  The Extent instance can still be used
     * as a parameter of Query.setCandidates, and to get an Iterator.
     *
     * @param iter an iterator obtained by the method iterator() on this Extent
     * instance.
     */
    public void close(Iterator iter)
    {
        QueryResult qr = (QueryResult) queryResultsByIterator.remove(iter);

        qr.close();
    }

    /**
     * Close all Iterators associated with this Extent instance.  Iterators
     * closed by this method will return false to hasNext() and will throw
     * NoSuchElementException on next().  The Extent instance can still be used
     * as a parameter of Query.setCandidates, and to get an Iterator.
     */
    public void closeAll()
    {
        Iterator iter = queryResultsByIterator.values().iterator();
        while (iter.hasNext())
        {
            QueryResult qr = (QueryResult) iter.next();

            qr.close();
            iter.remove();
        }
    }

    /**
     * This method retrieves the fetch plan associated with the Extent. It
     * always returns the identical instance for the same Extent instance. Any
     * change made to the fetch plan affects subsequent instance retrievals via
     * next(). Only instances not already in memory are affected by the fetch
     * plan. Fetch plan is described in Section 12.7.
     * @return the FetchPlan
     */
    public FetchPlan getFetchPlan()
    {
        return query.getFetchPlan();
    }
    
    /**
     * Stringifier method.
     * @return Stringified form of this object
     */
    public String toString()
    {
        return LOCALISER.msg("Extent.ClassExtentInfo", candidateClass.getName(), "" + subclasses);
    }
}