/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
	...
**********************************************************************/
package org.jpox.store.mapping;

import java.sql.PreparedStatement;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer;

import org.jpox.util.Localiser;

/**
 * Parameter setter class to deal with application identity.
 * Parameters must be in the same order as the fields.
 * 
 * TODO to support other types of fields in the application id fields, 
 * we must delegate the store method tasks to a Mapping...
 *
 * @version $Revision: 1.5 $ 
 */
class AppIDObjectIdFieldConsumer implements ObjectIdFieldConsumer
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");

    /** The PreparedStatement to populate. */
    private PreparedStatement ps;

    /** Parameter positions in the PreparedStatement. */
    private int[] param;

    /** Next parameter index. Increased each call. */
    private int next;

    /**
     * Constructor.
     * @param ps Prepared Statement
     * @param param The parameters to populate in the statement.
     */
    public AppIDObjectIdFieldConsumer(PreparedStatement ps, int[] param)
    {
        this.ps = ps;
        this.param = param;
        next = 0;
    }

    /**
     * Method to store a boolean in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeBooleanField(int fieldNumber, boolean value)
    {
        try
        {
            ps.setBoolean(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a character in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeCharField(int fieldNumber, char value)
    {
        try
        {
            ps.setString(param[next++], String.valueOf(value));
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a byte in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeByteField(int fieldNumber, byte value)
    {
        try
        {
            ps.setByte(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a short in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeShortField(int fieldNumber, short value)
    {
        try
        {
            ps.setShort(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store an integer in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeIntField(int fieldNumber, int value)
    {
        try
        {
            ps.setInt(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a long in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeLongField(int fieldNumber, long value)
    {
        try
        {
            ps.setLong(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a float in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeFloatField(int fieldNumber, float value)
    {
        try
        {
            ps.setFloat(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a double in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeDoubleField(int fieldNumber, double value)
    {
        try
        {
            ps.setDouble(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store a String in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeStringField(int fieldNumber, String value)
    {
        try
        {
            ps.setString(param[next++], value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }

    /**
     * Method to store an object in a field.
     * @param fieldNumber Number of the field
     * @param value The value to use
     */
    public void storeObjectField(int fieldNumber, Object value)
    {
        try
        {
            // this is used in indentifying relationships
            if (value instanceof PersistenceCapable)
            {
                PersistenceCapable pc = (PersistenceCapable) value;
				pc.jdoCopyKeyFieldsFromObjectId(this, pc.jdoGetObjectId());
            }
            else
            {
                ps.setObject(param[next++], value);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Mapping.CannotSetAppIDParameter", String.valueOf(value)), e);
        }
    }
}