/**********************************************************************
Copyright (c) 2003 Mike Martin and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - implementation of newScalarExpression, newLiteral
2005 Andy Jefferson - basic serialisation support
2005 Andy Jefferson - updated serialisation using SCOUtils methods
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.HashMap;
import java.util.Iterator;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.sco.Map;
import org.jpox.sco.SCO;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.MapExpression;
import org.jpox.store.expression.MapLiteral;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.rdbms.scostore.FKMapStore;
import org.jpox.store.scostore.MapStore;

/**
 * SCO Mapping for Map types.
 *
 * @version $Revision: 1.30 $
 **/
public class MapMapping extends AbstractContainerMapping implements MappingCallbacks
{
    /**
     * Accessor for the Store of the contents.
     * @param clr The ClassLoaderResolver
     * @return The store of the contents
     **/
    public synchronized MapStore getBackingStore(ClassLoaderResolver clr)
    {
        if (containerIsStoredInSingleColumn())
        {
            return null;
        }

        if (contentsStore == null)
        {
            contentsStore = storeMgr.getBackingStoreForMap(fmd, contentsTable, clr);
        }

        return (MapStore)contentsStore;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!obj.getClass().equals(getClass()))
        {
            return false;
        }

        MapMapping sm = (MapMapping)obj;

        return fmd.equals(sm.fmd) && storeMgr.equals(sm.storeMgr);
    }

    /**
     * Accessor for the Java type represented here.
     * @return The java type
     */
    public Class getJavaType()
    {
        return Map.class;
    }

    // ---------------- Implementation of MappingCallbacks --------------------

    /**
     * Method to be called after the insert of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postInsert(StateManager sm)
    {
        java.util.Map value = (java.util.Map) sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when serialised since we are handled in the main request
            if (value != null)
            {
                validateElementsForWriting(sm, value.keySet());
                validateElementsForWriting(sm, value.values());
            }
            return;
        }

        if (value == null)
        {
            // replace null map with an empty SCO wrapper
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
            return;
        }

        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).putAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
    }

    /**
     * Method to be called after any update of the owner class element.
     * @param sm StateManager of the owner
     */
    public void postUpdate(StateManager sm)
    {
        java.util.Map value = (java.util.Map) sm.provideField(fmd.getAbsoluteFieldNumber());
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when serialised since we are handled in the main request
            if (value != null)
            {
                validateElementsForWriting(sm, value.keySet());
                validateElementsForWriting(sm, value.values());
            }
            return;
        }

        if (value == null)
        {
            // replace null map with empty SCO wrapper
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
            return;
        }

        if (value instanceof SCO)
        {
            SCO sco = (SCO) value;

            if (sm.getObject() == sco.getOwner() && fieldName.equals(sco.getFieldName()))
            {
                return;
            }

            if (sco.getOwner() != null)
            {
                throw new JDOFatalInternalException("Owned second-class object was somehow assigned to a field other than its owner's");
            }
        }

        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
        getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).putAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newWrapper(sm, fieldName));
    }

    /**
     * Method to be called before any delete of the owner class element.
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        // Do nothing - dependent deletion is performed by deleteDependent()
    }

    /**
     * Method to delete all dependent objects (the contents of the Collection).
     * @param sm State Manager of the owner
     */
    public void deleteDependent(StateManager sm)
    {
        if (containerIsStoredInSingleColumn())
        {
            // Do nothing when serialised since we are handled in the main request
            return;
        }

        // makes sure field is loaded
        sm.isLoaded(sm.getObject(), fmd.getAbsoluteFieldNumber());
        java.util.Map value = (java.util.Map) sm.provideField(fmd.getAbsoluteFieldNumber());
        if (value != null && !value.isEmpty())
        {
            // if to delete key or values
            if (fmd.getMap().isDependentValue() || fmd.getMap().isDependentKey())
            {
                java.util.Map m = new HashMap();
                m.putAll(value);

                // clear the collection
                getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);

                // Delete the key and value objects since dependent
                for (Iterator it = m.entrySet().iterator(); it.hasNext();)
                {
                    Map.Entry entry = (Map.Entry) it.next();
                    Object val = entry.getValue();
                    Object key = entry.getKey();
                    if (fmd.getMap().isDependentKey() && key instanceof PersistenceCapable)
                    {
                        // TODO This should never have references to InverseMapStore, NormalMapStore (RDBMS-specific) information - remove it
                        // clear reference to the key in value if inverse relationship
                        // if key field is non nullable, the value is deleted too
                        if (getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()) instanceof FKMapStore)
                        {
                            if (val instanceof PersistenceCapable)
                            {
                                ((FKMapStore) getBackingStore(sm.getPersistenceManager().getClassLoaderResolver())).clearKeyOfValue(sm, key, val);
                            }
                        }
                        sm.getPersistenceManager().deletePersistent(key);
                    }
                    if (fmd.getMap().isDependentValue() && val instanceof PersistenceCapable)
                    {
                        sm.getPersistenceManager().deletePersistent(val);
                    }
                }
            }
            else
            {
                // clear the collection
                getBackingStore(sm.getPersistenceManager().getClassLoaderResolver()).clear(sm);
            }
        }
    }

    // -------------------------------- JDOQL Query Methods --------------------------------------

    /**
     * Accessor for a literal representing this type.
     * @param qs The Query
     * @param value the value of this object in the literal
     * @return The literal
     */
    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ContainerMapping.QueryInvalidIfSerialised", fmd.getFullFieldName()));
        }
        return new MapLiteral(qs, this, (java.util.Map)value);
    }

    /**
     * Method to return a scalar expression for the Map.
     * @param qs The QueryStatement
     * @param te The TableExpression
     * @return The ScalarExpression 
     **/
    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        if (containerIsStoredInSingleColumn())
        {
            throw new JDOFatalUserException(LOCALISER.msg("ContainerMapping.QueryInvalidIfSerialised", fmd.getFullFieldName()));
        }
        return new MapExpression(qs, datastoreContainer.getIDMapping(), te, getBackingStore(qs.getClassLoaderResolver()), fieldName);
    }
}