/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - added javadocs
2004 Andy Jefferson - changed to use Column spec from MetaData
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.sql.Timestamp;

import javax.jdo.JDOException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.VersionMetaData;
import org.jpox.metadata.VersionStrategy;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;

/**
 * Mapping class for mapping version state/timestamp columns in the database.
 * This class is for internal use only. It should not be used in user mappings.
 * 
 * @version $Revision: 1.24 $
 */
public final class VersionMapping extends SingleFieldMapping implements SimpleDatastoreRepresentation
{
    private final JavaTypeMapping delegate;

    private final VersionMetaData versionMetaData;

    /**
     * Constructor.
     * @param dba Datastore Adapter
     * @param datastoreContainer Datastore table
     * @param delegate The JavaTypeMapping to delegate the storage
     */
    public VersionMapping(DatastoreAdapter dba, DatastoreContainerObject datastoreContainer, JavaTypeMapping delegate)
    {
        initialize(dba, delegate.getType());
        this.delegate = delegate;
        this.versionMetaData = datastoreContainer.getVersionMetaData();

        // Currently we only use a single column mapping for versioning.
        // The MetaData supports multiple columns and so we could extend this in the future
        // to use all MetaData information.
        ColumnMetaData[] versionColumnMetaData = versionMetaData.getColumnMetaData();
        ColumnMetaData colmd;
        IdentifierFactory idFactory = datastoreContainer.getStoreManager().getIdentifierFactory();
        DatastoreIdentifier id = null;
        if (versionColumnMetaData.length == 0)
        {
            // No column name so generate a default
            id = idFactory.newVersionFieldIdentifier();
            colmd = new ColumnMetaData(versionMetaData, id.getIdentifier());
            datastoreContainer.getVersionMetaData().addColumn(colmd);
        }
        else
        {
            // Column name defined
            colmd = versionColumnMetaData[0];
            id = idFactory.newDatastoreFieldIdentifier(colmd.getName());
        }
        DatastoreField column = datastoreContainer.addDatastoreField(getType(), id, this, colmd);
        datastoreContainer.getStoreManager().getMappingManager().createDatastoreMapping(delegate, datastoreContainer.getStoreManager(), 
            column, getType());
    }

    /**
     * Accessor for a sample value for this column
     * @return The sample value
     */
    public Object getSampleValue(ClassLoaderResolver clr)
    {
        return delegate.getSampleValue(clr);
    }

    /**
     * Accessor for a new literal for this mapping.
     * @param qs The QueryStatement
     * @param value The value of the object
     * @return The new literal
     */
    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
        return delegate.newLiteral(qs, value);
    }

    /**
     * Accessor for a new literal for this mapping.
     * @param qs The QueryStatement
     * @param te The table expression
     * @return The new literal
     */
    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        return delegate.newScalarExpression(qs, te);
    }

    /**
     * Accessor for whether to include this column in any fetch statement
     * @return Whether to include the column when fetching.
     */
    public boolean includeInFetchStatement()
    {
        return false;
    }

    /**
     * Accessor for the number of datastore fields.
     * @return Number of datastore fields.
     */
    public int getNumberOfDatastoreFields()
    {
        return delegate.getNumberOfDatastoreFields();
    }

    /**
     * Accessor for a datastore mapping.
     * @param index The mapping index
     * @return the datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        return delegate.getDataStoreMapping(index);
    }

    /**
     * Method to add a datastore mapping.
     * @param datastoreMapping The mapping
     */
    public void addDataStoreMapping(DatastoreMapping datastoreMapping)
    {
        delegate.addDataStoreMapping(datastoreMapping);
    }

    /**
     * Accessor for the next version of the object
     * @param actualVersion The actual version
     * @return The next version
     */
    public Object getNextVersion(Object actualVersion)
    {
        if (versionMetaData == null)
        {
            // TODO Support optimistic when not using strategy
            throw new JDOException("JPOX doesnt currently support the use of optimistic transactions when no strategy is specified");
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.NONE)
        {
            // Just increment the version
            if (actualVersion == null)
            {
                return new Long(1);
            }
            return new Long(((Long)actualVersion).longValue()+1);
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.DATE_TIME)
        {
            return new Timestamp(System.currentTimeMillis());
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
        {
            if (actualVersion == null)
            {
                return new Long(1);
            }
            return new Long(((Long)actualVersion).longValue()+1);
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.STATE_IMAGE)
        {
            // TODO Support state-image strategy
            throw new JDOException("JPOX doesnt currently support JDO2 optimistic version strategy \"state-image\"");
        }
        else
        {
            throw new JDOException("Unknown optimistic version strategy - not supported");
        }
    }

    /**
     * Compare the transaction vs datastore version.
     * @param versionDatastore
     * @param versionTransaction
     * @return true if transaction version == datastore version
     */
    public boolean compareVersion(Object versionDatastore, Object versionTransaction)
    {
        if (versionTransaction == null)
        {
            return true;
        }

        if (versionMetaData == null)
        {
            // TODO Support optimistic when not using strategy
            throw new JDOException("JPOX doesnt currently support the use of optimistic transactions when no strategy is specified");
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.DATE_TIME)
        {
            return ((Timestamp)versionTransaction).getTime() == ((Timestamp)versionDatastore).getTime();
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
        {
            return ((Long)versionTransaction).longValue() == ((Long)versionDatastore).longValue();
        }
        else if (versionMetaData.getVersionStrategy() == VersionStrategy.STATE_IMAGE)
        {
            // TODO Support state-image strategy
            throw new JDOException("JPOX doesnt currently support JDO2 optimistic version strategy \"state-image\"");
        }
        else
        {
            throw new JDOException("Unknown optimistic version strategy - not supported");
        }
    }

    /**
     * Accessor for the type represented here, returning the class itself
     * @return This class.
     */
    public Class getJavaType()
    {
        return VersionMapping.class;
    }

    /**
     * Mutator for the object in this column
     * @param pm The Persistence Manager
     * @param preparedStatement The statement
     * @param exprIndex The indexes
     * @param value The value to set it to
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        delegate.setObject(pm, preparedStatement, exprIndex, value);
    }

    /**
     * Accessor for the object in this column
     * @param pm The Persistence manager
     * @param resultSet The ResultSet to get the value from
     * @param exprIndex The indexes
     * @return The object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        return delegate.getObject(pm, resultSet, exprIndex);
    }
}