/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved.   
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.poid;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.NoSuchElementException;

import org.jpox.util.StringUtils;

/**
 * Representation of a block of ids.
 *
 * @version $Revision: 1.3 $
 */
public class PoidBlock implements Serializable
{
    /** The next id position. */
    private int nextIndex = 0;

    /** The list of ids in this block. */
    private final List poidBlock;

    /**
     * Constructor for PoidBlock.
     * @param oid The block of objects that will be considered the "IDs"
     */
    public PoidBlock(Object[] oid)
    {
        poidBlock = Arrays.asList(oid);
    }

    /**
     * Constructor for PoIdBlock.
     * @param oid The list of objects that will be considered the "IDs"
     */
    public PoidBlock(List oid)
    {
        poidBlock = new ArrayList(oid);
    }

    /**
     * Accessor for the current Poid.
     * @return The current poid
     * @throws NoSuchElementException Thrown if no current poid
     */
    public Poid current()
    {
        if (nextIndex == 0 || (nextIndex-1) >= poidBlock.size())
        {
            throw new NoSuchElementException();
        }
        return new Poid(poidBlock.get(nextIndex-1));
    }

    /**
     * Accessor for the next Poid, or null if block ids exhausted
     * @return The next POID
     */
    public Poid next()
    {
        if (nextIndex >= poidBlock.size())
        {
            throw new NoSuchElementException();
        }

        return new Poid(poidBlock.get(nextIndex++));
    }

    /**
     * Accessor for whether there are more elements remaining in the block.
     * @return True when has more elements
     */
    public boolean hasNext()
    {
        return (nextIndex < poidBlock.size());
    }

    /**
     * Method to append a block onto this block.
     * This is used where we have some ids left, and we want to allocate more
     * to go into this block.
     * @param block The other block
     */
    public void addBlock(PoidBlock block)
    {
        if (block == null)
        {
            return;
        }

        while (block.hasNext())
        {
            poidBlock.add(block.next());
        }
    }

    /**
     * Stringify method.
     * @return A string version of this object
     */
    public String toString()
    {
        return "PoidBlock : " + StringUtils.collectionToString(poidBlock);
    }
}