/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - fixed creation of repository when block failed
2004 Andy Jefferson - removed MetaData requirement
2006 Andy Jefferson - changed to hold the generator rather than creating one each time
2006 Andy Jefferson - rewritten to handle the creation of PoidGenerator nad lookup.
    ...
**********************************************************************/
package org.jpox.store.poid;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.store.StoreManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Manager for the creation of PoidGenerators.
 * Allows creation of generators and provides lookup by symbolic name.
 *
 * @version $Revision $ 
 */
public class PoidManager
{
    /** Localisation of messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");

    /** Map of PoidGenerator keyed by the symbolic name. */
    protected Map generatorsByName = new HashMap();

    /**
     * Constructor.
     */
    public PoidManager()
    {
    }

    /**
     * Method to clear out the generators managed by this manager.
     */
    public void clear()
    {
        generatorsByName.clear();
    }

    /**
     * Accessor for the PoidGenerator with the given symbolic name.
     * @param name Name of the PoidGenerator when created
     * @return The PoidGenerator with this name
     */
    public PoidGenerator getPoidGenerator(String name)
    {
        if (name == null)
        {
            return null;
        }
        return (PoidGenerator)generatorsByName.get(name);
    }

    /**
     * Method to create a PoidGenerator.
     * @param clr ClassLoader resolver
     * @param name Symbolic name of the generator
     * @param generatorClassName Class name for the generator type
     * @param props Properties to control the generator
     * @return The PoidGenerator
     */
    public PoidGenerator createPoidGenerator(ClassLoaderResolver clr, String name, String generatorClassName, Properties props)
    {
        return createPoidGenerator(clr, name, generatorClassName, props, null, null);
    }

    /**
     * Method to create a PoidGenerator when the generator is datastore based.
     * @param clr ClassLoader resolver
     * @param name Symbolic name of the generator
     * @param generatorClassName Class name for the generator type
     * @param props Properties to control the generator
     * @param storeMgr Manager for the store
     * @param connectionProvider Provider for connections
     * @return The PoidGenerator
     */
    public PoidGenerator createPoidGenerator(ClassLoaderResolver clr, String name, String generatorClassName, Properties props, 
            StoreManager storeMgr, PoidConnectionProvider connectionProvider)
    {
        // Create the requested generator
        PoidGenerator generator;
        try
        {
            if( JPOXLogger.POID.isDebugEnabled() )
            {
                JPOXLogger.POID.debug("Creating PoidGenerator instance of "+generatorClassName+" for "+name);
            }
            Class[] argTypes = new Class[] {String.class, Properties.class};
            Object[] args = new Object[] {name, props};
            Class generatorClass = clr.classForName(generatorClassName, AbstractPersistenceManagerFactory.class.getClassLoader(), true);
            Constructor ctor = generatorClass.getConstructor(argTypes);
            generator = (PoidGenerator)ctor.newInstance(args);
        }
        catch (Exception e)
        {
            JPOXLogger.JDO.error(e);
            throw new PoidException(LOCALISER.msg("PoidManager.InvalidGeneratorType", generatorClassName,e),e);
        }

        if (generator instanceof AbstractDatastorePoidGenerator && storeMgr != null)
        {
            // Set the store manager and connection provider for any datastore-based generators
            ((AbstractDatastorePoidGenerator)generator).setStoreManager(storeMgr);
            ((AbstractDatastorePoidGenerator)generator).setConnectionProvider(connectionProvider);
        }

        // Store the generator
        generatorsByName.put(name, generator);

        return generator;
    }
}