/**********************************************************************
 Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

 Contributors:
 ...
 **********************************************************************/
package org.jpox.store.query;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.util.Localiser;

/**
 * Abstract representation of a statement that is used for iterating through a list of
 * objects of a class, possibly including subclasses.
 *
 * @version $Revision: 1.11 $
 */
public abstract class AbstractIteratorStatement
{
    /** Localisation or messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");

    /** Whether to include iteration through subclasses of the candidate. */
    protected final boolean includeSubclasses;

    /** Manager for the Store. */
    protected final StoreManager storeMgr;

    /** Datastore adapter */
    protected final DatastoreAdapter dba;
    
    /** full class name for the candidate class **/
    protected final String candidateFullClassName;

    /** Table where the candidate objects are stored. */
    protected DatastoreClass candidateTable;

    /** ClassLoader resolver. */
    protected final ClassLoaderResolver clr;

    /**
     * Constructor.
     * @param type Class that we are querying for as our base.
     * @param clr ClassLoaderResolver
     * @param subclasses Whether to include subclasses in the iteration.
     * @param storeMgr Manager for the store
     */
    public AbstractIteratorStatement(Class type, ClassLoaderResolver clr, boolean subclasses, StoreManager storeMgr)
    {
        this.includeSubclasses = subclasses;
        this.storeMgr = storeMgr;
        this.dba = storeMgr.getDatastoreAdapter();
        this.clr = clr;
        AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(type, clr);
        if( cmd == null )
        {
            candidateFullClassName = type.getName();
        }
        else
        {
            candidateFullClassName = cmd.getFullClassName();
        }

        if (!storeMgr.getPMFContext().getTypeManager().isSupportedType(candidateFullClassName))
        {
            candidateTable = storeMgr.getDatastoreClass(candidateFullClassName, clr);
        }
    }

    /**
     * Accessor for the QueryStatement that will return the objects of the candidate
     * type and its subclasses (if required).
     * @return The QueryExpression
     */
    public abstract QueryExpression getQueryStatement();
}