/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - commented and localised
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - addition of setUnique() and setRange()
2005 Andy Jefferson - added timeout support.
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.exceptions.ClassNotResolvedException;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.QueryNotUniqueException;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.util.Imports;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Abstract implementation of the JDO query interface.
 * Provides a base for all query languages supported by JPOX.
 *
 * @see javax.jdo.Query
 * @version $Revision: 1.59 $
 */
public abstract class Query implements javax.jdo.Query
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    protected final transient PersistenceManager pm;
    protected final transient DatastoreAdapter dba;

    /** The candidate class for this query. */
    protected Class candidateClass;

    /** Name of the candidate class (used when specified via Single-String). */
    protected String candidateClassName;

    /** Whether to allow subclasses of the candidate class be returned. */
    protected boolean subclasses = true;

    /** Whether to return single value, or collection from the query. */
    protected boolean unique = false;

    /** Specification of the result of the query e.g aggregates etc. */
    protected String result = null;    

    /** User-defined class that best represents the results of a query. Populated if specified via setResultClass(). */
    protected Class resultClass = null;

    /** Name of user-defined class to use as the result class. */
    protected String resultClassName = null;

    /** The filter for the query. */
    protected String filter;

    /** Any import declarations for the types used in the query. */
    protected String imports;

    /** Any explicit variables defined for this query. */
    protected String variables;

    /** Any explicit parameters defined for this query. */
    protected String parameters;

    /** Ordering clause for the query, governing the order objects are returned. */
    protected String ordering;

    /** Grouping clause for the query, for use with aggregate expressions. */
    protected String grouping;

    /** String form of the query result range. Only populated if specified via String. */
    protected String range;

    /** Query result range start position (included). Either specified, or compiled from "range". */
    protected long fromInclNo = 0;

    /** Query result range end position (excluded). Either specified, or compiled from "range". */
    protected long toExclNo = Long.MAX_VALUE;

    /** Whether the query can be modified */
    protected boolean unmodifiable = false;

    /** Whether to ignore dirty instances in the query. */
    protected boolean ignoreCache = false;

    /** Fetch Plan to use for the query. */
    protected final FetchPlan fetchPlan;

    /** State variable for the compilation state */
    protected transient boolean isCompiled = false;

    /** The imports definition. */
    protected transient Imports parsedImports = null;

    /** The parameter names. */
    protected transient List parameterNames = null;

    /** Look-up for the parameter types, keyed by the name. */
    protected transient Map parameterTypesByName = null;

    /** The variables names. */
    protected transient List variableNames = null;

    /** Look-up for the variables types, keyed by the name. */
    protected transient Map variableTypesByName = null;

    /**
     * All query results obtained from this query.
     * This is required because the query can be executed multiple times changing
     * the input slightly each time for example.
     */
    protected transient HashSet queryResults = new HashSet();

    /** State variable for whether the query hasnt yet passed the JDO "compile" step (preCompile=true). */
    protected transient boolean isPreCompile = true;

    /** Any JPOX extensions */
    protected Map extensions = null;

    /**
     * Constructs a new query instance that uses the given persistence manager.
     * @param pm The Persistence Manager for this query.
     */
    public Query(PersistenceManager pm)
    {
        this.pm = pm;

        // PM should always be provided so throw exception if null
        if (pm == null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.NoPMSpecified"));
        }

        this.fetchPlan = ((FetchPlanImpl)pm.getFetchPlan()).getCopy();

        // Inherit IgnoreCache from PersistenceManager (JDO 1.0 $12.6.3)
        this.ignoreCache = pm.getIgnoreCache();

        dba = (pm.getStoreManager() == null ? null : pm.getStoreManager().getDatastoreAdapter());
    }

    /**
     * Returns the class loader of the candidade class 
     * @return the ClassLoader for the candidate class
     */
    protected ClassLoader getCandidateLoader()
    {
        return candidateClass == null ? null : candidateClass.getClassLoader();
    }

    /**
     * Utility to remove any previous compilation of this Query.
     **/
    protected void discardCompiled()
    {
        isCompiled = false;
        parsedImports = null;
        parameterNames = null;
        parameterTypesByName = null;
        variableNames = null;
        variableTypesByName = null;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether this and the other object are equal.
     **/
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof Query))
        {
            return false;
        }

        Query q = (Query)obj;

        if (candidateClass == null) 
        {
            if (q.candidateClass != null)
            {
                return false;
            }
        }
        else if (!candidateClass.equals(q.candidateClass))
        {
            return false;
        }

        if (filter == null) 
        {
            if (q.filter != null)
            {
                return false;
            }
        }
        else if (!filter.equals(q.filter))
        {
            return false;
        }

        if (imports == null)
        {
            if (q.imports != null)
            {
                return false;
            }
        }
        else if (!imports.equals(q.imports)) 
        {
            return false;
        }

        if (parameters == null)
        {
            if (q.parameters != null)
            {
                return false;
            }
        }
        else if (!parameters.equals(q.parameters))
        {
            return false;
        }

        if (variables == null)
        {
            if (q.variables != null)
            {
                return false;
            }
        }
        else if (!variables.equals(q.variables))
        {
            return false;
        }

        if (unique != q.unique)
        {
            return false;
        }

        if (unmodifiable != q.unmodifiable)
        {
            return false;
        }

        if (resultClass != q.resultClass)
        {
            return false;
        }

        if (grouping == null)
        {
            if (q.grouping != null)
            {
                return false;
            }
        }
        else if (!grouping.equals(q.grouping))
        {
            return false;
        }

        if (ordering == null)
        {
            if (q.ordering != null)
            {
                return false;
            }
        }
        else if (!ordering.equals(q.ordering))
        {
            return false;
        }

        return true;
    }

    /**
     * Hashcode generator.
     * @return The Hashcode for this object.
     */
    public int hashCode()
    {
        return (candidateClass == null ? 0 : candidateClass.hashCode()) ^
            (result == null ? 0 : result.hashCode()) ^
        	(filter == null ? 0 : filter.hashCode()) ^
        	(imports == null ? 0 : imports.hashCode()) ^
        	(parameters == null ? 0 : parameters.hashCode()) ^
        	(variables == null ? 0 : variables.hashCode()) ^
        	(resultClass == null ? 0 : resultClass.hashCode()) ^
        	(grouping == null ? 0 : grouping.hashCode()) ^
        	(ordering == null ? 0 : ordering.hashCode()) ^
            (range == null ? 0 : range.hashCode());
    }

    /**
     * Accessor for the StoreManager associated with this Query.
     * @return the StoreManager associated with this Query.
     */
    public StoreManager getStoreManager()
    {
        return pm.getStoreManager();
    }

    /**
     * Accessor for the PersistenceManager associated with this Query.
     * @return the PersistenceManager associated with this Query.
     * @see javax.jdo.Query#getPersistenceManager
     */
    public javax.jdo.PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * Add a vendor-specific extension this query. The key and value
     * are not standard.
     * An implementation must ignore keys that are not recognized.
     * @param key the extension key
     * @param value the extension value
     * @since JDO 2.0
     */
    public void addExtension(String key, Object value)
    {
        if (extensions == null)
        {
            extensions = new HashMap();
        }
        extensions.put(key, value);
    }

    /**
     * Set multiple extensions, or use null to clear extensions.
     * Map keys and values are not standard.
     * An implementation must ignore entries that are not recognized.
     * @param extensions
     * @see #addExtension
     * @since 1.1
     */
    public void setExtensions(Map extensions)
    {
        extensions = new HashMap(extensions);
    }

    /**
     * Accessor for the value of an extension for this query.
     * @param key The key
     * @return The value (if this extension is specified)
     */
    public Object getExtension(String key)
    {
        return (extensions != null ? extensions.get(key) : null);
    }

    /**
     * This method retrieves the fetch plan associated with the Query. It always
     * returns the identical instance for the same Query instance. Any change
     * made to the fetch plan affects subsequent query execution. Fetch plan is
     * described in Section 12.7
     * @return the FetchPlan
     */
    public FetchPlan getFetchPlan()
    {
        return fetchPlan;
    }
    
    /**
     * Accessor for the class of the candidate instances of the query.
     * @return the Class of the candidate instances.
     * @see javax.jdo.Query#setClass
     */
    public Class getCandidateClass()
    {
        return candidateClass;
    }

    /**
     * Mutator for the class of the candidate instances of the query.
     * @param candidateClass the Class of the candidate instances.
     * @see javax.jdo.Query#setClass
     */
    public void setClass(Class candidateClass)
    {
        discardCompiled();
        assertIsModifiable();

        this.candidateClassName = (candidateClass != null ? candidateClass.getName() : null);
        this.candidateClass = candidateClass;
    }

    /**
     * Convenience method to set the name of the candidate class.
     * @param candidateClassName Name of the candidate class
     */
    public void setCandidateClassName(String candidateClassName)
    {
        this.candidateClassName = (candidateClassName != null ? candidateClassName.trim() : null);
    }

    /**
     * Set the candidate Extent to query. To be implemented by extensions.
     * @param pcs the Candidate Extent.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public abstract void setCandidates(Extent pcs);

    /**
     * Set the candidate Collection to query. To be implemented by extensions.
     * @param pcs the Candidate collection.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public abstract void setCandidates(Collection pcs);

    /**
     * Set the filter for the query.
     * @param filter the query filter.
     * @see javax.jdo.Query#setFilter
     */
    public void setFilter(String filter)
    {
        discardCompiled();
        assertIsModifiable();

        this.filter = (filter != null ? filter.trim() : null);
    }

    /**
     * Set the import statements to be used to identify the fully qualified name
     * of variables or parameters.
     * @param imports import statements separated by semicolons.
     * @see javax.jdo.Query#declareImports
     */
    public void declareImports(String imports)
    {
        discardCompiled();
        assertIsModifiable();

        this.imports = (imports != null ? imports.trim() : null);
    }

    /**
     * Declare the list of parameters query execution.
     * @param parameters the list of parameters separated by commas.
     * @see javax.jdo.Query#declareParameters
     */
    public void declareParameters(String parameters)
    {
        discardCompiled();
        assertIsModifiable();

        this.parameters = (parameters != null ? parameters.trim() : null);
    }

    /**
     * Declare the unbound variables to be used in the query.
     * @param variables the variables separated by semicolons.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareVariables(String variables)
    {
        discardCompiled();
        assertIsModifiable();

        this.variables = (variables != null ? variables.trim() : null);
    }

    /**
     * Set the ordering specification for the result Collection.
     * @param ordering the ordering specification.
     * @see javax.jdo.Query#setOrdering
     */
    public void setOrdering(String ordering)
    {
        discardCompiled();
        assertIsModifiable();

        this.ordering = (ordering != null ? ordering.trim() : null);
    }

    /**
     * Set the grouping specification for the result Collection.
     * @param grouping the grouping specification.
     * @see javax.jdo.Query#setGrouping
     */
    public void setGrouping(String grouping)
    {
        discardCompiled();
        assertIsModifiable();

        this.grouping = (grouping != null ? grouping.trim() : null);
    }

    /**
     * Set the uniqueness of the results. A value of true will return a single
     * value (or null) where the application knows that there are 0 or 1
     * objects to be returned. See JDO 2.0 specification section 14.6
     * @param unique whether the result is unique
     * @see javax.jdo.Query#setUnique
     * @since 1.1 
     */
    public void setUnique(boolean unique)
    {
        discardCompiled();
        assertIsModifiable();

        this.unique = unique;
    }

    /**
     * Set the range of the results. By default all results are returned but
     * this allows specification of a range of elements required. See JDO 2.0
     * specification section 14.6.8
     * @param fromIncl From element no (inclusive) to return
     * @param toExcl To element no (exclusive) to return
     * @see javax.jdo.Query#setRange(long, long)
     * @since 1.1
     */
    public void setRange(long fromIncl, long toExcl)
    {
        discardCompiled();

        // JDO2 spec 14.6 setRange is valid when unmodifiable so dont check it
        this.range = null;
        this.fromInclNo = fromIncl;
        this.toExclNo = toExcl;
    }

    /**
     * Set the range of the results. By default all results are returned but
     * this allows specification of a range of elements required. See JDO 2.0
     * specification section 14.6.8
     * @param range Range string
     * @see javax.jdo.Query#setRange(java.lang.String)
     * @since 1.1
     */
    public void setRange(String range)
    {
        discardCompiled();

        // JDO2 spec 14.6 setRange is valid when unmodifiable so dont check it
        // fromInclNo, toExclNo will be extracted when compiling
        this.range = (range != null ? range.trim() : null);
        fromInclNo = 0;
        toExclNo = Long.MAX_VALUE;
    }

    /**
     * Set the result for the results. The application might want to get results
     * from a query that are not instances of the candidate class. The results
     * might be fields of persistent instances, instances of classes other than
     * the candidate class, or aggregates of fields.
     * @param result The result parameter consists of the optional keyword
     * distinct followed by a commaseparated list of named result expressions or
     * a result class specification.
     * @see javax.jdo.Query#setResult
     * @since 1.1
     */
    public void setResult(String result)
    {
        discardCompiled();
        assertIsModifiable();

        this.result = (result != null ? result.trim() : null);
    }

    /**
     * Set the result class for the results.
     * The result class must obey various things as per the JDO 2.0 spec 14.6.12.
     * @param result_cls The result class
     * @see javax.jdo.Query#setResultClass
     * @since 1.1
     */
    public void setResultClass(Class result_cls)
    {
        discardCompiled();

        // JDO2 spec 14.6 setResultClass is valid when unmodifiable so dont check it
        this.resultClassName = (result_cls != null ? result_cls.getName() : null);
        this.resultClass = result_cls;
    }

    /**
     * Convenience method to set the name of the result class.
     * @param resultClassName Name of the result class
     */
    public void setResultClassName(String resultClassName)
    {
        this.resultClassName = resultClassName;
    }

    /**
     * Set the ignoreCache option. Currently this simply stores the ignoreCache
     * value, and doesn't necessarily use it. The parameter is a "hint" to the
     * query engine. TODO : Implement this fully.
     * @param ignoreCache the setting of the ignoreCache option.
     * @see javax.jdo.Query#setIgnoreCache
     * @see javax.jdo.PersistenceManager#setIgnoreCache
     */
    public void setIgnoreCache(boolean ignoreCache)
    {
        discardCompiled();

        // JDO2 spec 14.6 setIgnoreCache is valid when unmodifiable so dont check it
        this.ignoreCache = ignoreCache;
    }

    /**
     * Accessor for the ignoreCache option setting.
     * @return the ignoreCache option setting
     * @see #setIgnoreCache
     * @see javax.jdo.Query#getIgnoreCache
     * @see javax.jdo.PersistenceManager#getIgnoreCache
     */
    public boolean getIgnoreCache()
    {
        return ignoreCache;
    }

    /**
     * Accessor for whether this query includes subclasses
     * @return Returns whether the query includes subclasses.
     */
    public boolean isSubclasses()
    {
        return subclasses;
    }

    /**
     * Mutator for whether this query includes subclasses
     * @param subclasses Where subclasses of the candidate class are to be included.
     */
    public void setSubclasses(boolean subclasses)
    {
        discardCompiled();
        assertIsModifiable();

        this.subclasses = subclasses;

        // Set the candidates also if the candidate has been set
        /*if (candidateClass != null)
        {
            setCandidates(pm.getExtent(candidateClass, subclasses));
        }*/
    }

    /**
     * Accessor for unmodifiable.
     * @return Returns the unmodifiable.
     */
    public boolean isUnmodifiable()
    {
        return unmodifiable;
    }

    /**
     * Method to throw a JDOUserException if the query is currently
     * not modifiable.
     * @throws JDOUserException Thrown when it is unmodifiable
     */
    protected void assertIsModifiable()
    {
        if (unmodifiable)
        {
            throw new JDOUserException(LOCALISER.msg("Query.Unmodifiable"));
        }
    }

    /**
     * Mutator for unmodifiable.
     */
    public void setUnmodifiable()
    {
        this.unmodifiable = true;
    }

    /**
     * Verify the elements of the query and provide a hint to the query to prepare and optimize an execution plan.
     * Processes the imports, explicit parameters, and explicit variables.
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        if (isCompiled)
        {
            return;
        }
        if (pm == null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.NoPersistenceManagerError"));
        }

        boolean done = false;
        try
        {
            // Parse imports list
            parsedImports = new Imports();
            if (candidateClassName != null)
            {
                parsedImports.importPackage(candidateClassName);
            }
            if (imports != null)
            {
                parsedImports.parseImports(imports);
            }

            // Parse explicit parameters list
            parameterNames = new ArrayList();
            parameterTypesByName = new HashMap();
            if (parameters != null && parameters.length() > 0)
            {
                // Explicit parameters defined, so validate them
                StringTokenizer t1 = new StringTokenizer(parameters, ",");
                while (t1.hasMoreTokens())
                {
                    StringTokenizer t2 = new StringTokenizer(t1.nextToken(), " ");
                    if (t2.countTokens() != 2)
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.InvalidParameterListError",parameters));
                    }

                    String classDecl = t2.nextToken();
                    String parameterName = t2.nextToken();
                    if (!StringUtils.isValidJavaIdentifierForJDOQL(parameterName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.IllegalParameterNameError",parameterName));
                    }

                    if (parameterNames.contains(parameterName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.DuplicateParameterNameError",parameterName));
                    }

                    parameterNames.add(parameterName);
                    parameterTypesByName.put(parameterName, resolveClassDeclaration(classDecl));
                }
            }

            // Parse explicit variables list
            variableNames = new ArrayList();
            variableTypesByName = new HashMap();
            if (variables != null && variables.length() > 0)
            {
                // Explicit variables defined, so validate them
                StringTokenizer t1 = new StringTokenizer(variables, ";");
                while (t1.hasMoreTokens())
                {
                    StringTokenizer t2 = new StringTokenizer(t1.nextToken(), " ");
                    if (t2.countTokens() != 2)
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.InvalidVariableListError",variables));
                    }

                    String classDecl = t2.nextToken();
                    String variableName = t2.nextToken();
                    if (!StringUtils.isValidJavaIdentifierForJDOQL(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.IllegalVariableNameError",variableName));
                    }

                    if (parameterNames.contains(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.ConflictingVariableNameError",variableName));
                    }

                    if (variableNames.contains(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.DuplicateVariableNameError",variableName));
                    }

                    variableNames.add(variableName);
                    variableTypesByName.put(variableName, resolveClassDeclaration(classDecl));
                }
            }

            done = true;
        }
        finally
        {
            if (!done)
            {
                discardCompiled();
            }
        }
    }

    /**
     * Utility to resolve the declaration to a particular class.
     * Takes the passed in name, together with the defined import declarations and returns the
     * class represented by the declaration.
     * @param classDecl The declaration
     * @return The class it resolves to (if any)
     * @throws JDOUserException Thrown if the class cannot be resolved.
     */
    public Class resolveClassDeclaration(String classDecl)
    {
        try
        {
            return parsedImports.resolveClassDeclaration(classDecl, pm.getClassLoaderResolver(), getCandidateLoader());
        }
        catch (ClassNotResolvedException e)
        {
            throw new JDOUserException(LOCALISER.msg("Query.ClassNameNotFoundError", classDecl));
        }
    }

    // ----------------------------- Execute -----------------------------------

    /**
     * Execute the query and return the filtered List.
     * @return the filtered List.
     * @see javax.jdo.Query#execute()
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute()
    {
        return executeWithArray(new Object[0]);
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1)
    {
        return executeWithArray(new Object[] { p1 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object,Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1, Object p2)
    {
        return executeWithArray(new Object[] { p1, p2 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     * @param p3 the value of the third parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object,Object,Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1, Object p2, Object p3)
    {
        return executeWithArray(new Object[] { p1, p2, p3 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param parameterValues the Object array with all of the parameters.
     * @return the filtered List.
     * @see javax.jdo.Query#executeWithArray(Object[])
     */
    public Object executeWithArray(Object[] parameterValues)
    {
        // Compile the imports, explicit parameters, explicit variables
        isPreCompile = false;
        compile();

        // Build the parameter values into a Map so we have a common interface
        HashMap parameterMap = new HashMap();
        if (parameterNames.size() > 0)
        {
            // Explicit parameters defined
            for (int i = 0; i < parameterValues.length; ++i)
            {
                parameterMap.put(parameterNames.get(i), parameterValues[i]);
            }
        }
        else
        {
            // Must be implicit parameters, so give them dummy names for now
            for (int i = 0; i < parameterValues.length; ++i)
            {
                parameterMap.put("JPOX_" + i, parameterValues[i]);
            }
        }

        Object theReturn = executeWithMap(parameterMap);

        isPreCompile = true;
        return theReturn;
    }

    /**
     * Convenience method to return whether the query should return a single row.
     * @return Whether a single row should result
     */
    protected abstract boolean shouldReturnSingleRow();

    /**
     * Execute the query and return the filtered result(s).
     * @param parameters the Map containing all of the parameters.
     * @return the filtered results (List, or Object)
     * @see javax.jdo.Query#executeWithMap(Map)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object executeWithMap(Map parameters)
    {
        // JDO 2 $14.6.1 : If PM is closed, throw JDOFatalUserException
        if (pm.isClosed())
        {
            throw new JDOFatalUserException(LOCALISER.msg("Query.PMIsClosed"));
        }

        Collection qr = performExecute(parameters);

        // Gather information about whether we should apply range limits in this post-processing step
        RDBMSAdapter rdba = (RDBMSAdapter)dba; // TODO Abstract this into RDBMSQuery since is RDBMS-specific
        boolean range_specified = (fromInclNo >= 0 && toExclNo >= 0 && (fromInclNo != 0 || toExclNo != Long.MAX_VALUE));
        boolean using_limit_select_clause = (rdba.getRangeByLimitSelectClause(fromInclNo, toExclNo).length() > 0);
        boolean using_limit_where_clause = (rdba.getRangeByLimitWhereClause(fromInclNo, toExclNo).length() > 0);
        boolean using_rownum = (rdba.getRangeByRowNumberColumn().length() > 0);
        boolean applyRangeChecks = (range_specified && !using_limit_select_clause && !using_limit_where_clause && !using_rownum);

        if (shouldReturnSingleRow())
        {
            try
            {
                // Single row only needed (unique specified, or using aggregates etc), so just take the first row of the results
                if (qr == null || qr.size() == 0)
                {
                    return null;
                }
                else
                {
                    if (applyRangeChecks && (toExclNo - fromInclNo <= 0))
                    {
                        // JDO2 spec 14.6.8 (range excludes instances, so return null)
                        return null;
                    }

                    Iterator qrIter = qr.iterator();
                    Object firstRow = qrIter.next();
                    if (qrIter.hasNext())
                    {
                        throw new QueryNotUniqueException();
                    }
                    return firstRow;
                }
            }
            finally 
            {
                // can close qr right now because we don't return it,
                // also user cannot close it otherwise except for calling
                // closeAll()
                if (qr!=null) {
                    close(qr);
                }
            }
        }

        // Process any specified range
        // Use LIMIT where supported by the RDBMS, and otherwise just throw away ones outside the range
        if (applyRangeChecks)
        {
            int i=0;
            Iterator qr_iter=qr.iterator();
            Collection res=new ArrayList();
            while (qr_iter.hasNext())
            {
                Object obj=qr_iter.next();
                if (i >= fromInclNo && i < toExclNo)
                {
                    // Accept the item if within range
                    res.add(obj);
                }
                i++;
            }
            return res;
        }
        else
        {
            return qr;
        }
    }

    /**
     * Method to actually execute the query. To be implemented by extending
     * classes for the particular query language.
     * @param parameters Map containing the parameters.
     * @return The filtered QueryResult.
     */
    protected abstract Collection performExecute(Map parameters);

    // ------------------------- Delete Persistent -----------------------------

    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @return The number of deleted objects.
     * @see javax.jdo.Query#deletePersistentAll()
     */
    public long deletePersistentAll()
    {
        return deletePersistentAll(new Object[0]);
    }

    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @param parameters the Object array with all of the parameters.
     * @return the filtered Collection.
     * @see javax.jdo.Query#deletePersistentAll(Object[])
     */
    public long deletePersistentAll(Object[] parameters)
    {
        isPreCompile = false;
        compile();

        if (parameters.length != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER.msg("Query.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.length));
        }

        HashMap parameterMap = new HashMap();
        for (int i=0; i<parameters.length; ++i)
        {
            parameterMap.put(parameterNames.get(i), parameters[i]);
        }

        long theReturn = deletePersistentAll(parameterMap);
        isPreCompile = true;
        return theReturn;
    }

    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @param parameters Map of parameters for the query
     * @return the number of deleted objects
     * @see javax.jdo.Query#deletePersistentAll(Map)
     */
    public long deletePersistentAll(Map parameters)
    {
        // Check for specification of any illegal attributes
        if (result != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultSpecifiedError"));
        }
        if (resultClass != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultClassSpecifiedError"));
        }
        if (resultClassName != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultClassSpecifiedError"));
        }
        if (ordering != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.OrderingSpecifiedError"));
        }
        if (grouping != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.GroupingSpecifiedError"));
        }
        if (range != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultRangeSpecifiedError"));
        }
        if (fromInclNo >= 0 && toExclNo >= 0 &&
            (fromInclNo != 0 || toExclNo != Long.MAX_VALUE))
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultRangeSpecifiedError"));
        }

        long qr = performDeletePersistentAll(parameters);

        return qr;
    }

    /**
     * Method to actually execute the deletion of objects.
     * To be implemented by extending classes.
     * @param parameters Map containing the parameters.
     * @return The filtered QueryResult.
     */
    protected abstract long performDeletePersistentAll(Map parameters);

    // -------------------------------------------------------------------------

    /**
     * Close a query result and release any resources associated with it.
     * @param queryResult the result of execute(...) on this Query instance.
     * @see javax.jdo.Query#close
     */
    public void close(Object queryResult)
    {
        if (queryResult != null && queryResult instanceof QueryResult)
        {
            ((QueryResult)queryResult).close();
            queryResults.remove(queryResult);
        }
    }

    /**
     * Close all query results associated with this Query instance,
     * and release all resources associated with them.
     * @see javax.jdo.Query#closeAll
     */
    public void closeAll()
    {
        QueryResult[] qrs = (QueryResult[])queryResults.toArray(new QueryResult[queryResults.size()]);

        for (int i = 0; i < qrs.length; ++i)
        {
            close(qrs[i]);
        }
    }
}