/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.sql.DataSource;
import javax.transaction.Synchronization;

import org.jpox.exceptions.TransactionCommiting;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.util.Localiser;

/**
 * Abstract representation of a Transaction for RDBMS datastores.
 *
 * @version $Revision: 1.2 $
 **/ 
public abstract class AbstractRDBMSTransaction implements RDBMSTransaction
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /** Adapter for this datastore */
    protected RDBMSAdapter dba;

    /** Transactional data source(s) */
    protected DataSource[] tds;

    /** Whether the transaction is active. */
    protected boolean active = false;

    /** Whether the transaction is in the process of commiting or rollingback. */
    protected boolean commiting = false;

    protected boolean retainValues;
    protected boolean restoreValues;
    protected boolean optimistic;
    protected boolean nontransactionalRead;
    protected boolean nontransactionalWrite;

    /** Synchronisation for this transaction. */
    protected Synchronization sync = null;

    /** Whether the transaction is marked for rollback. JDO 2.0 section 13.4.5 */
    protected boolean rollbackOnly = false;

    /** Isolation level for this txn. JPOX extension. */
    protected int isolationLevel;

    /** Whether we lock objects for update. JPOX extension. */
    protected boolean useUpdateLock;

    /** Whether the JDOConnection is in use currently. */
    protected boolean isJDOConnectionInUse = false;

    /** Connection to use. */
    protected Connection conn = null;

    /**
     * Checks whether a connection to the datasource is open.
     * @return true if the Connection is open
     */
    public synchronized boolean isConnectionOpen()
    {
        return conn != null;
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param forQuery Whether the connection is to be used by a query (i.e.JDOQUERY).
     * @return The Connection
     * @throws SQLException
     */
    public synchronized Connection getConnection(boolean forWriting, boolean forQuery)
    throws SQLException
    {
        return getConnection(forWriting, forQuery, false);
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param forQuery Whether the connection is to be used by a query (i.e.JDOQUERY).
     * @param useOpenedConnection whether to use an already opened connection, if one is opened. 
     * @return The Connection
     * @throws SQLException
     */
    public abstract Connection getConnection(boolean forWriting, boolean forQuery, boolean useOpenedConnection)
    throws SQLException;

    /**
     * Method to release the Connection.
     * @param conn The Connection to release.
     * @throws SQLException Thrown when an error occurs releasing the
     * connection.
     **/    
    public abstract void releaseConnection(Connection conn)
    throws SQLException;

    /**
     * Mutator for the setting of retain values.
     * @param retainValues Whether to retain values at commit
     */
    public synchronized void setRetainValues(boolean retainValues)
    {
        assertNotCommiting();

        this.retainValues = retainValues;

        if (retainValues)
        {
            nontransactionalRead = true;
        }
    }

    /**
     * Accessor for the retainValues flag for this transaction.
     * @return Whether retainValues is set.
     */
    public boolean getRetainValues()
    {
        return retainValues;
    }

    /**
     * Mutator for the setting of restore values.
     * @param restoreValues Whether to restore values at commit
     */
    public synchronized void setRestoreValues(boolean restoreValues)
    {
        assertNotInUse();
        assertNotCommiting();

        this.restoreValues = restoreValues;
    }

    /**
     * Accessor for the restoreValues flag for this transaction.
     * @return Whether restoreValues is set.
     */
    public boolean getRestoreValues()
    {
        return this.restoreValues;
    }

    /**
     * Mutator for the optimistic transaction setting.
     * @param optimistic The optimistic transaction setting.
     **/
    public synchronized void setOptimistic(boolean optimistic)
    {
        assertNotInUse();
        assertNotCommiting();

        this.optimistic = optimistic;
    }

    /**
     * Accessor for the Optimistic setting
     * @return Whether optimistic transactions are in operation.
     **/
    public boolean getOptimistic()
    {
        return optimistic;
    }

    /**
     * Mutator for the setting of nontransactional read.
     * @param nontransactionalRead Whether to allow nontransactional read operations
     */
    public synchronized void setNontransactionalRead(boolean nontransactionalRead)
    {
        assertNotCommiting();

        this.nontransactionalRead = nontransactionalRead;
    }

    /**
     * Accessor for the nontransactionalRead flag for this transaction.
     * @return Whether nontransactionalRead is set.
     */
    public boolean getNontransactionalRead()
    {
        return nontransactionalRead;
    }

    /**
     * Mutator for the setting of nontransactional write.
     * @param nontransactionalWrite Whether to allow nontransactional write operations
     */
    public synchronized void setNontransactionalWrite(boolean nontransactionalWrite)
    {
        assertNotInUse();
        assertNotCommiting();

        if (nontransactionalWrite)
        {
            throw new JDOUnsupportedOptionException("Non-transactional write mode not (yet) supported");
        }

        this.nontransactionalWrite = nontransactionalWrite;
    }

    /**
     * Accessor for the nontransactionalWrite flag for this transaction.
     * @return Whether nontransactionalWrite is set.
     */
    public boolean getNontransactionalWrite()
    {
        return nontransactionalWrite;
    }

    /**
     * Mutator for the synchronization object to be notified on transaction
     * completion.
     * @param sync The synchronization object to be notified on transaction
     * completion
     **/
    public synchronized void setSynchronization(Synchronization sync)
    {
        assertNotCommiting();
        
        this.sync = sync;
    }

    /**
     * Accessor for the synchronization object to be notified on transaction
     * completion.
     * @return The synchronization instance ot be notified on transaction
     * completion.
     **/
    public Synchronization getSynchronization()
    {
        return sync;
    }

    /**
     * Accessor for the "rollback only" flag.
     * @return The rollback only flag
     * @since 1.1
     */
    public boolean getRollbackOnly()
    {
        return rollbackOnly;
    }

    /**
     * Mutator for the "rollback only" flag. Sets the transaction as for
     * rollback only.
     * @since 1.1
     */
    public void setRollbackOnly()
    {
        // Only apply to active transactions
        if (active)
        {
            rollbackOnly = true;
        }
    }

    /**
     * Accessor for the transaction isolation level.
     * @return The transaction isolation level.
     **/
    public int getTransactionIsolation()
    {
        return isolationLevel;
    }

    /**
     * Mutator for the transaction isolation level.
     * @param isolationLevel The isolation level required.
     **/
    public void setTransactionIsolation(int isolationLevel)
    {
        assertNotInUse();
        assertNotCommiting();

        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("Transaction.IsolationLevelInvalid", isolationLevel));
        }

        this.isolationLevel = isolationLevel;
    }

    /**
     * Accessor for whether to use an update lock when fetching data.
     * Will return false if isolation levels of TRANSACTION_REPEATABLE_READ
     * or TRANSACTION_SERIALIZABLE are selected. Otherwise returns the local
     * settings of the useUpdateLock flag.
     * @return Whether to use an update lock.
     **/
    public boolean useUpdateLockOnFetch()
    {
        // If we aren't active, or if doing optimistic transaction then don't lock
        if (!active || optimistic)
        {
            return false;
        }
        else
        {
            if (isolationLevel == Connection.TRANSACTION_REPEATABLE_READ ||
                isolationLevel == Connection.TRANSACTION_SERIALIZABLE)
            {
                return false;
            }
            else
            {
                // TODO Extend the code here to allow multiple locking strategies
                // so that we can just lock the first query in a transaction for example
                return useUpdateLock;
            }
        }
    }

    /**
     * Mutator for whether to lock queried/fetched objects in this txn.
     * @param lock Whether to use an update lock.
     **/
    public void setUseUpdateLock(boolean lock)
    {
        assertNotInUse();
        assertNotCommiting();

        this.useUpdateLock = lock;
    }

    /**
     * Accessor for whether the transaction is active.
     * @return Whether the transaction is active.
     **/
    public boolean isActive()
    {
        return active;
    }

    /**
     * Accessor for whether the transaction is comitting.
     * @return Whether the transaction is committing.
     **/
    public boolean isCommitting()
    {
        return commiting;
    }

    /**
     * Asserts that the transaction is not in use.
     **/
    protected abstract void assertNotInUse();

    /**
     * Asserts that the transaction is not commiting or rolling back.
     **/
    protected void assertNotCommiting()
    {
        if (commiting)
        {
            throw new TransactionCommiting(this);
        }
    }

    /**
     * Asserts that's the connection is not hold by the developer's application
     * @throws JDOUserException if hold by developer
     */
    protected void assertJDOConnectionNotInUse()
    {
        if (isJDOConnectionInUse)
        {
            // TODO Localise this message
            throw new JDOUserException("The JDOConnection was acquired by the developer and must be closed before using the JDO API.");
        }
    }
}