/**********************************************************************
Copyright (c) 2006 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.Types;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import org.jpox.store.DatastoreAdapter;
import org.jpox.util.StringUtils;

/**
 * Convenience helper for JDBC.
 * 
 * @version $Revision: 1.11 $
 */
public class JDBCUtils
{
    private static Map supportedJdbcTypesById = new HashMap();
    private static Map unsupportedJdbcTypesById = new HashMap();

    static
    {
        // Add the supported types
        supportedJdbcTypesById.put(new Integer(Types.BIGINT), "BIGINT");
        supportedJdbcTypesById.put(new Integer(Types.BIT), "BIT");
        supportedJdbcTypesById.put(new Integer(Types.BLOB), "BLOB");
        supportedJdbcTypesById.put(new Integer(Types.BOOLEAN), "BOOLEAN");
        supportedJdbcTypesById.put(new Integer(Types.CHAR), "CHAR");
        supportedJdbcTypesById.put(new Integer(Types.CLOB), "CLOB");
        supportedJdbcTypesById.put(new Integer(Types.DATALINK), "DATALINK");
        supportedJdbcTypesById.put(new Integer(Types.DATE), "DATE");
        supportedJdbcTypesById.put(new Integer(Types.DECIMAL), "DECIMAL");
        supportedJdbcTypesById.put(new Integer(Types.DOUBLE), "DOUBLE");
        supportedJdbcTypesById.put(new Integer(Types.FLOAT), "FLOAT");
        supportedJdbcTypesById.put(new Integer(Types.INTEGER), "INTEGER");
        supportedJdbcTypesById.put(new Integer(Types.LONGVARBINARY), "LONGVARBINARY");
        supportedJdbcTypesById.put(new Integer(Types.LONGVARCHAR), "LONGVARCHAR");
        supportedJdbcTypesById.put(new Integer(Types.NUMERIC), "NUMERIC");
        supportedJdbcTypesById.put(new Integer(Types.REAL), "REAL");
        supportedJdbcTypesById.put(new Integer(Types.SMALLINT), "SMALLINT");
        supportedJdbcTypesById.put(new Integer(Types.TIME), "TIME");
        supportedJdbcTypesById.put(new Integer(Types.TIMESTAMP), "TIMESTAMP");
        supportedJdbcTypesById.put(new Integer(Types.TINYINT), "TINYINT");
        supportedJdbcTypesById.put(new Integer(Types.VARBINARY), "VARBINARY");
        supportedJdbcTypesById.put(new Integer(Types.VARCHAR), "VARCHAR");

        // Add the unsupported types
        unsupportedJdbcTypesById.put(new Integer(Types.ARRAY), "ARRAY");
        unsupportedJdbcTypesById.put(new Integer(Types.BINARY), "BINARY");
        unsupportedJdbcTypesById.put(new Integer(Types.DISTINCT), "DISTINCT");
        unsupportedJdbcTypesById.put(new Integer(Types.JAVA_OBJECT), "JAVA_OBJECT");
        unsupportedJdbcTypesById.put(new Integer(Types.NULL), "NULL");
        unsupportedJdbcTypesById.put(new Integer(Types.OTHER), "OTHER");
        unsupportedJdbcTypesById.put(new Integer(Types.REF), "REF");
        unsupportedJdbcTypesById.put(new Integer(Types.STRUCT), "STRUCT");
    }
                        
    /**
     * Method to return whether the passed "jdbc-type" is a valid JDBC type for use in JPOX.
     * @param typeName "jdbc-type" name
     * @return Whether it is valid
     */
    public static boolean isValidJDBCType(String typeName)
    {
        if (typeName == null)
        {
            return false;
        }

        Iterator namesIter = supportedJdbcTypesById.values().iterator();
        while (namesIter.hasNext())
        {
            String name = (String)namesIter.next();
            if (name.equalsIgnoreCase(typeName))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the JDBC types.
     * @return JDBC types.
     */
    public static final int[] getJDBCTypes()
    {
        int[] types = new int[supportedJdbcTypesById.size() + unsupportedJdbcTypesById.size()];
        int i = 0;
        Iterator suppIter = supportedJdbcTypesById.keySet().iterator();
        while (suppIter.hasNext())
        {
            types[i++] = ((Integer)suppIter.next()).intValue();
        }
        Iterator unsuppIter = unsupportedJdbcTypesById.keySet().iterator();
        while (unsuppIter.hasNext())
        {
            types[i++] = ((Integer)unsuppIter.next()).intValue();
        }
        return types;
    }

    /**
     * Accessor for a string name of a JDBC Type
     * @param jdbcType The JDBC Type
     * @return The name
     */
    public static String getNameForJDBCType(int jdbcType)
    {
        String typeName = (String)supportedJdbcTypesById.get(new Integer(jdbcType));
        if (typeName == null)
        {
            typeName = (String)unsupportedJdbcTypesById.get(new Integer(jdbcType));
        }
        return typeName;
    }

    /**
     * Accessor for a string name of a JDBC isolation level
     * @param isolation The isolation level
     * @return The name
     */
    public static String getNameForJDBCIsolationLevel(int isolation)
    {
        if (isolation == Connection.TRANSACTION_NONE)
        {
            return "NONE";
        }
        else if (isolation == Connection.TRANSACTION_READ_COMMITTED)
        {
            return "READ_COMMITTED";
        }
        else if (isolation == Connection.TRANSACTION_READ_UNCOMMITTED)
        {
            return "READ_UNCOMMITTED";
        }
        else if (isolation == Connection.TRANSACTION_REPEATABLE_READ)
        {
            return "REPEATABLE_READ";
        }
        else if (isolation == Connection.TRANSACTION_SERIALIZABLE)
        {
            return "SERIALIZABLE";
        }
        else
        {
            return "UNKNOWN";
        }
    }

    /**
     * Convenience method to convert the supplied isolation level name into the
     * associated JDBC Connection type number.
     * @param isolationName The name of the isolation level
     * @return Isolation level type
     */
    public static int getJDBCIsolationLevelForName(String isolationName)
    {
        if (isolationName.equalsIgnoreCase("NONE"))
        {
            return Connection.TRANSACTION_NONE;
        }
        else if (isolationName.equalsIgnoreCase("READ_COMMITTED"))
        {
            return Connection.TRANSACTION_READ_COMMITTED;
        }
        else if (isolationName.equalsIgnoreCase("READ_UNCOMMITTED"))
        {
            return Connection.TRANSACTION_READ_UNCOMMITTED;
        }
        else if (isolationName.equalsIgnoreCase("REPEATABLE_READ"))
        {
            return Connection.TRANSACTION_REPEATABLE_READ;
        }
        else if (isolationName.equalsIgnoreCase("SERIALIZABLE"))
        {
            return Connection.TRANSACTION_SERIALIZABLE;
        }
        else
        {
            return -1;
        }
    }

    /**
     * Utility to return an identifier name without any leading/trailing "quote" characters.
     * @param name The original name
     * @param dba The database adapter
     * @return The stripped name with no quotes
     */
    public static String getIdentifierNameStripped(String name, DatastoreAdapter dba)
    {
        if (name == null)
        {
            return null;
        }
        if (dba.getIdentifierQuoteString() != null)
        {
            // Just remove all quote strings
            return StringUtils.replaceAll(name, dba.getIdentifierQuoteString(), "");
        }
        return name;
    }

    /**
     * Method to return the "subprotocol" for a JDBC URL.
     * A JDBC URL is made up of
     * "jdbc:{subprotocol}:...". For example, "jdbc:mysql:..." or "jdbc:hsqldb:...".
     * @param url The JDBC URL
     * @return The subprotocol
     */
    public static String getSubprotocolForURL(String url)
    {
        StringTokenizer tokeniser = new StringTokenizer(url, ":");
        tokeniser.nextToken();
        return tokeniser.nextToken();
    }
}