/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Savepoint;
import java.sql.Statement;
import java.util.Map;

import javax.jdo.JDOUserException;
import javax.jdo.datastore.JDOConnection;

import org.jpox.util.Localiser;

/**
 * Implementation of a JDO connection for RDBMS datastores.
 * Provides a wrapper to a java.sql.Connection in this implementation.
 * 
 * @version $Revision: 1.1 $
 */
final public class JDOConnectionImpl implements Connection, JDOConnection
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** The JDBC connection **/
    private final Connection conn;

    /** run onClose on close call */
    private final Runnable onClose;

    /** whether this connection is available to the developer */
    private boolean isAvailable = true;

    /**
     * Constructor.
     * @param conn the connection
     * @param onClose the Runnable operation that will be called at user call to close()
     */
    public JDOConnectionImpl(Connection conn, Runnable onClose)
    {
        this.conn = conn;
        this.onClose = onClose;
    }

    /**
     * Access the holdability of the connection.
     * @return The holdability
     * @throws SQLException if there is a problem
     */
    public int getHoldability() throws SQLException
    {
        assertAvailable();
        return this.conn.getHoldability();
    }

    /**
     * Accessor for the transaction isolation level.
     * @return The isolation level
     * @throws SQLException if there is a problem
     */
    public int getTransactionIsolation() throws SQLException
    {
        assertAvailable();
        return this.conn.getTransactionIsolation();
    }

    /**
     * Method to clear warnings from the connection.
     * @throws SQLException thrown if there is a problem
     */
    public void clearWarnings() throws SQLException
    {
        assertAvailable();
        this.conn.clearWarnings();
    }

    /**
     * Method to close the connection.
     */
    public void close()
    {
        assertAvailable();
        isAvailable = false;
        onClose.run();
    }

    /**
     * Method to commit the connection.
     * @throws SQLException thrown if an error occurs.
     */
    public void commit() throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "commit"));	
    }

    /**
     * Method to rollback the connection.
     * @throws SQLException thrown if an error occurs.
     */
    public void rollback() throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "rollback"));
    }

    /**
     * Accessor for the autocommit setting.
     * @return Whether to autocommit
     * @throws SQLException thrown if an error occurs.
     */
    public boolean getAutoCommit() throws SQLException
    {
        assertAvailable();
        return this.conn.getAutoCommit();
    }

    /**
     * Accessor for whether the connection is closed.
     * @return Whether it is closed.
     * @throws SQLException thrown if an error occurs.
     */
    public boolean isClosed() throws SQLException
    {
        if (isAvailable)
        {
            // When the connection is available we use the underlying connection response
            return this.conn.isClosed();
        }
        // Now closed (even though the underlying connection may be still open back in the PM)
        return false;
    }

    /**
     * Accessor for whether the connection is read only.
     * @return Whether it is read only.
     * @throws SQLException thrown if an error occurs.
     */
    public boolean isReadOnly() throws SQLException
    {
        assertAvailable();
        return this.conn.isReadOnly();
    }

    /**
     * Method to set the holdability for the connection.
     * @param holdability The holdability
     * @throws SQLException thrown if an error occurs.
     */
    public void setHoldability(int holdability) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setHoldability"));
    }

    /**
     * Method to set the transaction isolation for the connection.
     * @param level The transaction isolation
     * @throws SQLException thrown if an error occurs.
     */
    public void setTransactionIsolation(int level) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setTransactionIsolation"));
    }

    /**
     * Method to set the autocommit for the connection.
     * @param autoCommit Whether to autocommit
     * @throws SQLException thrown if an error occurs.
     */
    public void setAutoCommit(boolean autoCommit) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setAutoCommit"));
    }

    /**
     * Method to set the read-only nature for the connection.
     * @param readOnly whether it is read-only
     * @throws SQLException thrown if an error occurs.
     */
    public void setReadOnly(boolean readOnly) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setReadOnly"));
    }

    /**
     * Accessor for the catalog.
     * @return The catalog
     * @throws SQLException Thrown if an error occurs
     */
    public String getCatalog() throws SQLException
    {
        assertAvailable();
        return this.conn.getCatalog();
    }

    /**
     * Mutator for the catalog
     * @param catalog The catalog
     * @throws SQLException Thrown if an error occurs
     */
    public void setCatalog(String catalog) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setCatalog"));
    }

    /**
     * Accessor for the database metadata.
     * @return The database metadata
     * @throws SQLException Thrown if an error occurs
     */
    public DatabaseMetaData getMetaData() throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "getMetaData"));
    }

    /**
     * Accessor for the warnings.
     * @return The warnings
     * @throws SQLException Thrown if an error occurs
     */
    public SQLWarning getWarnings() throws SQLException
    {
        assertAvailable();
        return this.conn.getWarnings();
    }

    /**
     * Mutator to set the savepoint.
     * @return The savepoint
     * @throws SQLException Thrown if an error occurs
     */
    public Savepoint setSavepoint() throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setSavepoint"));
    }

    /**
     * Mutator to release the savepoint.
     * @param pt The savepoint
     * @throws SQLException Thrown if an error occurs
     */
    public void releaseSavepoint(Savepoint pt) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "releaseSavepoint"));
    }

    /**
     * Mutator to rollback the savepoint.
     * @param pt The savepoint
     * @throws SQLException Thrown if an error occurs
     */
    public void rollback(Savepoint pt) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "rollback"));
    }

    /**
     * Accessor for a statement.
     * @return The statement
     * @throws SQLException Thrown if an error occurs
     */
    public Statement createStatement() throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement();
    }

    /**
     * Accessor for a statement.
     * @param resultSetType type of results
     * @param resultSetConcurrency The concurrency
     * @return The statement
     * @throws SQLException Thrown if an error occurs
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement(resultSetType,resultSetConcurrency);
    }

    /**
     * Accessor for a statement.
     * @param resultSetType type of results
     * @param resultSetConcurrency The concurrency
     * @param resultSetHoldability The holdability
     * @return The statement
     * @throws SQLException Thrown if an error occurs
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency, int resultSetHoldability) throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement(resultSetType, resultSetConcurrency, resultSetHoldability);
    }

    /**
     * Accessor for the type map.
     * @return The type map
     * @throws SQLException Thrown if an error occurs
     */
    public Map getTypeMap() throws SQLException
    {
        assertAvailable();
        return this.conn.getTypeMap();
    }

    /**
     * Method to set the type map
     * @param map The type map
     * @throws SQLException Thrown if an error occurs
     */
    public void setTypeMap(Map map) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setTypeMap"));
    }

    /**
     * Accessor for using native SQL.
     * @param sql The sql
     * @return The native SQL
     * @throws SQLException Thrown if an error occurs
     */
    public String nativeSQL(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.nativeSQL(sql);
    }

    public CallableStatement prepareCall(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(sql);
    }

    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(sql,resultSetType,resultSetConcurrency);
    }

    public CallableStatement prepareCall(String arg0, int arg1, int arg2, int arg3) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(arg0,arg1,arg2,arg3);
    }

    public PreparedStatement prepareStatement(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(sql);
    }

    public PreparedStatement prepareStatement(String arg0, int arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(sql,resultSetType,resultSetConcurrency);
    }

    public PreparedStatement prepareStatement(String arg0, int arg1, int arg2, int arg3) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1,arg2,arg3);
    }

    public PreparedStatement prepareStatement(String arg0, int[] arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }

    public Savepoint setSavepoint(String arg0) throws SQLException
    {
        throw new JDOUserException(LOCALISER.msg("Connection.InvalidCall", "setSavepoint"));
    }

    public PreparedStatement prepareStatement(String arg0, String[] arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }

    /**
     * Assert the JDOConnection is available
     * @throws JDOUserException if the connection is no longer available
     */
    public void assertAvailable()
    {
        if (!isAvailable)
        {
            throw new JDOUserException(LOCALISER.msg("Connection.NotAvailable"));
        }
    }

    /* (non-Javadoc)
     * @see javax.jdo.JDOConnection#getNativeConnection()
     */
    public Object getNativeConnection()
    {
        // Return the SQL connection
        return conn;
    }
}