/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - coding standards
2003 Erik Bengtson - optimistic transactions
2004 Andy Jefferson - addition of rollbackOnly
2005 Marco Schulze - bugfix in getConnection methods for non-transactional reads
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalDataStoreException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.jdo.datastore.JDOConnection;
import javax.resource.spi.security.PasswordCredential;
import javax.sql.XADataSource;
import javax.transaction.Status;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.resource.ManagedConnectionImpl;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.util.JPOXLogger;

/**
 * Internal JDO transaction class used by ManagedConnectionImpl to perform the 
 * transaction operations for an RDBMS datastore. It should not be used by clients.
 * The connection handle provides a wrapper to this.
 * The ManagedConnectionImpl wraps this to provide a javax.resource.spi.LocalTransaction
 * that converts JDO exceptions to ResourceExceptions.
 *
 * @version $Revision: 1.4 $
 */
public class RDBMSManagedTransaction extends AbstractRDBMSTransaction
{
    /** ManagedConnection for this transaction. */
    private final ManagedConnectionImpl mc;

    /** Password credentials (typically username, password). */
    private final PasswordCredential credential;

    /**
     * Constructor.
     * @param mc The connection
     * @param credential Password credentials
     */
    public RDBMSManagedTransaction(ManagedConnectionImpl mc, PasswordCredential credential)
    {
        this.mc = mc;
        this.dba = (RDBMSAdapter)mc.getStoreManager().getDatastoreAdapter();

        AbstractPersistenceManagerFactory apmf = mc.getAbstractPersistenceManagerFactory();

        this.tds = apmf.getTransactionalDataSource();
        this.useUpdateLock = apmf.getUseUpdateLock();
        this.credential = credential;

        setRetainValues(apmf.getRetainValues());
        setOptimistic(apmf.getOptimistic());
        setNontransactionalRead(apmf.getNontransactionalRead());
        setNontransactionalWrite(apmf.getNontransactionalWrite());
        setTransactionIsolation(apmf.getTransactionIsolation());
    }

    /**
     * Accessor for the persistence manager
     * @return The Persistence Manager.
     */
    public PersistenceManager getPersistenceManager()
    {
        return mc.getPMHandle();
    }

    /**
     * Accessor for the connection to the datasource.
     * If this method is called while a datastore transaction is active,
     * the object returned will be enlisted in the current transaction.
     * If called in an optimistic transaction or outside an active
     * transaction, the object returned will not be enlisted in any transaction.
     * 
     * The object must be returned to the JDO implementation prior to calling
     * any JDO method or performing any action on any persistent instance that
     * might require the JDO implementation to use a connection. If the object
     * has not been returned and the JDO implementation needs a connection,
     * a JDOUserException is thrown. The object is returned to the JDO
     * implementation by calling the standard method on the object.
     * 
     * @return The Connection
     */
    public synchronized JDOConnection getJDOConnection()
    {
        assertJDOConnectionNotInUse();
        
        isJDOConnectionInUse = true;
        return new JDOConnectionImpl(conn, new Runnable()
        {
            public void run()
            {
                isJDOConnectionInUse = false;
            }
        });
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param forQuery Whether the connection is to be used by a query (i.e.JDOQUERY).
     * @param useOpenedConnection whether to use an already opened connection, if one is opened. 
     * @return The Connection
     * @throws SQLException
     */    
    public Connection getConnection(boolean forWriting, boolean forQuery, boolean useOpenedConnection)
    throws SQLException
    {
        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug("getConnection(forWriting="+forWriting+", forQuery="+forQuery+", useOpenedConnection="+
                useOpenedConnection+"): entered");
        }

        assertJDOConnectionNotInUse();
        if (forWriting && !active)
        {
            throw new JDOUserException(LOCALISER.msg("Transaction.NontransactionalWriteUnsupported"));
        }

        // TODO maybe MEMORY LEAK! The connection is eventually never released in
        // non-transactional read mode! Could someone verify this?
        // And we should check whether we are allowed to do non-transactional-reads at all!
        if (conn == null)
        {
            // I think, it is OK (no memory leak), if we check active here, because
            // releaseConnection(...) will close the connection if active == false.
            // Marco :-)
            if (active)
            {
                throw new IllegalStateException("conn == null, but transaction active! " + 
                    "When a transaction is active, the connection must already exist here!");
            }

            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug("getConnection(forWriting="+forWriting+", forQuery="+forQuery+", useOpenedConnection="+
                    useOpenedConnection+"): conn==null => calling dba.getConnection(...)");
            }

            conn = dba.getConnection(tds,
                credential == null ? null : credential.getUserName(),
                credential == null ? null : new String(credential.getPassword()), isolationLevel);
        }

        if (active)
        {
            // Before performing any (more) SQL operations in this transaction,
            // flush out any updates pending in the persistence manager.
            mc.flush();
        }

        return conn;
    }

    /**
     * Method to release the connection to the datastore.
     * @param conn The connection
     * @throws SQLException Thrown if an error occurs
     */
    public synchronized void releaseConnection(Connection conn)
    throws SQLException
    {
        // We must close non-transactional connections here, because they obviously won't be closed in
        // commit/rollback. If we close all connections (including transactional conns) here, a lot
        // of exceptions occur. This is, because in a J2EE container many things happen, when the
        // Transaction is committed or rolled back and this is AFTER releaseConnection was called.
        // That's why we close the connection only, if we are not in a transaction (i.e. active == false).
        // Marco :-)

        if (conn != null && conn == this.conn)
        {
            if (!active)
            {
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug("releaseConnection(Connection): A transaction is NOT active => closing NTR connection now!");
                }

                closeConnection();
            }
            else
            {
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug("releaseConnection(Connection): A transaction is active! " +
                        "Ignoring this call, because we close only connections for non-transactional reads here.");
                }
            }
        }
        else {
            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                if (conn == null)
                {
                    JPOXLogger.TRANSACTION.debug("releaseConnection(Connection): conn is null! Ignoring this call.");
                }
                else
                {
                    JPOXLogger.TRANSACTION.debug("releaseConnection(Connection): conn != this.conn: conn=" + conn + " this.conn=" + this.conn);
                }
            }
        }
    }

    /**
     * Convenience method to close the connection to the datastore.
     * @throws SQLException Thrown if an error occurs
     */
    public void closeConnection()
    throws SQLException
    {
        //internally we close the connection during commit and rollback, any later call we ignore closing
        //the connection
        // Additionally, closeConnection is called by releaseConnection(Connection conn) if there is
        // no transaction active, because a non-transactional connection obviously won't be closed in
        // commit/rollback.
        // Marco :-)
        if (conn != null)
        {
            try
            {
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg(
                        optimistic ? "Transaction.Optimistic.ConnectionClosed" : "Transaction.Pessimistic.ConnectionClosed", conn));
                }
                dba.closeConnection(conn);
            }
            finally
            {
                conn = null;
            }
        }
    }

    /**
     * Method to start the Transaction
     **/
    public synchronized void begin()
    {
        assertNotInUse();
        try
        {
            conn = dba.getConnection(tds,  credential == null ? null : credential.getUserName(),  
                    credential == null ? null : new String(credential.getPassword()), isolationLevel);
        }
        catch (SQLException sqle)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.ConnectionUnobtainable"), sqle);
        }

        active = true;

        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.StartedForConnection", conn));
        }
    }

    /**
     * Method to commit the Transaction
     **/
    public synchronized void commit()
    {
        if (!active)
        {
            throw new org.jpox.exceptions.TransactionActiveException(this);
        }

        assertJDOConnectionNotInUse();
        
        // JDO 2.0 section 13.4.5 rollbackOnly functionality
        // It isn't clear from the spec if we are expected to do the rollback
        // here. The spec simply says that we throw an exception. This is
        // assumed as meaning that the users code will catch the exception and
        // call rollback themselves. i.e we don't need to close the DB
        // connection or set "active" to false. 
        if (rollbackOnly)
        {
            // Throw an exception since can only exit via rollback
            String msg = LOCALISER.msg("Transaction.CommitWithRollbackOnly");
            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug(msg);
            }
            throw new JDOFatalDataStoreException(msg);
        }

        try
        {
            commiting = true;

            // Perform any precommit checks such as reachability
            mc.preCommit(); // This has to be after setting "commiting"

            boolean success = false;

            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.CommitForConnection", conn));
                    }

                    mc.flush();

                    if (sync != null)
                    {
                        sync.beforeCompletion();
                    }
                    if (!(this.tds[0] instanceof XADataSource))
                    {
                        conn.commit();
                    }

                    success = true;
                }
                finally
                {
                    if (!success)
                    {
                        try
                        {
                            mc.preRollback();
                        }
                        finally
                        {
                            if (!(this.tds[0] instanceof XADataSource))
                            {
                                conn.rollback();
                            }
                        }
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    closeConnection();
                }
                finally
                {
                    try
                    {
                        if (success)
                        {
                            mc.postCommit();
                        }
                    }
                    finally
                    {
                        commiting = false;
                        if (sync != null)
                        {
                            if (success)
                            {
                                sync.afterCompletion(Status.STATUS_COMMITTED);
                            }
                            else
                            {
                                sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                            }
                        }
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToCommit"), new Exception[] { e });
        }
    }

    /**
     * Method to rollback the Transaction
     **/
    public synchronized void rollback()
    {
        if (!active)
        {
            throw new org.jpox.exceptions.TransactionNotActiveException();
        }

        assertJDOConnectionNotInUse();

        try
        {
            commiting = true;
            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.RollbackForConnection", conn));
                    }

                    mc.preRollback();
                }
                finally
                {
                    if (!(this.tds[0] instanceof XADataSource))
                    {
                        conn.rollback();
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    closeConnection();
                }
                finally
                {
                    commiting = false;
                    if (sync != null)
                    {
                        sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToRollback"), new Exception[] { e });
        }

        // Reset rollbackOnly flag
        rollbackOnly = false;
    }

    /**
     * Asserts that the transaction is not in use.
     **/
    protected void assertNotInUse()
    {
        if (active)
        {
            throw new org.jpox.exceptions.TransactionActiveException(this);
        }
    }
}