/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - commented
2003 Erik Bengtson - added Optimistic Transactions
2003 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - addition of rollbackOnly
2004 Andy Jefferson - updated useUpdate treatment to ignore when on Optimistic
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalDataStoreException;
import javax.jdo.PersistenceManager;
import javax.jdo.datastore.JDOConnection;
import javax.sql.DataSource;
import javax.transaction.Status;

import org.jpox.PersistenceManagerFactoryImpl;
import org.jpox.PersistenceManagerImpl;
import org.jpox.exceptions.ConnectionInUseException;
import org.jpox.exceptions.TransactionActiveException;
import org.jpox.exceptions.TransactionNotActiveException;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.util.JPOXLogger;

/**
 * Implementation of a non managed transaction for RDBMS datastores.
 * Provides the begin, commit and rollback methods and, in addition,
 * manages a connection to the datastore.
 *
 * @version $Revision: 1.5 $
 **/
public class RDBMSNonmanagedTransaction extends AbstractRDBMSTransaction
{
    /** PersistenceManager for this transaction */
    private final PersistenceManagerImpl pm;

    /** Nontransactional data source. */
    private final DataSource[] ntds;

    private final String userName;
    private final String password;

    /**
     * optimistic transaction normally release connection after reading
     * Connection are pinned (hold opened during transaction) when 
     * ignoreCache == false 
     */
    private boolean pinnedConnection = false;

    /**
     * Constructor.
     * @param pm The PersistenceManager
     * @param userName The username for the datasource
     * @param password The password for the datasource
     **/
    public RDBMSNonmanagedTransaction(PersistenceManagerImpl pm, String userName, String password)
    {
        this.pm = pm;
        this.dba = (RDBMSAdapter)pm.getStoreManager().getDatastoreAdapter();

        PersistenceManagerFactoryImpl pmf = (PersistenceManagerFactoryImpl)pm.getPersistenceManagerFactory();

        this.tds = pmf.getTransactionalDataSource();
        this.ntds = pmf.getNontransactionalDataSource();
        this.userName = userName;
        this.password = password;

        setRetainValues(pmf.getRetainValues());
        setRestoreValues(pmf.getRestoreValues());
        setOptimistic(pmf.getOptimistic());
        setNontransactionalRead(pmf.getNontransactionalRead());
        setNontransactionalWrite(pmf.getNontransactionalWrite());

        setTransactionIsolation(pmf.getTransactionIsolation());
        this.useUpdateLock = pmf.getUseUpdateLock();
    }

    /**
     * Accessor for the Persistence Manager.
     * @return Persistence Manager
     **/
    public PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * Accessor for the connection to the datasource.
     * If this method is called while a datastore transaction is active,
     * the object returned will be enlisted in the current transaction.
     * If called in an optimistic transaction or outside an active
     * transaction, the object returned will not be enlisted in any transaction.
     * 
     * The object must be returned to the JDO implementation prior to calling
     * any JDO method or performing any action on any persistent instance that
     * might require the JDO implementation to use a connection. If the object
     * has not been returned and the JDO implementation needs a connection,
     * a JDOUserException is thrown. The object is returned to the JDO
     * implementation by calling the standard method on the object.
     * 
     * @return The Connection
     */
    public synchronized JDOConnection getJDOConnection()
    {
        assertJDOConnectionNotInUse();
        
        final Connection nativeConnection;
        final boolean createNew;
        if (conn != null)
        {
            isJDOConnectionInUse = true;
            nativeConnection = conn;
            createNew = false;
        }
        else
        {
	        try
	        {
	            nativeConnection = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                if (pm.getPMFContext().getPmfConfiguration().getReadOnlyDatastore())
                {
                    nativeConnection.setReadOnly(true);
                }
                createNew = true;
	        }
	        catch (SQLException e)
	        {
	            throw new JDODataStoreException(e.getMessage());
	        }
        }

        return new JDOConnectionImpl(nativeConnection, new Runnable()
            {
                public void run()
                {
                    isJDOConnectionInUse = false;
                    if (createNew)
                    {
                        try
                        {
                            nativeConnection.close();
                        }
                        catch (SQLException sqle)
                        {
                            throw new JDODataStoreException(sqle.getMessage());
                        }
                    }
                }
            });
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param forQuery Whether the connection is to be used by a query (i.e.JDOQUERY).
     * @param useOpenedConnection whether to use an already opened connection, if one is opened
     * @return The Connection
     * @throws SQLException
     */
    public synchronized Connection getConnection(boolean forWriting, boolean forQuery, boolean useOpenedConnection)
    throws SQLException
    {
        assertJDOConnectionNotInUse();
        
        if (active)
        {
            // Before performing any (more) SQL operations in this transaction,
            // flush out any updates pending in the persistence manager.
            if ((!getPersistenceManager().getIgnoreCache() && forQuery) || forWriting)
            {
                if (optimistic && !pinnedConnection)
                {
                    pinnedConnection = true;
                    conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                    if (pm.getPMFContext().getPmfConfiguration().getReadOnlyDatastore())
                    {
                        conn.setReadOnly(true);
                    }
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Optimistic.ConnectionRetrieved", conn));
                    }
                }
                pm.flush();
            }
            if (optimistic && !pinnedConnection)
            {
                conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                if (pm.getPMFContext().getPmfConfiguration().getReadOnlyDatastore())
                {
                    conn.setReadOnly(true);
                }
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Optimistic.ConnectionRetrieved", conn));
                }
            }
        }
        else
        {
            if (!nontransactionalRead || (forWriting && !nontransactionalWrite))
            {
                throw new TransactionNotActiveException();
            }

            if (!useOpenedConnection)
            {
                while (conn != null)
                {
                    try
                    {
                        // Wait for a connection
                        wait();
                    }
                    catch (InterruptedException e)
                    {
                    }
                }
            }

            if (conn == null)
            {
                conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                if (pm.getPMFContext().getPmfConfiguration().getReadOnlyDatastore())
                {
                    conn.setReadOnly(true);
                }
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg(
                        optimistic ? "Transaction.Optimistic.ConnectionRetrieved" : "Transaction.Pessimistic.ConnectionRetrieved", conn));
                }
            }
        }

        return conn;
    }

    /**
     * Method to release the Connection.
     * @param conn The Connection to release.
     * @throws SQLException Thrown when an error occurs releasing the
     * connection.
     **/
    public synchronized void releaseConnection(Connection conn)
    throws SQLException
    {
        assertJDOConnectionNotInUse();

        if (conn == this.conn)
        {
            if (!active || (optimistic && !pinnedConnection))
            {
                if (conn != null)
                {
                    if (!conn.getAutoCommit())
                    {
                        conn.commit();
                    }
                }
                closeConnection();
            }
        }
    }

    /**
     * Method to close the Connection.
     * @throws SQLException Thrown when an error occurs closing the connection.
     **/
    private void closeConnection() 
    throws SQLException
    {
        try
        {
            if (conn != null)
            {
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg(
                        optimistic ? "Transaction.Optimistic.ConnectionClosed" : "Transaction.Pessimistic.ConnectionClosed", conn));
                }
                dba.closeConnection(conn);
            }
        }
        finally
        {
            pinnedConnection = false;
            conn = null;

            // Notify any waiting threads that we can get a connection
            notifyAll();
        }
    }

    /**
     * Method to start the transaction.
     **/
    public synchronized void begin()
    {
        assertNotInUse();

        try
        {
            if (!optimistic)
            {
                conn = dba.getConnection(tds, userName, password, getTransactionIsolation());
                if (pm.getPMFContext().getPmfConfiguration().getReadOnlyDatastore())
                {
                    conn.setReadOnly(true);
                }
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Pessimistic.ConnectionRetrieved", conn));
                }
            }

            active = true;

            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.StartedForConnection",conn));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToStart"),e);
        }
    }

    /**
     * Method to commit the transaction.
     **/
    public synchronized void commit()
    {
		if (!active)
        {
            throw new TransactionNotActiveException();
        }

        assertJDOConnectionNotInUse();

		// JDO 2.0 section 13.4.5 rollbackOnly functionality
		// It isn't clear from the spec if we are expected to do the rollback
		// here. The spec simply says that we throw an exception. This is
		// assumed as meaning that the users code will catch the exception and
		// call rollback themselves. i.e we don't need to close the DB
		// connection or set "active" to false. 
		if (rollbackOnly)
		{
			// Throw an exception since can only exit via rollback
			if (JPOXLogger.TRANSACTION.isDebugEnabled())
			{
				JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.CommitWithRollbackOnly"));
			}
			
			throw new JDOFatalDataStoreException(LOCALISER.msg("Transaction.CommitWithRollbackOnly"));
		}

		try
		{
			commiting = true;

            // Perform any pre-commit operations
            pm.preCommit(); // This has to be after setting "commiting"

			boolean success = false;
			
			try
			{
				try
				{
					if (JPOXLogger.TRANSACTION.isDebugEnabled())
					{
						JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.CommitForConnection",conn));
					}
					
					pm.flush();
					
					if (sync != null)
					{
						sync.beforeCompletion();
					}
					
					//optimistic transactions that don't have dirty
					if (conn != null)
					{
                        if (!conn.getAutoCommit())
                        {
                            conn.commit();
                        }
					}
					
					success = true;
				}
				finally
				{
					if (!success)
					{
						try
						{
							pm.preRollback();
						}
						finally
						{
							if (conn != null)
							{
								conn.rollback();
							}
						}
					}
				}
			}
			finally
			{
				try
				{
					active = false;
					closeConnection();
				}
				finally
				{
					try
					{
						if (success)
						{
							pm.postCommit();
						}
					}
					finally
					{
						commiting = false;
						if (sync != null)
						{
							if (success)
							{
								sync.afterCompletion(Status.STATUS_COMMITTED);
							}
							else
							{
								sync.afterCompletion(Status.STATUS_ROLLEDBACK);
							}
						}
					}
				}
			}
		}
		catch (SQLException e)
		{
			throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToCommit"),e);
		}
    }

    /**
     * Method to rollback the transaction.
     **/
    public synchronized void rollback()
    {
        if (!active)
        {
            throw new TransactionNotActiveException();
        }
        
        assertJDOConnectionNotInUse();

        try
        {
			commiting = true;
            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.RollbackForConnection",conn));
                    }

                    pm.preRollback();
                }
                finally
                {
                    if (conn != null)
                    {
                        conn.rollback();
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    closeConnection();
                }
                finally
                {
					commiting = false;
                    if (sync != null)
                    {
                        sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToRollback"),e);
        }

        // Reset rollbackOnly flag
        rollbackOnly = false;
    }

    /**
     * Asserts that the transaction is not in use.
     **/
    protected void assertNotInUse()
    {
        if (active)
        {
            throw new TransactionActiveException(this);
        }

        if (conn != null)
        {
            throw new ConnectionInUseException();
        }
    }
}