/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Properties;

import javax.jdo.JDODataStoreException;
import javax.jdo.datastore.Sequence;

import org.jpox.PersistenceManager;
import org.jpox.metadata.ExtensionMetaData;
import org.jpox.metadata.SequenceMetaData;
import org.jpox.store.StoreManager;
import org.jpox.store.poid.PoidGenerator;
import org.jpox.store.poid.PoidConnectionProvider;
import org.jpox.store.poid.PoidManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Basic implementation of a JDO 2 datastore sequence for RDBMS datastores.
 * Utilises the <B>org.jpox.store.poid</B> classes to generate
 * sequence values.
 * @version $Revision: 1.1 $
 */
public class SequenceImpl implements Sequence
{
    /** Localisation of messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** Store Manager where we obtain our sequence. */
    protected final StoreManager storeManager;

    /** Name of the sequence. */
    protected final SequenceMetaData seqMetaData;

    /** The generator for the sequence. */
    protected final PoidGenerator generator;

    /** The controlling PersistenceManager. */
    protected final PersistenceManager pm;

    /**
     * Constructor.
     * @param persistenceManager The PersistenceManager managing the sequence
     * @param storeMgr Manager of the store where we obtain the sequence
     * @param seqmd MetaData defining the sequence
     */
    public SequenceImpl(PersistenceManager persistenceManager, StoreManager storeMgr, SequenceMetaData seqmd)
    {
        this.pm = persistenceManager;
        this.storeManager = storeMgr;
        this.seqMetaData = seqmd;

        // Allocate the PoidManager for this sequence
        String poidGeneratorName = null;
        if (storeMgr.getDatastoreAdapter().supportsSequences())
        {
            poidGeneratorName = "org.jpox.store.rdbms.poid.SequencePoidGenerator";
        }
        else
        {
            poidGeneratorName = "org.jpox.store.rdbms.poid.SequenceTablePoidGenerator";
        }

        // Create the controlling properties for this sequence
        Properties props = new Properties();
        ExtensionMetaData[] seqExtensions = seqmd.getExtensions();
        if (seqExtensions != null && seqExtensions.length > 0)
        {
            // Add all MetaData extension properties provided
            for (int i=0;i<seqExtensions.length;i++)
            {
                props.put(seqExtensions[i].getKey(), seqExtensions[i].getValue());
            }
        }
        if (poidGeneratorName.equals("org.jpox.store.rdbms.poid.SequencePoidGenerator"))
        {
            // Use the provided name as the sequence name (what if it is null ?)
            // TODO Fix this to avoid name clashes
            props.put("sequence-name", seqMetaData.getDatastoreSequence());
        }
        else if (poidGeneratorName.equals("org.jpox.store.rdbms.poid.SequenceTablePoidGenerator"))
        {
            // Use the provided name as the table name (what if it is null ?)
            // TODO Fix this to avoid name clashes
            props.put("sequence-name", seqMetaData.getDatastoreSequence());
        }

        // Generate a PoidManager to create the generator
        // TODO Take this manager from the StoreManager
        PoidManager mgr = new PoidManager();

        PoidConnectionProvider connProvider = new PoidConnectionProvider()
            {
                Connection conn;

                public Object retrieveConnection()
                {
                    try
                    {
                        // Obtain a new connection
                        // Note : it may be worthwhile to use the PM's connection here however where a Sequence doesnt yet
                        // exist the connection would then be effectively dead until the end of the tx
                        // The way around this would be to find a way of checking for existence of the sequence
                        this.conn = ((RDBMSManager)storeManager).getConnection(pm.getPMFContext().getPmfConfiguration().getPoidTransactionIsolationLevel());
                    }
                    catch (SQLException e)
                    {
                        String msg = LOCALISER.msg("Sequence.POIDConnectionOpenError", e);
                        JPOXLogger.JDO.error(msg);
                        throw new JDODataStoreException(msg,e);
                    }
                    return conn;
                }

                public void releaseConnection()
                {
                    try
                    {
                        // Release the connection
                        boolean success = false;
                        try
                        {
                            if (!conn.getAutoCommit())
                            {
                                conn.commit();
                            }
                            success = true;
                        }
                        finally
                        {
                            if (!success)
                            {
                                if (!conn.getAutoCommit())
                                {
                                    conn.rollback();
                                }
                            }
                        }
                        ((RDBMSManager)storeManager).closeConnection(conn);
                    }
                    catch (SQLException e)
                    {
                        String msg = LOCALISER.msg("Sequence.POIDConnectionCloseError", e);
                        JPOXLogger.JDO.error(msg);
                        throw new JDODataStoreException(msg);
                    }
                }
            };
        generator = mgr.createPoidGenerator(pm.getClassLoaderResolver(), seqMetaData.getName(), poidGeneratorName, props, 
            storeManager, connProvider);

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("Sequence.Allocated", seqMetaData.getName(), poidGeneratorName));
        }
    }

    /**
     * Accessor for the sequence name.
     * @return The sequence name
     */
    public String getName()
    {
        return seqMetaData.getName();
    }

    /**
     * Method to allocate a set of elements.
     * @param additional The number of additional elements to allocate
     */
    public void allocate(int additional)
    {
        generator.allocate(additional);
    }

    /**
     * Accessor for the next element in the sequence.
     * @return The next element
     */
    public Object next()
    {
        return generator.next();
    }

    /**
     * Accessor for the next element in the sequence as a long.
     * @return The next element
     * @throws JDODataStoreException Thrown if not numeric
     */
    public long nextValue()
    {
        return generator.nextValue();
    }

    /**
     * Accessor for the current element.
     * @return The current element.
     */
    public Object current()
    {
        return generator.current();
    }

    /**
     * Accessor for the current element in the sequence as a long.
     * @return The current element
     * @throws JDODataStoreException Thrown if not numeric
     */
    public long currentValue()
    {
        return generator.currentValue();
    }
}