/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Iterator;
import java.util.Properties;

import javax.sql.DataSource;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.exceptions.UnsupportedDataTypeException;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.table.ViewImpl;
import org.jpox.store.rdbms.typeinfo.ForeignKeyInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Adapter for relational databases 
 */
public interface RDBMSAdapter extends DatastoreAdapter
{
    /**
     * Accessor for a Connection to the datastore.
     * @param ds The data source. Possible to have more than one datasource for failover
     * @param userName The username for the datastore
     * @param password The password for the datastore
     * @param isolationLevel The level of transaction isolation
     * @return The Connection
     * @throws SQLException Thrown when an error occurs in the creation.
     **/
    Connection getConnection(DataSource[] ds, String userName, String password, int isolationLevel)
    throws SQLException;

    /**
     * Method to close a Connection to the datastore.
     * @param conn The connection
     * @throws SQLException Thrown if error occurs on the close.
     **/
    void closeConnection(Connection conn)
    throws SQLException;

    /**
     * Method to return the SQL to append to the SELECT clause of a SELECT statement to handle
     * restriction of ranges using the LIMIT keyword. Defaults to an empty
     * string (not supported).
     * SELECT {LIMIT} params
     * @param offset The offset to return from
     * @param count The number of items to return
     * @return The SQL to append to allow for ranges using LIMIT.
     */
    String getRangeByLimitSelectClause(long offset, long count);

    /**
     * Method to return the SQL to append to the WHERE clause of a SELECT statement to handle
     * restriction of ranges using the LIMIT keyword. Defaults to an empty
     * string (not supported).
     * SELECT param ... WHERE {LIMIT}
     * @param offset The offset to return from
     * @param count The number of items to return
     * @return The SQL to append to allow for ranges using LIMIT.
     */
    String getRangeByLimitWhereClause(long offset, long count);
    
    /**
     * Method to return the column name to use when handling ranges via
     * a rownumber on the select. Defaults to an empty string (not supported).
     * @return The row number column.
     */
    String getRangeByRowNumberColumn();
    
    /**
     * Accessor for the maximum foreign keys by table permitted
     * datastore.
     * @return Max table name length
     **/
    int getMaxForeignKeys();
    
    /**
     * Accessor for the maximum indexes by schema permitted
     * datastore.
     * @return Max table name length
     **/
    int getMaxIndexes();

    /**
     * Accessor for the catalog separator.
     * @return Catalog separator string.
     **/
    String getCatalogSeparator();
    
    /**
     * A factory for ColumnInfo objects.  This method should always be used
     * instead of directly constructing ColumnInfo objects in order to give the
     * DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The column information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getColumns().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs    The result set returned from DatabaseMetaData.getColumns().
     * @return The column info 
     */
    ColumnInfo newColumnInfo(ResultSet rs);

    /**
     * A factory for ForeignKeyInfo objects.  This method should always be used
     * instead of directly constructing ForeignKeyInfo objects in order to give
     * the DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The column information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getImportedKeys() or
     * DatabaseMetaData.getExportedKeys().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs The result set returned from DatabaseMetaData.get??portedKeys()
     * @return The foreign key info 
     */
    ForeignKeyInfo newForeignKeyInfo(ResultSet rs);
    
    /**
     * Accessor for an identifier quote string.
     * @return Identifier quote string.
     **/
    String getIdentifierQuoteString();
    
    /**
     * Whether this datastore supports the use of the catalog name in SQL
     * table definitions (DDL).
     * @return whether we can use the catalog name in DDL
     **/
    boolean supportsCatalogsInTableDefinitions();

    /**
     * Whether this datastore supports the use of the schema name in SQL
     * table definitions (DDL).
     * @return whether we can use the schema name in DDL
     **/
    boolean supportsSchemasInTableDefinitions();

    /**
     * Whether this datastore supports batching of statements.
     * @return whether we can use statement batching
     **/
    boolean supportsStatementBatching();

    /**
     * Whether the RDBMS supports use of UNION syntax.
     * @return Whether UNION is supported.
     */
    boolean supportsUnionSyntax();

    /**
     * Whether the RDBMS supports use of EXISTS syntax.
     * @return Whether EXISTS is supported.
     */
    boolean supportsExistsSyntax();

    /**
     * Accessor for whether the SQL extensions CUBE, ROLLUP are supported.
     * @return Whether the SQL extensions CUBE, ROLLUP are supported.
     */
    boolean supportsAnalysisMethods();

    /**
     * Accessor for table and column information for a catalog/schema in this datastore.
     * @param conn Connection to use
     * @param catalog The catalog (null if none)
     * @param schema The schema (null if none)
     * @param table The table (null if all)
     * @return ResultSet containing the table/column information
     * @throws SQLException Thrown if an error occurs
     */
    ResultSet getColumns(Connection conn, String catalog, String schema, String table)
    throws SQLException;

    /**
     * Accessor for table and column information for a catalog/schema in this datastore.
     * @param conn Connection to use
     * @param catalog The catalog (null if none)
     * @param schema The schema (null if none)
     * @param table The table (null if all)
     * @param columnNamePattern The column name (null if all)
     * @return ResultSet containing the table/column information
     * @throws SQLException Thrown if an error occurs
     */
    ResultSet getColumns(Connection conn, String catalog, String schema, String table, String columnNamePattern)
    throws SQLException;
    
    /**
     * Accessor for table information for a catalog/schema in this datastore.
     * @param conn Connection to use
     * @param catalog The catalog (null if none)
     * @param schema The schema (null if none)
     * @return ResultSet containing the table/column information
     * @throws SQLException Thrown if an error occurs
     */
    ResultSet getTables(Connection conn, String catalog, String schema)
    throws SQLException;    
    
    /**
     * Whether this datastore supports the specified foreign key update action
     * @param action The update action
     * @return Whether it is supported
     */
    boolean supportsForeignKeyUpdateAction(ForeignKey.FKAction action);

    /**
     * Whether this datastore supports the specified foreign key delete action
     * @param action The delete action
     * @return Whether it is supported
     */
    boolean supportsForeignKeyDeleteAction(ForeignKey.FKAction action);

    /**
     * Whether this datastore supports deferred constraints.
     * @return whether we support deferred constraints.
     **/
    boolean supportsDeferredConstraints();
    
    /**
     * Whether to include ORDER BY columns in a SELECT.
     * @return Whether to include ORDER BY in SELECTs
     **/
    boolean includeOrderByColumnsInSelect();

    /**
     * Method to retutn the INSERT statement to use when inserting into a table that has no
     * columns specified. This is the case when we have a single column in the table and that column
     * is autoincrement/identity (and so is assigned automatically in the datastore).
     * @param table The table
     * @return The statement for the INSERT
     */
    String getInsertStatementForNoColumns(Table table);
    
    /**
     * Accessor for the maximum constraint name length permitted on this
     * datastore.
     * @return Max constraint name length
     **/
    int getMaxConstraintNameLength();
    
    /**
     * Accessor for the maximum column name length permitted on this datastore.
     * @return Max column name length
     **/
    int getMaxColumnNameLength();
    
    /**
     * Accessor for whether a JDBC Types.BIT is really mapped in the datastore
     * to a boolean field.
     * @return Whether it is really a boolean
     */
    boolean isBitReallyBoolean();

    /**
     * Accessor for whether setting a BLOB value allows use of PreparedStatement.setString()
     * @return Whether setString is allowed.
     */
    boolean supportsSettingBlobUsingSetString();

    /**
     * Accessor for whether setting a CLOB value allows use of PreparedStatement.setString()
     * @return Whether setString is allowed.
     */
    boolean supportsSettingClobUsingSetString();

    /**
     * Returns the precision value to be used when creating string columns of
     * "unlimited" length.  Usually, if this value is needed it is provided in
     * the database metadata ({@link TypeInfo#precision}).  However, for some
     * types in some databases the value must be computed specially.
     *
     * @param typeInfo  the typeInfo object for which the precision value is
     *                  needed.
     *
     * @return  the precision value to be used when creating the column, or -1
     *          if no value should be used.
     */
    int getUnlimitedLengthPrecisionValue(TypeInfo typeInfo);

    /**
     * Whether any DEFAULT tag will be before any NULL/NOT NULL in the column options.
     * @return Whether to put DEFAULT before NULL
     */
    boolean supportsDefaultBeforeNullInColumnOptions();
    
    /**
     * Whether we support DEFAULT tag in CREATE TABLE statements
     * @return whether we support the DEFAULT tag in CREATE TABLE
     **/
    boolean supportsDefaultKeywordInColumnOptions();

    /**
     * Whether we support DEFAULT tag together with NOT NULL in CREATE TABLE statements.
     * <pre>
     * CREATE TABLE X ( MEMORY_SIZE BIGINT DEFAULT 0 NOT NULL )
     * </pre>
     * Some databases only support <i>DEFAULT {ConstantExpression | NULL}</i>
     * 
     * @return whether we support the DEFAULT tag together with NOT NULL in CREATE TABLE
     **/
    boolean supportsDefaultKeywordWithNotNullInColumnOptions();

    /**
     * Whether the database support NULLs in the column options for table creation.
     * @return whether the database support NULLs in the column options for table creation.
     **/
    boolean supportsNullsKeywordInColumnOptions();

    /**
     * Whether this datastore supports the use of CHECK in CREATE TABLE
     * statements (DDL).
     * @return whether we can use CHECK in CREATE TABLE.
     **/
    boolean supportsCheckInCreateStatements();

    /**
     * Accessor for the auto-increment/identity sql statement for this datastore.
     * @param tableName Name of table (that the autoincrement is for)
     * @param columnName (that the autoincrement is for)
     * @return The statement for getting the latest auto-increment/identity key
     **/
    String getAutoIncrementStmt(String tableName, String columnName);

    /**
     * Accessor for the auto-increment/identity keyword for generating DDLs.
     * (CREATE TABLEs...).
     * @return The keyword for a column using auto-increment/identity
     **/
    String getAutoIncrementKeyword();

    /**
     * Whether we support auto-increment/identity keys with nullability specification.
     * @return whether we support auto-increment keys with nullability spec.
     **/
    boolean supportsAutoIncrementKeysNullSpecification();

    /**
     * Whether we support auto-increment/identity keys with column type specification.
     * @return whether we support auto-increment keys with column type spec.
     **/
    boolean supportsAutoIncrementColumnTypeSpecification();

    /**
     * Whether this datastore supports the use of UNIQUE after the column
     * definitions in CREATE TABLE statements (DDL).
     * e.g.<PRE>
     * CREATE TABLE XXX
     * (
     *     COL_A int,
     *     COL_B char(1),
     *     PRIMARY KEY (COL_A),
     *     UNIQUE (COL_B ...)
     * )
     * </PRE> 
     * @return whether we can use UNIQUE after the column definitions CREATE TABLE.
     **/
    boolean supportsUniqueConstraintsInEndCreateStatements();
    
    /**
     * Accessor for the maximum table name length permitted on this
     * datastore.
     * @return Max table name length
     **/
    int getMaxTableNameLength();
    
    /**
     * Accessor for the maximum index name length permitted on this datastore.
     * @return Max index name length
     **/
    int getMaxIndexNameLength();

    /**
     * Tests if a given string is a SQL key word.
     * <p>
     * The list of key words tested against is defined to contain all SQL/92 key
     * words, plus any additional key words reported by the JDBC driver for this
     * adapter via <code>DatabaseMetaData.getSQLKeywords()</code>.
     * <p>
     * In general, use of a SQL key word as an identifier should be avoided.
     * SQL/92 key words are divided into reserved and non-reserved words. If a
     * reserved word is used as an identifier it must be quoted with double
     * quotes. Strictly speaking, the same is not true of non-reserved words.
     * However, as C.J. Date writes in <u>A Guide To The SQL Standard </u>:
     * <blockquote>The rule by which it is determined within the standard that
     * one key word needs to be reserved while another need not is not clear to
     * this writer. In practice, it is probably wise to treat all key words as
     * reserved. </blockquote>
     * @param word The word to test.
     * @return <code>true</code> if <var>word </var> is a SQL key word for
     * this DBMS. The comparison is case-insensitive.
     * @see SQLConstants
     */
    boolean isSQLKeyword(String word);
    
    /**
     * Returns the appropriate SQL to drop the given table.
     * It should return something like:
     * <p>
     * <pre>
     * DROP TABLE FOO CASCADE
     * </pre>
     *
     * @param table The table to drop.
     * @return The text of the SQL statement.
     */
    String getDropTableStatement(DatastoreContainerObject table);

    /**
     * Returns the appropriate SQL to add a candidate key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_CK UNIQUE (BAZ)
     * </pre>
     *
     * @param ck An object describing the candidate key.
     * @param factory Identifier factory
     * @return  The text of the SQL statement.
     */
    String getAddCandidateKeyStatement(CandidateKey ck, IdentifierFactory factory);

    /**
     * Whether we support NULLs in candidate keys.
     * @return whether we support NULLs in candidate keys.
     **/
    boolean supportsNullsInCandidateKeys();
    
    /**
     * Method to return whether the specified JDBC type is valid for use in a PrimaryKey.
     * @param datatype The JDBC type.
     * @return Whether it is valid for use in the PK
     */
    boolean isValidPrimaryKeyType(int datatype);
    
    /**
     * Accessor for the SQL statement to add a column to a table.
     * @param table The table
     * @param col The column
     * @return The SQL necessary to add the column
     */
    String getAddColumnStatement(DatastoreContainerObject table, Column col);

    /**
     * Whether to create indexes before foreign keys.
     * @return Whether to create indexes before foreign keys
     **/
    boolean createIndexesBeforeForeignKeys();

    /**
     * Whether this datastore supports ALTER TABLE DROP constraints
     * @return whether we support ALTER TABLE DROP constraints
     **/
    boolean supportsAlterTableDropConstraint();
    
    /**
     * Whether this datastore supports ALTER TABLE DROP FOREIGN KEY constraints
     * @return whether we support ALTER TABLE DROP FOREIGN KEY constraints
     **/
    boolean supportsAlterTableDropForeignKeyConstraint();
    
    /**
     * Returns the appropriate SQL to add an index to its table.
     * It should return something like:
     * <p>
     * <pre>
     * CREATE INDEX FOO_N1 ON FOO (BAR,BAZ)
     * CREATE UNIQUE INDEX FOO_U1 ON FOO (BAR,BAZ)
     * </pre>
     *
     * @param idx An object describing the index.
     * @param factory Identifier factory
     * @return The text of the SQL statement.
     */
    String getCreateIndexStatement(Index idx, IdentifierFactory factory);

    /**
     * Provide the existing indexes in the database for the table
     * @param conn the JDBC connection
     * @param dmd the DatabaseMetaData
     * @param catalog the catalog name
     * @param schema the schema name
     * @param table the table name
     * @return a ResultSet with the format @see DatabaseMetaData#getIndexInfo(java.lang.String, java.lang.String, java.lang.String, boolean, boolean)
     * @throws SQLException
     */
    ResultSet getExistingIndexes(Connection conn, DatabaseMetaData dmd, String catalog, String schema, String table)
    throws SQLException;
    
    /**
     * Returns the appropriate SQL to create the given table having the given
     * columns. No column constraints or key definitions should be included.
     * It should return something like:
     * <p>
     * <pre>
     * CREATE TABLE FOO (BAR VARCHAR(30), BAZ INTEGER)
     * </pre>
     *
     * @param table The table to create.
     * @param columns The columns of the table.
     * @param props Properties for controlling the table creation
     * @return The text of the SQL statement.
     */
    String getCreateTableStatement(TableImpl table, Column[] columns, Properties props);

    /**
     * Returns the appropriate SQL to add a primary key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_PK PRIMARY KEY (BAR)
     * </pre>
     *
     * @param pk An object describing the primary key.
     * @param factory Identifier factory
     * @return The text of the SQL statement.
     */
    String getAddPrimaryKeyStatement(PrimaryKey pk, IdentifierFactory factory);

    /**
     * Returns the appropriate SQL to add a foreign key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_FK1 FOREIGN KEY (BAR, BAZ) REFERENCES ABC (COL1, COL2)
     * </pre>
     * @param fk An object describing the foreign key.
     * @param factory Identifier factory
     * @return  The text of the SQL statement.
     */
    String getAddForeignKeyStatement(ForeignKey fk, IdentifierFactory factory);

    /**
     * Returns the appropriate SQL to drop the given view.
     * It should return something like:
     * <p>
     * <pre>
     * DROP VIEW FOO
     * </pre>
     *
     * @param view The view to drop.
     * @return The text of the SQL statement.
     */
    String getDropViewStatement(ViewImpl view);
    
    /**
     * Returns type information for the database type that best implements the
     * given JDBC type.
     * @param dataType JDBC type number of the data type.
     * @return type information for the best matching type.
     * @throws UnsupportedDataTypeException Thrown if the specified datatype is
     * not supported.
     */
    TypeInfo getTypeInfo(int dataType)
    throws UnsupportedDataTypeException;

    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return whether the database treats an empty string as null
     */
    boolean isNullEqualsEmptyStrings();
    
    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return returns a surrogate to replace the empty string in the database
     * otherwise it would be treated as null
     */
    String getSurrogateForEmptyStrings();

    /**
     * Whether we support sequences.
     * @return whether we support sequences.
     **/
    boolean supportsSequences();
    
    /**
     * Accessor for whether the column defaults need initialising for this datastore
     * @return Whether to initialise the column defaults
     */
    boolean requiresColumnDefaultsInitialising();
    
    /**
     * Accessor for the transaction isolation level to use during schema creation.
     * @return The transaction isolation level for schema generation process
     */
    int getTransactionIsolationForSchemaCreation();
    
    /**
     * Accessor for the Catalog Name for this datastore.
     * @param conn Connection to the datastore
     * @return The catalog name
     * @throws SQLException Thrown if error occurs in determining the
     * catalog name.
     **/
    String getCatalogName(Connection conn)
    throws SQLException;

    /**
     * Accessor for the Schema Name for this datastore.
     * @param conn Connection to the datastore
     * @return The schema name
     * @throws SQLException Thrown if error occurs in determining the
     * schema name.
     **/
    String getSchemaName(Connection conn)
    throws SQLException;

    /**
     * Accessor for whether the RDBMS supports ANSI join syntax.
     * @return Whether the RDBMS supports ANSI join syntax
     */
    boolean supportsAnsiJoinSyntax();
    
    /**
     * Accessor for the WHERE clause to add to provide a LEFT OUTER JOIN
     * using non-ANSI syntax.
     * @param col1 The main datastore column
     * @param col2 The secondary column to join to
     * @return The Left Outer Join WHERE clause.
     */
    String getNonAnsiLeftOuterJoinWhereClause(String col1, String col2);

    /**
     * Accessor for the WHERE clause to add to provide an INNER JOIN
     * using non-ANSI syntax.
     * @param col1 The main datastore column
     * @param col2 The secondary column to join to
     * @return The Inner Join WHERE clause.
     */
    String getNonAnsiInnerJoinWhereClause(String col1, String col2);

    /**
     * Whether this datastore supports SELECT ... FOR UPDATE.
     * @return whether we support SELECT ... FOR UPDATE.
     **/
    boolean supportsLockWithSelectForUpdate();
    
    /**
     * The option to specify in "SELECT ... FROM TABLE ... WITH (option)" to lock instances
     * Null if not supported.
     * @return The option to specify with "SELECT ... FROM TABLE ... WITH (option)"
     **/
    String getSelectWithLockOption();

    /**
     * Determines whether the {@link #getSelectWithLockOption()} is to be placed right
     * before the FROM clause, or at the end of the statement 
     * @return true if option has to be placed right after the from clause
     */
    public boolean getPlaceWithOptionAfterFromClause();
    
    /**
     * Determines whether lock option has to be placed within JOIN clauses as well.
     * @return true if option has to be placed right after the join clause
     */    
    public boolean getPlaceWithOptionWithinJoinClauses();
    
    /**
     * Returns a new TableExpression object appropriate for this DBMS.
     * This should be an instance of one of the three built-in styles of table
     * expression:
     * <ul>
     *   <li>TableExprAsJoins</li>
     *   <li>TableExprAsSubjoins</li>
     *   <li>TableExprAsSubquery</li>
     * </ul>
     * TableExprAsSubjoins is the default, which arguably produces the most
     * readable SQL but doesn't work on all DBMS's.  TableExprAsSubjoins
     * should work anywhere, but may be less efficient.
     *
     * @param qs        The query statement in which the table expression will
     *                  be included.
     * @param table     The main table in the expression.
     * @param rangeVar  The SQL alias, or "range variable", to assign to the
     *                  expression or to the main table.
     * @return The expression
     */
    LogicSetExpression newTableExpression(QueryExpression qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar);

    /**
     * Union combines the results of two or more queries into a single result
     * set. Union include only distinct rows and Union all may include
     * duplicates.
     * 
     * When using the UNION statement, keep in mind that, by
     * default, it performs the equivalent of a SELECT DISTINCT on the final
     * result set. In other words, UNION takes the results of two like
     * recordsets, combines them, and then performs a SELECT DISTINCT in order
     * to eliminate any duplicate rows. This process occurs even if there are no
     * duplicate records in the final recordset. If you know that there are
     * duplicate records, and this presents a problem for your application, then
     * by all means use the UNION statement to eliminate the duplicate rows. On
     * the other hand, if you know that there will never be any duplicate rows,
     * or if there are, and this presents no problem to your application, then
     * you should use the UNION ALL statement instead of the UNION statement.
     * The advantage of the UNION ALL is that is does not perform the SELECT
     * DISTINCT function, which saves a lot of unnecessary SQL Server resources
     * from being using.
     * 
     * @return Whether to use UNION ALL instead of UNION
     */
    boolean useUnionAll();
    
    /**
     * Whether this datastore supports using DISTINCT when using SELECT ... FOR UPDATE.
     * @return whether the datastore supports DISTINCT in same statement as FOR UPDATE
     **/
    boolean supportsDistinctWithSelectForUpdate();
    
    /**
     * The function to creates a unique value of type uniqueidentifier.
     * @return The function. e.g. "SELECT NEWID()"
     **/
    String getSelectNewUUIDStmt();

    /**
     * Accessor for the sequence statement to get the next id for this 
     * datastore.
     * @param sequence_name Name of the sequence 
     * @return The statement for getting the next id for the sequence
     **/
    String getSequenceNextStmt(String sequence_name);

    /**
     * Accessor for the sequence create statement for this datastore.
     * TODO Change param types to int.
     * @param sequence_name Name of the sequence 
     * @param min Minimum value for the sequence
     * @param max Maximum value for the sequence
     * @param start Start value for the sequence
     * @param increment Increment value for the sequence
     * @param cache_size Cache size for the sequence
     * @return The statement for getting the next id from the sequence
     **/
    String getSequenceCreateStmt(String sequence_name,
                                        String min,String max,
                                        String start,String increment,
                                        String cache_size);

    /**
     * Iterator for the TypeInfo objects constructed from the method
     * DataBaseMetaData.getTypeInfo
     * @return an Iterator with org.jpox.store.rdbms.typeinfo.TypeInfo
     */
    Iterator iteratorTypeInfo();

    /**
     * Iterator for the reserved words constructed from the method
     * DataBaseMetaData.getSQLKeywords + standard SQL reserved words
     * @return an Iterator with a set of reserved words
     */
    Iterator iteratorReservedWords();
    
    /**
     * Whether the datastore supports specification of the primary key in
     * CREATE TABLE statements.
     * @return Whetehr it allows "PRIMARY KEY ..."
     */
    boolean supportsPrimaryKeyInCreateStatements();
    
    /**
     * Whether the datastore supports {@link Statement#getGeneratedKeys()}
     * @return true if it is supported
     */
    boolean supportsStatementGetGeneratedKeys();
}