/**********************************************************************
Copyright (c) 2006 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.columninfo;

import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDOFatalDataStoreException;

/**
 * Represents the metadata of a specific table.  This class is
 * basically a data structure that makes accessing the JDBC column metadata
 * easier.  Each of the items returned by 
 * {@link java.sql.DatabaseMetaData#getTables(String,String,String,String[])}
 * is represented by a public field in this class.
 *
 * Subclasses of TableInfo can be created on a per-DBMS basis to supply missing
 * metadata or correct faulty metadata obtained from that DBMS's JDBC driver(s).
 *
 * @version $Revision$ 
 */
public class TableInfo
{
    /**
     * The table catalog, which may be <tt>null</tt>.
     */
    protected String tableCat;

    /**
     * The table schema, which may be <tt>null</tt>.
     */
    protected String tableSchem;

    /**
     * The table name.
     */
    protected String tableName;

    /**
     * Table type. Typical types are "TABLE", "VIEW", "SYSTEM TABLE", "GLOBAL TEMPORARY", "LOCAL TEMPORARY", "ALIAS", "SYNONYM".
     */
    protected String tableType;

    /**
     * explanatory comment on the table.
     */
    protected String remarks;
    
    /**
     * The local type catalog used by the data source, which may be <tt>null</tt>.
     */
    protected String typeCat;    
    
    /**
     * The local type schema used by the data source, which may be <tt>null</tt>.
     */
    protected String typeSchem;    

    /**
     * The local type name used by the data source, which may be <tt>null</tt>.
     */
    protected String typeName;

    /**
     * (SELF_REFERENCING_COL_NAME) The local name of the designated "identifier" column of a typed table, which may be <tt>null</tt>.
     */
    protected String selfRefColName;

    /**
     * Specifies how values in SELF_REFERENCING_COL_NAME are created. Values are "SYSTEM", "USER", "DERIVED", which may be <tt>null</tt>.
     */
    protected String refGenName;
    
    private int hash = 0;

    /**
     * Accessor for table remarks
     * @return The remarks on the t
     **/
    public String getRemarks()
    {
        return remarks;
    }

    /**
     * Accessor for type name
     * @return The type name for the column
     **/
    public String getTypeName()
    {
        return typeName;
    }

    /**
     * Accessor for table name
     * @return The table name for the column
     **/
    public String getTableName()
    {
        return tableName;
    }

    /**
     * Accessor for table catalog
     * @return The table catalog
     **/
    public String getTableCat()
    {
        return tableCat;
    }

    /**
     * Accessor for table schema
     * @return The table schema
     **/
    public String getTableSchem()
    {
        return tableSchem;
    }

    /**
     * Constructs a column information object from the current row of the given
     * result set.  The {@link ResultSet} object passed must have been obtained
     * from a call to java.sql.DatabaseMetaData.getColumns().
     *
     * <p>This method only retrieves the values from the current row; the caller
     * is required to advance to the next row with {@link ResultSet#next}.
     *
     * @param rs    The result set returned from 
     *              java.sql.DatabaseMetaData.getColumns().
     *
     * @exception JDOFatalDataStoreException
     *      if a column of column information could not be retrieved from the
     *      result set.
     */
    public TableInfo(ResultSet rs) throws JDOFatalDataStoreException
    {
        try
        {
            tableCat        = rs.getString(1);
            tableSchem      = rs.getString(2);
            tableName       = rs.getString(3);
            tableType       = rs.getString(4);
            remarks         = rs.getString(5);
            if (rs.getMetaData().getColumnCount()>5)
            {
	            typeCat         = rs.getString(6);
	            typeSchem       = rs.getString(7);
	            typeName        = rs.getString(8);
	            selfRefColName  = rs.getString(9);
	            refGenName      = rs.getString(10);
            }
        }
        catch (SQLException e)
        {
            throw new JDOFatalDataStoreException("Can't read JDBC metadata from result set", e);
        }

    }

    /**
     * Indicates whether some object is "equal to" this one.  Two <tt>ColumnInfo
     * </tt> objects are considered equal if their catalog, schema, table, and
     * column names are all equal.
     *
     * @param   obj     the reference object with which to compare
     *
     * @return  <tt>true</tt> if this object is equal to the obj argument;
     *          <tt>false</tt> otherwise.
     */
    public final boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof ColumnInfo))
        {
            return false;
        }

        ColumnInfo ci = (ColumnInfo)obj;

        return (tableCat   == null ? ci.tableCat   == null : tableCat.equals(ci.tableCat)) &&
            (tableSchem == null ? ci.tableSchem == null : tableSchem.equals(ci.tableSchem)) &&
            tableName.equals(ci.tableName);
    }

    /**
     * Returns a hash code value for this object.
     *
     * @return  a hash code value for this object.
     */
    public final int hashCode()
    {
        if (hash == 0)
        {
            hash = (tableCat   == null ? 0 : tableCat.hashCode()) ^
                 (tableSchem == null ? 0 : tableSchem.hashCode()) ^
                 tableName.hashCode();
        }

        return hash;
    }

    /**
     * Returns the string representation of this object.
     *
     * @return  string representation of this object.
     */
    public String toString()
    {
        StringBuffer str=new StringBuffer();
        str.append(this.getClass().getName() + "\n");
        str.append("  tableCat        = " + tableCat + "\n");
        str.append("  tableSchem      = " + tableSchem + "\n");
        str.append("  tableName       = " + tableName + "\n");
        str.append("  tableType       = " + tableType + "\n");
        str.append("  remarks         = " + remarks + "\n");
        str.append("  typeCat         = " + typeCat + "\n");
        str.append("  typeSchem       = " + typeSchem + "\n");
        str.append("  typeName        = " + typeName + "\n");
        str.append("  selfRefColName  = " + selfRefColName + "\n");
        str.append("  refGenName      = " + refGenName + "\n");

        return str.toString();
    }
}
