/**********************************************************************
Copyright (c) 2003 Mike Martin and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - split out AbstractRDBMSExtent
    ...
**********************************************************************/
package org.jpox.store.rdbms.extent;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.extent.AbstractExtent;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.query.TransientIDROF;
import org.jpox.util.Localiser;

/**
 * An Extent of all persistent objects backed by a view.
 * @version $Revision: 1.13 $
 */
public class ClassViewExtent extends AbstractExtent implements Queryable
{
    /** Localised messages */
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    private final DatastoreClass view;

    private final int fieldCount;
    private final int[] prefetchFieldNumbers;
	private final StatementExpressionIndex[] statementExpressionIndex;

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param view The view
     * @param cls Candidate class
     * @param subclasses Whether to include subclasses
     */
    public ClassViewExtent(PersistenceManager pm, DatastoreClass view, Class cls, boolean subclasses)
    {
        super(pm, cls, subclasses);

        this.view = view;

        // Set up list of fields required for the candidate class
        fieldCount = cmd.getNoOfManagedFields();
        int[] fn = new int[fieldCount];
		statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        int prefetchFieldCount = 0;

        for (int i=0; i<fieldCount; ++i)
        {
            JavaTypeMapping m = view.getFieldMapping(cmd.getManagedFieldAbsolute(i));
            if (m != null)
            {
                if (!m.includeInFetchStatement() || m instanceof MappingCallbacks)
                {
                    throw new JDOFatalInternalException(LOCALISER_RDBMS.msg("RDBMS.Extent.MappingInvalidForView",m, candidateClass.getName()));
                }
                
				statementExpressionIndex[i] = new StatementExpressionIndex();
				statementExpressionIndex[i].setMapping(m);
                fn[prefetchFieldCount++] = i;
            }
        }

        if (prefetchFieldCount == 0)
        {
            throw new JDOFatalUserException(LOCALISER_RDBMS.msg("RDBMS.Extent.ClassHasNoPersistentFields", candidateClass.getName()));
        }

        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);
    }

    /**
     * Create a new query to search for the candidate classes etc.
     * @return The new QueryStatement.
     */
    public QueryExpression newQueryStatement()
    {
        return storeMgr.getDatastoreAdapter().newQueryStatement(view,pm.getClassLoaderResolver());
    }

    /**
     * Create a query to search for the candidateClass and subclasses if true
     * @param candidateClass The class to use
     * @return The new QueryStatement
     */
    public QueryExpression newQueryStatement(Class candidateClass)
    {
        String extentType = view.getType();
        if (!extentType.equals(candidateClass.getName()))
        {
            throw new IncompatibleQueryElementTypeException(extentType, candidateClass.getName());
        }

        return storeMgr.getDatastoreAdapter().newQueryStatement(view,pm.getClassLoaderResolver());
    }

    /**
     * Create a new factory for objects from the ResultSet.
     * @param stmt The Query Statement
     * @param ignoreCache Whether to ignore dirty objects
     * @param resultClass Whether to create objects of a particular class
     * @param useFetchPlan whether to use the fetch plan to retrieve fields in the same query
     * @return The result object factory
     */
    public ResultObjectFactory newResultObjectFactory(QueryExpression stmt,boolean ignoreCache,Class resultClass, boolean useFetchPlan)
    {
        Mappings.selectMapping(stmt,statementExpressionIndex);

        return new TransientIDROF(getCandidateClass(), prefetchFieldNumbers, statementExpressionIndex);
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public boolean isEmpty()
    {
        // we don't know if we are empty until it calls the datastore
        return false;
    }
}