/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a Float RDBMS type.
 * 
 * @version $Revision: 1.6 $
 */
public class FloatRDBMSMapping extends DoubleRDBMSMapping
{
    /**
     * @param storeMgr Store Manager
     * @param mapping THe java type mapping
     */
    protected FloatRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    /**
     * Constructor.
     * @param mapping The java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public FloatRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    private void initialize()
    {
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.FLOAT);
    }
    
    public float getFloat(Object rs, int param)
    {
        float value;

        try
        {
            value = ((ResultSet) rs).getFloat(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","float","" + param, column, e.getMessage()), e);
        }

        return value;
    }
    
    public void setFloat(Object ps, int param, float value)
    {
        try
        {
            ((PreparedStatement) ps).setFloat(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","float","" + value, column, e.getMessage()), e);
        }
    }
    
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else if (value instanceof Integer)
            {
                ((PreparedStatement) ps).setFloat(param, ((Integer) value).floatValue());
            }
            else if (value instanceof Long)
            {
                ((PreparedStatement) ps).setFloat(param, ((Long) value).floatValue());
            }
            else if (value instanceof Float)
            {
                ((PreparedStatement) ps).setFloat(param, ((Float) value).floatValue());
            }
            else if (value instanceof Double)
            {
                ((PreparedStatement) ps).setDouble(param, ((Double) value).doubleValue());
            }
            else
            {
                ((PreparedStatement) ps).setFloat(param, ((Double) value).floatValue());
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam", "Object", "" + value, column, e.getMessage()), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        Object value;

        try
        {
            float d = ((ResultSet) rs).getFloat(param);
            if (getJavaTypeMapping().getJavaType().getName().equals(Integer.class.getName()))
            {
                value = ((ResultSet) rs).wasNull() ? null : new Integer((int) d);
            }
            else if (getJavaTypeMapping().getJavaType().getName().equals(Long.class.getName()))
            {
                value = ((ResultSet) rs).wasNull() ? null : new Long((long) d);
            }
            else if (getJavaTypeMapping().getJavaType().getName().equals(Float.class.getName()))
            {
                value = ((ResultSet) rs).wasNull() ? null : new Float(d);
            }
            else if (getJavaTypeMapping().getJavaType().getName().equals(Double.class.getName()))
            {
                double dbl = ((ResultSet)rs).getDouble(param);
                value = ((ResultSet) rs).wasNull() ? null : new Double(dbl);
            }
            else
            {
                value = ((ResultSet) rs).wasNull() ? null : new Double(d);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","Object","" + param, column, e.getMessage()), e);
        }

        return value;
    }    
}