/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2004 Andy Jefferson - localised messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.ClassNameConstants;
import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.SingleFieldMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a character RDBMS type.
 * 
 * @version $Revision: 1.17 $
 */
public class NumericRDBMSMapping extends ColumnMapping
{
    private static final int INT_MAX_DECIMAL_DIGITS = 10;

    /**
     * @param storeMgr Store Manager
     * @param mapping Java type mapping
     */
    protected NumericRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public NumericRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    /**
     * Initialise the mapping, setting any default precision.
     */
    private void initialize()
    {
        if (column != null)
        {
            // Valid Values
            if (getJavaTypeMapping() instanceof SingleFieldMapping)
            {
                Object[] validValues = ((SingleFieldMapping)getJavaTypeMapping()).getValidValues(0);
                if (validValues != null)
                {
                    applyValidValuesToColumn(column, validValues);
                }
            }

            /*
             * In case the default DECIMAL precision is less than the number of
             * digits we need, set it manually.
             */
            if (getJavaTypeMapping().getJavaType().getName().equals(ClassNameConstants.JAVA_LANG_INTEGER))
            {
                // If the user hasn't set the precision, set it to the size necessary for
                // the Java type
                if (column.getColumnMetaData().getLength() == null)
                {
                    column.getColumnMetaData().setLength(INT_MAX_DECIMAL_DIGITS);
                }
            }
            else if (getJavaTypeMapping().getJavaType().getName().equals(ClassNameConstants.JAVA_LANG_BOOLEAN))
            {
                column.getColumnMetaData().setLength(1);

                StringBuffer constraints = new StringBuffer("CHECK (" + column.getIdentifier() + " IN (1,0)");
                if (column.isNullable())
                {
                    constraints.append(" OR " + column.getIdentifier() + " IS NULL");
                }
                constraints.append(')');
                column.setConstraints(constraints.toString());
                column.checkDecimal();
            }
        }
        initTypeInfo();
    }

    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.NUMERIC);
    }

    public void setChar(Object ps, int param, char value)
    {
        try
        {
            ((PreparedStatement) ps).setInt(param,value);                 
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","char", "" + value, column, e.getMessage()), e);
        }
    }

    public char getChar(Object rs, int param)
    {
        char value;

        try
        {
            value = (char)((ResultSet) rs).getInt(param);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","char", "" + param, column, e.getMessage()), e);
        }
        return value;
    }
    
    public void setInt(Object ps, int param, int value)
    {
        try
        {
            ((PreparedStatement) ps).setInt(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","int", "" + value, column, e.getMessage()), e);
        }
    }

    public int getInt(Object rs, int param)
    {
        int value;

        try
        {
            value = ((ResultSet) rs).getInt(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","int", "" + param, column, e.getMessage()), e);
        }

        return value;
    }

    public void setByte(Object ps, int param, byte value)
    {
        try
        {
            //TODO bug with SQL SERVER DRIVER. It doesnt accept Byte -128
//          ps.setByte(param, value);
            ((PreparedStatement) ps).setInt(param, value);
            
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","byte", "" + value, column, e.getMessage()), e);
        }
    }

    public byte getByte(Object rs, int param)
    {
        byte value;

        try
        {
            value = ((ResultSet) rs).getByte(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","byte", "" + param, column, e.getMessage()), e);
        }

        return value;
    }
    
    public void setBoolean(Object ps, int param, boolean value)
    {
        try
        {
            ((PreparedStatement) ps).setBoolean(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","boolean", "" + value, column, e.getMessage()), e);
        }
    }

    public boolean getBoolean(Object rs, int param)
    {
        boolean value;

        try
        {
            value = ((ResultSet) rs).getBoolean(param);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","boolean", "" + param, column, e.getMessage()), e);
        }

        return value;
    }
    
    public void setShort(Object ps, int param, short value)
    {
        try
        {
            ((PreparedStatement) ps).setShort(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","short", "" + value,column, e.getMessage()), e);
        }
    }

    public short getShort(Object rs, int param)
    {
        short value;

        try
        {
            value = ((ResultSet) rs).getShort(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","short", "" + param, column, e.getMessage()), e);
        }

        return value;
    }
    
    public void setLong(Object ps, int param, long value)
    {
        try
        {
            ((PreparedStatement) ps).setLong(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","long", "" + value, column, e.getMessage()), e);
        }
    }

    public long getLong(Object rs, int param)
    {
        long value;

        try
        {
            value = ((ResultSet) rs).getLong(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMS.Mapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","long", "" + param, column, e.getMessage()), e);
        }

        return value;
    }  
    
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else if (value instanceof Byte)
            {
                ((PreparedStatement) ps).setInt(param, ((Byte)value).byteValue());
            }
            else if (value instanceof Integer)
            {
                ((PreparedStatement) ps).setInt(param, ((Integer)value).intValue());
            }
            else if (value instanceof Character)
            {
                String s = value.toString();
                ((PreparedStatement) ps).setInt(param,s.charAt(0));                  
            }
            else if (value instanceof String)
            {
                String s = (String)value;
                ((PreparedStatement) ps).setInt(param,s.charAt(0));                  
            }             
            else if (value instanceof Long)
            {
                ((PreparedStatement) ps).setLong(param, ((Long)value).longValue());
            }
            else if (value instanceof Short)
            {
                ((PreparedStatement) ps).setShort(param, ((Short)value).shortValue());
            }
            else if( value instanceof BigDecimal )
            {
                ((PreparedStatement) ps).setBigDecimal(param, (BigDecimal)value);
            }            
            else if( value instanceof Boolean )
            {
                ((PreparedStatement) ps).setBoolean(param, ((Boolean)value).booleanValue());
            }
            else
            {
                ((PreparedStatement) ps).setBigDecimal(param, new BigDecimal((BigInteger)value));
            }            
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToSetParam","Numeric", "" + value, column, e.getMessage()), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        try
        {
            BigDecimal value = ((ResultSet) rs).getBigDecimal(param);
            
            if (value == null)
            {
                return null;
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(BigInteger.class.getName()) )
            {
                return value.toBigInteger();
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Integer.class.getName()) )
            {
                return new Integer(value.intValue());
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Long.class.getName()) )
            {
                return new Long(value.longValue());
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Boolean.class.getName()) )
            {
                return new Boolean(value.intValue() == 1 ? true : false );
            }        
            else if( getJavaTypeMapping().getJavaType().getName().equals(Byte.class.getName()) )
            {
                return new Byte(value.byteValue());
            }            
            else if( getJavaTypeMapping().getJavaType().getName().equals(Short.class.getName()) )
            {
                return new Short(value.shortValue());
            }  
            else if( getJavaTypeMapping().getJavaType().getName().equals(Character.class.getName()))
            {
            	return new Character((char)value.intValue());
            }            
            else if( getJavaTypeMapping().getJavaType().getName().equals(String.class.getName()))
            {
            	return new Character((char)value.intValue()).toString();
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(BigDecimal.class.getName()))
            {
                return value;
            }
            else
            {
                return new Long(value.longValue());
            }            
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.Mapping.UnableToGetParam","Numeric", "" + param, column, e.getMessage()),e);
        }
    }  
}