/******************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
2004 Erik Bengtson - addition of JDBC, SQL type maps
2004 Andy Jefferson - fixed use of SQL/JDBC to find the one applicable to
                      the specified field and type.
    ...
*****************************************************************/
package org.jpox.store.rdbms.mapping;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.NullValue;
import org.jpox.plugin.ConfigurationElement;
import org.jpox.plugin.ExtensionPoint;
import org.jpox.plugin.PluginManager;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.AbstractMappingManager;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.DatastoreMappingFactory;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.sqlidentifier.RDBMSIdentifierFactory;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;
import org.jpox.util.MultiMap;
import org.jpox.util.StringUtils;

/**
 * Mapping manager for RDBMS datastores.
 * Provides mappings from standard Java types defined in org.jpox.store.mapping to
 * datastore mappings for JDBC types.
 * 
 * @version $Revision: 1.45 $
 */
public class RDBMSMappingManager extends AbstractMappingManager
{
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    MultiMap datastoreMappingsByJavaType;
    MultiMap datastoreMappingsByJDBCType;
    MultiMap datastoreMappingsBySQLType;

    /**
     * Constructor.
     */
    public RDBMSMappingManager()
    {
        super();
    }

    /**
     * Load all datastore mappings defined in the associated plugins.
     * We handle RDBMS datastore mappings so refer to rdbms-mapping-class, jdbc-type, sql-type in particular.
     * @param mgr the PluginManager
     * @param clr the ClassLoaderResolver
     * @param vendorId the datastore vendor id
     */
    public void loadDatastoreMapping(PluginManager mgr, ClassLoaderResolver clr, String vendorId)
    {
        if (datastoreMappingsByJavaType != null)
        {
            // Already loaded
            return;
        }

        datastoreMappingsByJDBCType = new MultiMap();
        datastoreMappingsBySQLType = new MultiMap();
        datastoreMappingsByJavaType = new MultiMap();

        ExtensionPoint exPoint = mgr.getExtensionPoint("org.jpox.store_datastoremapping");
        for (int i=0; i<exPoint.getExtensions().length; i++)
        {
            ConfigurationElement[] elms = exPoint.getExtensions()[i].getConfigurationElements();
            for (int e=0; e<elms.length; e++)
            {
                String javaName = elms[e].getAttribute("java-type").trim();
                String rdbmsMappingClassName = elms[e].getAttribute("rdbms-mapping-class");
                String jdbcType = elms[e].getAttribute("jdbc-type");
                String sqlType = elms[e].getAttribute("sql-type");
                String defaultJava = elms[e].getAttribute("default");
                String javaVersion = elms[e].getAttribute("java-version");
                String javaVersionRestricted = elms[e].getAttribute("java-version-restricted");

                boolean defaultForJavaType = false;
                if (defaultJava != null)
                {
                    if (defaultJava.equalsIgnoreCase("true"))
                    {
                        defaultForJavaType = Boolean.TRUE.booleanValue();
                    }
                }

                boolean javaRestricted = false;
                if (javaVersionRestricted != null)
                {
                    if (javaVersionRestricted.equalsIgnoreCase("true"))
                    {
                        javaRestricted = Boolean.TRUE.booleanValue();
                    }
                }

                if (javaVersion == null || javaVersion.length() < 1)
                {
                    // Default to use for all JDKs 1.3+
                    javaVersion = "1.3";
                }

                if ((JavaUtils.isGreaterEqualsThan(javaVersion) && !javaRestricted) ||
                    (JavaUtils.isEqualsThan(javaVersion) && javaRestricted))
                {
                    Class mappingType = null;
                    if (!StringUtils.isWhitespace(rdbmsMappingClassName))
                    {
                        try
                        {
                            mappingType = mgr.loadClass(elms[e].getExtension().getPlugin().getId(), rdbmsMappingClassName);
                        }
                        catch (JDOException jdoe)
                        {
                            JPOXLogger.METADATA.error(LOCALISER.msg("Mapping.DatastoreMappingClassNotFound", rdbmsMappingClassName));
                        }
                        Set includes = new HashSet();
                        Set excludes = new HashSet();
                        ConfigurationElement[] childElm = elms[e].getChildren();
                        for (int j=0; j<childElm.length; j++)
                        {
                            if (childElm[j].getName().equals("includes"))
                            {
                                includes.add(childElm[j].getAttribute("vendor-id"));
                            }
                            else if (childElm[j].getName().equals("excludes"))
                            {
                                excludes.add(childElm[j].getAttribute("vendor-id"));
                            }
                        }

                        if (!excludes.contains(vendorId))
                        {
                            if (includes.isEmpty() || includes.contains(vendorId))
                            {
                                registerDatastoreMapping(javaName, mappingType, jdbcType, sqlType, defaultForJavaType);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Utility to register a datastore mapping for a java type, and the SQL/JDBC types it can be mapped to.
     * This can also be called to change the default setting of a mapping - just supply the same
     * values of java/JDBC/SQL types and a different default value
     * @param javaTypeName Name of the java type
     * @param datastoreMappingType The datastore mapping
     * @param jdbcType The JDBC type that can be used
     * @param sqlType The SQL type that can be used
     * @param dflt Whether this type should be used as the default mapping for this Java type
     */
    public void registerDatastoreMapping(String javaTypeName, Class datastoreMappingType, String jdbcType, String sqlType, boolean dflt)
    {
        boolean mappingRequired = true;
        Collection coll = (Collection)datastoreMappingsByJavaType.get(javaTypeName);
        if (coll != null && coll.size() > 0)
        {
            Iterator collIter = coll.iterator();
            while (collIter.hasNext())
            {
                RDBMSTypeMapping typeMapping = (RDBMSTypeMapping)collIter.next();
                if (typeMapping.getJdbcType().equals(jdbcType) && typeMapping.getSqlType().equals(sqlType))
                {
                    // Same mapping
                    mappingRequired = false;
                    if (typeMapping.isDefault() != dflt)
                    {
                        typeMapping.setDefault(dflt);
                    }
                }
                else if (dflt)
                {
                    // New one should be default so change all others not to be
                    typeMapping.setDefault(false);
                }
            }
        }

        // Add the mapping if not already existing
        if (mappingRequired)
        {
            RDBMSTypeMapping mapping = new RDBMSTypeMapping(datastoreMappingType, dflt, javaTypeName, jdbcType, sqlType);
            datastoreMappingsByJDBCType.put(jdbcType,mapping);
            datastoreMappingsBySQLType.put(sqlType,mapping);
            datastoreMappingsByJavaType.put(javaTypeName,mapping);
        }
    }

    /**
     * Utility to deregister all mappings for a JDBC type.
     * @param jdbcTypeName The JDBC type name
     */
    public void deregisterDatastoreMappingsForJDBCType(String jdbcTypeName)
    {
        // Find the mappings that we have for this JDBC type
        Collection coll = (Collection)datastoreMappingsByJDBCType.get(jdbcTypeName);
        if (coll == null || coll.size() == 0)
        {
            return;
        }

        // Take a copy so we have no ConcurrentModificationException issues
        Collection mappings = new HashSet(coll);

        // Delete the mapping from the 3 maps
        Iterator iter = mappings.iterator();
        while (iter.hasNext())
        {
            RDBMSTypeMapping mapping = (RDBMSTypeMapping)iter.next();
            datastoreMappingsByJavaType.remove(mapping.getJavaType(), mapping);
            datastoreMappingsBySQLType.remove(mapping.getSqlType(), mapping);
            datastoreMappingsByJDBCType.remove(mapping.getJdbcType(), mapping);
        }
    }

    /**
     * Accessor for a datastore mapping class for the specified java type 
     * (and optional jdbc type or sql type).
     * @param javaType The java type
     * @param jdbcType The JDBC type
     * @param sqlType The SQL Type
     * @param clr ClassLoader resolver to use
     * @return The datastore mapping class
     */
    protected Class getDatastoreMappingClass(String javaType, String jdbcType, String sqlType, ClassLoaderResolver clr)
    {
        RDBMSTypeMapping datastoreMapping = null;

        if (sqlType != null)
        {
            // First look for "sql-type"
            if (datastoreMappingsBySQLType.get(sqlType.toUpperCase()) == null)
            {
                throw new JDOFatalException(LOCALISER_RDBMS.msg("RDBMS.MappingManager.UnknownSQLTypeForJavaType", javaType,
                    sqlType));
            }

            // Find if this sql-type has been defined for this java-type
            Iterator sqlTypeIter = ((Collection) datastoreMappingsBySQLType.get(sqlType.toUpperCase())).iterator();
            while (sqlTypeIter.hasNext())
            {
                RDBMSTypeMapping sqlTypeMapping = (RDBMSTypeMapping)sqlTypeIter.next();
                if (sqlTypeMapping.getJavaType().equals(javaType))
                {
                    datastoreMapping = sqlTypeMapping;
                    break;
                }
            }
        }
        else if (jdbcType != null)
        {
            // Then look for "jdbc-type"
            if (datastoreMappingsByJDBCType.get(jdbcType.toUpperCase()) == null)
            {
                throw new JDOFatalException(LOCALISER_RDBMS.msg("RDBMS.MappingManager.UnknownJDBCTypeForJavaType", javaType, jdbcType));
            }

            // Find if this jdbc-type has been defined for this java-type
            Iterator jdbcTypeIter = ((Collection) datastoreMappingsByJDBCType.get(jdbcType.toUpperCase())).iterator();
            while (jdbcTypeIter.hasNext())
            {
                RDBMSTypeMapping jdbcTypeMapping = (RDBMSTypeMapping)jdbcTypeIter.next();
                if (jdbcTypeMapping.getJavaType().equals(javaType))
                {
                    datastoreMapping = jdbcTypeMapping;
                    break;
                }
            }
        }

        if (datastoreMapping == null)
        {
            // No specified type so get the best for this java-type (if primitve then use the wrapper java-type)
            // Use wrapper type instead of primitive type since primitives have no java-type entries for datastore mappings
            String type = ClassUtils.getWrapperTypeNameForPrimitiveTypeName(javaType);
            Collection mappings = (Collection)datastoreMappingsByJavaType.get(type);
            if (mappings == null)
            {
                // This java-type isnt specifically supported so maybe its superclass is
                Class javaTypeClass = clr.classForName(type);
                Class superClass = javaTypeClass.getSuperclass();
                while (superClass != null && !superClass.getName().equals(java.lang.Object.class.getName()) && mappings == null)
                {
                    mappings = (Collection) datastoreMappingsByJavaType.get(superClass.getName());
                    superClass = superClass.getSuperclass();
                }
            }
            if (mappings != null)
            {
                if (mappings.size() == 1)
                {
                    datastoreMapping = (RDBMSTypeMapping) mappings.iterator().next();
                }
                else
                {
                    // More than 1 so take the default
                    Iterator mappingsIter = mappings.iterator();
                    while (mappingsIter.hasNext())
                    {
                        RDBMSTypeMapping rdbmsMapping = (RDBMSTypeMapping)mappingsIter.next();
                        if (rdbmsMapping.isDefault())
                        {
                            // default matched so take it
                            datastoreMapping = rdbmsMapping;
                            break;
                        }
                    }

                    // No default set, so use the first one
                    if (datastoreMapping == null && mappings.size() > 0)
                    {
                        datastoreMapping = (RDBMSTypeMapping) mappings.iterator().next();
                    }
                }
            }
        }

        if (datastoreMapping == null)
        {
            throw new JDOFatalException(LOCALISER_RDBMS.msg("RDBMS.MappingManager.UnsupportedJavaType", javaType, jdbcType, sqlType));
        }
        return datastoreMapping.getMappingType();
    }

    protected class RDBMSTypeMapping extends TypeMapping
    {
        private String javaType;
        private String jdbcType;
        private String sqlType;

        /**
         * @param mappingType Datastroe mapping type
         * @param isDefault Whether is the default for this java type
         * @param javaType the java type
         * @param jdbcType the jdbc type
         * @param sqlType the sql type
         */
        public RDBMSTypeMapping(Class mappingType, boolean isDefault, String javaType, String jdbcType, String sqlType)
        {
            super(mappingType, isDefault);
            this.javaType = javaType;
            this.jdbcType = jdbcType;
            this.sqlType = sqlType;
        }

        /**
         * @return Returns the javaType.
         */
        public String getJavaType()
        {
            return javaType;
        }

        /**
         * @return Returns the jdbcType.
         */
        public String getJdbcType()
        {
            return jdbcType;
        }

        /**
         * @return Returns the sqlType.
         */
        public String getSqlType()
        {
            return sqlType;
        }
    }

    /**
     * Method to create the datastore mapping for a java type mapping at a particular index.
     * @param mapping The java mapping
     * @param fmd MetaData for the field
     * @param index Index of the datastore field
     * @param storeMgr Store Manager
     * @param column The column
     * @return The datastore mapping
     */
    public DatastoreMapping createDatastoreMapping(JavaTypeMapping mapping, AbstractPropertyMetaData fmd, int index, 
            StoreManager storeMgr, DatastoreField column)
    {
        Class datastoreMappingClass = null;

        ClassLoaderResolver clr = storeMgr.getPMFContext().getClassLoaderResolver(null);
        if (fmd.getColumnMetaData().length > 0)
        {
            // Use "datastore-mapping-class" extension if provided
            if (fmd.getColumnMetaData()[index].hasExtension("datastore-mapping-class"))
            {
                datastoreMappingClass = clr.classForName(fmd.getColumnMetaData()[index].getValueForExtension("datastore-mapping-class"));
            }
        }
        if (datastoreMappingClass == null)
        {
            String javaType = mapping.getJavaTypeForDatastoreMapping(index);
            String jdbcType = null;
            String sqlType = null;
            if (fmd.getColumnMetaData().length > 0)
            {
                // Utilise the jdbc and sql types if specified
                jdbcType = fmd.getColumnMetaData()[index].getJdbcType();
                sqlType = fmd.getColumnMetaData()[index].getSqlType();
            }
            if (fmd.isSerialized())
            {
                javaType = java.io.Serializable.class.getName();
            }
            datastoreMappingClass = getDatastoreMappingClass(javaType, jdbcType, sqlType, clr);
        }

        return DatastoreMappingFactory.createMapping(datastoreMappingClass, mapping, storeMgr, column);
    }

    /**
     * Method to create the datastore mapping for a particular column and java type.
     * @param mapping The java mapping
     * @param storeMgr Store Manager
     * @param column The column
     * @param javaType The java type (isnt this stored in the java mapping ?)
     * @return The datastore mapping
     */
    public DatastoreMapping createDatastoreMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField column, String javaType)
    {
        ClassLoaderResolver clr = storeMgr.getPMFContext().getClassLoaderResolver(null);
        Column col = (Column)column;
        String jdbcType = null;
        String sqlType = null;
        if (col != null && col.getColumnMetaData() != null)
        {
            // Utilise the jdbc and sql types if specified
            jdbcType = col.getColumnMetaData().getJdbcType();
            sqlType = col.getColumnMetaData().getSqlType();
        }
        Class datastoreMappingClass = getDatastoreMappingClass(javaType, jdbcType, sqlType, clr);

        return DatastoreMappingFactory.createMapping(datastoreMappingClass, mapping, storeMgr, column);
    }

    /**
     * Method to create a datastore field for a Java type mapping.
     * This is NOT used for PersistenceCapable mappings - see method below.
     * @param fmd MetaData for the field
     * @param datastoreContainer The container to add the datastore field to
     * @param dba Datastore adapter
     * @param mapping Java type mapping for the field
     * @param javaType The type of field being stored in this column
     * @param datastoreFieldIndex Index of the datastore field to use
     * @return The datastore field
     */
    public DatastoreField createDatastoreField(AbstractPropertyMetaData fmd,
                                               DatastoreContainerObject datastoreContainer,
                                               DatastoreAdapter dba,
                                               JavaTypeMapping mapping,
                                               String javaType,
                                               int datastoreFieldIndex)
    {
        Column col;
        ColumnMetaData colmd = null;
        if (fmd.getColumnMetaData().length > datastoreFieldIndex)
        {
            colmd = fmd.getColumnMetaData()[datastoreFieldIndex];
        }
        else
        {
            // If column specified add one (use any column name specified on field element)
            colmd = new ColumnMetaData(fmd, fmd.getColumn());
            fmd.addColumn(colmd);
        }

        // Generate the column identifier
        IdentifierFactory idFactory = datastoreContainer.getStoreManager().getIdentifierFactory();
        DatastoreIdentifier identifier = null;
        if (colmd.getName() == null)
        {
            // No name specified, so generate the identifier from the field name
			identifier = idFactory.newIdentifier(IdentifierFactory.COLUMN, fmd.getName());
			int i=0;
			while (datastoreContainer.hasDatastoreField(identifier))
			{
				identifier = idFactory.newIdentifier(IdentifierFactory.COLUMN, fmd.getName() + "_" + i);
                i++;
			}

            colmd.setName(identifier.getIdentifier());
        }
        else
        {
            // User has specified a name, so try to keep this unmodified
            identifier = idFactory.newDatastoreFieldIdentifier(fmd.getColumnMetaData()[datastoreFieldIndex].getName(), 
                datastoreContainer.getStoreManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(fmd.getType()), 
                DatastoreField.ROLE_CUSTOM);
        }

        // Create the column
        col = (Column) datastoreContainer.addDatastoreField(javaType, identifier, mapping, colmd);

        if (fmd.isPrimaryKey())
        {
            col.setAsPrimaryKey();
        }

        if (datastoreContainer.getStoreManager().isStrategyDatastoreAttributed(fmd.getValueStrategy()))
        {
            if (datastoreContainer instanceof DatastoreClass)
            {
                if ((fmd.isPrimaryKey() && ((DatastoreClass)datastoreContainer).isBaseDatastoreClass()) || !fmd.isPrimaryKey())
                {
                    // Increment any PK field if we are in base class, and increment any other field
                    col.setAutoIncrement(true);
                }
            }
        }

        if (fmd.getValueForExtension("select-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("select-function"),Column.WRAPPER_FUNCTION_SELECT);
        }
        if (fmd.getValueForExtension("insert-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("insert-function"),Column.WRAPPER_FUNCTION_INSERT);
        }
        if (fmd.getValueForExtension("update-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("update-function"),Column.WRAPPER_FUNCTION_UPDATE);
        }

        setDatastoreFieldNullability(fmd,colmd,col);

        if (fmd.getNullValue() == NullValue.DEFAULT)
        {
            col.setDefaultable();
        }

        return col;
    }

    /**
     * Method to create a datastore field for a Java type mapping.
     * This is used for serialised elements/keys/values in a join table.
     * @param fmd MetaData for the field
     * @param datastoreContainer The container to add the datastore field to
     * @param dba Datastore adapter
     * @param mapping Java type mapping for the field
     * @param javaType The type of field being stored in this column
     * @param colmd MetaData for the column
     * @return The datastore field
     */
    public DatastoreField createDatastoreField(AbstractPropertyMetaData fmd,
                                               DatastoreContainerObject datastoreContainer,
                                               DatastoreAdapter dba,
                                               JavaTypeMapping mapping,
                                               String javaType,
                                               ColumnMetaData colmd)
    {
        Column col;
        if (colmd == null)
        {
            // If column specified add one (use any column name specified on field element)
            colmd = new ColumnMetaData(fmd, fmd.getColumn());
            fmd.addColumn(colmd);
        }

        IdentifierFactory idFactory = datastoreContainer.getStoreManager().getIdentifierFactory();
        if (colmd.getName() == null)
        {
            // No name specified, so generate the identifier from the field name
            DatastoreIdentifier identifier = idFactory.newIdentifier(IdentifierFactory.COLUMN, fmd.getName());
            int i=0;
            while (datastoreContainer.hasDatastoreField(identifier))
            {
                identifier = idFactory.newIdentifier(IdentifierFactory.COLUMN, fmd.getName() + "_" + i);
                i++;
            }

            colmd.setName(identifier.getIdentifier());
            col = (Column) datastoreContainer.addDatastoreField(javaType, identifier, mapping, colmd);
        }
        else
        {
            // User has specified a name, so try to keep this unmodified
            col = (Column) datastoreContainer.addDatastoreField(javaType, 
                idFactory.newDatastoreFieldIdentifier(colmd.getName(), 
                    datastoreContainer.getStoreManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(fmd.getType()), 
                    DatastoreField.ROLE_CUSTOM), 
                mapping, colmd);
        }

        setDatastoreFieldNullability(fmd,colmd,col);

        if (fmd.getNullValue() == NullValue.DEFAULT)
        {
            col.setDefaultable();
        }

        return col;
    }

    /**
     * Method to create a datastore field for a PersistenceCapable mapping.
     * @param fmd MetaData for the field whose mapping it is
     * @param datastoreContainer Datastore class where we create the datastore field
     * @param dba The datastore adapter
     * @param mapping The Java type for this field
     * @param colmd The columnMetaData for this datastore field
     * @param reference The datastore field we are referencing
     * @param clr ClassLoader resolver
     * @return The DatastoreField
     */
    public DatastoreField createDatastoreField(AbstractPropertyMetaData fmd,
                                               DatastoreContainerObject datastoreContainer,
                                               DatastoreAdapter dba,
                                               JavaTypeMapping mapping,
                                               ColumnMetaData colmd,
                                               DatastoreField reference,
                                               ClassLoaderResolver clr)
    {
        IdentifierFactory idFactory = datastoreContainer.getStoreManager().getIdentifierFactory();
        DatastoreIdentifier identifier = null;
        if (colmd.getName() == null)
        {
            // No name specified, so generate the identifier from the field name
            identifier = ((RDBMSIdentifierFactory)idFactory).newForeignKeyFieldIdentifier(fmd.getRelatedFieldMetaData(clr), fmd, reference.getIdentifier(), datastoreContainer.getStoreManager().getPMFContext().getTypeManager().isDefaultEmbeddedType(fmd.getType()), 
                DatastoreField.ROLE_OWNER);
            colmd.setName(identifier.getIdentifier());
        }
        else
        {
            // User has specified a name, so try to keep this unmodified
            identifier = idFactory.newDatastoreFieldIdentifier(colmd.getName(), false, DatastoreField.ROLE_CUSTOM);
        }
        Column col = (Column)datastoreContainer.addDatastoreField(fmd.getType().getName(), identifier, mapping, colmd);

        // Copy the characteristics of the reference column to this one
        reference.copyConfigurationTo(col);

        if (fmd.isPrimaryKey())
        {
            col.setAsPrimaryKey();
        }

        if (datastoreContainer.getStoreManager().isStrategyDatastoreAttributed(fmd.getValueStrategy()))
        {
            if ((fmd.isPrimaryKey() && ((DatastoreClass)datastoreContainer).isBaseDatastoreClass()) || !fmd.isPrimaryKey())
            {
                // Increment any PK field if we are in base class, and increment any other field
                col.setAutoIncrement(true);
            }
        }

        if (fmd.getValueForExtension("select-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("select-function"),Column.WRAPPER_FUNCTION_SELECT);
        }
        if (fmd.getValueForExtension("insert-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("insert-function"),Column.WRAPPER_FUNCTION_INSERT);
        }
        if (fmd.getValueForExtension("update-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("update-function"),Column.WRAPPER_FUNCTION_UPDATE);
        }

        setDatastoreFieldNullability(fmd,colmd,col);

        if (fmd.getNullValue() == NullValue.DEFAULT)
        {
            col.setDefaultable();
        }

        return col;        
    }

    /**
     * Sets the DatastoreField nullability based on metadata configuration.
     * <P>
     * Configuration is taken in this order:
     * <UL>
     * <LI>ColumnMetaData (allows-null)</LI>
     * <LI>AbstractPropertyMetaData (null-value)</LI>
     * <LI>Field type (primitive does not allows null)</LI>
     * </UL>
     * @param fmd the AbstractPropertyMetaData
     * @param colmd the ColumnMetaData
     * @param field the DatastoreField
     */
    private void setDatastoreFieldNullability(AbstractPropertyMetaData fmd, ColumnMetaData colmd, DatastoreField field)
    {
        // Provide defaults for "allows-null" in ColumnMetaData
        if (colmd != null && !colmd.isAllowsNullSet())
        {
            if (fmd.isPrimaryKey())
            {
                colmd.setAllowsNull(new Boolean(false));
            }
            else if (!fmd.getType().isPrimitive() && fmd.getNullValue() != NullValue.EXCEPTION)
            {
                colmd.setAllowsNull(new Boolean(true));
            }
            else
            {
                colmd.setAllowsNull(new Boolean(false));
            }
            if (colmd.isAllowsNull())
            {
                field.setNullable();
            }
        }
        // Set the nullability of the column in the datastore
        else if (colmd != null && colmd.isAllowsNullSet())
        {
            if (colmd.isAllowsNull())
            {
                field.setNullable();
            }
        }
        else if (fmd.isPrimaryKey())
        {
            // field is never nullable
        }
        else if (!fmd.getType().isPrimitive() && fmd.getNullValue() != NullValue.EXCEPTION)
        {
            field.setNullable();
        }
    }
}