/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


Contributors:
2005 Andy Jefferson - added support for bidirectional iterator
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;
import java.util.NoSuchElementException;

import javax.jdo.FetchPlan;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalException;

import org.jpox.PersistenceManager;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.util.SQLWarnings;
import org.jpox.util.WeakValueMap;

/**
 * Lazy collection results from a Query with the ResultSet insensitive to direction.
 * The actual result elements are only loaded when accessed. Once a result
 * has been loaded it will be stored in a local cache for any subsequent
 * access.
 *
 * @version $Revision: 1.2 $
 */
public final class InsensitiveQueryResult extends AbstractQueryResult
    implements QueryResult, java.io.Serializable
{
    /** Size of the ResultSet. */
    private final int size;

    /** Map of ResultSet object values, keyed by the list index ("0", "1", etc). */
    private Map resultsObjsByIndex = new WeakValueMap();

    /**
     * Constructor of the result from a Query. 
     * @param qs The Query Statement
     * @param query The Query
     * @param rof The factory to retrieve results from
     * @param rs The ResultSet from the Query Statement
     * @param candidates the Candidates collection. Pass this argument only when distinct = false
     */
    public InsensitiveQueryResult(QueryExpression qs, 
                                  Query query,
                                  ResultObjectFactory rof,
                                  ResultSet rs,
                                  Collection candidates)
    {
        super(qs, query, rof, rs);

        if (candidates != null)
        {
            //TODO support this feature
            throw new JDOFatalException("Unsupported Feature: Candidate Collection is only allowed using ForwardQueryResults");
        }
        
        int fetchSize = query.getFetchPlan().getFetchSize();

        // Set the list size from the ResultSet
        try
        {
            // Set the list size based on the last row number in the ResultSet
            boolean hasLast = rs.last();
            if (!hasLast)
            {
                // No results in ResultSet
                size = 0;
            }
            else
            {
                // ResultSet has results so return the row number as the list size
                size = rs.getRow();
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("QueryResult.ReadError",e));
        }

        // Load up result objects as per "fetchSize" requirements
        if (!query.getPersistenceManager().currentTransaction().isActive() ||
            query.getPersistenceManager().currentTransaction().getOptimistic() ||
            fetchSize == FetchPlan.FETCH_SIZE_GREEDY)
        {
            for (int i=0;i<size;i++)
            {
                getObjectForIndex(i);
            }
        }
        else if (fetchSize > 0)
        {
            for (int i=0;i<fetchSize;i++)
            {
                getObjectForIndex(i);
            }
        }
    }

    /**
     * Accessor for the result object at an index.
     * If the object has already been processed will return that object,
     * otherwise will retrieve the object using the factory.
     * @param index The list index position
     * @return The result object
     */
    protected Object getObjectForIndex(int index)
    {
        Object obj = resultsObjsByIndex.get("" + index);
        if (obj != null)
        {
            // Already retrieved so return it
            return obj;
        }

        try
        {
            // ResultSet is numbered 1, 2, ... N
            // List is indexed 0, 1, 2, ... N-1
            rs.absolute(index+1);
            obj = rof.getObject((PersistenceManager)query.getPersistenceManager(), rs, query.getCandidateClass());
            SQLWarnings.log(rs);

            // Put it in our cache, keyed by the list index
            resultsObjsByIndex.put("" + index, obj);
            return obj;
        }
        catch (SQLException sqe)
        {
            throw new JDODataStoreException(LOCALISER.msg("QueryResult.ReadError", sqe));
        }
    }

    /**
     * Method to close the results.
     */
    public synchronized void close()
    {
        resultsObjsByIndex.clear();
        super.close();
    }

    /**
     * Inform the query result that the connection is being closed so perform
     * any operations now, or rest in peace.
     */
    public void closingConnection()
    {
        boolean loadAllResults = true;
        Object value = query.getExtension("org.jpox.query.loadResultsAtCommit");
        if (value != null && ((String)value).equalsIgnoreCase("false"))
        {
            // User has turned off loading of results at commit
            loadAllResults = false;
        }

        // Make sure all rows are loaded.
        if (loadAllResults)
        {
            for (int i=0;i<size();i++)
            {
                getObjectForIndex(i);
            }
        }
    }

    // ------------------------- Implementation of List methods ----------------------

    /**
     * Accessor for an iterator for the results.
     * @return The iterator
     */
    public Iterator iterator()
    {
        return new QueryResultIterator();
    }

    /**
     * Accessor for an iterator for the results.
     * @return The iterator
     */
    public ListIterator listIterator()
    {
        return new QueryResultIterator();
    }

    /**
     * An Iterator results of a pm.query.execute().iterator()
     */
    private class QueryResultIterator implements ListIterator
    {
        private int iterRowNum = 0; // The index of the next object

        /**
         * Constructor
         */
        public QueryResultIterator()
        {
        }

        public void add(Object obj)
        {
            throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
        }

        public boolean hasNext()
        {
            synchronized (InsensitiveQueryResult.this)
            {
                if (!isOpen())
                {
                    // Spec 14.6.7 Calling hasNext() on closed Query will return false
                    return false;
                }

                // When we aren't at size()-1 we have at least one more element
                return (iterRowNum <= (size() - 1));
            }
        }

        public boolean hasPrevious()
        {
            synchronized (InsensitiveQueryResult.this)
            {
                if (!isOpen())
                {
                    // Spec 14.6.7 Calling hasPrevious() on closed Query will return false
                    return false;
                }

                // A List has indices starting at 0 so when we have > 0 we have a previous
                return (iterRowNum > 0);
            }
        }

        public Object next()
        {
            synchronized (InsensitiveQueryResult.this)
            {
                if (!isOpen())
                {
                    // Spec 14.6.7 Calling next() on closed Query will throw NoSuchElementException
                    throw new NoSuchElementException(LOCALISER.msg("QueryResult.Closed"));
                }

                if (!hasNext())
                {
                    throw new NoSuchElementException("No next element");
                }
                Object obj = getObjectForIndex(iterRowNum);
                iterRowNum++;

                return obj;
            }
        }

        public int nextIndex()
        {
            if (hasNext())
            {
                return iterRowNum;
            }
            return size();
        }

        public Object previous()
        {
            synchronized (InsensitiveQueryResult.this)
            {
                if (!isOpen())
                {
                    // Spec 14.6.7 Calling previous() on closed Query will throw NoSuchElementException
                    throw new NoSuchElementException(LOCALISER.msg("QueryResult.Closed"));
                }

                if (!hasPrevious())
                {
                    throw new NoSuchElementException("No previous element");
                }

                iterRowNum--;
                return getObjectForIndex(iterRowNum);
            }
        }

        public int previousIndex()
        {
            if (iterRowNum == 0)
            {
                return -1;
            }
            return iterRowNum-1;
        }

        public void remove()
        {
            throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
        }

        public void set(Object obj)
        {
            throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
        }
    }

    /**
     * Equality operator for QueryResults.
     * Overrides the AbstractList implementation since that uses 
     * size() and iterator() and that would cause problems when closed.
     * @param o The object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object o)
    {
        if (o == null || !(o instanceof InsensitiveQueryResult))
        {
            return false;
        }
        InsensitiveQueryResult other = (InsensitiveQueryResult)o;
        
        // Just base equality on the QueryStatement for now
        // NB For SQL queries the QueryStatement will be null so use Query
        if (qs != null)
        {
            return other.qs == qs;
        }
        else
        {
            return other.query == query;
        }
    }

    /**
     * Method to retrieve a particular element from the list.
     * @param index The index of the element
     * @return The element at index
     */
    public synchronized Object get(int index)
    {
        assertIsOpen();
        return getObjectForIndex(index);
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * Checks if the result set is empty.
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public synchronized boolean isEmpty()
    {
        assertIsOpen();
        return size() < 1;
    }

    /**
     * Method to return the size of the result set.
     * @return The size of the result set.
     */
    public synchronized int size()
    {
        assertIsOpen();
        return size;
    }
}