/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.util.StringTokenizer;

import javax.jdo.JDOUserException;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Parser for handling JDOQL Single-String queries.
 * Takes a JDOQLQuery and the query string and parses it into its constituent parts, updating
 * the JDOQLQuery accordingly with the result that after calling the parse() method the JDOQLQuery
 * is populated.
 * <pre>
 * select [unique] [ <result> ] [into <result-class-name>]
 *                              [from <candidate-class-name> [exclude subclasses] ]
 *                              [where <filter>]
 *                              [variables <variables-clause> ]
 *                              [parameters <parameters-clause>]
 *                              [<imports-clause>]
 *                              [group by <grouping-clause> ]
 *                              [order by <ordering-clause>]
 *                              [range <from-range> ,<to-range>]                                       
 * </pre>
 * @version $Revision: 1.1 $
 */
public class JDOQLSingleStringParser
{    
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    /** The JDOQL query to populate. */
    private JDOQLQuery query;

    /** The single-string query string. */
    private String queryString;

    /** Record of the keyword currently being processed, so we can check for out of order keywords. */
    int keywordPosition = -1;

    /**
     * Constructor for the Single-String parser.
     * @param query The query
     * @param queryString The Single-String query
     */
    public JDOQLSingleStringParser(JDOQLQuery query, String queryString)
    {
        JPOXLogger.JDO_QUERY.debug("JDOQL Single-String with \"" + queryString + "\"");
        this.query = query;
        this.queryString = queryString;
    }

    /**
     * Method to parse the Single-String query
     */
    public void parse()
    {
        new Compiler(new Parser(queryString)).compile();
    }

    /**
     * Method to detect whether this token is a keyword for JDOQL Single-String
     * @param token The token
     * @return Whether it is a keyword
     */
    private boolean isKeyword(String token)
    {
        // Compare the passed token against the provided keyword list, or their lowercase form
        for (int i=0;i<JDOQLQuery.SINGLE_STRING_KEYWORDS.length;i++)
        {
            if (token.equals(JDOQLQuery.SINGLE_STRING_KEYWORDS[i]) || token.equals(JDOQLQuery.SINGLE_STRING_KEYWORDS_LOWERCASE[i]))
            {
                return true;
            }
        }
        if (token.equals("IMPORT") || token.equals("import"))
        {
            return true;
        }
        return false;
    }
    
    /**
     * Compiler to process keywords contents. In the query the keywords often have
     * content values following them that represent the constituent parts of the query. This takes the keyword
     * and sets the constituent part accordingly.
     */
    private class Compiler
    {
        Parser tokenizer;

        Compiler(Parser tokenizer)
        {
            this.tokenizer = tokenizer;
        }

        private void compile()
        {
            compileSelect();
            // any keyword after compiling the SELECT is an error
            String keyword = tokenizer.parseKeyword();
            if (keyword != null)
            {
                if (isKeyword(keyword))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.QueryContainsOutOfOrderKeyword", keyword));
                }
                else
                {
                    // unexpected token
                }
            }
        }

        private void compileSelect()
        {
            if (!tokenizer.parseKeyword("SELECT") && !tokenizer.parseKeyword("select"))
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.NoSelectStart"));
            }
            if (tokenizer.parseKeyword("UNIQUE") || tokenizer.parseKeyword("unique"))
            {
                compileUnique();
            }
            compileResult();
            if (tokenizer.parseKeyword("INTO") || tokenizer.parseKeyword("into"))
            {
                compileInto();
            }
            if (tokenizer.parseKeyword("FROM") || tokenizer.parseKeyword("from"))
            {
                compileFrom();
            }
            if (tokenizer.parseKeyword("WHERE") || tokenizer.parseKeyword("where"))
            {
                compileWhere();
            }
            if (tokenizer.parseKeyword("VARIABLES") || tokenizer.parseKeyword("variables"))
            {
                compileVariables();
            }
            if (tokenizer.parseKeyword("PARAMETERS") || tokenizer.parseKeyword("parameters"))
            {
                compileParameters();
            }
            if (tokenizer.parseKeyword("import"))
            {
                compileImport();
            }
            if (tokenizer.parseKeyword("GROUP") || tokenizer.parseKeyword("group"))
            {
                compileGroup();
            }
            if (tokenizer.parseKeyword("ORDER") || tokenizer.parseKeyword("order"))
            {
                compileOrder();
            }
            if (tokenizer.parseKeyword("RANGE") || tokenizer.parseKeyword("range"))
            {
                compileRange();
            }
        }
        
        private void compileUnique()
        {
            query.setUnique(true);
        }

        private void compileResult()
        {
            String content = tokenizer.parseContent();
            //content may be empty
            if (content.length() > 0)
            {
                query.setResult(content);
            }
        }

        private void compileInto()
        {
            String content = tokenizer.parseContent();
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "INTO", "<result class>"));
            }

            query.setResultClassName(content);
        }

        private void compileFrom()
        {
            String content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "FROM", "<candidate class>"));
            }

            if (content.indexOf(' ') > 0)
            {
                // Cannot have spaces in class names (maybe an incorrect word specified after the class name)
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithIllegalValue", "FROM", content));
            }
            query.setCandidateClassName(content);

            if (tokenizer.parseKeyword("EXCLUDE") || tokenizer.parseKeyword("exclude"))
            {
                if (!tokenizer.parseKeyword("SUBCLASSES") && !tokenizer.parseKeyword("subclasses"))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.ExpectedKeywordAfter", "SUBCLASSES", "EXCLUDE"));
                }
                content = tokenizer.parseContent();
                if (content.length() > 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithIllegalValue", "EXCLUDE SUBCLASSES", content));
                }
                query.setSubclasses(false);
            }
        }

        private void compileWhere()
        {
            String content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "WHERE", "<filter>"));
            }
            query.setFilter(content);
        }

        private void compileVariables()
        {
            String content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "VARIABLES", "<variable declarations>"));
            }
            query.declareVariables(content);
        }

        private void compileParameters()
        {
            String content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "PARAMETERS", "<parameter declarations>"));
            }
            query.declareParameters(content);
        }

        private void compileImport()
        {
            String content = "import " + tokenizer.parseContent();
            while (tokenizer.parseKeyword("import"))
            {
                content += "import " + tokenizer.parseContent();
            }
            query.declareImports(content);
        }

        private void compileGroup()
        {
            String content = tokenizer.parseContent();
            // content must be empty
            if (!tokenizer.parseKeyword("BY") && !tokenizer.parseKeyword("by"))
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.ExpectedKeywordAfter", "BY", "GROUP"));
            }
            content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "GROUP BY", "<grouping>"));
            }
            query.setGrouping(content);
        }

        private void compileOrder()
        {
            String content = tokenizer.parseContent();
            // content must be empty
            if (!tokenizer.parseKeyword("BY") && !tokenizer.parseKeyword("by"))
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.ExpectedKeywordAfter", "BY", "ORDER"));
            }
            content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "ORDER BY", "<ordering>"));
            }
            query.setOrdering(content);
        }

        private void compileRange()
        {
            String content = tokenizer.parseContent();
            // content cannot be empty
            if (content.length() == 0)
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", "RANGE", "<range>"));
            }
            query.setRange(content);
        }
    }

    /**
     * Tokenizer that provides access to current token
     */
    private class Parser
    {

        /** tokens * */
        final String[] tokens;

        /** keywords * */
        final String[] keywords;

        /** current token cursor position */
        int pos = -1;

        /**
         * Constructor
         * @param str
         */
        public Parser(String str)
        {
            StringTokenizer tokenizer = new StringTokenizer(str);
            tokens = new String[tokenizer.countTokens()];
            keywords = new String[tokenizer.countTokens()];
            int i = 0;
            while (tokenizer.hasMoreTokens())
            {
                tokens[i] = tokenizer.nextToken();
                if (isKeyword(tokens[i]))
                {
                    keywords[i] = tokens[i];
                }
                i++;
            }
        }

        /**
         * Parse the content until a keyword is found
         * @return the content
         */
        public String parseContent()
        {
            String content = "";
            while (pos < tokens.length - 1)
            {
                pos++;
                if (isKeyword(tokens[pos]))
                {
                    pos--;
                    break;
                }
                else
                {
                    if (content.length() == 0)
                    {
                        content = tokens[pos];
                    }
                    else
                    {
                        content += " " + tokens[pos];
                    }
                }
            }
            return content;
        }

        /**
         * Parse the next token looking for a keyword. The cursor position is
         * skipped in one tick if a keyword is found
         * @param keyword the searched keyword
         * @return true if the keyword
         */
        public boolean parseKeyword(String keyword)
        {
            if (pos < tokens.length - 1)
            {
                pos++;
                if (keywords[pos] != null)
                {
                    if (keywords[pos].equals(keyword))
                    {
                        return true;
                    }
                }
                pos--;
            }
            return false;
        }

        /**
         * Parse the next token looking for a keyword. The cursor position is
         * skipped in one tick if a keyword is found
         * @return the parsed keyword or null
         */
        public String parseKeyword()
        {
            if (pos < tokens.length - 1)
            {
                pos++;
                if (keywords[pos] != null)
                {
                    return keywords[pos];
                }
                pos--;
            }
            return null;
        }
    }
}