/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson - removed toStatement and toString methods enherited
                    from QueryStatement. setOrdering now goes thought
                    union queries.                  
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.BooleanCharColumnExpression;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.CharacterExpression;
import org.jpox.store.expression.ObjectExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.util.Localiser;

/**
 * Representation of a Query Statement in Oracle.
 *
 * @version $Revision: 1.22 $
 **/ 
public class OracleQueryStatement extends QueryStatement
{
    /** Localisation for messages. */
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /**
     * The system property that selects the "linguistic definition" to be used
     * for native language sorting of String fields.  This is the string
     * "org.jpox.oracle.nlsSortOrder".  The default value is "LATIN".
     * A value of "BINARY" disables native language sorting.
     */
    public static final String NLS_SORT_ORDER_PROPERTY = "org.jpox.oracle.nlsSortOrder";

    private final String nlsSortOrder = System.getProperty(NLS_SORT_ORDER_PROPERTY, "LATIN").toUpperCase();

    /**
     * QueryStatement constructor 
     * @param initialTable The main table for this statement.
     * @param clr ClassLoader resolver
     */
    public OracleQueryStatement(DatastoreContainerObject initialTable, ClassLoaderResolver clr)
    {
        super(initialTable, clr);
        
        /*
         * Oracle 8 or below doesnt support SQL 92
         * Joins are added to the WHERE clause instead of the FROM clause.
         * e.g.
         * FROM TABLE1,TABLE2
         * WHERE
         * TABLE1.KEY = TABLE2.KEY
         */
		stmtJoinsSeparator = ',';
    }

    /**
     * QueryStatement constructor 
     * @param initialTable The main table for this statement.
     * @param initialRangeVar The alias for the main table
     * @param clr ClassLoader resolver
     */
    public OracleQueryStatement(DatastoreContainerObject initialTable, DatastoreIdentifier initialRangeVar, ClassLoaderResolver clr)
    {
        super(initialTable, initialRangeVar, clr);
		/*
		 * Oracle 8 or below doesnt support SQL 92
		 * Joins are added to the WHERE clause instead of the FROM clause.
		 * e.g.
		 * FROM TABLE1,TABLE2
		 * WHERE
		 * TABLE1.KEY = TABLE2.KEY
		 */
		stmtJoinsSeparator = ',';
    }

    /**
     * Convenience method to generate the ordering statement to add to the overall
     * query statement.
     * @return The ordering statement
     */
    protected StatementText generateOrderingStatement()
    {
        StatementText orderByStmt = null;
        if (orderingExpressions != null && orderingExpressions.length > 0)
        {
            orderByStmt = new StatementText();
            boolean needsSelect = ((RDBMSAdapter)dba).includeOrderByColumnsInSelect();
            for (int i=0; i<orderingExpressions.length; ++i)
            {
                String orderExpr = "JPOXORDER" + i;
                if (i > 0)
                {
                    orderByStmt.append(',');
                }

                if (needsSelect && !hasAggregateExpression)
                {
                    // Order by the "JPOXORDER?" if we need them to be selected and it isn't an aggregate
                    if (orderingExpressions[i] instanceof CharacterExpression && !nlsSortOrder.equals("BINARY"))
                    {
                        orderByStmt.append("NLSSORT(").append(orderExpr).append(", 'NLS_SORT = ").append(nlsSortOrder).append("')");
                    }
                    else if (orderingExpressions[i] instanceof BooleanExpression && !(orderingExpressions[i] instanceof BooleanCharColumnExpression))
                    {
                        throw new JDOFatalInternalException(LOCALISER_RDBMS.msg("RDBMS.QueryStatement.OracleBooleanExpression"));
                    }
                    else
                    {
                        orderByStmt.append(orderExpr);
                    }

                }
                else
                {
                    // Order by the "THIS.COLUMN" otherwise
                    if (orderingExpressions[i] instanceof CharacterExpression && !nlsSortOrder.equals("BINARY"))
                    {
                        orderByStmt.append("NLSSORT(").append(orderingExpressions[i].toStatementText(ScalarExpression.PROJECTION).toStatementString(ScalarExpression.PROJECTION)).append(", 'NLS_SORT = ").append(nlsSortOrder).append("')");
                    }
                    else if (orderingExpressions[i] instanceof BooleanExpression && !(orderingExpressions[i] instanceof BooleanCharColumnExpression))
                    {
                        throw new JDOFatalInternalException(LOCALISER_RDBMS.msg("RDBMS.QueryStatement.OracleBooleanExpression"));
                    }
                    else
                    {
                        orderByStmt.append(orderingExpressions[i].toStatementText(ScalarExpression.PROJECTION).toStatementString(ScalarExpression.PROJECTION));
                    }
                }

                if (orderingDirections[i])
                {
                    orderByStmt.append(" DESC");
                }
            }
        }
        return orderByStmt;
    }

    /**
     * Method to do an inner join to another table.
     * Requires ANSI-86 syntax, providing the table to the FROM clause
     * and the join within the WHERE clause.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression
     **/
    public void innerJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr)
    {
        assertNotFrozen();
        //TODO copy implementation from QueryStatement
        Join join = new Join(expr, expr2, tblExpr);

        if (tblExpr.equals(getInitialTableExpr()))
        {
            andCondition(expr.eq(expr2));
            return;
        }
        if (crossJoins.contains(tblExpr))
        {
            crossJoins.remove(tblExpr);
        }
        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());
        joins.add(join);

        andCondition(expr.eq(expr2));
    }

    /**
     * Method to do a left outer join to another table.
     * Requires ANSI-86 syntax, providing the table to the FROM clause
     * and the join within the WHERE clause.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression
     **/
    public void leftOuterJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr)
    {
        assertNotFrozen();
        //TODO copy implementation from QueryStatement
        Join join = new Join(expr, expr2, tblExpr);

        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());
        joins.add(join);

        // Update the right-hand-side to use Oracle outer join syntax
        if (expr2 instanceof ObjectExpression)
        {
            ((ObjectExpression)expr2).addOuterJoinSuffix("(+)");
        }
        andCondition(expr.eq(expr2));
    }

    /**
     * Method to do a right outer join to another table.
     * Requires ANSI-86 syntax, providing the table to the FROM clause
     * and the join within the WHERE clause.
     * @param expr the left hand expression
     * @param expr2 the right hand expression
     * @param tblExpr The table expression
     **/
    public void rightOuterJoin(ScalarExpression expr, ScalarExpression expr2, LogicSetExpression tblExpr)
    {
        assertNotFrozen();
        //TODO copy implementation from QueryStatement
        Join join = new Join(expr, expr2, tblExpr);

        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());
        joins.add(join);

        // Update the left-hand-side to use Oracle outer join syntax
        if (expr instanceof ObjectExpression)
        {
            ((ObjectExpression)expr).addOuterJoinSuffix("(+)");
        }
        andCondition(expr.eq(expr2));
    }
}