/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Marco Schulze - replaced catch(NotPersistenceCapable) with TypeManager.isSupportedType()
2004 Andy Jefferson - added restriction on Discriminator, added check on subclass using same table
2004 Andy Jefferson - added comments. Added fix to inverse map with non primitive keys
2004 Andy Jefferson - added way of enforcing inclusion of "JPOXMETADATA"
2005 Andy Jefferson - catch "subclass-table" cases where user has a strange inheritance setup
2005 Andy Jefferson - added discriminator for element table for 1-N join table cases
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.util.Iterator;
import java.util.Set;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.MetaDataStringLiteral;
import org.jpox.store.expression.NullLiteral;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.AbstractIteratorStatement;
import org.jpox.store.rdbms.table.SCOTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Class to generate a QueryStatement for iterating through the elements
 * of a Set. This can be in an Extent/Query for example where the user has
 * selected a candidate class and to include/exclude subclasses. The generated
 * query statement typically contains UNIONs for each of the possible classes
 * involved.
 * <P>
 * To give an example, lets assume that we have class A which is the candidate
 * and this has a subclass B. We want to find all objects of the candidate type
 * and optionally its subclasses and we want information about what type the object
 * is (A or B). The query will be of the form
 * <PRE>
 * SELECT THIS.A_ID,'org.jpox.samples.A' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
 * FROM A THIS
 *   LEFT OUTER JOIN B SUBELEMENT0 ON SUBELEMENT0.B_ID = THIS.A_ID
 * WHERE SUBELEMENT0.B_ID IS NULL
 * 
 * UNION
 * 
 * SELECT THIS.A_ID,'org.jpox.samples.B' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
 * FROM A THIS
 *   INNER JOIN B 'ELEMENT' ON 'ELEMENT'.B_ID = THIS.A_ID
 * </PRE>
 * <H3>Extent</H3>
 * Here we use the above arrangement, with the "AssociationEnd" being the element
 * table and the mapping being the ID column mapping.
 * <H3>"Normal" Relationship</H3>
 * Here the join table is the candidate table, and the "AssociationEnd" would be 
 * the element table, and the mapping in the join table to join to this element table 
 * PK column.
 * <H3>"Inverse" Relationship</H3>
 * Here we use the same arrangement as the Extent case. The "AssociationEnd" is the
 * element table, and the mapping being the ID column mapping.
 * <H3>Map "Key" table</H3>
 * Here we have the key table as the candidate, and we join to the value table. So
 * the "AssociationEnd" is the value table, with the mapping being the mapping of the
 * key column in the value table.
 * <H3>Primitive key/value in SCO table</H3>
 * This also supports retrieval of keys or values of a map that are stored as a column
 * in a different table.
 * TODO Add full documentation for all combinations supported.
 * TODO Commonise the code with DiscriminatorIteratorStatement
 *
 * @version $Revision: 1.1 $ 
 */
public class UnionIteratorStatement extends AbstractIteratorStatement
{
    /**
     * One side of an association.
     * TODO Replace this with just the parameters (like in DiscriminatorIteratorStatement)
     */
    public static interface AssociationEnd
    {
        /** 
         * The element type.
         * @return the element type
         */
        Class getType();

        /** 
         * Mapping in the source table to get to join to the target.
         * @return The JavaTypeMapping
         */
        JavaTypeMapping getMapping();

        /**
         * The join table. Can be the same as the target table
         * @return The DatastoreContainerObject 
         */
        DatastoreContainerObject getDatastoreContainerObject();

        /** 
         * Whether to use a join. Only used by map key tables.
         * @return true if to use join
         */
        boolean useJoin();
    }

    /** the source association end */
    AssociationEnd source;

    /** Whether to include the JPOXMETADATA tag in the query to distinguish classes. */
    Boolean withMetadata = null;

    /**
     * Constructor. 
     * @param clr The ClassLoaderResolver
     * @param candidateType the candidate is who are are looking to
     * @param includeSubclasses if the subclasses of the candidate should be
     *                          included in the result
     * @param storeMgr the store manager
     * @param source the source for the association
     */
    public UnionIteratorStatement(ClassLoaderResolver clr,
                                  Class candidateType, 
                                  boolean includeSubclasses,
                                  StoreManager storeMgr,
                                  AssociationEnd source)
    {
        this(clr, candidateType, includeSubclasses, storeMgr, source, null);
    }

    /**
     * Constructor.
     * @param clr The ClassLoaderResolver
     * @param candidateType the candidate is who we are looking to
     * @param includeSubclasses if the subclasses of the candidate should be
     *                          included in the result
     * @param storeMgr the store manager
     * @param source the source for the association
     * @param withMetadata if has subclasses, include the metadata in the query
     */
    public UnionIteratorStatement(ClassLoaderResolver clr,
                                  Class candidateType,
                                  boolean includeSubclasses,
                                  StoreManager storeMgr,
                                  AssociationEnd source,
                                  boolean withMetadata)
    {
        this(clr, candidateType, includeSubclasses, storeMgr, source, (withMetadata ? Boolean.TRUE : Boolean.FALSE));
    }

    /**
     * Constructor.
     * @param clr The ClassLoaderResolver
     * @param candidateType the candidate is who we are looking to
     * @param includeSubclasses if the subclasses of the candidate should be
     *                          included in the result
     * @param storeMgr the store manager
     * @param source the source for the association
     * @param withMetadata if has subclasses, include the metadata in the query
     */
    public UnionIteratorStatement(ClassLoaderResolver clr,
                                  Class candidateType,
                                  boolean includeSubclasses,
                                  StoreManager storeMgr,
                                  AssociationEnd source,
                                  Boolean withMetadata)
    {
        super(candidateType, clr, includeSubclasses, storeMgr);

        this.source = source;
        this.withMetadata = withMetadata;

        // check if the element target is assignable from the candidate
        if (!source.getType().isAssignableFrom(candidateType))
        {
            if (candidateType.isInterface())
            {
                if (!candidateType.isAssignableFrom(source.getType()))
                {
                    throw new IncompatibleQueryElementTypeException(candidateFullClassName,source.getType().getName());
                }
            }
            else
            {
                throw new IncompatibleQueryElementTypeException(source.getType().getName(), candidateType.getName());
            }
        }

        // Give warning if table is incorrect for candidate and subclasses not included in query.
        if (!source.getDatastoreContainerObject().equals(candidateTable) && !includeSubclasses)
        {
            JPOXLogger.RDBMS.warn(LOCALISER.msg("Extent.NoSubclassesWarning", this, candidateFullClassName));
        }
    }

    /**
     * Accessor for the Query Statement.
     * @return The Query Statement for iterating through objects
     */
    public QueryExpression getQueryStatement()
    {
        if (storeMgr.getPMFContext().getTypeManager().isSupportedType(candidateFullClassName))
        {
            // Non-FCO candidates are embedded in the source
            QueryExpression qs = dba.newQueryStatement(source.getDatastoreContainerObject(), clr);
            qs.select(source.getMapping());
            return qs;
        }
        else
        {
            // FCO candidate
            if (!includeSubclasses)
            {
                // No need to include the JPOXMETADATA
                withMetadata = Boolean.FALSE;
            }

            Set subclasses = null;
            if (includeSubclasses)
            {
                // Add a select for each subclass elements, and union the selects
                subclasses = storeMgr.getSubClassesForClass(candidateFullClassName, true, clr);
            }
            
            // Create the main query
            QueryExpression stmt = getQueryForElement(candidateFullClassName, storeMgr.getDatastoreClass(candidateFullClassName, clr));

            if (includeSubclasses)
            {
                // Add a select for each subclass elements, and union the selects
                Iterator iterator = subclasses.iterator();
                while (iterator.hasNext())
                {
                    String subCandidateType = (String) iterator.next();
                    DatastoreClass subCandidateTable = storeMgr.getDatastoreClass(subCandidateType, clr);
                    if (subCandidateTable != null)
                    {
                        // Add UNION to any subclasses that have their own table
                        // This omits those using "subclass-table". Those will have their own subclasses where they
                        // store their objects.
                        QueryExpression stmt_subclass = getQueryForElement(subCandidateType, subCandidateTable);
                        stmt.union(stmt_subclass);
                    }
                }
            }
            return stmt;
        }
    }

    /**
     * Utility method to generate a select statement to find objects of
     * subclasses of the element class in the Set.
     *  <p>
     * In simple terms this creates a SELECT like
     * <pre>
     * SELECT THIS.A_ID,'org.jpox.samples.A' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
     * FROM A THIS
     *   LEFT OUTER JOIN B SUBELEMENT0 ON SUBELEMENT0.B_ID = THIS.A_ID
     * WHERE SUBELEMENT0.B_ID IS NULL
     * </pre>
     * The actual query will vary if this subclass has its own subclasses. If
     * there are no subclasses of this class then the LEFT OUTER JOIN parts will
     * not be present, nor will the SUBELEMENT_? parts of the WHERE clause.
     * </p>
     * <p>
     * If the target element shares the same table as the source element then
     * no (left outer) join will be made and it will be a simple select.
     * </p>
     * <p>
     * If a discriminator column is present on the table, it will be used to
     * separate the classes in that table rather than using the left outer join
     * strategy
     * </p>
     * @param targetElementType The Class for the element
     * @param targetElementTable The table for this class (will not be null)
     * @return QueryStatement
     */
    private QueryExpression getQueryForElement(String targetElementType, DatastoreClass targetElementTable)
    {
        QueryExpression stmt;

        JavaTypeMapping discriminatorMapping = null;
        DiscriminatorMetaData discriminatorMetaData = null;
        LogicSetExpression discriminatorTableExpr = null;
        if (source.getDatastoreContainerObject() instanceof SCOTable)
        {
            // * Selecting the join table of a JoinTable (normal) relationship, and joining to the element table
            stmt = dba.newQueryStatement(source.getDatastoreContainerObject(), clr);

            // Add inner joins to all classes above up to elementType and across to the join table
            DatastoreIdentifier targetTableIdentifier = joinSourceToTargetElement(stmt, targetElementTable);

            discriminatorMapping = targetElementTable.getDiscriminatorMapping();
            discriminatorMetaData = targetElementTable.getDiscriminatorMetaData();
            discriminatorTableExpr = stmt.getTableExpression(targetTableIdentifier);

            // Add left outer joins to exclude any target element subclasses
            joinToExcludeTargetWhenSubElementsExists(stmt, source.getMapping(), targetElementType);
        }
        else
        {
            // * Selecting FCO objects directly (Extents etc)
            // * Selecting the element table of a ForeignKey (inverse) relationship
            stmt = dba.newQueryStatement(candidateTable, clr);

            discriminatorMapping = source.getDatastoreContainerObject().getDiscriminatorMapping();
            discriminatorMetaData = source.getDatastoreContainerObject().getDiscriminatorMetaData();
            discriminatorTableExpr = stmt.getDefaultTableExpression();

            // in case of the elementType is a subClass of the element for the candidateTable
            // if the root (candidate) element type is not the same as the current target element type
            // joins the root to the current element
            if ((!targetElementTable.toString().equals(source.getDatastoreContainerObject().toString()) &&
                 !candidateTable.getType().equals(targetElementType)) || 
                 source.useJoin())
            {
                // Add inner joins to all classes above up to elementType and across to the join table
                if (source.useJoin())
                {
                    joinTargetToSourceElement(stmt, targetElementTable);
                }
                else
                {
                    joinSourceToTargetElement(stmt, targetElementTable);
                }
            }
            joinToExcludeTargetWhenSubElementsExists(stmt, candidateTable.getIDMapping(), targetElementType);
        }

        if (discriminatorMapping != null && discriminatorMetaData.getStrategy() != DiscriminatorStrategy.NONE)
        {
            // Restrict to valid discriminator values where we have a discriminator specified on this table
            String discriminatorValue = targetElementType;
            if (discriminatorMetaData.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
            {
                // Get the MetaData for the target class since that holds the "value"
                AbstractClassMetaData targetCmd = storeMgr.getMetaDataManager().getMetaDataForClass(targetElementType, clr);
                discriminatorValue = targetCmd.getInheritanceMetaData().getDiscriminatorMetaData().getValue();
            }
            ScalarExpression discrExpr = discriminatorMapping.newScalarExpression(stmt, discriminatorTableExpr);
            ScalarExpression discrVal = discriminatorMapping.newLiteral(stmt, discriminatorValue);
            stmt.andCondition(discrExpr.eq(discrVal));
        }

        if (withMetadata == null || withMetadata.booleanValue())
        {
            // Select JPOXMETADATA if required
            selectMetadata(stmt, targetElementType);
        }

        return stmt;
    }

    /**
     * Joins the container table (source) to the element table (target).
     * This is used when the source mapping is from target to the source (normal
     * relationships where the target is SCO).
     * @param stmt The Query Statement
     * @param elementTargetTable the table that contains the elements
     * @return Table identifier for the target table in the query
     */
    private DatastoreIdentifier joinSourceToTargetElement(QueryExpression stmt, DatastoreClass elementTargetTable)
    {
        // TODO Make this identifier specifiable as input to the iterator statement
        DatastoreIdentifier targetTableIdentifier = stmt.getStoreManager().getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "ELEMENT");
        LogicSetExpression teTargetElement = stmt.newTableExpression(elementTargetTable, targetTableIdentifier);
        ScalarExpression sourceExpr = source.getMapping().newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression targetExpr = elementTargetTable.getIDMapping().newScalarExpression(stmt, teTargetElement);
        stmt.innerJoin(sourceExpr, targetExpr, teTargetElement, true);
        return targetTableIdentifier;
    }

    /**
     * Joins the element table (target) to the container table (source).
     * This is used when the source mapping is from the source to the target.
     * @param stmt The Query Statement
     * @param elementTargetTable the table that contains the elements
     */
    private void joinTargetToSourceElement(QueryExpression stmt, DatastoreClass elementTargetTable)
    {
        // TODO Make this identifier specifiable as input to the iterator statement
        DatastoreIdentifier sourceTableIdentifier = stmt.getStoreManager().getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "VALUE");
        LogicSetExpression teSourceElement = stmt.newTableExpression(source.getDatastoreContainerObject(), sourceTableIdentifier);
        ScalarExpression sourceExpr = source.getMapping().newScalarExpression(stmt, teSourceElement);
        ScalarExpression targetExpr = elementTargetTable.getIDMapping().newScalarExpression(stmt, stmt.getDefaultTableExpression());
        stmt.innerJoin(sourceExpr, targetExpr, teSourceElement, true);
    }

    /**
     * Join to exclude targetElementType instances when exists sub elements
     * instances of targetElementType. This is made to avoid returning
     * targetElement when there are sub classes instances of targetElement
     * <pre>
     * ...
     * ELEMENT_TABLE TARGETELEMENT
     * LEFT OUTER JOIN SUB_TARGETELEMENT ON TARGETELEMENT.ELEMENT_ID = SUBCLASS_TARGETELEMENT.SUB_TARGETELEMENT_ID
     * WHERE SUB_TARGETELEMENT.SUB_TARGETELEMENT_ID IS NULL
     * ...
     * </pre>
     * <p>When the subelements are using the same table as the target element, no
     * exclusion is performed</p>
     *
     * @param stmt The Query Statement
     * @param targetElementMapping the JavaTypeMapping for the target element for joining to
     * @param targetElementType The target element
     */
    private void joinToExcludeTargetWhenSubElementsExists(
            QueryExpression stmt,
            JavaTypeMapping targetElementMapping,
            String targetElementType)
    {
        // Add LEFT OUTER JOIN to all direct subclasses of the targetRootElementQCL - if any
        Iterator iterTargetSubElementType = storeMgr.getSubClassesForClass(targetElementType, false, stmt.getClassLoaderResolver()).iterator();
        int subSequenceIdentifier = 0;
        while (iterTargetSubElementType.hasNext())
        {
            String targetSubElementType = (String) iterTargetSubElementType.next();

            QueryExpression targetQS;
            DatastoreClass cbtTargetSubElementType = storeMgr.getDatastoreClass(targetSubElementType, clr);
            DatastoreIdentifier tiTargetSubElementType;
            LogicSetExpression teTargetSubElementType;
            ScalarExpression targetSubElementTypeExpr;

            DatastoreClass[] targetSubElementTypes = null;
            if (cbtTargetSubElementType == null)
            {
                AbstractClassMetaData targetSubCmd = storeMgr.getMetaDataManager().getMetaDataForClass(targetSubElementType, stmt.getClassLoaderResolver());
                AbstractClassMetaData[] targetSubCmds = storeMgr.getClassesManagingTableForClass(targetSubCmd, clr);
                targetSubElementTypes = new DatastoreClass[targetSubCmds.length];
                for (int i=0;i<targetSubCmds.length;i++)
                {
                    targetSubElementTypes[i] = storeMgr.getDatastoreClass(targetSubCmds[i].getFullClassName(), clr);
                }
            }
            else
            {
                targetSubElementTypes = new DatastoreClass[1];
                targetSubElementTypes[0] = cbtTargetSubElementType;
            }

            for (int i=0;i<targetSubElementTypes.length;i++)
            {
                // Add leftOuterJoin to sub-target elements if they are not in the same table as the target
                if (!targetSubElementTypes[i].toString().equals(storeMgr.getDatastoreClass(targetElementType, clr).toString()))
                {
                    tiTargetSubElementType = storeMgr.getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "SUBELEMENT" + (subSequenceIdentifier++));
                    
                    // create a new statement
                    targetQS = new QueryStatement(targetSubElementTypes[i], tiTargetSubElementType, stmt.getClassLoaderResolver());

                    // table expression from the table identifier
                    teTargetSubElementType = targetQS.newTableExpression(targetSubElementTypes[i], tiTargetSubElementType);

                    JavaTypeMapping targetMapping = targetSubElementTypes[i].getIDMapping();
                    ScalarExpression targetElementExpr = targetElementMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                    targetSubElementTypeExpr = targetMapping.newScalarExpression(stmt,teTargetSubElementType);
                    stmt.leftOuterJoin(targetElementExpr,targetSubElementTypeExpr, teTargetSubElementType, true);
                    ScalarExpression seTargetSubElementType = targetMapping.newScalarExpression(targetQS, teTargetSubElementType);
                    stmt.andCondition(new NullLiteral(stmt).eq(seTargetSubElementType));
                }
            }
        }
    }

    /**
     * Adds an expression to the statement. The expression is a metadata that
     * identifies an element in a hierarchy level of classes. This will appear
     * in the SELECT as "'org.jpox.samples.MyClass     ' AS JPOXMETADATA".
     * 
     * @param stmt Query Statement
     * @param pcClass The Persistence Capable class
     */
    private void selectMetadata(QueryExpression stmt, String pcClass)
    {
        // If we are looking for subclasses, we need to add the class to the
        // select, and unfortunately some RDBMS set an odd limit on the first
        // row in the output set column for the length of the name column.
        // This fix looks through the classes concerned in our query and takes
        // the longest name and pads all other names to this length. This is
        // used in the "'package.classname' AS JPOXMETADATA" parts of the query.
        int maxClassNameLen = candidateFullClassName.length();
        boolean hasSubClasses = false;
        Iterator iterator = storeMgr.getSubClassesForClass(candidateFullClassName, true, clr).iterator();
        if (iterator.hasNext())
        {
            hasSubClasses = true;
            while (iterator.hasNext())
            {
                String nextClass = (String) iterator.next();
                if (nextClass.length() > maxClassNameLen)
                {
                    maxClassNameLen = nextClass.length();
                }
            }
        }

        // Include metadata if explicitly told to include it, or if subclasses present
        if ((withMetadata == null && hasSubClasses) ||
            (withMetadata != null && withMetadata.booleanValue()))
        {
            String classname = pcClass;
            if (maxClassNameLen > classname.length())
            {
                classname = StringUtils.leftAlignedPaddedString(pcClass, maxClassNameLen);
            }
            stmt.selectScalarExpression(new MetaDataStringLiteral(stmt,classname));
        }
    }
}