/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
2005 Andy Jefferson - added handling for updating FK in related object
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOOptimisticVerificationException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InterfaceMetaData;
import org.jpox.metadata.Relation;
import org.jpox.metadata.VersionStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Class to provide a means of deletion of records from a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * delete operation.
 *
 * @version $Revision: 1.41 $ 
 **/
public class DeleteRequest extends Request
{
    private static final int IDPARAMNUMBER = 1;

    private final MappingCallbacks[] callbacks;

    /** Statement for retrieving the datastore version of the object to be deleted. */
    private final String versionStmt;

    /** Whether to perform version checks */
    private final boolean versionChecks;

    /** Statement for deleting the object from the datastore. */
    private final String deleteStmt;

    /** one StatementExpressionIndex for each field **/
    private StatementExpressionIndex[] statementExpressionIndex;

    /** PK fields to be provided in defining the record to be deleted (app identity cases). */
    private final int[] pkfieldsToBeProvided;

    /** MetaData for fields with related objects to those in this class. */
    private final AbstractPropertyMetaData[] fieldsWithRelatedObjects;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to delete.
     * @param cls Class of objects being deleted
     * @param clr ClassLoader resolver
     **/
    public DeleteRequest(DatastoreClass table, Class cls, ClassLoaderResolver clr)
    {
        super(table);

        AbstractClassMetaData cmd = table.getStoreManager().getMetaDataManager().getMetaDataForClass(cls, clr);        

        int paramIndex = IDPARAMNUMBER;

        // Construct the statement for the Delete
        StringBuffer where = new StringBuffer();

        if (table.getIdentityType() == IdentityType.DATASTORE)
        {
            Iterator iterator = key.getColumns().iterator();
            where.append(((DatastoreField)iterator.next()).getIdentifier().toString());
            where.append("=?");
            paramIndex++;
        }

        DeleteMappingConsumer consumer = new DeleteMappingConsumer(clr, cmd, where, paramIndex);
        table.providePrimaryKeyMappings(consumer);
        table.provideDatastoreIdMappings(consumer);
        table.provideNonPrimaryKeyMappings(consumer); // to compute callbacks
        pkfieldsToBeProvided = consumer.getPrimaryKeysToBeProvided();
        statementExpressionIndex = consumer.getStatementExpressionIndex();
        callbacks = (MappingCallbacks[])consumer.getMappingCallBacks().toArray(new MappingCallbacks[consumer.getMappingCallBacks().size()]);
        fieldsWithRelatedObjects = consumer.getFieldsWithRelatedObjects();

        // delete statement
        deleteStmt = "DELETE FROM " + table.toString() + " WHERE " + where;

        // version retrieval statement
        if (table.getVersionMapping() != null)
        {
            versionStmt = "SELECT " +
                table.getVersionMapping().getDataStoreMapping(0).getDatastoreField() +
                " FROM " + table.toString() +   " WHERE " + where;
        }
        else
        {
            versionStmt = null;
        }
        if (table.getVersionMapping() != null && table.getVersionMetaData().getVersionStrategy() != VersionStrategy.NONE)
        {
            versionChecks = true;
        }
        else
        {
            versionChecks = false;
        }
    }

    /**
     * Method performing the deletion of the record from the datastore. Takes
     * the constructed deletion query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be deleted.
     **/
    public void execute(StateManager sm)
    {
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            // Debug information about what we are deleting
            StringBuffer str = new StringBuffer("Deleting object " + StringUtils.toJVMIDString(sm.getObject()) + 
                " of type " + sm.getObject().getClass().getName() + " from table " + table);
            JPOXLogger.JDO.debug(str);
        }

        // Execute any mapping actions before the delete (things like clearing
        // the links to other objects that we have relationships with).
        for (int i = 0; i < callbacks.length; ++i)
        {
            callbacks[i].preDelete(sm);
        }

        // Update any related objects that may have a FK back to this record, before we delete it
        if (fieldsWithRelatedObjects != null && fieldsWithRelatedObjects.length > 0)
        {
            for (int i=0;i<fieldsWithRelatedObjects.length;i++)
            {
                AbstractPropertyMetaData relatedFmd = fieldsWithRelatedObjects[i];
                updateRelatedObjectsForField(sm, relatedFmd);
            }
        }

        PersistenceManager pm = sm.getPersistenceManager();
        RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
        try
        {
            Connection conn = storeMgr.getConnection(pm, true, false);

            try
            {
                PreparedStatement ps;
                if (pm.currentTransaction().getOptimistic())
                {
                    if (versionStmt != null)
                    {
                        // Version available, so check it for validity of deletion
                        ps = storeMgr.getStatement(conn, versionStmt, false);

                        try
                        {
                            if (sm.getInternalObjectId() instanceof OID)
                            {
                                // a). datastore identity
                                int[] paramNumber = { IDPARAMNUMBER };
                                table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                            }
                            else
                            {
                                // b). application identity
                                sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, false));
                            }

                            ResultSet rs = storeMgr.executeStatementQuery(versionStmt, ps);
                            try
                            {
                                if (rs.next())
                                {
                                    Object datastoreVersion = table.getVersionMapping().getObject(pm, rs, new int[]{1});
                                    if (versionChecks &&
                                        !((VersionMapping)table.getVersionMapping()).compareVersion(datastoreVersion, sm.getTransactionalVersion(sm.getObject())))
                                    {
                                        String msg = LOCALISER.msg("RDBMS.Request.OptimisticVersionMismatch", 
                                            sm.getInternalObjectId(), table.toString(), datastoreVersion, sm.getTransactionalVersion(sm.getObject()));
                                        JPOXLogger.RDBMS.error(msg);
                                        throw new JDOOptimisticVerificationException(msg, sm.getObject());
                                    }
                                }
                                else
                                {
                                    // Nothing found, so already deleted presumably
                                    String msg = LOCALISER.msg("RDBMS.Request.OptimisticVersionMissing", 
                                        sm.getInternalObjectId(), table.toString());
                                    JPOXLogger.RDBMS.error(msg);
                                    throw new JDOOptimisticVerificationException(msg, sm.getObject());
                                }
                            }
                            finally
                            {
                                rs.close();
                            }
                        }
                        finally
                        {
                            ps.close();
                        }
                    }
                    else
                    {
                        // TODO No version available, so retrieve the datastore object and do a field-by-field check
                    }
                }

                ps = storeMgr.getStatement(conn, deleteStmt, false);

                try
                {
                    // provide primary key fields
                    if (sm.getInternalObjectId() instanceof OID)
                    {
                        // a). datastore identity
                        int[] paramNumber = { IDPARAMNUMBER };
                        table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                    }
                    else
                    {
                        // b). application identity
                        sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, false));
                    }

                    storeMgr.executeStatementUpdate(deleteStmt, ps);
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Delete request failed: " + deleteStmt, e);
        }
    }

    /**
     * Method to update any related objects to remove any links to the object being deleted.
     * @param sm StateManager of this object
     * @param fmd MetaData for field that has related objects
     */
    private void updateRelatedObjectsForField(StateManager sm, AbstractPropertyMetaData fmd)
    {
        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(sm.getPersistenceManager().getClassLoaderResolver());
        String fullClassName = ((AbstractClassMetaData)relatedFmd.getParent()).getFullClassName();

        //TODO I'm not sure that we need to loop all implementations. will we have the fk set to all tables, if many?
        String[] classes;
        if (((AbstractClassMetaData)relatedFmd.getParent()) instanceof InterfaceMetaData)
        {
            classes = sm.getStoreManager().getMetaDataManager().getClassesImplementingInterface(fullClassName, sm.getPersistenceManager().getClassLoaderResolver());
        }
        else
        {
            classes = new String[] {fullClassName};
        }
        Set datastoreClasses = new HashSet();
        for( int i=0; i<classes.length; i++)
        {
            // just remove duplicates
            datastoreClasses.add(sm.getStoreManager().getDatastoreClass(classes[i], sm.getPersistenceManager().getClassLoaderResolver()));
        }

        PersistenceManager pm = sm.getPersistenceManager();
        RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
        Iterator it = datastoreClasses.iterator();
        while (it.hasNext())
        {
            DatastoreClass refTable = (DatastoreClass)it.next();
            JavaTypeMapping refMapping = refTable.getFieldMapping(fmd.getMappedBy());
            if (refMapping.isNullable()) // Only clear the references that can be cleared
            {
                // Create a statement to clear the link from the previous related object
                StringBuffer clearLinkStmt = new StringBuffer("UPDATE " + refTable.toString() + " SET ");
                for (int j=0;j<refMapping.getNumberOfDatastoreFields();j++)
                {
                    if (j > 0)
                    {
                        clearLinkStmt.append(",");
                    }
                    clearLinkStmt.append(refMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier());
                    clearLinkStmt.append("=NULL");
                }
                clearLinkStmt.append(" WHERE ");
                for (int j=0;j<refMapping.getNumberOfDatastoreFields();j++)
                {
                    if (j > 0)
                    {
                        clearLinkStmt.append(" AND ");
                    }
                    clearLinkStmt.append(refMapping.getDataStoreMapping(j).getDatastoreField().toString());
                    clearLinkStmt.append("=?");
                }
                
                try
                {
                    Connection conn = storeMgr.getConnection(pm, true, false);
                    try
                    {
                        // Null out the relationship to the object being deleted.
                        PreparedStatement ps = null;
                        try
                        {
                            ps = storeMgr.getStatement(conn, clearLinkStmt.toString(), false);
                            refMapping.setObject(pm, ps, Mappings.getParametersIndex(1, refMapping), sm.getObject());
                            storeMgr.executeStatementUpdate(clearLinkStmt.toString(), ps);
                        }
                        finally
                        {
                            if (ps != null)
                            {
                                ps.close();
                            }
                        }
                    }
                    finally
                    {
                        storeMgr.releaseConnection(pm, conn);
                    }
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                    throw new JDODataStoreException("Update request failed", e);
                }
            }
        }
    }

    private class DeleteMappingConsumer implements MappingConsumer
    {
        private final StringBuffer where;
        private int paramIndex;

        /** one StatementExpressionIndex for each field **/
        private StatementExpressionIndex[] statementExpressionIndex;

        /** Primary Key fields to use in identifying the record to delete. */
        private List pkbp = new ArrayList();

        /** Fields with a related object storing a FK back to this object (that need nulling). */
        private List fwro = new ArrayList();

        /** Mapping Callbacks to invoke at deletion. */
        private List mc = new ArrayList();

        /** ClassLoaderResolver **/
        private final ClassLoaderResolver clr;

        private final AbstractClassMetaData cmd;

        /**
         * Constructor
         * @param clr the ClassLoaderResolver
         * @param cmd AbstractClassMetaData
         * @param where where clause
         * @param initialParamIndex initial index for JDBC Parameter
         */
        public DeleteMappingConsumer(ClassLoaderResolver clr, AbstractClassMetaData cmd, StringBuffer where, int initialParamIndex)
        {
            this.clr = clr;
            this.cmd = cmd;
            this.where = where;
            this.paramIndex = initialParamIndex;
        }

        public void preConsumeMapping(int highestFieldNumberIndex)
        {
            if (statementExpressionIndex == null)
            {
                statementExpressionIndex = new StatementExpressionIndex[highestFieldNumberIndex];
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (!fmd.getAbstractClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            if (m.includeInUpdateStatement())
            {
                if (fmd.isPrimaryKey())
                {
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        if (where.length() > 0)
                        {
                            where.append(" AND ");
                        }
                        where.append(m.getDataStoreMapping(j).getDatastoreField().getIdentifier());
                        where.append(" = ");
                        where.append(((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter());

                        pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        parametersIndex[j] = paramIndex++;
                    }
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);                     
                }
                else if (m.getNumberOfDatastoreFields() == 0 && (m instanceof PersistenceCapableMapping || m instanceof InterfaceMapping))
                {
                    int relationType = fmd.getRelationType(clr);
                    if (relationType == Relation.ONE_TO_ONE_BI)
                    {
                        if (fmd.getMappedBy() != null)
                        {
                            // 1-1 bidirectional field without datastore column(s) (with single FK)
                            fwro.add(fmd);
                        }
                    }
                    else if (relationType == Relation.MANY_TO_ONE_BI)
                    {
                        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                        if (fmd.getJoinMetaData() != null || relatedFmd.getJoinMetaData() != null)
                        {
                            // 1-N bidirectional using join table for relation
                            // TODO Anything to do here ?
                        }
                    }
                }
            }

            // Build up list of mappings callbacks for the fields of this class.
            // The Mapping callback called delete is the preDelete
            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field
         * @param m
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
        }

        /**
         * All the pk fields that must be provided by the StateManager.
         * The StateManager will call the PC class to provide fields
         * @return array of absolute primary key field numbers
         */        
        public int[] getPrimaryKeysToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer)pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        /**
         * All fields that have related objects and where we use the related object to
         * maintain a relationship. These are used where we have 1-1 bidirectional
         * relationships with a single FK (in the other object).
         * @return The fields with related objects
         */
        public AbstractPropertyMetaData[] getFieldsWithRelatedObjects()
        {
            AbstractPropertyMetaData[] fieldsWithRelatedObjects = new AbstractPropertyMetaData[fwro.size()];
            for (int i = 0; i < fwro.size(); ++i)
            {
                fieldsWithRelatedObjects[i] = (AbstractPropertyMetaData) fwro.get(i);
            }
            return fieldsWithRelatedObjects;
        }

        /**
         * Obtain a List of mapping callbacks
         * @return the mapping callbacks
         */
        public List getMappingCallBacks()
        {
            return mc;
        }

        /** 
         * Obtain the StatementExpressionIndex
         * @return the array of StatementExpressionIndex indexed by absolute field numbers
         */
        public StatementExpressionIndex[] getStatementExpressionIndex()
        {
            return statementExpressionIndex;
        }
    }
}