/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson - throws a JDOObjectNotFoundException
2003 Erik Bengtson - removed unused variable and import
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - conversion to use Logger
2003 Erik Bengtson - added optimistic
2004 Erik Bengtson - added consumer
2004 Andy Jefferson - added discriminator
2005 Andy Jefferson - added handling of 1-1 bidir FKs
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOObjectNotFoundException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.Relation;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.FetchStatement;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.RDBMSTransaction;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.fieldmanager.ResultSetGetter;
import org.jpox.store.rdbms.table.ListTable;
import org.jpox.store.rdbms.table.SetTable;
import org.jpox.util.JPOXLogger;

/**
 * Class to provide a means of selection of records from a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * select operation.
 *
 * @version $Revision: 1.47 $ 
 **/
public class FetchRequest extends Request
{
    private final MappingCallbacks[] callbacks;

    private FetchMappingConsumer consumer;

    private MappingStatementIndex mappingStatementIndex;

    private FetchStatement fetchStatement;

    private final AbstractPropertyMetaData[] fieldMetaData;

    private AbstractClassMetaData cmd;
    
    /** cache the statement when generated **/
    private String[] stmt = new String[2];

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to retrieve
     * @param fieldMetaData MetaData of the fields to retrieve
     * @param cls Class of objects being fetched
     * @param clr ClassLoader resolver
     **/
    public FetchRequest(DatastoreClass table, AbstractPropertyMetaData[] fieldMetaData, Class cls, ClassLoaderResolver clr)
    {
        super(table);

        this.fieldMetaData = fieldMetaData;

        cmd = table.getStoreManager().getMetaDataManager().getMetaDataForClass(cls, clr);

        mappingStatementIndex = new MappingStatementIndex();
        fetchStatement = new RDBMSFetchStatement(table);

        // Extract the fields to be retrieved
        consumer = new FetchMappingConsumer(fetchStatement, clr);
        table.provideMappingsForFields(consumer, fieldMetaData, true);
        table.provideDatastoreIdMappings(consumer);
        table.provideDiscriminatorMappings(consumer);

        callbacks = (MappingCallbacks[]) consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);

        // Construct the statement for the Fetch (the WHERE conditions)
        if (consumer.getFieldsToBeReplaced().length > 0)
        {
            table.provideVersionMappings(consumer);
            table.providePrimaryKeyMappings(consumer);
        }
        else
        {
            fetchStatement = null;
        }
    }

    /**
     * Method performing the retrieval of the record from the datastore. Takes
     * the constructed retrieval query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be retrieved
     **/
    public void execute(StateManager sm)
    {
        if (fieldMetaData != null && JPOXLogger.JDO.isDebugEnabled())
        {
            // Debug information about what we are retrieving
            StringBuffer str = new StringBuffer("Fetching class " + cmd.getFullClassName() + " fields [");
            for (int i=0;i<fieldMetaData.length;i++)
            {
                if (i > 0)
                {
                    str.append(",");
                }
                str.append(fieldMetaData[i].getName());
            }
            str.append("] for object ").append(sm.getObjectId(sm.getObject()));
            str.append(" in table ").append(table);
            JPOXLogger.JDO.debug(str);
        }

        if (fetchStatement != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            RDBMSTransaction tx = (RDBMSTransaction) pm.currentTransaction();
            if (stmt[tx.useUpdateLockOnFetch() ? 1 : 0] == null)
            {
                stmt[tx.useUpdateLockOnFetch()?1:0] = fetchStatement.toString(tx.useUpdateLockOnFetch());
            }
            try
            {
                Connection conn = storeMgr.getConnection(pm, false, false);

                try
                {
                    PreparedStatement ps = storeMgr.getStatement(conn, stmt[tx.useUpdateLockOnFetch() ? 1 : 0], true);

                    try
                    {
                        // Provide the primary key field(s)
                        if (sm.getInternalObjectId() instanceof OID)
                        {
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, 
                                mappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
                        }
                        else
                        {
                            sm.provideFields(consumer.getPrimaryKeyFieldsToBeProvided(), 
                                new ParameterSetter(sm, ps, mappingStatementIndex.getFields(), false));
                        }

                        ResultSet rs = storeMgr.executeStatementQuery(stmt[tx.useUpdateLockOnFetch()?1:0], ps);
                        try
                        {
                            if (!rs.next())
                            {
                                JPOXLogger.RDBMS_SQL.warn(LOCALISER.msg("RDBMS.Manager.ObjectWithIdNotFound",sm.getInternalObjectId()));
                                throw new JDOObjectNotFoundException("No such database row", sm.getInternalObjectId());
                            }
                            sm.replaceFields(consumer.getFieldsToBeReplaced(), new ResultSetGetter(sm, rs, mappingStatementIndex.getFields()));

                            if (mappingStatementIndex.getVersion().getExpressionIndex() != null)
                            {
                                Object datastoreVersion = table.getVersionMapping().getObject(pm, rs,  mappingStatementIndex.getVersion().getExpressionIndex());
                                sm.setVersion(datastoreVersion);
                            }
                        }
                        finally
                        {
                            rs.close();
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Fetch request failed: " + stmt[tx.useUpdateLockOnFetch()?1:0], e);
            }
        }

        // Execute any mapping actions now that we have fetched the element
        for (int i = 0; i < callbacks.length; ++i)
        {
            callbacks[i].postFetch(sm);
        }
    }

    private class FetchMappingConsumer implements MappingConsumer
    {
        private int paramIndex = 1;

        final FetchStatement fetchStmt;

        /** fields to be replaced */
        List ftbr = new ArrayList();

        /** pk fields to be provided */
        List pkbp = new ArrayList();

        List mappingCallbacks = new ArrayList();

        boolean initialized = false;
        
        final ClassLoaderResolver clr;

        /**
         * @param fetchStmt The Fetch Statement
         * @param clr The ClassLoaderResolver
         */
        public FetchMappingConsumer(final FetchStatement fetchStmt, ClassLoaderResolver clr)
        {
            super();
            this.fetchStmt = fetchStmt;
            this.clr = clr;
        }

        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (!fmd.getAbstractClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            if (m.includeInFetchStatement())
            {
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setMapping(m);

                // create the expressions index (columns index)
                int expressionsIndex[] = null;
                if (m.getNumberOfDatastoreFields() == 0 && (m instanceof PersistenceCapableMapping || m instanceof InterfaceMapping))
                {
                    int relationType = fmd.getRelationType(clr);
                    if (relationType == Relation.ONE_TO_ONE_BI)
                    {
                        if (fmd.getMappedBy() != null)
                        {
                            // 1-1 bidirectional field without datastore column(s) (with single FK)
                            Class fieldType = fmd.getType();
                            DatastoreClass relatedTable;
                            String[] clsNames;
                            if (fieldType.isInterface())
                            {
                                clsNames = table.getStoreManager().getMetaDataManager().getClassesImplementingInterface(fieldType.getName(), clr);
                            }
                            else
                            {
                                clsNames = new String[] { fieldType.getName() };
                            }
                            //only allowed to have association to one table. implementations can be many, but must store to the same table
                            relatedTable = table.getStoreManager().getDatastoreClass(clsNames[0], clr);
                            
                            JavaTypeMapping referenceMapping = relatedTable.getFieldMapping(fmd.getMappedBy());
                            JavaTypeMapping selectMapping = relatedTable.getIDMapping();
                            JavaTypeMapping refDiscrimMapping = relatedTable.getDiscriminatorMapping();
                            Object[] discrimValues = null;
                            JavaTypeMapping refTypeMapping = null;

                            expressionsIndex = new int[referenceMapping.getNumberOfDatastoreFields()];
                            if (refDiscrimMapping != null)
                            {
                                // Related table has a discriminator
                                DiscriminatorStrategy strategy = fmd.getAbstractClassMetaData().getDiscriminatorStrategy();
                                if (strategy == DiscriminatorStrategy.CLASS_NAME)
                                {
                                    HashSet allclasses = new HashSet();
                                    if (clsNames != null)
                                    {
                                        for (int i=0;i<clsNames.length;i++)
                                        {
                                            allclasses.add(clsNames[i]);
                                        }
                                    }
                                 //   Collections.addAll(allclasses, clsNames);
                                    for( int i=0; i<clsNames.length; i++)
                                    {
                                        HashSet subclasses = table.getStoreManager().getSubClassesForClass(clsNames[i], true, clr);
                                        if (subclasses != null && subclasses.size() > 0)
                                        {
                                            allclasses.addAll(subclasses);
                                        }
                                    }
                                    discrimValues = allclasses.toArray();
                                }
                                else if (strategy == DiscriminatorStrategy.VALUE_MAP)
                                {
                                    HashSet subclasses = table.getStoreManager().getSubClassesForClass(fmd.getTypeName(), true, clr);
                                    if (subclasses != null && subclasses.size() > 0)
                                    {
                                        int valNo = 0;
                                        discrimValues = new Object[subclasses.size()+1];
                                        discrimValues[valNo++] = fmd.getAbstractClassMetaData().getInheritanceMetaData().getDiscriminatorMetaData().getValue();
                                        Iterator subclassesIter = subclasses.iterator();
                                        while (subclassesIter.hasNext())
                                        {
                                            String subclass = (String)subclassesIter.next();
                                            AbstractClassMetaData subclassCmd = table.getStoreManager().getMetaDataManager().getMetaDataForClass(subclass, clr);
                                            discrimValues[valNo++] = subclassCmd.getInheritanceMetaData().getDiscriminatorMetaData().getValue();
                                        }
                                    }
                                    else
                                    {
                                        discrimValues = new Object[1];
                                        discrimValues[0] = fmd.getAbstractClassMetaData().getMetaDataManager().getMetaDataForClass(fmd.getType(), clr).getInheritanceMetaData().getDiscriminatorMetaData().getValue();
                                    }
                                }
                            }
                            else if (relatedTable != referenceMapping.getDataStoreMapping(0).getDatastoreField().getDatastoreContainerObject())
                            {
                                // The reference mapping is stored in a base class and the field actually stores a subclass
                                // so we need to provide the reference type mapping too
                                refTypeMapping = relatedTable.getIDMapping();
                            }

                            if (refTypeMapping == null)
                            {
                                expressionsIndex = fetchStatement.selectOuterJoin(referenceMapping, selectMapping, refDiscrimMapping, discrimValues);
                            }
                            else
                            {
                                expressionsIndex = fetchStatement.selectOuterJoin(referenceMapping, selectMapping, refTypeMapping);
                            }
                        }
                    }
                    else if (relationType == Relation.MANY_TO_ONE_BI)
                    {
                        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                        if (fmd.getJoinMetaData() != null || relatedFmd.getJoinMetaData() != null)
                        {
                            // N-1 (Join Table) bidirectional field in element class without datastore column(s)
                            DatastoreContainerObject joinTable = table.getStoreManager().getDatastoreContainerObject(relatedFmd);
                            if (joinTable != null)
                            {
                                JavaTypeMapping referenceMapping = null;
                                JavaTypeMapping selectMapping = null;
                                if (java.util.List.class.isAssignableFrom(relatedFmd.getType()))
                                {
                                    ListTable listTable = (ListTable)joinTable;
                                    referenceMapping = listTable.getElementMapping();
                                    selectMapping = listTable.getOwnerMapping();
                                }
                                else
                                {
                                    SetTable setTable = (SetTable)joinTable;
                                    referenceMapping = setTable.getElementMapping();
                                    selectMapping = setTable.getOwnerMapping();
                                }
                                expressionsIndex = new int[referenceMapping.getNumberOfDatastoreFields()];
                                // TODO Do we need to allow for a discriminator in the owner class here ?
                                // TODO Do we need to allow for subclass table here (like in 1-1 case) ?
                                expressionsIndex = fetchStatement.selectOuterJoin(referenceMapping, selectMapping, null, null);
                            }
                        }
                    }
                }
                else
                {
                    // All other cases, where the field has datastore columns in the same table
                    expressionsIndex = new int[m.getNumberOfDatastoreFields()];
                    for (int i = 0; i < expressionsIndex.length; i++)
                    {
                        expressionsIndex[i] = fetchStmt.select(m.getDataStoreMapping(i).getDatastoreField());
                    }
                }
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setExpressionIndex(expressionsIndex);

                /*
                 * check if the our column is primary key. if the
                 * first column for the mapping is primary key,
                 * all columns for the same mapping must be primary key
                 */
                Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                if (fmd.isPrimaryKey())
                {
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        fetchStmt.andCondition(fetchStmt.referenceDatastoreField(m.getDataStoreMapping(j).getDatastoreField()) + "=?");

                        if (!pkbp.contains(abs_field_num))
                        {
                            pkbp.add(abs_field_num);
                        }

                        parametersIndex[j] = paramIndex++;
                    }
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                }
                else
                {
                    if (!ftbr.contains(abs_field_num))
                    {
                        ftbr.add(abs_field_num);
                    }
                }
            }

            if (m instanceof MappingCallbacks)
            {
                mappingCallbacks.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field
         * @param m The mapping
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                // version column
                mappingStatementIndex.getVersion().setMapping(table.getVersionMapping());
                //create the expressions index (columns index)
                int expressionsIndex[] = new int[table.getVersionMapping().getNumberOfDatastoreFields()];
                for (int j = 0; j < expressionsIndex.length; j++)
                {
                    expressionsIndex[j] = fetchStmt.select(table.getVersionMapping().getDataStoreMapping(j).getDatastoreField());
                }
                mappingStatementIndex.getVersion().setExpressionIndex(expressionsIndex);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                fetchStmt.andCondition(fetchStmt.referenceDatastoreField((DatastoreField) key.getColumns().get(0)) + "=?");
                mappingStatementIndex.getDatastoreId().setParameterIndex(new int[]{paramIndex});
                paramIndex++;
            }
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mappingCallbacks;
        }

        /**
         * Obtain the field numbers which values are to be replaced
         * @return the array of field numbers 
         */        
        public int[] getFieldsToBeReplaced()
        {
            int[] fieldsToBeReplaced = new int[ftbr.size()];
            for (int i = 0; i < ftbr.size(); ++i)
            {
                fieldsToBeReplaced[i] = ((Integer) ftbr.get(i)).intValue();
            }
            return fieldsToBeReplaced;
        }

        /**
         * Obtain the primary key field numbers which values are to be provided by the StateManager
         * @return the array of primary key field numbers 
         */           
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }
    }
}
