/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson  - optimistic transaction support
2004 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
2005 Andy Jefferson - added handling for updating FK in related object
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOException;
import javax.jdo.JDOOptimisticVerificationException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.Relation;
import org.jpox.metadata.VersionStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.FetchStatement;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.AbstractClassTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Class to provide a means of update of records in a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * update operation.
 *
 * @version $Revision: 1.51 $ 
 **/
public class UpdateRequest extends Request
{
    /** SQL statement for the update. */
    private final String updateStmt;

    /** callback mappings will have their postUpdate method called after the update */
    private final MappingCallbacks[] callbacks;
    
    /** the index for the expression in the update sql statement. */
    private MappingStatementIndex mappingStatementIndex;

    /** Fields to be provided for the update. */
    private final int[] fieldsToBeProvided;

    /** PK fields to be provided (if using app identity) for the update. */
    private final int[] pkFieldsToBeProvided;

    /** MetaData for fields with related objects to those in this class. */
    private final AbstractPropertyMetaData[] fieldsWithRelatedObjects;

    /** Highest field number for this class being updated. */
    private final int highestFieldNumber;

    /** index for the expression in the version sql statement */
    private MappingStatementIndex versionMappingStatementIndex;

    /** SQL statement for retrieving the version. */
    private final String versionStmt;

    private final boolean versionChecks;

    /** PK fields to be provided (if using app identity) for the version query. */
    private final int[] versionPkFieldsToBeProvided;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query. The query will be of the form
     * <PRE>
     * UPDATE table-name
     * SET param1=?,param2=?
     * WHERE param3=? AND param4=?
     *</PRE> 
     *
     * @param table The Class Table representing the datastore table to update
     * @param reqFieldMetaData MetaData of the fields to update
     * @param cls Class of objects being updated
     * @param clr ClassLoader resolver
     **/
    public UpdateRequest(DatastoreClass table, AbstractPropertyMetaData[] reqFieldMetaData, Class cls, ClassLoaderResolver clr)
    {
        super(table);

        AbstractClassMetaData cmd = table.getStoreManager().getMetaDataManager().getMetaDataForClass(cls, clr);

        // Set up the basic mapping information
        mappingStatementIndex = new MappingStatementIndex(); // Populated using the subsequent lines
        UpdateMappingConsumer consumer = new UpdateMappingConsumer(clr, cmd);
        table.provideMappingsForFields(consumer, reqFieldMetaData, false);
        table.provideVersionMappings(consumer);
        table.providePrimaryKeyMappings(consumer);
        table.provideDatastoreIdMappings(consumer);
        updateStmt = consumer.getStatement();
        callbacks = (MappingCallbacks[])consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);
        pkFieldsToBeProvided = consumer.getPrimaryKeyFieldsToBeProvided();
        fieldsToBeProvided = consumer.getFieldsToBeProvided();
        fieldsWithRelatedObjects = consumer.getFieldsWithRelatedObjects();
        highestFieldNumber = consumer.getHighestFieldNumber();

        // Set up any version mapping information
        versionMappingStatementIndex = new MappingStatementIndex(); // Populated using the subsequent lines
        VersionMappingConsumer consumerVersion = new VersionMappingConsumer(cmd);
        table.provideVersionMappings(consumerVersion);
        table.providePrimaryKeyMappings(consumerVersion);
        table.provideDatastoreIdMappings(consumerVersion);
        versionStmt = consumerVersion.getStatement();
        versionPkFieldsToBeProvided = consumerVersion.getPrimaryKeyFieldsToBeProvided();
        if (cmd.getVersionMetaData() != null && cmd.getVersionMetaData().getVersionStrategy() != VersionStrategy.NONE)
        {
            versionChecks = true;
        }
        else
        {
            versionChecks = false;
        }
    }

    /**
     * Method performing the update of the record in the datastore. Takes
     * the constructed update query and populates with the specific record
     * information.
     * @param sm The state manager for the record to be updated
     **/
    public void execute(StateManager sm)
    {
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            // Debug information about what we are deleting
            StringBuffer str = new StringBuffer("Updating object " + StringUtils.toJVMIDString(sm.getObject()) + 
                " of type " + sm.getObject().getClass().getName() + " in table " + table);
            JPOXLogger.JDO.debug(str);
        }

        if (updateStmt != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
            try
            {
                Connection conn = storeMgr.getConnection(pm, true, false);

                try
                {
                    PreparedStatement ps;
                    if (pm.currentTransaction().getOptimistic())
                    {
                        if (versionStmt != null)
                        {
                            // Version mapping present
                            ps = storeMgr.getStatement(conn, versionStmt, false);

                            try
                            {
                                // provide primary key fields for WHERE clause
                                if (table.getIdentityType() == IdentityType.DATASTORE)
                                {
                                    // a). datastore identity
                                    table.getDataStoreObjectIdMapping().setObject(pm, ps, 
                                        versionMappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
                                }
                                else if (table.getIdentityType() == IdentityType.APPLICATION)
                                {
                                    // b). application identity
                                    sm.provideFields(versionPkFieldsToBeProvided,
                                        new ParameterSetter(sm, ps, versionMappingStatementIndex.getPrimaryKeys(), true));
                                }

                                ResultSet rs = storeMgr.executeStatementQuery(versionStmt, ps);
                                try
                                {
                                    if (rs.next())
                                    {
                                        Object datastoreVersion = table.getVersionMapping().getObject(pm, rs, 
                                            versionMappingStatementIndex.getVersion().getExpressionIndex());
                                        boolean passedVerification = true;
                                        if (versionChecks)
                                        {
                                            passedVerification = ((VersionMapping)table.getVersionMapping()).compareVersion(datastoreVersion, 
                                                sm.getTransactionalVersion(sm.getObject()));
                                        }
                                        if (versionChecks && !passedVerification)
                                        {
                                            String msg = LOCALISER.msg("RDBMS.Request.OptimisticVersionMismatch", 
                                                sm.getInternalObjectId(), table.toString(), datastoreVersion, sm.getTransactionalVersion(sm.getObject()));
                                            JPOXLogger.RDBMS.error(msg);
                                            throw new JDOOptimisticVerificationException(msg, sm.getObject());
                                        }
                                    }
                                    else
                                    {
                                        String msg = LOCALISER.msg("RDBMS.Request.OptimisticVersionMissing", 
                                            sm.getInternalObjectId(), table.toString());
                                        JPOXLogger.RDBMS.error(msg);
                                        throw new JDOOptimisticVerificationException(msg, sm.getObject());
                                    }
                                }
                                finally
                                {
                                    rs.close();
                                }
                            }
                            finally
                            {
                                ps.close();
                            }
                        }
                        else
                        {
                            // TODO No version mapping so retrieve the datastore object and do a field-by-field comparison check
                        }
                    }

                    ps = storeMgr.getStatement(conn, updateStmt, false);
                    try
                    {
                        // Set the required fields to be updated
                        if (fieldsToBeProvided != null)
                        {
                            sm.provideFields(fieldsToBeProvided, new ParameterSetter(sm, ps, mappingStatementIndex.getFields(),true));
                        }

                        // provide primary key fields for WHERE clause
                        if (table.getIdentityType() == IdentityType.DATASTORE)
                        {
                            // a). datastore identity
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, mappingStatementIndex.getDatastoreId().getParameterIndex(), 
                                sm.getInternalObjectId());
                        }
                        else if (table.getIdentityType() == IdentityType.APPLICATION)
                        {
                            // b). application identity
                            sm.provideFields(pkFieldsToBeProvided, new ParameterSetter(sm, ps, mappingStatementIndex.getPrimaryKeys(),true));
                        }

                        if (versionStmt != null)
                        {
                            // Using version so update the latest version
                            Object nextOptimisticVersion = ((VersionMapping)table.getVersionMapping()).getNextVersion(sm.getTransactionalVersion(sm.getObject()));
                            table.getVersionMapping().setObject(pm, ps, mappingStatementIndex.getVersion().getParameterIndex(), 
                                nextOptimisticVersion);
                            sm.setTransactionalVersion(nextOptimisticVersion);
                        }

                        storeMgr.executeStatementUpdate(updateStmt, ps);
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    storeMgr.releaseConnection(pm, conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Update request failed: " + updateStmt, e);
            }
        }

        // Update any related objects where we use their FK, and the FK is to be updated.
        // This is used where we have a 1-1 bidirectional and this object has no actual field,
        // instead using an FK in the related object to maintain the linkage.
        if (fieldsWithRelatedObjects != null && fieldsWithRelatedObjects.length > 0)
        {
            for (int i=0;i<fieldsWithRelatedObjects.length;i++)
            {
                updateRelatedObjectsForField(sm, fieldsWithRelatedObjects[i]);
            }
        }

        // Execute any mapping actions now that we have done the update
        for (int i=0; i<callbacks.length; ++i)
        {
            try
            {
                callbacks[i].postUpdate(sm);
            }
            catch (NotYetFlushedException e)
            {
                sm.updateFieldAfterInsert(e.getPersistenceCapable(),((JavaTypeMapping)callbacks[i]).getFieldMetaData().getAbsoluteFieldNumber());
            }
        }
    }

    /**
     * Method to update any related objects.
     * @param sm StateManager of this object
     * @param fmd MetaData for field that has related objects
     */
    private void updateRelatedObjectsForField(StateManager sm, AbstractPropertyMetaData fmd)
    {
        ClassLoaderResolver clr = sm.getPersistenceManager().getClassLoaderResolver();
        DatastoreClass refTable = sm.getStoreManager().getDatastoreClass(fmd.getTypeName(), clr);
        //TODO to support javax.jdo.option.mapping.RelationSubclassTable, we should go through all sub types if subclass inheritance strategy is used
        // and fire many UPDATE statements (one for each table)
        if( refTable == null )
        {
        	throw new JDOException("Could not find table for for type "+fmd.getTypeName()+". You should verify if the inheritance strategy is valid.");
        }
        AbstractPropertyMetaData refFmd = fmd.getRelatedFieldMetaData(clr);
        JavaTypeMapping refMapping = refTable.getFieldMapping(fmd.getMappedBy());

        PersistenceManager pm = sm.getPersistenceManager();
        RDBMSManager storeMgr = (RDBMSManager)pm.getStoreManager();
        try
        {
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                JavaTypeMapping refIdMapping;
                if (refFmd.getAbstractClassMetaData().getIdentityType() == IdentityType.DATASTORE)
                {
                    refIdMapping = refTable.getDataStoreObjectIdMapping();
                }
                else
                {
                    refIdMapping = refTable.getIDMapping();
                }
                // A. Null out the existing related object FK link
                // Create a statement to clear the link from the previous related object
                // Only clear the link if the new related object has not changed from the previous related object
                // UPDATE tbl SET tbl.FK_FIELD=NULL WHERE tbl.FK_FIELD=?THIS_ID? AND tbl.ID<>?FK_ID?
                StringBuffer clearLinkStmt = new StringBuffer("UPDATE " + refTable.toString() + " SET ");
                for (int j=0;j<refMapping.getNumberOfDatastoreFields();j++)
                {
                    if (j > 0)
                    {
                        clearLinkStmt.append(",");
                    }
                    clearLinkStmt.append(refMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier());
                    clearLinkStmt.append(" =NULL ");
                }
                clearLinkStmt.append(" WHERE ");
                for (int j=0;j<refMapping.getNumberOfDatastoreFields();j++)
                {
                    if (j > 0)
                    {
                        clearLinkStmt.append(" AND ");
                    }
                    clearLinkStmt.append(refMapping.getDataStoreMapping(j).getDatastoreField().toString());
                    clearLinkStmt.append(" =? ");
                }
                if( sm.provideField(fmd.getAbsoluteFieldNumber()) != null )
                {
                    for (int j=0;j<refIdMapping.getNumberOfDatastoreFields();j++)
                    {
                        clearLinkStmt.append(" AND ");
                        clearLinkStmt.append(refIdMapping.getDataStoreMapping(j).getDatastoreField().toString());
                        clearLinkStmt.append(" <>? ");
                    }
                }
                PreparedStatement ps = null;
                try
                {
                    ps = storeMgr.getStatement(conn, clearLinkStmt.toString(), false);
                    refMapping.setObject(pm, ps, Mappings.getParametersIndex(1, refMapping), sm.getObject());
                    if( sm.provideField(fmd.getAbsoluteFieldNumber()) != null )
                    {
                        StatementExpressionIndex[] indexes = new StatementExpressionIndex[highestFieldNumber];
                        StatementExpressionIndex index = new StatementExpressionIndex();
                        index.setMapping(refIdMapping);
                        int[] paramIndexes = new int[refIdMapping.getNumberOfDatastoreFields()];
                        for (int i=0;i<paramIndexes.length;i++)
                        {
                            paramIndexes[i] = refMapping.getNumberOfDatastoreFields() + 1 + i;
                        }
                        index.setParameterIndex(paramIndexes);
                        indexes[fmd.getAbsoluteFieldNumber()] = index;
                        sm.provideFields(new int[] {fmd.getAbsoluteFieldNumber()}, new ParameterSetter(sm, ps, indexes, false));
                    }
                    storeMgr.executeStatementUpdate(clearLinkStmt.toString(), ps);
                }
                finally
                {
                    if (ps != null)
                    {
                        ps.close();
                    }
                }
                if( sm.provideField(fmd.getAbsoluteFieldNumber()) != null )
                {
                    // B. Set the FK link on the new object (if any)
                    // TODO ONLY do this if we are setting the link to a NEW VALUE(CHANGED)
                    // Only do this if we are setting the link to a new value (and not null)
                    ps = null;
                    try
                    {
                        // Create a statement to set the link on the new related object (if any)
                        // UPDATE tbl SET FK_FIELD=THIS_ID WHERE ID=FK_ID
                        StringBuffer newLinkStmt = new StringBuffer("UPDATE " + refTable.toString() + " SET ");
                        for (int j=0;j<refMapping.getNumberOfDatastoreFields();j++)
                        {
                            if (j > 0)
                            {
                                newLinkStmt.append(",");
                            }
                            newLinkStmt.append(refMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier());
                            newLinkStmt.append("=?");
                        }
                        newLinkStmt.append(" WHERE ");
                        for (int j=0;j<refIdMapping.getNumberOfDatastoreFields();j++)
                        {
                            if (j > 0)
                            {
                                newLinkStmt.append(" AND ");
                            }
                            newLinkStmt.append(refIdMapping.getDataStoreMapping(j).getDatastoreField().toString());
                            newLinkStmt.append("=?");
                        }
    
                        ps = storeMgr.getStatement(conn, newLinkStmt.toString(), false);
                        refMapping.setObject(pm, ps, Mappings.getParametersIndex(1, refMapping), sm.getObject());
    
                        StatementExpressionIndex[] indexes = new StatementExpressionIndex[highestFieldNumber];
                        StatementExpressionIndex index = new StatementExpressionIndex();
                        index.setMapping(refIdMapping);
                        int[] paramIndexes = new int[refIdMapping.getNumberOfDatastoreFields()];
                        for (int i=0;i<paramIndexes.length;i++)
                        {
                            paramIndexes[i] = refMapping.getNumberOfDatastoreFields() + 1 + i;
                        }
                        index.setParameterIndex(paramIndexes);
                        indexes[fmd.getAbsoluteFieldNumber()] = index;
                        sm.provideFields(new int[] {fmd.getAbsoluteFieldNumber()}, new ParameterSetter(sm, ps, indexes, false));
                        storeMgr.executeStatementUpdate(newLinkStmt.toString(), ps);
                    }
                    finally
                    {
                        if (ps != null)
                        {
                            ps.close();
                        }
                    }
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new JDODataStoreException("Update request failed", e);
        }
    }

    private class UpdateMappingConsumer implements MappingConsumer
    {
        boolean initialized = false;

        int paramIndex = 1;
        int highestFieldNumber = 0;

        List ftbp = new ArrayList();
        List mc = new ArrayList();
        List pkbp = new ArrayList();
        List fwro = new ArrayList();

        /** for UPDATE statement **/
        StringBuffer columnAssignments = new StringBuffer();

        Map assignedColumns = new HashMap();

        StringBuffer where = new StringBuffer();
        
        private final ClassLoaderResolver clr;
        private final AbstractClassMetaData cmd;

        /**
         * Constructor
         * @param clr the ClassLoaderResolver
         */
        public UpdateMappingConsumer(ClassLoaderResolver clr, AbstractClassMetaData cmd)
        {
            super();
            this.clr = clr;
            this.cmd = cmd;
        }

        /**
         * Accessor for the highest field number consumed.
         * @return The highest (absolute) field number
         */
        public int getHighestFieldNumber()
        {
            return highestFieldNumber;
        }

        public void preConsumeMapping(int highest)
        {
            if (!initialized)
            {
                highestFieldNumber = highest;
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highest]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highest]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (!fmd.getAbstractClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            if (m.includeInUpdateStatement())
            {
                Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];

                if (m.getNumberOfDatastoreFields() == 0 && m instanceof PersistenceCapableMapping)
                {
                    int relationType = fmd.getRelationType(clr);
                    if (relationType == Relation.ONE_TO_ONE_BI)
                    {
                        if (fmd.getMappedBy() != null)
                        {
                            // 1-1 bidirectional field without datastore column(s) (with single FK)
                            fwro.add(fmd);
                        }
                    }
                    else if (relationType == Relation.MANY_TO_ONE_BI)
                    {
                        AbstractPropertyMetaData relatedFmd = fmd.getRelatedFieldMetaData(clr);
                        if (fmd.getJoinMetaData() != null || relatedFmd.getJoinMetaData() != null)
                        {
                            // N-1 bidirectional using join table for relation.
                            // TODO Need to update the join table with this relation
                        }
                    }
                }

                StatementExpressionIndex sei = new StatementExpressionIndex();
                sei.setMapping(m);
                sei.setParameterIndex(parametersIndex);

                if (fmd.isPrimaryKey())
                {
                    //primary keys goes to the UPDATE SET COLUMN = ?, ....
                    mappingStatementIndex.getPrimaryKeys()[fmd.getAbsoluteFieldNumber()] = sei; 
                    for (int j=0; j<parametersIndex.length; j++)
                    {
                        if (where.length() > 0)
                        {
                            where.append(" AND ");
                        }
                        String condition = m.getDataStoreMapping(j).getDatastoreField().getIdentifier() + 
                            "=" + ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();
                        where.append(condition);

                        if (!pkbp.contains(abs_field_num))
                        {
                            pkbp.add(abs_field_num);
                        }
                        parametersIndex[j] = paramIndex++;
                    }
                }
                else
                {
                    //non primary keys goes to the WHERE COLUMN = ?, ....
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = sei;

                    for (int j = 0; j < parametersIndex.length; j++)
	                {
	                    // check if the column was not already assigned
	                    if (!assignedColumns.containsKey(m.getDataStoreMapping(j).getDatastoreField().getIdentifier().toString()))
	                    {
	                        if (columnAssignments.length() > 0)
	                        {
	                            columnAssignments.append(", ");
	                        }

                            String param = ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();
	                        columnAssignments.append(m.getDataStoreMapping(j).getDatastoreField().getIdentifier()).append("=").append(param);
                            if (param.indexOf("?") > -1) // only add fields to be replaced by the real values only if the param value has '?'
                            {
                                if (!ftbp.contains(abs_field_num))
                                {
	                                ftbp.add(abs_field_num);
                                }
	                            parametersIndex[j] = paramIndex++;
                            }
	                        assignedColumns.put(m.getDataStoreMapping(j).getDatastoreField().getIdentifier().toString(), 
                                new Integer(fmd.getAbsoluteFieldNumber()));
	                    }
	                    else
	                    {
	                        parametersIndex[j] = ((Integer)assignedColumns.get(m.getDataStoreMapping(j).getDatastoreField().getIdentifier().toString())).intValue();
	                    }
	                }
                }
            }

            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                String val = ((RDBMSMapping)m.getDataStoreMapping(0)).getUpdateInputParameter();
                if (columnAssignments.length() > 0)
                {
                    columnAssignments.append(", ");
                }
                String condition = m.getDataStoreMapping(0).getDatastoreField().getIdentifier() + "=" + val;
                columnAssignments.append(condition);

                mappingStatementIndex.getVersion().setMapping(m);
                int[] param = { paramIndex++ };
                mappingStatementIndex.getVersion().setParameterIndex(param);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                String condition = ((DatastoreField)key.getColumns().get(0)).getIdentifier() + "=? ";
                where.append(condition);

                int[] param = { paramIndex++ };
                mappingStatementIndex.getDatastoreId().setParameterIndex(param);
            }         
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mc;
        }

        /**
         * All non-primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         * @return array of absolute field numbers
         */
        public int[] getFieldsToBeProvided()
        {
            int[] fieldsToBeProvided = new int[ftbp.size()];
            for (int i = 0; i < ftbp.size(); ++i)
            {
                fieldsToBeProvided[i] = ((Integer) ftbp.get(i)).intValue();
            }
            return fieldsToBeProvided;
        }

        /**
         * All primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         * @return array of absolute primary key field numbers
         */
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        /**
         * All fields that have related objects and where we use the related object to
         * maintain a relationship. These are used where we have 1-1 bidirectional
         * relationships with a single FK (in the other object).
         * @return The fields with related objects
         */
        public AbstractPropertyMetaData[] getFieldsWithRelatedObjects()
        {
            AbstractPropertyMetaData[] fieldsWithRelatedObjects = new AbstractPropertyMetaData[fwro.size()];
            for (int i = 0; i < fwro.size(); ++i)
            {
                fieldsWithRelatedObjects[i] = (AbstractPropertyMetaData) fwro.get(i);
            }
            return fieldsWithRelatedObjects;
        }

        /**
         * Accessor for the basic update SQL statement.
         * <PRE>
         * UPDATE TABLE SET COL1 = ?, COL2 = ? WHERE COL3 = ? AND COL4 = ? 
         * </PRE>
         * @return The update SQL statement
         */
        public String getStatement()
        {
            if (columnAssignments.length() < 1)
            {
                return null;
            }
            return "UPDATE " + table.toString() + " SET " + columnAssignments + " WHERE " + where;
        }
    }

    private class VersionMappingConsumer implements MappingConsumer
    {
        boolean initialized = false;
        boolean hasVersionMapping = false;
        int paramIndex = 1;
        int exprIndex = 1;
        
        List pkbp = new ArrayList();

        /** for SELECT statement **/
        FetchStatement fetchStmt = new RDBMSFetchStatement(((AbstractClassTable)table).getPrimaryDatastoreContainerObject());

        private final AbstractClassMetaData cmd;

        /**
         * Default constructor
         */
        public VersionMappingConsumer(AbstractClassMetaData cmd)
        {
            this.cmd = cmd;
        }
        
        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                versionMappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, AbstractPropertyMetaData fmd)
        {
            if (!fmd.getAbstractClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }            
            if (m.includeInUpdateStatement())
            {
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];

                StatementExpressionIndex sei = new StatementExpressionIndex();
                sei.setMapping(m);
                sei.setParameterIndex(parametersIndex);

                if (fmd.isPrimaryKey())
                {
                    versionMappingStatementIndex.getPrimaryKeys()[fmd.getAbsoluteFieldNumber()] = sei; 
                    for (int j=0; j<parametersIndex.length; j++)
                    {                   
                        String condition = m.getDataStoreMapping(j).getDatastoreField().getIdentifier() + 
                            "=" + ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();
                        fetchStmt.andCondition(condition);

                        pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        parametersIndex[j] = paramIndex++;
                    }                       
                }
            }
        }

        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                hasVersionMapping = true;

                fetchStmt.select(m.getDataStoreMapping(0).getDatastoreField());
                
                versionMappingStatementIndex.getVersion().setMapping(m);
                versionMappingStatementIndex.getVersion().setExpressionIndex(new int[]{ exprIndex++ });
            }             
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                String condition = ((DatastoreField)key.getColumns().get(0)).getIdentifier() + "=? ";
                fetchStmt.andCondition(condition);
                
                int[] param = { paramIndex++ };
                versionMappingStatementIndex.getDatastoreId().setParameterIndex(param);
            }
        }
        
        /**
         * All primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         * @return array of absolute primary key field numbers
         */        
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        /**
         * Obtain the SQL statement
         * @return the SQL statement
         */
        public String getStatement()
        {
            if (!hasVersionMapping)
            {
                return null;
            }
            return fetchStmt.toString();
        }
    }
}