/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 


Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.lang.reflect.Array;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.StringLiteral;
import org.jpox.store.mapping.EmbeddedElementPCMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.mapping.SerialisedPCMapping;
import org.jpox.store.mapping.SerialisedReferenceMapping;
import org.jpox.store.query.ResultObjectFactory;
import org.jpox.store.rdbms.RDBMSTransaction;
import org.jpox.store.rdbms.table.JoinTable;
import org.jpox.store.scostore.ArrayStore;
import org.jpox.util.JPOXLogger;

/**
 * Abstract representation of the backing store for an array.
 * 
 * @version $Revision: 1.15 $
 */
public abstract class AbstractArrayStore extends ElementContainerStore implements ArrayStore
{
    /** Statement for adding an element to an array. */
    protected String addStmt;

    /**
     * Constructor.
     * @param storeMgr Manager for the store
     * @param clr ClassLoader resolver
     */
    protected AbstractArrayStore(StoreManager storeMgr, ClassLoaderResolver clr)
    {
        super(storeMgr, clr);
    }

    public List getArray(StateManager ownerSM)
    {
        Iterator iter = iterator(ownerSM);
        List elements = new ArrayList();
        while (iter.hasNext())
        {
            Object obj = iter.next();
            elements.add(obj);
        }

        return elements;
    }

    /**
     * Clear the association from owner to all elements.
     * Observes the necessary dependent field settings with respect to whether
     * it should delete the element when doing so.
     * @param ownerSM State Manager for the container. 
     */
    public void clear(StateManager ownerSM)
    {
        Collection dependentElements = null;
        if (ownerFieldMetaData.getArray().isDependentElement())
        {
            // Retain the dependent elements that need deleting after clearing
            dependentElements = new HashSet();
            Iterator iter = iterator(ownerSM);
            while (iter.hasNext())
            {
                dependentElements.add(iter.next());
            }
        }

        try
        {
            PersistenceManager pm = ownerSM.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, clearStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    jdbcPosition = populateOwnerInStatement(ownerSM, pm, ps, jdbcPosition);
                    if (relationDiscriminatorMapping != null)
                    {
                        jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                    }

                    storeMgr.executeStatementUpdate(clearStmt, ps);
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.ClearRequestFailed",clearStmt),e);
        }

        if (dependentElements != null && dependentElements.size() > 0)
        {
            ownerSM.getPersistenceManager().deletePersistentAll(dependentElements);
        }
    }

    /**
     * Method to set the array for the specified owner to the passed value.
     * @param ownerSM State Manager for the owner
     * @param array the array
     * @return Whether the array was updated successfully
     */
    public boolean set(StateManager ownerSM, Object array)
    {
        if (array == null || Array.getLength(array) == 0)
        {
            return true;
        }

        // Validate all elements for writing
        int length = Array.getLength(array);
        for (int i=0;i<length;i++)
        {
            Object obj = Array.get(array, i);
            validateElementForWriting(ownerSM, obj);
        }

        boolean modified = false;
        List exceptions = new ArrayList();
        boolean batched = allowsBatching() && length > 1;

        try
        {
            PersistenceManager pm = ownerSM.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, addStmt, false);

            try
            {
                // Loop through all elements to be added
                Object element = null;
                for (int i=0;i<length;i++)
                {
                    element = Array.get(array, i);

                    // Add the row to the join table
                    try
                    {
                        modified = internalAdd(ownerSM, element, ps, batched, i);
                    }
                    catch (SQLException sqle)
                    {
                        sqle.printStackTrace();
                        exceptions.add(sqle);
                        JPOXLogger.RDBMS.error(sqle);
                    }
                }

                if (batched && exceptions.size() == 0)
                {
                    // Execute the batch
                    int[] num = storeMgr.executeStatementBatch(addStmt, ps);
                    if (num == null)
                    {
                        modified = false;
                    }
                    else
                    {
                        for (int i=0;i<num.length;i++)
                        {
                            if (num[i] > 0)
                            {
                                modified = true;
                            }
                        }
                    }
                }
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            e.printStackTrace();
            exceptions.add(e);
            JPOXLogger.RDBMS.error(e);
        }

        if (!exceptions.isEmpty())
        {
            // Throw all exceptions received as the cause of a JDODataStoreException so the user
            // can see which record(s) didn't persist
            String msg = LOCALISER.msg("RDBMS.SCO.AddRequestFailed", addStmt);
            JPOXLogger.RDBMS.error(msg);
            throw new JDODataStoreException(msg, (Throwable[])exceptions.toArray(new Throwable[exceptions.size()]), ownerSM.getObject());
        }

        return modified;
    }

    /**
     * Adds one element to the association owner vs elements
     * @param sm State Manager for the container
     * @param element The element to add
     * @param position The position to add this element at
     * @return Whether it was successful
     */
    public boolean add(StateManager sm, Object element, int position)
    {
        validateElementForWriting(sm, element);

        boolean modified = false;

        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, addStmt, false);

            try
            {
                // Add a row to the join table
                modified = internalAdd(sm, element, ps, false, position);
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.AddRequestFailed",addStmt),e);
        }

        return modified;
    }

    /**
     * Internal method to add a row to the join table.
     * Used by add() and set() to add a row to the join table.
     * @param sm StateManager for the owner of the collection
     * @param element The element to add the relation to
     * @param ps PreparedStatement to use
     * @param batched Whether we are batching
     * @param orderId The order id to use for this element relation
     * @return Whether a row was inserted
     * @throws SQLException Thrown if an error occurs
     */
    protected boolean internalAdd(StateManager sm, Object element, PreparedStatement ps, boolean batched, int orderId)
    throws SQLException
    {
        boolean modified = false;

        PersistenceManager pm = sm.getPersistenceManager();

        // Insert the join table row
        int jdbcPosition = 1;
        jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
        jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
        jdbcPosition = populateOrderInStatement(pm, ps, orderId, jdbcPosition);
        if (relationDiscriminatorMapping != null)
        {
            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
        }

        if (batched)
        {
            // Batch the statement
            ps.addBatch();
        }
        else
        {
            int num = storeMgr.executeStatementUpdate(addStmt, ps);
            if (num > 0)
            {
                modified = true;
            }
        }

        return modified;
    }

    /**
     * Accessor for the statement for the iterator.
     * @param ownerSM the owner StateManager
     * @return The iterator Query Statement.
     **/
    protected abstract QueryExpression getIteratorStatement(StateManager ownerSM);

    /**
     * Accessor for an iterator through the array elements.
     * @param ownerSM State Manager for the container.
     * @return The Iterator
     **/
    public Iterator iterator(StateManager ownerSM)
    {
        // Create the basic statement (without any selected columns)
        QueryExpression stmt = getIteratorStatement(ownerSM);
        if (stmt == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.SCO.IteratorStatementIsNull"));
        }

        // Add the required field selections to the statement and get a factory for extracting them
        ResultObjectFactory rof = newResultObjectFactory(ownerSM, stmt, false, true);

        PersistenceManager pm = ownerSM.getPersistenceManager();
        RDBMSTransaction tx = (RDBMSTransaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        if (statement == null)
        {
            // this happens when collection is null
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.SCO.IteratorStatementIsNull"));
        }

        Iterator iter;
        try
        {
            Connection conn = storeMgr.getConnection(pm, false, false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm, conn);
                try
                {
                    ResultSet rs = storeMgr.executeStatementQuery(statement, ps);
                    try
                    {
                        iter = new ArrayStoreIterator(ownerSM, rs, rof);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.IteratorRequestFailed",statement),e);
        }

        return iter;
    }

    /**
     * Inner class providing iterator.
     **/
    private class ArrayStoreIterator implements Iterator
    {
        private final PersistenceManager pm;
        private final Iterator delegate;
        private Object lastElement = null;

        /**
         * Constructor
         * @param sm the StateManager
         * @param rs the ResultSet
         * @param rof the Query.ResultObjectFactory
         * @throws SQLException
         */
        public ArrayStoreIterator(StateManager sm, ResultSet rs, ResultObjectFactory rof) 
        throws SQLException
        {
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();
            if (rs != null)
            {
                while (rs.next())
                {
                    Object nextElement;
                    if (elementsAreEmbedded || elementsAreSerialised)
                    {
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }

                        if (elementMapping instanceof SerialisedPCMapping ||
                            elementMapping instanceof SerialisedReferenceMapping ||
                            elementMapping instanceof EmbeddedElementPCMapping)
                        {
                            // Element = Serialised
                            int ownerFieldNumber = -1;
                            if (containerTable != null)
                            {
                                ownerFieldNumber = ((JoinTable)containerTable).getOwnerFieldMetaData().getAbsoluteFieldNumber();
                            }
                            nextElement = elementMapping.getObject(pm, rs, param, sm, ownerFieldNumber);
                        }
                        else
                        {
                            // Element = Non-PC
                            nextElement = elementMapping.getObject(pm, rs, param);
                        }
                    }
                    else if (elementMapping instanceof ReferenceMapping)
                    {
                        // Element = Reference (Interface/Object)
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }
                        nextElement = elementMapping.getObject(pm, rs, param);
                    }
                    else
                    {
                        // Element = PC
                        nextElement = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                    }

                    results.add(nextElement);
                }
            }
            delegate = results.iterator();
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public Object next()
        {
            lastElement = delegate.next();

            return lastElement;
        }

        public synchronized void remove()
        {
            // Do nothing
        }
    }
    
    /**
     * JDOQL utility to generate an exists() statement for an element.
     * The generated query will be of the form
     * <PRE>
     * SELECT 1 FROM JOINTABLE THIS_JOIN WHERE THIS_JOIN.OWNER_ID_OID = THIS.OWNER_ID
     * </PRE>
     * @param qs The JDOQL query statement
     * @param mapping mapping of the field
     * @param ownerTe Expression for the table
     * @param collectionRangeVar range variable for the collection
     * @return The JDOQL query statement
     */
    public QueryExpression getExistsSubquery(QueryExpression qs,
                                             JavaTypeMapping mapping,
                                             LogicSetExpression ownerTe,
                                             DatastoreIdentifier collectionRangeVar)
    {
        QueryExpression stmt = dba.newQueryStatement(containerTable, collectionRangeVar, qs.getClassLoaderResolver());

        // Join for the owner
        ScalarExpression ownerExpr = mapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerInCollectionExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(collectionRangeVar));
        stmt.andCondition(ownerExpr.eq(ownerInCollectionExpr));

        stmt.select(collectionRangeVar, elementMapping);

        return stmt;
    }
    
    /**
     * JDOQL utility to generate a subquery for the size() of the collection.
     * The generated query will be of the form
     * <PRE>
     * SELECT COUNT(*) FROM JOINTABLE THIS_JOIN WHERE THIS_JOIN.OWNER_ID_OID = THIS.OWNER_ID
     * </PRE>
     * @param qs The JDOQL query statement
     * @param mapping mapping of the field
     * @param ownerTe Expression for the table
     * @param collectionRangeVar range variable for the collection
     * @return The JDOQL query statement
     */
    public QueryExpression getSizeSubquery(QueryExpression qs,
                                           JavaTypeMapping mapping,
                                           LogicSetExpression ownerTe,
                                           DatastoreIdentifier collectionRangeVar)
    {
        QueryExpression stmt = dba.newQueryStatement(containerTable, collectionRangeVar, qs.getClassLoaderResolver());

        // Join for the owner
        ScalarExpression ownerExpr = mapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerInCollectionExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(collectionRangeVar));
        stmt.andCondition(ownerExpr.eq(ownerInCollectionExpr));

        // Select COUNT(*)
        JavaTypeMapping m = dba.getMapping(String.class, storeMgr);
        StringLiteral lit = (StringLiteral)m.newLiteral(stmt, "COUNT(*)");
        lit.generateStatementWithoutQuotes();
        stmt.selectScalarExpression(lit);

        return stmt;
    }    
}