/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.PreparedStatement;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.state.StateManagerFactory;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.RDBMSAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.JoinTable;
import org.jpox.util.Localiser;

/**
 * Base class for all RDBMS container stores (collections, maps, arrays).
 * Provides a series of helper methods for handling the store process.
 *
 * @version $revision$
 */
abstract class BaseContainerStore
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /** Manager for the store. */
    protected RDBMSManager storeMgr;

    /** Datastore adapter in use by this store. */
    protected DatastoreAdapter dba;

    /** Mapping to the owner of the container. */
    protected JavaTypeMapping ownerMapping;

    /** MetaData for the field in the owner with this container. */
    protected AbstractPropertyMetaData ownerFieldMetaData;

    /** Identifier for the container in JDOQL queries. */
    protected final DatastoreIdentifier thisIdentifier;

    /**
     * Constructor.
     * @param storeMgr Manager for the datastore being used
     */
    protected BaseContainerStore(StoreManager storeMgr)
    {
        this.storeMgr = (RDBMSManager)storeMgr;
        this.dba = storeMgr.getDatastoreAdapter();

        thisIdentifier = storeMgr.getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, "this");
    }

    /**
     * Accessor for the RDBMSManager.
     * @return The RDBMSManager.
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the owner mapping.
     * @return Owner mapping.
     **/
    public JavaTypeMapping getOwnerMapping()
    {
        return ownerMapping;
    }

    /**
     * Check if the mapping correspond to a non pc object or embedded field
     * @param mapping the mapping
     * @return true if the field is embedded into one column
     */
    protected boolean isEmbeddedMapping(JavaTypeMapping mapping)
    {
        return !InterfaceMapping.class.isAssignableFrom(mapping.getClass()) &&
               !OIDMapping.class.isAssignableFrom(mapping.getClass());
    }

    /**
     * Convenience method to populate the passed PreparedStatement with the value from the owner.
     * @param sm State Manager
     * @param pm Persistence Manager
     * @param ps The PreparedStatement
     * @param jdbcPosition Position in JDBC statement to populate
     * @return The next position in the JDBC statement
     */
    protected int populateOwnerInStatement(StateManager sm, PersistenceManager pm, PreparedStatement ps, int jdbcPosition)
    {
        if (((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getInsertionInputParameter().indexOf('?') < 0)
        {
            // Dont try to insert any mappings with insert parameter that isnt ? (e.g Oracle)
            return jdbcPosition;
        }
        if (ownerFieldMetaData != null)
        {
            ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(jdbcPosition, ownerMapping), sm.getObject(), 
                sm, ownerFieldMetaData.getAbsoluteFieldNumber());
        }
        else
        {
            ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(jdbcPosition, ownerMapping), sm.getObject());
        }
        return jdbcPosition + ownerMapping.getNumberOfDatastoreFields();
    }

    /**
     * Method to return the StateManager for an embedded PC object (element, key, value).
     * It creates one if the element is not currently managed.
     * @param sm State Manager of the owner
     * @param obj The embedded PC object
     * @param table Join table where the objects are stored
     * @return The state manager
     */
    protected StateManager getStateManagerForEmbeddedPCObject(StateManager sm, Object obj, JoinTable table)
    {
        PersistenceManager pm = sm.getPersistenceManager();
        StateManager objSM = pm.findStateManager((PersistenceCapable)obj);
        if (objSM == null)
        {
            objSM = StateManagerFactory.newStateManager((PersistenceCapable)obj, pm, false);

            AbstractPropertyMetaData ownerFmd = table.getOwnerFieldMetaData();
            objSM.addEmbeddedOwner(sm, ownerFmd.getAbsoluteFieldNumber());
        }
        return objSM;
    }

    /**
     * Convenience method to return if the RDBMS supports batching and the user wants batching.
     * @return If batching of statements is permissible
     */
    protected boolean allowsBatching()
    {
        if (((RDBMSAdapter)dba).supportsStatementBatching() &&
            storeMgr.getPMFContext().getPmfConfiguration().getStatementBatchLimit() != 0)
        {
            return true;
        }
        return false;
    }
}