/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson  - removed unused import
2003 Erik Bengtson  - fixed bug [832635] Imcompatiblities in casting
                     inherited classes in query
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - added getSizeStmt
2003 Andy Jefferson - updated retrieval of ColumnList's
2004 Andy Jefferson - updated to support inherited objects
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...) 
                     by advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt.
2004 Andy Jefferson - added removeAll
2004 Andy Jefferson - moved all statement construction to AbstractListStore
2005 Andy Jefferson - added dependent-element when removed from collection
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.MetaDataUtils;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.QueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.LogicSetExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.query.DiscriminatorIteratorStatement;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.query.UnionIteratorStatement;
import org.jpox.store.rdbms.table.CollectionTable;
import org.jpox.util.JPOXLogger;

/**
 * Representation of the backing store for a Normal List.
 * Uses a Join table, so we use 3 tables - owner table, join table and element
 * table.
 *
 * @version $Revision: 1.16 $
 **/
public class JoinListStore extends AbstractListStore
{
    /**
     * Constructor.
     * @param joinTable The join table
     * @param clr ClassLoader resolver
     */
    public JoinListStore(AbstractPropertyMetaData fmd, CollectionTable joinTable, ClassLoaderResolver clr)
    {
        super(joinTable.getStoreManager(), clr);

        // A List really needs a ListTable, but we need to cope with the situation
        // where a user declares a field as Collection but is instantiated as a List or a Set
        // so we just accept CollectionTable and rely on it being adequate
        this.containerTable = joinTable;
        this.ownerFieldMetaData = fmd;
        listName = "list";

        ownerMapping = joinTable.getOwnerMapping();
        elementMapping = joinTable.getElementMapping();
        orderMapping = joinTable.getOrderMapping();
        if (orderMapping == null)
        {
            // If the user declares a field as java.util.Collection JPOX will use SetTable to generate the join table
            // If they then instantiate it as a List type it will come through here, so we need to ensure the order column exists
            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.List.OrderColumnMissing", ownerFieldMetaData.getFullFieldName(), joinTable.toString()));
        }
        relationDiscriminatorMapping = joinTable.getRelationDiscriminatorMapping();
        relationDiscriminatorValue = joinTable.getRelationDiscriminatorValue();

        elementType = fmd.getCollection().getElementType();
        elementsAreEmbedded = joinTable.isEmbeddedElement();
        elementsAreSerialised = joinTable.isSerialisedElement();

        if (elementsAreSerialised)
        {
            elementInfo = null;
        }
        else
        {
            Class element_class = clr.classForName(elementType);
            if (storeMgr.getPMFContext().getTypeManager().isReferenceType(element_class))
            {
                // Collection of reference types (interfaces/Objects)
                String[] implNames = MetaDataUtils.getInstance().getImplementationNamesForReferenceField(ownerFieldMetaData, 
                    DatastoreField.ROLE_COLLECTION_ELEMENT, clr);
                elementInfo = new ElementInfo[implNames.length];
                for (int i=0;i<implNames.length;i++)
                {
                    DatastoreClass table = storeMgr.getDatastoreClass(implNames[i], clr);
                    AbstractClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(implNames[i], clr);
                    elementInfo[i] = new ElementInfo(cmd,table);
                }
            }
            else
            {
                // Collection of PC or non-PC
                // Generate the information for the possible elements
                emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class, clr);
                if (emd != null)
                {
                    if (!elementsAreEmbedded)
                    {
                        elementInfo = getElementInformationForClass();
                        /*if (elementInfo != null && elementInfo.length > 1)
                        {
                            throw new JDOUserException(LOCALISER.msg("RDBMS.SCO.Collection.ElementWithSubclassTableWithMultipleTablesNotSupported", 
                                ownerFieldMetaData.getFullFieldName()));
                        }*/
                    }
                    else
                    {
                        elementInfo = null;
                    }
                }
                else
                {
                    elementInfo = null;
                }
            }
        }

        // Generate the statements
        initialiseStatements(); 
    }

    /**
     * Remove all elements from a collection from the association owner vs
     * elements. Performs the removal in 3 steps. The first gets the indices
     * that will be removed (and the highest index present). The second step
     * removes these elements from the list. The third step updates the indices
     * of the remaining indices to fill the holes created.
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated 
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;

        // Get the current size of the list (and hence maximum index size)
        int currentListSize = size(sm);

        // Get the indices of the elements we are going to remove (highest first)
        int[] indices = getIndicesOf(sm, elements);

        // Remove the specified elements from the join table
        String removeAllStmt = getRemoveAllStmt(elements);
        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            try
            {
                PreparedStatement ps = storeMgr.getStatement(conn, removeAllStmt, false);
                try
                {
                    int jdbcPosition = 1;
                    Iterator iter = elements.iterator();
                    while (iter.hasNext())
                    {
                        Object element = iter.next();
                        jdbcPosition = populateOwnerInStatement(sm, pm, ps, jdbcPosition);
                        jdbcPosition = populateElementInStatement(pm, ps, element, jdbcPosition);
                        if (relationDiscriminatorMapping != null)
                        {
                            jdbcPosition = populateRelationDiscriminatorInStatement(pm, ps, jdbcPosition);
                        }
                    }

                    int number = storeMgr.executeStatementUpdate(removeAllStmt, ps);
                    if (number > 0)
                    {
                        modified = true;
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.RemoveRequestFailed", removeAllStmt), e);
        }

        try
        {
            // Shift the remaining indices to remove the holes in ordering
            boolean batched = allowsBatching();

            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = storeMgr.getConnection(pm, true, false);
            PreparedStatement ps = storeMgr.getStatement(conn, shiftStmt, false);
            try
            {
                for (int i=0;i<currentListSize;i++)
                {
                    // Find the number of deleted indexes above this index
                    int shift = 0;
                    boolean removed = false;
                    for (int j=0;j<indices.length;j++)
                    {
                        if (indices[j] == i)
                        {
                            removed = true;
                            break;
                        }
                        if (indices[j] < i)
                        {
                            shift++;
                        }
                    }
                    if (!removed && shift > 0)
                    {
                        processShift(sm, ps, batched, i, -1*shift);
                    }
                }
                if (batched)
                {
                    storeMgr.executeStatementBatch(shiftStmt, ps);
                }
            }
            finally
            {
                ps.close();
                storeMgr.releaseConnection(pm, conn);
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS.error(e);
            throw new JDODataStoreException(LOCALISER.msg("RDBMS.SCO.RemoveRequestFailed", removeAllStmt), e);
        }

        if (ownerFieldMetaData.getCollection().isDependentElement())
        {
            // "delete-dependent" : delete elements if the collection is marked as dependent
            // TODO What if the collection contains elements that are not in the List ? should not delete them
            sm.getPersistenceManager().deletePersistentAll(elements);
        }

        return modified;
    }

    /**
     * Generate statement for removing a collection of items from the List.
     * <PRE>
     * DELETE FROM LISTTABLE 
     * WHERE (OWNERCOL=? AND ELEMENTCOL=?) OR
     *      (OWNERCOL=? AND ELEMENTCOL=?) OR
     *      (OWNERCOL=? AND ELEMENTCOL=?)
     * </PRE>
     * @param elements Collection of elements to remove
     * @return Statement for deleting items from the List.
     **/
    protected String getRemoveAllStmt(Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return null;
        }

        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(containerTable.toString());
        stmt.append(" WHERE ");
        Iterator elementsIter = elements.iterator();
        boolean first = true;
        while (elementsIter.hasNext())
        {
            elementsIter.next(); // Not really used at the moment except to size the Collection
            if (first)
            {
                stmt.append("(");
            }
            else
            {
                stmt.append(" OR (");
            }
            for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
            {
                if (i > 0)
                {
                    stmt.append(" AND ");
                }
                stmt.append(ownerMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
            for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(elementMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                if (elementsAreSerialised)
                {
                    // Can't directly compare serialised element fields
                    stmt.append(" LIKE ");
                }
                else
                {
                    stmt.append(" = ");
                }
                stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }

            if (relationDiscriminatorMapping != null)
            {
                for (int i=0; i<relationDiscriminatorMapping.getNumberOfDatastoreFields(); i++)
                {
                    stmt.append(" AND ");
                    stmt.append(relationDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField().getIdentifier().toString());
                    stmt.append(" = ");
                    stmt.append(((RDBMSMapping)relationDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
                }
            }

            stmt.append(")");
            first = false;
        }

        return stmt.toString();
    }

    /**
     * Accessor for the iterator statement to retrieve element(s) in a range
     * from the List. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that
     * element</LI>
     * <LI>start and end have values &ge;0 and are different then get a
     * range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI> 
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL>
     * @param ownerSM The StateManager
     * @param start_index The start position in the List. 
     * @param end_index The end position in the List. 
     * @return The QueryStatement.
     **/
    protected QueryExpression getIteratorStatement(StateManager ownerSM, int start_index, int end_index)
    {
        QueryExpression stmt = null;
        final ClassLoaderResolver clr = ownerSM.getPersistenceManager().getClassLoaderResolver();
        if (elementsAreEmbedded || elementsAreSerialised)
        {
            // Element = embedded/serialised
            // Just select the join table since we're going to return the embedded/serialised columns from it
            stmt = dba.newQueryStatement(containerTable, clr);
            stmt.select(elementMapping);
        }
        else if (elementMapping instanceof ReferenceMapping)
        {
            // Element = Reference type (interface/Object)
            // Just select the join table since we're going to return the implementation id columns only
            stmt = dba.newQueryStatement(containerTable, clr);
        }
        else if (elementInfo != null)
        {
            // Element = PC
            // Join to the element table(s)
            // TODO This will mean we cant have NULLs in the List. Needs fixing
            for (int i=0;i<elementInfo.length;i++)
            {
                // TODO This will only work if all element types have a discriminator
                final int elementNo = i;
                final Class elementCls = clr.classForName(elementInfo[elementNo].getClassName());
                QueryExpression elementStmt = null;
                if (elementInfo[elementNo].getDiscriminatorStrategy() != null &&
                    elementInfo[elementNo].getDiscriminatorStrategy() != DiscriminatorStrategy.NONE)
                {
                    // The element uses a discriminator so just use that in the SELECT
                    if (storeMgr.getPMFContext().getTypeManager().isReferenceType(clr.classForName(ownerFieldMetaData.getCollection().getElementType())))
                    {                
                        // Take the metadata for the first implementation of the reference type
                        String[] clsNames = storeMgr.getMetaDataManager().getClassesImplementingInterface(ownerFieldMetaData.getCollection().getElementType(), clr);
                        Class[] cls = new Class[clsNames.length];
                        for (int j=0; j<clsNames.length; j++)
                        {
                            cls[j] = clr.classForName(clsNames[j]);
                        }
                        elementStmt = new DiscriminatorIteratorStatement(clr, cls, true, this.storeMgr, true,
                            containerTable, elementMapping, elmIdentifier).getQueryStatement();
                    }
                    else
                    {
                        elementStmt = new DiscriminatorIteratorStatement(clr, new Class[] {elementCls}, true, this.storeMgr, true,
                            containerTable, elementMapping, elmIdentifier).getQueryStatement();
                    }
                    iterateUsingDiscriminator = true;
                }
                else
                {
                    // The element has potential subclasses and no discrim so use UNIONs
                    elementStmt = new UnionIteratorStatement(clr, elementCls, true, this.storeMgr,
                        new UnionIteratorStatement.AssociationEnd()
                        {
                        public JavaTypeMapping getMapping()
                        {
                            return elementMapping;
                        }
                        public Class getType()
                        {
                            return elementCls;
                        }
                        public DatastoreContainerObject getDatastoreContainerObject()
                        {
                            return containerTable;
                        }
                        public boolean useJoin()
                        {
                            return false;
                        }
                        }, true).getQueryStatement();
                }
                
                if (stmt == null)
                {
                    stmt = elementStmt;
                }
                else
                {
                    stmt.union(elementStmt);
                }
            }
        }
        else
        {
            // Element = PC
            stmt = new UnionIteratorStatement(clr, clr.classForName(this.elementType), true, this.storeMgr,
                new UnionIteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return containerTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }
                }).getQueryStatement();
        }

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, ownerSM.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        // Apply condition on distinguisher field to filter by distinguisher (when present)
        if (relationDiscriminatorMapping != null)
        {
            ScalarExpression distinguisherExpr = relationDiscriminatorMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression distinguisherVal = relationDiscriminatorMapping.newLiteral(stmt, relationDiscriminatorValue);
            stmt.andCondition(distinguisherExpr.eq(distinguisherVal), true);
        }

        // Add restriction(s) on index values (where required)
        boolean returning_range=false;
        if (start_index == -1 && end_index == -1)
        {
            // Add no more constraints. We want all elements.
            returning_range = true;
        }
        else if (start_index == end_index && start_index >= 0)
        {
            // Apply condition on join-table owner field to filter by index
            ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(start_index));
            stmt.andCondition(indexExpr.eq(indexVal), true);
        }
        else
        {
            // Apply filter to retrieve indexes in the range
            // a). >= start_index 
            if (start_index >= 0)
            {
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(start_index));
                stmt.andCondition(indexExpr.gteq(indexVal), true);
            }
            // b). < end_index
            if (end_index >= 0)
            {
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                ScalarExpression indexVal = orderMapping.newLiteral(stmt, new Integer(end_index));
                stmt.andCondition(indexExpr.lt(indexVal), true);
            }
            returning_range = true; 
        }

        // If returning a range, add ordering by index
        if (returning_range)
        { 
            ScalarExpression exprIndex[] = new ScalarExpression[orderMapping.getNumberOfDatastoreFields()];
            boolean descendingOrder[] = new boolean[orderMapping.getNumberOfDatastoreFields()];
            exprIndex = orderMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression()).getExpressionList().toArray();
            stmt.setOrdering(exprIndex, descendingOrder);
        }
 
        return stmt;
    }

    // ---------------------------- Query Methods ------------------------------
 
    /**
     * Utility to return a new QueryStatement.
     * @param sm The state manager
     * @param candidateClass The base class
     * @return The QueryStatement
     **/
    public QueryExpression newQueryStatement(StateManager sm,
                                             String candidateClass)
    {
        if (elementsAreEmbedded || elementsAreSerialised)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("RDBMS.SCO.QueryOverCollectionImpossible"));
        }

        DatastoreIdentifier listRangeVar    = storeMgr.getIdentifierFactory().newIdentifier(IdentifierFactory.TABLE, listName);
        DatastoreIdentifier elementRangeVar = thisIdentifier;

        QueryExpression stmt = dba.newQueryStatement(containerTable, listRangeVar, sm.getPersistenceManager().getClassLoaderResolver());

        // Add join to owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        if (!storeMgr.getPMFContext().getTypeManager().isSupportedType(candidateClass))
        {
            DatastoreClass candidateTable=storeMgr.getDatastoreClass(candidateClass, sm.getPersistenceManager().getClassLoaderResolver());
            JavaTypeMapping elementTableID=candidateTable.getIDMapping();

            stmt.newTableExpression(candidateTable, elementRangeVar);

            ScalarExpression elmListExpr = elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
            ScalarExpression elmExpr = elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));

            stmt.innerJoin(elmExpr,elmListExpr,stmt.getTableExpression(elementRangeVar), true, true);
            
            stmt.select(elementTableID);
        }
        else
        {
            stmt.select(listRangeVar, elementMapping);
        }

        return stmt;
    }

    /**
     * Method used in queries when contains() has been invoked.
     * @param stmt The Query Statement 
     * @param parentStmt the parent Query Statement. If there is no parent, <code>parentStmt</code> must be equals to <code>stmt</code> 
     * @param ownerMapping the mapping for the owner. 
     * @param ownerTe Table Expression for the owner
     * @param listRangeVar   The range variable for the "List" table. 
     * @param filteredElementType The Class Type for the filtered element
     * @param elmExpr The Expression for the element
     * @param elementRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the element table.
     * @return expression to the join
     **/
    public ScalarExpression joinElementsTo(QueryExpression stmt, 
                                           QueryExpression parentStmt,
                                           JavaTypeMapping ownerMapping,
                                           LogicSetExpression ownerTe,
                                           DatastoreIdentifier listRangeVar,
                                           Class filteredElementType,
                                           ScalarExpression elmExpr,
                                           DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=stmt.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType, filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType, elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        /* TODO this comment is no longer valid, since we use sub queries
         * create a Join from the owner to the set table based on the owner ID column.
         * FROM OWNER THIS
         * INNER JOIN OWNER_ELEMENT THIS_SETTABLE 
         * ON THIS.OWNER_ID = THIS_SETTABLE.OWNER_OID */
        LogicSetExpression ownTblExpr = stmt.newTableExpression(containerTable, listRangeVar);
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        if( !parentStmt.hasCrossJoin(ownTblExpr) )
        {
            stmt.crossJoin(ownTblExpr, true);
        }
        stmt.andCondition(ownerExpr.eq(ownerSetExpr),true);

        if (storeMgr.getPMFContext().getTypeManager().isSupportedType(filteredElementType.getName()))
        {
            // Element = Non-PC(embedded)
            return elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        }
        else if (elementsAreEmbedded || elementsAreSerialised)
        {
            // Element = PC(embedded), PC(serialised)
            return elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        }
        else
        {
            // Element = PC
            // Join to element table on id column(s) of element
            DatastoreClass elementTable = storeMgr.getDatastoreClass(filteredElementType.getName(), stmt.getClassLoaderResolver());
            DatastoreClass joiningClass = (elmExpr.getLogicSetExpression() == null ? elementTable : (DatastoreClass)elmExpr.getLogicSetExpression().getMainTable());
            JavaTypeMapping elementTableID = joiningClass.getIDMapping();

            /* TODO this comment is no longer valid, since we use sub queries
             * if elementType == filteredElementType
             *    INNER JOIN ELEMENT THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.ELEMENT_ID
             * 
             * else if elementType != filteredElementType but filteredElementType is a superclass of elementType
             * 
             *    INNER JOIN FILTER_ELEMENT_TYPE THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.FILTER_ELEMENT_TYPE_ID */
            LogicSetExpression elmTblExpr = stmt.getTableExpression(elementRangeVar);
            if (elmTblExpr==null)
            {
                elmTblExpr = stmt.newTableExpression(elementTable,elementRangeVar);
            }
            ScalarExpression elmSetExpr = elementMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
            if (!parentStmt.hasCrossJoin(elmTblExpr))
            {
                stmt.crossJoin(elmTblExpr, true);
            }
            if( elmExpr.getLogicSetExpression()!= null && !elementTable.equals(elmExpr.getLogicSetExpression().getMainTable()) )
            {
                //elmExpr might express a FK in another to the ELEMENT table 
                stmt.andCondition(elmSetExpr.eq(elmExpr),true);
                return this.elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
            }
            else
            {
                //elmExpr might be a PK of the ELEMENT table 
                ScalarExpression elementExpr = elementTableID.newScalarExpression(stmt, stmt.getTableExpression(elementRangeVar));
                stmt.andCondition(elmSetExpr.eq(elementExpr),true);
                return elementExpr;
            }                    
        }
    }
}