/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - replaced TableMetadata with SQLIdentifier and java name
2004 Andy Jefferson - changed to extend TableImpl
2005 Andy Jefferson - moved discrim/version, owner, PK methods from subclasses
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import javax.jdo.JDOFatalInternalException;

import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.PrimaryKey;

/**
 * Abstract class representing a field that maps to a table in the datastore.
 * This will be something like an SCO, such as Set, Map, List where a link table is used.
 * It could potentially be used where the user wants to map some field into its own SCO table.
 *
 * @version $Revision: 1.11 $
 **/
public abstract class JoinTable extends TableImpl implements DatastoreContainerObject
{
    /** FieldMetaData for the field in the owner class. */
    protected final AbstractPropertyMetaData fmd;

    /** Mapping of an owner (OID) column back to the owner table PK */
    protected JavaTypeMapping ownerMapping;

    /** Object type of the owner. */
    protected final String ownerType;

    /**
     * Constructor.
     * @param tableName The Table SQL identifier
     * @param fmd Field meta data for the owner field
     * @param storeMgr Manager for the datastore.
     **/
    protected JoinTable(DatastoreIdentifier tableName, AbstractPropertyMetaData fmd, RDBMSManager storeMgr)
    {
        super(tableName, storeMgr);

        this.fmd = fmd;
        this.ownerType = fmd.getClassName(true);

        if (fmd.getPersistenceModifier() == FieldPersistenceModifier.NONE)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("RDBMS.Table.InvalidPersistenceModifierForField",fmd.getName()));
        }
    }

    /**
     * Accessor for the primary key for this table. Overrides the method in TableImpl
     * to add on any specification of PK name in the &lt;join&gt; metadata.
     * @return The primary key.
     */
    public PrimaryKey getPrimaryKey()
    {
        PrimaryKey pk = super.getPrimaryKey();
        if (fmd.getJoinMetaData() != null)
        {
            PrimaryKeyMetaData pkmd = fmd.getJoinMetaData().getPrimaryKeyMetaData();
            if (pkmd != null && pkmd.getName() != null)
            {
                pk.setName(pkmd.getName());
            }
        }

        return pk;
    }

    /**
     * Convenience method for whether a PK is required for the join table.
     * Makes use of the JPOX extension "primary-key" (within &lt;join&gt;) to allow turning
     * off PK generation.
     * @return Whether a PK is required
     */
    protected boolean requiresPrimaryKey()
    {
        boolean pkRequired = true;
        if (fmd.getJoinMetaData() != null && fmd.getJoinMetaData().hasExtension("primary-key") &&
            fmd.getJoinMetaData().getValueForExtension("primary-key").equalsIgnoreCase("false"))
        {
            pkRequired = false;
        }
        return pkRequired;
    }

    /**
     * Accessor for the "owner" mapping end of the relationship. This will be
     * the primary key of the owner table.
     * @return The column mapping for the owner.
     */
    public JavaTypeMapping getOwnerMapping()
    {
        assertIsInitialized();
        return ownerMapping;
    }

    /**
     * Accessor for the MetaData for the owner field for this container.
     * @return The FieldMetaData
     */
    public AbstractPropertyMetaData getOwnerFieldMetaData()
    {
        return fmd;
    }

    /**
     * Accessor for a mapping for the ID (PersistenceCapable) for this table.
     * This is not supported by join tables since they don't represent FCOs.
     * @return The (PersistenceCapable) ID mapping.
     **/
    public JavaTypeMapping getIDMapping()
    {
        throw new JDOFatalInternalException("Unsupported ID mapping in join table");
    }
}