/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - replaced TableMetadata with identifier
2004 Marco Schulze - added advance-check via TypeManager.isSupportedType(...)
2005 Andy Jefferson - only create ADPT column when necessary
2005 Andy Jefferson - enabled ability to have embedded keys/values
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.IndexMetaData;
import org.jpox.metadata.MapMetaData;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.metadata.UniqueMetaData;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.mapping.EmbeddedKeyPCMapping;
import org.jpox.store.mapping.EmbeddedValuePCMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.mapping.ReferenceMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;

/**
 * Representation of a join table for a Map relationship.
 * A Map can be configured in various ways using JPOX.
 * <UL>
 * <LI><B>Map&lt;Simple, PC&gt;</B> - the key is simple and is embedded in this
 * join table in its own column, and we have a FK to the value table</LI>
 * <LI><B>Map&lt;Simple, Simple&gt;</B> - the key and value are simple and are
 * both embedded in this join table in their own column</LI>
 * <LI><B>Map&lt;PC, PC&gt;</B> - the key and value both have their own tables
 * and we have FK's to these tables in this join table</LI>
 * </UL>
 * This join table will always have a FK back to the owning object.
 * A variant on the above configurations is where we have a PC key/value is where we can 
 * embed that in this join table too. This is specified using the <embedded> mapping
 * within <key>, or <value>.
 *
 * @version $Revision: 1.44 $
 **/
public class MapTable extends JoinTable implements SCOTable
{
    /** Mapping to the key object (if required). */
    private JavaTypeMapping keyMapping;

    /** Mapping to the value object (if required). */
    private JavaTypeMapping valueMapping;

    /**
     * when the key mapping columns can't be part of the primary key
     * by datastore limitations like BLOB types.
     * An adapter mapping is used to be a kind of "index"
     */
    private JavaTypeMapping adapterMapping;

    /** Map of field mappings when containing an embedded PC key. Keyed by the FieldMetaData of the field. */
    protected Map embeddedKeyMappingsMap;

    /** Map of field mappings when containing an embedded PC value. Keyed by the FieldMetaData of the field. */
    protected Map embeddedValueMappingsMap;

    /**
     * Constructor.
     * @param tableName Identifier name of the table
     * @param fmd MetaData for the field of the owner
     * @param storeMgr The Store Manager managing these tables.
     **/
    public MapTable(DatastoreIdentifier tableName, AbstractPropertyMetaData fmd, RDBMSManager storeMgr)
    {
        super(tableName,fmd,storeMgr);
    }

    /**
     * Method to initialise the table definition.
     * @param clr The ClassLoaderResolver
     **/
    public void initialize(ClassLoaderResolver clr)
    {
        assertIsUninitialized();

        MapMetaData mmd = fmd.getMap();
        if (mmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("RDBMS.Table.MapMetaDataError",fmd));
        }

        // Extract necessary information for assigning the primary-key columns
        PrimaryKeyMetaData pkmd = (fmd.getJoinMetaData() != null ? fmd.getJoinMetaData().getPrimaryKeyMetaData() : null);
        boolean pkColsSpecified = (pkmd != null && pkmd.getColumnMetaData() != null);
        boolean pkRequired = requiresPrimaryKey();

        // Add owner mapping
        ownerMapping = ColumnCreator.createColumnsForJoinTables(clr.classForName(ownerType), fmd,
            fmd.getJoinMetaData() == null ? null : fmd.getJoinMetaData().getColumnMetaData(),
            storeMgr, this, true, false, false, false, DatastoreField.ROLE_OWNER, clr, null, false);

        // Add key mapping
        if (isSerialisedKeyPC() || isEmbeddedKeyPC())
        {
            // Key = PC(embedded), PC(serialised)
            if (pkRequired && !pkColsSpecified &&
                !(isEmbeddedKeyPC() && fmd.getMap().getKeyClassMetaData().getIdentityType() == IdentityType.APPLICATION))
            {
                // We have a key mapping that cannot be part of the PK, so we need some other column
                ColumnMetaData colmd = null;
                if (fmd.hasExtension("adapter-column-name"))
                {
                    colmd = new ColumnMetaData(fmd, fmd.getValueForExtension("adapter-column-name"));
                }
                adapterMapping = dba.getMapping(long.class, storeMgr);
                ColumnCreator.createAdapterColumn(adapterMapping, storeMgr, clr, this, colmd, pkRequired);
            }

            // Add field(s) for key
            keyMapping = dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_MAP_KEY);

            if (pkRequired && !pkColsSpecified &&
                isEmbeddedKeyPC() && fmd.getMap().getKeyClassMetaData().getIdentityType() == IdentityType.APPLICATION)
            {
                // Key is embedded but uses app id so assign all app id PK fields to be part of PK
                int numberOfKeyColumns = keyMapping.getNumberOfDatastoreFields();
                for (int i=0;i<numberOfKeyColumns;i++)
                {
                    if (keyMapping.getDataStoreMapping(i).getJavaTypeMapping().getFieldMetaData().isPrimaryKey())
                    {
                        keyMapping.getDataStoreMapping(i).getDatastoreField().setAsPrimaryKey();
                    }
                }
            }
        }
        else
        {
            // Key = PC, Non-PC(embedded)
            adapterMapping = dba.getMapping(long.class, storeMgr);
            keyMapping = ColumnCreator.createColumnsForJoinTables(clr.classForName(mmd.getKeyType()), fmd,
                fmd.getKeyMetaData() == null ? null : fmd.getKeyMetaData().getColumnMetaData(), storeMgr, this, 
                pkRequired && !pkColsSpecified, false, fmd.getMap().isSerializedKey(), false,
                DatastoreField.ROLE_MAP_KEY, clr, adapterMapping, false);
            if (adapterMapping.getNumberOfDatastoreFields() < 1)
            {
                adapterMapping = null;
            }
        }

        // Add value mapping
        if (isSerialisedValuePC() || isEmbeddedValuePC())
        {
            // Value = PC(embedded), PC(serialised)
            // Add field(s) for value
            valueMapping = dba.getMappingManager().getMapping(this, fmd, dba, clr, MappingManager.MAPPING_MAP_VALUE);
        }
        else
        {
            // Value = PC, Non-PC(embedded)
            valueMapping = ColumnCreator.createColumnsForJoinTables(clr.classForName(mmd.getValueType()), fmd,
                fmd.getValueMetaData() == null ? null : fmd.getValueMetaData().getColumnMetaData(), storeMgr, this,
                false, true, fmd.getMap().isSerializedValue(), false,
                DatastoreField.ROLE_MAP_VALUE, clr, null, false);
        }

        if (pkRequired && pkColsSpecified)
        {
            // Apply the users PK specification
            applyUserPrimaryKeySpecification(pkmd);
        }

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Convenience method to apply the user specification of <primary-key> columns
     * @param pkmd MetaData for the primary key
     */
    protected void applyUserPrimaryKeySpecification(PrimaryKeyMetaData pkmd)
    {
        ColumnMetaData[] pkCols = pkmd.getColumnMetaData();
        for (int i=0;i<pkCols.length;i++)
        {
            String colName = pkCols[i].getName();
            boolean found = false;
            for (int j=0;j<ownerMapping.getNumberOfDatastoreFields();j++)
            {
                if (ownerMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier().getIdentifier().equals(colName))
                {
                    ownerMapping.getDataStoreMapping(j).getDatastoreField().setAsPrimaryKey();
                    found = true;
                }
            }
            
            if (!found)
            {
                for (int j=0;j<keyMapping.getNumberOfDatastoreFields();j++)
                {
                    if (keyMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier().getIdentifier().equals(colName))
                    {
                        keyMapping.getDataStoreMapping(j).getDatastoreField().setAsPrimaryKey();
                        found = true;
                    }
                }
            }
            
            if (!found)
            {
                for (int j=0;j<valueMapping.getNumberOfDatastoreFields();j++)
                {
                    if (valueMapping.getDataStoreMapping(j).getDatastoreField().getIdentifier().getIdentifier().equals(colName))
                    {
                        valueMapping.getDataStoreMapping(j).getDatastoreField().setAsPrimaryKey();
                        found = true;
                    }
                }
            }
            
            if (!found)
            {
                throw new JDOUserException(LOCALISER.msg("RDBMS.Table.PrimaryKeyColumnNotFound", toString(), colName));
            }
        }
    }

    /**
     * Accessor for whether the key is embedded into this table.
     * This can be an embedded PersistenceCapable, or an embedded simple type.
     * @return Whether the key is embedded.
     */
    public boolean isEmbeddedKey()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedKey())
        {
            // Serialized takes precedence over embedded
            return false;
        }
        if (fmd.getMap().getKeyClassMetaData() == null)
        {
            // Non-PC(embedded)
            return true;
        }
        else
        {
            // PC(embedded)
            return (fmd.getKeyMetaData() != null && fmd.getKeyMetaData().getEmbeddedMetaData() != null);
        }
    }

    /**
     * Accessor for whether the key is serialised into this table.
     * This can be an serialised PersistenceCapable, or a serialised simple type.
     * @return Whether the key is serialised.
     */
    public boolean isSerialisedKey()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedKey())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the key is a PersistenceCapable(serialised)
     * @return Whether the key is PC and is serialised
     */
    public boolean isSerialisedKeyPC()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedKey())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the key is a PersistenceCapable(embedded)
     * @return Whether the key is PC and is embedded
     */
    public boolean isEmbeddedKeyPC()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedKey())
        {
            // Serialized takes precedence over embedded
            return false;
        }
        if (fmd.getKeyMetaData() != null && fmd.getKeyMetaData().getEmbeddedMetaData() != null)
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the value is embedded into this table.
     * This can be an embedded PersistenceCapable, or an embedded simple type.
     * @return Whether the value is embedded.
     */
    public boolean isEmbeddedValue()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedValue())
        {
            // Serialized takes precedence over embedded
            return false;
        }
        if (fmd.getMap().getValueClassMetaData() == null)
        {
            return true;
        }
        else
        {
            return (fmd.getValueMetaData() != null && fmd.getValueMetaData().getEmbeddedMetaData() != null);
        }
    }

    /**
     * Accessor for whether the value is serialised into this table.
     * This can be an serialised PersistenceCapable, or a serialised simple type.
     * @return Whether the value is serialised.
     */
    public boolean isSerialisedValue()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedValue())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the value is a PersistenceCapable(serialised)
     * @return Whether the value is PC and is serialised
     */
    public boolean isSerialisedValuePC()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedValue())
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for whether the value is a PersistenceCapable(embedded)
     * @return Whether the value is PC and is embedded
     */
    public boolean isEmbeddedValuePC()
    {
        if (fmd.getMap() != null && fmd.getMap().isSerializedValue())
        {
            // Serialized takes precedence over embedded
            return false;
        }
        if (fmd.getValueMetaData() != null && fmd.getValueMetaData().getEmbeddedMetaData() != null)
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for the "key" mapping end of the relationship.
     * @return The column mapping for the element.
     **/
    public JavaTypeMapping getKeyMapping()
    {
        assertIsInitialized();

        return keyMapping;
    }

    /**
     * Accessor for the "value" mapping end of the relationship.
     * @return The column mapping for the element.
     **/
    public JavaTypeMapping getValueMapping()
    {
        assertIsInitialized();

        return valueMapping;
    }

    /**
     * Accessor for the key type for this Map.
     * @return Name of key type.
     */
    public String getKeyType()
    {
        return fmd.getMap().getKeyType();
    }

    /**
     * Accessor for the value type for this Map.
     * @return Name of value type.
     */
    public String getValueType()
    {
        return fmd.getMap().getValueType();
    }

    /**
     * Accessor for adapeter mapping.
     * The columns in this mapping are part of the primary key.
     * @return The column mapping for objects without identity or not
     *         supported as part of the primary keys
     **/
    public JavaTypeMapping getAdapterMapping()
    {
        assertIsInitialized();

        return adapterMapping;
    }

    /**
     * Accessor for the expected foreign keys for this table.
     * @param clr The ClassLoaderResolver
     * @return The expected foreign keys.
     */
    protected List getExpectedForeignKeys(ClassLoaderResolver clr)
    {
        assertIsInitialized();

        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();
        try
        {
            // FK from join table to owner table
            DatastoreClass referencedTable = storeMgr.getDatastoreClass(ownerType, clr);
            if (referencedTable != null)
            {
                // Take <foreign-key> from <join>
                ForeignKeyMetaData fkmd = null;
                if (fmd.getJoinMetaData() != null)
                {
                    fkmd = fmd.getJoinMetaData().getForeignKeyMetaData();
                }
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(ownerMapping,dba,referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }

            if (!isSerialisedValuePC())
            {
                if (isEmbeddedValuePC())
                {
                    // Add any FKs for the fields of the (embedded) value
                    EmbeddedValuePCMapping embMapping = (EmbeddedValuePCMapping)valueMapping;
                    for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
                    {
                        JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                        AbstractPropertyMetaData embFmd = embFieldMapping.getFieldMetaData();
                        if (this.storeMgr.getPMFContext().getTypeManager().isReferenceType(embFmd.getType()) && embFieldMapping instanceof ReferenceMapping)
                        {
                            // Field is a reference type, so add a FK to the table of the PC for each PC implementation
                            Collection fks = TableUtils.getForeignKeysForReferenceField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                            foreignKeys.addAll(fks);
                        }
                        else if (storeMgr.getMetaDataManager().getMetaDataForClass(embFmd.getType(), clr) != null &&
                                embFieldMapping.getNumberOfDatastoreFields() > 0 &&
                                embFieldMapping instanceof PersistenceCapableMapping)
                        {
                            // Field is for a PC class with the FK at this side, so add a FK to the table of this PC
                            ForeignKey fk = TableUtils.getForeignKeyForPCField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                            if (fk != null)
                            {
                                foreignKeys.add(fk);
                            }
                        }
                    }
                }
                else
                {
                    // FK from join table to value table (if present)
                    referencedTable = storeMgr.getDatastoreClass(fmd.getMap().getValueType(),clr);
                    if (referencedTable != null)
                    {
                        // Take <foreign-key> from <value>
                        ForeignKeyMetaData fkmd = null;
                        if (fmd.getValueMetaData() != null)
                        {
                            fkmd = fmd.getValueMetaData().getForeignKeyMetaData();
                        }
                        if (fkmd != null || autoMode)
                        {
                            ForeignKey fk = new ForeignKey(valueMapping,dba,referencedTable, true);
                            fk.setForMetaData(fkmd);
                            foreignKeys.add(fk);
                        }
                    }
                    
                }
            }

            if (!isSerialisedKeyPC())
            {
                if (isEmbeddedKeyPC())
                {
                    // Add any FKs for the fields of the (embedded) key
                    EmbeddedKeyPCMapping embMapping = (EmbeddedKeyPCMapping)keyMapping;
                    for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
                    {
                        JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                        AbstractPropertyMetaData embFmd = embFieldMapping.getFieldMetaData();
                        if (this.storeMgr.getPMFContext().getTypeManager().isReferenceType(embFmd.getType()) && embFieldMapping instanceof ReferenceMapping)
                        {
                            // Field is a reference type, so add a FK to the table of the PC for each PC implementation
                            Collection fks = TableUtils.getForeignKeysForReferenceField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                            foreignKeys.addAll(fks);
                        }
                        else if (storeMgr.getMetaDataManager().getMetaDataForClass(embFmd.getType(), clr) != null &&
                                embFieldMapping.getNumberOfDatastoreFields() > 0 &&
                                embFieldMapping instanceof PersistenceCapableMapping)
                        {
                            // Field is for a PC class with the FK at this side, so add a FK to the table of this PC
                            ForeignKey fk = TableUtils.getForeignKeyForPCField(embFieldMapping, embFmd, autoMode, storeMgr, clr);
                            if (fk != null)
                            {
                                foreignKeys.add(fk);
                            }
                        }
                    }
                }
                else
                {
                    // FK from join table to key table (if present)
                    referencedTable = storeMgr.getDatastoreClass(fmd.getMap().getKeyType(),clr);
                    if (referencedTable != null)
                    {
                        // Take <foreign-key> from <key>
                        ForeignKeyMetaData fkmd = null;
                        if (fmd.getKeyMetaData() != null)
                        {
                            fkmd = fmd.getKeyMetaData().getForeignKeyMetaData();
                        }
                        if (fkmd != null || autoMode)
                        {
                            ForeignKey fk = new ForeignKey(keyMapping,dba,referencedTable, true);
                            fk.setForMetaData(fkmd);
                            foreignKeys.add(fk);
                        }
                    }
                }
            }
        }
        catch (NoTableManagedException e)
        {
            // expected when no table exists
        }
        return foreignKeys;
    }

    /**
     * Accessor for the indices for this table. 
     * This includes both the user-defined indices (via MetaData), and the ones required by 
     * foreign keys (required by relationships).
     * @param clr The ClassLoaderResolver
     * @return The indices
     */
    protected Set getExpectedIndices(ClassLoaderResolver clr)
    {
        // The indices required by foreign keys (BaseTable)
        Set indices = super.getExpectedIndices(clr);

        if (keyMapping instanceof EmbeddedKeyPCMapping)
        {
            // Add all indices required by fields of the embedded key
            EmbeddedKeyPCMapping embMapping = (EmbeddedKeyPCMapping)keyMapping;
            for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
            {
                JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                IndexMetaData imd = embFieldMapping.getFieldMetaData().getIndexMetaData();
                if (imd != null)
                {
                    Index index = TableUtils.getIndexForField(this, imd, embFieldMapping);
                    if (index != null)
                    {
                        indices.add(index);
                    }
                }
            }
        }

        if (valueMapping instanceof EmbeddedValuePCMapping)
        {
            // Add all indices required by fields of the embedded value
            EmbeddedValuePCMapping embMapping = (EmbeddedValuePCMapping)valueMapping;
            for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
            {
                JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                IndexMetaData imd = embFieldMapping.getFieldMetaData().getIndexMetaData();
                if (imd != null)
                {
                    Index index = TableUtils.getIndexForField(this, imd, embFieldMapping);
                    if (index != null)
                    {
                        indices.add(index);
                    }
                }
            }
        }

        return indices;
    }

    /**
     * Accessor for the candidate keys for this table.
     * @return The indices
     */
    protected List getExpectedCandidateKeys()
    {
        // The indices required by foreign keys (BaseTable)
        List candidateKeys = super.getExpectedCandidateKeys();

        if (keyMapping instanceof EmbeddedKeyPCMapping)
        {
            // Add all candidate keys required by fields of the embedded key
            EmbeddedKeyPCMapping embMapping = (EmbeddedKeyPCMapping)keyMapping;
            for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
            {
                JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                UniqueMetaData umd = embFieldMapping.getFieldMetaData().getUniqueMetaData();
                if (umd != null)
                {
                    CandidateKey ck = TableUtils.getCandidateKeyForField(this, umd, embFieldMapping);
                    if (ck != null)
                    {
                        candidateKeys.add(ck);
                    }
                }
            }
        }

        if (valueMapping instanceof EmbeddedValuePCMapping)
        {
            // Add all candidate keys required by fields of the embedded value
            EmbeddedValuePCMapping embMapping = (EmbeddedValuePCMapping)valueMapping;
            for (int i=0;i<embMapping.getNumberOfJavaTypeMappings();i++)
            {
                JavaTypeMapping embFieldMapping = embMapping.getJavaTypeMapping(i);
                UniqueMetaData umd = embFieldMapping.getFieldMetaData().getUniqueMetaData();
                if (umd != null)
                {
                    CandidateKey ck = TableUtils.getCandidateKeyForField(this, umd, embFieldMapping);
                    if (ck != null)
                    {
                        candidateKeys.add(ck);
                    }
                }
            }
        }

        return candidateKeys;
    }

    /**
     * Accessor the for the mapping for a field stored in this table
     * @param fmd MetaData for the field whose mapping we want
     * @return The mapping
     */
    public JavaTypeMapping getFieldMapping(AbstractPropertyMetaData fmd)
    {
        return null;
    }
}