/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 

Contributors:
 	...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ForeignKeyAction;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.JoinMetaData;
import org.jpox.metadata.PrimaryKeyMetaData;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.PrimaryKey;

/**
 * Representation of a secondary table for a class.
 * Has a primary table that manages the main fields for that class
 * and some fields are defined (using <field table="...">) to be stored
 * in the secondary table.
 * Many of the methods in this class defer to the equivalent method
 * in the primary table.
 *
 * @version $Revision: 1.36 $
 **/
public class SecondaryTable extends AbstractClassTable implements DatastoreClass
{
    /** Primary table */
    private ClassTable primaryTable;

    /** MetaData for the join to the primary table. */
    private JoinMetaData joinMetaData;

    /**
     * Constructor. This is package level so that it SecondaryTables are only created by other tables.
     * @param tableName Table name SQL identifier
     * @param storeMgr Store Manager to manage this table
     * @param primaryTable The primary table for the class
     * @param jmd MetaData for the join to the primary table
     * @param clr The ClassLoaderResolver
     */
    SecondaryTable(DatastoreIdentifier tableName, RDBMSManager storeMgr, ClassTable primaryTable, JoinMetaData jmd, ClassLoaderResolver clr)
    {
        super(tableName, storeMgr);
        if (primaryTable == null)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("RDBMS.Table.Secondary.PrimaryTableNotSpecified", tableName.getIdentifier()));
        }
        this.primaryTable = primaryTable;
        this.joinMetaData = jmd;
    }

    /**
     * Pre initilize. For things that must be initialized right after constructor 
     * @param clr the ClassLoaderResolver
     */
    public void preInitialize(final ClassLoaderResolver clr)
    {
        assertIsUninitialized();
        
        // Initialise the PK field(s)
        if (!isPKInitialized())
        {
            initializePK(clr);
        }
    }
    
    /**
     * Post initilize. For things that must be set after all classes have been initialized before 
     * @param clr the ClassLoaderResolver
     */
    public void postInitialize(final ClassLoaderResolver clr)
    {
        assertIsInitialized();
    }
    
    /**
     * Method to initialise the table.
     * This will initialise the primary key columns for the table. Any other columns
     * are added via the addFieldMapping() method.
     * @param clr The ClassLoaderResolver
     */
    public void initialize(ClassLoaderResolver clr)
    {
        assertIsUninitialized();

        // TODO Support <primary-key> specification of the columns to use in the PK.
        // This doesnt seem too important IMHO since the secondary table should *always* be
        // joined using the PK of the primary table (and one row in the secondary table matches
        // one row in the primary table).

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Method to initialise the primary key of this table.
     * @see org.jpox.store.rdbms.table.AbstractClassTable#initializePK(ClassLoaderResolver)
     */
    protected void initializePK(ClassLoaderResolver clr)
    {
        assertIsPKUninitialized();

        // Generate the primary key using the primary table as reference
        if (primaryTable.getIdentityType() == IdentityType.APPLICATION)
        {
            addApplicationIdUsingClassTableId(joinMetaData, primaryTable, clr, primaryTable.getClassMetaData());
        }
        else if (primaryTable.getIdentityType() == IdentityType.DATASTORE)
        {
            addDatastoreId(joinMetaData, primaryTable, primaryTable.getClassMetaData());
        }

        state = TABLE_STATE_PK_INITIALIZED;
    }

    /**
     * Accessor for the primary key for this table. Overrides the method in TableImpl
     * to add on any specification of PK name in the metadata.
     * @return The primary key.
     */
    public PrimaryKey getPrimaryKey()
    {
        PrimaryKey pk = super.getPrimaryKey();
        if( joinMetaData == null )
        {
            throw new JDOUserException("A relationship to a secondary table requires a <join> specification. The secondary table is "+this.getDatastoreIdentifierFullyQualified()+" and the primary table is "+this.getPrimaryDatastoreContainerObject()+". The fields mapped to this secondary table are: "+fieldMappingsMap.keySet().toString());
        }
        PrimaryKeyMetaData pkmd = joinMetaData.getPrimaryKeyMetaData();
        if (pkmd != null && pkmd.getName() != null)
        {
            pk.setName(pkmd.getName());
        }

        return pk;
    }

    /**
     * Accessor for the primary table associated with this table
     * @return The associated primary table
     */
    public DatastoreContainerObject getPrimaryDatastoreContainerObject()
    {
        return primaryTable;
    }

    /**
     * Accessor for the identity-type.
     * Simply returns the same as the primary table
     * @return identity-type tag value
     */    
    public IdentityType getIdentityType()
    {
        return primaryTable.getIdentityType();
    }

    /**
     * Accessor for the main type represented here.
     * @return Name of the principal class represented
     */
    public String getType()
    {
        return primaryTable.getType();
    }

    /**
     * Whether this table or super table has id (primary key) attributed by the datastore
     * @return true if the id attributed by the datastore
     */
    public boolean isObjectIDDatastoreAttributed()
    {
        // We never use auto increment since we only use the same value as the primary table entry
        return false;
    }

    /**
     * Whether this table is the base in the inheritance hierarchy.
     * @return true if this table is a root table
     */
    public boolean isBaseDatastoreClass()
    {
        return primaryTable.isBaseDatastoreClass();
    }

    /**
     * Convenience accessor for the base table for this table which has the specified field.
     * @param fmd Field MetaData for this field
     * @return The base table which has the field specified
     */
    public DatastoreClass getBaseDatastoreClassWithField(AbstractPropertyMetaData fmd)
    {
        return primaryTable.getBaseDatastoreClassWithField(fmd);
    }

    /**
     * Accessor for the supertable for this table.
     * @return The supertable
     **/
    public DatastoreClass getSuperDatastoreClass()
    {
        return null;
    }

    /**
     * Accessor for whether this table manages the specified class
     * @param className Name of the class
     * @return Whether it is managed by this table
     */
    public boolean managesClass(String className)
    {
        // We dont manage classes as such (the primary table does that) so just return false
        return false;
    }

    /**
     * Utility to throw an exception if the object is not PersistenceCapable.
     * @param sm The StateManager for the object
     **/ 
    private void assertPCClass(StateManager sm)
    {
        primaryTable.assertPCClass(sm);
    }

    /**
     * Method to insert an object for this class.
     * Will insert any superclass records as well (recursive).
     * @param sm StateManager for the object to insert.
     **/
    public void insert(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        storeMgr.getInsertRequest(this, sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver()).execute(sm);
    }

    /**
     * Method to fetch an object for this class.
     * Will fetch any superclass records as well (recursive).
     * @param sm StateManager for the object to fetch.
     * @param fieldMetaData MetaData for the fields to be fetched
     */
    public void fetch(StateManager sm, AbstractPropertyMetaData[] fieldMetaData)
    {
        // Do nothing since the retrieval is done using the primary table
    }

    /**
     * Method to update an object for this class.
     * Will update any superclass records as well (recursive).
     * @param sm StateManager for the object to update.
     * @param fieldMetaData MetaData for the fields to be updated
     */
    public void update(StateManager sm, AbstractPropertyMetaData[] fieldMetaData)
    {
        assertIsInitialized();
        assertPCClass(sm);

        storeMgr.getUpdateRequest(this, fieldMetaData, sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver()).execute(sm);
    }

    /**
     * Method to delete an object for this class.
     * Will delete any superclass records as well (recursive).
     * @param sm StateManager for the object to delete.
     **/
    public void delete(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        storeMgr.getDeleteRequest(this, sm.getObject().getClass(), sm.getPersistenceManager().getClassLoaderResolver()).execute(sm);
    }

    /**
     * Locates this object in the datastore.
     * @param sm The StateManager for the object to be found 
     * @throws javax.jdo.JDOObjectNotFoundException If the instance does not exist in
     * the datastore
     */
    public void locate(StateManager sm)
    {
        // Do nothing - we have no need of locating here. The primary table is the thing to be located
    }

    /**
     * Method to delete all dependent objects.
     * Not used here since a secondary table has no such dependents
     * @param sm The State Manager
     * @param fields MetaData for the fields
     */
    public void deleteDependent(StateManager sm, AbstractPropertyMetaData[] fields)
    {
        // Do nothing since we have no such dependent behaviour
    }

    /**
     * Accessor for the expected foreign keys for this table.
     * @return The expected foreign keys.
     **/
    protected List getExpectedForeignKeys()
    {
        assertIsInitialized();

        // Auto mode allows JPOX to decide which FKs are needed as well as using
        // what is in the users MetaData.
        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();

        // Add FK back to the primary table unless requested not to
        ForeignKeyMetaData fkmd = joinMetaData != null ? joinMetaData.getForeignKeyMetaData() : null;
        if (autoMode || (fkmd != null && fkmd.getDeleteAction() != ForeignKeyAction.NONE))
        {
            ForeignKey fk = new ForeignKey(getIDMapping(), dba, primaryTable, fkmd != null && fkmd.isDeferred() ? true : false);
            if (fkmd != null && fkmd.getName() != null)
            {
                fk.setName(fkmd.getName());
            }
            foreignKeys.add(0, fk);
        }

        return foreignKeys;
    }

    /**
     * Accessor for the field Mapping.
     * Returns the mapping if it is present in this table.
     * @param fmd Field MetaData for this field
     * @return the Mapping for the field
     */
    public JavaTypeMapping getFieldMapping(AbstractPropertyMetaData fmd)
    {
        assertIsInitialized();

        // Check if we manage this field
        JavaTypeMapping m = (JavaTypeMapping)fieldMappingsMap.get(fmd);
        if (m != null)
        {
            return m;
        }

        return null;
    }

    /**
     * Accessor for the field mapping for the specified field.
     * TODO Use of this is discouraged since the fieldName is not fully qualified
     * and if a superclass-table inheritance is used we could have 2 fields of that name here.
     * @param fieldName Name of the field
     * @return The mapping for the field
     */
    public JavaTypeMapping getFieldMapping(String fieldName)
    {
        return getFieldMapping(primaryTable.getFieldMetaData(fieldName));
    }

    /**
     * Accessor for the ID mapping for this table.
     * @return The ID mapping
     */
    public JavaTypeMapping getIDMapping()
    {
        if (idMapping != null)
        {
            return idMapping;
        }

        PersistenceCapableMapping mapping = new PersistenceCapableMapping();
        mapping.initialize(getStoreManager().getDatastoreAdapter(),primaryTable.getClassMetaData().getFullClassName());
        if (getIdentityType() == IdentityType.DATASTORE)
        {
            mapping.addJavaTypeMapping(datastoreIDMapping);
        }
        else if (getIdentityType() == IdentityType.APPLICATION)
        {
            for (int i = 0; i < pkMappings.length; i++)
            {
                mapping.addJavaTypeMapping(pkMappings[i]);
            }
        }
        idMapping = mapping;
        return mapping;
    }

    /**
     * Provide the mappings to the consumer for all primary-key fields mapped to
     * this table (for application identity).
     * @param consumer Consumer for the mappings
     */
    public void providePrimaryKeyMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        ClassMetaData cmd = primaryTable.getClassMetaData();
        if (pkMappings != null)
        {
            // Application identity
            int[] primaryKeyFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
            for (int i=0;i<pkMappings.length;i++)
            {
                // Make the assumption that the pkMappings are in the same order as the absolute field numbers
                AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(primaryKeyFieldNumbers[i]);
                consumer.consumeMapping(pkMappings[i], fmd);
            }
        }
        else
        {
            // Datastore identity
            int[] primaryKeyFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
            int countPkFields = cmd.getNoOfPrimaryKeyFields();
            for (int i = 0; i < countPkFields; i++)
            {
                AbstractPropertyMetaData pkfmd = cmd.getManagedFieldAbsolute(primaryKeyFieldNumbers[i]);
                consumer.consumeMapping(getFieldMapping(pkfmd), pkfmd);
            }
        }
    }

    public void provideExternalMappings(MappingConsumer consumer, int mappingType)
    {
    }

    public Object newOID(PersistenceManager sm, PersistenceCapable pcClass)
    {
        throw new UnsupportedOperationException("Currently, there is no support for creating ids for secondary tables");
    }

    public JavaTypeMapping getExternalMapping(AbstractPropertyMetaData fmd, int mappingType)
    {
        throw new JDOFatalInternalException("N/A");
    }
}