/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2004 Andy Jefferson - changed to pass Properties rather than MetaData
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashSet;

import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.IdentifierFactory;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.poid.PoidException;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;

/**
 * Class defining a table sequence for unique identifiers
 *  
 * @version $Revision: 1.46 $
 **/ 
public class SequenceTable extends TableImpl
{
    /** Mapping for the sequence name column. */
    private JavaTypeMapping sequenceNameMapping = null;

    /** Mapping for the next value column */
    private JavaTypeMapping nextValMapping = null;

    private String insertStmt = null;
    private String incrementByStmt = null;
    private String deleteStmt = null;
    private String deleteAllStmt = null;
    private String fetchAllStmt = null;
    private String fetchStmt = null;

    /** the sequence column name */
    private String sequenceNameColumnName;

    /** the next value column name */
    private String nextValColumnName;

    /**
     * Constructor
     * @param identifier Datastore identifier for this table
     * @param storeMgr The RDBMSManager for this datastore
     * @param seqNameColName Name for the "sequence name" column
     * @param nextValColName Name for the "next value" column
     **/
    public SequenceTable(DatastoreIdentifier identifier, RDBMSManager storeMgr, String seqNameColName, String nextValColName)
    {
        super(identifier, storeMgr);

        this.sequenceNameColumnName = seqNameColName;
        this.nextValColumnName = nextValColName;
    }

    /**
     * Method to initialise the table.
     * @param clr The ClassLoaderResolver
     **/
    public void initialize(ClassLoaderResolver clr)
    {
        assertIsUninitialized();

        IdentifierFactory idFactory = storeMgr.getIdentifierFactory();

        // "SEQUENCE_NAME" column
        sequenceNameMapping = dba.getMapping(String.class, storeMgr);
        Column colSequenceName=(Column) addDatastoreField(String.class.getName(),
            idFactory.newDatastoreFieldIdentifier(sequenceNameColumnName), sequenceNameMapping, null);
        colSequenceName.setAsPrimaryKey();
        colSequenceName.getColumnMetaData().setLength(Integer.valueOf("255"));
        colSequenceName.getColumnMetaData().setJdbcType("VARCHAR");
		getStoreManager().getMappingManager().createDatastoreMapping(sequenceNameMapping, storeMgr, colSequenceName, 
            String.class.getName());

        // "NEXT_VAL" column
        nextValMapping = dba.getMapping(Long.class, storeMgr);
        DatastoreField colNextVal=addDatastoreField(Long.class.getName(),
            idFactory.newDatastoreFieldIdentifier(nextValColumnName), nextValMapping, null);
		getStoreManager().getMappingManager().createDatastoreMapping(nextValMapping, storeMgr, colNextVal, Long.class.getName());

        // Set up JDBC statements for supported operations
        insertStmt = "INSERT INTO " + toString() + " (" + colSequenceName.getIdentifier() + "," + colNextVal.getIdentifier() + 
            ") VALUES (?,?)";
        incrementByStmt = "UPDATE " + toString() + " SET " + colNextVal.getIdentifier() + "=(" + 
            colNextVal.getIdentifier() + "+?) WHERE " + colSequenceName.getIdentifier() + "=?";
        deleteStmt = "DELETE FROM " + toString();
        deleteAllStmt = "DELETE FROM " + toString();
        fetchStmt = "SELECT " + colNextVal.getIdentifier() + " FROM " + toString() + " WHERE " + 
            colSequenceName.getIdentifier() + "=?";
        if (dba.supportsLockWithSelectForUpdate())
        {
            fetchStmt += " FOR UPDATE";
        }
        fetchAllStmt = "SELECT " + colNextVal.getIdentifier() + "," + colSequenceName.getIdentifier() + 
            " FROM " + toString() + " ORDER BY " + colSequenceName.getIdentifier();

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for a mapping for the ID (PersistenceCapable) for this table.
     * @return The (PersistenceCapable) ID mapping.
     **/
    public JavaTypeMapping getIDMapping()
    {
        throw new JDOFatalInternalException("Attempt to get ID mapping of Sequence table!");
    }

    /**
     * Accessor for the sequences
     * @param conn Connection for this datastore.
     * @return The HashSet of Sequence names
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public HashSet getFetchAllSequences(Connection conn)
    throws SQLException
    {
        HashSet sequenceNames = new HashSet();

        PreparedStatement ps = null;
        try
        {
            ps = storeMgr.getStatement(conn, fetchAllStmt, true);

            ResultSet rs = storeMgr.executeStatementQuery(fetchAllStmt, ps);
            try
            {
	            while (rs.next())
	            {
	                sequenceNames.add(rs.getString(2));
	            }
            }
            finally
            {
                rs.close();
            }
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }

        return sequenceNames;
    }

    /**
     * Accessor for the nextval of a sequence
     * @param conn Connection for this datastore.
     * @param sequenceName The sequence name
     * @param incrementBy The amount to increment
     * @param tableIdentifier Identifier for the table being incremented
     * @param columnName Name of the column being incremented
     * @param initialValue Initial value (otherwise use tableIdentifier/columnName to find the initial value)
     * @return The HashSet of Sequence names
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public Long getNextVal(String sequenceName, Connection conn, int incrementBy, 
            DatastoreIdentifier tableIdentifier, String columnName, int initialValue)
    throws SQLException
    {
        PreparedStatement ps = null;
        Long nextVal = null;
        try
        {
            ps = storeMgr.getStatement(conn, fetchStmt, true);
            sequenceNameMapping.setString(null, ps, new int[]{1}, sequenceName);

            ResultSet rs = storeMgr.executeStatementQuery(fetchStmt, ps);
            try
            {
                if (!rs.next())
                {
                    // No data in the SEQUENCE_TABLE for this sequence so check the table itself for records
                    // TODO Support initialValue (when specified should be used before the tableName/columnName)
                    boolean addedSequence = false;
                    if (columnName != null && tableIdentifier != null)
                    {
                        // Table/Column specified so find the current max id value
                        PreparedStatement ps2 = null;
                        ResultSet rs2 = null;
                        try
                        {
                            String fetchInitStmt = "SELECT MAX(" + columnName + ") FROM " + tableIdentifier.getFullyQualifiedName(true);
                            ps2 = storeMgr.getStatement(conn, fetchInitStmt, true);

                            rs2 = storeMgr.executeStatementQuery(fetchInitStmt, ps2);
                            if (rs2.next())
                            {
                                long val = rs2.getLong(1);
                                addSequence(sequenceName, new Long(incrementBy + 1 + val), conn);
                                nextVal = new Long(1 + val);
                                addedSequence = true;
                            }
                        }
                        catch (Exception e)
                        {
                            // Do nothing - since if the table is empty we get this
                        }
                        finally
                        {
                            if (rs2 != null)
                            {
                                rs2.close();
                            }
                            if (ps2 != null)
                            {
                                ps2.close();
                            }
                        }
                    }
                    if (!addedSequence)
                    {
                        addSequence(sequenceName, new Long(incrementBy + 1), conn);
                        nextVal = new Long(1);
                    }
                }
                else
                {
                    nextVal = new Long(rs.getLong(1));
                    incrementSequence(sequenceName, incrementBy, conn);
                }
            }
            finally
            {
                rs.close();
            }
        }
        catch (SQLException e)
        {
            throw new PoidException(LOCALISER.msg("SequencePoidGenerator.CantObtainSequenceId", e.getMessage()));
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }

        return nextVal;
    }

    /**
     * Method to increment a sequence
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs incrementing the sequence. 
     **/
    private void incrementSequence(String sequenceName, long incrementBy, Connection conn)
    throws SQLException
    {
        PreparedStatement ps = null;
        try
        {
            ps = storeMgr.getStatement(conn, incrementByStmt, false);
            nextValMapping.setLong(null, ps, new int[] {1}, incrementBy);

            sequenceNameMapping.setString(null, ps, new int[] {2}, sequenceName);

            storeMgr.executeStatementUpdate(incrementByStmt, ps);

            // TODO : handle any warning messages
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }
    }

    /**
     * Method to insert a row in the SequenceTable
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs inserting the sequence. 
     **/
    private void addSequence(String sequenceName, Long nextVal, Connection conn)
    throws SQLException
    {
        PreparedStatement ps=null;
        try
        {
            ps = storeMgr.getStatement(conn, insertStmt, false);
            sequenceNameMapping.setString(null, ps, new int[] {1}, sequenceName);
            nextValMapping.setLong(null, ps, new int[] {2}, nextVal.longValue());

            storeMgr.executeStatementUpdate(insertStmt, ps);

            // TODO : handle any warning messages
        }
        catch (SQLException e)
        {
            e.printStackTrace();
            throw e;
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }
    }

    /**
     * Method to delete a sequence
     *
     * @param sequenceName
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting the schema. 
     **/
    public void deleteSequence(String sequenceName, Connection conn)
    throws SQLException
    {
        PreparedStatement ps=null;
        try
        {
            ps = storeMgr.getStatement(conn, deleteStmt, false);
            ps.setString(1, sequenceName);

            storeMgr.executeStatementUpdate(deleteStmt, ps);

            // TODO : handle any warning messages
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }
    }

    /**
     * Method to delete all sequences
     *
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting. 
     **/
    public void deleteAllSequences(Connection conn)
    throws SQLException
    {
        PreparedStatement ps=null;
        try
        {
            ps = storeMgr.getStatement(conn, deleteAllStmt, false);

            storeMgr.executeStatementUpdate(deleteAllStmt, ps);

            // TODO : handle any warning messages
        }
        finally
        {
            if (ps != null)
            {
                ps.close();
            }
        }
    }

    /**
     * Accessor the for the mapping for a field store in this table
     * @param fmd MetaData for the field whose mapping we want
     * @return The mapping
     */
    public JavaTypeMapping getFieldMapping(AbstractPropertyMetaData fmd)
    {
        return null;
    }
}