/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - commented
2004 Andy Jefferson - changed to extend CollectionStore
2004 Andy Jefferson - added subList
    ...
**********************************************************************/
package org.jpox.store.scostore;

import java.util.Collection;
import java.util.ListIterator;

import org.jpox.StateManager;

/**
 * Interface representation of the backing store for a List.
 * Takes the collection methods and extends them for lists.
 *
 * @version $Revision: 1.4 $
 **/
public interface ListStore extends CollectionStore
{
    /**
     * Method to add an element to the List.
     * @param ownerSM State Manager for the List. 
     * @param index Position to add the element. 
     * @param element Element to add
     **/
    void add(StateManager ownerSM, int index, Object element);

    /**
     * Method to add a collection of elements to the List.
     * @param ownerSM State Manager for the List.
     * @param index Position to add the elements.
     * @param c Collection of elements to add
     * @return Whether the elements were added ok
     **/
	boolean addAll(StateManager ownerSM, int index, Collection c);

    /**
     * Method to add a collection of elements to the List.
     * @param ownerSM State Manager for the List.
     * @param c Collection of elements to add
     * @param initialising Whether we are initialising the list with these elements
     * @return Whether the elements were added ok
     **/
    boolean addAll(StateManager ownerSM, Collection c, boolean initialising);

    /**
     * Method to remove an element from the List.
     * @param ownerSM State Manager for the List.
     * @param index Position to remove the element.
     * @return The element that was removed.
     **/
	Object remove(StateManager ownerSM, int index);

    /**
     * Method to retrieve an element from a position in the List.
     * @param ownerSM State Manager for the List.
     * @param index Position of the element.
     * @return The element at that position.
     **/
	Object get(StateManager ownerSM, int index);

    /**
     * Method to update an element at a position in the List.
     * @param ownerSM State Manager for the List.
     * @param index Position of the element.
     * @param element The element value
     * @param allowDependentField Whether to enable dependent field during this operation
     * @return The previous element at that position.
     **/
	Object set(StateManager ownerSM, int index, Object element, boolean allowDependentField);

    /**
     * Accessor for a sublist of elements between from and to indices.
     * @param ownerSM State Manager for the List.
     * @param from Start position (inclusive)
     * @param to End position (exclusive)
     * @return List of elements in this range.
     **/
	java.util.List subList(StateManager ownerSM, int from, int to);

    /**
     * Method to return the position of an element in the List.
     * @param ownerSM State Manager for the List.
     * @param element The element value
     * @return The position of the element.
     **/
	int indexOf(StateManager ownerSM, Object element);

    /**
     * Method to return the last position of an element in the List.
     * @param ownerSM State Manager for the List.
     * @param element The element value
     * @return The last position of the element.
     **/
	int lastIndexOf(StateManager ownerSM, Object element);

    /**
     * Accessor for a list iterator for the List.
     * @param ownerSM State Manager for the List. 
     * @return List iterator for the List.
     **/
	ListIterator listIterator(StateManager ownerSM);

    /**
     * Accessor for a list iterator for the List.
     * @param ownerSM State Manager for the List. 
     * @param index start position of the iterator.
     * @return List iterator for the List.
     **/ 
	ListIterator listIterator(StateManager ownerSM, int index);
}
