/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 

Contributors:
2004 Andy Jefferson - added getNewSingleFieldIdentity reflection method
    ...
**********************************************************************/
package org.jpox.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.identity.ByteIdentity;
import javax.jdo.identity.CharIdentity;
import javax.jdo.identity.IntIdentity;
import javax.jdo.identity.LongIdentity;
import javax.jdo.identity.ShortIdentity;
import javax.jdo.identity.SingleFieldIdentity;
import javax.jdo.identity.StringIdentity;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassNameConstants;
import org.jpox.metadata.AbstractClassMetaData;
import org.jpox.metadata.AbstractPropertyMetaData;
import org.jpox.metadata.IdentityType;

/**
 * Application Identity 'id' utils.
 *
 * @version $Revision: 1.14 $ 
 */
public class AIDUtils
{
    /** Localiser for messages. */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /**
     * Constructor
     */
    private AIDUtils()
    {
    }

    /**
     * Method to extract the value for a field of a PK from the identity object.
     * @param id The identity value
     * @param fieldNumber Absolute field number of the field
     * @param cmd MetaData for the class that this id pertains to
     * @return Value for the PK field
     */
    public static Object getPKFieldValueFromId(Object id, int fieldNumber, AbstractClassMetaData cmd)
    {
        if (cmd.getIdentityType() != IdentityType.APPLICATION)
        {
            // Not application identity
            return null;
        }
        // TODO Check for SingleFieldIdentity

        AbstractPropertyMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        if (!fmd.isPrimaryKey())
        {
            // Not a PK field
            return null;
        }
        String fieldName = fmd.getName();
        try
        {
            Field fld = id.getClass().getDeclaredField(fieldName);
            Object value = fld.get(id);
            return value;
        }
        catch (Exception e)
        {
            return null;
        }
    }

    /**
     * Create a new ObjectId instance and copy the key fields from the
     * persistence capable instance to the objectId instance
     * @param pc The persistence capable object
     * @param cmd The AbstractClassMetaData
     * @return The object id for the persistence capable object
     */
    public static Object getNewObjectId(PersistenceCapable pc, AbstractClassMetaData cmd)
    {
        Object userOid = pc.jdoNewObjectIdInstance();

        if (!cmd.usesSingleFieldIdentityClass())
        {
            pc.jdoCopyKeyFieldsToObjectId(userOid);
        }
        return userOid;
    }

    /**
     * Utility to create a new SingleFieldIdentity using reflection when you know the
     * type of the PersistenceCapable, and also which SingleFieldIdentity, and the value
     * of the key.
     * @param idType Type of SingleFieldIdentity
     * @param pcType Type of the PersistenceCapable
     * @param value The value for the identity (the Long, or Int, or ... etc).
     * @return The SingleFieldIdentity key
     */
    public static SingleFieldIdentity getNewSingleFieldIdentity(Class idType, Class pcType, Object value)
    {
        if (idType == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.IdTypeNull", pcType));
        }
        if (pcType == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.PcTypeNull", idType));
        }
        if (value == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueNull", idType, pcType));
        }
        if (!SingleFieldIdentity.class.isAssignableFrom(idType))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.IdTypeInvalid", idType.getName(), pcType.getName()));
        }

        SingleFieldIdentity id = null;
        Class keyType = null;
        if (idType == LongIdentity.class)
        {
            keyType = Long.class;
            if (!(value instanceof Long))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "Long"));
            }
        }
        else if (idType == IntIdentity.class)
        {
            keyType = Integer.class;
            if (!(value instanceof Integer))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "Integer"));
            }
        }
        else if (idType == StringIdentity.class)
        {
            keyType = String.class;
            if (!(value instanceof String))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "String"));
            }
        }
        else if (idType == ByteIdentity.class)
        {
            keyType = Byte.class;
            if (!(value instanceof Byte))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "Byte"));
            }
        }
        else if (idType == ShortIdentity.class)
        {
            keyType = Short.class;
            if (!(value instanceof Short))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "Short"));
            }
        }
        else if (idType == CharIdentity.class)
        {
            keyType = Character.class;
            if (!(value instanceof Character))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("SingleFieldIdentity.ValueTypeInvalid", idType.getName(), 
                    pcType.getName(), value.getClass().getName(), "Character"));
            }
        }
        else
        {
            // ObjectIdentity
            keyType = Object.class;
        }

        try
        {
            Class[] ctrArgs = new Class[] {Class.class, keyType};
            Constructor ctr = idType.getConstructor(ctrArgs);

            Object[] args = new Object[] {pcType, value};
            id = (SingleFieldIdentity)ctr.newInstance(args);
        }
        catch (Exception e)
        {
            JPOXLogger.JDO.error("Error encountered while creating SingleFieldIdentity instance of type \"" + idType.getName() + "\"");
            JPOXLogger.JDO.error(e);

            return null;
        }

        return id;
    }

    /**
     * Checks whether the argument is equals to one the single field identity class names
     * @param className the full class name
     * @return true if the argument is equals to one the single field identity class names 
     */
    public static boolean isSingleFieldIdentityClass(String className)
    {
        if (className == null || className.length() < 1)
        {
            return false;
        }

        return (className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_BYTE_IDENTITY) || 
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_CHAR_IDENTITY) || 
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_INT_IDENTITY) ||
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_LONG_IDENTITY) || 
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_OBJECT_IDENTITY) || 
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_SHORT_IDENTITY) ||
                className.equals(ClassNameConstants.JAVAX_JDO_IDENTITY_STRING_IDENTITY));
    }
}