/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.jpox.util;

/**
 * Logging framework for JPOX.
 * Allows use of Log4J, JDK1.4, or no logging.
 * Performs a similar role to Apache CommonsLogging yet doesnt need
 * an extra jar to be present in the CLASSPATH and also allows for no
 * available logger.
 *
 * @version $Revision: 1.14 $   
 **/
public abstract class JPOXLogger
{
    /** Log for JDO issues */
    public static final JPOXLogger JDO;

    /** Log for JDO REACHABILITY issues */
    public static final JPOXLogger JDO_REACHABILITY;

    /** Log for JDO Lifecycle issues */
    public static final JPOXLogger JDO_LIFECYCLE;

    /** Log for JDO Query issues */
    public static final JPOXLogger JDO_QUERY;

    /** Log for JPA issues */
    public static final JPOXLogger JPA;

    /** Log for METADATA issues */
    public static final JPOXLogger METADATA;

    /** Log for Cache issues */
    public static final JPOXLogger CACHE;

    /** Log for General issues */
    public static final JPOXLogger GENERAL;

    /** Log for Transaction issues */
    public static final JPOXLogger TRANSACTION;

    /** Log for RDBMS issues */
    public static final JPOXLogger RDBMS;

    /** Log for RDBMS DDL issues */
    public static final JPOXLogger RDBMS_DDL;

    /** Log for RDBMS SQL issues */
    public static final JPOXLogger RDBMS_SQL;

    /** Log for RDBMS Schema issues */
    public static final JPOXLogger RDBMS_SCHEMA;

    /** Log for ClassLoading issues */
    public static final JPOXLogger CLASSLOADING;

    /** Log for IDE */
    public static final JPOXLogger IDE;

    /** Log for Enhancer */
    public static final JPOXLogger ENHANCER;

    /** Log for SchemaTool */
    public static final JPOXLogger SCHEMATOOL;

    /** Log for PLUGIN issues */
    public static final JPOXLogger PLUGIN;

    /** Log for POID issues */
    public static final JPOXLogger POID;

    static
    {
        // Set the log type to be used based on what is available from this ClassLoader
        // Note that we could have registered in the PluginManager but that needs to log too
        Class loggerClass = org.jpox.util.NullLogger.class;
        try
        {
            JPOXLogger.class.getClassLoader().loadClass("org.apache.log4j.Logger");
            loggerClass = org.jpox.util.Log4JLogger.class;
        }
        catch (Exception e)
        {
            if (JavaUtils.isJRE1_4OrAbove())
            {
                loggerClass = org.jpox.util.JDK14Logger.class;
            }
        }

        // Create the Loggers for our predefined categories
        // TODO Would be nice to not have to do this, and just allocate them when the first request
        // comes in for a particular one, but that would impact on all calling methods
        JDO = getLoggerInstance(loggerClass, "JPOX.JDO");
        JDO_REACHABILITY = getLoggerInstance(loggerClass, "JPOX.JDO.Reachability");
        JDO_LIFECYCLE = getLoggerInstance(loggerClass, "JPOX.JDO.Lifecycle");
        JDO_QUERY = getLoggerInstance(loggerClass, "JPOX.JDO.Query");
        JPA = getLoggerInstance(loggerClass, "JPOX.JPA");
        METADATA = getLoggerInstance(loggerClass, "JPOX.MetaData");
        CACHE = getLoggerInstance(loggerClass, "JPOX.Cache");
        GENERAL = getLoggerInstance(loggerClass, "JPOX.General");
        TRANSACTION = getLoggerInstance(loggerClass, "JPOX.Transaction");
        RDBMS = getLoggerInstance(loggerClass, "JPOX.RDBMS");
        RDBMS_DDL = getLoggerInstance(loggerClass, "JPOX.RDBMS.DDL");
        RDBMS_SQL = getLoggerInstance(loggerClass, "JPOX.RDBMS.SQL");
        RDBMS_SCHEMA = getLoggerInstance(loggerClass, "JPOX.RDBMS.Schema");
        IDE = getLoggerInstance(loggerClass, "JPOX.IDE");
        ENHANCER = getLoggerInstance(loggerClass, "JPOX.Enhancer");
        SCHEMATOOL = getLoggerInstance(loggerClass, "JPOX.SchemaTool");
        PLUGIN = getLoggerInstance(loggerClass, "JPOX.Plugin");
        POID = getLoggerInstance(loggerClass, "JPOX.Store.Poid");
        CLASSLOADING = getLoggerInstance(loggerClass, "JPOX.ClassLoading");
    }

    /**
     * Method to create a logger instance.
     * @param loggerClass The logger class
     * @param logCategory The category (or null)
     * @return The logger
     */
    private static JPOXLogger getLoggerInstance(Class loggerClass, String logCategory)
    {
        return (JPOXLogger)ClassUtils.newInstance(loggerClass, new Class[] {String.class}, new Object[] {logCategory});
    }

    /**
     * Log a debug message.
     * @param msg The message
     */
    public abstract void debug(Object msg);

    /**
     * Log a debug message with throwable.
     * @param msg The message
     * @param thr A throwable
     */
    public abstract void debug(Object msg, Throwable thr);

    /**
     * Log an info message.
     * @param msg The message
     */
    public abstract void info(Object msg);

    /**
     * Log an info message with throwable.
     * @param msg The message
     * @param thr A throwable
     */
    public abstract void info(Object msg, Throwable thr);

    /**
     * Log a warning message.
     * @param msg The message
     */
    public abstract void warn(Object msg);

    /**
     * Log a warning message with throwable.
     * @param msg The message
     * @param thr A throwable
     */
    public abstract void warn(Object msg, Throwable thr);

    /**
     * Log an error message.
     * @param msg The message
     */
    public abstract void error(Object msg);

    /**
     * Log an error message with throwable.
     * @param msg The message
     * @param thr A throwable
     */
    public abstract void error(Object msg, Throwable thr);

    /**
     * Log a fatal message.
     * @param msg The message
     */
    public abstract void fatal(Object msg);

    /**
     * Log a fatal message with throwable.
     * @param msg The message
     * @param thr A throwable
     */
    public abstract void fatal(Object msg, Throwable thr);

    /**
     * Accessor for whether debug logging is enabled
     * @return Whether it is enabled
     */
    public abstract boolean isDebugEnabled();

    /**
     * Accessor for whether info logging is enabled
     * @return Whether it is enabled
     */
    public abstract boolean isInfoEnabled();
}